/*	%Z%$RCSfile: gips.h,v $	GIPS $Revision: 3.14 $ (MIT/CSR) $Date: 2004/11/29 18:48:48 $
**
**		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		gips.h		Creation Date:	12/29/83
**	Version:	$Revision: 3.14 $		Update:		$Date: 2004/11/29 18:48:48 $
**	Function:	Supply definitions and macros to gips sources
**	Components:	none		C #include file
*/
	/*
	**	BUFLEN	- maximum length of incore header string variable
	**	FONTLEN	- maximum length of any filename
	**	IMAGE	- settype() code to select image titles
	**	GRAPHIC - settype() code to select graphic titles
	**	ROWS	- number of rasters in display device
	**	COLS	- number of pixels per raster in display device
	**	ISCOL	- test for legal column number
	**	ISROW	- test for legal row number
	**	TOPIX	- return pixel offset in image
	**	SETGRAPH   - set pixel in graphic plane
	**	UNSETGRAPH - clear pixel in graphic plane
	**	PIXEL	- reference pixel in image plane
	**	NIMAGE	- number of bytes in image
	**	NGRAPH	- number of bytes in graphic
	*/
#ifndef	FILE
#include	<stdio.h>
#include	<unistd.h>
#endif

#define	BUFLEN			4096
#define	FONTLEN			100
#define	IMAGE			0
#define	GRAPHIC			1

#ifndef MAX
#define MAX(x, y)     ( ((x) > (y)) ? (x) : (y) )
#endif
#ifndef MIN
#define MIN(x, y)     ( ((x) < (y)) ? (x) : (y) )
#endif

#ifdef comtal
#define	ROWS			512
#define	COLS			512
#define	ISCOL(x)		(!((x)&~0777))
#define	ISROW(x)		(!((x)&~0777))
#define	TOPIX(x,y)		(((y)<<9)|(x))
#define	SETGRAPH(g,x,y)		((g)[((y)<<6)|((x)>>3)]|=(0200>>((x)&7)))
#define	UNSETGRAPH(g,x,y)	((g)[((y)<<6)|((x)>>3)]&= ~(0200>>((x)&7)))
#define	PIXEL(i,x,y)		((i)[((y)<<9)|(x)])
#define	GRAPHADDR(x)		((x)^1)
#else
#define	ROWS			1024
#define	COLS			1024
#define	ISCOL(x)		(!((x)&~01777))
#define	ISROW(x)		(!((x)&~01777))
#define	TOPIX(x,y)		(((y)<<10)|(x))
#define	SETGRAPH(g,x,y)		((g)[((y)<<7)|((x)>>3)]|=(0200>>((x)&7)))
#define	UNSETGRAPH(g,x,y)	((g)[((y)<<7)|((x)>>3)]&= ~(0200>>((x)&7)))
#define	PIXEL(i,x,y)		((i)[((y)<<10)|(x)])
#define	GRAPHADDR(x)		(x)
#endif

#define	NIMAGE			(ROWS * COLS)
#define	NGRAPH			(NIMAGE / 8)
	/*
	**	default: header description, header directory, font directory
	*/
#define	GIPSPROFILE		"cprofile"
#define	GINPROFILE		"ginprofile"
#define	GIPSPROFDIR		gipsprofdir()
#define	GIPSFONTDIR		"/usr/lib/vfont/"
#ifdef sun
#define	GIPSAUXFONTDIR		"/usr/lib/fonts/fixedwidthfonts/"
#endif
#define	GIPSCMAPDIR		gipscmapdir()
#define	GIPSBOXDIR		gipsboxdir()
#define	GIPSVERSIONS		"gipsversions"
	/*
	**	gips() returned values indicating pixel out of range
	*/
#define	GIPSLOW		(-1.e30)
#define	GIPSHIGH	(1.e30)
	/*
	**	command argument definition structure
	*/
#define	flag	char

typedef	struct	flagentry {
	char *	flag_name;
	char *	flag_oper;
	flag *	flag_ptr;
} flagentry;
	/*
	**	display device: pseudocolor and function table elements
	*/
typedef	short	PSEUDO;
typedef	short	FUNCTION;
typedef	char	gimage;
	/*
	**	GIPS functions
	*/
extern	FILE	*command_open();	/* open stream to display controller */
extern	FILE	*fopens();		/* return FILE pointer to string */
extern	FILE	*ihread();		/* read general image header */
extern	FILE	*ihwrit();		/* write general image header */
extern	FILE	*ihmerg();		/* merge general image header */
extern	void	gipsint();		/* specify GIPS error handler */
extern	char	*gipsboxdir();		/* returns GIPS boxcar directory */
extern	char	*gipscmapdir();		/* returns GIPS colormap directory */
extern	char	*gipsprofdir();		/* returns GIPS profile directory */
extern	void	saveargs();		/* save argv[] list */
extern	void	putargs();		/* write argv[] as history entry */
extern	void	gconv();		/* convert between arithmetic fields */
extern	void	clear();		/* set block to zero */
extern	void	move();			/* copy block */
extern	void	poly();			/* apply polynomial transform */
extern	char	*alloc();		/* call calloc(), test for error */
extern	char	*setfont();		/* redefine fonts for comlabel() */
extern	char	*setauxfont();		/* redefine font for comlabel() */
extern	char	setink();		/* redefine dn for comlabel() */
extern	gimage	*gopen();		/* open image formatted by ihgfmt */
extern	char	*gcell();		/* point to image pixel */
extern	char	*ginterp();		/* interpolate image pixel value */
extern	char	*g_getname();		/* return name of current mapping */
extern	int	gipserror();		/* error handler */

#define	SIGTYPE	void
extern	SIGTYPE	exit();
extern	char	*getenv();

#ifdef	sun
#define	clear(c,l)	(bzero((c),(l)))
#define	move(f,t,l)	(bcopy((f),(t),(l)))
#endif

extern	char	_ihlen_[];	/* table of lengths of dtype items */
extern	char	_ihcvt_[];	/* table of CVT types */
extern	char	_ihdim_[];	/* table of lengths .v. CVT type */
extern	char	*gipserrs[];	/* table of error message strings */
extern	int	gipserrno;	/* pointer to error message */
extern	int	errno;		/* UNIX error number */
	/*
	**	gips() and egips() modes
	*/
#define		GIPS_X_VAL	0	/* real x-value -> x-pixel address */
#define		GIPS_Y_VAL	1	/* real y-value -> y-pixel address */
#define		GIPS_D_VAL	2	/* array element -> 8-bit pixel value */
#define		GIPS_XY_IVAL	3	/* real xy-values -> xy-pixel addrs */
#define		GIPS_D_DVAL	4	/* real d-value -> 8-bit pixel value */
#define		GIPS_XY_FVAL	5	/* real xy-values -> float xy-addrs */
#define		GIPS_R_VAL	6	/* array element -> real d-value */
#define		EGIPS_X_VAL	0	/* x-pixel address -> real x-value */
#define		EGIPS_Y_VAL	1	/* y-pixel address -> real y-value */
#define		EGIPS_D_VAL	2	/* 8-bit pixel value -> real d-value */
#define		EGIPS_XY_VAL	3	/* xy-pixel addrs -> real xy-values */
	/*
	**	return codes from ihget(), ihput(), ihdel(), gips(),
	**			  egips(), gopen(), gcell(), ginterp().
	**	also returned in external variable: gipserrno
	*/
#define		GIPSBADARG	1
#define		GIPSNOHEAD	2
#define		GIPSBADOPEN	3
#define		GIPSBADREAD	4
#define		GIPSBADDATA	5
#define		GIPSNOVAR	6
#define		GIPSNOVAL	7
#define		GIPSNOEOF	8
#define		GIPSBADVAR	9
#define		GIPSBADVAL	10
#define		GIPSBADTYPE	11
#define		GIPSTRUNC	12
#define		GIPSBADLIM	13
#define		GIPSUNSCALE	14
#define		GIPSNOSTR	15
#define		GIPSNOCORE	16
#define		GIPSBADWRITE	17
#define		GIPSBADCONV	18
#define		GIPSBADMAP	19
	/*
	**	return length of pixel and conversion code of dtype=c
	*/
#define	ihdtype(c)	((_ihlen_+1)[c] & 0177)
#define	ihdcvt(c)	((_ihcvt_+1)[c] & 0177)
#define	ihdlen(c)	((_ihdim_+1)[c] & 0177)

#ifdef _IOSTRG
#define	isastring(f) ((f) != (FILE *)0 && ((f)->_flag & _IOSTRG))
#else
extern	char	*ihdstr;	/* pointer to stream string table */
extern	int	ihdlstr;	/* length of string table */
#define	isastring(f) ((f) != (FILE *)0 && ihdstr && ihdstr[fileno(f)] != 0)
#endif

#ifdef	SVR4
#define	WAITYPE	int *
#else
#define	WAITYPE	union wait *
#endif

	/*
	**	gconv() conversion codes
	*/
#define		CVT_S		0
#define		CVT_B		1
#define		CVT_H		2
#define		CVT_U		3
#define		CVT_I		4
#define		CVT_L		5
#define		CVT_F		6
#define		CVT_D		7
#define		CVT_C		8
	/*
	**	error message generators:
	**	gipserr()  - after failing gips routine
	**	gipserr1() - user-supplied error condition
	**	gipserr2() - user-supplied error message
	*/
#define gipserr()	(error(gipserrs[gipserrno],gipserrs[0]))
#define gipserr1(x)	(error(gipserrs[x],gipserrs[0]))
#define gipserr2(x,y)	(error(gipserrs[x],(y)))

#ifndef IGED
#define	error(x,y)	((void)gipserror((x),(y)))
#endif

#ifdef	GRID
#define	GIPS_GRID	"Gips control grid: xnum=%d ynum=%d rasters=%d"
#define	GIPS_MAPPING	"name=%s type=%d origin=%lf,%lf radius=%lf euler=%lf,%lf,%lf lat=%lf,%lf xyrot=%d"
#define	GIPS_MAPPING_P	"name=%s type=%d origin=%.12g,%.12g radius=%.12g euler=%.12g,%.12g,%.12g lat=%.12g,%.12g xyrot=%d"
#ifndef	M_PI
#define	M_PI	3.14159265358979323846
#define	M_PI_2	1.57079632679489661923
#define	M_PI_4	0.78539816339744830962
#endif
#define	DEG_RAD	(M_PI/180)
#define	RAD_DEG	(180/M_PI)
	/*
	 *	master projection table
	 */
typedef	struct	g_trans {
	int	(*proc)();	/* projection subroutine */
	char	*name;		/* projection name */
	char	*desc;		/* short description */
	char	*init;		/* */
} G_TRANS;
	/*
	 *	projection parameters
	 */
typedef	struct	g_parm	{
	flag    deg_sw;         /* =1 if transformation in degrees */
	flag    euler_sw;       /* =1 if Euler rotation specified */
	int	xyrot;		/* x/y axis rotation */
	double	xorg, yorg;	/* output map origin */
	double	e[3][3];	/* Euler rotation matrix */
	double	radius, rlim;	/* radius and limit */
	double	fact0, fact1;	/* scaling factors */
	double	fact2;		/* scaling factor */
	double	cphi, sphi;	/* cos,sin longitude */
	double	rcphi, tphi2;	/* */
	double	csphi, sphi0;	/* */
	double	csphi0, rcphi0;	/* */
	double	rctphi;		/* */
	G_TRANS	*tproj;		/* pointer to projection table */
} G_PARM;

extern	G_PARM	*g_proj;	/* master projection structure */
#endif

#ifdef lint
	/*LINTLIBRARY*/
	/*VARARGS*/
int	gipsflags(n,a) int n; char **a; { return gipsflags(n,a); }
	/*VARARGS*/
int	syscall() { syscall(); }
/*
	/*VARARGS
	execle(f,a) char *f,*a; { execle(f,a);}
	/*VARARGS
	execve(s,v,e) char *s, **v, **e; { execve(s,v,e);}
char  *	strcat(a,b) char *a,*b; { return strcat(a,b); }
char  *	strcpy(a,b) char *a,*b; { return strcpy(a,b); }
*/

char	_ihlen_[];
char	_ihcvt_[];
char	_ihdim_[];
char	chrtab[][16];
char	*gipserrs[];
char	*argsav[];
char	*ihdstr;
int	ihdlstr;
int	gipserrno;
int	arginit;

char	*alloc(n,m) { return alloc(n,m); }
#ifndef sun
void	clear(c,n) char *c; { clear(c,n); }
void	move(c,d,n) char *c,*d; { move(c,d,n); }
#endif
	/*ARGSUSED*/
void	poly(c,d,n,p,m) double *c,*d,*p; { poly(c,d,n,p,m); }
char	*setfont(c) char *c; { return setfont(c); }
char	*setauxfont(c,n) char *c; { return setauxfont(c,n); }
char	setink(c) char c; { return setink(c); }
int	setorigin(n) { return setorigin(n); }
int	settype(n) { return settype(n); }
int	comlabel(n,m,l,i,j,c) char *c; { return comlabel(n,m,l,i,j,c); }
int	commarker(n,m,l,i,j,c) char *c; { return commarker(n,m,l,i,j,c); }
int	memlabel(c,m,l,i,j,d) char *c,*d; { return memlabel(c,m,l,i,j,d); }
int	memmarker(c,m,l,i,j,d) char *c,*d; { return memmarker(c,m,l,i,j,d); }
int	comheight(c) char *c; { return comheight(c); }
int	comwidth(c) char *c; { return comwidth(c); }
char	*gipsboxdir() { return gipsboxdir(); }
char	*gipscmapdir() { return gipscmapdir(); }
char	*gipsprofdir() { return gipsprofdir(); }
void	gipsint(p) int (*p)(); { gipsint(p) ; }
gimage	*gopen(f,n) FILE *f; { return gopen(f,n); }
int	gclose(g) gimage *g; { return gclose(g); }
char	*gcell(n,m,g) gimage *g; { return gcell(n,m,g); }
char	*ginterp(x,y,g) float x, y; gimage *g; { return ginterp(x,y,g); }
void	g_listproj(f) FILE *f; { g_listproj(f); }
int	g_setup() { return g_setup(); }
char	*g_getname(n) { return g_getname(n); }
int	g_setproc(n,p) int (*p)(); { return g_setproc(n,p); }
int	g_map(d,e,f) double d,e,f[2]; { return g_map(d,e,f); }
int	g_unmap(x,y,f) double f[2]; { return g_unmap(x,y,f); }
int	gips(d,i,n,m) double *d; int *i; { return gips(d,i,n,m); }
int	egips(i,d,n,m) int *i; double *d; { return egips(i,d,n,m); }
int	gips_(d,i,n,m) double *d; int *i,*n,*m; { return gips_(d,i,n,m); }
int	egips_(i,d,n,m) int *i,*n,*m; double *d; { return egips_(i,d,n,m); }
int	ihbin(c,f,n) char *c; FILE *f; { return ihbin(c,f,n); }
int	ihbout(c,f,n) char *c; FILE *f; { return ihbout(c,f,n); }
int	ihbcl(f) FILE *f; { return ihbcl(f); }
int	ihbin_(c,f,n) char *c; FILE **f; { return ihbin_(c,f,n); }
int	ihbout_(c,f,n) char *c; FILE **f; { return ihbout_(c,f,n); }
int	ihbcl_(f) FILE **f; { return ihbcl_(f); }
FILE	*ihread(c,d) char *c,*d; { return ihread(c,d); }
FILE	*ihmerg(c) char *c; { return ihmerg(c); }
int	ihdel(c) char *c; { return ihdel(c); }
int	ihget(c,d,m) char *c,*d; { return ihget(c,d,m); }
int	ihput(c,d) char *c,*d; { return ihput(c,d); }
int	ihtype(c,d,m) char *c,*d; { return ihtype(c,d,m); }
FILE	*ihwrit(c,n) char *c; { return ihwrit(c,n); }
void	putargs() { putargs(); }
void	saveargs(c) char **c; { saveargs(c); }
	telltale(n) int n; { telltale(n); }
void	gconv(c,d,l,n,m) char *c,*d; { gconv(c,d,l,n,m); }
int	icvti(n,m,l) int *n,*m; { return icvti(n,m,l); }
int	icvtf(f,g,n) float *f,*g; { return icvtf(f,g,n); }
int	icvtd(d,e,n) double *d,*e; { return icvtd(d,e,n); }
int	icvti_(n,m,l) int *n,*m,*l; { return icvti_(n,m,l); }
int	icvtf_(f,g,n) float *f,*g; int *n; { return icvtf_(f,g,n); }
int	icvtd_(d,e,n) double *d,*e; int *n; { return icvtd_(d,e,n); }
FILE	*command_open() { return command_open(); }
int	image_open(n,m) { return image_open(n,m); }
int	graphic_open(n,m) { return graphic_open(n,m); }
int	pseudo_open(n,m) { return pseudo_open(n,m); }
int	function_open(n,m) { return function_open(n,m); }
int	target_get(n,m) int *m; { return target_get(n,m); }
int	target_put(n,m) int *m; { return target_put(n,m); }
int	target_roam(n) { return target_roam(n); }
int	target_dump(n) { return target_dump(n); }
int	color_get(n) { return color_get(n); }
int	color_put(n,m) { return color_put(n,m); }
#ifdef IGED
int	error(c,d) char *c,*d; { return error(c,d); }
#else
int	gipserror(c,d) char *c,*d; { return gipserror(c,d); }
#endif
#endif
