#ifndef F77
#ifndef lint
static char *sccsid = "%Z%$RCSfile: gimandel.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/11/13 23:53:36 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		gimandel.c	Creation Date:	12/09/86
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/11/13 23:53:36 $
**	Function:	Generate General Image of portion of Mandelbrot set
**	Components:	main()		GIPS/UNIX command
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"

static	flag	Sflag	= 0;
static	int	xnum	= 256;
static	int	ynum	= 256;
static	char	*xlims	= NULL;
static	char	*ylims	= NULL;
static	char	*dlims	= NULL;

static flagentry flagtable[] = {
	"x",	"%d num",	(flag *) &xnum,
	"y",	"%d num",	(flag *) &ynum,
	"D",	"-> dmin,dmax",	(flag *) &dlims,
	"S",	"1",		(flag *) &Sflag,
	"X",	"-> xmin,xmax",	(flag *) &xlims,
	"Y",	"-> ymin,ymax",	(flag *) &ylims,
	NULL
};

main(argc,argv)

	int	argc;
	char	**argv;
{
	register FILE	*fp;
	register int	x, y, l, n;
	char	buf[256];
	int	m;
	double	fx, fy, dx, dy, dfact, r, i, rr, ri, ii;
	double	xmin	= -2.5;
	double	ymin	= 1.5;
	double	xmax	= 1.0;
	double	ymax	= -1.5;
	double	dmin	= 0;
	double	dmax	= 1000;

	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 1)
		error("bad argument: %s",*++argv);
	if (Sflag)
		telltale(0);

	if ((fp = ihread(GINPROFILE,GIPSPROFILE)) == NULL)
		gipserr();
	else
		(void)(ihbcl(fp));

	if (dlims != NULL)
		(void)sscanf(dlims, "%lf,%lf", &dmin, &dmax);
	if (xlims != NULL)
		(void)sscanf(xlims, "%lf,%lf", &xmin, &xmax);
	if (ylims != NULL)
		(void)sscanf(ylims, "%lf,%lf", &ymin, &ymax);

	(void)ihput("title", "The Mandelbrot Fractal");

	(void)ihput("xtitle", "Re z");
	(void)(ihput("xnum",(char *)&xnum));
	(void)(ihput("xmin",(char *)&xmin));
	(void)(ihput("xmax",(char *)&xmax));
	(void)ihput("xfont", "I.6");
	(void)sprintf(buf, "place=tbco font=screen.r.7 by=%d", xnum/4);
	(void)ihput("xaxis", buf);

	(void)ihput("ytitle", "Im z");
	(void)(ihput("ynum",(char *)&ynum));
	(void)(ihput("ymin",(char *)&ymin));
	(void)(ihput("ymax",(char *)&ymax));
	(void)ihput("yfont", "I.6r");
	(void)sprintf(buf, "place=tbco font=screen.r.7 by=%d", ynum/4);
	(void)ihput("yaxis", buf);

	(void)ihput("dtitle", "z->z(z+1)");
	(void)(ihput("dscale","n"));
	(void)(ihput("dtype","b"));
	(void)(ihput("dmin",(char *)&dmin));
	(void)(ihput("dmax",(char *)&dmax));
	(void)(ihput("dzero",(char *)&dmin));
	m = (int)dmax;
	n = (int)dmin;
	dfact = (dmax - dmin)/255;
	(void)(ihput("dfact",(char *)&dfact));
	dfact = 1/dfact;

	putargs();
	if ((fp = ihwrit("-",1)) == NULL)
		gipserr();
	(void)(signal(SIGPIPE,exit));

	dx = (xmax-xmin)/(double)xnum;
	dy = (ymin-ymax)/(double)ynum;

	for (fy = ymin, y = 0; y < ynum; y++, fy -= dy)
		for (fx = xmin, x = 0; x < xnum; x++, fx += dx)
			putc((int)((mandel_(&fx, &fy, &m) - n)*dfact), fp);
	(void)(ihbcl(fp));
	if (Sflag)
		telltale(1);
	exit(0);
}

#ifndef NF77
static
mandel_(x, y, m)
	register double	*x, *y;
	register *m;
{
	double	r = *x, i = *y, rr, ii, ri;
	register l;

	for (l = 0; l < *m; l++) {
		rr = r*r;
		ii = i*i;
		if ((rr + ii) >= 4)
			break;
		ri = r*i;
		r = rr - ii + *x;
		i = ri + ri + *y;
	}
	return l;
}
#endif
#else
	integer function mandel(x, y, m)
	real*8 x, y, r, i, ri
	r = x
	i = y
	do 10 l=0,m
		mandel = l
		if ((r*r+i*i).gt.4) return
		ri = r*i
		r = r*r-i*i+x
10		i = ri+ri+y
	return
	end
#endif
