#ifndef lint
static char *sccsid = "%Z%$RCSfile: gipsmongo.c,v $	GIPS $Revision: 2.9 $ (MIT/CSR) $Date: 1998/11/13 23:28:10 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		gipsmongo.c	Creation Date:	11/26/85
**	Version:	$Revision: 2.9 $		Update:		$Date: 1998/11/13 23:28:10 $
**	Components:	main()		GIPS/UNIX command
**	Subsystem:	mongo_()	(c) J. Tonry, 1983
**
**	Function:	Execute mongo with data from General Image
*/

#include	<stdio.h>
#include	<ctype.h>
#include	<signal.h>
#include	"gips.h"
#include	"mongo.h"
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/un.h>

#define		comp(a,b,c,d)	(d-c >= strlen(b) && !strncmp(a,c,d-c))
#define		min(x,y)	(((x)<(y))?(x):(y))
#define		max(x,y)	(((x)>(y))?(x):(y))
#define		print(s)	fprintf(stderr ,(s))
	/*
	**	flag fields filled by gipsflags()
	*/
static	char	*ifile;			/* input mongo command file */
static	char	*ofile;			/* output device */
static	flag	Sflag;			/* generate statistics */
	/*
	**	table used by gipsflags()
	*/
static	flagentry	flagtable[]	= {
	"S",	"1",		&Sflag,
	"I",	"-> file",	(flag *) &ifile,
	"o",	"-> device",	(flag *) &ofile,
	NULL,	"[file ...]",	NULL
};
	/*
	**	Gips header variables
	*/
static	char	title[BUFSIZ];	/* Gips image title */
static	char	dtitle[BUFSIZ];	/* Gips data title */
static	int	xnum, ynum;	/* number of columns and rows*/
static	char	dtype;		/* pixel data type code */
static	char	dscale;		/* pixel data scaling code */
static	float	fmin, fmax;	/* real data value limits */
	/*
	**	Socket stuff
	*/
static	struct	sockaddr_un su	= { AF_UNIX, "/tmp/gips_socket" };
static	int	ns	= -1;	/* input accept socket */
	/*
	**	Miscellaneous
	*/
static	float	*data;		/* pixel array */
static	int	cont_nlev;	/* number of contour levels */
static	float	*cont_level;	/* contour levels */
extern	int	errno;		/* system error code */
extern	char	*index();	/* C library */
static	float	fact_3d[4] =	/* 3d scaling factors */
		{20, 45, 1, 0};
static	read_array(), process_command(), usage(), mongoparms();
static	draw_contour(), draw_3dplot(), draw_halftone(), gips_title();
	/*
	**	Remove the socket name
	*/
static
bailout()
{
	if (gipserrno == 0 && errno != 0)
		perror(su.sun_path);
	unlink(su.sun_path);
	exit(1);
}

main(argc, argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	int	idev;		/* mongo output device type */
	int	s;		/* input listen socket */
	int	rc;		/* return code from process_command() */
	char	buf[5][32];	/* scratch for "color" command */

	/*
	**	save and parse callers argument list
	*/
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (Sflag)
		telltale(0);
	/*
	**	determine output device
	*/
	if ((idev = get_mongo_device(&ofile)) == 0)
		error("unsupported output device: %s", ofile);
	/*
	**	Initialize the output device
	*/
	f_init();
	device(idev);
	if (idev > 0) {
		tsetup();
		erase();
	} else {
		vtswapxy();
		setlweight(2);
		fileinit();
	}
	/*
	**	Initialize the color command
	*/
	(void)sprintf(buf[0], "%-32.32s", "define color");
	(void)sprintf(buf[3], "set \\9 %s + &1", COLOR_CMAGIC);
	(void)sprintf(buf[1], "%-32.32s", buf[3]);
	(void)sprintf(buf[2], "%-32.32s", "end");
	(void)sprintf(buf[3], "%-32.32s", "end");
	mongo(4, buf, 0, 0, NULL, 32);
	/*
	**	If no input file specified, listen on a socket
	*/
	if (argc == 1) {
		unlink(su.sun_path);
		if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1 ||
			bind(s, &su, strlen(su.sun_path)+2) || listen(s, 1))
			bailout();
		gipsint(bailout);
		ns = 0;
	}
	/*
	**	Loop over input files or socket connections
	*/
	while (ns != -1 || *++argv) {
		if (ns != -1) {
			fprintf(stderr,
				"GipsMongo%s waiting for connection...\n",
				index("2.2", '%') ? "" : " 2.2");
			fflush(stdout);
			fclose(stdin);
			errno = 0;
			while ((ns = accept(s, NULL, NULL)) == -1)
				if (errno != 4)
					bailout();
			print("connected\n");
			fflush(stdout);
			fdopen(ns, "r");
			*argv = "-";
		}
		read_array(*argv);
		while ((rc = process_command()) > 0)
			if (idev == 3)
				ttflush_();
		if (rc < 0)
			break;
	}
	if (idev < 0)
		fileplot();
	else
		tidle();
	f_exit();
	if (ns != -1)
		unlink(su.sun_path);
	if (Sflag)
		telltale(1);
	exit(0);
}
	/*
	**	Read the array into allocated storage
	*/
static
read_array(name)
	register char	*name;		/* input file name or "-" */
{
	register FILE	*fp;		/* input stream pointer */
	register int	size;		/* size of input pixel */
	register char	*inbuf;		/* raster input buffer */
	register float	*f;		/* scratch */
	register int	x;		/* scratch */
	register int	y;		/* scratch */
	double		dzero, dfact;	/* scratch */
	float		fzero, ffact;	/* base and scale for dscale == 'n' */
	double		xmin, xmax;	/* x-axis realworld limits */
	double		ymin, ymax;	/* y-axis realworld limits */
	/*
	**	Read the header
	*/
	if ((fp = ihread(name, GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		name = (ns == -1) ? "'stdin'" : "'socket'";
	/*
	**	Access header variables
	*/
	if (ihget("xnum", &xnum, 0) || ihget("ynum", &ynum, 0) ||
	    ihget("dtype", &dtype, 1) || (size = ihdtype(dtype)) == 0)
		gipserr();
	if (ihget("dscale", &dscale, 1) == 0 && dscale == 'n')
		if (ihget("dzero", &dzero, 0) || ihget("dfact", &dfact, 0))
			gipserr();
		else
			fzero = dzero, ffact = dfact;
	if (fp != stdin || ihget("title", title, sizeof(title)))
		(void) strcpy(title, name);
	if (ihget("xmin", &xmin, 0)) xmin = 0;
	if (ihget("xmax", &xmax, 0)) xmax = xnum;
	if (ihget("ymax", &ymin, 0)) ymin = 0;
	if (ihget("ymin", &ymax, 0)) ymax = ynum;
	setlim(xmin, ymin, xmax, ymax);
	(void) ihget("dtitle", dtitle, sizeof(dtitle)-1);
	/*
	**	Allocate buffers
	*/
	if (data != NULL)
		free((char *)data);
	inbuf = alloc(size, xnum);
	data = (float *)alloc(sizeof(float), xnum*ynum);
	if (inbuf == NULL || data == NULL)
		gipserr();
	/*
	**	Read the pixels, converting them to floats
	*/
	fmin = GIPSHIGH;
	fmax = GIPSLOW;
	fprintf(stderr, "Reading \"%s\" (%dx%d pixels)\n", title, xnum, ynum);
	for (y = ynum; --y >= 0; ) {
		if (fread(inbuf, size, xnum, fp) != xnum)
			gipserr2(GIPSBADREAD, name);
		gconv(f = &data[y*xnum], inbuf, xnum, CVT_F, ihdcvt(dtype));
		for (x = 0; x < xnum; x++, f++) {
			if (dscale == 'n')
				*f = fzero + ffact * *f;
			if (fmin > *f) fmin = *f;
			if (fmax < *f) fmax = *f;
		}
	}
	/*
	**	Close the input file and reaccess the standard input
	*/
	free(inbuf);
	ihbcl(fp);
	if (fp == stdin)
		freopen("/dev/tty", "r", stdin);
}
	/*
	**	Prompt, read and execute command
	*/
static
process_command()
{
	char	buf[BUFSIZ];
	char	*getenv();
	register char	*c, *d, *e;

	if (ifile) {
		(void)sprintf(c = buf, "input %s\n", ifile);
		ifile = NULL;
	} else {
		print(" > ");
		if ((c = fgets(buf, sizeof(buf), stdin)) == NULL) {
			return -1;
		}
	}
	while (isspace(*c)) c++;
	if (d = index(c, '\n')) *d = '\0';
	switch (*c) {
	case '!':
		while (isspace(*++c)) ;
		if (! *c)
			c = (d = getenv("SHELL")) ? d : "/bin/csh";
		system(c);
	case '\0':
	case '#':
		return 1;
	default:
		break;
	}
	for (d = c; *d && ! isspace(*d); d++)
		if (isupper(*d))
			*d = tolower(*d);
	if (comp("contour", "cont", c, d))
		draw_contour(d);
	else if (comp("plot3d", "plo", c, d))
		draw_3dplot(d);
	else if (comp("halftone", "half", c, d))
		draw_halftone(d);
	else if (comp("header", "hea", c, d))
		(void) ihwrit("-", 1);
	else if (comp("inform", "inf", c, d))
		usage();
	else if (comp("show", "sh", c, d))
		mongoparms();
	else if (comp("quit", "quit", c, d))
		return -1;
	else if (comp("next", "next", c, d))
		return 0;
	else if (comp("mongo", "m", c, d))
		mongo(0, " ", xnum, ynum, data, 1);
	else {
		if (comp("xlabel", "xla", c, d))
			gips_title("xtitle", d, buf+sizeof(buf));
		else if (comp("ylabel", "yla", c, d))
			gips_title("ytitle", d, buf+sizeof(buf));
		for (d = c+strlen(c), e = max(d,c+80); d < e; *d++ = ' ');
		e = d + (d-c);
		*d++ = 'e'; *d++ = 'n'; *d++ = 'd';
		while (d < e) *d++ = ' ';
		mongo(2, c, xnum, ynum, data, (d-c)/2);
	}
	return 1;
}
	/*
	**	Write information
	*/
static
usage()
{
	print("	FUNCTION	COMMAND <options>\n");
	print("	Plot contours:	contour <#levels> <lev1> <lev2> ... <levN>\n");
	print("	List header:	header\n");
	print("	Display info: 	inform\n");
	print("	Enter Mongo:  	mongo\n");
	print("	Halftone plot: 	halftone <low> <high> <contrast>\n");
	print("	Next file:	next\n");
	print("	Plot 3d figure:	plot3d <zenith> <azimuth> <offset> <origin>\n");
	print("	Quit GipsMongo:	quit\n");
	print("	Display params:	show\n");
	print("	Plot X-label:	xlabel <label>\n");
	print("	Plot Y-label:	ylabel <label>\n");
	print("	Execute shell:	! <command>\n");
}
	/*
	**	Write mongo parameter block
	*/
static
mongoparms()
{
	register i, n;

	fprintf(stderr,
		"Input GIPS file:  \"%s\"  %dx%d elements\n",
		title, xnum, ynum);

	fprintf(stderr,
		"Realworld cursor: (%g,%g)\tx: %g to %g\ty: %g to %g\n",
		mongopar_.c_cord[0], mongopar_.c_cord[1],
		mongopar_.u_cord[0], mongopar_.u_cord[1],
		mongopar_.u_cord[2], mongopar_.u_cord[3]);

  	fprintf(stderr,
		"Plot position:    (%g,%g)\tx: %g to %g\ty: %g to %g\n",
		mongopar_.p_cord[0], mongopar_.p_cord[1],
		mongopar_.g_cord[0], mongopar_.g_cord[1],
		mongopar_.g_cord[2], mongopar_.g_cord[3]);

	fprintf(stderr,
		"Physical limits:  \tx: %d to %d\ty: %d to %d\n",
		mongopar_.l_cord[0], mongopar_.l_cord[1],
		mongopar_.l_cord[2], mongopar_.l_cord[3]);

	print("User variables:  ");
	for (i = n = 0; i < 10; i++)
		if (mongopar_.uservar[i] != 0)
			fprintf(stderr, " %d:%g", i ,mongopar_.uservar[i]), n++;
	print(n ? "\n" : " nil\n");

	fprintf(stderr,
		"Pixel range:      %f to %f (%s)\n",
		fmin, fmax, dtitle);

	fprintf(stderr,
		"Contour levels:   (%d)", cont_nlev);
	for (i = 0; i < cont_nlev; i++)
		fprintf(stderr, " %d:%g", i, cont_level[i]);

	fprintf(stderr,
	"\n3-d plot factors: Zenith %g, Azimuth %g, Scale %g, Offset %g\n",
		fact_3d[0], fact_3d[1], fact_3d[2], fact_3d[3]);

	fprintf(stderr,
	"Cht %g, Cwd %g, Cxdef %g, Cydef %g, Pxdef %g, Pydef %g Coff %g\n",
		mongopar_.c_dim[0], mongopar_.c_dim[1], mongopar_.c_dim[2],
		mongopar_.c_dim[3], mongopar_.c_dim[4], mongopar_.c_dim[5],
		mongopar_.c_dim[6]);

	fprintf(stderr,
	"Expand %g, Angle %g, Ltype %d, Lweight %d, Term %d, Numdev %d\n",
		mongopar_.expand, mongopar_.angle, mongopar_.ltype,
		mongopar_.lweight, mongopar_.termout, mongopar_.numdev);
}
	/*
	**	Draw a contour plot
	*/
static
draw_contour(c)
	register char	*c;		/* string pointer */
{
	register int	i, rc;		/* scratch */

	while (isspace(*c)) c++;
	if (sscanf(c, "%d", &cont_nlev) != 1 || cont_nlev < 1) {
		if (*c) {
			fprintf("Bad number of contours: %s\n", c);
			return;
		}
		cont_nlev = 8;
	}
	if (cont_level != NULL)
		free((char *)cont_level);
	cont_level = (float *)alloc(sizeof(float), cont_nlev);
	for (i = 0; i < cont_nlev; i++) {
		if (! *c)
			cont_level[i] = fmin +
				(i*(fmax-fmin))/(cont_nlev-1);
		while (*c && ! isspace(*c)) c++;
		while (isspace(*c)) c++;
		if (*c && *c != '*' && sscanf(c, "%f", &cont_level[i]) != 1) {
			fprintf("Bad contour level: %s\n", c);
			return;
		}
	}
	gipscont(data, xnum, ynum, cont_level, cont_nlev);
}
	/*
	**	Draw a 3-dimensional plot
	*/
static
draw_3dplot(c)
	register char	*c;		/* string pointer */
{
	float		fact, fave;	/* data scaling factor and average */
	register float	*f;		/* scratch */
	register float	*zbuf;		/* scratch */
	register int	i;		/* scratch */
	/*
	**	scan arguments
	*/
	for (i = 0; i < 4; i++) {
		while (*c && ! isspace(*c)) c++;
		while (isspace(*c)) c++;
		if (! *c)
			break;
		if (*c != '*' && sscanf(c, "%f", &fact_3d[i]) != 1) {
			fprintf("Bad 3-d plot factor: %s\n", c);
			return;
		}
	}
	/*
	**	rescale data in place, plot, then scale it back again
	*/
	if (fmin == fmax) {
		print("Can't plot: dmin = dmax\n");
		return;
	}
	zbuf = (float *)alloc(sizeof(float), 4*min(xnum, ynum));
	if (zbuf == NULL) {
		print("Insufficient memory to rescale image\n");
		return;
	}
	for (fact = 1/(fmax - fmin), f = data, i = xnum*ynum; --i >= 0; f++) {
		*f = fact * (*f - fmin);
		fave += *f;
	}	
	plt3d(data, xnum, ynum, zbuf, fact_3d[0], fact_3d[1], fact_3d[2],
		fact_3d[3] - fave/(xnum*ynum));
	for (fact = fmax - fmin; --f >= data; )
		*f = fmin + fact * *f;
	free(zbuf);
}
	/*
	**	Draw a halftone plot
	*/
static
draw_halftone(c)
	register char	*c;		/* string pointer */
{
	float	dmin, dmax;		/* min,max data to plot */
	float	contrast = 1.e7;	/* contrast stretch */
	register int	i;		/* scratch */
	/*
	**	scan arguments
	*/
	switch (sscanf(c, " %f %f %f", &dmin, &dmax, &contrast)) {
	case 1:
		contrast = dmin;
	case EOF:
	case 0:
		dmin = fmin;
		dmax = fmax;
	default:
		break;
	}
	halftone(data, xnum, ynum, dmin, dmax, contrast);
}
	/*
	**	If not on command line, insert label title into buffer
	*/
static
gips_title(name, c, d)
	register char	*name, *c, *d;
{
	while (isspace(*c)) c++;
	if (! *c) {
		*c++ = ' ', *c = 0;
		(void)ihget(name, c, d-c);
	}
}
