#ifndef lint
static char *sccsid = "%Z%$RCSfile: axis.c,v $	GIPS $Revision: 3.7 $ (MIT/CSR) $Date: 1998/05/13 01:10:06 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		axis.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.7 $		Update:		$Date: 1998/05/13 01:10:06 $
**	Function:	Draw axes
**	Components:	convert_scale_value()	generate scale string
**			convert_to_real()	get variable in real & image
**			draw_axis()		top-level axis routine
**			scan_axis_place()	decode place= argument
**			scan_axis_string()	decode axis= variable
**			truncate_gips_value()	convert realworld to ASCII
**			update_gips_axis()	make updates to axis=
*/
#include	"gipstool.h"
#define		AXINIT		"place=cbi font=R.6"
#define		MAXNUM		20
	/*
	**	Fields filled directly from header
	*/
static	char	axis[BUFLEN];	/* axis descriptor fields */
static	char	scale;		/* axis scale type */
/*static	char	font[FONTLEN];*/	/* title or axis font name */
	/*
	**	Subfields of "xaxis" or "yaxis"
	*/
static	int	mark;		/* number of axis scale marks */
static	int	tic;		/* number of axis tics per scale */
static	int	line;		/* line drawing mask */
static	int	ticlen;		/* length of axis tic mark */
static	char	form[FONTLEN];	/* axis scale format */
static	char	place[FONTLEN];	/* position code(s) for titles */
static	char	ascale[2];	/* scale type */
static	char	sfrom[MAXNUM];	/* low scale value */
static	char	sto[MAXNUM];	/* high scale value */
static	char	sby[MAXNUM];	/* scale increment */
static	char	sloc[MAXNUM];	/* location on other axis */
	/*
	**	Definition of "axis" subfields
	*/
#define		A_FOUND		01
#define		A_INT		02
#define		A_REAL		04
static	struct	scan	{
	char	*name[2];	/* Variable names */	
	char	*addr;		/* address of value field */
	int	flg;		/* see above */
} axscan[] = {
	" mark	= %d",	" mark=%d",	(char *) &mark,		A_INT,
	" tic	= %d",	" tic=%d",	(char *) &tic,		A_INT,
	" place	= %s",	" place=%s",	(char *) place,		0,
	" font	= %99s"," font=%s",	(char *) font,		0,
	" form	= %99s"," form=%s",	(char *) form,		0,
	" scale = %1s",	" scale=%.1s",	(char *) ascale,	0,
	" line	= %d",	" line=%d",	(char *) &line,		A_INT,
	" ticlen = %d",	" ticlen=%d",	(char *) &ticlen,	A_INT,
	" loc	= %19s"," loc=%s",	(char *) sloc,		A_REAL,
	" from	= %19s"," from=%s",	(char *) sfrom,		A_REAL,
	" to	= %19s"," to=%s",	(char *) sto,		A_REAL,
	" by	= %19s"," by=%s",	(char *) sby,		A_REAL,
	NULL
};
	/*
	 *	axis place code lookup table
	 */
#define		P_BOTTOM	01
#define		P_MIDDLE	02
#define		P_TOP		04
#define		P_LEFT		010
#define		P_CENTER	020
#define		P_RIGHT		040
#define		P_INSIDE	0100
#define		P_OUTSIDE	0200
#define		P_ENCLOSED	0400
#define		P_UNENCLOSED	01000
static	struct	ptab {
	char	ch;		/* place code character */
	int	mask;		/* flag field */
	int	exclude;	/* conflicts */
}  ptable[] = {
	'b',	P_BOTTOM,	0,
	'm',	P_MIDDLE,	0,
	't',	P_TOP,		0,
	'l',	P_LEFT,		P_LEFT | P_RIGHT | P_CENTER,
	'r',	P_RIGHT,	P_LEFT | P_RIGHT | P_CENTER,
	'c',	P_CENTER,	P_LEFT | P_RIGHT | P_CENTER,
	'i',	P_INSIDE,	P_OUTSIDE,
	'o',	P_OUTSIDE,	P_INSIDE,
	'e',	P_ENCLOSED,	P_UNENCLOSED,
	'u',	P_UNENCLOSED,	P_ENCLOSED,
	'\0'
};

static	origins[2][2][2][3] =	/* setorigin() codes */
	{ 6, 7, 8, 1, 2, 3,
	  3, 5, 8, 1, 4, 6,
	  1, 4, 6, 3, 5, 8,
	  8, 7, 6, 3, 2, 1 };

static	int	egips_mode[3]	= { EGIPS_X_VAL, EGIPS_Y_VAL, EGIPS_D_VAL };

static	double	*xyval[2];		/* scale coordinate values */
static	int	nxyval[2];		/* previous coord dimensions */
extern	int	arginit;		/* gips variable update flag */
extern	double	atof(), fabs();		/* C library */
static	plot_axis(), scan_axis_string(), scan_axis_place();
static	convert_scale_value(), convert_to_real(), update_gips_axis();
	/*
	**	Draw non-separable axes
	*/
draw_map_axes(buffer)
	char	*buffer;
{
	double	d, x[2], mins[2], maxs[2], incs[2], steps[2], wrap;
	register xy, x1, y1;
	register struct scan *sc;	/* pointer to header desc table */

	steps[0] = steps[1] = 1;

	if (g_unmap(0, ynum/2, mins) || g_unmap(xnum, ynum/2, x)) {
		if (ihget("xmin", (char *)&mins[0], 1))
			mins[0] = 0;
		if (ihget("xmax", (char *)&maxs[0], 1))
			maxs[0] = xnum;
	} else
		maxs[0] = x[0];

	if (g_unmap(xnum/2, 0, x) || g_unmap(xnum/2, ynum, incs)) {
		if (ihget("ymin", (char *)&mins[1], 1))
			mins[1] = 0;
		if (ihget("ymax", (char *)&maxs[1], 1))
			maxs[1] = ynum;
	} else
		mins[1] = x[1], maxs[1] = incs[1];

	if (d = maxs[0] - mins[0])
		mins[0] = d * (int)(10 * fabs(mins[0] / d));

	if (d = maxs[1] - mins[1])
		mins[1] = d * (int)(10 * fabs(mins[1] / d));

	if (ihget("xscale", &scale, 1) || scale != 'w')
		wrap = 0;
	else if (ihget("xwrap", &wrap, 0))
		wrap = 360;

	for (xy = 0; xy < 2; xy++) {
		if (ihget(xy ? "yaxis"  : "xaxis",  axis, BUFLEN))
			*axis = 0;
		for (sc = axscan; sc->name[0] != NULL; (sc++)->flg &= ~A_FOUND)
			;
		*sfrom = *sto = *sby = *sloc = *form = 0;
		mark = tic = line = 0;
		ticlen = 4;
		if (scan_axis_string(AXINIT, axscan) ||
	    		scan_axis_string(axis, axscan) ||
	    		scan_axis_string(buffer, axscan))
			return;
		if (mark == 0)
			mark = 30;
		if (*sfrom)
			mins[xy] = atof(sfrom);
		if (*sto)
			maxs[xy] = atof(sto);
		if (*sby)
			incs[xy] = atof(sby);
		else
			incs[xy] = (maxs[xy] - mins[xy])/(1 + mark);
		incs[xy] = fabs(incs[xy]);
		if (mins[xy] > maxs[xy]) {
			d = maxs[xy];
			maxs[xy] = mins[xy];
			mins[xy] = d;
		}
	}

	for (xy = 0; xy < 2; xy++)
		for (x[xy] = mins[xy]; x[xy] <= maxs[xy]; x[xy] += incs[xy])
			plot_axis(1-xy, x, mins, maxs, steps, wrap);

	/*
	**	Clear borders, draw box
	*/

	x1 = image_rect.r_left + image_rect.r_width;
	y1 = image_rect.r_top + image_rect.r_height;

	draw_background();

	if (image_rect.r_top > 0)
		draw_line(image_rect.r_left, image_rect.r_top,
			x1, image_rect.r_top);
	if (image_rect.r_left > 0)
		draw_line(image_rect.r_left, image_rect.r_top,
			image_rect.r_left, y1);
	if (x1 < sw_rect.r_width)
		draw_line(x1, image_rect.r_top, x1, y1);
	if (y1 < sw_rect.r_height)
		draw_line(image_rect.r_left, y1, x1, y1);
}

static
plot_axis(yx, x, mins, maxs, steps, wrap)
	register yx;
	double	x[2], mins[2], maxs[2], steps[2], wrap;
{
	int	ix0 = 0, iy0 = 0, ix1, iy1;
	double	d, dd[2], dx;
	register sw1, sw2 = 0;

	for (d = mins[yx]; d <= maxs[yx]; d += steps[yx], sw2 = sw1) {
		sw1 = 0;
		x[yx] = d;
		if (g_map(x[0], x[1], dd))
			continue;
		ix1 = dd[0] + 0.5;
		iy1 = dd[1] + 0.5;
		if (g_unmap(ix1, iy1, dd) || fabs(x[1] - dd[1]) > 1)
			continue;
		dx = x[0] - dd[0];
		if (fabs(dx) > 1 && fabs(dx - wrap) > 1 && fabs(dx + wrap) > 1)
			continue;
		if (sw2 && 4*abs(ix0-ix1)<image_rect.r_width &&
				4*abs(iy0-iy1)<image_rect.r_height)
			draw_line(X_IMAGE_TO_RECT(ix0), Y_IMAGE_TO_RECT(iy0),
			X_IMAGE_TO_RECT(ix1), Y_IMAGE_TO_RECT(iy1));
		sw1 = 1;
		ix0 = ix1;
		iy0 = iy1;
	}
}
	/*
	**	All coordinates are expressed in image_pw values
	**	ip[] contains the three or fewer co-axis coordinates
	**	jp[] contains -1, 1 for scales and titles below, above the image
	*/
draw_axis(mode, buffer)
	int	mode;			/* 0 or 1 */
	char	*buffer;		/* command string */
{
	int	comode;			/* = 1, 0 for x, y axis */
	int	ipl;			/* axis place flags */
	int	ip[3];			/* axis coordinates */
	int	jp[3];			/* scale directions +- */
	int	x[2];			/* scale coordinates */
	int	y[2];			/* scale limits */
	int	org[2];			/* origin of displayed image */
	int	lim[2];			/* size of displayed image */
	int	ifrom;			/* pixel starting value */
	int	ito;			/* pixel ending value */
	int	iby;			/* pixel stepping value */
	int	fby;			/* flag =0, -1 if "by" is fixed/float */
	int	nmark;			/* number of scale markers */
	int	fontrot;		/* rotated font flag */
	int	height;			/* height of scale markers */
	char	buf[BUFLEN];		/* scratch */
	char	*isfont;		/* NULL if no scale font */
	double	dloc;			/* scratch */
	double	from;			/* float starting value */
	double	to;			/* float ending value */
	double	by;			/* float step */
	double	truncate_gips_value();	/* truncate a string value */
	register norg;			/* origin code 0-2 */
	register struct scan *sc;	/* pointer to header desc table */
	register i;			/* scratch */
	/*
	 *	Get values for header items
	 */
	if (ihget(mode ? "yaxis"  : "xaxis",  axis, BUFLEN)) *axis = 0;
	if (ihget(mode ? "yscale" : "xscale", &scale, 1)) scale = 's';
	/*
	 *	Parse the "axis"  and interactive subfields
	 */
	for (sc = axscan; sc->name[0] != NULL; (sc++)->flg &= ~A_FOUND)
		;
	sfrom[0] = sto[0] = sby[0] = sloc[0] = form[0] = 0;
	jp[0] = jp[1] = jp[2] = height = mark = tic = line = 0;
	ticlen = 4;
	if (scan_axis_string(AXINIT, axscan) ||
	    scan_axis_string(axis, axscan) ||
	    scan_axis_string(buffer, axscan))
		return 0;
	/*
	**	Collect display coordinates
	*/
	comode = 1 - mode;
	org[0] = image_rect.r_left;
	org[1] = image_rect.r_top;
	lim[0] = org[0] + image_rect.r_width;
	lim[1] = org[1] + image_rect.r_height;
	/*
	 *	Scan the 'place' string and set flags in 'ipl'
	 */
	if ((ipl = scan_axis_place(place, ptable)) == -1) {
		display_error_message("Invalid axis place field: \"%s\"",place);
		return 0;
	}
	if ((ipl & P_MIDDLE) && *sloc &&
			convert_to_real(sloc, &dloc, ip, 0, comode))
		jp[0] = (ip[0] <= 0) ? 1 : -1;

	if ((mode == 0 && (ipl & P_TOP)) || (mode == 1 && (ipl & P_BOTTOM))) {
		ip[1] = org[comode];
		if (ipl & P_INSIDE)
			jp[1] = 1;
		else if (ipl & P_OUTSIDE)
			jp[1] = -1;
		else
			jp[1] = (ip[1] <= 0) ? 1 : -1;
	}
	if ((mode == 0 && (ipl & P_BOTTOM)) || (mode == 1 && (ipl & P_TOP))) {
		ip[2] = lim[comode];
		if (ipl & P_INSIDE)
			jp[2] = -1;
		else if (ipl & P_OUTSIDE)
			jp[2] = 1;
		else
			jp[2] = (ip[2] >= COLS) ? -1 : 1;
	}
	/*
	 *	Update the "axis" header item
	 */
	if (update_gips_axis(axis, axscan, mode)) {
		GIPS_ERROR;
		return 1;
	}
	/*
	**	Form correct font name for scale numerics
	*/
	if (fontrot = (font[i = strlen(font)-1] == 'r'))
		font[i] = '\0';
	if ((isfont = setfont(font)) == NULL) 
		display_system_message(font);
	/*
	**	Fill the pixel coordinate table
	*/
	i = 1 + (mode ? ynum : xnum);
	if (nxyval[mode] < i || xyval[mode] == NULL) {
		if (xyval[mode] != NULL)
			free((char *)xyval[mode]);
		if (!(xyval[mode] = (double *)alloc(i, sizeof(double)))) {
			no_more_memory();;
			return 0;
		} else
			nxyval[mode] = i;
	} else
		i = 0;

	if (i > 0 || arginit)
		if (egips((int *)NULL, xyval[mode], i, egips_mode[mode])) {
			GIPS_ERROR;
			return 1;
		}
	/*
	**	Loop over scale marks
	*/
	if (convert_to_real(sfrom, &from, &ifrom, org[mode], mode) > 0) {
		display_error_message("Bad axis field: \"from=%s\"", sfrom);
		return 0;
	}
	if (convert_to_real(sto, &to, &ito, lim[mode], mode) > 0) { 
		display_error_message("Bad axis field: \"to=%s\"", sto);
		return 0;
	}
	nmark = (mark > 1) ? mark : 10;
	fby = convert_to_real(sby, &by, &iby,
		(lim[mode]-org[mode])/(nmark-1), mode);
	if (fby > 0) {
		display_error_message("Bad axis field: \"by=%s\"", sby);
		return 0;
	}
	if (fby == -1) {
		iby = 1;
		if ((by < 0 && from < to) || (by > 0 && from > to))
			by = -by;
	}
	if ((iby < 0 && ifrom < ito) || (iby > 0 && ifrom > ito))
		iby = -iby;

	for (x[mode] = ifrom; --nmark >= 0; ) {
		for (i = 0; i < 3; i++) {
			int	j, k;
			double	xmin, xmax;	/* wrap x-coordinate limits */
			double	d;

			if (jp[i] == 0)
				continue;
			x[comode] = ip[i];
			y[comode] = x[comode] + jp[i]*(ticlen+2);
			y[mode] = x[mode];
			draw_line(x[0], x[1], y[0], y[1]);
			x[comode] = y[comode] + jp[i]*2;
			d = (fby == -1) ? from :
				truncate_gips_value(x[mode], mode);
			if (d <= GIPSLOW || d >= GIPSHIGH || ! isfont)
				continue;
			convert_scale_value(d, buf, *ascale, form);
			write_text(x[0], x[1], buf,
				origins[fontrot][mode][jp[i]==1][1],
				fontrot ? 90 : 0);
			j = (fontrot == mode) ?
				comheight(buf) : comwidth(buf);
			if (j > height)
				height = j;
			if (mode != 0 || scale != 'w')
				continue;
			if (x[0] == org[0])
				j = lim[0];
			else if (x[0] == lim[0])
				j = org[0];
			else
				continue;
			k = RECT_TO_IMAGE(x[0], 0);
			if (egips(&k, &xmin, 1, egips_mode[0]))
				continue;
			k = RECT_TO_IMAGE(j, 0);
			if (egips(&k, &xmax, 1, egips_mode[0]))
				continue;
			if (xmin != xmax)
				continue;
			if (ihget("xmin", (char *)&xmin, 1))
				xmin = 0;
			if (ihget("xmax", (char *)&xmax, 1))
				xmax = xnum;
			if (d == xmin)
				d = xmax;
			else if (d == xmax)
				d = xmin;
			else
				continue;
			draw_line(j, ip[i], j, y[comode]);
			convert_scale_value(d, buf, *ascale, form, mode);
			write_text(j, x[1], buf,
				origins[fontrot][mode][jp[i]==1][1],
				fontrot ? 90 : 0);
		}
		/*
		**	If requested, draw axis lines across image
		*/
		if (line) {
			y[mode] = x[mode];
			x[comode] = org[comode];
			y[comode] = lim[comode];
			draw_line(x[0], x[1], y[0], y[1]);
		}
		if (fby == -1) {
			from += by;
			if (gips(&from, &x[mode], 1,
					mode ? GIPS_Y_VAL : GIPS_X_VAL))
				break;
			x[mode] = IMAGE_TO_RECT(x[mode], mode);
		} else
			x[mode] += iby;
		if ((iby >= 0 && x[mode] > ito+1) ||
		    (iby < 0 && x[mode] < ito-1))
			break;
	}
	/*
	**	If font and title exist, insert the title.
	*/
	if (ipl & P_LEFT)
		norg = 0;
	else if (ipl & P_CENTER)
		norg = 1;
	else if (ipl & P_RIGHT)
		norg = 2;
	else
		norg = -1;

	if (norg >= 0 &&
			ihget(mode ? "ytitle" : "xtitle", buf, BUFLEN-1) == 0 &&
			ihget(mode ? "yfont"  : "xfont", font, FONTLEN) == 0 &&
			(i = strlen(font)) > 0) {
		if (fontrot = (font[i-1] == 'r'))
			font[i-1] = '\0';
		if (setfont(font) == NULL) {
			display_system_message(font);
			return 0;
		}
		x[mode] = org[mode] + (norg * (lim[mode]-org[mode]))/2;
		for (i = 1; i < 3; i++) {
			if (jp[i] == 0)
				continue;
			x[comode] = ip[i] + jp[i]*(height+ticlen+4);
			write_text(x[0], x[1], buf,
				origins[fontrot][mode][jp[i]==1][norg],
				fontrot ? 90 : 0);
		}
	}
	/*
	**	If requested, draw box around image
	*/
	x[mode] = org[mode];
	y[mode] = lim[mode];
	ip[1] = org[comode];
	ip[2] = lim[comode];
	for (i = 0; i < 3; i++) {
		if ((i == 0 && jp[0] == 0) ||
		    (i > 0 && (ipl & P_UNENCLOSED)))
			continue;
		x[comode] = y[comode] = ip[i];
		draw_line(x[0], x[1], y[0], y[1]);
	}
	return 1;
}
	/*
	**	Use fscanf() to decode the "axis" header item and the
	**	subcommand parameter strings. The simplest way
	**	(which may be UNIX release dependent) is to create
	**	a dummy FILE structure, just the way sscanf() does.
	*/
static
scan_axis_string(s, ax)

	register char	*s;		/* "axis" string */
	struct	 scan	*ax;		/* lookup table */
{
	char		buf[32];	/* scratch */
	int		off;		/* length of keyword */
	register struct scan *sc;	/* pointer to scan table */

	for (sc = ax; sc->name[0] != NULL; sc++) {
		sprintf(buf, "%s%%n", sc->name[0]);
		if (sscanf(s, buf, sc->addr, &off) == 1) {
			sc->flg |= A_FOUND;
			sc = ax - 1;
			s += off;
		}
	}
	/*
	**	Flush the remainder of the string 
	*/
	while (isspace(*s))
		s++;
	if (! *s)
		return 0;
	display_error_message("Bad axis string field: \"%s\"", s);
	return 1;
}
	/*
	**	Scan the "place" string, setting "ipl" flags
	**	and detecting illegal combinations
	*/
static
scan_axis_place(s, pt)

	register char		*s;	/* place string */
	register struct ptab	*pt;	/* base of scan table */
{
	register int		flg;	/* returned flag field */
	register struct ptab	*p;	/* pointer to scan table */
	
	for (flg = 0; *s; flg |= p->mask, s++)
		for (p = pt; p->ch; )
			if (p->ch == *s)
				if (flg & p->exclude)
					return -1;
				else	
					break;
			else if ((++p)->ch == '\0')
					return -1;
	return flg;
}
	/*
	**	Convert the axis value "d" to ASCII according to scale code
	*/
static
convert_scale_value(d, s, sc, form, mode)

	double	d;		/* number to convert */
	register char *s;	/* output buffer */
	register char sc;	/* axis scale code */
	register char *form;	/* user-specified format list */
	int	mode;		/* =0,1 for x,y axis */
{
	char	xscale;		/* x-axis scale factor */
	double	xwrap;		/* x-axis azimuthal limit */

	if (sc == 'd' || sc == 'h') {
		register deg, mins;
		register char *t;
		double	 sec;
	
		if (d < 0) { *s++ = '-'; d = -d; }
		mins = 60 * (d - (deg = (int)d));
		sec = 3600 * d - 60 * (mins + 60*deg);
		if (*form)
			t = form;
		else if (sec != 0.0)
			t = (sc == 'd') ? "%d\013%d'%.12g\"" : "%dh%dm%.12gs";
		else if (mins != 0)
			t = (sc == 'd') ? "%d\013%d'" : "%dh%dm";
		else
			t = (sc == 'd') ? "%d\013" : "%dh";
		(void)sprintf(s, t, deg, mins, sec);
		return;
	}

	if (mode == 0 && ihget("xscale", &xscale, 1) == 0 && xscale == 'w') {
		if (ihget("xwrap", &xwrap, 0))
			xwrap = 360;
		if (d < 0)
			d += xwrap;
		else if (d > xwrap)
			d -= xwrap;
	}

	if (*form)
		(void)sprintf(s, form, d);
	else
		(void)gcvt(d, precision, s);
}
	/*
	**	Convert string "s" to float or fixed, then call gips()
	**	or egips() to convert to real or pixel.
	**	Returns:    0 - fixed value converted to realworld
	**		   -1 -	float value converted to fixed
	**		  > 0 -	conversion error - message generated
	*/
static
convert_to_real(s, d, i, n, m)

	register char	*s;	/* string value */
	register double	*d;	/* pointer to float output */
	register int	*i;	/* pointer to integer value */
	int		n;	/* pixel value if null string */
	register	m;	/* mode value */
{
	*i = n;
	if (*s) {
		register char	*c = ".edED";	/* float indicators */

		while (*c && index(s, *c) == NULL)
			c++;
		if (*c && sscanf(s, "%lf", d) == 1) {
			if (gips(d, i, 1, m ? GIPS_Y_VAL : GIPS_X_VAL))
				return 1;
			*i = IMAGE_TO_RECT(*i, m);
			return -1;
		}
		else if (*c || (sscanf(s, "%d", i) != 1))
			return 1;
	}
	n = RECT_TO_IMAGE(*i, m);
	return egips(&n, d, 1, egips_mode[m]);
}
	/*
	**	Generate a new "axis" string and put it
	**	into the incore GIPS header
	*/
static
update_gips_axis(axis, sc, i)

	register char 		*axis;	/* x, y axis string */
	register struct scan 	*sc;	/* pointer to axis table */
	register		i;	/* =0, 1 for x, y */
{
	for (*axis = 0; sc->name[0] != NULL; sc++) {
		if ((sc->flg & A_FOUND) == 0)
			continue;
		sc->flg &= ~A_FOUND;
		if (sc->flg & A_INT)
			(void)sprintf(axis+strlen(axis), sc->name[1],
				*((int *)sc->addr));
		else
			(void)sprintf(axis+strlen(axis), sc->name[1], sc->addr);
	}
	return ihput(i ? "yaxis" : "xaxis", ++axis);
}
	/*
	**	Return a "representative" pixel coordinate value
	*/
double
truncate_gips_value(pix, mode)

	int	pix;			/* image pixel address */
	int	mode;			/* axis mode, 0=x, 1=y */
{
	struct	{
		char	msign;
		char	lead;
		char	period;
		char	mant[15];
		char	ee;
		char	esign;
		char	exp[2];
		char	pad;
	} cbuf[2];
	register i0, i1;
	register char *s, *t;
	register double	*dlim;		/* pointer to coord values */
	/*
	**	Get dlim -> realworld pixel coordinate
	*/
	dlim = &xyval[mode][RECT_TO_IMAGE(pix, mode)];
	/*
	**	Detect error conditions
	*/
	if (dlim[0] == GIPSLOW || dlim[0] == GIPSHIGH || dlim[0] == dlim[1])
		return dlim[0];
	if (dlim[1] == GIPSLOW || dlim[1] == GIPSHIGH)
		return dlim[1];
	/*
	**	Determine order of limits
	*/
	i0 = ((dlim[0] > dlim[1] && dlim[0] > 0) || 
		(dlim[0] < dlim[1] && dlim[0] < 0));
	/*
	**	Convert limits to ascii
	*/
	(void)sprintf((char *)&cbuf[0], "%22.15e", dlim[i0]);
	(void)sprintf((char *)&cbuf[1], "%22.15e", dlim[1-i0]);
	cbuf[0].pad = ' ';
	/*
	**	If limits are of opposite sign, the answer must be 0
	*/
	if (cbuf[0].msign != cbuf[1].msign)
		return 0;
	/*
	**	Determine difference in exponents
	*/
	i0 = 10 * cbuf[0].exp[0] + cbuf[0].exp[1] - 11 * '0';
	if (cbuf[0].esign == '-') i0 = -i0;
	i1 = 10 * cbuf[1].exp[0] + cbuf[1].exp[1] - 11 * '0';
	if (cbuf[1].esign == '-') i1 = -i1;
	/*
	**	Hide the periods for string search
	*/
	if (i0 == i1)
		cbuf[0].period = cbuf[0].lead;
	else {
		cbuf[0].period = '0';
		cbuf[0].mant[0] = (i1 > ++i0) ? '0' : cbuf[0].lead;
	}
	cbuf[0].ee = '0';
	cbuf[1].period = cbuf[1].lead;
	/*
	**	Determine the leading difference digit
	*/
	for (s = &cbuf[0].lead, t = &cbuf[1].lead; *++s == *++t; )
		;
	/*
	**	Inspect the differing digits and replace them
	*/
	if (*t != 'e')
		if (*t > *s + 1)
			*s = (*s < '5' && *t > '5') ? '5' :
				(*s + *t) >> 1;
		else if (t[1] == '0' && s[1] != '9') {
			s++;
			*s = (*s < '5') ? '5' : (*s + '0' + 10) >> 1;
		} else
			(*s)++;
	/*
	**	Put the period and exponent back again
	*/
	cbuf[0].lead = cbuf[0].period;
	cbuf[0].period = '.';
	(void)strcpy(++s, &cbuf[1].ee);
	/*
	**	Convert to binary
	*/
	(void)sscanf((char *)cbuf, "%lf", dlim);
	return dlim[0];
}
