#ifndef lint
static char *sccsid = "%Z%$RCSfile: color.c,v $	GIPS $Revision: 3.12 $ (MIT/CSR) $Date: 1998/11/13 23:33:05 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		color.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.12 $		Update:		$Date: 1998/11/13 23:33:05 $
**	Components:	create_color_palette()	create palette pop-up
**			display_color_wedge()	create wedge pup-up
**			error_diffuse_wedge()	fill B/W wedge popup
**			flood_overlay()		fill closed paint boundary
**			flood_raster()		fill paint raster
**			get_color()		read colormap file
**			handle_palette_event()	mouse down in palette
**			handle_wedge_event()	mouse down in wedge
**			highlight_color()	emphasize palette item
**			init_color()		generate shared colormap name
**			paint_pixel()		brush stroke in paint
**			read_paint_overlay()	input saved paint file
**			stretch_colormap()	stretch pixel values
**			wedge_repaint()		repair wedge damage
**			write_colormap()	write colormap file
**			write_paint_overlay()	read saved paint file
**
**	Function:	Allocate colormap segment, manage color objects
*/
#include	"gipstool.h"
#include	<sunwindow/cms.h>

static	short	image_color_icon[] = {
#include "color.icon"
};
DEFINE_ICON_FROM_IMAGE(color_icon, image_color_icon);

static	Rect	color_icon_rect = { 0, 48, 64, 16 };

static	char	colormap_name[CMS_NAMESIZE];	/* shared colormap */
static	char	colormap_file[FONTLEN];		/* colormap file name */
static	wedge_repaint(), error_diffuse_wedge(), highlight_color();
	/*
	**	Initialize colormap->colormap_name and afile->colormap_file
	*/
init_color()
{
	if (afile == NULL)
		afile = "bw";
	if (colormap == NULL)
		if ((colormap = rindex(afile, '/')) == NULL)
			colormap = afile;
		else
			colormap++;
	colormap = strncpy(colormap_name, colormap, CMS_NAMESIZE-1);
	if (index(colormap, '.') != NULL)
		*index(colormap, '.') = '\0';
	pw_setcmsname(image_pw, colormap);
	return get_color(afile);
}
	/*
	**	Read the colormap from "afile" or "afile.cmap".
	**	Look first in current directory, then GIPSCMAPDIR.
	*/
get_color(name)
	char	*name;			/* colormap file name */
{
	char	buf[BUFSIZ];		/* input buffer */
	unsigned char rgb0[3];		/* conversion buffer */
	double	d[3];			/* input {rgb} color triplet */
	register FILE	*fp;		/* colormap file pointer */
	register i, n, m;		/* scratch */

	(void) strcpy(buf, name);
	if (name[0] == '!')
		fp = popen(name, "r");
	else {
		fp = fopen(name, "r");
		if (fp == NULL)
			fp = fopen(strcat(buf,".cmap"), "r");
		if (fp == NULL && *name != '/') {
			fp = fopen(strcat(strcpy(buf,GIPSCMAPDIR),name), "r");
			if (fp == NULL)
				fp = fopen(strcat(buf,".cmap"), "r");
		}
	}
	if (fp == NULL)
		return 1;
	m = 255 >> (8 - cshift);
	/*
	**	Disregard comment lines beginning with "#"
	**	If -N flag set, accept all RGBSIZE entries.
	**	Otherwise, skip every other entry.
	*/
	for (i = n = 0; i < RGBSIZE; ) {
		if (fgets(buf,sizeof(buf)-1,fp) == NULL)
			return 1;
		if (buf[0] == '#')
			continue;
		if (sscanf(buf," %lf %lf %lf ",&d[0],&d[1],&d[2]) != 3) {
			if (*name == '!')
				(void) pclose(fp);
			return 1;
		}
		if (++i & m)
			continue;
		gconv((char *)rgb0,(char *)d,3,CVT_B,CVT_D);
		if (rgb0[0] == 0 && rgb0[1] == 0 && rgb0[2] == 0)
			rgb0[0] = rgb0[1] = rgb0[2] = 1;
		else if (rgb0[0] == 255 && rgb0[1] == 255 && rgb0[2] == 255)
			rgb0[0] = rgb0[1] = rgb0[2] = 254;
		rgb[0][n] = rgb0[0];
		rgb[1][n] = rgb0[1];
		rgb[2][n] = rgb0[2];
		n++;
	}
	if (*name == '!') {
		(void) pclose(fp);
		name++;
	} else
		(void) fclose(fp);
	afile = strncpy(colormap_file, name, FONTLEN-1);
	move((char *)rgb, (char *)newrgb, sizeof(rgb));
	if (pw_putcolormap(image_pw, 0, colors,
			rgb[0], rgb[1], rgb[2]) == PIX_ERR)
		return 1;
	return 0;
}
	/*
	**	Transform colormap via image_hist[] array
	*/
stretch_colormap(percent)
	int	percent;
{
	int	total, size;
	float	scale, fract;
	register i, j, l;
	/*
	**	Toggle on stretchflag
	*/
	size = colors;
	if (percent == 0) { 
		move((char *)rgb, (char *)newrgb, sizeof(rgb));
		pw_putcolormap(image_pw, 0, size, rgb[0], rgb[1], rgb[2]);
		return;
	}
	/*
	**	Get total count of non-zero pixels
	*/
	for (total = 0, i = 1; i < (size-1); i++)
		total += image_hist[i];
	if (total == 0)
		return;
	scale = ((float)percent * size) / (float)total;
	/*
	**	Construct new stretched colormap
	*/
	for (fract = 0, i = 1; i < (size-1); i++) {
		fract += (float) image_hist[i];
		l = ((int)(scale*fract) + (100-percent)*i)/100;
		for (j = 0; j < 3; j++)
			newrgb[j][i] = rgb[j][l];
	}
	pw_putcolormap(image_pw, 1, size-2, newrgb[0], newrgb[1], newrgb[2]);
}
	/*
	**	Transform colormap linearly
	*/
linstretch_colormap()
{
	double	dmin, dmax;
	register i, j, l;
	/*
	**	Toggle on stretchflag
	*/
	if (ihget("dmin", &dmin)) dmin = 0;
	if (ihget("dmax", &dmax)) dmax = 255;
	if (stretch_min == stretch_max || dmin == dmax) {
		display_error_message("bad stretch limits", (char *)NULL);
		return;
	}
	if (stretch_min == dmin && stretch_max == dmax) { 
		move((char *)rgb, (char *)newrgb, sizeof(rgb));
		pw_putcolormap(image_pw, 0, colors, rgb[0], rgb[1], rgb[2]);
		return;
	}
	/*
	**	Construct new stretched colormap
	*/
	for (i = 1; i < (colors-1); i++) {
		l = (i * (dmax - dmin) + (dmin - stretch_min) * colors)
			/ (stretch_max - stretch_min);
		l = MIN(colors-1, MAX(1, l));
		for (j = 0; j < 3; j++)
			newrgb[j][i] = rgb[j][l];
	}
	pw_putcolormap(image_pw, 1, colors-2, newrgb[0], newrgb[1], newrgb[2]);
}
	/*
	**	Write out a colormap
	*/
write_colormap(c)
	char	*c;			/* name of output file */
{
	register FILE	*fp;		/* output stream pointers */
	register int	i, m;		/* scratch */

	if ((fp = fopen(c, "w")) == NULL) {
		display_system_message(c);
		return;
	}
	for (m = RGBSIZE/colors, i = 0; i < 256; i++)
		(void)fprintf(fp, "\t%d\t%d\t%d\n",
			newrgb[0][i/m], newrgb[1][i/m], newrgb[2][i/m]);
	(void) fclose(fp);
}

	/*
	**	Popups for color wedge
	*/
static	Frame		wedge_frame;	/* Frame for wedge popup */
static	Canvas		wedge_canvas;	/* Canvas for wedge popup */
static	Pixwin		*wedge_pw;	/* Pixwin for wedge popup */
static	struct pixrect	*wedge_pr;	/* Pixrect for wedge popup */
static	Menu		wedge_menu;	/* Popup menu within wedge */
static	int	wedge_width	= 256;	/* width of wedge window */
static	int	wedge_height	= 64;	/* height of wedge window */
static	int	wedge_sense	= 1;	/* painting direction */
	/*
	**	Event handler to pass menu button
	*/
/*ARGSUSED*/
static
handle_wedge_event(canvas, event, arg)
	Canvas	canvas;
	Event	*event;
	caddr_t arg;
{
	int	sw;

	if (event_id(event) == MS_RIGHT) {
		sw = (int)menu_show(wedge_menu, wedge_canvas, event, 0);
		if (sw == 1)
			window_set(wedge_frame, WIN_SHOW, FALSE, 0);
		else if (sw > 1 && sw < 6) {
			wedge_sense = sw - 1;
			wedge_repaint(wedge_canvas, wedge_pw, (Rectlist *)NULL);
		}
	} else if (event_is_ascii(event) && event_is_down(event))
		image_selected(image_sw, event, 0);
}
	/*
	**	Repaint procedure for wedge canvas
	*/
/*ARGSUSED*/
static
wedge_repaint(canvas, pixwin, repaint_area)
	Canvas		canvas;
	Pixwin		*pixwin;
	Rectlist	*repaint_area;
{
	Rect		r;
	register int	x, y, i, l, rflag, shift;
	register char	*p;

	r = *((Rect *)window_get(canvas, WIN_RECT));

	if (! wedge_pr || wedge_sense || r.r_width != wedge_width ||
			r.r_height != wedge_height) {
		shift = 8 - cshift;
		rflag = wedge_sense & 01;

		if (wedge_pr)
			pr_close(wedge_pr);
		wedge_width = r.r_width;
		wedge_height = r.r_height;
		wedge_pr = mem_create(wedge_width, wedge_height, 8);
		p = (char *)mpr_d(wedge_pr)->md_image;
		l = mpr_d(wedge_pr)->md_linebytes;
		switch (wedge_sense) {
		case 1:
		case 2:
			for (x = 0; x < wedge_width; x++)
				p[x] = ((rflag ? x : (wedge_width - x))
					<< shift) / wedge_width;
			for (y = 1; y < wedge_height; y++, p += l)
				bcopy(p, p+l, wedge_width);
			break;
		case 3:
		case 4:
			for (y = 0; y < wedge_height; y++, p += l) {
				i = ((rflag ? (wedge_height - y) : y)
					<< shift) / wedge_height;
				for (x = 0; x < wedge_width; x++)
					p[x] = i;
			}
			break;
		}
		if (wedge_pw->pw_pixrect->pr_depth == 1)
			error_diffuse_wedge();
		window_set(canvas,
			WIN_WIDTH,		wedge_width,
			WIN_HEIGHT,		wedge_height,
			CANVAS_WIDTH,		wedge_width,
			CANVAS_HEIGHT,		wedge_height,
			0);
		window_fit(wedge_frame);
		wedge_sense = 0;
	}
	pw_write(pixwin, 0, 0, wedge_width, wedge_height, 
		PIX_SRC, wedge_pr, 0, 0);
}

static
error_diffuse_wedge()
{
	short	*zs[2];
	register Pixrect *pr;
	register y;
	register char	*p, *q;

	zs[0] = (short *)alloc(wedge_width+2, sizeof(short));
	zs[1] = (short *)alloc(wedge_width+2, sizeof(short));
	pr = mem_create(wedge_width, wedge_height, 1);

	p = (char *)mpr_d(wedge_pr)->md_image;
	q = (char *)mpr_d(pr)->md_image;
	for (y = 0; y < wedge_height; y++) {
		if (Zflag)
			diffuse_error(p, q, wedge_width,
				zs[y & 1] + 1, zs[1 - (y & 1)] + 1);
		else
			shift_bw(p, q, wedge_width);
		p += mpr_d(wedge_pr)->md_linebytes;
		q += mpr_d(pr)->md_linebytes;
	}

	pr_close(wedge_pr);
	free(zs[0]);
	free(zs[1]);
	wedge_pr = pr;
}

	/*
	**	Create wedge window
	*/
display_color_wedge(ie, s)
	register Event	*ie;
	register char	*s;
{
	Rect		r;
	register char	*c;
	register int	x, y;

	(void)sscanf(s, "%d %d", &wedge_width, &wedge_height);

	if (wedge_canvas && !window_get(wedge_canvas, WIN_RECT)) {
		wedge_frame = NULL;
		if (wedge_menu)
			menu_destroy(wedge_menu);
	}
	if (! wedge_frame) {
		wedge_frame = window_create(0, FRAME,
			FRAME_LABEL,	window_get(base_frame, FRAME_LABEL),
			FRAME_SHOW_LABEL,	FALSE,
			0);
		c = rindex(imfile, '/');
		icon_set(&color_icon,
			ICON_LABEL,		c ? c+1 : imfile,
			ICON_LABEL_RECT,	&color_icon_rect,
			0);
		window_set(wedge_frame,
			FRAME_ICON,		&color_icon,
			FRAME_NO_CONFIRM,	TRUE,
			0);
		wedge_canvas = window_create(wedge_frame, CANVAS,
			WIN_EVENT_PROC,		handle_wedge_event,
			CANVAS_AUTO_CLEAR,	FALSE,
			CANVAS_REPAINT_PROC,	wedge_repaint,
			WIN_CONSUME_KBD_EVENTS,
				WIN_NO_EVENTS,
				WIN_ASCII_EVENTS,
				0,
			WIN_CONSUME_PICK_EVENTS,
				WIN_NO_EVENTS,
				WIN_MOUSE_BUTTONS,
				0,
			0);
		wedge_pw = (Pixwin *)window_get(wedge_canvas, WIN_PIXWIN);
		if (wedge_pw->pw_pixrect->pr_depth != 1) {
			pw_setcmsname(wedge_pw, colormap);
			pw_putcolormap(wedge_pw, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
		}
		wedge_pw = canvas_pixwin(wedge_canvas);
		if (wedge_pw->pw_pixrect->pr_depth != 1) {
			pw_setcmsname(wedge_pw, colormap);
			pw_putcolormap(wedge_pw, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
		}
		wedge_menu = menu_create(
			MENU_STRINGS,
				"Hide Insert",
				"Left to Right",
				"Right to Left",
				"Bottom to Top",
				"Top to Bottom",
				0,
			MENU_PULLRIGHT_ITEM,	"Frame",
				(Menu)window_get(wedge_frame, WIN_MENU),
			0);
	}
	r = *((Rect *)window_get(base_frame, WIN_RECT));
	x = r.r_left + event_x(ie) - wedge_width/2
		+ (int)window_get(base_frame, WIN_LEFT_MARGIN)
		- (int)window_get(wedge_frame, WIN_LEFT_MARGIN);
	y = r.r_top + event_y(ie) - wedge_height/2
		+ (int)window_get(base_frame, WIN_TOP_MARGIN)
		- (int)window_get(wedge_frame, WIN_TOP_MARGIN);
	window_set(wedge_frame,
		WIN_X,			MAX(x, 0),
		WIN_Y,			MAX(y, 0),
		0);
	window_set(wedge_canvas,
		WIN_X,			0,
		WIN_Y,			0,
		WIN_WIDTH,		wedge_width,
		WIN_HEIGHT,		wedge_height,
		CANVAS_WIDTH,		wedge_width,
		CANVAS_HEIGHT,		wedge_height,
		0);
	window_fit(wedge_frame);
	wedge_repaint(wedge_canvas, wedge_pw, (Rectlist *)NULL);
	window_set(wedge_frame, WIN_SHOW, TRUE, 0);
}

	/*
	**	Paint pixels into an overlay graphic
	*/
paint_pixel(ie)
	register Event	*ie;
{
	register int	x, y, dx, dy, x0, y0, dx0, dy0, sw;

	if (depth == 1)
		return;
	if (paint_pr == NULL) {
		paint_pr = mem_create(saved_rect.r_width,
			saved_rect.r_height, depth);
		if (paint_pr == NULL) {
			no_more_memory();
			return;
		}
	}
	x = event_x(ie)-brush_size;
	y = event_y(ie)-brush_size;
	dx = MAX(0, MIN(brush_size, sw_rect.r_width - x));
	dy = MAX(0, MIN(brush_size, sw_rect.r_height - y));
	if (dx <= 0 || dy <= 0)
		return;
	if (event_shift_is_down(ie)) {
		sw = PIX_SRC | PIX_COLOR(0);
	} else {
		sw = PIX_SRC | PIX_COLOR(graphic_color >> cshift);
		pw_write(image_pw, x, y, dx, dy, sw, NULL, 0, 0);
	}
	x0 = x + (xorg+xoff)*xzoom - image_rect.r_left,
	y0 = y + (yorg+yoff)*yzoom - image_rect.r_top,
	dx0 = MIN(brush_size, saved_rect.r_width - x0);
	dy0 = MIN(brush_size, saved_rect.r_height - y0);
	if (dx0 <= 0 || dy0 <= 0)
		return;
	dx = MIN(dx, dx0);
	dy = MIN(dy, dy0);
	if (event_shift_is_down(ie))
		pw_write(image_pw, x, y, dx, dy, PIX_SRC, saved_pr, x0, y0);
	pr_rop(paint_pr, x0, y0, dx0, dy0, sw, NULL, 0, 0);
}

	/*
	**	Flood a region in the paint overlay
	*/

flood_overlay(ie)
	Event	*ie;
{
	if (depth == 1)
		return;
	if (paint_pr == NULL) {
		display_error_message("no paint overlay exists", (char *)0);
		return;
	}
	flood_raster(event_x(ie)+(xorg+xoff)*xzoom - image_rect.r_left,
		event_y(ie)+(yorg+yoff)*yzoom - image_rect.r_top,
		event_shift_is_down(ie), 3);
}

	/*
	**	Fill the current y-raster
	*/

flood_raster(x, y, sw, dir)
	register int	x, y;		/* starting pixel */
	register int	sw;		/* =1 to clear */
	register int	dir;		/* 2=go up, 1=go down */
{
	register char	*c;
	register int	len, i, color;

	if (y < 0 || y >= saved_rect.r_height)
		return;
	len = mpr_d(paint_pr)->md_linebytes;
	c = (char *)(mpr_d(paint_pr)->md_image) + len*y;
	color = graphic_color >> cshift;

	for (i = x; i > 0 && c[i-1] == sw ? color : 0; i--)
		;

	for (; i < saved_rect.r_width && c[i] == sw ? color : 0; i++) {
		c[i] = sw ? 0 : color;
		if ((dir & 2) && y > 0 && c[i-len] == sw ? color : 0)
			flood_raster(i, y-1, sw, 2);
		if ((dir & 1) && y < saved_rect.r_height-1 &&
				c[i+len] == sw ? color : 0)
			flood_raster(i, y+1, sw, 1);
	}
	must_update = 1;
}

	/*
	**	Read paint overlay file
	*/

read_paint_overlay()
{
	register FILE	*fp;
	register char	*c;
	register int	n, y;
	char		buf[6];

	if (paint_pr == NULL) {
		paint_pr = mem_create(saved_rect.r_width,
			saved_rect.r_height, depth);
		if (paint_pr == NULL) {
			no_more_memory();
			return;
		}
	}

	if ((fp = fopen(paint_overlay_file, "r")) == NULL) {
		display_system_message(paint_overlay_file);
		return;
	}

	for (c = buf; (n = getc(fp)) != EOF; *c++ = n)
		if (n == 'F')
			c = buf;
		else if (c < buf+6)
			;
		else if (strncmp(buf, "FINIS=", 6) == 0)
			break;
		else
			c = buf;

	c = (char *)(mpr_d(paint_pr)->md_image);
	n = mpr_d(paint_pr)->md_linebytes;

	for (y = 0; y < saved_rect.r_height; y++, c += n)
		if (fread(c, saved_rect.r_width, 1, fp) != 1) {
			display_system_message(paint_overlay_file);
			break;
		}

	(void)fclose(fp);
	must_update++;
}

	/*
	**	Write paint overlay into file
	*/

write_paint_overlay()
{
	register int	x, y;
	register FILE	*fp, *fq;
	register char	*c;
	char	buf[1024];

	if ((fp = fopen(paint_overlay_file, "w")) == NULL) {
		display_system_message(paint_overlay_file);
		return;
	}

	/*
	**	Open the default GIPS header file
	*/

	(void) sprintf(buf, "%s%s", GIPSPROFDIR, "GIPSinit");
	if ((fq = fopen(buf, "r")) == NULL) {
		display_system_message(buf);
		(void) fclose(fp);
		return;
	}

	(void)fprintf(fp, "dtype = b\n");
	(void)fprintf(fp, "dnull = 0\n");
	(void)fprintf(fp, "xnum = %d\n", saved_rect.r_width);
	(void)fprintf(fp, "ynum = %d\n", saved_rect.r_height);

	copy_header_variable(fp, "title", "title");
	copy_header_variable(fp, "xtitle", "xtitle");
	copy_header_variable(fp, "ytitle", "ytitle");
	copy_header_variable(fp, "dtitle", "dtitle");
	copy_header_variable(fp, "owner", "owner");
	copy_header_variable(fp, "origin", "origin");
	copy_header_variable(fp, "date", "date");
	copy_header_variable(fp, "time", "time");

	/*
	**	Append the default header file
	*/

	while ((x = getc(fq)) != EOF)
		(void)putc(x, fp);
	(void) fclose(fq);

	c = (char *)(mpr_d(paint_pr)->md_image);
	x = mpr_d(paint_pr)->md_linebytes;

	for (y = 0; y < saved_rect.r_height; y++, c += x)
		if (fwrite(c, saved_rect.r_width, 1, fp) != 1) {
			display_system_message(paint_overlay_file);
			break;
		}

	(void)fclose(fp);
}

static	Frame	palette_frame;
static	Canvas	palette_canvas;
static	Pixwin	*palette_pixwin;
static	int	palette_width, palette_height;

#define	PALETTE_SIZE	16

/*ARGSUSED*/
static
handle_palette_event(canvas, event, arg)
	Canvas	canvas;
	Event	*event;
	caddr_t arg;
{
	int	n;

	if (event_id(event) == MS_LEFT && event_is_down(event)) {
		highlight_color(event, graphic_color, 0);
		graphic_color = palette_width * (event_y(event)/PALETTE_SIZE)
			+ event_x(event)/PALETTE_SIZE;
		graphic_color <<= cshift;
		highlight_color(event, graphic_color, 1);
		update_info();
	} else if (event_is_ascii(event) && event_is_down(event))
		image_selected(image_sw, event, 0);
}

create_color_palette(event)
	Event	*event;
{
	int	i, x, y;

	if (palette_pixwin == NULL) {
		palette_height = 1 << ((8 - cshift) / 2);
		palette_width = colors / palette_height;
		palette_frame = window_create(base_frame, FRAME,
			FRAME_LABEL,	window_get(base_frame, FRAME_LABEL),
			FRAME_SHOW_LABEL,		TRUE,
			0);	
		palette_canvas = window_create(palette_frame, CANVAS,
			WIN_X,			0,
			WIN_Y,			0,
			WIN_WIDTH,		palette_width * PALETTE_SIZE,
			WIN_HEIGHT,		palette_height * PALETTE_SIZE,
			WIN_EVENT_PROC,		handle_palette_event,
			WIN_CONSUME_KBD_EVENTS,
				WIN_NO_EVENTS,
				WIN_ASCII_EVENTS,
				0,
			WIN_CONSUME_PICK_EVENTS,
				WIN_NO_EVENTS,
				WIN_MOUSE_BUTTONS,
				0,
			CANVAS_WIDTH,		palette_width * PALETTE_SIZE,
			CANVAS_HEIGHT,		palette_height * PALETTE_SIZE,
			0);
		window_fit(palette_frame);
		palette_pixwin = (Pixwin *)
			window_get(palette_canvas, WIN_PIXWIN);
		pw_setcmsname(palette_pixwin, colormap);
		pw_putcolormap(palette_pixwin, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
		palette_pixwin = canvas_pixwin(palette_canvas);
		pw_setcmsname(palette_pixwin, colormap);
		pw_putcolormap(palette_pixwin, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
	}
	for (i = y = 0; y < palette_height; y++)
		for (x = 0; x < palette_width; x++, i++)
			pw_rop(palette_pixwin,
				x * PALETTE_SIZE, y * PALETTE_SIZE,
				PALETTE_SIZE, PALETTE_SIZE,
				PIX_SRC | PIX_COLOR(i), NULL, 0, 0);
	highlight_color(event, graphic_color, 1);
}

static
highlight_color(event, color, sw)
	Event	*event;
	int	color, sw;
{
	int	x, y, d, value;
	char	buf[32], *b;

	color >>= cshift;
	value = sw ? 0 : (graphic_color >> cshift);
	if (sw && color == 0 && value == 0)
		value = colors - 1;
	x = PALETTE_SIZE * (color % palette_width);
	y = PALETTE_SIZE * (color / palette_width);
	d = PALETTE_SIZE - 1;
	pw_vector(palette_pixwin, x, y, x + d, y, PIX_SRC, value);
	pw_vector(palette_pixwin, x, y, x, y + d, PIX_SRC, value);
	pw_vector(palette_pixwin, x + d, y, x + d, y + d, PIX_SRC, value);
	pw_vector(palette_pixwin, x, y + d, x + d, y + d, PIX_SRC, value);
	if (sw) {
		(void)sprintf(b = buf, "C%d\r", graphic_color);
		window_set(palette_frame,
			FRAME_LABEL,	buf,
			0);
		if (event == NULL)
			return;
		if ((x = no_display) == 0)
			window_set(palette_frame, WIN_SHOW, TRUE, 0);
		no_display = 1;
		while (*b) {
			event_set_id(event, *b++);
			image_selected(image_sw, event, 0);
		}
		no_display = x;
	}
}
