#ifndef lint
static char *sccsid = "%Z%$RCSfile: format.c,v $	GIPS $Revision: 3.4 $ (MIT/CSR) $Date: 1998/05/13 01:12:19 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		format.c	Creation Date:	03/19/86
**	Version:	$Revision: 3.4 $		Update:		$Date: 1998/05/13 01:12:19 $
**	Components:	get_coordinates	transform pixel to ASCII value
**			shift_bw	transform input B/W pixel to bits
**
**	Function:	Handle translation between d/n value and pixel
*/
#include	"gipstool.h"
	/*
	**	Generate coordinate description strings
	*/
char	*
get_coordinates(outbuf, fmt, ie)
	char		*outbuf;	/* output buffer */
	char		*fmt;		/* format string */
	Event		*ie;		/* input event structure */
{
	register char	*s;		/* pointer to buffer */
	register char	*f;		/* pointer to format */
	register 	sw;		/* conversion code */
	register 	mode;		/* 0,1,2 = x,y,d */
	int		x, y;		/* cursor location */
	int		v;		/* pixel value */
	double		d;		/* realworld value */
	double		pix[2];		/* scratch */
	char		buf[7];		/* for ihget() */

	for (s = outbuf, *s = '\0', f = fmt; *f; f++) {
		if (*f != '%' || (f[1] && *++f == '%')) {
			*s++ = *f;
			continue;
		}
		while (isspace(*f)) f++;
		switch (*f) {
		case 'r':
			sw = 1;
			f++;
			break;
		case 't':
			sw = 2;
			f++;
			break;
		default:
			sw = 0;
			break;
		}
		while (isspace(*f)) f++;
		switch (*f) {
		case 'X':
		case 'x':
			mode = EGIPS_X_VAL;
			x = v = X_RECT_TO_IMAGE(event_x(ie));
			if (map_sw)
				y = Y_RECT_TO_IMAGE(event_y(ie));
			break;
		case 'Y':
		case 'y':
			mode = EGIPS_Y_VAL;
			y = v = Y_RECT_TO_IMAGE(event_y(ie));
			if (map_sw)
				x = X_RECT_TO_IMAGE(event_x(ie));
			break;
		case 'D':
		case 'd':
			mode = EGIPS_D_VAL;
			x = event_x(ie) + saved_rect.r_left - image_rect.r_left;
			y = event_y(ie) + saved_rect.r_top - image_rect.r_top;
			if (depth == 1)
				v = pr_get(data_pr, x/xzoom, y/yzoom) & 0377;
			else
				v = (pr_get(saved_pr, x, y) << cshift) & 0377;
			if (*f == 'D')
				v += RGBSIZE/colors;
			break;
		default:
			display_error_message("bad format: %s", fmt);
			outbuf[0] = '\0';
			return outbuf;
		}
		if (isupper(*f))
			v++;
		switch (sw) {
		case 0:
			(void) sprintf(s, "%d", v);
			break;
		case 1:
			if (map_sw && mode != EGIPS_D_VAL)
				d = (g_unmap(x, y, pix)) ? GIPSHIGH : pix[mode];
			else
				(void) egips(&v, &d, 1, mode);
			if (d >= GIPSHIGH || (mode == EGIPS_D_VAL && v >= 256))
				(void) strcpy(s, "MAX");
			else if (d <= GIPSLOW || (mode == EGIPS_D_VAL && v <= 0))
				(void) strcpy(s, "MIN");
			else if (d < 1.e-12 && d > -1.e-12)
				(void) strcpy(s, "0");
			else {
				char	xscale;
				double	xwrap;

				if (mode != EGIPS_X_VAL)
					;
				else if (ihget("xscale", &xscale, 1))
					;
				else if (xscale == 'w') {
					if (ihget("xwrap", &xwrap, 0))
						xwrap = 360;
					if (d < 0)
						d += xwrap;
					else if (d > xwrap)
						d -= xwrap;
				}
				(void) gcvt(d, precision, s);
			}
			break;
		case 2:
			sprintf(buf, "%ctitle", "xyd"[mode]);
			if (ihget(buf, s, 64))
				(void)sprintf(s, "%c-value", buf[0]);
			break;
		}
		s += strlen(s);
	}
	*s = '\0';
	return outbuf;
}
	/*
	**	Byte-pixel to bit-array lookup array
	*/
static	char	pixel_bit_value[4][64] = {
	{ 0, /*1x1*/
	  1 },
	{ 0,  2,  2,  3, /*2x2*/
	  0,  0,  1,  2 },
	{ 0,  0,  1,  4,  2,  2,  6,  3, /*3x3*/
	  0,  2,  0,  2,  5,  7,  5,  7,
	  0,  0,  4,  1,  2,  2,  3,  6 },
	{ 0,  0,  0,  2,  4,  5,  4,  4,  8, 11, 11, 10, 11, 13, 15, 15, /*4x4*/
	  0,  4,  2,  0,  1,  8, 12, 12,  5,  3,  3,  7, 14, 15, 13, 11,
	  0,  0,  0, 10,  8,  2,  3,  3,  8, 12, 12, 13,  7,  5, 15, 15,
	  0,  0,  8,  0,  2,  8,  2, 10,  5,  5, 13,  7, 13, 15,  7, 15 }
};
	/*
	**	Pack byte-pixel input array to pixrect
	*/
shift_bw(in, out, npix)
	char	*in;
	short	*out;
	int	npix;
{
	register	dim = xzoom, acc, shift, n, m;
	char	*mask	= pixel_bit_value[xzoom-1];
	int	d	= yzoom;

	while (--d >= 0) {
		for (acc = shift = n = 0; n < npix; n++) {
			m = in[n] & 0377;
			image_hist[m]++;
			acc = (acc << dim) | ((m == 255) ?
				(m >> (8-dim)) : (mask[m >> (8-dim)] & 0377));
			if ((shift += dim) > 16) {
				*out++ = m = (acc >> (shift -= 16));
				acc ^= (m << shift);
			}
		}
		if (shift) *out++ = (acc << (16-shift));
		mask += (1 << dim);
	}
}
