#ifndef lint
static char *sccsid = "%Z%$RCSfile: graphics.c,v $	GIPS $Revision: 3.7 $ (MIT/CSR) $Date: 1998/05/13 01:12:20 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		graphics.c	Creation Date:	03/19/86
**	Version:	$Revision: 3.7 $		Update:		$Date: 1998/05/13 01:12:20 $
**	Function:	Handle all graphics
**	Components:	clear_graphic		remove a graphic index
**			display_graphics	display a graphic index
**			draw_line		add a graphic line
**			rotate_pixrect		generate rotated pixrect
**			write_text		add a graphic text line
*/
#include	"gipstool.h"
	/*
	**	Each graphic entity is stored in the following
	*/
static	struct	g_elem	{
	struct	g_elem	*ge_next;	/* forward chain */
	int		ge_index;	/* graphic index number */
	int		ge_flags;	/* graphic item flags */
#define		GE_LINE		01	/* .. Element is a line */
#define		GE_RECT		02	/* .. Element is a pixrect */
#define		GE_BKGRND	04	/* .. Element is a background */
	int		ge_color;	/* graphic color */
	struct	pr_pos	ge_pos;		/* location in pixwin */
	struct	pr_pos	ge_lim;		/* end location of line */
	Pixrect		*ge_pr;		/* pixrect to hold text */
} *g_anch, *g_next = (struct g_elem *)&g_anch;

static	clear_background(), rotate_pixrect();

	/*
	**	Add graphic overlays
	*/
display_graphics(pw)
	register Pixwin *pw;
{
	register struct g_elem *ge;
	register x, y;

	pw_lock(pw, &image_rect);

	for (ge = g_anch; ge != NULL; ge = ge->ge_next) {
		if (graphic_sw[ge->ge_index] != 1)
			continue;
		if (ge->ge_flags & GE_BKGRND) {
			clear_background();
			continue;
		}
		x = ge->ge_pos.x;
		y = ge->ge_pos.y;
		if (ge->ge_index != 0) {
			x -= (xorg+xoff)*xzoom - image_rect.r_left;
			y -= (yorg+yoff)*yzoom - image_rect.r_top;
		}
		if (ge->ge_flags & GE_LINE) {
			pw_vector(GIPSWIN(pw),
				x, y,
				x + ge->ge_lim.x, y + ge->ge_lim.y,
				PIX_SRC, ge->ge_color);
		} else if (ge->ge_flags & GE_RECT) {
			if (depth == 1)
				pw_write(GIPSWIN(pw),
					x, y,
					ge->ge_pr->pr_size.x,
					ge->ge_pr->pr_size.y,
					PIX_SRC | PIX_DST,
					ge->ge_pr, 0, 0);
			else
				pw_stencil(GIPSWIN(pw),
					x, y,
					ge->ge_pr->pr_size.x,
					ge->ge_pr->pr_size.y,
					PIX_SRC | PIX_COLOR(ge->ge_color),
					ge->ge_pr, 0, 0, NULL, 0, 0);
		}
	}
	pw_unlock(pw);
}
	/*
	**	Remove last graphic item at current index level
	*/
undo_graphic(i)
	register i;			/* desired index level */
{
	register struct g_elem *gl, *ge, *gs = NULL;

	for (gl = (struct g_elem *)&g_anch; ge = gl->ge_next; gl = ge)
		if (ge->ge_index == i)
			gs = gl;
	if (gs == NULL)
		return 0;
	ge = gs->ge_next;
	if (ge->ge_flags & GE_RECT)
		(void)pr_close(ge->ge_pr);
	gs->ge_next = ge->ge_next;
	free((char *)ge);
	if (gs->ge_next == NULL)
		g_next = gs;
	return 1;
}
	/*
	**	Remove an index level from the graphic chain
	*/
clear_graphic(i)
	register i;			/* desired index level */
{
	register struct g_elem *gl, *ge;

	for (gl = (struct g_elem *)&g_anch; ge = gl->ge_next; gl = ge)
		if (ge->ge_index == i) {
			if (ge->ge_flags & GE_RECT)
				(void)pr_close(ge->ge_pr);
			gl->ge_next = ge->ge_next;
			free((char *)ge);
			ge = gl;
		}
	g_next = gl;
	if (i != graphic_index)
		graphic_sw[i] = 0;

	if (paint_pr) {
		pr_close(paint_pr);
		paint_pr = (Pixrect *)NULL;
	}
}
	/*
	**	Create background mask element
	*/
draw_background()
{
	register struct g_elem *ge;

	if (ge = (struct g_elem *)alloc(sizeof(struct g_elem), 1)) {
		g_next->ge_next = ge;
		g_next = ge;
		ge->ge_index = graphic_index;
		ge->ge_flags = GE_BKGRND;
		if (graphic_sw[graphic_index] == 1)
			clear_background();
	} else
		no_more_memory();
}
	/*
	**	Set all pixels outside image to background value
	*/
static
clear_background()
{
	register x, y;

	x = image_rect.r_left + image_rect.r_width;
	y = image_rect.r_top + image_rect.r_height;
	if (image_rect.r_top > 0)
		pw_writebackground(GIPSWIN(image_pw), 0, 0,
			sw_rect.r_width, image_rect.r_top,
			PIX_CLR);
	if (image_rect.r_left > 0)
		pw_writebackground(GIPSWIN(image_pw), 0, 0,
			image_rect.r_left, sw_rect.r_height,
			PIX_CLR);
	if (x < sw_rect.r_width)
		pw_writebackground(GIPSWIN(image_pw), x, 0,
			sw_rect.r_width - x, sw_rect.r_height,
			PIX_CLR);
	if (y < sw_rect.r_height)
		pw_writebackground(GIPSWIN(image_pw), 0, y,
			sw_rect.r_width, sw_rect.r_height - y,
			PIX_CLR);
}
	/*
	**	Insert straight line into display and current graphic
	*/
draw_line(x0, y0, x1, y1)
	int	x0;			/* starting x-location */
	int	y0;			/* starting y-location */
	int	x1;			/* ending x-location */
	int	y1;			/* ending y-location */
{
	register struct g_elem *ge;

	x1 -= x0;
	y1 -= y0;

	if (x1 < -1000000 || x1 > 1000000 || y1 < -1000000 || y1 > 1000000)
		return;

	if (graphic_sw[graphic_index] == 1)
		pw_vector(GIPSWIN(image_pw), x0, y0, x1+x0, y1+y0, PIX_SRC,
			graphic_color >> cshift);
	
	if (graphic_index != 0) {
		x0 += (xorg+xoff)*xzoom - image_rect.r_left;
		y0 += (xorg+yoff)*yzoom - image_rect.r_top;
	}

	if ((ge = (struct g_elem *)alloc(sizeof(struct g_elem), 1)) == NULL) {
		no_more_memory();
		return;
	}

	g_next->ge_next = ge;
	g_next = ge;

	ge->ge_index = graphic_index;
	ge->ge_flags = GE_LINE;
	ge->ge_color = graphic_color >> cshift;
	ge->ge_pos.x = x0;
	ge->ge_pos.y = y0;
	ge->ge_lim.x = x1;
	ge->ge_lim.y = y1;
}
	/*
	**	Insert text into display and current graphic
	*/
write_text(xloc, yloc, str, origin, angle)
	int	xloc;			/* x-location */
	int	yloc;			/* y-location */
	char	*str;			/* the string */
	int	origin;			/* the relative origin mode */
	int	angle;			/* the angle */
{
	register struct g_elem *ge;	/* pointer to new graphic element */
	struct	pr_subregion prb;	/* unrotated string boundary */
	/*
	**	Check that font is defined
	*/
	if (setauxfont((char *)NULL, 0) == NULL) {
		display_error_message("No font defined", "");
		return;
	}
	/*
	**	First get unrotated string size and create pixrect
	*/
	if ((ge = (struct g_elem *)alloc(sizeof(struct g_elem), 1)) == NULL) {
		no_more_memory();
		return;
	}
	(void)settype(GRAPHIC);
	(void)setink(1);
	(void)setorigin(1);
	prb.size.x = comwidth(str)+1;
	prb.size.y = comheight(str)+2;
	if ((ge->ge_pr = mem_create(prb.size.x, prb.size.y, 1)) == NULL) {
		no_more_memory();
		free((char *)ge);
		return;
	}
	(void)memmarker((char *)
		(((struct mpr_data *)ge->ge_pr->pr_data)->md_image),
		prb.size.y,
		8 * ((struct mpr_data *)ge->ge_pr->pr_data)->md_linebytes,
		0, 0, str);
	/*
	**	Relocate the pixrect origin (prb.pos) according to specs
	*/
	switch (origin) {
	case 0:
		prb.pos.x = -prb.size.x/2;
		prb.pos.y = -prb.size.y/2;
		break;
	case 1:
		prb.pos.x = 0;
		prb.pos.y = 0;
		break;
	case 2:
		prb.pos.x = -prb.size.x/2;
		prb.pos.y = 0;
		break;
	case 3:
		prb.pos.x = -prb.size.x;
		prb.pos.y = 0;
		break;
	case 4:
		prb.pos.x = 0;
		prb.pos.y = -prb.size.y/2;
		break;
	case 5:
		prb.pos.x = -prb.size.x;
		prb.pos.y = -prb.size.y/2;
		break;
	case 6:
		prb.pos.x = 0;
		prb.pos.y = -prb.size.y;
		break;
	case 7:
		prb.pos.x = -prb.size.x/2;
		prb.pos.y = -prb.size.y;
		break;
	case 8:
		prb.pos.x = -prb.size.x;
		prb.pos.y = -prb.size.y;
		break;
	}
	/*
	**	Now rotate the pixrect if necessary
	*/
	if (angle != 0 && rotate_pixrect(&ge->ge_pr, &prb, angle)) {
		free((char *)ge);
		no_more_memory();
		return;
	}
	/*
	**	fill out the structure
	*/
	g_next->ge_next = ge;
	g_next = ge;
	ge->ge_index = graphic_index;
	ge->ge_flags = GE_RECT;
	ge->ge_color = graphic_color >> cshift;
	/*
	**	determine origin
	*/
	xloc += prb.pos.x;
	yloc += prb.pos.y;
	/*
	**	write to screen
	*/
	if (graphic_sw[graphic_index] == 1) {
		if (depth == 1)
			pw_write(GIPSWIN(image_pw), xloc, yloc,
				prb.size.x, prb.size.y,
				PIX_SRC | PIX_DST, ge->ge_pr, 0, 0);
		else
			pw_stencil(GIPSWIN(image_pw), xloc, yloc,
				prb.size.x, prb.size.y,
				PIX_SRC | PIX_COLOR(graphic_color),
				ge->ge_pr, 0, 0, NULL, 0, 0);
	}
	/*
	**	If not graphic 0, reset the origin
	*/
	if (graphic_index != 0) {
		xloc += (xorg+xoff)*xzoom - image_rect.r_left;
		yloc += (yorg+yoff)*yzoom - image_rect.r_top;
	}
	ge->ge_pos.x = xloc;
	ge->ge_pos.y = yloc;
	ge->ge_lim.x = prb.size.x;
	ge->ge_lim.y = prb.size.y;
}
	/*
	**	Routine to rotate pixrect
	*/
static
rotate_pixrect(pr, prb, angle)
	Pixrect	**pr;			/* pixrect to be rotated */
	struct	pr_subregion *prb;	/* origin and size */
	int	angle;			/* the angle */
{
	Pixrect	*pr_new;
	struct	pr_subregion prb_new;
	double	d, sin(), cos(), atan();
	int	y, xx, yy;
	register x, xxx, yyy, cos_a, sin_a;
#define	x_rot(x,y)	(( cos_a * (x) + sin_a * (y)) >> 16)
#define	y_rot(x,y)	((-sin_a * (x) + cos_a * (y)) >> 16)
	/*
	**	Find size of rotated pixrect
	*/
	d = angle * atan(1.0) / 45.0;
	cos_a = 65536.0 * cos(d);
	sin_a = 65536.0 * sin(d);
	x = x_rot(prb->size.x, 0);
	prb_new.pos.x = MIN(0, x);
	prb_new.size.x = MAX(0, x);
	y = y_rot(prb->size.x, 0);
	prb_new.pos.y = MIN(0, y);
	prb_new.size.y = MAX(0, y);
	x = x_rot(0, prb->size.y);
	prb_new.pos.x = MIN(prb_new.pos.x, x);
	prb_new.size.x = MAX(prb_new.size.x, x);
	y = y_rot(0, prb->size.y);
	prb_new.pos.y = MIN(prb_new.pos.y, y);
	prb_new.size.y = MAX(prb_new.size.y, y);
	x = x_rot(prb->size.x, prb->size.y);
	prb_new.pos.x = MIN(prb_new.pos.x, x);
	prb_new.size.x = MAX(prb_new.size.x, x);
	y = y_rot(prb->size.x, prb->size.y);
	prb_new.pos.y = MIN(prb_new.pos.y, y);
	prb_new.size.y = MAX(prb_new.size.y, y);
	/*
	**	Allocate the new pixrect
	*/
	prb_new.size.x -= prb_new.pos.x;
	prb_new.size.y -= prb_new.pos.y;
	if ((pr_new = mem_create(prb_new.size.x, prb_new.size.y, 1)) == NULL)
		return 1;
	/*
	**	Project the old pixrect, pixel by pixel
	*/
	xx = (-prb_new.pos.x) << 16;
	yy = (-prb_new.pos.y) << 16;
	for (y = 0; y < prb->size.y; y++, xx += sin_a, yy += cos_a)
		for (x = 0, xxx = xx, yyy = yy; x < prb->size.x; x++) {
			if (pr_get(*pr, x, y))
				(void)pr_put(pr_new, xxx >> 16, yyy >> 16, 1);
			xxx += cos_a;
			yyy -= sin_a;
		}
	/*
	**	Rotate the pixrect origin address
	*/
	prb_new.pos.x = x_rot(prb->pos.x + prb->size.x/2,
		prb->pos.y + prb->size.y/2) - prb_new.size.x/2;
	prb_new.pos.y = y_rot(prb->pos.x + prb->size.x/2,
		prb->pos.y + prb->size.y/2) - prb_new.size.y/2;
	/*
	**	Remove the non-rotated pixrect
	**	and return the new pointer and sub-region
	*/
	(void)pr_close(*pr);
	*pr = pr_new;
	*prb = prb_new;
	return 0;
}
