#ifndef lint
static char *sccsid = "%Z%$RCSfile: header.c,v $	GIPS $Revision: 3.7 $ (MIT/CSR) $Date: 1998/05/13 01:12:21 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		header.c	Creation Date:	03/19/86
**	Version:	$Revision: 3.7 $		Update:		$Date: 1998/05/13 01:12:21 $
**	Components:	open_gips_socket()	open socket to mongo
**			write_subimage()	write GIPS sub-image
**			write_histogram()	write histogram for mongo
**			write_line_trace()	write line track for mongo
**			copy_header_variable()	copy GIPS string variable
**
**	Function:	Write a sub-image to disk or socket
*/
#define		GRID
#include	"gipstool.h"
#include	<sys/socket.h>
#include	<sys/un.h>
	/*
	**	gips header items needing to be saved
	*/
#define		H_FOUND		01	/* item present in header */
#define		H_CHAR		02	/* character-string item */
#define		H_DEL		04	/* delete if found */
#define		H_NUM		00	/* numeric item */
static	struct	hvar	{
	char	*h_title;		/* header item name */
	int	h_flag;			/* =1 if item found */
	double	h_dval;			/* numeric item value */
	char	*h_cval;		/* character string value */
} headers[] = {
	{ "dtype",	H_NUM },	/* Pixel length */
	{ "dscale",	H_NUM },	/* Type of data mapping */
	{ "dnum",	H_NUM },	/* Number of data components */
	{ "dmax",	H_NUM },	/* Maximum float data value */
	{ "dmin",	H_NUM },	/* Minimum float data value */
	{ "dfile",	H_CHAR|H_DEL },	/* file containing byte-pixel mapping */
	{ "dpoly",	H_NUM|H_DEL },	/* Order of polynomial dtype=p */
	{ "dparm",	H_CHAR|H_DEL },	/* Polynomial coefficients */
	{ "dnull",	H_NUM },	/* Data value denoting none available */
	{ "dfact",	H_NUM|H_DEL },	/* Pixel scaling factor for dscale=n */
	{ "dzero",	H_NUM|H_DEL },	/* Pixel baseline value for dscale=n */
	{ "geometry",	H_CHAR|H_DEL },	/* Cellular format */
	{ "mapping",	H_CHAR },	/* Coordinate projection */
	{ "xscale",	H_NUM },	/* Type of x-axis mapping */
	{ "xnum",	H_NUM },	/* Number of x-axis pixels. */
	{ "xorg",	H_NUM|H_DEL },	/* Lower pixel X-limit */
	{ "xlim",	H_NUM|H_DEL },	/* Upper displayed pixel X-limit */
	{ "xmax",	H_NUM },	/* Maximum float x-axis value */
	{ "xmin",	H_NUM },	/* Minimum float x-axis value */
	{ "xfile",	H_CHAR|H_DEL },	/* File for xscale=f mapping. */
	{ "xpoly",	H_NUM|H_DEL },	/* Order of polynomial xscale=p */
	{ "xparm",	H_CHAR|H_DEL },	/* Polynomial coefficients xscale=p */
	{ "yscale",	H_NUM },	/* Type of y-axis mapping */
	{ "ynum",	H_NUM },	/* Number of y-axis pixels. */
	{ "yorg",	H_NUM|H_DEL },	/* Lower pixel Y-limit */
	{ "ylim",	H_NUM|H_DEL },	/* Upper displayed pixel Y-limit */
	{ "ymax",	H_NUM },	/* Maximum float y-axis value */
	{ "ymin",	H_NUM },	/* Minimum float y-axis value */
	{ "yfile",	H_CHAR|H_DEL },	/* File for yscale=f mapping. */
	{ "ypoly",	H_NUM|H_DEL },	/* Order of polynomial yscale=p */
	{ "yparm",	H_CHAR|H_DEL },	/* Polynomial coefficients yscale=p */
	NULL
};
	/*
	**	Socket to GipsMongo
	*/
/*static	struct	sockaddr_un	gips_su	= { AF_UNIX, GIPSSOCKET }; */
static	write_subimage_2();
	/*
	**	open a client socket to gipsmongo
	*/
static	FILE *
open_gips_socket()
{
	register int	s;		/* stream socket */
	register FILE	*fp;		/* stream I/O pointer */

static	struct	sockaddr_un	gips_su	= { AF_UNIX, GIPSSOCKET };

	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1 ||
		connect(s, (struct sockaddr *)&gips_su,
			strlen(gips_su.sun_path)+2) == -1 ||
		(fp = fdopen(s, "w")) == NULL) {
		display_system_message(gips_su.sun_path);
		return NULL;
	} else
		return fp;
}
	/*
	**	write a sub-image to disk or socket
	*/
write_subimage(name, locx, locy)
	char	*name;			/* output file name */
	int	locx, locy;		/* cursor address */
{
	char	buf[BUFLEN];		/* scratch */
	int	nx[2], ny[2], nd[2];	/* scratch */
	double	dx[2], dy[2], dd[2];	/* scratch */
	double	dfact;			/* scratch */
	char	dscale;			/* input dscale value */
	register struct hvar *h;	/* pointer to header names */
	register x, y;			/* scratch */
	/*
	**	save pixel coordinate and realworld limits
	*/
	if (ihget("dscale", &dscale, 1))
		dscale = 's';
	x = X_RECT_TO_IMAGE(locx);
	nx[0] = MIN(x, image_mark[0]);
	nx[0] = MAX(nx[0], xorg);
	nx[1] = MAX(x, image_mark[0])+ 1;
	nx[1] = MIN(nx[1], xorg+xdim);
	(void) egips(nx, dx, 2, EGIPS_X_VAL);
	y = Y_RECT_TO_IMAGE(locy);
	ny[0] = MIN(y, image_mark[1]);
	ny[0] = MAX(ny[0], yorg);
	ny[1] = MAX(y, image_mark[1])+ 1;
	ny[1] = MIN(ny[1], yorg+ydim);
	(void) egips(ny, dy, 2, EGIPS_Y_VAL);
	nd[0] = 0;
	nd[1] = 256;
	(void) egips(nd, dd, 2, EGIPS_D_VAL);
	dfact = (dd[1] - dd[0])/256;
	nd[0] = nx[1] - nx[0];
	nd[1] = ny[1] - ny[0];
	/*
	**	save some header variables, delete others
	*/
	for (h = headers; h->h_title; h++)
		if (! ihget(h->h_title, (h->h_flag & H_CHAR) ?
				buf : (char *)&h->h_dval, BUFLEN-1)) {
			h->h_flag |= H_FOUND;
			if (h->h_flag & H_DEL)
				(void) ihdel(h->h_title);
			if (h->h_flag & H_CHAR) {
				h->h_cval = alloc(1, strlen(buf)+1);
				if (h->h_cval == NULL) {
					no_more_memory();
					return;
				}
				(void) strcpy(h->h_cval, buf);
			}
		}
	/*
	**	Fix header variables
	*/
	(void) ihput("xnum", (char *)&nd[0]);
	(void) ihput("xscale", "s");
	(void) ihput("xmin", (char *)&dx[0]);
	(void) ihput("xmax", (char *)&dx[1]);
	(void) ihput("ynum", (char *)&nd[1]);
	(void) ihput("yscale", "s");
	(void) ihput("ymin", (char *)&dy[0]);
	(void) ihput("ymax", (char *)&dy[1]);
	(void) ihput("dmin", (char *)&dd[0]);
	(void) ihput("dmax", (char *)&dd[1]);
	/*
	**	Compute index limits within the pixrects
	*/
	nx[0] -= xorg;
	nx[1] -= xorg;
	ny[0] -= yorg;
	ny[1] -= yorg;
	if (depth != 1) {
		nx[0] *= xzoom;
		nx[1] *= xzoom;
		ny[0] *= yzoom;
		ny[1] *= yzoom;
	}
	/*
	**	Handle maping projection
	*/
	if (ihget("mapping", buf, sizeof(buf)) == 0) {
		char	name[32];
		int	type, xyrot = 0;
		double	org[2], radius, e[3], lat[2];
		(void)sscanf(buf, GIPS_MAPPING, name, &type, &org[0], &org[1],
			&radius, &e[0], &e[1], &e[2], &lat[0], &lat[1], &xyrot);
		radius *= xzoom;
		org[0] = (org[0] - xorg - xoff) * xzoom - nx[0];
		org[1] = (org[1] - yorg - xoff) * yzoom - ny[0];
		(void)sprintf(buf, GIPS_MAPPING_P, name, type, org[0], org[1],
			radius, e[0], e[1], e[2], lat[0], lat[1], xyrot);
		ihput("mapping", buf);
	}
	/*
	**	Write out
	*/
	if (dscale == 'l') {
		double	*dbuf;
		x = sizeof(double) + sizeof(float) + sizeof(int);
		if (dbuf = (double *)alloc(x, nd[0])) {
			(void) ihput("dtype", "f");
			(void) ihput("dscale", "s");
			write_subimage_2(name, nx, ny, nd[0], dbuf);
			(void)free((char *)dbuf);
		}
	} else {
		(void) ihput("dtype", "b");
		(void) ihput("dscale", "n");
		(void) ihput("dzero", (char *)&dd[0]);
		(void) ihput("dfact", (char *)&dfact);
		write_subimage_2(name, nx, ny, nd[0], (double *)NULL);
	}
	/*
	**	Replace altered header variables
	*/
	for (h = headers; h->h_title; h++)
		if (h->h_flag & H_FOUND) {
			h->h_flag &= ~H_FOUND;
			(void)ihput(h->h_title, (h->h_flag & H_CHAR) ?
				h->h_cval : (char *)&h->h_dval);
			if (h->h_flag & H_CHAR)
				free(h->h_cval);
		}
}
	/*
	**	Write the subimage pixels
	*/
static
write_subimage_2(name, nx, ny, xout, dbuf)
	char	*name;			/* output file name */
	int	nx[2];			/* x-pixel limits */
	int	ny[2];			/* y-pixel limits */
	int	xout;			/* number of output x-pixels */
	double	*dbuf;			/* dscale=l conversion buffer */
{
	register FILE *fp;		/* output file pointer */
	int	*ibuf, *ii;		/* dscale=l pixel value buffer */
	float	*fbuf;			/* dscale=l output buffer */
	register x, y, i;		/* scratch */
	/*
	**	Open the output stream
	*/
	if ( ! *name ) {
		(void)fclose(stdout);
		if ((fp = open_gips_socket()) == NULL)
			return;
		name = "-";
	}
	if ((fp = ihwrit(name, 1)) == NULL) {
		GIPS_ERROR;
		return;
	}
	if (dbuf != NULL) {
		fbuf = (float *)&dbuf[xout];
		ii = ibuf = (int *)&fbuf[xout];
	}
	/*
	**	Write the image
	*/
	for (y = ny[0]; y < ny[1]; y += (depth == 1) ? 1 : yzoom) {
		for (x = nx[0]; x < nx[1]; ) {
			if (depth == 1)
				i = pr_get(data_pr, x++, y) & 0377;
			else {
				i = (pr_get(saved_pr, x, y) << cshift) & 0377;
				x += xzoom;
			}
			if (dbuf != NULL)
				*ii++ = i;
			else
				(void) putc(i, fp);
		}
		if (dbuf != NULL) {
			if (egips(ii = ibuf, dbuf, xout, EGIPS_D_VAL)) {
				gipserr();
				break;
			}
			(void) gconv((char *)fbuf, (char *)dbuf,
				xout, CVT_F, CVT_D);
			if (fwrite((char *)fbuf, sizeof(float), xout, fp)
					!= xout) {
				gipserr2(GIPSBADWRITE, name);
				break;
			}
		}
	}
	(void)fclose(fp);
}
	/*
	**	Write out a sub-image histogram
	*/
write_histogram(c, locx, locy, sw)
	char	*c;			/* name of output file */
	int	locx, locy;		/* cursor location */
	int	sw;			/* =1 for histogram of full image */
{
	register FILE	*fp, *fq;	/* output/input stream pointers */
	float		out[7][256];	/* mongo input record */
	double		pixels[256];	/* realworld pixel values */
	char		buf[BUFLEN];	/* scratch */
	int		hist[256];	/* image histogram */
	register int	x, n, y;	/* scratch */
	/*
	**	Open file or interprocess socket
	*/
	if (clean_whitespace(c)) {
		if ((fp = fopen(c, "w")) == NULL) {
			display_system_message(c);
			return;
		}
	} else if ((fp = open_gips_socket()) == NULL)
		return;
	/*
	**	Open the default gips header file
	*/
	(void) sprintf(buf, "%s%s", GIPSPROFDIR, "GIPSinit");
	if ((fq = fopen(buf, "r")) == NULL) {
		display_system_message(buf);
		(void) fclose(fp);
		return;
	}
	/*
	**	Write non-default header stuff
	*/
	(void)fprintf(fp, "xnum\t = 256\nynum\t = 7\ndtype\t = f\n");
	copy_header_variable(fp, "title", "title");
	copy_header_variable(fp, "dtitle", "xtitle");
	copy_header_variable(fp, "owner", "owner");
	copy_header_variable(fp, "origin", "origin");
	copy_header_variable(fp, "date", "date");
	copy_header_variable(fp, "time", "time");
	/*
	**	Append the default header file
	*/
	while ((n = getc(fq)) != EOF)
		(void)putc(n, fp);
	(void) fclose(fq);
	/*
	**	Construct the histogram
	*/
	if (sw)
		move((char *)image_hist, (char *)hist, sizeof(image_hist));
	else {
		int	nx[2], ny[2];

		clear((char *)hist, sizeof(hist));

		x = (depth == 1) ? 1 : xzoom;
		locx = X_RECT_TO_IMAGE(locx);
		nx[0] = MIN(locx, image_mark[0])-xorg;
		nx[0] = MAX(nx[0], 0)*x;
		nx[1] = MAX(locx, image_mark[0])-xorg+1;
		nx[1] = MIN(nx[1], xdim)*x;

		y = (depth == 1) ? 1 : yzoom;
		locy = Y_RECT_TO_IMAGE(locy);
		ny[0] = MIN(locy, image_mark[1])-yorg;
		ny[0] = MAX(ny[0], 0)*y;
		ny[1] = MAX(locy, image_mark[1])-yorg+1;
		ny[1] = MIN(ny[1], ydim)*y;

		for (y = ny[0]; y < ny[1]; y += (depth == 1) ? 1 : yzoom)
			for (x = nx[0]; x < nx[1]; hist[n & 0377]++)
				if (depth == 1)
					n = pr_get(data_pr, x++, y);
				else {
					n = pr_get(saved_pr, x, y) << cshift;
					x += xzoom;
				}
	}
	/*
	**	Construct the histogram
	*/
	(void)egips((int *)NULL, pixels, 256, EGIPS_D_VAL);
	for (x = y = 0; y < 256; y++) {
		n = y >> cshift;
		out[0][y] = y ? (x += hist[y]) : 0;
		out[1][y] = y ? hist[y] : 0;
		out[2][y] = newrgb[2][n];
		out[3][y] = newrgb[1][n];
		out[4][y] = newrgb[0][n];
		out[5][y] = pixels[y];
		out[6][y] = y;
	}
	if (fwrite((char *)out, sizeof(out), 1, fp) != 1) 
		display_system_message(*c ? c : "'socket'");
	(void) fclose(fp);
}
	/*
	**	Write out a line-track image
	*/
write_line_trace(c, locx, locy)
	char	*c;			/* name of output file */
	int	locx, locy;		/* cursor location */
{
	register FILE	*fp, *fq;	/* output/input stream pointers */
	float		*out;		/* mongo input record */
	char		buf[BUFLEN];	/* scratch */
	double		d;		/* scratch */
	double		lonlat[2];	/* scratch */
	int		x, x0, x1, dx;	/* x indices */
	int		y, y0, y1, dy;	/* y indices */
	int		i;		/* scratch */
	register float	*o;		/* scratch */
	register	n, sw, npix;	/* scratch */
	/*
	**	Open file or interprocess socket
	*/
	if (clean_whitespace(c)) {
		if ((fp = fopen(c, "w")) == NULL) {
			display_system_message(c);
			return;
		}
	} else if ((fp = open_gips_socket()) == NULL)
		return;
	/*
	**	Open the default gips header file
	*/
	(void) sprintf(buf, "%s%s", GIPSPROFDIR, "GIPSinit");
	if ((fq = fopen(buf, "r")) == NULL) {
		display_system_message(buf);
		(void) fclose(fp);
		return;
	}
	/*
	**	Get line length
	*/
	locx = X_RECT_TO_IMAGE(locx);
	x = x0 = MIN(MAX(image_mark[0]-xorg, 0), xdim);
	x1 = MIN(MAX(locx-xorg, 0), xdim);
	dx = (x1 > x0) ? 1 : -1;

	locy = Y_RECT_TO_IMAGE(locy);
	y = y0 = MIN(MAX(image_mark[1]-yorg, 0), ydim);
	y1 = MIN(MAX(locy-yorg, 0), ydim);
	dy = (y1 > y0) ? 1 : -1;

	if (abs(y0-y1) > abs(x0-x1)) {
		npix = abs(y0-y1) + 1;
		sw = 1;
	} else {
		npix = abs(x0-x1) + 1;
		sw = 0;
	}
	/*
	**	Allocate array
	*/
	o = out = (float *)alloc(sizeof(float), 7*npix);
	/*
	**	Fill the array
	*/
	for (n = 0; n < npix; n++, x += dx, y += dy) {
		o[6*npix] = n;
		if (sw)
			x = (x0*(npix-n)+x1*n)/npix;
		else
			y = (y0*(npix-n)+y1*n)/npix;
		if (! map_sw) {
			i = x + xorg;
			(void)egips(&i, &lonlat[0], 1, EGIPS_X_VAL);
			if (lonlat[0] <= GIPSLOW || lonlat[0] >= GIPSHIGH)
				continue;
			i = y + yorg;
			(void)egips(&i, &lonlat[1], 1, EGIPS_Y_VAL);
			if (lonlat[1] <= GIPSLOW || lonlat[1] >= GIPSHIGH)
				continue;
		} else if (g_unmap(x + xorg, y + yorg, lonlat))
			continue;
		o[5*npix] = x + xorg;
		o[4*npix] = lonlat[0];
		o[3*npix] = y + yorg;
		o[2*npix] = lonlat[1];
		if (depth == 1)
			i = pr_get(data_pr, x, y);
		else
			i = pr_get(saved_pr, xzoom*x, yzoom*y) << cshift;
		i &= 0377;
		o[npix] = i;
		(void)egips(&i, &d, 1, EGIPS_D_VAL);
		*o++ = d;
	}
	/*
	**	Determine whether any line is to be written
	*/
	if (n = o - out) {
		/*
		**	Write non-default header stuff
		*/
		(void)fprintf(fp, "xnum\t = %d\nynum\t = 7\ndtype\t = f\n", n);
		copy_header_variable(fp, "title", "title");
		copy_header_variable(fp, "dtitle", "ytitle");
		copy_header_variable(fp, "owner", "owner");
		copy_header_variable(fp, "origin", "origin");
		copy_header_variable(fp, "date", "date");
		copy_header_variable(fp, "time", "time");
		/*
		**	Append the default header file
		*/
		while ((i = getc(fq)) != EOF)
			(void)putc(i, fp);
		/*
		**	Write the array
		*/
		for (o = out, i = 0; i < 7; i++, o += npix)
			if (fwrite((char *)o, sizeof(float), n, fp) != n) 
			display_system_message(*c ? c : "'socket'");
	} else {
		display_error_message("No points to plot", "");
	}
	free((char *)out);
	(void) fclose(fp);
	(void) fclose(fq);
}
	/*
	**	If header variable present, copy to stream
	*/
copy_header_variable(fp, name, newname)
	register FILE	*fp;		/* output stream */
	register char	*name;		/* header variable name */
	register char	*newname;	/* new output name */
{
	char		buf[BUFLEN];	/* scratch */
	register char	*c;		/* scratch */

	if (ihget(name, c = buf, BUFLEN-1))
		return;
	(void)fprintf(fp, "%s\t= \"", newname);
	while (*c) {
		if (*c == '"')
			(void)putc('\\', fp);
		(void)putc(*c++, fp);
	}
	(void)fprintf(fp, "\"\n");
}
