#ifndef lint
static char *sccsid = "%Z%$RCSfile: info.c,v $	GIPS $Revision: 3.6 $ (MIT/CSR) $Date: 1998/11/13 23:33:06 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1989 Massachusetts Institute of Technology
**
**	Module:		info.c			Creation Date:	07/19/89
**	Version:	$Revision: 3.6 $		Update:		$Date: 1998/11/13 23:33:06 $
**	Components:	info_events		menu handler
**			set_info_panel		create line of display
**			display_info		create info window
**
**	Function:	Display information window
*/
#include	"gipstool.h"
#include	"selections.h"

#define		FONT_MASK	0x8000	/* flag to specify font type */
	/*
	**	Sunview info popup structures
	*/
static	Frame		info_frame;	/* describes the info window */
static	Panel		info_panel;	/* describes the info panel */
static	Panel_item	info_item[64];	/* describes the info items */
static	int		info_line;	/* line counter */
static	void		(*info_proc)();	/* panel event handler */
	/*
	**	Event handler to pass menu button
	*/
static
handle_info_event(canvas, event, arg)
	Canvas	canvas;
	Event	*event;
	caddr_t	arg;
{
	switch (event_id(event)) {
	case MS_LEFT:
		info_proc(canvas, event, arg);
		break;
	case MS_MIDDLE:
		break;
	case MS_RIGHT:
		(void)menu_show((Menu)window_get(
			info_frame, WIN_MENU), canvas, event, 0);
		break;
	default:
		if (event_is_ascii(event) && event_is_down(event))
			image_selected(image_sw, event, 0);
	}
}
	/*
	**	Panel button handler
	*/
static
handle_info_item(item, event)
	Panel_item	item;
	Event		*event;
{
	int	id;
	Rect	r;

	if (event_id(event) != MS_LEFT)
		return;
	id = (int)panel_get(item, PANEL_CLIENT_DATA);
	if (id & FONT_MASK) {
		event_set_id(event, 'F');
		image_selected(image_sw, event, 0);
		event_set_id(event, id & ~FONT_MASK);
		image_selected(image_sw, event, 0);
		id = ' ';
	}
	event_set_id(event, id);
	image_selected(image_sw, event, 0);
}
	/*
	**	Create a pair of panel messges on one line
	*/
static
set_info_panel(str1, value, str2)
	int	value;
	char	*str1, *str2;
{
	static	int	width;

	if (! info_frame) {
		info_frame = window_create(base_frame, FRAME,
			WIN_SHOW,		FALSE,
			FRAME_SHOW_LABEL,	TRUE,
			WIN_RIGHT_OF,		base_frame,
			0);
		info_panel = window_create(info_frame, PANEL, 0);
		info_proc = (void (*)())window_get(info_panel, WIN_EVENT_PROC);
		window_set(info_panel, WIN_EVENT_PROC, handle_info_event, 0);
		width = (panel_button_image(info_panel, "", 2, 0))->pr_size.x;
	}
	if (! info_item[info_line]) {
		if (value) {
			char buf[3], *b = buf;
			if (value & FONT_MASK)
				(void)sprintf(buf, "F%c", value & ~FONT_MASK);
			else
				b = form_char(value);
			(void)panel_create_item(info_panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	handle_info_item,
			PANEL_ITEM_X,		10,
			PANEL_ITEM_Y,		ATTR_ROW(info_line)-4,
			PANEL_LABEL_IMAGE,	panel_button_image(info_panel,
						b, 2, 0),
			PANEL_CLIENT_DATA,	(caddr_t)value,
			0);
		}
		(void)panel_create_item(info_panel, PANEL_MESSAGE,
			PANEL_ITEM_X,		10 + width + 10,
			PANEL_ITEM_Y,		ATTR_ROW(info_line),
			PANEL_LABEL_STRING,	str1,
			0);
		info_item[info_line] = panel_create_item(info_panel,
			PANEL_MESSAGE,
			PANEL_ITEM_X,		20 + width + ATTR_COL(20),
			PANEL_ITEM_Y,		ATTR_ROW(info_line),
			0);
	}
	panel_set(info_item[info_line++],
		PANEL_LABEL_STRING,	str2 ? str2 : "none",
		0);
}
	/*
	**	Return file basename
	*/
static	char	*
basename(name)
	char	*name;
{
	char	*s, *rindex();

	if (name && (s = rindex(name, '/')))
		name = s+1;
	return name;
}
	/*
	**	Display image information
	*/
display_info()
{
	register i, n;
	int	create;
	static	char *title[] = { "Roman", "Italic", "Bold", "Symbol" };
	char	buf[256], *s, *label;
	Rect	r;

	if (no_display)
		return;
	create = info_frame == NULL;
	info_line = 0;
	set_info_panel("Frame title:", SET_FRAME_LABEL,
		label = (char *)window_get(base_frame, FRAME_LABEL));
	set_info_panel("Program version:", 0,
		index(TOOL_NAME(3.2), '%') ? "test" : TOOL_NAME(3.2));
	set_info_panel("Original image:", 0,
		basename(imfile));
	(void)sprintf(buf, "%dx%d pixels", xnum, ynum);
	set_info_panel("Original size:", 0, buf);
	(void)sprintf(buf, "%dx%d pixels", xdim, ydim);
	set_info_panel("In-core size:", 0, buf);
	(void)sprintf(buf, "%d,%d", xorg, yorg);
	set_info_panel("In-core origin:", 0, buf);
	(void)sprintf(buf, "%dx%d pixels",
		image_rect.r_width, image_rect.r_height);
	set_info_panel("Display size:", SET_BOTTOM_RIGHT, buf);
	(void)sprintf(buf, "%d,%d", xoff, yoff);
	set_info_panel("Display origin:", 0, buf);
	(void)sprintf(buf, "%d,%d", image_rect.r_left, image_rect.r_top);
	set_info_panel("Display margins:", SET_TOP_LEFT, buf);
	(void)sprintf(buf, "%d,%d", image_mark[0], image_mark[1]);
	set_info_panel("Image mark:", SET_LINE_MARK, buf);
	set_info_panel("Colormap name:", 0, depth > 1 ? colormap : NULL);
	set_info_panel("Colormap file:", SET_COLORMAP, depth > 1 ?
		basename(afile) : NULL);
	(void)sprintf(buf, "%d",stretch_percent);
	set_info_panel("Color stretch:", SET_COLORMAP_STRETCH, buf);
	(void)sprintf(buf, "%.*g - %.*g",
			precision, stretch_min, precision, stretch_max);
	set_info_panel("Linear stretch:", SET_COLORMAP_LINSTRETCH, buf);
	(void)sprintf(buf, "%d", graphic_index);
	set_info_panel("Graphic index:", ADD_GRAPHIC, buf);
	(void)sprintf(buf, "%d", brush_size);
	set_info_panel("Brush size:", SET_PAINT_SIZE, buf);
	(void)sprintf(buf, "%d", graphic_color);
	set_info_panel("Graphic color:", SET_TEXT_COLOR, buf);
	(void)sprintf(buf, "%d degrees", graphic_angle);
	set_info_panel("Graphic angle", SET_TEXT_ANGLE, buf);
	for (n = 0; n < 2; n++) {
		buf[0] = 0;
		for (i = 0; i < NGRAPHICS; i++)
			if ((n == 0 && graphic_sw[i] == 1) ||
					(n == 1 && graphic_sw[i] > 1))
				(void)sprintf(buf+strlen(buf), "%d ", i);
		set_info_panel(n ? "Indices `off':" : "Indices `on': ",
			ADD_GRAPHIC, buf[0] ? buf : (char *)NULL);
	}
	for (i = 0; i < 4; i++) {
		(void)sprintf(buf, "%s font:", title[i]);
		set_info_panel(buf, FONT_MASK | title[i][0],
			basename(setauxfont(0, i)));
	}
	(void)sprintf(buf, "%d%s", graphic_origin,
		rotated_font ? " (rotated)" : "");
	set_info_panel("Text orientation:", SET_TEXT_ORIGIN, buf);
	set_info_panel("Command log file:", LOG_INPUT_COMMANDS,
		ofile && *ofile ? basename(ofile) : NULL);
	set_info_panel("Pixel log file:", BEGIN_PIXEL_LOGGING,
		pixel_stream ? basename(pixel_log_file) : NULL);
	set_info_panel("Tag file:", READ_TAG_FILE,
		basename(tag_file));
	(void)sprintf(buf, "[%s] %s", command_text[i-1], left_button_value+1);
	set_info_panel("Left button:", SET_LEFT_BUTTON,
		(i = command_index[left_button_value[0] & 0377]) ?
			buf : "unassigned");
	set_info_panel("Middle button:", 0,
		"display zoomed insert");
	set_info_panel("Right button:", 0,
		"display pop-up menus");

	window_fit(info_panel);
	window_fit(info_frame);

	if (create) {
		r = *((Rect *)window_get(base_frame, WIN_RECT));
		i = (int)window_get(info_frame, WIN_WIDTH);
		if (r.r_left + r.r_width + i > display_rect.r_width)
			window_set(info_frame,
				WIN_X, display_rect.r_width - i - r.r_left,
				0);
	}

	window_set(info_frame,
		FRAME_LABEL,	label,
		WIN_SHOW,	TRUE,
		0);
}

update_info()
{
	if (info_frame && (int)window_get(info_frame, WIN_SHOW))
		display_info();
}
