#ifndef lint
static char *sccsid = "%Z%$RCSfile: main.c,v $	GIPS $Revision: 3.7 $ (MIT/CSR) $Date: 1992/01/02 18:58:51 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		main.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.7 $		Update:		$Date: 1992/01/02 18:58:51 $
**	Function:	Display GIPS image within customized Sun window
**	Components:	main()		GIPS/UNIX SunTool main program
*/
#define		MAIN
#include	"gipstool.h"
#include	<signal.h>
	/*
	**	Pointers to compound input variables
	*/
static	char	*xbuf;		/* input pixel width, origin, offset */
static	char	*ybuf;		/* input pixel width, origin, offset */
static	char	*xview;		/* output window size, left-right border */
static	char	*yview;		/* output window size, left-right border */
static	int	nring;		/* number of ring buffers for dtype=G */
static	flag	Kflag;		/* run as zombie */
static	flag	Nflag;		/* use unshared colormap (colors=256) */
	/*
	**	table used by gipsflags()
	*/
static	flagentry	flagtable[] = {
	"",	"%d depth",		(flag *) &depth,
	"C",	"-> map",		(flag *) &colormap,
	"I",	"-> file",		(flag *) &gipsrc,
	"K",	"1",			(flag *) &Kflag,
	"N",	"1",			(flag *) &Nflag,
	"S",	"1",			(flag *) &Sflag,
	"X",	"-> size,left,right",	(flag *) &xview,
	"Y",	"-> size,left,right",	(flag *) &yview,
	"a",	"-> file",		(flag *) &afile,
	"c",	"1",			(flag *) &cflag,
	"e",	"-> coeffs",		(flag *) &ecoeffs,
	"i",	"-> file",		(flag *) &ifile,
	"m",	"-> file",		(flag *) &mfile,
	"n",	"%d colors",		(flag *) &colors,
	"o",	"-> file",		(flag *) &ofile,
	"r",	"%d ringbufs",		(flag *) &nring,
	"x",	"-> size,org,off",	(flag *) &xbuf,
	"y",	"-> size,org,off",	(flag *) &ybuf,
	"z",	"%d xzoom",		(flag *) &xzoom,
	"Z",	"%d yzoom",		(flag *) &yzoom,
	NULL,	"[file]",		NULL
};
	/*
	**	The tool icon
	*/
static	short	image_icon[] = {
#include "gipstool.icon"
};
DEFINE_ICON_FROM_IMAGE(icon, image_icon);

static	Rect	icon_rect = { 0, 48, 64, 16 };

#include "cursor.icon"

	/*
	**	Start here
	*/
main(argc, argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fm;		/* input GIPS merge-header file stream */
	char	buf[BUFLEN];	/* scratch */
	char	*icon_name;	/* title for icon */
	int	width;		/* initial width of image sub-window */
	int	height;		/* initial height of image sub-window */
	int	nolocate = 0;	/* =1 if user specifies window location */
	short	rborder	= 0;	/* right viewport border */
	short	bborder	= 0;	/* bottom viewport border */
	char	*c;		/* scratch */
	int	i;		/* scratch */
	/*
	**	Check for window placement override
	*/
	for (i = 1; i < argc; i++)
		if (argv[i] && strncmp(argv[i], "-Wp", 3) == 0)
			nolocate++;
	/*
	**	parse tool-specific arguments
	*/
	base_frame = window_create(NULL, FRAME,
		FRAME_INHERIT_COLORS,		TRUE,
		FRAME_SUBWINDOWS_ADJUSTABLE,	TRUE,
		FRAME_ARGC_PTR_ARGV,		&argc, argv,
		WIN_ERROR_MSG,			"Unable to create window.",
		0);
	/*
	**	save and parse callers arguments
	*/
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...", argv[2]);
	if (argc == 2)
		imfile = argv[1];
	if (depth != 0 && depth != 1 && depth != 8)
		error("bad pixel depth: %d", (char *)depth);
	if (Sflag)
		telltale(0);
	/*
	**	do we run as a zombie?
	*/
	if (Kflag && fork()) {
		i = getdtablesize();
		while (i >= 0)
			(void)close(i--);
		exit(0);
	}
	/*
	**	read input GIPS header and possible merge file
	*/
	if ((fp = ihread(imfile, GIPSPROFILE)) == NULL)
		gipserr();
	if (*mfile != '-')
		if ((fm = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void) ihbcl(fm);
	/*
	**	If we are reading the standard input,
	**	get the title from the GIPS header.
	**	Then construct a short name for the icon
	*/
	if (fp == stdin) {
		imfile = save_string(ihget("title", buf, BUFLEN-1) ?
			"'stdin'" : buf, (char *)NULL);
		icon_name = "GipsTool";
	} else if (icon_name = rindex(imfile, '/'))
		icon_name++;
	else
		icon_name = imfile;
	icon_set(&icon, ICON_LABEL, icon_name, ICON_LABEL_RECT, &icon_rect, 0);
	window_set(base_frame, FRAME_ICON, &icon, 0);
	/*
	**	If input image is seek-able, record its location
	*/
	if (cflag && (strncmp(imfile, "/dev/", 5) == 0 ||
		(image_seek = ftell(fp)) <= 0 || ferror(fp)))
		cflag = 0;
	/*
	**	see whether this is a cellular image
	*/
	if (ihget("dtype", &dtype, 1))
		gipserr();
	if (dtype == 'G') {
		(void)ihget("xnum", (char *)&xdim, 0);
		(void)ihget("ynum", (char *)&ydim, 0);
		if ((input_gimage = gopen(fp, nring)) == NULL)
			gipserr2(GIPSBADOPEN, imfile);
		if (ihget("geometry", buf, sizeof(buf)))
			gipserr();
		dtype = (buf[5] == 'M') ? 'c' : buf[5];
		(void)ihput("dtype", &dtype);
		cflag++;
	}
	if (ihdtype(dtype) == 0)
		gipserr1(GIPSBADVAL);
	if (ihget("xnum", (char *)&xnum, 0) ||
			ihget("ynum", (char *)&ynum, 0))
		gipserr();
	map_sw = (g_setup() == 0);
	/*
	**	get input windowing parameters
	*/
	if (xbuf && sscanf(xbuf, "%d,%d,%d", &xdim, &xorg, &xoff) < 1)
		error("bad x-pixel origin, size, or offset: %s", xbuf);

	if (ybuf && sscanf(ybuf, "%d,%d,%d", &ydim, &yorg, &yoff) < 1)
		error("bad y-pixel origin, size, or offset: %s", ybuf);

	if (xorg != NIX) (void) ihput("xorg", (char *)&xorg);
	else if (ihget("xorg", (char *)&xorg, 0)) xorg = 0;

	if (xdim != NIX) (void) ihput("xlim", (char *)&xdim);
	else if (ihget("xlim", (char *)&xdim, 0)) xdim = xnum;

	if (yorg != NIX) (void) ihput("yorg", (char *)&yorg);
	else if (ihget("yorg", (char *)&yorg, 0)) yorg = 0;

	if (ydim != NIX) (void) ihput("ylim", (char *)&ydim);
	else if (ihget("ylim", (char *)&ydim, 0)) ydim = ynum;
	/*
	**	inspect window limits for consistency
	*/
	if (xorg < 0 || xorg >= xnum)
		error("x-limits outside image", "");
	if (yorg < 0 || yorg >= ynum)
		error("y-limits outside image", "");
	/*
	**	Determine size of saved pixrect
	*/
	if (yzoom == 0 && xzoom == 1)
		Zflag++, yzoom = 1;
	else if (yzoom == NIX)
		yzoom = xzoom;
	if (xzoom < 1)
		error("bad -z value: %d", (char *)xzoom);
	if (yzoom < 1)
		error("bad -Z value: %d", (char *)yzoom);
	image_rect.r_width = saved_rect.r_width = xzoom*MIN(xnum-xorg, xdim);
	image_rect.r_height = saved_rect.r_height = yzoom*MIN(ynum-yorg, ydim);
	/*
	**	Read user-supplied viewport dimensions
	*/
	if (xview && sscanf(xview, "%hd,%hd,%hd",
			&image_rect.r_width, &image_rect.r_left, &rborder) < 1)
		error("bad x-view size, or left/right border: %s", xview);
	if (yview && sscanf(yview, "%hd,%hd,%hd",
			&image_rect.r_height, &image_rect.r_top, &bborder) < 1)
		error("bad y-view size, or left/right border: %s", yview);
	/*
	**	create the base_frame, using the command name if
	**	TOOL_NAME is an SCCS temporary.
	*/
	display_rect = *((Rect *)window_get(base_frame, WIN_SCREEN_RECT));

	width = MIN(image_rect.r_width + image_rect.r_left + rborder,
			display_rect.r_width
				- (int)window_get(base_frame, WIN_LEFT_MARGIN)
				- (int)window_get(base_frame, WIN_RIGHT_MARGIN));
	height = MIN(image_rect.r_height + image_rect.r_top + bborder,
		display_rect.r_height
			- (int)window_get(base_frame, WIN_TOP_MARGIN)
			- (int)window_get(base_frame, WIN_BOTTOM_MARGIN));
	if (depth == 0)
		depth = ((Pixwin *)
			window_get(base_frame, WIN_PIXWIN))->pw_pixrect->pr_depth;

	if (depth == 1) {
		if (xzoom > 4)
			error("-z value too large: %d", (char *)xzoom);
		if (xzoom != yzoom)
			error("xzoom != yzoom not implemented for depth=1", "");
	}
	image_cursor = cursor_create(CURSOR_IMAGE, &image_cursor_pr, 0);
	drag_cursor = cursor_create(CURSOR_IMAGE, &drag_cursor_pr,
		CURSOR_XHOT, 7, CURSOR_YHOT, 7, 0);
	hourglass_cursor = cursor_create(CURSOR_IMAGE, &hourglass_cursor_pr, 0);
	load_cursor[0] = cursor_create(CURSOR_IMAGE, &load_cursor_pr0, 0);
	load_cursor[1] = cursor_create(CURSOR_IMAGE, &load_cursor_pr1, 0);
	load_cursor[2] = cursor_create(CURSOR_IMAGE, &load_cursor_pr2, 0);
	load_cursor[3] = cursor_create(CURSOR_IMAGE, &load_cursor_pr3, 0);
	load_cursor[4] = cursor_create(CURSOR_IMAGE, &load_cursor_pr4, 0);
	load_cursor[5] = cursor_create(CURSOR_IMAGE, &load_cursor_pr5, 0);
	load_cursor[6] = cursor_create(CURSOR_IMAGE, &load_cursor_pr6, 0);
	load_cursor[7] = cursor_create(CURSOR_IMAGE, &load_cursor_pr7, 0);
	load_cursor[8] = cursor_create(CURSOR_IMAGE, &load_cursor_pr8, 0);
	load_cursor[9] = cursor_create(CURSOR_IMAGE, &load_cursor_pr9, 0);
	load_cursor[10] = cursor_create(CURSOR_IMAGE, &load_cursor_pr10, 0);
	load_cursor[11] = cursor_create(CURSOR_IMAGE, &load_cursor_pr11, 0);
	load_cursor[12] = cursor_create(CURSOR_IMAGE, &load_cursor_pr12, 0);
	load_cursor[13] = cursor_create(CURSOR_IMAGE, &load_cursor_pr13, 0);
	load_cursor[14] = cursor_create(CURSOR_IMAGE, &load_cursor_pr14, 0);
	load_cursor[15] = cursor_create(CURSOR_IMAGE, &load_cursor_pr15, 0);
	/*
	**	create a canvas sub-window and associated pixwin
	*/
	image_sw = window_create(base_frame, CANVAS,
		CANVAS_AUTO_SHRINK,	TRUE,
		CANVAS_AUTO_EXPAND,	TRUE,
		CANVAS_RETAINED,	FALSE,
		CANVAS_AUTO_CLEAR,	FALSE,
		CANVAS_WIDTH,		width,
		CANVAS_HEIGHT,		height,
		CANVAS_REPAINT_PROC,	image_repaint,
		WIN_EVENT_PROC,		image_selected, 
		WIN_CURSOR,		image_cursor,
		WIN_ERROR_MSG,		"Can't create image sub-window.",
		WIN_CONSUME_KBD_EVENTS,
			WIN_NO_EVENTS,
			WIN_ASCII_EVENTS,
			KEY_TOP_EDGE,
			KEY_BOTTOM_EDGE,
			KEY_LEFT_EDGE,
			KEY_RIGHT_EDGE,
			KEY_REDRAW,
			0,
		WIN_CONSUME_PICK_EVENTS,
			WIN_NO_EVENTS,
			WIN_MOUSE_BUTTONS,
			LOC_DRAG,
			0,
		0);
	i = (int)window_get(image_sw, CANVAS_MARGIN);	
	window_set(image_sw,
		WIN_WIDTH,	width + i, 
		WIN_HEIGHT,	height + i, 
		0);
	window_fit(base_frame);
	c = rindex(imfile, '/');
	window_set(base_frame,
		FRAME_LABEL,	c ? c+1 : imfile,
		0);
	if (! nolocate)
		window_set(base_frame,
			WIN_X,	MAX(0, (display_rect.r_width - width - i)/2),
			WIN_Y,	MAX(0, (display_rect.r_height - height - i)/2),
			0);
	image_pw = (Pixwin *)window_get(image_sw, CANVAS_PIXWIN);
	image_rect.r_width = (int)window_get(image_sw, CANVAS_WIDTH)
		- image_rect.r_left - rborder;
	image_rect.r_height = (int)window_get(image_sw, CANVAS_HEIGHT)
		- image_rect.r_top - bborder;
	/*
	**	verify colormap depth
	*/
	if (Nflag || Zflag)
		colors = 256;
	for (cshift = 0, i = 256; i != colors; i >>= 1, cshift++)
		if (i == 2)
			error("illegal number of colors: %d", (char *)colors);
	/*
	**	load the colormap and install the menu tree and event mask
	*/
	if (ihget("dmin", &stretch_min)) stretch_min = 0;
	if (ihget("dmax", &stretch_max)) stretch_max = 255;
	graphic_color = 255;
	graphic_sw[0] = 1;
	setup_command_menu();
	default_font = pf_default();
	(void)setfont(strcpy(font, GIPSTOOLFONT));
	(void)setauxfont("S.6", 3);
	set_gips_error();
	/*
	**	If requested, open the logging file
	*/
	if (ofile != NULL) {
		ofile = save_string(ofile, (char *)NULL);
		if ((output_stream = fopen(ofile, "w")) == NULL)
			gipserr2(GIPSBADOPEN, ofile);
	}
	/*
	**	Finally, enable interrupts and start up the tool
	**	Let the window manager read the image
	*/
	window_main_loop(base_frame);
	exit(0);
}
