#ifndef lint
static char *sccsid = "%Z%$RCSfile: menu.c,v $	GIPS $Revision: 3.4 $ (MIT/CSR) $Date: 1998/11/13 23:33:08 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		menu.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.4 $		Update:		$Date: 1998/11/13 23:33:08 $
**	Components:	gips_win_error()	error handler
**			set_gips_error()	initialize error handler
**			display_help_window()	add help window to display
**			setup_command_menu()	read command file
**
**	Function:	Handle menu prompts and help messages
*/

#include	"gipstool.h"
#include	<suntool/panel.h>
	/*
	**	Window error messages
	*/
static	char	*err_title[] = {
#include	"winerrs.h"
	0
};
#define		WIN_NERR	(sizeof(err_title)/sizeof(char *)-1)
	/*
	**	Static
	*/
static	int	(*old_win_error)();		/* system error handler */
extern	int	(*win_errorhandler())();	/* change system handler */
static	int	win_errorcount;			/* error level index */
	/*
	**	Our window error handler
	*/
static
gips_win_error(errnum, iocode)
	register errnum, iocode;
{
	register n;
	char	buf[256];

	if (! win_errorcount++ && errnum && image_sw)
		if (errno > 0 && errno < sys_nerr) {
			display_error_message(sys_errlist[errno], "");
		} else if ((n = (iocode-1) & 0xff) < WIN_NERR && *err_title[n]) {
			(void)sprintf(buf, "win_%%s: error code %d", errnum);
			display_error_message(buf, err_title[n]);
		}
	errnum = (*old_win_error)(errnum, iocode);
	win_errorcount--;
	return errnum;
}
	/*
	**	Setup our error handling routine
	*/
set_gips_error()
{
	if (! old_win_error)
		old_win_error = win_errorhandler(gips_win_error);
}
	/*
	**	read the menu structures
	*/
setup_command_menu()
{
	long	l, lseek();
	int	menu_count;
	char	*buf, file[256], **sub_menu_name;
	Menu	*sub_menu, parent_menu;
	register i, n;
	register char *c;
	/*
	**	Firstly, read the menu file into *buf
	*/
	(void) sprintf(file, GIPSTOOLMENU);
	if ((i = open(file, 0)) < 0 || (l = lseek(i, 0L, 2)) < 0)
		perror(file), exit(1);
	if ((buf = alloc((int)l+2, 1)) == NULL)
		exit(2);
	if (lseek(i, 0L, 0) != 0 || read(i, buf, (int)l) != l)
		perror(file), exit(1);
	(void)close(i);
	/*
	**	Determine number of root menus
	*/
	for (c = buf, menu_count = 0; *c != '\n'; *c++ = 0, menu_count++)
		c = index(c, '\n');
	/*
	**	Allocate menu structures
	*/
	menu = menu_create(MENU_TITLE_ITEM, "Gipstool",
		MENU_LEFT_MARGIN,	4,
		0);
	sub_menu = (Menu *)alloc(sizeof(Menu), menu_count);
	sub_menu_name = (char **)alloc(sizeof(char *), menu_count);
	/*
	**	Fill the menu structure fields
	*/
	for (i = 0, c = buf; *c != '\n'; c++, i++) {
		while (isspace(*c)) c++;
		sub_menu_name[i] = c;
		while (!isspace(*c)) c++;
		*c++ = 0;
		while (isspace(*c)) c++;
		sub_menu[i] = menu_create(MENU_TITLE_ITEM, c,
			MENU_LEFT_MARGIN,       4,
			0);
		(void)sprintf(file, "   %s", c);
		menu_set(menu, MENU_PULLRIGHT_ITEM,
			save_string(file, (char *)NULL),
			sub_menu[i],
			0);
		c += strlen(c);
	}
	/*
	**	Count the number of sub-menu items
	*/
	for (buf = c, i = 0; *++c; i++)
		do c = index(c, '\n');
			while (*++c && *c != '\n');
	/*
	**	Allocate pointer tables to the sub-menu items
	*/
	command_prompt = (char **)alloc(sizeof(char *), i);
	command_text = (char **)alloc(sizeof(char *), i);
	command_help = (char **)alloc(sizeof(char *), i);
	/*
	**	Scan the items again and fill structures
	*/
	for (c = buf, i = 0; *++c; i++) {
		/*
		**	First determine which menu it belongs in
		*/
		parent_menu = NULL;
		for (n = 0; ! parent_menu && n < menu_count; n++) {
			char *d = sub_menu_name[n];
			if (strncmp(d, c, strlen(d)) == 0)
				parent_menu = sub_menu[n];
		}
		/*
		**	Now isolate any command prompt
		*/
		while (!isspace(*c)) c++;
		while (*c == ' ' || *c == '\t') c++;
		if (*c != '\n') command_prompt[i] = c;
		c = index(c, '\n'); *c++ = 0;
		/*
		**	Save the menu item string and decode the key value
		*/
		command_text[i] = c;
		if (c[0] == 'M' && c[1] == '-')
			n = 0200, c += 2;
		else
			n = 0;
		if (c[0] == 'D' && c[1] == 'E' && c[2] == 'L')
			n += 0177;
		else 
			n += (*c == '^' && ! isspace(c[1])) ?
				*++c & 037 : *c & 0177;
		command_index[n] = i+1;
		c = index(c, '\n'); *c++ = 0;
		menu_set(parent_menu ? parent_menu : menu,
			MENU_STRING_ITEM, command_text[i], (caddr_t)n,
			0);
		/*
		**	Finally, save the help lines,
		**	replacing single newlines with blanks
		*/
		command_help[i] = c;
		do	{
			c = index(c, '\n') + 1;
		} while (*c && *c != '\n');
		c[-1] = 0;
	}
	free ((char *)sub_menu_name);
	/*
	**	Add the frame menu
	*/
	menu_set(menu, MENU_PULLRIGHT_ITEM, "   Frame",
		(Menu)window_get(base_frame, WIN_MENU), 0);
}
	/*
	**	Display a menu
	*/
display_canvas_menu(buf, ie)
	char	buf[NUMBUF][TEXTLEN];
	Event	*ie;
{
	Menu	temp_menu;

	if (! event_shift_is_down(ie))
		return (int)menu_show(menu, image_sw, ie, 0);

	temp_menu = menu_create(MENU_STRINGS,
		get_coordinates(buf[0], "x: %x, %tx (%rx,%rX)", ie),
		get_coordinates(buf[1], "y: %y, %ty (%ry,%rY)", ie),
		get_coordinates(buf[2], "d: %d, %td (%rd,%rD)", ie),
		0,
		MENU_NOTIFY_PROC,	menu_return_item,
		0);
	(void)menu_show(temp_menu, image_sw, ie, 0);
	menu_destroy(temp_menu);
	return 0;
}
	/*
	**	Button notifier
	*/
/*ARGSUSED*/
static
reply_to_button(item, event)
	Panel_item	item;	
	Event		*event;
{
	window_return(panel_get(item, PANEL_CLIENT_DATA));
}
	/*
	**	Display the help message for a particular command key
	*/
display_help_window(code, str1, str2)
	int		code;
	char		*str1, *str2;
{
#ifdef sunview_defined
#include	<suntool/alert.h>
	int		narg, i, rc;
	char		ch, *b, *c, buf[50], *arg[20];

	if (i = command_index[code])
		b = c = command_help[i-1];
	else
		(void)sprintf(b = c = buf, "Key %s not assigned", c);
	for (narg = 0; narg < 18; c = b) {
		while ((ch = *b) && ch != '\n') b++;
		*b = 0;
		arg[narg++] = save_string(c, NULL);
		if (! (*b++ = ch))
			break;
	}
	arg[narg] = arg[narg+1] = NULL;
	(void)sprintf(buf, "The '%s' key:", form_char(code)),
	rc = alert_prompt(base_frame, (Event *)NULL,
		ALERT_BUTTON_YES,	str1,
		(str2 && *str2) ? ALERT_BUTTON_NO : ALERT_NO_BEEPING,
		(str2 && *str2) ? (caddr_t)str2 : (caddr_t)TRUE,
		ALERT_POSITION,		ALERT_CLIENT_CENTERED,
		ALERT_MESSAGE_STRINGS, buf,
			arg[0], arg[1], arg[2], arg[3], arg[4],
			arg[5], arg[6], arg[7], arg[8], arg[9],
			arg[10], arg[11], arg[12], arg[13], arg[14],
			arg[15], arg[16], arg[17], arg[18], arg[19],
			0,
		0);
	for (i = 0; i < narg; i++)
		free(arg[narg]);
	switch (rc) {
	case ALERT_YES:
		return 1;
	case ALERT_NO:
		return 2;
	default:
		return 0;
	}
#else
	int		i, width, height, left, top;
	Frame		popup;
	Panel		panel;
	Panel_item	msg;
	struct	pixrect *pr1, *pr2;
	Rect		*r;
	char		*c, *b, buf[50], ch;

	popup = window_create(0, FRAME,
		FRAME_SHOW_LABEL, FALSE,
		0);
	if (depth > 1) {
		static  struct  singlecolor red = { 255, 0, 0 };
		window_set(popup,
			FRAME_FOREGROUND_COLOR, &red,
			0);
	}
	panel = window_create(popup, PANEL,
		PANEL_LAYOUT,	PANEL_VERTICAL,
		0);
	(void)sprintf(buf, "The '%s' key:", c = form_char(code));
	msg = panel_create_item(panel, PANEL_MESSAGE,
		PANEL_ITEM_X,		10,
		PANEL_ITEM_Y,		10,
		PANEL_LABEL_STRING,	buf,
		PANEL_LABEL_BOLD,	TRUE,
		0);
	if (i = command_index[code])
		c = command_help[i-1];
	else
		(void)sprintf(c = buf, "Key %s not assigned", c);
	for (b = c, width = 0, i = 1; ; c = b) {
		while ((ch = *b) && ch != '\n') b++;
		*b = 0;
		msg = panel_create_item(panel, PANEL_MESSAGE,
			PANEL_ITEM_X,		10,
			PANEL_ITEM_Y,		10 + ATTR_ROW(i++),
			PANEL_LABEL_STRING,	c,
			0);
		r = (Rect *)panel_get(msg, PANEL_ITEM_RECT);
		width = max(width, r->r_width);
		if (! (*b++ = ch))
			break;
	}
	pr1 = panel_button_image(panel, str1, strlen(str1),
		(struct pixfont *)NULL);
	width -= pr1->pr_width;
	if (str2 && *str2) {
		pr2 = panel_button_image(panel, str2, strlen(str2),
			(struct pixfont *)NULL);
		width -= pr2->pr_width + 10;
	}
	msg = panel_create_item(panel, PANEL_BUTTON,
		PANEL_ITEM_X,		10 + max(0, width/2),
		PANEL_ITEM_Y,		10 + ATTR_ROW(i),
		PANEL_LABEL_IMAGE,	pr1,
		PANEL_NOTIFY_PROC,	reply_to_button,
		PANEL_CLIENT_DATA,	1,
		0);
	if (str2 && *str2)
		msg = panel_create_item(panel, PANEL_BUTTON,
			PANEL_ITEM_Y,		10 + ATTR_ROW(i),
			PANEL_LABEL_IMAGE,	pr2,
			PANEL_NOTIFY_PROC,	reply_to_button,
			PANEL_CLIENT_DATA,	2,
			0);
	window_fit(panel);
	window_fit(popup);
	width = (int)window_get(popup, WIN_WIDTH) + 10;
	height = (int)window_get(popup, WIN_HEIGHT) + 10;
	r = (Rect *)window_get(base_frame, WIN_SCREEN_RECT);
	left = max(0, (r->r_width - width)/2);
	top = max(0, (r->r_height - height)/2);
	window_set(popup,
		WIN_WIDTH,	width,
		WIN_HEIGHT,	height,
		WIN_X,		left,
		WIN_Y,		top,
		WIN_MOUSE_XY,	left + 15 + (int)panel_get(msg, PANEL_ITEM_X),
				top + 15 + (int)panel_get(msg, PANEL_ITEM_Y),
		0);
	flash_window(0);
	code = (int)window_loop(popup);
	window_set(popup, FRAME_NO_CONFIRM, TRUE, 0);
	window_destroy(popup);
	return code;
#endif sunview_defined
}
