#ifndef lint
static char *sccsid = "%Z%$RCSfile: select.c,v $	GIPS $Revision: 3.18 $ (MIT/CSR) $Date: 1998/11/13 23:33:09 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		select.c	Creation Date:	03/19/86
**	Version:	$Revision: 3.18 $		Update:		$Date: 1998/11/13 23:33:09 $
**	Components:
**		bad_depth()			generate error message
**		clean_whitespace()		remove spaces around input
**		create_ask_window()		setup text input insert
**		display_error_message()		setup error message insert
**		display_system_message()	setup system message insert
**		end_zoomed_pixels()		remove image detail
**		handle_input()			general keyboard parser
**		image_selected()		top level interrupt handler
**		log_command()			write command to log file
**		no_more_memory()		setup out-of-memory insert
**		parse_input_string()		parse text input
**		read_startup_file()		execute the .gipstoolrc file
**		xor_marked_cursor()		temporary cursor insert
**
**	Function:	Handle all keyboard and mouse input
*/

#include	"gipstool.h"
#include	<setjmp.h>
#include	<sys/file.h>
#include	"selections.h"
	/*
	**	Miscellaneous
	*/
static	char	buf[NUMBUF][TEXTLEN];		/* message buffers */
static	Event	*ie;				/* whatever happened */
static	int	cursor_dragging;		/* dragging flag */
static	int	input_alpha;			/* accept alpha input */
static	int	input_escape;			/* reading ESC sequence */
static	int	begin_logging;			/* first logged command */
static	jmp_buf	quit_macro;			/* register storage */
extern	char	*getenv();			/* C library */
static	parse_input_string(), end_zoomed_pixels();
static	log_command(), xor_marked_cursor(), toggle_cursor();
	/*
	**	Read and execute the GIPSTOOLRC file
	**	If an error occurs during execution, quit gipstool
	*/
read_startup_file()
{
	if (gipsrc) {
		if (access(strcpy(buf[0], gipsrc), R_OK))
			gipserr2(GIPSBADOPEN, gipsrc);
	} else {
		(void)sprintf(buf[0], GIPSTOOLRC, ".");
		if (access(buf[0], R_OK)) {
			if (getenv("HOME") == NULL)
				return;
			(void)sprintf(buf[0], GIPSTOOLRC, getenv("HOME"));
			if (access(buf[0], R_OK))
				return;
		}
	}
	if (setjmp(quit_macro)) {
		window_destroy(base_frame);
		(void)fprintf(stderr, "%s\n", buf[0]);
		exit(1);
	}
	begin_logging = 1;
	read_commands(buf[0], ie);
	if (input_stream)
		(void)fclose(input_stream);
	input_stream = NULL;
	update_info();
}
	/*
	**	Error handlers - force message even if no_display set
	*/
display_error_message(msg, insert)
	register char	*msg, *insert;
{
	register display_status;

	if (insert == NULL)
		(void)sprintf(buf[0], "%s: \"%s\"", msg, buf[1]);
	else
		(void)sprintf(buf[0], msg, insert);
	display_status = no_display;
	errno = gipserrno = no_display = 0;
	create_cursor_window(buf, 1);
	if (display_status)
		longjmp(quit_macro, 1);
	flash_window(0);
}
	/*
	**	Handle non-zero errno error.
	**	Take care if insert == buf[1]
	*/
display_system_message(insert)
	register char	*insert;
{
	insert = insert ? strcpy(buf[2], insert) : "";
	if (errno > 0 && errno < sys_nerr)
		(void) strcpy(buf[1], sys_errlist[errno]);
	else
		(void) sprintf(buf[1], "unknown errno = %d", errno);
	display_error_message(insert, (char *)NULL);
}
	/*
	**	A common error message
	*/
no_more_memory()
{
	display_error_message("Insufficient core to hold new graphic", "");
}
	/*
	**	Another common error message
	*/
bad_depth()
{
	display_error_message("%s command illegal for depth=1 display",
		form_char(event_id(ie) & 0377));
}
	/*
	**	Open a window to display keyboard input
	*/
static
create_ask_window(c)
	char	*c;
{
	(void)sprintf(buf[0], command_prompt[command_index[event_id(ie)]-1], c);
	(void)strcpy(buf[1], "_");
	create_cursor_window(buf, 2);
	input_alpha = event_id(ie);
}
	/*
	**	Keyboard/mouse handler
	*/
/*ARGSUSED*/
image_selected(canvas, event, arg)
	Canvas	canvas;			/* image sub-window */
	Event	*event;			/* the event */
	caddr_t	arg;			/* event argument */
{
	int	i;

	ie = event;
	must_update = 0;
	handle_input();
	if (must_update) {
		clear_cursor_window();
		image_repaint(image_sw, image_pw, (Rectlist *)NULL);
		must_update = 0;
		update_info();
	}
}
	/*
	**	This routine is called for each accepted input event
	**	and from read_commands for external input
	*/
handle_input()
{
	register int	button	= 0;	/* set for button input */
	register char	*c;		/* scratch */
	register int	i;		/* scratch */
	/*
	**	Throw away repaints
	*/
	switch (event_id(ie)) {
	case WIN_REPAINT:
	case WIN_RESIZE:
		return;
	case LOC_RGNENTER:
		if ((int)window_get(image_sw, WIN_EVENT_STATE, MS_LEFT))
			{ event_set_id(ie, MS_LEFT); }
		else if ((int)window_get(image_sw, WIN_EVENT_STATE, MS_MIDDLE))
			{ event_set_id(ie, MS_MIDDLE); }
		else if ((int)window_get(image_sw, WIN_EVENT_STATE, MS_RIGHT))
			{ event_set_id(ie, MS_RIGHT); }
		else
			return;
		break;
	}
	/*
	**	Translate ESC[A through ESC[D to numeric codes
	*/
	switch (input_escape) {
	case CTRL('['):
		if (event_id(ie) == '[') {
			input_escape = '[';
			return;
		}
		input_escape = 0;
		break;
	case '[':
		switch (event_id(ie)) {
		case 'A':
			event_set_id(ie, PAN_UP);
			break;
		case 'B':
			event_set_id(ie, PAN_DOWN);
			break;
		case 'C':
			event_set_id(ie, PAN_RIGHT);
			break;
		case 'D':
			event_set_id(ie, PAN_LEFT);
			break;
		default:
			break;
		}
		input_escape = 0;
		break;
	default:
		if (event_id(ie) == CTRL('[')) {
			input_escape = CTRL('[');
			cursor_dragging = 0;
			return;
		}
	}
	/*
	**	If the cursor is being actively sensed,
	**	go and execute the dragging function
	*/
	if (event_id(ie) == LOC_DRAG)
		if ((int)window_get(image_sw, WIN_EVENT_STATE, MS_MIDDLE)) {
			event_set_id(ie, SHOW_ZOOMED_PIXELS);
		} else if (cursor_dragging) {
			event_set_id(ie, cursor_dragging);
		} else
			return;
	else if (cursor_dragging && end_zoomed_pixels()) {
		if (input_alpha)
			return;
		switch (cursor_dragging) {
		case SET_LINE_MARK:
		case SET_BOX_MARK:
			if (event_id(ie) == MS_LEFT)
				return;
			xor_marked_cursor(last_mark.x, last_mark.y);
		default:
			break;
		}
		cursor_dragging = 0;
		window_set(image_sw, WIN_CURSOR, image_cursor, 0);
	} else if (event_shift_is_down(ie) && ! end_zoomed_pixels()) {
		event_set_id(ie, SHOW_ZOOMED_PIXELS);
	}
	/*
	**	Now check to see whether the interrupt code is
	**	special, and needs translating into something else.
	**	[1] text input is continuing
	**	[2] left button - translate to keystroke
	**	[4] right mouse button - display the menu stack
	*/
	if (input_alpha) {
		if ((button = parse_input_string()) == 0)
			return;
	} else if (event_is_button(ie))
		switch (event_id(ie)) {
		case MS_LEFT:
			if (! left_button_value[0])
				return;
			button = -1;
			event_set_id(ie, left_button_value[0] & 0377);
			(void)strcpy(buf[1], left_button_value+1);
			cursor_dragging = left_button_value[0] & 0377;
			break;
		case MS_MIDDLE:
			event_set_id(ie, SHOW_ZOOMED_PIXELS);
			break;
		case MS_RIGHT:
			if ((i = display_canvas_menu(buf, ie)) == 0)
				return;
			event_set_id(ie, i);
			break;
		}
	if (output_stream && ! cursor_dragging)
		log_command(button);
	/*
	**	Now handle the command
	*/
	switch (event_id(ie)) {
	case DEFINE_COMMENT:
		if (! button)
			create_ask_window("");
		break;
	case SET_LINE_MARK:
	case SET_BOX_MARK:
		event_set_x(ie, X_MOD_RECT(event_x(ie)));
		event_set_y(ie, Y_MOD_RECT(event_y(ie)));
		if (cursor_dragging) {
			xor_marked_cursor(last_mark.x, last_mark.y);
		} else {
			cursor_dragging = event_id(ie);
			window_set(image_sw, WIN_CURSOR, drag_cursor, 0);
			set_image_mark(event_x(ie), event_y(ie));
		}
		xor_marked_cursor(last_mark.x = event_x(ie),
			last_mark.y = event_y(ie));
		break;
	case DRAW_LINE:
		last_mark.x = X_MOD_RECT(event_x(ie));
		last_mark.y = Y_MOD_RECT(event_y(ie));
		draw_line(X_IMAGE_TO_RECT(image_mark[0]),
			Y_IMAGE_TO_RECT(image_mark[1]),
			last_mark.x, last_mark.y);
		set_image_mark(last_mark.x, last_mark.y);
		cursor_dragging = SET_LINE_MARK;
		window_set(image_sw, WIN_CURSOR, drag_cursor, 0);
		break;
	case SCROLL_TAG_LIST:
		display_tag_list();
		break;
	case RECALL_NAMED_TAG:
		if (! button)
			create_ask_window(get_tag_name());
		else {
			(void) clean_whitespace(buf[1]);
			recall_named_tag(buf[1], ie);
		}
		break;
	case SET_NAMED_TAG:
		if (! button)
			create_ask_window(get_tag_name());
		else {
			(void) clean_whitespace(buf[1]);
			set_named_tag(buf[1], event_x(ie), event_y(ie));
		}
		break;
	case WRITE_LINE_TRACE:
		if (! button) {
			draw_line(X_IMAGE_TO_RECT(image_mark[0]),
				Y_IMAGE_TO_RECT(image_mark[1]),
				X_MOD_RECT(event_x(ie)),
				Y_MOD_RECT(event_y(ie)));
			create_ask_window("");
		} else {
			(void) clean_whitespace(buf[1]);
			write_line_trace(buf[1], event_x(ie), event_y(ie));
		}
		break;
	case WRITE_SUBIMAGE:
		if (! button)
			create_ask_window("");
		else {
			(void) clean_whitespace(buf[1]);
			write_subimage(buf[1], event_x(ie), event_y(ie));
		}
		break;
	case READ_INPUT_COMMANDS:
		if (! button)
			create_ask_window((ifile && *ifile) ? ifile : "none");
		else if (input_stream)
			display_error_message("Illegal nested input",
				(char *)NULL);
		else if (clean_whitespace(buf[1])) {
			if (! setjmp(quit_macro))
				read_commands(buf[1], ie);
			if (input_stream) {
				(void) fclose(input_stream);
				input_stream = NULL;
			}
			update_info();
		}
		break;
	case READ_TAG_FILE:
		if (! button)
			create_ask_window(tag_file ? tag_file : "none");
		else if (clean_whitespace(buf[1]))
			read_tag_file(buf[1]);
		break;
	case WRITE_TAG_FILE:
		if (! button)
			create_ask_window(tag_file ? tag_file : "none");
		else if (clean_whitespace(buf[1]))
			write_tag_file(buf[1]);
		break;
	case LOG_INPUT_COMMANDS:
		if (! button) {
			create_ask_window((ofile && *ofile) ? ofile : "none");
			begin_logging = 1;
		} else {
			(void) clean_whitespace(buf[1]);
			start_log(buf[1]);
		}
		break;
	case SET_FRAME_LABEL:
		if (! button) {
			c = (char *)window_get(base_frame, FRAME_LABEL);
			create_ask_window(c ? c : "none");
		} else {
			(void) clean_whitespace(buf[1]);
			window_set(base_frame,
				FRAME_LABEL,		buf[1],
				FRAME_SHOW_LABEL,	buf[1][0] ?
							TRUE : FALSE,
				0);
			update_info();
		}
		break;
	case DRAW_X_AXIS:
	case DRAW_Y_AXIS:
	case DRAW_BOTH_AXES:
		if (! button)
			create_ask_window("");
		else {
			if (event_id(ie) == DRAW_Y_AXIS ||
				(draw_axis(0, buf[1]) &&
				event_id(ie) == DRAW_BOTH_AXES))
					(void)draw_axis(1, buf[1]);
			(void)setfont(font);
		}
		break;
	case DRAW_MAP_AXES:
		if (g_setup())
			display_error_message("Image has separable axes", "");
		else if (! button)
			create_ask_window("");
		else
			draw_map_axes(buf[1]);
		break;
	case SET_COLORMAP:
		if (depth == 1)
			bad_depth();
		else if (! button)
			create_ask_window((afile && *afile) ? afile : "none");
		else if (! clean_whitespace(buf[1]))
			;
		else if (! get_color(buf[1])) {
			(void)strcpy(afile, buf[1]);
			update_info();
			stretch_percent = 0;
		} else if (errno)
			display_system_message(buf[1]);
		else
			display_error_message("Open or read error on",
				(char *)NULL);
		break;
	case SET_COLORMAP_STRETCH:
		if (depth == 1)
			bad_depth();
		else if (! button)
			create_ask_window((char *)stretch_percent);
		else if (! clean_whitespace(buf[1]))
			;
		else if ((i = atoi(buf[1])) >= 0 && i <= 100) {
			stretch_colormap(stretch_percent = i);
			if (ihget("dmin", &stretch_min)) stretch_min = 0;
			if (ihget("dmax", &stretch_max)) stretch_max = 255;
			update_info();
		} else
			display_error_message("Bad stretch percentage",
				(char *)NULL);
		break;
	case SET_COLORMAP_LINSTRETCH:
		if (depth == 1) {
			bad_depth();
			break;
		} else if (! button) {
			(void)sprintf(buf[1], "%g %g",
				stretch_min, stretch_max);
			create_ask_window(buf[1]);
			break;
		} else if (buf[1][0]) {
			double	d1, d2;
			if (ihget("dmin", &d1)) d1 = 0;
			if (ihget("dmax", &d2)) d2 = 255;
			if (sscanf(buf[1], "%lf %lf", &d1, &d2) <= 0) {
				display_error_message("Bad stretch range",
					(char *)NULL);
				break;
			}
			stretch_min = d1;
			stretch_max = d2;
		} else {
			if (ihget("dmin", &stretch_min)) stretch_min = 0;
			if (ihget("dmax", &stretch_max)) stretch_max = 255;
		}
		linstretch_colormap();
		stretch_percent = 0;
		update_info();
		break;
	case WRITE_COLORMAP:
		if (depth == 1)
			bad_depth();
		else if (! button)
			create_ask_window("");
		else if (clean_whitespace(buf[1]))
			write_colormap(buf[1]);
		break;
	case WRITE_HISTOGRAM:
		if (! button)
			create_ask_window("");
		else
			write_histogram(buf[1], event_x(ie), event_y(ie), 0);
		break;
	case DISPLAY_COLOR_WEDGE:
		if (! button)
			create_ask_window("");
		else
			display_color_wedge(ie, buf[1]);
		break;
	case SET_FOREGROUND_COLOR:
	case SET_BACKGROUND_COLOR:
		if (depth == 1)
			bad_depth();
		else if (! button) {
			(void)sprintf(buf[1], "%d %d %d",
				rgb[0][0], rgb[1][0], rgb[2][0]);
			create_ask_window(buf[1]);
		} else if (buf[1][0]) {
			int	i1, i2, i3;
			switch (sscanf(buf[1], " %d %d %d", &i1, &i2, &i3)) {
			case 3:
				rgb[2][0] = i3;
			case 2:
				rgb[1][0] = i2;
			case 1:
				rgb[0][0] = i1;
				pw_putcolormap(image_pw,
					(event_id(ie) == SET_BACKGROUND_COLOR)
						? 0 : (colors-1), 1,
					rgb[0], rgb[1], rgb[2]);
				break;
			default:
				display_error_message("Bad color triplet",
					(char *)NULL);
				break;
			}
		}
		break;
	case SET_TEXT_COLOR:
		if (depth == 1)
			bad_depth();
		else if (! button) {
			if (no_display)
				create_ask_window((char *)graphic_color);
			else
				create_color_palette(ie);
		} else if (! buf[1][0])
			;
		else if ((i = atoi(buf[1])) >= 0 && i <= 255) {
			graphic_color = i;
			update_info();
			create_color_palette(0);
		} else
			display_error_message("Bad color value", (char *)NULL);
		break;
	case SET_PAINT_SIZE:
		if (depth == 1)
			bad_depth();
		else if (! button)
			create_ask_window((char *)brush_size);
		else if ((i = atoi(buf[1])) >= 0 && i <= 64) {
			brush_size = i;
			update_info();
		} else
			display_error_message("Bad paint size", (char *)NULL);
		break;
	case PAINT_PIXEL:
		if (depth == 1)
			bad_depth();
		else 
			paint_pixel(ie);
		break;
	case FLOOD_OVERLAY:
		if (depth == 1)
			bad_depth();
		else 
			flood_overlay(ie);
		break;
	case READ_PAINT_OVERLAY:
	case WRITE_PAINT_OVERLAY:
		if (depth == 1)
			bad_depth();
		else if (event_id(ie) == WRITE_PAINT_OVERLAY && ! paint_pr)
			display_error_message("no overlay created", "");
		else if (! button)
			create_ask_window(paint_overlay_file ?
				paint_overlay_file : "none");
		else if (! clean_whitespace(buf[1]))
			;
		else {
			paint_overlay_file =
				save_string(buf[1], paint_overlay_file);
			if (event_id(ie) == WRITE_PAINT_OVERLAY)
				write_paint_overlay();
			else
				read_paint_overlay();
		}
		break;
	case SET_TEXT_FONT:
		if (! button)
			create_ask_window(font);
		else if (clean_whitespace(buf[1])) {
			char	*auxlst = "0rR1iI2bB3sS";
			c = &buf[1][strlen(buf[1])-1];
			i = rotated_font;
			if (*c == 'r') {
				*c = '\0';
				rotated_font = 90;
			} else
				rotated_font = 0;
			c = buf[1];
			if (c[1] == ' ' && (c = index(auxlst, *c))) {
				buf[1][0] = ' ';
				if (! clean_whitespace(buf[1]))
					break;
				c = setauxfont(buf[1], (c - auxlst)/3);
			} else
				c = setfont(buf[1]);
			if (c == NULL) {
				display_system_message(buf[1]);
				(void)setfont(font);
				rotated_font = i;
				break;
			}
			(void)strcpy(font, buf[1]);
			update_info();
		}
		break;
	case DISPLAY_IMAGE_INFO:
		display_info();
		break;
	case ADD_GRAPHIC:
	case REMOVE_GRAPHIC:
		if (! button)
			create_ask_window((char *)graphic_index);
		else if (! buf[1][0] || 
				((i = atoi(buf[1])) >= 0 && i < NGRAPHICS)) {
			if (! buf[1][0])
				i = graphic_index;
			if (event_id(ie) == ADD_GRAPHIC) {
				graphic_index = i;
				if (graphic_sw[i] == -1)
					must_update++;
				graphic_sw[i] = 1;
			} else {
				if (graphic_sw[i] == 1)
					must_update++;
				graphic_sw[i] = -1;
			}
			update_info();
		} else
			display_error_message("Illegal graphic number",
				(char *)NULL);
		break;
	case GET_HELP:
		create_ask_window("");
		break;
	case SET_NUMERIC_PRECISION:
		if (! button)
			create_ask_window((char *)precision);
		else if (! buf[1][0])
			;
		else if ((i = atoi(buf[1])) > 0 && i < 15) {
			precision = i;
			update_info();
		} else
			display_error_message("Bad precision", (char *)NULL);
		break;
	case SET_PAN_INCREMENT:
		if (! button)
			create_ask_window((char *)pan_count);
		else if (! buf[1][0])
			;
		else if ((i = atoi(buf[1])) >= 0) {
			pan_count = i;
			update_info();
		} else
			display_error_message("Bad pan increment",
				(char *)NULL);
		break;
	case SET_TEXT_ANGLE:
		if (! button)
			create_ask_window((char *)graphic_angle);
		else if (buf[1][0]) {
			graphic_angle = atoi(buf[1]);
			update_info();
		}
		break;
	case SET_TEXT_ORIGIN:
		if (! button)
			create_ask_window((char *)graphic_origin);
		else if (! buf[1][0])
			;
		else if ((i = atoi(buf[1])) >= 0) {
			graphic_origin = i;
			update_info();
		} else
			display_error_message("Bad graphic origin code",
				(char *)NULL);
		break;
	case SET_LEFT_BUTTON:
		if (! button)
			create_ask_window(left_button_value);
		else if (buf[1][0]) {
			register char *s;
			for (s = buf[1]; isspace(*s); s++)
				;
			if (*s == '^' && s[1])
				*++s &= 037;
			if (command_index[*s & 0377] == 0)
				display_error_message("Unknown command key: ",
					form_char(*s & 0377));
			else {
				left_button_value =
					save_string(s, left_button_value);
				update_info();
			}
		}
		break;
	case CLEAR_CURRENT_GRAPHIC:
		clear_graphic(graphic_index);
		must_update++;
		break;
	case UNDO_GRAPHIC:
		if (undo_graphic(graphic_index))
			must_update++;
		break;
	case SET_GIPS_VARIABLE:
		if (! button)
			create_ask_window("");
		else if (! buf[1][0])
			;
		else
			switch (set_gips_variable(buf[0], buf[1])) {
			case 1:
				create_cursor_window(buf, 2);
				break;
			case 0:
				clear_cursor_window();
			default:
				break;
			}
		break;
	case ENTER_TEXT_STRING:
		if (! button)
			create_ask_window("");
		else
			write_text(event_x(ie), event_y(ie),
				buf[1][0] ? buf[1] : get_tag_name(),
				graphic_origin, graphic_angle+rotated_font);
		break;
	case ENTER_HEADER_ITEM:
		if (! button)
			create_ask_window("");
		else
			switch (set_gips_variable(buf[0], c = buf[1])) {
			case 1:
				if (gipserrno) {
					create_cursor_window(buf, 2);
					break;
				}
				c[strlen(c)-1] = '\0';
				write_text(event_x(ie), event_y(ie), c+8,
					graphic_origin,
					graphic_angle+rotated_font);
			case 0:
				clear_cursor_window();
			default:
				break;
			}
		break;
	case ENTER_CURSOR_LOCATION:
		if (! button)
			create_ask_window(cursor_format);
		else {
			c = buf[1][0] ? buf[1] : cursor_format;
			if (! *get_coordinates(buf[0], c, ie))
				break;
			write_text(event_x(ie), event_y(ie), buf[0],
				graphic_origin, graphic_angle+rotated_font);
			if (c != cursor_format)
				cursor_format = save_string(c, cursor_format);
		}
		break;
	case BEGIN_PIXEL_LOGGING:
		if (! button)
			create_ask_window((pixel_stream && pixel_log_file) ?
				pixel_log_file : "none");
		else {
			if (pixel_stream) {
				(void) fclose(pixel_stream);
				pixel_stream = NULL;
			}
			if (! clean_whitespace(buf[1]))
				;
			else if ((pixel_stream = fopen(buf[1], "a")) == NULL)
				GIPS_ERROR;
			else {
				pixel_log_file = save_string(buf[1],
					pixel_log_file);
				update_info();
			}
		}
		break;
	case LOG_PIXEL_VALUE:
		if (! button)
			create_ask_window(logging_format);
		else {
			c = buf[1][0] ? buf[1] : logging_format;
			if (! *get_coordinates(buf[0], c, ie))
				break;
			if (pixel_stream)
				(void)fprintf(pixel_stream, "%s\n", buf[0]);
			else
				create_cursor_window(buf, 1);
			if (c != logging_format)
				logging_format = save_string(c, logging_format);
		}
		break;
	case SHOW_ZOOMED_PIXELS:
		create_zoomed_window(ie);
		create_value_pixrect(ie);
		break;
	case ADDRESS_IMAGE_PIXEL:
		if (! button) {
			c = get_coordinates(buf[1], " [x: %x, %rx]", ie);
			c = get_coordinates(c+strlen(c), " [y: %y, %ry]", ie);
			create_ask_window(buf[1]);
		} else if (buf[1][0])
			decode_address(buf[1], ie);
		break;
	case SET_ZOOM_SIZE:
		if (! button) {
			(void)sprintf(buf[1], "%d %d %d",
				item_size, item_zoom, item_color);
			create_ask_window(buf[1]);
		} else if (buf[1][0]) {
			int	i1, i2, i3;
			switch (sscanf(buf[1], " %d %d %d", &i1, &i2, &i3)) {
			case 3:
				if (i3 >= 0 && i3 <= 255)
					item_color = i3;
			case 2:
				if (i2 > 0)
					item_zoom = i2;
			case 1:
				if (i1 >= 0)
					item_size = i1 | 1;
				update_info();
				break;
			default:
				display_error_message("Bad window parms",
					(char *)NULL);
				break;
			}
		}
		break;
	case BLINK_CROSS_HAIRS:
		blink_cross_hairs();
		break;
	case RUN_MONGO:
		if (access(GIPSSOCKET, 0) == EOF)
			system("gfxtool gipsmongo &");
		else
			display_error_message("%s: already exists", GIPSSOCKET);
		break;
	case KEY_REDRAW:
		toggle_cursor();
		break;
	case REPAINT_IMAGE:
		must_update++;
		break;
	case PAN_LEFT:
		xoff -= (pan_count ? pan_count : image_rect.r_width/2)/xzoom;
		must_update++;
		break;
	case PAN_RIGHT:
		xoff += (pan_count ? pan_count : image_rect.r_width/2)/xzoom;
		must_update++;
		break;
	case PAN_UP:
		yoff -= (pan_count ? pan_count : image_rect.r_height/2)/yzoom;
		must_update++;
		break;
	case PAN_DOWN:
		yoff += (pan_count ? pan_count : image_rect.r_height/2)/yzoom;
		must_update++;
		break;
	case KEY_LEFT_EDGE:
	case PAN_LEFT_EDGE:
		xoff = 0;
		must_update++;
		break;
	case KEY_TOP_EDGE:
	case PAN_TOP_EDGE:
		yoff = 0;
		must_update++;
		break;
	case KEY_RIGHT_EDGE:
	case PAN_RIGHT_EDGE:
		xoff = 999999;
		must_update++;
		break;
	case KEY_BOTTOM_EDGE:
	case PAN_BOTTOM_EDGE:
		yoff = 999999;
		must_update++;
		break;
	case SET_TOP_LEFT:
		image_rect.r_width = MIN(sw_rect.r_width - event_x(ie),
			saved_rect.r_width);
		image_rect.r_left = event_x(ie);
		image_rect.r_height = MIN(sw_rect.r_height - event_y(ie),
			saved_rect.r_height);
		image_rect.r_top = event_y(ie);
		must_update++;
		break;
	case SET_BOTTOM_RIGHT:
		if (event_x(ie) > image_rect.r_left)
			image_rect.r_width = MIN(saved_rect.r_width,
				event_x(ie) - image_rect.r_left);
		if (event_y(ie) > image_rect.r_top)
			image_rect.r_height = MIN(saved_rect.r_height,
				event_y(ie) - image_rect.r_top);
		must_update++;
		break;
	case CANCEL_GIPSTOOL:
		if (display_help_window(event_id(ie),
				"EXIT FROM PROGRAM", "CANCEL REQUEST") == 1) {
			window_set(base_frame, FRAME_NO_CONFIRM, TRUE, 0);
			destroy_tag_frame();
			window_destroy(base_frame);
			if (Sflag)
				telltale(1);
			(void) exit(0);
		}
		break;
	case EXECUTE_SHELL:
		if (! button)
			create_ask_window("");
		else if (clean_whitespace(buf[1]))
			execute_shell(buf[1]);
		break;
	case CLEAR_INSERTS:
		clear_cursor_window();
		break;
	case 0:
	case ' ':
	case MS_RIGHT:
		break;
	default:
		if (event_is_ascii(ie)) {
			(void)sprintf(buf[0], "Key %s not assigned",
				form_char(event_id(ie)));
			create_cursor_window(buf, 1);
		}
		break;
	}
}
	/*
	**	Accumulate input characters in buf[1]
	*/
static
parse_input_string()
{
	register char *c;
	register i;
	register id = event_id(ie);
	/*
	**	Help requested from command level
	*/
	if (input_alpha == GET_HELP) {
		if (id == MS_LEFT)
			id = left_button_value[0] & 0377;
		else if (id == MS_MIDDLE)
			id = SHOW_ZOOMED_PIXELS;
		if (id < 0 || id > 255)
			id = 0;	
		(void) display_help_window(id, "OK", "");
		input_alpha = 0;
		clear_cursor_window();
		return 0;
	}
	/*
	**	Help requested during input
	*/
	if (id == GET_HELP) {
		if (display_help_window(input_alpha,
				 "CONTINUE TYPING", "CANCEL SUBCOMMAND") == 2) {
			clear_cursor_window();
			input_alpha = 0;
			buf[1][0] = '\0';
		}
		return 0;
	}
	/*
	**	Ignore bad bytes and bug-out on long lines
	*/
	if (id < 0 || id > 255 || ! isascii(id))
		return 0;
	if ((i = strlen(buf[1])) >= TEXTLEN) {
		display_error_message("line too long","");
		input_alpha = 0;
		return 0;
	}
	c = buf[1]+i-1;
	/*
	**	Carriage return terminates the input
	*/
	if (id == '\r') {
		event_set_id(ie, input_alpha);
		input_alpha = *c = 0;
		clear_cursor_window();
		return 1;
	}
	/*
	**	Normal input accumulates in buf[1]
	*/
	if (isprint(id))
		*c++ = id;
	else if (id == 0177 && c > buf[1])
		c--;
	*c++ = '_';
	*c = 0;
	create_cursor_window(buf, 2);
	return 0;
}
	/*
	**	End cursor-directed zooming unless incremental command
	*/
static
end_zoomed_pixels()
{
	switch (event_id(ie)) {
	case PAN_LEFT:
		event_set_x(ie, event_x(ie) - xzoom);
		break;
	case PAN_RIGHT:
		event_set_x(ie, event_x(ie) + xzoom);
		break;
	case PAN_UP:
		event_set_y(ie, event_y(ie) - yzoom);
		break;
	case PAN_DOWN:
		event_set_y(ie, event_y(ie) + yzoom);
		break;
	case KEY_REDRAW:
		toggle_cursor();
		break;
	case REPAINT_IMAGE:
		break;
	case KEY_TOP_EDGE:
	case PAN_TOP_EDGE:
		event_set_y(ie, event_y(ie) - item_size * yzoom);
		break;
	case KEY_BOTTOM_EDGE:
	case PAN_BOTTOM_EDGE:
		event_set_y(ie, event_y(ie) + item_size * yzoom);
		break;
	case KEY_LEFT_EDGE:
	case PAN_LEFT_EDGE:
		event_set_x(ie, event_x(ie) - item_size * xzoom);
		break;
	case KEY_RIGHT_EDGE:
	case PAN_RIGHT_EDGE:
		event_set_x(ie, event_x(ie) + item_size * xzoom);
		break;
	default:
		return 1;
	}
	window_set(image_sw, WIN_MOUSE_XY, event_x(ie), event_y(ie), 0);
	event_set_id(ie, cursor_dragging);
	return 0;
}
	/*
	**	Remove leading and trailing whitespace
	**	Returns 0 if input is blank, 1 otherwise
	*/
clean_whitespace(c)
	register char *c;
{
	register char *d, *e;

	for (d = e = c; isspace(*c); c++)
		;
	if (c > d)
		while (*d++ = *c++)
			;
	while (d > e)
		if (isspace(*--d))
			*d = '\0';
	return (*d != '\0');
}
	/*
	**	Copy the command to the logging stream
	**	     sw	= 0	simple keystroke
	**		= 1	end of text entry (drop prior newline)
	**		= -1	keystroke and text via mouse button
	*/
static
log_command(sw)
	register int	sw;	/* button code (see above) */
{
	if (sw != 1) {
		switch (event_id(ie)) {
		case KEY_TOP_EDGE:
			event_set_id(ie, PAN_TOP_EDGE);
			break;
		case KEY_BOTTOM_EDGE:
			event_set_id(ie, PAN_BOTTOM_EDGE);
			break;
		case KEY_LEFT_EDGE:
			event_set_id(ie, PAN_LEFT_EDGE);
			break;
		case KEY_RIGHT_EDGE:
			event_set_id(ie, PAN_RIGHT_EDGE);
			break;
		case KEY_REDRAW:
			return;
		case LOG_INPUT_COMMANDS:
			return;
		default:
			break;
		}
		if (begin_logging)
			begin_logging = 0;
		else
			(void)putc('\n', output_stream);
		fputs(form_char(event_id(ie) & 0377), output_stream);
	}
	if (sw != 0 && event_id(ie) != LOG_INPUT_COMMANDS)
		(void)fprintf(output_stream, "\t%s", buf[1]);
}
	/*
	**	Xor cursor box into displayed image
	*/
static
xor_marked_cursor(x0, y0)
	register x0, y0;
{
	register x1, y1;

	x1 = X_IMAGE_TO_RECT(image_mark[0]);
	y1 = Y_IMAGE_TO_RECT(image_mark[1]);

	if (cursor_dragging == SET_LINE_MARK)
		pw_vector(image_pw, x0, y0, x1, y1, PIX_SRC ^ PIX_DST, 255);
	else {
		pw_lock(image_pw, &sw_rect);
		pw_vector(image_pw, x0, y0, x0, y1, PIX_SRC ^ PIX_DST, 255);
		pw_vector(image_pw, x0, y1, x1, y1, PIX_SRC ^ PIX_DST, 255);
		pw_vector(image_pw, x1, y1, x1, y0, PIX_SRC ^ PIX_DST, 255);
		pw_vector(image_pw, x1, y0, x0, y0, PIX_SRC ^ PIX_DST, 255);
		pw_unlock(image_pw);
	}
}

static
toggle_cursor()
{
	(void)cursor_set(image_cursor, CURSOR_OP,
                (int)cursor_get(image_cursor, CURSOR_OP)==(PIX_SRC | PIX_DST) ?
                        (PIX_SRC ^ PIX_DST) : (PIX_SRC | PIX_DST),
                0);
	(void)cursor_set(drag_cursor, CURSOR_OP,
                (int)cursor_get(drag_cursor, CURSOR_OP)==(PIX_SRC | PIX_DST) ?
                        (PIX_SRC ^ PIX_DST) : (PIX_SRC | PIX_DST),
                0);
        (void)window_set(image_sw, WIN_CURSOR, image_cursor, 0);
}
