#ifndef lint
static char *sccsid = "%Z%$RCSfile: sigwinch.c,v $	GIPS $Revision: 3.5 $ (MIT/CSR) $Date: 1990/12/12 12:17:44 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		sigwinch.c	Creation Date:	03/19/86
**	Version:	$Revision: 3.5 $		Update:		$Date: 1990/12/12 12:17:44 $
**	Function:	Handle all window changes
**	Components:	image_sigwinched()	window damage repairer
*/

#include	"gipstool.h"
	/*
	**	Only called when the image has been damaged.
	**	Always restore the entire image (via clipping list)
	**	and repaint the background (if any)
	*/
/*ARGSUSED*/
image_repaint(canvas, pixwin, repaint_area)
	Canvas		canvas;
	Pixwin		*pixwin;
	Rectlist	*repaint_area;
{
	int		x, y, width, height;
	/*
	**	First time through, read the input file
	*/
	if (saved_pr == NULL) {
		read_image();
		read_startup_file();
		return;
	}
	/*
	**	Get the window size
	*/
	width = (int)window_get(canvas, WIN_WIDTH);
	height = (int)window_get(canvas, WIN_HEIGHT);
	/*
	**	Determine whether window has stretched
	*/
	x = width - sw_rect.r_width;
	y = height - sw_rect.r_height;
	if (x != 0 || y != 0)  {
		image_rect.r_width += x;
		sw_rect.r_width = width;
		image_rect.r_height += y;
		sw_rect.r_height = height;
		clear_graphic(0);
	}
	/*
	**	Determine whether window origin has changed
	**	If -c flag used, change of offset causes a new read
	*/
	if (cflag) { 
		if (xoff != 0 || yoff != 0) {
			xorg = MIN(xnum-saved_rect.r_width/xzoom, xorg+xoff);
			yorg = MIN(ynum-saved_rect.r_height/yzoom, yorg+yoff);
			xorg = MAX(0, xorg);
			yorg = MAX(0, yorg);
			xoff = yoff = 0;
			read_image();
			display_graphics(pixwin);
			return;
		}
	} else {
		xoff = MIN(xoff, (saved_rect.r_width-image_rect.r_width)/xzoom);
		xoff = MAX(xoff, 0);
		x = xoff * xzoom;
		yoff= MIN(yoff,(saved_rect.r_height-image_rect.r_height)/yzoom);
		yoff = MAX(yoff, 0);
		y = yoff * yzoom;
		if (x != saved_rect.r_left || y != saved_rect.r_top) {
			saved_rect.r_left = x;
			saved_rect.r_top = y;
		}
	}
	/*
	**	If reading from log file, don't repaint
	*/
	if (no_display)
		return;
	/*
	**	If a border is necessary, paint the background
	*/
	x = saved_rect.r_left + image_rect.r_width;
	y = saved_rect.r_top + image_rect.r_height;
	if (image_rect.r_left > 0 || image_rect.r_top > 0 ||
		image_rect.r_width < sw_rect.r_width ||
		image_rect.r_height < sw_rect.r_height ||
		x > saved_rect.r_width || y > saved_rect.r_height)
		pw_writebackground(pixwin, 0, 0,
			sw_rect.r_width, sw_rect.r_height, PIX_CLR);
	/*
	**	For color displays, pre-access the saved pixrect
	**	to ensure that it's available in virtual memory
	*/
	if (depth > 1)
		for (y = MIN(y,saved_rect.r_height); y >= saved_rect.r_top; y--)
			(void) pr_get(saved_pr, saved_rect.r_left, y);
	/*
	**	Paint the image
	*/
	x = MIN(image_rect.r_width, saved_rect.r_width-saved_rect.r_left);
	y = MIN(image_rect.r_height, saved_rect.r_height-saved_rect.r_top);
	pw_write(pixwin, image_rect.r_left, image_rect.r_top, x, y,
		PIX_SRC, saved_pr, saved_rect.r_left, saved_rect.r_top);
	/*
	**	Add graphic overlays
	*/
	display_graphics(pixwin);
	/*
	**	Add the paint overlay
	*/
	if (paint_pr && graphic_index == 0) {
#ifdef OLD
		pw_write(pixwin, image_rect.r_left, image_rect.r_top, x, y,
			PIX_SRC | PIX_DST,
			paint_pr, saved_rect.r_left, saved_rect.r_top);
#else
		register int	len, xx, yy, n;
		register char	*c;
		Rect	*r;

		r = (Rect *)window_get(canvas, WIN_RECT); 
		pw_lock(pixwin, r);
		len = mpr_d(paint_pr)->md_linebytes;
		c = (char *)(mpr_d(paint_pr)->md_image)
			+ len*saved_rect.r_top + saved_rect.r_left;
		for (yy = 0; yy < y; yy++, c += len)
			for (xx = 0; xx < x; xx++)
				if (n = c[xx])
					pw_put(pixwin,
						xx+image_rect.r_left,
						yy+image_rect.r_top,
						n);
		pw_unlock(pixwin);
#endif
	}
}
