#ifndef lint
static char *sccsid = "%Z%$RCSfile: utils.c,v $	GIPS $Revision: 3.6 $ (MIT/CSR) $Date: 1998/11/13 23:33:04 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		utils.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.6 $		Update:		$Date: 1998/11/13 23:33:04 $
**	Components:	display_gips_variable	display header variable
**			form_char		convert byte to ASCII
**			set_gips_variable	redefine header variable
**			save_string()		alloc & save string
**			read_commands()		read commands from file
**			start_log()		commence command logging
**			log_command()		log single command
**
**	Function:	Handle interaction between windows & gips header
*/

#include	"gipstool.h"
#include	<sunwindow/notify.h>
#include	<sys/wait.h>
#include	<sys/time.h>
#include	<sys/resource.h>
	/*
	**	Set value of gips variable
	*/
set_gips_variable(info, value)
	char	*info, *value;
{
	char	abuf[BUFLEN];
	int	i;
	double	d;
	register char	*name;
	register char	*c;
	/*
	**	Scan "value" for variable name
	*/
	*info = 0;
	name = info;
	for (c = value; isspace(*c); )
		if (! *++c) return 0;
	while (*c && ! isspace(*c) && *c != '=')
		*info++ = *c++;
	while (isspace(*c)) *c++ = 0;
	*info = 0;
	/*
	**	Remember the delimiter and get variable characteristics
	*/
	i = *c;
	if (ihtype(name, abuf, sizeof(abuf))) {
		GIPS_ERROR;
		return -1;
	}
	/*
	**	If name followed by '=', continue parsing
	*/
	if (i == '=') {
		while (isspace(*++c)) c++;
		if (*c == '\0') {
			(void)ihdel(name);
		} else  {
			gipserrno = GIPSBADTYPE;
			if (*abuf == 'i' && sscanf(c,"%d",&i) == 1)
				(void)ihput(name,(char *)&i);
			else if (*abuf == 'd' && sscanf(c,"%lf",&d) == 1)
				(void)ihput(name,(char *)&d);
			else if (*abuf == 's' || *abuf == 'h' || *abuf == 'c')
				(void)ihput(name,c);
			else
				gipserrno = GIPSBADTYPE;
		}
	}
	/*
	**	Write the value back
	*/
	value = strcpy(value, "value: '")+8;
	if (gipserrno == 0)
		switch (*abuf) {
		case 'c':
		case 'R':
		case 'V':
		case 'h':
		case 's':
			if (ihget(name, value, 200) == 0 && *abuf == 'c')
				value[1] = 0;
			break;
		case 'i':
			if (ihget(name, (char *)&i, 0) == 0)
				(void)sprintf(value, "%d", i);
			break;
		case 'd':
			if (ihget(name, (char *)&d, 0) == 0)
				(void)gcvt(d, precision, value);
			break;
		}
	/*
	**	Handle errors
	*/
	switch (gipserrno) {
	case 0:
		(void)strcat(value,"'");
		break;
	case GIPSNOVAL:
		(void)strcpy(--value, "none");
		break;
	default:
		(void)sprintf(value-8, gipserrs[gipserrno], gipserrs[0]);
		break;
	}
	(void)sprintf(info, " (type=%c) %s", *abuf, abuf+1);
	return 1;
}
	/*
	**	Allocate and save a string
	*/
char	*
save_string(s, t)
	register char	*s;		/* string to save */
	register char	*t;		/* previous value, if any */
{
	if (t != NULL)
		free(t);
	if ((t = alloc(1,strlen(s)+1)) == NULL)
		gipserr();
	return strcpy(t, s);
}
	/*
	**	Return printable version of character
	*/
char *
form_char(i)
	register i;
{
	static	char	b[12];
	register char	*c = b;

	if (i > 255 || i < -255) {
		(void)sprintf(b, "%d", i);
		return b;
	}
	if ((i &= 0377) & 0200)
		*c++ = 'M', *c++ = '-', i &= 0177;
	if (i < 040)
		*c++ = '^', i += '@';
	else if (i == 0177)
		*c++ = 'D', *c++ = 'E', i = 'L';
	*c++ = i;
	*c = 0;
	return b;
}
	/*
	**	Read commands from file
	*/
read_commands(name, ie)
	char	*name;		/* input file name */
	Event	*ie;		/* from select.c */
{
	char	buf[BUFSIZ];	/* input buffer */
	register char *c;	/* scratch */

	no_display = 1;
	ifile = save_string(name, ifile);
	if ((input_stream = fopen(ifile, "r")) == NULL)
		display_system_message(ifile);;
	while (fgets(buf, sizeof(buf)-1, input_stream)) {
		for (c = buf; *c == ' ' || *c == '\t'; c++)
			;
		if (*c == '^')
			*++c &= 037;
		else if (strncmp(c, "DEL", 3) == 0)
			c++,  *++c = 0177;
		while (*c) {
			if (*c == '\n')
				*c = '\r';
			event_set_id(ie, *c++ & 0377);
			handle_input();
			if (must_update)
				image_repaint(image_sw, image_pw,
					(Rectlist *)NULL);
		}
	}
	no_display = 0;
}
	/*
	**	Redirect command logging
	*/
start_log(c)
	register char *c;		/* output file name */
{
	register char *mode;		/* file opening mode */

	if (output_stream) {
		(void)putc('\n', output_stream);
		(void) fclose(output_stream);
	}
	if (*c == '>') {
		mode = "a";
		while (isspace(*++c)) ;
		if (! *c) c = ofile;
	} else
		mode = "w";
	if (*c) {
		if (c != ofile)
			ofile = save_string(c, ofile);
		if ((output_stream = fopen(ofile, mode)) == NULL)
			display_system_message(ofile);
		else
			display_error_message("Logging to: \"%s\"", ofile);
	} else {
		output_stream = NULL;
		display_error_message("Log file closed", "");
	}
}
	/*
	**	Remove tabs from string
	*/
static	char *
notab(s)
	register char	*s;
{
	static	char	buf[TEXTLEN];
	register char	*t;

	for (t = buf; *t++ = (*s == '\t') ? ' ' : *s; s++) ;
	return buf;
}

static	Frame	text_frame;		/* frame for text popup */
	/*
	**	Trap menu for text window
	*/
static
handle_text_event(canvas, event, arg)
        Canvas  canvas;
        Event   *event;
        caddr_t arg;
{
        if (event_id(event) == MS_RIGHT)
                (void)menu_show((Menu)window_get(text_frame, WIN_MENU),
			canvas, event,
			0);
	else if (event_is_ascii(event) && event_is_down(event))
		image_selected(image_sw, event, 0);
}
	/*
	**	Save corner of image, insert text
	*/
create_cursor_window(buf, narg)
	char	buf[NUMBUF][TEXTLEN];	/* text to insert */
	int	narg;			/* number of lines */
{
	static	Panel		text_panel;	/* panel for text insert */
	static	Panel_item	text_item[NUMBUF];	/* text items */
	int	i;

	if (no_display)
		return;

	if (! text_frame) {
		text_frame = window_create(base_frame, FRAME,
			FRAME_SHOW_LABEL,	FALSE,
			WIN_X,			0,
			WIN_Y,			(int)window_get(base_frame,
						WIN_TOP_MARGIN),
			0);
		if (depth > 1) {
			static	struct	singlecolor red = { 255, 0, 0 };
			window_set(text_frame,
				FRAME_FOREGROUND_COLOR,	&red,
				0);
		}
		text_panel = window_create(text_frame, PANEL,
			WIN_EVENT_PROC,		handle_text_event,
			0);
		for (i = 0; i < NUMBUF; i++)
			text_item[i] = panel_create_item(text_panel,
				PANEL_MESSAGE,
				PANEL_ITEM_X,		0,
				PANEL_ITEM_Y,		ATTR_ROW(i),
				0);
		panel_set(text_item[0], PANEL_LABEL_BOLD, TRUE, 0);
	}

	for (i = 0; i < NUMBUF; i++)
		panel_set(text_item[i],
			PANEL_LABEL_STRING, i < narg ? notab(buf[i]) : "",
			0);

	window_fit(text_panel);
	window_fit(text_frame);
	window_set(text_frame, WIN_SHOW, TRUE, 0);
}
	/*
	**	Remove the window
	*/
clear_cursor_window()
{
	if (text_frame)
		window_set(text_frame, WIN_SHOW, FALSE, 0);
}
	/*
	**	Blink the cursor cross hairs
	*/
blink_cross_hairs()
{
	static	int	state;

	cursor_set(image_cursor, CURSOR_SHOW_CROSSHAIRS,
		(state ^= 1) ? TRUE : FALSE,
		0);
	window_set(image_sw, WIN_CURSOR, image_cursor, 0);
}

	/*
	**	Routine to clean up after child terminates
	*/

static	Notify_value
shell_handler(id, pid, status, rusage)
	int	*id, pid;
	union	wait	*status;
	struct	rusage	*rusage;
{
	if (! WIFEXITED(*status))
		return NOTIFY_IGNORED;
	cursor_set(image_cursor, CURSOR_SHOW_CURSOR, TRUE, 0);
	window_set(image_sw, WIN_CURSOR, image_cursor, 0);
	flash_window(image_pw);
	return NOTIFY_DONE;
}

	/*
	**	Execute a shell within gipstool
	*/

execute_shell(str)
	char	*str;
{
	static	int	pid;
	int		i;

	clear_cursor_window();
	cursor_set(image_cursor, CURSOR_SHOW_CURSOR, FALSE, 0);
	window_set(image_sw, WIN_CURSOR, image_cursor, 0);
	image_repaint(image_sw, image_pw, (Rectlist *)NULL);
	if ((pid = vfork()) == 0) {
		for (i = 3; i < getdtablesize(); i++)
			(void)close(i);
		sleep(1);
		execl("/bin/sh", "sh", "-c", str, 0);
		_exit(127);
	}
	(void) notify_set_wait3_func(&pid, shell_handler, pid);
}

	/*
	**	Flash the image, beep the buzzer
	*/

flash_window(pw)
	Pixwin	*pw;
{
	static	struct	timeval	tv = { 0, 100000} ;

	win_bell((int)window_get(image_sw, WIN_FD), tv, pw);
}
