#ifndef lint
static char *sccsid = "%Z%$RCSfile: zoom.c,v $	GIPS $Revision: 3.6 $ (MIT/CSR) $Date: 1998/05/13 01:12:28 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		zoom.c		Creation Date:	03/19/86
**	Version:	$Revision: 3.6 $		Update:		$Date: 1998/05/13 01:12:28 $
**	Components:	create_bit_pixrect	zoom a 1-bit raster
**			create_item_pixrect	zoom an 8-bit raster
**
**	Function:	Create zoomed pixrects from the saved image
*/

#include	"gipstool.h"
	/*
	**	Popups for coordinate values and zoom box
	*/
static	Frame		zoom_frame;	/* Frame for zoom popup */
static	Canvas		zoom_canvas;	/* Canvas for zoom popup */
static	Pixwin		*zoom_pixwin;	/* Pixwin for zoom popup */
static	struct pixrect	*zoom_pr;	/* Pixrect for zoom popup */
static	int		zoom_width;	/* width of zoom window */
static	int		zoom_height;	/* height of zoom window */
static	Frame		value_frame;	/* Frame for value popup */
static	Panel		value_panel;	/* Panel for value popup */
static	Panel_item	value_item[3];	/* Panel items */
 
static	char	*fmt[]	= {
		" x: %x, %tx (%rx,%rX) ",
		" y: %y, %ty (%ry,%rY) ",
		" d: %d, %td (%rd,%rD) "
};

static	create_bit_pixrect();
	/*
	**	Create pixrect and fill with zoomed pixels
	*/
static
create_item_pixrect(locx, locy)
	int	locx;	/* x coordinate of source pixel in pixwin */
	int	locy;	/* y coordinate of source pixel in pixwin */
{
	int	nx = item_zoom*xzoom;
	int	ny = item_zoom*yzoom;
	char	*in, *out;
	int	inlen, outlen, outsize, x0, y0, xlim, ylim;
	register char *cin, *cout;
	register i, p, x;
	/*
	**	create a pixrect to hold the result
	*/
	xlim = nx*item_size;
	ylim = ny*item_size;
	if (! zoom_pr ||
		xlim != zoom_pr->pr_width ||
			ylim != zoom_pr->pr_height) {
		if (zoom_pr)
			(void)pr_close(zoom_pr);
		zoom_pr = mem_create(nx*item_size, ny*item_size, depth);
		if (zoom_pr == NULL) {
			no_more_memory();
			return;
		}
	}
	x0 = (X_RECT_TO_IMAGE(locx) - xorg - item_size/2)*xzoom;
	y0 = (Y_RECT_TO_IMAGE(locy) - yorg - item_size/2)*yzoom;

	in = (char *)(mpr_d(saved_pr)->md_image);
	out = (char *)(mpr_d(zoom_pr)->md_image);
	inlen = (int)(mpr_d(saved_pr)->md_linebytes);
	outlen = (int)(mpr_d(zoom_pr)->md_linebytes);
	bzero(out, outlen * ylim);
	if (depth == 1) {
		ny = item_zoom;
		outsize = outlen;
		outlen *= yzoom;
	}
	xlim = MIN(saved_rect.r_width, x0 + xzoom*item_size);
	ylim = MIN(saved_rect.r_height, y0 + yzoom*item_size);
	/*
	**	Outer loop over non-identical rasters
	*/
	for ( ; y0 < ylim; y0 += yzoom, out += ny * outlen) {
		if (y0 < 0)
			continue;
		cin = in + y0 * inlen;
		if (depth == 1)
			create_bit_pixrect(cin, inlen, out, outsize, x0, xlim);
		else {
			cout = out;
			for (x = x0; x < xlim; x += xzoom, cout += nx)
				if (x >= 0)
					for (p = cin[x], i = 0; i < nx; i++)
						cout[i] = p;
		}
		/*
		**	Replicate the identical rasters
		*/
		for (cout = out, i = ny; --i > 0; cout += outlen)
			move(cout, cout+outlen, outlen);
	}
	/*
	**	Insert borders and cross-hair
	*/
	x0 = item_zoom*xzoom*item_size-1;
	y0 = item_zoom*yzoom*item_size-1;
	(void)pr_vector(zoom_pr, 0, 0, x0, 0, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, x0, 0, x0, y0, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, 0, y0, x0, y0, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, 0, 0, 0, y0, PIX_SRC, item_color);
	x0 = item_zoom*xzoom*(item_size/2);
	y0 = item_zoom*yzoom*(item_size/2);
	nx = item_zoom*xzoom;
	ny = item_zoom*yzoom;
	(void)pr_vector(zoom_pr, x0+nx, y0+ny, x0+2*nx-1, y0+2*ny-1, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, x0+nx, y0-1, x0+2*nx-1, y0-ny, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, x0-nx, y0-ny, x0-1, y0-1, PIX_SRC, item_color);
	(void)pr_vector(zoom_pr, x0-1, y0+ny, x0-nx, y0+2*ny-1, PIX_SRC, item_color);
}
	/*
	**	Replicate yzoom rasters in depth=1 pixrect
	*/
static
create_bit_pixrect(in, inlen, out, outlen, x0, xlim)
	char	*in, *out;
	int	inlen, outlen, x0, xlim;
{
	register xz = xzoom;
	int	yz = yzoom;

	for (in += 2*(x0 >> 4); --yz >= 0; in += inlen, out += outlen) {
		short	*inptr = (short *)in;
		unsigned inacc = (x0 >= 0) ? *inptr : 0;
		int	infill = 16 - (x0 & 017);
		short	*outptr = (short *) out;
		register unsigned outacc = 0;
		register outfill = 0;
		register x;

		for (x = x0; x < xlim; x += xz) {
			register unsigned pix = 0;
			register i;

			for (i = xz; i; ) {
				if (i >= infill) {
					pix <<= infill;
					pix |= inacc &
						(0xffff >> (16-infill));
					i -= infill;
					inptr++;
					if (x+xz >= i)
						inacc = *inptr;
					infill = 16;
				} else {
					pix <<= i;
					pix |= (inacc >> (infill -= i))
						& (0xffff >> (16-i));
					break;
				}
			}
			for (i = item_zoom; --i >= 0; outfill += xz) {
				if (outfill >= 16) {
					outfill -= 16;
					*outptr++ = (outacc>>outfill) & 0xffff;
				}
				outacc = (outacc << xz) | pix;
			}
		}
		if (outfill > 0)
			*outptr++ = outacc << (16 - outfill);
	}
}
	/*
	**	Event handler to pass menu button
	*/
/*ARGSUSED*/
static
handle_zoom_event(canvas, event, arg)
	Canvas	canvas;
	Event	*event;
	caddr_t	arg;
{
	if (event_id(event) == MS_RIGHT)
		(void)menu_show((Menu)window_get(
			(canvas == zoom_canvas) ? zoom_frame : value_frame,
			WIN_MENU), canvas, event, 0);
	else if (event_is_ascii(event) && event_is_down(event))
		image_selected(image_sw, event, 0);
}
	/*
	**	Repaint procedure for zoomed window
	*/
/*ARGSUSED*/
static
zoom_repaint(canvas, pixwin, repaint_area)
	Canvas		canvas;
	Pixwin		*pixwin;
	Rectlist	*repaint_area;
{
	if (zoom_pr)
		pw_write(pixwin, 0, 0, zoom_width, zoom_height, 
			PIX_SRC, zoom_pr, 0, 0);
}
	/*
	**	Create zoomed window and possible info insert
	*/
create_zoomed_window(ie)
	Event	*ie;
{
	Rect	r;
	int	create = 0;

	if (no_display)
		return;
	create_item_pixrect(event_x(ie), event_y(ie));

	if (! zoom_frame || ! zoom_canvas ||
		zoom_pr->pr_width != zoom_width ||
			zoom_pr->pr_height != zoom_height) {
		if (! zoom_frame) zoom_frame = window_create(base_frame, FRAME,
			FRAME_LABEL,	window_get(base_frame, FRAME_LABEL),
			FRAME_SHOW_LABEL,		TRUE,
			WIN_BELOW,			base_frame,
			0);
		create++;
		zoom_height = zoom_pr->pr_height;
		zoom_width = zoom_pr->pr_width;
		zoom_canvas = window_create(zoom_frame, CANVAS,
			WIN_X,			0,
			WIN_Y,			0,
			WIN_WIDTH,		zoom_width,
			WIN_HEIGHT,		zoom_height,
			WIN_EVENT_PROC,		handle_zoom_event,
			WIN_CONSUME_KBD_EVENTS,
				WIN_NO_EVENTS,
				WIN_ASCII_EVENTS,
				0,
			WIN_CONSUME_PICK_EVENTS,
				WIN_NO_EVENTS,
				WIN_MOUSE_BUTTONS,
				0,
			CANVAS_AUTO_SHRINK,	FALSE,
			CANVAS_RETAINED,	FALSE,
			CANVAS_WIDTH,		zoom_width,
			CANVAS_HEIGHT,		zoom_height,
			CANVAS_FIXED_IMAGE,	TRUE,
			CANVAS_AUTO_CLEAR,	FALSE,
			CANVAS_REPAINT_PROC,	zoom_repaint,
			0);
		window_fit(zoom_frame);
		zoom_pixwin = (Pixwin *)window_get(zoom_canvas, WIN_PIXWIN);
		if (depth != 1 && zoom_pixwin->pw_pixrect->pr_depth != 1) {
			pw_setcmsname(zoom_pixwin, colormap);
			pw_putcolormap(zoom_pixwin, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
		}
		zoom_pixwin = canvas_pixwin(zoom_canvas);
		if (depth != 1 && zoom_pixwin->pw_pixrect->pr_depth != 1) {
			pw_setcmsname(zoom_pixwin, colormap);
			pw_putcolormap(zoom_pixwin, 0, colors,
				newrgb[0], newrgb[1], newrgb[2]);
		}
	}
	if (create) {
		r = *((Rect *)window_get(base_frame, WIN_RECT));
		if (r.r_top + r.r_height + zoom_height > display_rect.r_height)
			window_set(zoom_frame,
				WIN_Y, display_rect.r_height
					- r.r_top - zoom_height,
				0);
	}
	zoom_repaint(zoom_canvas, zoom_pixwin, (Rectlist *)NULL);
	window_set(zoom_frame, WIN_SHOW, TRUE, 0);
}
	/*
	**	Coordinate value window
	*/
create_value_pixrect(ie)
	Event	*ie;
{
	char	buf[256];
	int	i, create = 0;
	Rect	r, t;

	if (no_display)
		return;
	if (! value_frame) {
		value_frame = window_create(base_frame, FRAME,
			FRAME_LABEL,	window_get(base_frame, FRAME_LABEL),
			FRAME_SHOW_LABEL,	TRUE,
			WIN_BELOW,		base_frame,
			WIN_RIGHT_OF,		zoom_frame,
			0);
		create++;
		value_panel = window_create(value_frame, PANEL,
			WIN_EVENT_PROC,	handle_zoom_event,
			0);
	}
	for (i = 0; i < 3; i++) {
		if (! value_item[i])
			value_item[i] = panel_create_item(value_panel,
				PANEL_MESSAGE,
				PANEL_ITEM_X,	0,
				PANEL_ITEM_Y,	ATTR_ROW(i),
				0);
		(void)panel_set(value_item[i], PANEL_LABEL_STRING,
				get_coordinates(buf, fmt[i], ie),
			0);
	}
	window_fit(value_panel);
	window_fit(value_frame);
	if (create) {
		r = *((Rect *)window_get(base_frame, WIN_RECT));
		i = (int)window_get(value_frame, WIN_HEIGHT);
		if (r.r_top + r.r_height + i > display_rect.r_height)
			window_set(value_frame,
				WIN_Y, display_rect.r_height - r.r_top - i,
				0);
	}
	if (create && zoom_frame) {
		t = *((Rect *)window_get(zoom_frame, WIN_RECT));
		i = (int)window_get(value_frame, WIN_WIDTH);
		if (r.r_left + t.r_left + t.r_width + i > display_rect.r_width)
			window_set(value_frame,
				WIN_X, display_rect.r_width - r.r_width - i,
				0);
	}
	window_set(value_frame, WIN_SHOW, TRUE, 0);
}
