#ifndef lint
static char *sccsid = "%Z%$RCSfile: git.c,v $	GIPS $Revision: 2.12 $ (MIT/CSR) $Date: 2004/11/02 20:39:08 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		gin.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.12 $		Update:		$Date: 2004/11/02 20:39:08 $
**	Function:	Display general image pixels
**	Components:	main()		GIPS/UNIX command
*/

#include	<stdio.h>
#include	"gips.h"

static	flag	aflag;			/* display abs(pixel) for dtype=c */
static	flag	cflag;			/* compressed output */
static	flag	hflag;			/* display header */
static	flag	nflag;			/* display only header */
static	flag	sflag;			/* insert /n between rasters */
static	char	*mfile	= "-";		/* merge input header */
static	int	prec	= 4;		/* specify output precision */

static	flagentry flagtable[] = {
	"a",	"1",		(flag *) &aflag,
	"c",	"1",		(flag *) &cflag,
	"h",	"1",		(flag *) &hflag,
	"m",	"-> file",	(flag *) &mfile,
	"n",	"1",		(flag *) &nflag,
	"p",	"%d precision",	(flag *) &prec,
	"s",	"1",		(flag *) &sflag,
	NULL,	"[file]",	NULL
};

static	double	pixmin;
extern  char    *gcvt();
static		complex();

main(argc,argv)
	
	int	argc;
	char	**argv;
{
	char	dtype;
	char	dscale	= 's';
	double	dfact;
	double	dzero;
	int	xnum;
	int	ynum;
	int	ny;
	int	dsize;
	int	fnull;
	char	buf[20];
	char	*inbuf;
	char	*nullbuf	= NULL;
	double	*convbuf;
	double	dnull;
	double	pow();
	FILE	*fm;

	register FILE	*fp;
	register char	*name;

	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	name = (argc > 1) ? *++argv : "-";
	if (prec < 0 || prec > 12)
		error("illegal -p value: %d",(char *)prec);
	pixmin = pow((double)10,(double)(-prec));
	if ((fp = ihread(name,GIPSPROFILE)) == NULL)
		gipserr();
	if (*mfile != '-')
		if ((fm = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void) ihbcl(fm);
	if (hflag || nflag)
		(void)(ihwrit("-", 1));
	if (nflag) {
		(void)(ihbcl(fp));
		exit(0);
	}
	if (ihget("xnum",(char *)&xnum,0) ||
	    ihget("ynum",(char *)&ynum,0) ||
	    ihget("dtype",(char *)&dtype,1))
		gipserr();
	if (! ihdtype(dtype))
		gipserr1(GIPSBADTYPE);
	dsize = xnum * ihdtype(dtype);
	inbuf = alloc(xnum, ihdtype(dtype));
	convbuf = (double *)alloc(xnum, sizeof(double));
	if (inbuf == NULL || convbuf == NULL)
		gipserr();
	(void)(ihget("dscale",(char *)&dscale,1));
	fnull = ihget("dnull",(char *)&dnull,0);
	if (dscale == 'n')
		if (ihget("dfact",(char *)&dfact,0) ||
		    ihget("dzero",(char *)&dzero,0))
			gipserr();
	if (aflag && dtype != 'c')
		error("dtype=%.1s conflicts with -a",&dtype);

	for (ny = 0; ny < ynum; ny++) {
		register	i;
		register	nu = 0;
		register double	*d;
		register char	*b;

		if (sflag)
			(void) printf("\n%d:\t", ny);

		if (fread(inbuf, ihdtype(dtype), xnum, fp) != xnum)
			break;
		if (dtype == 'c' && ! aflag) {
			complex((float *)inbuf, 2*xnum);
			continue;
		}
		if (nullbuf && ! bcmp(inbuf, nullbuf, dsize)) {
			(void) printf("[%d]\n", xnum);
			continue;
		}
		gconv((char*)convbuf, inbuf, xnum, CVT_D, ihdcvt(dtype));
		
		for (d = convbuf, i = 0; i++ < xnum; d++) {
			if (fnull == 0 && dtype != 'c' && *d == dnull) {
				nu++;
				continue;
			} else if (nu) {
				(void) printf("[%d] ",nu);
				nu = 0;
			}
			if (cflag) {
				register n = 0;
				while (i < xnum && d[0] == d[1])
					i++, d++, n++;
				if (n)
					(void) printf("%d*", n + 1);
				
			}
			if (dscale == 'n')
				*d = dfact * *d + dzero;
			(void) printf(cflag ? "%s " : "%3s ",
				(*d == 0 || *d > pixmin || *d < -pixmin) ?
				gcvt(*d, prec, buf) : "~0");
		}
		if (nu == xnum && nullbuf == NULL &&
				(nullbuf = alloc(xnum, ihdtype(dtype))))
			move(inbuf, nullbuf, dsize);
		(void) printf(nu ? "[%d]\n" : "\n",nu);
	}
	(void)(ihbcl(fp));
	exit(0);
}

static
complex(f, i)
	register float *f;
	register i;
{
	char	buf[20];

	for ( ; --i >= 0; f++) {
		if (i & 1) putchar('(');
		(void) fputs((*f == 0 || *f > pixmin || *f < -pixmin) ?
			gcvt(*f, prec, buf) : "~0", stdout);
		if (!(i & 1)) putchar(')');
		putchar(' ');
	}
	putchar('\n');
}
