/*	%Z%$RCSfile: iged.h,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 2004/11/29 18:54:08 $
**
**		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged.h		Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 2004/11/29 18:54:08 $
**	Components:	none		C include file
**
**	Function:	Supply definitions and macros to iged.c source
*/

#include	<stdio.h>
#ifndef		IGED
#define		IGED
#endif
#include	"gips.h"
	/*
	 *	Common subroutine return values
	 */
#define		GOOD		0
#define		EVIL		1
	/*
	 *	Flags defined in subcommand table
	 */
#define		UNDEF	01	/* subcommand not defined */
#define		IN	02	/* valid in input mode only */
#define		UP	04	/* subcommand updates a graphic */
#define		HD	010	/* header must be in core */
#define		NI	020	/* invalid during initialization */
#define		NR	040	/* invalid for repetition */
	/*
	 *	Routines to process subcommands
	 */
extern	iged_bf();		/* ] write filled box into graphic */
extern	iged_bx();		/* [ write open box into graphic */
extern	iged_com();		/* . issue command to display */
extern	iged_csh();		/* ! execute shell command */
extern	iged_eq();		/* = repeat previous command */
extern	iged_in();		/* < input commands from external file */
extern	iged_mac();		/* " store, report, or execute a macro */
extern	iged_qm();		/* ? write help message */
extern	iged_out();		/* > redirect terminal output */
extern	iged_debug();		/* ` internal debug command */

extern	iged_a();		/* insert x- and y- axes */
extern	iged_b();		/* enter block mode */
extern	iged_c();		/* change or report graphic color */
extern	iged_d();		/* draw line(s) */
extern	iged_e();		/* draw data scale */
extern	iged_f();		/* change or report font */
extern	iged_g();		/* change or report graphic */
extern	iged_h();		/* read or report GIPS header */
extern	iged_i();		/* define current image */
extern	iged_j();		/* write header item value into graphic */
extern	iged_k();		/* change or report linetype */
extern	iged_l();		/* list header item value(s) */
extern	iged_m();		/* read or report secondary GIPS header */
extern	iged_n();		/* display realworld pixel values */
extern	iged_o();		/* change or report text origin */
extern	iged_p();		/* change or report float precision */
extern	iged_q();		/* return from iged (repeated) */
extern	iged_r();		/* read graphic plane or file */
extern	iged_s();		/* enter single-command mode */
extern	iged_t();		/* enter text or reposition target */
extern	iged_u();		/* undo previous graphic change */
extern	iged_v();		/* assign value to header item */
extern	iged_w();		/* write graphic plane or file */
extern	iged_z();		/* report external GIPS file values */
extern	iged_C();		/* clear current graphic */
extern	iged_D();		/* draw lines and attempt to fill */
extern	iged_F();		/* change to alternate font */
extern	iged_G();		/* logically add another graphic plane */
extern	iged_H();		/* display image histogram */
extern	iged_L();		/* insert logo into graphic */
extern	iged_N();		/* display truncated realworld pixel values */
extern	iged_O();		/* insert circle into graphic */
extern	iged_P();		/* set an image pixel */
extern	iged_R();		/* Repetitively execute iged subcommand */
extern	iged_S();		/* Draw smooth curve through points */
extern	iged_T();		/* Write coordinate value into graphic */
extern	iged_U();		/* undo previous image change */
	/*
	 *	Utility routines 
	 */
extern	char  *	iged_exp();	/* shell-expand a file name */
extern	double	iged_sca();	/* truncate a realworld pixel value */
extern	iged_atoi();		/* convert ascii to integer with test */
extern	iged_atod();		/* convert ascii incore pixel value */
extern	iged_cmd();		/* parse an input iged command */
extern	iged_im();		/* return current image pixel value */
extern	iged_line();		/* draw line in current graphic */
extern	iged_loc();		/* parse pixel location string */
extern	iged_num();		/* display pixel value */
extern	iged_org();		/* get current image origin coordinates */
extern	iged_str();		/* parse command strings */
extern	iged_tg();		/* get current Comtal target location */

extern	char *	strcpy();	/* C library */
extern	char *	strncpy();	/* C library */
extern	char *	strcat();	/* C library */
extern	char *	strncat();	/* C library */
extern	char *	index();	/* C library */
extern	off_t	lseek();	/* C library */
extern	char *	ralloc();	/* C library */
extern	char *	rindex();	/* C library */
extern	char *	gcvt();		/* C library */
	/*
	 *	Access to globals defined in iged.c
	 */
extern	char	*c;		/* input line pointer */
extern	FILE	*fio;		/* terminal output file */
extern	FILE	*fimage;	/* current input image */
extern	int	fgraphic;	/* current output graphic */
extern	char	**igenvp;	/* entry shell environment */
extern	char	*shell;		/* pointer to absolute shell pathname */
extern	char	*fioname;	/* pointer to terminal file name */
extern	char	fiodflt[];	/* pointer to default file name */
extern	char	instr[];	/* string buffer */
extern	char	headfile[];	/* last header read */
extern	char	mergefile[];	/* last header merged */
extern	char	*graphic;	/* current edited graphic */
extern	char	*graphic1;	/* last written graphic */
extern	char	incom;		/* current command letter */
extern	char	*qfile;		/* name of help file */
extern	char	*qmacs;		/* name of help macro file */
extern	char	*qcommand;	/* qfile processing command */

extern	int	pixel[];	/* target location x-y */
extern	int	ngraph;		/* last accessed graphic plane */
extern	int	nimage;		/* last accessed image plane */
extern	int	blockmode;	/* graphic refresh update mode */
extern	int	batch;		/* =1 if iged invoked with args */
extern	int	maclev;		/* macro execution stack depth */
extern	int	prolog;		/* =1 if iged executing prolog file */
extern	int	curpnt;		/* line-drawing point counter */
extern	int	precision;	/* precision of float output */
extern	int	linetype;	/* code for drawing lines */
extern	int	debug;		/* internal debugging */
extern	int	gflg;		/* general flags */

#ifdef ISAXIS
	/*
	 *	scan.flg flags
	 */
#define		A_FOUND		01
#define		A_REQD		02
#define		A_INSERT	04
#define		A_INT		010
	/*
	 *	Definition of header fields
	 */
struct scan {
	char	*name[2];	/* Variable names */	
	char	*addr;		/* address of value field */
	int	maxlen;		/* maximum string length */
	int	flg;		/* = 01 if required, = 02 if found */
};
	/*
	 *	"ipl" flag fields
	 */
#define		P_BOTTOM	01
#define		P_MIDDLE	02
#define		P_TOP		04
#define		P_LEFT		010
#define		P_CENTER	020
#define		P_RIGHT		040
#define		P_INSIDE	0100
#define		P_OUTSIDE	0200
#define		P_ENCLOSED	0400
#define		P_UNENCLOSED	01000
	/*
	 *	axis place code lookup table
	 */
struct ptab {
	char	ch;
	int	mask;
	int	exclude;
};
	/*
	 * aux axis routines
	 */
extern	iged_aline();		/* write straight line into graphic */
extern	iged_ascan();		/* parse x/yaxis string */
extern	iged_conv();		/* convert axis label to ascii */
extern	iged_fcvt();		/* convert axis label to realworld */
extern	iged_pscan();		/* parse place= field of x/yaxis string */
extern	iged_upaxis();		/* update x/yaxis header string */

#endif ISAXIS
