#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_S.c,v $	GIPS $Revision: 2.3 $ (MIT/CSR) $Date: 2004/11/29 18:54:08 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_S.c	Creation Date:	03/27/84
**	Version:	$Revision: 2.3 $		Update:		$Date: 2004/11/29 18:54:08 $
**	Components:	iged_S()	iged subcommand
**			midge()		interpolation routine
**			dot()		dot product
**
**	Function:	Join points in splined curve
*/

#include	"iged.h"

static	int	lflag;		/* logic flag */
static	int	method;		/* flag for wrap begin to end */
static	double	density;	/* type of line density */
static	double	ddens[3]	= { 1.0, 0.5, 0.25 };
static	midge();

iged_S()
{
	int	wrap	= 0;	/* flag for wrap begin to end */
	int	repeat	= 0;	/* flag for infinite repetition */
	int	quiet;		/* display update flag */

	int	xy[5][2];
	register n = 1, i;

	lflag = 1;
	method = 0;
	density = 0.5;
	quiet = maclev > 1 || batch || blockmode;

	for ( --c; n == 1; ) switch (*++c) {
	case ' ':
		break;
	case 'm':
	case 'M':
		method = *++c - '0';
		if (method < 0 || method > 2)
			return(error("bad method: %s",c));
		break;
	case 'p':
	case 'P':
		i = *++c - '0';
		if (i < 0 || i > 2)
			return(error("bad point density: %s",c));
		density = ddens[i];
		break;
	case 'q':
	case 'Q':
		quiet = 1;
		break;
	case 'r':
	case 'R':
		repeat = 1;
		break;
	case 'w':
	case 'W':
		wrap = 1;
		break;
	default:
		if (repeat)
			return(error("bad args: %s",c));
	case '\0':
		n = 0;
		break;
	}

	while (*c || repeat) {
		if (repeat)
			c = "%";
		if (iged_loc())
			return(EVIL);
		for (i = 0; i < 2; i++) {
			xy[0][i] = xy[1][i];
			xy[1][i] = xy[2][i];
			xy[2][i] = pixel[i];
		}
		if (n++ > 1) {
			if (midge(xy))
				return(EVIL);
			if (! quiet) {
				(void)(lseek(fgraphic,(off_t)0,0));
				(void)(write(fgraphic,graphic,NGRAPH));
			}
			if (lflag == 1 && wrap)
				for (i = 0; i < 2; i++) {
					xy[3][i] = xy[0][i];
					xy[4][i] = xy[1][i];
				}
			lflag = 0;
		}
	}
	if (lflag)
		return(error("insufficient points",""));
	if (wrap)
		return(midge(xy+1) || midge(xy+2));
	lflag = 2;
	return(midge(xy+1));
}

static float
dot(x,y)
	register float *x, *y;
{
	return(x[0]*y[0] + x[1]*y[1]);
}

static
midge(XY)
	register int (*XY)[2];
{
	register i;
	static	float D1[2];
	float	D0[2], D2[2], A[2], B[2], C[2], T[2];
	float	d, t, f, h, p, q, r, u, v, w, dt, dt0;
	double	sqrt();
	int	from[2], to[2];

	for (i = 0; i < 2; i++) {
		D0[i] = D1[i];
		D1[i] = (from[i] = XY[1][i]) - (to[i] = XY[0][i]);
		D2[i] = XY[2][i]-from[i];
	}

	if ((d = dot(D1,D1)) == 0)
		return(iged_line(from,to,linetype));
	d = 1/d;

	switch (lflag) {
	case 0:
		break;
	case 1:
		t = dot(D1,D2)*d;
		D0[0] = 2*t*D1[0]-D2[0];
		D0[1] = 2*t*D1[1]-D2[1];
		break;
	case 2:
		t = dot(D0,D1)*d;
		D2[0] = 2*t*D1[0]-D0[0];
		D2[1] = 2*t*D1[1]-D0[1];
		break;
	default:
		return(EVIL);
	}
	
	p = dot(D0,D0)*d;
	u = dot(D2,D2)*d;

	if (method) {
		q = dot(D0,D1)*d;
		r = (float) sqrt((double)(p*(p+2*q+1)));
		v = dot(D1,D2)*d;
		w = (float) sqrt((double)(u*(u+2*v+1)));
	}

	switch (method) {
	case 0:
		if (p == 0 || u == 0)
			return(iged_line(from,to,linetype));
		f = 6/(7*(p+(float) sqrt((double)p)));
		h = 6/(7*(u+(float) sqrt((double)u)));
		break;
	case 1:
		f = p+q+r;
		h = u+v+w;
		if (f == 0 || h == 0)
			return(iged_line(from,to,linetype));
		f = 2/f;
		h = 2/h;
		break;
	case 2:
		if (w == 0 || (f = p+q+r*(2+(u+v)/w)) == 0 ||
		    r == 0 || (h = u+v+w*(2+(p+q)/r)) == 0)
			return(iged_line(from,to,linetype));
		f = 4/f;
		h = 4/h;
		break;
	default:
		return(EVIL);
	}

	for (i = 0; i < 2; i++) {
		A[i] = f*D0[i]+f*p*D1[i];
		B[i] = -2*f*D0[i]+(3-2*f*p-h*u)*D1[i]-h*D2[i];
		C[i] = f*D0[i]+(-2+f*p+h*u)*D1[i]+h*D2[i];
		T[i] = B[i]+3*C[i];
	}

	dt = dot(T,T);
	dt0 = dot(B,B);
	if (dt0 < dt) dt = dt0;
	dt = (float) sqrt((double)(density/(4*(float) sqrt((double)dt))));

	for (t = dt; t < 1; t += dt) {
		for (i = 0; i < 2; i++) {
			from[i] = to[i];
			to[i] = XY[0][i] + (int) ((A[i]+(B[i]+C[i]*t)*t)*t);
		}
		if (to[0] != from[0] || to[1] != from[1])
			(void)(iged_line(from,to,linetype));
	}
	return(GOOD);
}
