#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_a.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:17:22 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:17:22 $
**	Components:	iged_a()	iged() subroutine
**
**	Function:	Generate axis graphics
*/

#include        <string.h>   /* for strlen */
#define		ISAXIS
#include	"iged.h"
#include	<ctype.h>
#define		MAXNUM		20
	/*
	 *	Fields filled directly from header
	 */
static	char	title[BUFLEN];	/* title string for axis */
static	char	axis[BUFLEN];	/* axis descriptor fields */
static	char	scale;		/* axis scale type */
static	int	num[2];		/* number of axis pixels */
static	int	org[2];		/* axis origin in display coords */
static	int	lim[2];		/* dimension of displayed image */
static	char	tfont[FONTLEN];	/* title font type */
	/*
	 *	Subfields of "xaxis" or "yaxis"
	 */
static	int	mark;		/* number of axis scale marks */
static	int	tic;		/* number of axis tics per scale */
static	int	line;		/* line drawing mask */
static	int	ticlen;		/* length of axis tic mark */
static	char	font[FONTLEN];	/* axis scale font type */
static	char	form[FONTLEN];	/* axis scale format */
static	char	place[FONTLEN];	/* position code(s) for titles */
static	char	ascale[2];	/* scale type */
static	char	sfrom[MAXNUM];	/* low scale value */
static	char	sto[MAXNUM];	/* high scale value */
static	char	sby[MAXNUM];	/* scale increment */
static	char	sloc[MAXNUM];	/* location on other axis */
	/*
	 *	Definition of header fields
	 */
static struct scan inscan[] = {
	"xaxis",	"",	(char *) axis,		BUFLEN,	A_INSERT,
	"xscale",	"",	(char *) &scale,	1,	A_INSERT,
	"xnum",		"",	(char *) &num[0],	0,	A_REQD | A_INT,
	"ynum",		"",	(char *) &num[1],	0,	A_REQD | A_INT,
	"xorg",		"",	(char *) &org[0],	0,	A_INT,
	"yorg",		"",	(char *) &org[1],	0,	A_INT,
	"xlim",		"",	(char *) &lim[0],	0,	A_INT,
	"ylim",		"",	(char *) &lim[1],	0,	A_INT,
	"xtitle",	"",	(char *) title,		BUFLEN,	A_INSERT,
	"xfont",	"",	(char *) tfont,		FONTLEN,A_INSERT,
	NULL
};
	/*
	 *	Definition of "axis" subfields
	 */
static struct scan axscan[] = {
	" mark	= %d",	" mark=%d",	(char *) &mark,	0,	A_INT,
	" tic	= %d",	" tic=%d",	(char *) &tic,	0,	A_INT,
	" place	= %s",	" place=%s",	(char *) place,	0,	0,
	" font	= %99s"," font=%s",	(char *) font,	0,	0,
	" form	= %99s"," form=%s",	(char *) form,	0,	0,
	" scale = %1s",	" scale=%.1s",	(char *) ascale,0,	0,
	" line	= %d",	" line=%d",	(char *) &line,	0,	A_INT,
	" ticlen = %d",	" ticlen=%d",	(char *) &ticlen, 0,	A_INT,
	" loc	= %19s"," loc=%s",	(char *) sloc,	0,	0,
	" from	= %19s"," from=%s",	(char *) sfrom,	0,	0,
	" to	= %19s"," to=%s",	(char *) sto,	0,	0,
	" by	= %19s"," by=%s",	(char *) sby,	0,	0,
	NULL
};

static	char	*axinit	= "place=cbi font=R.6";

static	int	mode;			/* = 0,1 for x,y axis */
static	int	comode;			/* = 1,0 for x,y axis */
static	int	ipl;			/* axis place flags */
	/*
	 *	axis place code lookup table
	 */
static	struct ptab ptable[] = {
	'b',	P_BOTTOM,	0,
	'm',	P_MIDDLE,	0,
	't',	P_TOP,		0,
	'l',	P_LEFT,		P_LEFT | P_RIGHT | P_CENTER,
	'r',	P_RIGHT,	P_LEFT | P_RIGHT | P_CENTER,
	'c',	P_CENTER,	P_LEFT | P_RIGHT | P_CENTER,
	'i',	P_INSIDE,	P_OUTSIDE,
	'o',	P_OUTSIDE,	P_INSIDE,
	'e',	P_ENCLOSED,	P_UNENCLOSED,
	'u',	P_UNENCLOSED,	P_ENCLOSED,
	'\0'
};

static	iged_axis();

iged_a()
{
	char	axchar;			/* actual calling subcommand */
	int	ip[3];			/* axis coordinates */
	int	jp[3];			/* scale directions +- */
	int	originsave;		/* saved origin orientation */
	char	fontsave[45];		/* saved titling font */
	double	dloc;			/* scratch */
	register struct scan *sc;	/* pointer to header desc table */
	register i;			/* scratch */

	if ((axchar = incom) == 'a')
		incom = 'x';
	mode = (incom == 'y');
	comode = 1-mode;
	/*
	 *	Default values for header items
	 */
	axis[0] = title[0] = tfont[0] = '\0';
	org[0] = org[1] = lim[0] = lim[1] = 0;
	scale = ascale[0] = 's';
	/*
	 *	Read the header fields
	 */
	for (sc = inscan; sc->name[0] != NULL; sc++) {
		if (sc->flg & A_INSERT)
			sc->name[0][0] = incom;
		if (ihget(sc->name[0],sc->addr,sc->maxlen))
			if (gipserrno == GIPSNOVAR || (sc->flg & A_REQD))
				return(gipserr());
			else
				sc->flg &= ~A_FOUND;
		else 
			sc->flg |= A_FOUND;
	}
	/*
	 *	Parse the "axis"  and interactive subfields
	 */
	for (sc = axscan; sc->name[0] != NULL; (sc++)->flg &= ~A_FOUND)
		;
	sfrom[0] = sto[0] = sby[0] = sloc[0] = form[0] = '\0';
	jp[0] = jp[1] = jp[2] = mark = tic = line = 0;
	ticlen = 4;
	if (iged_ascan(axinit,axscan) ||
	    iged_ascan(axis,axscan) || iged_ascan(c,axscan))
		return(EVIL);
	/*
	 *	Determine the ip[] locations of axis lines
	 */
	lim[0] = org[0] - 1 + (num[0] = lim[0] ? lim[0] : num[0]);
	lim[1] = org[1] - 1 + (num[1] = lim[1] ? lim[1] : num[1]);
	/*
	 *	Scan the 'place' string and set flags in 'ipl'
	 */
	if ((ipl = iged_pscan(place,ptable)) == -1)
		return(error("invalid axis 'place' field: %s",place));

	if ((ipl & P_MIDDLE) && *sloc && iged_fcvt(sloc,&dloc,ip,0,comode))
		jp[0] = (ip[0] <= 0) ? 1 : -1;
	if ((mode == 0 && (ipl & P_TOP)) || (mode == 1 && (ipl & P_BOTTOM))) {
		ip[1] = org[comode];
		if (ipl & P_INSIDE)
			jp[1] = 1;
		else if (ipl & P_OUTSIDE)
			jp[1] = -1;
		else
			jp[1] = (ip[1] <= 0) ? 1 : -1;
	}
	if ((mode == 0 && (ipl & P_BOTTOM)) || (mode == 1 && (ipl & P_TOP))) {
		ip[2] = lim[comode];
		if (ipl & P_INSIDE)
			jp[2] = -1;
		else if (ipl & P_OUTSIDE)
			jp[2] = 1;
		else
			jp[2] = (ip[2] >= COLS) ? -1 : 1;
	}
	/*
	 *	Update the "axis" header item
	 */
	(void)(iged_upaxis(axis,axscan,mode));
	/*
	 *	Call iged_axis() to do all the work.
	 *	ip[] contains the three or fewer co-axis coordinates
	 *	jp[] contains -1,1 for scales and titles below,above the image
	 */
	(void)(strcpy(fontsave,setfont((char *)NULL)));
	originsave = setorigin(-1);
	i = iged_axis(ip,jp);
	if (axchar == 'a')
		incom = 'y';
	(void)(setfont(fontsave));
	(void)(setorigin(originsave));
	return(i);
}
/*
 *	This routine does all the work of drawing titles, axes, and
 *	scale marks. The ip[] array contains the co-axis coordinates
 *	of the 3 possible scale locations, jp[] is -1,1 according to
 *	whether the scale marks and titles are below,above the axis.
 */
static
iged_axis(ip,jp)

	int	ip[3];			/* axis locations */
	int	jp[3];			/* scale directions */
{
	int	x[2];			/* scale coordinates */
	int	y[2];			/* scale limits */
	register norg;			/* origin code 0-2 */
	register i;			/* scratch */
	char	buf[FONTLEN];		/* scale number buffer */
	static	origins[2][2][2][3] =	/* setorigin() codes */
		{ 6, 7, 8, 1, 2, 3,
		  3, 5, 8, 1, 4, 6,
		  1, 4, 6, 3, 5, 8,
		  1, 2, 3, 6, 7, 8 };
	double	from;			/* float starting value */
	double	to;			/* float ending value */
	double	by;			/* float step */
	int	ifrom;			/* pixel starting value */
	int	ito;			/* pixel ending value */
	int	iby;			/* pixel stepping value */
	int	fby;			/* flag =0,-1 if "by" is fixed/float */
	int	nmark;			/* number of scale markers */
	int	height = 0;	/* height of scale markers */
	int	fontrot;		/* =0,1 for non/rotated font */
	/*
	 *	Form correct font name for scale numerics
	 */
	if (font[0] == '\0')
		return(error("missing %.1saxis font","xy"+mode));
	if (setfont(font) == NULL)
		return(error("bad axis font: %s",font));
	fontrot = (font[strlen(font)-1] == 'r');
	/*
	 *	Loop over scale marks
	 */
	nmark = (mark > 1) ? mark : 5;

	if (iged_fcvt(sfrom,&from,&ifrom,0,mode) > 0 ||
	    iged_fcvt(sto,&to,&ito,num[mode],mode) > 0 ||
	    (fby = iged_fcvt(sby,&by,&iby,num[mode]/(nmark-1),mode)) > 0)
		return(EVIL);
	ifrom += org[mode];
	ito += org[mode];
	if (fby == -1) {
		iby = 1;
		if ((by < 0 && from < to) || (by > 0 && from > to))
			by = -by;
	}
	if ((iby < 0 && ifrom < ito) || (iby > 0 && ifrom > ito))
		iby = -iby;
	
	for (x[mode] = ifrom; --nmark >= 0; ) {
		for (i = 0; i < 3; i++) {
			register j;
			double	d;

			if (jp[i] == 0)
				continue;
			x[comode] = ip[i];
			y[comode] = x[comode] + jp[i]*(ticlen+2);

			x[mode] -= (j = (x[mode] >= lim[mode]-1));
			if ((y[mode] = x[mode]) <= 0)
				norg = 0;
			else if ((mode == 0 && x[mode] >= COLS) ||
				 (mode == 1 && x[mode] >= ROWS))
				norg = 2;
			else
				norg = 1;
			(void)(iged_aline(x,y,0377));
			x[mode] += j;
			x[comode] = y[comode] + jp[i]*2;
			(void)(setorigin(origins[fontrot][mode][jp[i]==1][norg]));
			d = (fby == -1) ?
				from : iged_sca(x[mode]-org[mode],mode);
			if (d != GIPSLOW && d != GIPSHIGH) {
				iged_conv(d,buf,*ascale,form);
				(void)(memmarker(graphic,ROWS,COLS,x[1],x[0],buf));
				j = (fontrot == mode) ?
					comheight(buf) : comwidth(buf);
				if (j > height)
					height = j;
			}
		}
		/*
		 *	If requested, draw axis lines across image
		 */
		if (line) {
			y[mode] = x[mode];
			x[comode] = org[comode];
			y[comode] = lim[comode];
			(void)(iged_aline(x,y,line));
		}
		if (fby == -1) {
			from += by;
			if (gips(&from,&x[mode],1,
					mode ? GIPS_Y_VAL : GIPS_X_VAL))
				break;
			x[mode] += org[mode];
		} else
			x[mode] += iby;
		if ((iby >= 0 && x[mode] > ito+1) ||
		    (iby < 0 && x[mode] < ito-1))
			break;
	}
	/*
	 *	If font and title exist, insert the title.
	 *	Call comheight() to get max. pixel height of scale
	 */
	if (ipl & P_LEFT)
		norg = 0;
	else if (ipl & P_CENTER)
		norg = 1;
	else if (ipl & P_RIGHT)
		norg = 2;
	else
		norg = -1;
	
	if (norg >= 0 && title[0] != '\0' && tfont[0] != '\0' &&
			setfont(tfont) != NULL) {
		fontrot = (tfont[strlen(tfont)-1] == 'r');
		x[mode] = org[mode] + (norg * num[mode])/2;
		for (i = 1; i < 3; i++) {
			if (jp[i] == 0)
				continue;
			(void)(setorigin(origins[fontrot][mode][jp[i]==1][norg]));
			x[comode] = ip[i] + jp[i]*(height+ticlen+4);
			(void)(memmarker(graphic,ROWS,COLS,x[1],x[0],title));
		}
	}
	/*
	 *	If requested, draw box around image
	 */
	x[mode] = org[mode];
	y[mode] = lim[mode];
	ip[1] = org[comode];
	ip[2] = lim[comode];
	for (i = 0; i < 3; i++) {
		if ((i == 0 && jp[0] == 0) ||
		    (i > 0 && (ipl & P_UNENCLOSED)))
			continue;
		x[comode] = y[comode] = ip[i];
		(void)(iged_aline(x,y,0377));
	}
	return(GOOD);
}
