#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_aaux.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 2004/11/02 20:39:47 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_aaux.c	Creation Date:	03/19/84
**	Version:	$Revision: 2.5 $		Update:		$Date: 2004/11/02 20:39:47 $
**	Components:	iged_ascan()	Parse x/yaxis header string
**			iged_pscan()	Parse place= option of x/yaxis
**			iged_conv()	Convert scale notation to ascii
**			iged_fcvt()	Convert scale to realworld
**			iged_upaxis()	Update x/yaxis header string
**			iged_aline()	Write axis line to graphic
**
**	Function:	Auxilliary routines used by iged_a()
*/

#include        <string.h>   /* for strlen */
#define		ISAXIS
#include	"iged.h"
#include	<ctype.h>
#if __SWR
#define _ptr _p
#define _cnt _r
#endif
	/*
	 *	Use fscanf() to decode the "axis" header item and the subcommand
	 *	parameter strings. The simplest way (which may be UNIX release
	 *	dependent) is to create a dummy FILE structure, just the way
	 *	sscanf() does.
	 */
iged_ascan(s,ax)

	register char	*s;		/* "axis" string */
	struct	 scan	*ax;		/* lookup table */
{
	register FILE	*ff;		/* string 'stream' pointer */
	register struct scan *sc;	/* pointer to scan table */
#ifdef sun
	register unsigned char	*t;	/* pointer to "axis" string */
#else
	register char	*t;		/* pointer to "axis" string */
#endif
	register	i;		/* scratch */
	/*
	 *	Setup ff so fscanf() thinks "axis" is an input stream
	 */
	ff = fopens(s);

	for (sc = ax; *(t = ff->_ptr) && sc->name[0] != NULL; sc++)
		if ((i = fscanf(ff,sc->name[0],sc->addr)) == 1) {
			sc->flg |= A_FOUND;
			sc = ax - 1;
		} else if (i == EOF)
			break;
		else
			ff->_cnt = strlen(ff->_ptr = t);
	/*
	 *	Flush the remainder of the string 
	 */
	while (isspace(*t))
		t++;
	return(*t ? error("bad 'axis' field: %s",t) : GOOD);
}
/*
 *	Scan the "place" string, setting flags in "ipl"
 *	and detecting illegal combinations
 */
iged_pscan(s,pt)

	register char		*s;	/* place string */
	register struct ptab	*pt;	/* base of scan table */
{
	register int		flg;	/* returned flag field */
	register struct ptab	*p;	/* pointer to scan table */
	
	for (flg = 0; *s; flg |= p->mask, s++)
		for (p = pt; p->ch; )
			if (p->ch == *s)
				if (flg & p->exclude)
					return(-1);
				else	
					break;
			else if ((++p)->ch == '\0')
					return(-1);
	return(flg);
}
/*
 *	Convert the axis value "d" to ASCII according to scale code
 */
iged_conv(d,s,sc,form)

	double	d;			/* number to convert */
	register char *s;		/* output buffer */
	register char sc;		/* axis scale code */
	register char *form;		/* user-specified format list */
{
	if (sc == 'd' || sc == 'h') {
		register deg, min;
		register char *t;
		double	 sec;
	
		if (d < 0) { *s++ = '-'; d = -d; }
		min = 60 * (d - (deg = (int)d));
		sec = 3600 * d - 60 * (min + 60*deg);
		if (*form)
			t = form;
		else if (sec != 0.0)
			t = (sc == 'd') ? "%d\013%d'%.12g\"" : "%dh%dm%.12gs";
		else if (min != 0)
			t = (sc == 'd') ? "%d\013%d'" : "%dh%dm";
		else
			t = (sc == 'd') ? "%d\013" : "%dh";
		(void)(sprintf(s,t,deg,min,sec));
	} else if (*form)
		(void)(sprintf(s,form,d));
	else
		(void)(gcvt(d,precision,s));
}
	/*
	 *	Convert string "s" to float or fixed, then call gips()
	 *	or egips() to convert to real or pixel.
	 *	Returns:    0 - fixed value converted to realworld
	 *		   -1 -	float value converted to fixed
	 *		  > 0 -	conversion error - message generated
	 */
iged_fcvt(s,d,i,n,m)

	register char	*s;		/* string value */
	register double	*d;		/* pointer to float output */
	register int	*i;		/* pointer to integer value */
	register	n;		/* pixel value if null string */
	register	m;		/* mode value */
{
	*i = n;
	if (*s) {
		register char	*c = ".edED";	/* float indicators */

		while (*c && index(s,*c) == NULL)
			c++;
		if (*c && sscanf(s,"%lf",d) == 1)
			return(gips(d,i,1,m ? GIPS_Y_VAL : GIPS_X_VAL) ?
				gipserr() : -1);
		else if (*c || (sscanf(s,"%d",i) != 1))
			return(error("bad axis subfield value: %s",s));
	}
	return(egips(i,d,1,m ? EGIPS_Y_VAL : EGIPS_X_VAL) ? gipserr() : 0);
}
/*
 *	Generate a new "axis" string and put it into the incore GIPS header
 */
iged_upaxis(axis,sc,i)

	register char 		*axis;	/* x,y axis string */
	register struct scan 	*sc;	/* pointer to axis table */
	register		i;	/* =0,1 for x,y */
{
	for (*axis = 0; sc->name[0] != NULL; sc++) {
		if ((sc->flg & A_FOUND) == 0)
			continue;
		sc->flg &= ~A_FOUND;
		if (sc->flg & A_INT)
			(void)(sprintf(axis+strlen(axis),sc->name[1],
				*((int *)sc->addr)));
		else
			(void)(sprintf(axis+strlen(axis),sc->name[1],sc->addr));
	}
	(void)(ihput(i ? "yaxis" : "xaxis",++axis));
}

iged_aline(from,to,line)

	int	 from[2];	/* start location */
	int	 to[2];		/* stop location */
	register int line;		/* line type */
{
	register ix;		/* x-address */
	register iy;		/* y-address */
	register count;		/* dot count */

	if (debug & 1)
		(void) fprintf(fio,"%d %d %d %d ",from[0],from[1],to[0],to[1]);

	for (ix = from[0], iy = from[1], count = 0; ; ) {
		if (ISCOL(ix) && ISROW(iy) && (line == 255 ||
				((line << (count++ & 07)) & 0200)))
			SETGRAPH(graphic,ix,iy);
		if (ix != to[0])
			if (ix > to[0])
				--ix;
			else
				ix++;
		else if (iy == to[1])
			return;
		if (iy != to[1])
			if (iy > to[1])
				--iy;
			else
				iy++;
	}
}
