#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_aux.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:16:56 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_aux.c	Creation Date:	03/16/84
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:16:56 $
**	Components: 	iged_atoi(n)	convert ascii to integer
**			iged_b()	switch to block mode
**			iged_c()	change graphic color
**			iged_cmd()	issue command to Comtal
**			iged_f()	change font type
**			iged_h()	read new incore header
**			iged_k()	report or set line type
**			iged_m()	update incore header
**			iged_o()	report or set text origin
**			iged_org()	adjust pixel[] for offset images
**			iged_p()	report or set float precision
**			iged_q()	quit from iged
**			iged_s()	switch out of block mode
**			iged_str()	parse input word or string
**			iged_u()	undo previous graphic update
**			iged_C()	clear current graphic
**			iged_F()	change alternate font type
**			error(s,t)	generate error message
**			*alloc(n,m)	allocate storage
**			*ralloc(s,n)	rallocate storage
**			headmerg()	utility used by iged_h() & iged_m()
**
**	Function:	Auxiliary routines for iged
*/

#include        <stdlib.h>     /* for calloc */
#include        <string.h>     /* for strlen */
#include	<ctype.h>
#include	<signal.h>
#include	<setjmp.h>
#include	"iged.h"

extern	jmp_buf	igenv;
static	headmerg();
int	quits;

iged_atoi(n)

	register *n;
{
	register char *s = instr;
	register i = 0;

	if (*s == '+' || *s == '-')
		s++;
	while (*s)
		if (isdigit(*s))
			i = 10*i + *s++ - '0';
		else
			return(EVIL);
	*n = (instr[0] == '-') ? -i : i;
	return(GOOD);
}

iged_b()
{
	blockmode = 1;
	return(GOOD);
}

iged_c()
{
	register int i;
	register char **col;
	static char *colors[] = {
		"BLA",	"RED",	"DGR",	"LOR",	"FGR",	"ORN",	"GRN",	"YEL",
		"BLU",	"MAG",	"DTU",	"PNK",	"TUR",	"LPI",	"CYN",	"WHT",
		NULL };

	if (! *instr) {
		if ((i = color_get(ngraph)) < 0)
			return(gipserr());
		(void) fprintf(stderr,
			"c: graphic %d color is %s\n",ngraph,colors[i]);
		return(GOOD);
	}
	for (col = colors; *col != NULL; col++) {
		for (i = 0; i < 3 && (*col)[i] == toupper(instr[i]); i++)
			;
		if (i != 3 || instr[i])
			continue;
		return(color_put(ngraph,col-colors) ? gipserr() : GOOD);
	}
	return(error("unknown color: %s",instr));
}

iged_cmd()
{
	register	char *s = c;
	register	FILE *fcom;

	if ((fcom = command_open()) == NULL)
		return(gipserr());
	while (*s)
		(void)(fputc(islower(*s) ? toupper(*s++) : *s++,fcom));
	(void)(fputc('\n',fcom));
	(void)(fflush(fcom));
	if (! ferror(fcom))
		return(GOOD);
	(void)(error ("'%s' rejected",c));
	return((command_open() == NULL) ? EVIL : GOOD);
}

iged_debug()
{
	int	n;

	if (! *instr) {
		(void) fprintf(stderr,"`: %d\n",debug);
		return(GOOD);
	}
	if (iged_atoi(&n))
		return(error("bad debug code: %s",instr));
	debug = n;
	return(GOOD);
}

iged_f()
{
	register char *s;
	char	buf[BUFLEN], *getenv();

	if (! *instr) {
		register i;
		static	char *tit[4] = {
			"major font:\t%s", "italic font:\t%s",
			"bold font:\t%s", "symbol font:\t%s"
		};
		for (*buf = i = 0; i < 4; i++)
			(void)(error(tit[i],
				(s = setauxfont((char *)NULL,i)) ? s : "none"));
		return(GOOD);
	}
	if (*instr == '*') {
		*instr = ' ';
		if (! instr[1] && iged_str())
			return(EVIL);
		s = getenv("GIPSFONTDIR");
		sprintf(c = buf,"cd %s ; ls %s",s ? s : GIPSFONTDIR,instr);
		return(iged_csh());
	}
	if ((s = iged_exp(instr)) == NULL)
		return(EVIL);
	return((setfont(s) != NULL) ? GOOD : error("bad font: %s",s));
}

iged_F()
{
static	char	 *fonts[] = { "roman", "italic", "bold", "symbol" };
	register char *s;
	register i = 0;

	if (! *instr)
		return(iged_f());
	while (strncmp(instr,s = fonts[i],strlen(instr)))
		if (instr[1] == 0 && instr[0] == toupper(*s))
			break;
		else if (i++ == 3)
			return(error("bad font type: %s",instr));
	if (iged_str())
		return(EVIL);
	if (! *instr)
		if ((s = setauxfont((char *)NULL,i)) == NULL)
			return(! error("no %s font",fonts[i]));
		else
			return(! error("font:\t%s",s));
	if ((s = iged_exp(instr)) == NULL)
		return(EVIL);
	if (setauxfont(s,i) != NULL)
		return(GOOD);
	return(error("bad font: %s",instr));
}

iged_h()
{
	return(headmerg(headfile,"read"));
}

iged_C()
{
	clear(graphic,NGRAPH);
	return(GOOD);
}

iged_k()
{
	register int i = 0, j = 0, k;

	if (! *instr) {
		for (i = 8, j = linetype; i--; j >>= 1) 
			instr[i] = (j & 1) ? 'x' : '_';
		return(! error("line type: \"%.8s\"",instr));
	}
	if (isdigit(*instr)) {
		k = (*instr == '0') ? 8 : 10;
		while (isdigit(instr[i]))
			j = j*k + instr[i++] - '0';
	} else
		while (instr[i])
			j += j+ (isalpha(instr[i++]) != 0);
	if (j < 0 || j > 255 || (! isdigit(*instr) && i != 8))
		return(error("illegal linetype: %s",instr));
	linetype = j;
	(void)(setink(j ? '\377' : '\0'));
	return(GOOD);
}

iged_m()
{
	return(headmerg(mergefile,"merged"));
}

iged_o()
{
	int	n;

	if (! *instr) {
		(void) fprintf(stderr,"o: %d\n",setorigin(-1));
		return(GOOD);
	}
	if (iged_atoi(&n) || n < 0 || n > 8)
		return(error("bad origin code: %s",instr));
	(void)(setorigin(n));
	return(GOOD);
}

iged_org(i,j)

	register i, j;
{
	int	org = 0;
	int	lim;

	(void)(ihget(i ? "yorg" : "xorg", (char *) &org, 1));
	if (j) return(org);
	if (ihget(i ? "ylim" : "xlim", (char *) &lim, 1))
		lim = i ? COLS-1 : ROWS-1;
	if (pixel[i] < 0 || pixel[i] > lim)
		return(error("%.1s coordinate outside image plane","xy"+i));
	pixel[i] += org;
	return(GOOD);
}

iged_p()
{
	int	n;

	if (! *instr) {
		(void) fprintf(stderr,"p: %d\n",precision);
		return(GOOD);
	}
	if (iged_atoi(&n) || n < 0 || n > 16)
		return(error("bad precision: %s",instr));
	precision = n;
	return(GOOD);
}

iged_q()
{
	if (batch || quits != GOOD) {
		(void)(signal(SIGINT,SIG_DFL));
		(void)(signal(SIGFPE,SIG_DFL));
		(void)(signal(SIGBUS,SIG_DFL));
		(void)(signal(SIGSEGV,SIG_DFL));
		exit(0);
	}
	return(quits = error ("repeat 'q' to exit",""));
}

iged_s()
{
	blockmode = 0;
	return(GOOD);
}

iged_str()
{
	register char *s, *t;

	while (isspace(*c))
		c++;
	s =  (*c == '"') ? c++ : c;
	for (t = instr; *c; *t++ = *c++)
		if (*c == '\\' && c[1] == '"')
			c++;
		else if ((*s != '"' && isspace(*c)) ||
				(*s == '"' && *c == '"'))
			break;
	*t = 0;
	return((*s != '"' || *c++ == '"') ? GOOD :
		error ("unbalanced quotes: %s",s));
}

iged_u()
{
	register char *s = graphic;

	graphic = graphic1;
	graphic1 = s;
	return(GOOD);
}

char	*argsav[3];

error(s,t)

	register char *s, *t;
{
	if (incom)
		(void) fprintf(stderr,"%c: ",incom);
	else if (argsav[0] != NULL)
		(void) fprintf(stderr,"%s: ",argsav[0]);
	(void) fprintf(stderr,s,t);
	(void) fprintf(stderr,"\n");
	if (batch || prolog)
		exit(1);
	return(EVIL);
}

char *
alloc(n,m)

	register n, m;
{
	register char *s;
/*	char	*calloc(); */   /* in <stdlib.h> */

	if ((s = calloc((unsigned) n,(unsigned) m)) != 0)
		return(s);
	(void)(error ("no core available",""));
	longjmp(igenv,0);
	return(s);
}

char *
ralloc(s,n)

	register char *s;
	register n;
{
/*	char	*realloc();   */    /* in <stdlib.h> */

	if (s == NULL)
		return(alloc(n,1));
	if ((s = realloc(s, (unsigned)n)) != 0)
		return(s);
	(void)(error ("no core available",""));
	longjmp(igenv,0);
	return(s);
}

static
headmerg(name,title)

	register char *name, *title;
{
	register char *s;
	register FILE *fp;

	if (! *instr) return(*name ?
		! error("last gips header read/merged from: %s",name) :
		error("no header %s",title));
	if ((s = iged_exp(instr)) == NULL)
		return(EVIL);
	*name = 0;
	fp = (*title == 'r') ? ihread(s,GIPSPROFILE) : ihmerg(s);
	if (fp == NULL)
		return(gipserr());
	if (fileno(fp) > 2)
		(void)(ihbcl(fp));
	(void)(strcpy(name,s));
	return(GOOD);
}
