#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_d.c,v $	GIPS $Revision: 2.3 $ (MIT/CSR) $Date: 2004/11/29 18:54:08 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_d.c	Creation Date:	03/16/84
**	Version:	$Revision: 2.3 $		Update:		$Date: 2004/11/29 18:54:08 $
**	Components:	iged_bf()	draw and fill rectangular box
**			iged_bx()	draw rectangular box
**			iged_d()	draw polygon
**			iged_e()	draw data scale & update image
**			iged_line()	draw line
**			iged_D()	draw and fill rectangular polygon
**			iged_H()	histogram above command
**			iged_L()	insert logo into graphic
**			iged_O()	draw circle
**			iged_P()	update image pixel
**			iged_U()	undo iged_e() or iged_P()
**
**	Function:	various line-drawing routines
*/

#include	"iged.h"
#include	<ctype.h>

#define		min(x,y)	(((x)<(y))?(x):(y))
#define		max(x,y)	(((x)>(y))?(x):(y))
	/*
	 *	hist.h_flag values
	 */
#define		T_HIST		01	/* generate long histogram table */
#define		V_HIST		02	/* generate value list */
#define		R_HIST		04	/* generate rounded value list */
#define		S_HIST		010	/* generate short histogram table */
#define		N_HIST		020	/* suppress null pixels */

static	struct {
	int	h_flag;			/* set if histogram required */
	int	h_table[256];		/* value histogram array */
	int	h_lim[2*ROWS];		/* block fill x-limits */
} hist;

static	struct {
	int	xfrom;			/* lower x-limit */
	int	xlen;			/* x-raster length */
	int	yfrom;			/* lower y-limit */
	int	yto;			/* upper y-limit */
	char	*save;			/* saved image buffer */
} isave;

static	iged_blk(), iged_Hs(), iged_Ht(), sethist();

iged_bx()
{
	int	from[2], to[2];

	curpnt = 0;
	to[0] = from[0] = pixel[0];
	from[1] = pixel[1];
	if (iged_loc())
		return(EVIL);
	to[1] = pixel[1];
	if (iged_line(from,to,linetype) || iged_line(to,pixel,linetype))
		return(EVIL);
	to[0] = pixel[0];
	to[1] = from[1];
	return(iged_line(from,to,linetype) || iged_line(to,pixel,linetype));
}

iged_d()
{
	int	from[2];

	curpnt = 0;
	do	{
		from[0] = pixel[0];
		from[1] = pixel[1];
		if ((*c && iged_loc()) || iged_line(from,pixel,linetype))
			return(EVIL);
	} while (*c);
	return(GOOD);
}

iged_bf()
{
	return(iged_blk(iged_bx));
}

iged_D()
{
	return(iged_blk(iged_d));
}

static
iged_blk(proc)

	register int	(*proc)();
{
	register	i, l, *h;
	int	from[2], to[2];

	for (i = 0; i < ROWS;)
		hist.h_lim[i++] = -1;
	l = linetype;
	linetype = -1;
	i = (*proc)();
	linetype = l;

	if (i == 0)
		for (curpnt = i = 0, h = hist.h_lim; i < ROWS; i++, h++)
			if (*h >= 0) {
				from[0] = *h;
				to[0] = h[ROWS];
				from[1] = to[1] = i;
				(void)(iged_line(from,to,l ? 0377 : 0));
			}
	return(GOOD);
}

iged_L()
{
	register	i, nx, ny;
	register int	x;
	register FILE	*fp;
	int	y;
	char	*s;
	float	fact = 1;

	if (iged_str())
		return(EVIL);
	s = iged_exp(instr);
	if (s == NULL || ! *s)
		return(error("%s not found",instr));
	if ((fp = fopen(s,"r")) == NULL)
		return(gipserr2(GIPSBADOPEN,s));
	if (iged_str())
		return(EVIL);
	if (*instr && (sscanf(instr,"%f",&fact) != 1 || fact <= 0))
		return(error("bad scale factor: %s",instr));
	for (x = y = 0; (i = getc(fp)) != EOF; )
		if (i == '\n') {
			x = 0;
			ny = pixel[1]+(int)(fact * ++y);
		} else if (ISROW(ny)) do {
			nx = pixel[0] + fact * x++;
			if (ISCOL(nx))
				if (i == ' ' || i == '\t')
					UNSETGRAPH(graphic,nx,ny);
				else
					SETGRAPH(graphic,nx,ny);
		} while (i == '\t' && (x & 7) != 0);
	(void)(fclose(fp));
	return(GOOD);
}

iged_e()
{
	int	p[2];
	int	fildes;
	int	ylen;
	char	*buf;
	register char	*s;
	register int	len;
	register	y, n, pix;

	p[0] = pixel[0];
	p[1] = pixel[1];
	if (iged_bx() || iged_str())
		return(EVIL);
	if (isave.save != NULL)
		free(isave.save);
	if (! ISCOL(isave.xfrom = min(p[0],pixel[0])) ||
	    ! ISCOL(len = max(p[0],pixel[0])) ||
	    ! ISROW(isave.yfrom = min(p[1],pixel[1])) ||
	    ! ISROW(isave.yto = max(p[1],pixel[1])))
		return(EVIL);
	len = isave.xlen = len - isave.xfrom + 1;
	ylen = isave.yto - isave.yfrom + 1;
	s = buf = isave.save = alloc(len,ylen + 1);
	if (fimage != NULL) {
		(void)(fclose(fimage));
		fimage = NULL;
	}
	if ((fildes = image_open(nimage,2)) < 0)
		return(gipserr());

	for (y = isave.yfrom; y <= isave.yto; y++) {
		(void)(lseek(fildes,(off_t)TOPIX(isave.xfrom,y),0));
		(void)(read(fildes,buf += len,len));
		(void)(lseek(fildes,(off_t)TOPIX(isave.xfrom,y),0));
		if (p[0] < pixel[0] && p[1] < pixel[1])
			for (n = 0; n < len; )
				*s++ = (n++ << 8) / len;
		else if (p[0] > pixel[0] && p[1] > pixel[1])
			for (n = len; --n >= 0; )
				*s++ = (n << 8) / len;
		else {
			if (p[0] < pixel[0])
				pix = (isave.yto-y)/ylen;
			else
				pix = (y-isave.yfrom)/ylen;
			for (n = len; --n >= 0; )
				*s++ = pix;
		}
		(void)(write(fildes,s = isave.save,len));
	}
	(void)(close(fildes));
	return(GOOD);
}

iged_U()
{
	register	char	*s;
	register		fildes;
	register		y;
	register		rc;

	if ((s = isave.save) == NULL)
		return(GOOD);
	if (fimage != NULL) {
		(void)(fclose(fimage));
		fimage = NULL;
	}
	if ((fildes = image_open(nimage,1)) < 0)
		return(gipserr());
	for (rc = GOOD, y = isave.yfrom; y <= isave.yto; s += isave.xlen)
		if (lseek(fildes,(off_t)TOPIX(isave.xfrom,y++),0) == -1 ||
				write(fildes,s,isave.xlen) != isave.xlen) {
			(void)(gipserr1(GIPSBADWRITE));
			rc = EVIL;
			break;
		}
	(void)(close(fildes));
	free(isave.save);
	isave.save = NULL;
	return(rc);
}

iged_line(from,to,n)

	int	from[2];		/* starting x,y */
	int	to[2];			/* ending x,y */
	register int	n;		/* line type mask */
{
	double	l[2];			/* float x,y */
	double	d[2];			/* scratch */
	int	il[2];			/* fixed x,y */
	register	i;		/* scratch */
	register	j;		/* scratch */
	register int	sw;		/* point index */

	if ((debug & 1) && n >= 0)
		(void) fprintf(fio,"%d %d %d %d ",from[0],from[1],to[0],to[1]);

	for (i = 0; i < 2; i++) {
		l[i] = il[i] = from[i];
		d[i] = to[i] - from[i];
	}
	i = (d[0]*d[0] > d[1]*d[1]) ? 0 : 1;
	j = (d[i] > 0) ? 1 : -1;
	if (d[i] == 0.0) j = 0;
	else d[0] = (d[1-i]/d[i])*j;
	for (sw = curpnt; ; sw++, il[i] += j) {
		if (ISCOL(il[0]) && ISROW(il[1]))
			if (n < 0) {
				register int *h = hist.h_lim+il[1];
				if (*h < 0)
					*h = h[ROWS] = il[0];
				else if (*h > il[0])
					*h = il[0];
				else if (h[ROWS] < il[0])
					h[ROWS] = il[0];
			} else {
				if (n == 255 || ((n << (sw & 07)) & 0200))
					SETGRAPH(graphic,il[0],il[1]);
				else if (n == 0)
					UNSETGRAPH(graphic,il[0],il[1]);
				if (hist.h_flag && sethist(il,sw))
					return(EVIL);
			}
		if (il[i] == to[i]) {
			curpnt = sw;
			return(GOOD);
		}
		il[1-i] = (l[1-i] += d[0]);
	}
}

iged_O()
{
	double	rad, th, incr, acos(), sin(), cos(), sqrt();
	int	from[2], to[2], cen[2];

	if (! *c)
		return(error("no point(s) defined",""));
	cen[0] = pixel[0];
	cen[1] = pixel[1];
	if (iged_loc())
		return(EVIL);
	to[0] = pixel[0] - cen[0];
	to[1] = pixel[1] - cen[1];
	rad = sqrt((double)(to[0]*to[0] + to[1]*to[1]));
	if (rad == 0)
		return(error("radius is nil",""));
	incr = acos((rad-0.5)/rad);
	from[0] = cen[0] + rad;
	from[1] = cen[1];
	for (curpnt = th = 0; (th += incr) < 6.2831852; ) {
		to[0] = cen[0] + rad*cos(th);
		to[1] = cen[1] + rad*sin(th);
		if (iged_line(from,to,linetype))
			return(EVIL);
		from[0] = to[0];
		from[1] = to[1];
	}
	return(GOOD);
}

iged_H()
{
	register	char *s;

	if (! *instr) {
		if (hist.h_flag & S_HIST)
			(void)(iged_Hs());
		if (hist.h_flag & T_HIST)
			(void)(iged_Ht());
		hist.h_flag = 0;
		return(GOOD);
	}

	for (s = instr; *s; s++)
		switch (*s) {
		case ' ':
		case '\t':
		case '\n':
			break;
		case 's':
			hist.h_flag |= S_HIST;
			clear((char *)hist.h_table,256*sizeof(int));
			break;
		case 't':
			hist.h_flag |= T_HIST;
			clear((char *)hist.h_table,256*sizeof(int));
			break;
		case 'v':
			hist.h_flag &= ~R_HIST;
			hist.h_flag |= V_HIST;
			break;
		case 'r':
			hist.h_flag &= ~V_HIST;
			hist.h_flag |= R_HIST;
			break;
		case 'n':
			hist.h_flag |= N_HIST;
			break;
		case 'z':
			hist.h_flag &= ~N_HIST;
			break;
		default:
			return(error("illegal code: %s",instr));
		}
	return(GOOD);
}

static
iged_Hs()
{
	register	i, n;
	register char	*s;
	register int	*h, *t;
	int	inc, tab[64];
	char	buf[65], *u;

	for (i = n = 0, h = hist.h_table, t = tab; i < 64; i++, h += 4, t++)
		if ((*t = h[0]+h[1]+h[2]+h[3]) > n) n = *t;
	if (n == 0)
		return(! error("no points accumulated",""));

	for (inc = (n+19)/20; n > 0; n -= inc) {
		for (i = 0, s = buf, t = tab; i < 64; i++) {
			if (*t++ >= n)
				*s++ = '*', u = s;
			else
				*s++ = ' ';
		}
		*u = 0;
		(void) fprintf(fio,"%5d:\t%s\n",n,buf);
	}
	(void) fprintf(fio,
	"\t|----+----+----+----+----+----+----+----+----+----+----+----+--|\n");
	(void) fprintf(fio,
	"\t0    20   40   60   80  100  120  140  160  180  200  220  240\n");
	(void)(fflush(fio));
	return(GOOD);
}

static
iged_Ht()
{
	register	j, n;
	register int	*h;
	int	i;
	double	d;
	char	buf[30];

	for (i = n = 0, h = hist.h_table; i < 256; i++, h++)
		if (*h > n) n = *h;
	if (n == 0)
		return(! error("no points accumulated",""));
	fputs("real       pixel count\n",fio);
	for (i = 0, h = hist.h_table; i < 256; i++, h++) {
		if (*h == 0)
			continue;
		if (*headfile && ! egips(&i,&d,1,EGIPS_D_VAL))
			(void)(gcvt(d,precision,buf));
		else
			buf[0] = 0;
		(void) fprintf(fio,"%12.12s %3d %5d: ",buf,i,*h);
		for (j = (48* *h)/n; j >= 8; j -= 8)
			(void)(putc('\t',fio));
		while (--j >= 0)
			(void)(putc(' ',fio));
		fputs("*\n",fio);
	}
	(void)(fflush(fio));
	return(GOOD);
}

static
sethist(il,sw)

	register	int *il;		/* pixel location */
	register	sw;		/* point count */
{
	if (sw && sw == curpnt)
		return(GOOD);
	if ((sw = iged_im(il)) < 0 || fimage == NULL)
		return(EVIL);
	if ((hist.h_flag & N_HIST) && sw == 0)
		return(GOOD);
	if (hist.h_flag & (S_HIST | T_HIST))
		hist.h_table[sw]++;
	if (hist.h_flag & V_HIST)
		return(iged_num(il,0,sw));
	return((hist.h_flag & R_HIST) ?
		iged_num(il,1,sw) : GOOD);
}

iged_P()
{
	register	fildes;
	register long	off;
	char	newch, oldch;

	if (iged_atod(&newch))
		return(EVIL);
	if (*c)
		return(error("extraneous text: %s",c));
	if (fimage != NULL) {
		(void)(fclose(fimage));
		fimage = NULL;
	}
	if ((fildes = image_open(nimage,1)) == -1)
		return(gipserr());
	off = TOPIX(pixel[0],pixel[1]);
	if (lseek(fildes,(off_t)off,0) == -1 || read(fildes,&oldch,1) != 1)
		return(gipserr1(GIPSBADREAD));
	if (lseek(fildes,(off_t)off,0) == -1 || write(fildes,&newch,1) != 1)
		return(gipserr1(GIPSBADWRITE));
	(void)(close(fildes));
	if (isave.save == NULL)
		isave.save = alloc(1,sizeof(char));
	isave.xfrom = pixel[0];
	isave.yfrom = isave.yto = pixel[1];
	isave.xlen = 1;
	isave.save[0] = oldch;
	return(GOOD);
}
