#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_g.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:17:27 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_g.c	Creation Date:	03/16/84
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:17:27 $
**	Components:	iged_g()	switch/report current graphic
**			iged_i()	switch/report current image
**			iged_r()	Read a graphic
**			iged_w()	Write a graphic
**			iged_rw()	Read/Write a graphic
**			iged_G()	Logically add another graphic
**			iged_im()	Return image pixel value
**
**	Function:	allocate, delete, and switch incore graphics/image
*/

#include	<ctype.h>
#include	"iged.h"

static	char *graphtab[129];		/* table of graphic pointers */
static	iged_rw();

iged_g()
{
	int	i;
	register	char *s;

	if (! *instr) {
		(void) fprintf(stderr,"%c: graphics:",incom);
		for (i = 0; i++ < 128; )
			if (graphtab[i] != NULL || i == ngraph)
				(void) fprintf(stderr,
					" %d%s",i,(i == ngraph)?"*":"");
		(void)(putc('\n',stderr));
		return(GOOD);
	}
	if (iged_atoi(&i) || i*i == 0 || i*i > 128*128)
		return(error("bad graphic number: %s",instr));
	if (i < 0)
		if (i == -ngraph)
			return(error("graphic %s is active",instr));
		else if ((s = graphtab[-i]) == NULL)
			return(error("graphic %s not defined",instr));
		else {
			graphtab[-i] = NULL;
			free(s);
			return(GOOD);
		}
	if (i == ngraph)
		return(((i = color_get(ngraph)) < 0 ||
		    (i == 0 && color_put(ngraph,15) < 0)) ? gipserr() : GOOD);
	if ((s = graphtab[ngraph]) == NULL)
		s = graphtab[ngraph] = alloc(NGRAPH,1);
	move(graphic,s,NGRAPH);
	if ((s = graphtab[ngraph = i]) != NULL)
		move(s,graphic,NGRAPH);
	if (fgraphic >= 0)
		(void)(close(fgraphic));
	if ((i = graphic_open(ngraph,0)) < 0)
		return(gipserr());
	(void)(read(i,graphic,NGRAPH));
	(void)(close(i));
	fgraphic = graphic_open(ngraph,1);
	gflg |= UP;
	return((fgraphic < 0) ? gipserr() : GOOD);
}

iged_G()
{
	register char	*s, *d;
	register	flg, m, n;

	for (n = 0, d = c; isdigit(*c); n = 10 * n + *c++ - '0')
		;
	if (n <= 0 || n > 128)
		return(error("bad graphic number: %s",d));
	while (isspace(*c)) c++;
	switch (*c) {
	case '|':	flg = 0;	break;
	case '&':	flg = 1;	break;
	case '^':	flg = 2;	break;
	case '=':	flg = 3;	break;
	default:
		return(error("bad operator: %s",c));
	}
	for (*c++ = 0; isspace(*c); c++);
	m = (*c == '~') ? 0377 : 0;
	if ((s = graphtab[n]) == NULL)
		return(error("graphic %s not defined",d));
	for (d = graphic, n = NGRAPH; --n >= 0; )
		switch (flg) {
		case 0:	*d++ |= *s++ ^ m; break;
		case 1:	*d++ &= *s++ ^ m; break;
		case 2:	*d++ ^= *s++ ^ m; break;
		case 3:	*d++  = *s++ ^ m; break;
		}
	return(GOOD);
}

iged_i()
{
	char		buf[10];
	int		ni;

	if (! *instr) {
		sprintf(buf,"%d",nimage);
		return(! error("current image: %s",buf));
	}

	if (iged_atoi(&ni) || ni < 1 || ni > 128)
		return(error("bad image number: %s",instr));

	if (fimage != NULL) {
		(void)(fclose(fimage));
		fimage = NULL;
	}
	nimage = ni;
	return(GOOD);
}

iged_im(xy)

	register	int	*xy;
{
	register		i;

	if (fimage == NULL)
		if ((i = image_open(nimage,0)) < 0 ||
		    (fimage = fdopen(i,"r")) == NULL) {
			(void)(gipserr());
			return(-1);
		}
	if (! ISCOL(xy[0]) || ! ISROW(xy[1]) ||
	    fseek(fimage,(long) TOPIX(xy[0],xy[1]),0) < 0)
		return(-1);
	return(getc(fimage));
}

iged_r()
{
	return(iged_rw(0));
}

iged_w()
{
	register char	*s = instr;

	if (*instr != '!')
		return(iged_rw(1));

	if (! instr[1]) {
		if (iged_str())
			return(EVIL);
	} else	{
		while (isspace(*++s))
			;
		(void)(strcpy(instr,s));
	}
	return(iged_rw(2));
}

static
iged_rw(sw)

	register int	sw;
{
	register	count;
	register	fildes;
	register	char *s = NULL;
	int		fnum;

	while (*c)
		if (! isspace(*c++))
			return(error("extraneous argument(s): %s",--c));

	if (! *instr)
		fnum = ngraph;
	else if (iged_atoi(&fnum) || fnum < 0 || fnum > 128) {
		if ((s = iged_exp(instr)) == NULL)
			return(EVIL);
		s = strcpy(instr,s);
		if (sw == 1 && access(s,0) == 0)
			return(error("file already exists - use w!%s",s));
		if ((fildes = sw ? creat(s,0644) : open(s,0)) < 0)
			return(gipserr2(GIPSBADOPEN,s));
		else
			swab(graphic,graphic,NGRAPH);
	}

	if (s == NULL && (fildes = graphic_open(fnum,sw)) < 0)
		return(gipserr());

	count = sw ? write(fildes,graphic,NGRAPH) : read(fildes,graphic,NGRAPH);

	if (s != NULL) {
		swab(graphic,graphic,NGRAPH);
		gflg &= ~UP;
	}
	(void)(close(fildes));
	if (count == NGRAPH)
		return(GOOD);
	return((s != NULL) ? gipserr2(GIPSBADWRITE,s) : gipserr1(GIPSBADWRITE));
}
