#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_l.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:17:28 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged.c		Creation Date:	03/29/84
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:17:28 $
**	Components:	iged_l()	Display header or header items
**			iged_j()	Insert header variable in graphic
**			putitem()	Return header variable value
**			iged_v()	Set value of header variable
**			iged_n()	Display image pixel value
**			iged_N()	Display truncated pixel value
**			iged_num()	Display pixel value
**			iged_t()	Set target or insert text
**			iged_tg()	Locate target
**			iged_T()	Write truncated value to graphic
**			iged_sca()	Get truncated pixel value
**
**	Function:	Miscellaneous iged subcommands
*/

#include        <string.h>   /* for strlen */
#include	<ctype.h>
#include	"iged.h"

static	int	egips_mode[3]	= { EGIPS_X_VAL, EGIPS_Y_VAL, EGIPS_D_VAL };
static	putitem();

iged_l()
{
	char	buf[BUFLEN];

	if (! *instr) {
		sprintf(buf,"%d",nimage);
		(void) printf("%c:\tcurrent output stream\t: %s\n\
	current graphic plane\t: %d\n\
	current image plane\t: %s\n\
	incore GIPS header\t: %s\n\
	last merged header\t: %s\n\
	iged online help file\t: %s\n\
	online help command\t: %s\n\
	command shell pathname\t: %s\n\
	assumed target location\t: (x=%d,y=%d)\n\
	current iged mode\t: (%sblock,%sbatch,%sprolog)\n\
	macro execution level\t: %d\n\
	line drawing mask\t: (%d,%d)\n\
	floating precision\t: %d\n\
	internal iged flags\t: 0x%x debug=%d\n\
	device control pseudocodes\n",
		incom,fioname,ngraph,buf,
		headfile,mergefile,qfile,qcommand,shell,pixel[0],pixel[1],
		blockmode ? "":"no", batch ? "":"no", prolog ? "":"no",
		maclev,linetype,curpnt,precision,gflg,debug);
		print_subtab("\t\t",stdout);
		(void)(fflush(stdout));
		return(GOOD);
	}
	if (*instr == '*') {
		register	char *s, *t;
		register	FILE *fo;

		s = (instr[1]) ? instr+1 : c;
		while (isspace(*s))
			s++;
		for (t = s; *t; )
			if (isspace(*t++))
				*--t = 0;
		s = *s ? iged_exp(s) : "-";
		if (*s && *s != '-') {
			sprintf(buf,"iged: -l * %s",s);
			(void)(ihput("history",buf));
		}
		fo = ihwrit(s,1);
		(void)(ihdel("history"));
		if (fo == NULL)
			return(gipserr());
		else if (fileno(fo) > 2)
			(void)(ihbcl(fo));
		return(GOOD);
	}
	if (! headfile) 
		return(gipserr1(GIPSNOHEAD));
	do {
		if (! putitem(buf,sizeof(buf)))
			(void) fprintf(fio,"%s:\t%s\n",instr,buf);
	} while (! iged_str() && *instr);
	(void)(fflush(fio));
	return(GOOD);
}

iged_j()
{
	char	buf[BUFLEN];

	if (*c)
		return(error("too many variable names: %s",c));
	return(putitem(c = buf,sizeof(buf)) ? EVIL : iged_t());
}
		
static
putitem(s,n)

	register char	*s;
	int		n;
{
	double	d;
	char	sw;

	if (ihtype(instr,&sw,1))
		return(gipserr());
	switch (sw) {
	case 'c':
	case 'R':
	case 'h':
	case 's':
		if (ihget(instr,s,n))
			return(gipserr());
		if (sw == 'c')
			*++s = 0;
		break;
	case 'i':
		if (ihget(instr,(char *)&n,0))
			return(gipserr());
		(void)(sprintf(s,"%d",n));
		break;
	case 'd':
		if (ihget(instr,(char *)&d,0))
			return(gipserr());
		(void)(gcvt(d,precision,s));
		break;
	default:
		return(error("cannot display: %s",instr));
	}
	return(GOOD);
}

iged_v()
{
	char	buf[BUFLEN];
	int	i;
	double	d;
	register char	*item;

	while (*c) {
		while (isspace(*c))
			c++;
		if (! *(item = c++))
			return(GOOD);
		while (! isspace(*c) && *c && *c != '=')
			c++;
		i = *c;
		*c++ = 0;
		if (ihtype(item,buf,sizeof(buf)))
			;
		else if (i != '=')
			(void) fprintf(fio,"%s\t%c\t%s\n",item,*buf,buf+1);
		else if (iged_str())
			return(EVIL);
		else if (! *instr)
			return(ihdel(item) ? gipserr() : GOOD);
		else {
			gipserrno = GIPSBADVAL;
			if (*buf == 'i' && sscanf(instr,"%d",&i) == 1)
				(void)(ihput(item,(char *)&i));
			else if (*buf == 'd' && sscanf(instr,"%lf",&d) == 1)
				(void)(ihput(item,(char *)&d));
			else if (*buf == 's' || *buf == 'h' || *buf == 'c')
				(void)(ihput(item,instr));
			else
				gipserrno = GIPSBADTYPE;
		}
		if (gipserrno)
			return(gipserr());
	}
	(void)(fflush(fio));
	return(GOOD);
}
	
iged_n()
{
	register	pix;

	return(((pix = iged_im(pixel)) < 0) ? EVIL : iged_num(pixel,0,pix));
}

iged_N()
{
	register	pix;

	return(((pix = iged_im(pixel)) < 0) ? EVIL : iged_num(pixel,1,pix));
}

iged_num(loc,sw,pix)

	register	int *loc;
	register	sw;
	register	pix;
{
	double	d[2];
	int	var[2];
	char	buf[31];
	register int	i;
	register int	j;

	for (buf[0] = ' ', i = 0; i < 3; i++) {
		var[1] = 1 + (var[0] = (i==2) ? pix : (loc[i] - iged_org(i,1)));
		(void) fprintf(fio,"%c: %d","xyd"[i],var[0]);
		if (! *headfile || (sw == 0 && egips(var,d,2,egips_mode[i]))) {
			fputs(" NUL NUL",fio);
		} else for (j = sw; j < 2; j++) {
			if (sw)
				d[1] = iged_sca(var[0],i);
			if (d[j] == GIPSLOW)
				fputs(" MIN",fio);
			else if (d[j] == GIPSHIGH)
				fputs(" MAX",fio);
			else
				fputs(gcvt(d[j],precision,buf+1)-1,fio);
		}
		(void)(putc((i == 2) ? '\n' : ' ',fio));
	}
	(void)(fflush(fio));
	return(GOOD);
}

iged_tg()
{
	register	i, j;

	if (batch || prolog)
		return(error("illegal target use",""));
	if (target_roam(0))
		return(gipserr());
	(void) printf("# ");
	for (j = 0; (i = getchar()) != '\n'; j++)
		if (i == EOF)
			exit(0);
	if (j) {
		(void)(command_open());
		return(EVIL);
	}
	return((command_open() == NULL || target_get(nimage,pixel)) ?
		gipserr() : GOOD);
}

iged_t()
{
	register	i;

	if (*c) {
		if (maclev <= 1 && ! batch)
			move(graphic,graphic1,NGRAPH);
		return(! memmarker(graphic,ROWS,COLS,pixel[1],pixel[0],c));
	}
	i = target_put(nimage,pixel);
	if (i > 0)
		return(error("bad target coordinate(s)",""));
	else
		return(i ? gipserr() : GOOD);
}

iged_T()
{
	double	d;	/* realworld value */
	register	pix;
	char	buf[BUFLEN];

	if (! *c)
		return(error("no keyword `xyd' specified",""));
	for (buf[0] = 0; *c; c++) {
		if (*c == 'd')
			if ((pix = iged_im(pixel)) < 0)
				return(EVIL);
			else
				d = iged_sca(pix,2);
		else if (*c == 'x')
			d = iged_sca(pixel[0]-iged_org(0,1),0);
		else if (*c == 'y')
			d = iged_sca(pixel[1]-iged_org(1,1),1);
		else
			return(error("illegal keyword: %s",c));
		if (d == GIPSLOW)
			(void)(strcat(buf," MIN"));
		else if (d == GIPSHIGH)
			(void)(strcat(buf," MAX"));
		else
			(void)(gcvt(d,precision,buf+strlen(buf)));
	}
	return(! memmarker(graphic,ROWS,COLS,pixel[1],pixel[0],buf));
}

double
iged_sca(pix,mode)

	int	pix;
	int	mode;
{
	int	ipix[2];
	double	dpix[2];
	char	buf[46];
	register int	i0;
	register	i1;
	register	char *s;
	register	char *t;

	ipix[1] = 1 + (ipix[0] = pix);
	if (egips(ipix,dpix,2,egips_mode[mode]))
		return(gipserr());
	if (dpix[0] == GIPSLOW || dpix[0] == GIPSHIGH || dpix[0] == dpix[1])
		return(dpix[0]);
	if (dpix[1] == GIPSLOW || dpix[1] == GIPSHIGH)
		return(dpix[1]);
	i0 = ((dpix[0] > dpix[1] && dpix[0] > 0) || 
		(dpix[0] < dpix[1] && dpix[0] < 0));
	(void)(sprintf(buf,"%22.15e %22.15e",dpix[i0],dpix[1-i0]));
	if (*(s = buf) != *(t = buf+23))
		return(0);
	i0 = 10*s[20]+s[21]-11*'0';
	if (s[19] == '-') i0 = -i0;
	i1 = 10*t[20]+t[21]-11*'0';
	if (t[19] == '-') i1 = -i1;
	if (i0 == i1)
		s[2] = s[1];
	else {
		s[2] = '0';
		s[3] = (i1 > ++i0) ? '0' : s[1];
	}
	for (t[2] = t[1], s[18] = '0', ++s, ++t; *++s == *++t; )
		;
	if (*t != 'e')
		if (*t > *s + 1)
			*s = (*s < '5' && *t > '5') ? '5' :
				(*s + *t) >> 1;
		else if (t[1] == '0' && s[1] != '9') {
			s++;
			*s = (*s < '5') ? '5' : (*s + '0' + 10) >> 1;
		} else
			(*s)++;
	buf[1] = buf[2];
	buf[2] = '.';
	(void)(strcpy(++s,buf+41));
	(void)(sscanf(buf,"%lf",dpix));
	return(dpix[0]);
}
