#ifndef lint
static char *sccsid = "%Z%$RCSfile: iged_loc.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1989/07/19 10:33:51 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		iged_loc.c	Creation Date:	03/16/84
**	Version:	$Revision: 2.2 $		Update:		$Date: 1989/07/19 10:33:51 $
**	Components:	iged_loc()	iged subroutine
**
**	Function:	decode specified image pixel address
*/

#include	"iged.h"

#define		ISSPACE		01
#define		ISCOMMA		02
#define		ISMETA		04
#define		ISARITH		010
#define		ISLOGIC		020
#define		ISMINUS		040
#define		ISMIN		ISMINUS | ISARITH
#define		ISLT		ISMINUS | ISLOGIC
#define		ISEXP		0100
#define		ISDIGIT		0200
#define		ISLEGAL		0377

#define		TEST(x,y)	(types[(x)&0377]&(y))

static	char	types[128] = {
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	ISSPACE,ISSPACE,0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	ISSPACE,0,	0,	0,	0,	ISMETA,	0,	0,
	0,	0,	0,	ISARITH,ISCOMMA,ISMIN,	0,	0,
	ISDIGIT,ISDIGIT,ISDIGIT,ISDIGIT,ISDIGIT,ISDIGIT,ISDIGIT,ISDIGIT,
	ISDIGIT,ISDIGIT,ISLOGIC,0,	ISLT,	0,	ISLOGIC,0,
	ISMETA,	0,	0,	0,	ISEXP,	ISEXP,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	ISEXP,	ISEXP,	0,	0
};

iged_loc()
{
	double		x[2];
	register	i;
	register char	*s;
	char		sign[2], esign[2];

	for (s = c, i = 0; i < 2; i++) {
		register	e;
		register int	f, ok = -1;

		x[i] = e = f = sign[i] = esign[i] = 0;

		while (TEST(*s,ISCOMMA | ISSPACE))
			s++;
		if (TEST(*s,ISMETA)) {
			sign[0] = sign[1] = *s++;
			esign[1] = x[1] = 0;
			break;
		}
		if (TEST(*s,ISARITH | ISLOGIC))
			sign[i] = *s++;
		while (TEST(*s,ISDIGIT))
			x[i] = ok = 10 * x[i] + *s++ - '0';
		if (ok != -1 && (! *s || TEST(sign[i],ISLOGIC)))
			continue;
		if (*s == '.')
			for (esign[i] = *s++; TEST(*s,ISDIGIT); f--)
				x[i] = ok = 10 * x[i] + *s++ - '0';
		if (ok == -1)
			return(error("missing target location: %s",c));
		while (TEST(*s,ISEXP)) {
			if (! TEST(s[1],ISDIGIT))
				if (TEST(s[1],ISARITH) && TEST(s[2],ISDIGIT))
					s++;
				else
					break;
			esign[i] = *s++;
			while (TEST(*s,ISDIGIT))
				e = 10*e + *s++ - '0';
			break;
		}
		e = (esign[i] == '+') ? f+e : f-e;
		if (e > 0)
			while (e--)
				x[i] *= 10.0;
		else
			while (e++)
				x[i] *= 0.1;
	}

	while (TEST(*s,ISSPACE))
		s++;
	c = s;

	if (sign[0] == '@')
		return(GOOD);

	if (((TEST(sign[0],ISMETA | ISARITH) && ! esign[0]) ||
		(TEST(sign[1],ISMETA | ISARITH) && ! esign[1])) && iged_tg())
		return(EVIL);

	for (i = 0; i < 2; i++) {
		if (TEST(sign[i],ISMINUS))
			x[i] = -x[i];
		if (esign[i]) {
			if (! *headfile)
				return(gipserr1(GIPSNOHEAD));
			if (gips(&x[i],pixel+i,1,i ? GIPS_Y_VAL : GIPS_X_VAL))
				return(gipserr());
		} else if (sign[i]) {
			if (sign[i] == ':')
				pixel[i] = 0;
			pixel[i] += x[i];
			continue;
		} else
			pixel[i] = x[i];
		if (iged_org(i,0))
			return(EVIL);
	}
	return(GOOD);
}

iged_atod(ival)

	char		*ival;
{
	register char	*s = c;
	register	e, f;
	register int	ok = -1;
	char		sign, esign;
	double		d;

	d = e = f = sign = esign = 0;

	while (TEST(*s,ISSPACE))
		s++;
	if (TEST(*s,ISARITH))
		sign = *s++;
	while (TEST(*s,ISDIGIT))
		d = ok = 10 * d + *s++ - '0';
	if (*s == '.')
		for (esign = *s++; TEST(*s,ISDIGIT); f--)
			d = ok = 10 * d + *s++ - '0';
	if (ok == -1)
		return(error("bad data value: ",c));
	while (TEST(*s,ISEXP)) {
		if (! TEST(s[1],ISDIGIT))
			if (TEST(s[1],ISARITH) && TEST(s[2],ISDIGIT))
				s++;
			else
				break;
		esign = *s++;
		while (TEST(*s,ISDIGIT))
			e = 10*e + *s++ - '0';
		break;
	}
	e = (esign == '+') ? f+e : f-e;
	if (e > 0)
		while (e--)
			d *= 10.0;
	else
		while (e++)
			d *= 0.1;
	if (sign == '-')
		d = -d;
	while (TEST(*s,ISSPACE))
		s++;
	c = s;
	*ival = 0;
	if (! esign) {
		if ((e = (int)d) < 0 || e > 255)
			return(error("bad pixel value: %d",(char *)e));
		*ival = e;
		return(GOOD);
	}
	return(gips(&d,(int *)ival,1,GIPS_D_DVAL) ? gipserr() : GOOD);
}
