#ifndef lint
static char *sccsid = "%Z%$RCSfile: get_table.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1998/05/13 01:20:00 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1985 Massachusetts Institute of Technology
**
**	Module:		get_table.c	Creation Date:	04/23/85
**	Version:	$Revision: 2.2 $		Update:		$Date: 1998/05/13 01:20:00 $
**	Components:	get_table()	iha subroutine
**
**	Function:	Construct an image descriptor table
*/

#include	"iha.h"

TABLE *
get_table(file)

	char	*file;		/* name of input file - "-" for stdin
				   or NULL if this is to be the output file */
{
	register TABLE *t;	/* pointer to new table */
	register i;		/* scratch */
	register char *s;	/* scratch */
	register float *d;	/* copy of t->dpoly */
	register FILE *ff;	/* string 'stream' pointer */

	static	char cstdin[] = "'stdin'";
	static	char cstdout[] = "'stdout'";
	char	dparm[BUFLEN];	/* polynomial coefficients */
	int	numx;		/* number of x pixels */
	int	numy;		/* number of y pixels */

	t = (TABLE *) alloc(1,sizeof(struct table));
	/*
	 *	read or write the image header
	 */
	if (file == NULL) {
		if (!pflag && (t->iop = ihwrit(ofile,0)) == stdout) {
			(void)(signal(SIGPIPE,cleanup));
			t->fname = cstdout;
		}
	} else
		if ((t->iop = ihread(t->fname = file,GIPSPROFILE)) == stdin)
			t->fname = cstdin;
	if (!pflag && t->iop == NULL)
		gipserr();
	/*
	 *	look-up vital header variables
	 */
	if (ihget("dscale",&t->dscale,1))
		t->dscale = 's';
	if (ihget("xnum",(char *)&numx,0) ||
	    ihget("ynum",(char *)&numy,0) ||
	    ihget("dtype",&t->dtype,1))
		gipserr();
	t->dlen = ihdtype(t->dtype);
	/*
	 *	all xnum,ynum variables must be identical
	 */
	if (xnum == -1)
		xnum = numx, ynum = numy;
	else if (xnum != numx)
		error("different xnum: %s",t->fname);
	else if (ynum != numy)
		error("different ynum: %s",t->fname);
	/*
	 *	save possible scaling coefficients
	 */
	switch (t->dscale) {
	case 'p':
		if (file == NULL)
			error("bad dscale=p for %s",t->fname);
		if (ihget("dpoly",(char *)&t->npoly,0) ||
	  	    ihget("dparm",dparm,sizeof(dparm)))
			gipserr();
		d = t->dpoly = (float *) alloc(i = 1 + t->npoly,sizeof(float));
		ff = fopens(dparm);
		while (--i >= 0)
			if (fscanf(ff,"%f",d++) != 1)
				gipserr1(GIPSBADVAL);
		t->flgs |= CONV|POLY;
		break;
	case 'n':
		if (ihget("dfact",(char *)&t->dfact,0) ||
		    ihget("dzero",(char *)&t->dzero,0))
			gipserr();
		t->flgs |= NORM|CONV;
		break;
	case 'f':
		error("dscale=f unsupported: %s",t->fname);
	default:
		break;
	}
	/*
	 *	look-up C language name for pixel type
	 */
	switch (t->dtype) {
		case 'b':	s = "unsigned char";	break;
		case 's':	s = "char";		break;
		case 'h':	s = "short";		break;
		case 'u':	s = "unsigned short";	break;
		case 'i':	s = "long";		break;
		case 'l':	s = "unsigned long";	break;
		case 'c':	t->flgs &= ~(CONV|NORM|POLY);
		case 'f':	s = "float";		break;
		case 'd':	s = "double";		break;
		default:	error("%s: bad dtype",t->fname);
	}
	(void)(strcpy(t->vartype,(t->flgs & CONV) ? "float" : s));
	/*
	 *	allocate buffers, allowing enough for lookup-raster
	 *	if it is possible to use a lookup table.
	 */
	if (file == NULL)
		;
	else if (t->dlen > 2)
		byteflag = 0;
	else if (byteflag)
		if (t->dlen == 1 && xnum < 256)
			numx = 256;
		else if (t->dlen == 2)
			if (byteflag == 2 || xnum >= 256*256)
				byteflag = 0;
			else
				byteflag = 2, numx = 256*256;
	t->buffer = alloc(t->dlen,numx);
	if (t->flgs & CONV) {
		t->plen = sizeof(float);
		t->fbuffer = (float *)alloc(sizeof(float),numx);
	} else
		t->plen = t->dlen;

	return(t);
}
