#ifndef lint
static char *sccsid = "%Z%$RCSfile: iha.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/11/13 23:11:12 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1985 Massachusetts Institute of Technology
**
**	Module:		iha.c		Creation Date:	04/23/85
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/11/13 23:11:12 $
**	Components:	main()		UNIX command
**
**	Function:	Performs pixel-by-pixel image transformation
*/

#include	"iha.h"
	/*
	 *	variables set by command flags and options
	 */
char	*afile	=	NULL;	/* arithmetic file */
char	*lfile	=	"";	/* library file list */
char	*mfile	=	"-";	/* secondary header file */
char	*ofile	=	"-";	/* output file */
char	*ddir	=	"/tmp";	/* scratch file directory */
flag	Bflag	=	0;	/* don't use lookup table */
flag	Sflag	=	0;	/* generate memory stats */
flag	pflag	=	0;	/* print expr routine only */
	/*
	 *	table used by gipsflags()
	 */
static	flagentry	flagtable[] = {
	"a",	"-> file",	(flag *) &afile,
	"d",	"-> dir",	(flag *) &ddir,
	"l",	"-> file",	(flag *) &lfile,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"p",	"1",		(flag *) &pflag,
	"B",	"1",		(flag *) &Bflag,
	"S",	"1",		(flag *) &Sflag,
	NULL,	"[expr] [file...]",	NULL
};
	/*
	 *	miscellaneous stuff
	 */
char	tmpl[]	=	"ihaXXXXXX";
char	*tempsrc;	/* temporary source file */
char	*tempexec;	/* temporary executable */
TABLE	*out;		/* output table structure */
int	xnum	= -1;	/* number of x pixels */
int	ynum	= -1;	/* number of y pixels */
flag	byteflag = 1;	/* =1 for all-byte or =2 halfword pixels */
/*
 *	remove the scratch files, close the rest
 */
void
cleanup()
{
	char	buf[256];

	if (lfile != NULL && *lfile != '\0') {
		(void)sprintf(buf, "%s.o", tempexec);
		(void)unlink(buf);
	}
	if (tempsrc != NULL) (void) unlink(tempsrc);
	if (tempexec != NULL) (void) unlink(tempexec);
	exit(0);
}
/*
 *	the iha command
 */
main(argc,argv,envp)

	int	argc;		/* argument count */
	char	**argv;		/* array of argument pointers */
	char	**envp;		/* unix environment */
{
	FILE	*fp;		/* scratch file pointer */
	TABLE	*t;		/* table pointer */
	int	nfile;		/* number of input files */
	/*
	 *	save and validate callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (Bflag) byteflag = 0;
	/*
	 *	determine whether operation is in external file
	 *	or is a simple argument
	 */
	if (afile == NULL)
		if (--argc == 0)
			error("expr missing","");
		else
			afile = *++argv;

	if (--argc == 0)
		argv[++argc] = "-";
	if (Sflag)
		telltale(0);
	/*
	 *	read first input image header, construct first table
	 */
	t = get_table(*++argv);
	/*
	 *	possibly augment header from -m file.
	 */
	if (*mfile != '-')
		if ((fp = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fp));
	putargs();
	/*
	 *	construct output image table
	 */
	out = get_table((char *)NULL);
	out->nexttab = t;
	nfile = argc;
	/*
	 *	read remaining image headers, construct tables
	 */
	while (--argc > 0)
		t = t->nexttab = get_table(*++argv);
	/*
	**	Read and parse the command script
	*/

	get_arith();

	/*
	 *	determine whether conversion is faster via lookup
	 */

	if (byteflag == 2 && nfile == 1 && (xnum*ynum) > 256*256) {
		put_arith(256*256,1);
		comp_arith(envp);
		exec_arith(256*256,1,envp);
		copy_file(2);
	} else if (byteflag == 1 && nfile == 1 && (xnum*ynum) > 256) {
		put_arith(256,1);
		comp_arith(envp);
		exec_arith(256,1,envp);
		copy_file(0);
	} else if (byteflag == 1 && nfile == 2 && (xnum*ynum) > 256*256) {
		put_arith(256,256);
		comp_arith(envp);
		exec_arith(256,256,envp);
		copy_file(1);
	} else {
		byteflag = 0;
		put_arith(xnum,ynum);
		comp_arith(envp);
		exec_arith(xnum,ynum,envp);
	}

        printf( "iha: before sflag \n" );


	/*
	 *	generate C source file, compile and execute it
	 */
	if (Sflag)
		telltale(1);
	(void) cleanup();
}
