#ifndef lint
static char *sccsid = "%Z%$RCSfile: put_arith.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1998/05/13 01:20:02 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1985 Massachusetts Institute of Technology
**
**	Module:		put_arith.c	Creation Date:	04/23/85
**	Version:	$Revision: 2.7 $		Update:		$Date: 1998/05/13 01:20:02 $
**	Components:	put_arith()	iha subroutine
**
**	Function:	Generate C source program to perform desired arithmetic
*/

#include        <string.h>    /* for strlen */
#include	"iha.h"

static	char	*vars[]	=	{
	"ddmax",	"ddmin",	"ddnull",	"dxmax",	
	"dxmin",	"dymax",	"dymin",	"zdzero",
	"fdfact",	"ixlfft",	"ixnum",	"iylfft",
	"iynum",	0
};

extern char *index();

put_arith(x,y)
int	x, y;
{
	register FILE *fp, *fq;
	register TABLE *t;
	register char *s;
	register i;
	int	ivar;
	double	dvar;

	if (pflag)
		fp = stdout;
	else {
		tempexec = alloc(1,strlen(ddir)+strlen(tmpl)+2);
		(void)(sprintf(tempexec, "%s/%s", ddir, mktemp(tmpl)));
		tempsrc = alloc(1,strlen(tempexec)+3);
		(void)(strcat(strcpy(tempsrc,tempexec),".c"));
		if ((fp = fopen(tempsrc,"w")) == NULL)
			gipserr2(GIPSBADOPEN,tempsrc);
		gipsint(cleanup);
	}

        
	(void) fprintf(fp, "#include\t<stdlib.h>\n");      /* added for exit() function */
	(void) fprintf(fp, "#include\t<stdio.h>\n#include\t<signal.h>\n");
	(void) fprintf(fp, "#include\t<math.h>\n\n");

	if (out->flgs & CONV)
		(void) fprintf(fp, "static\tdouble\t_B0[%d];\n",
			(x*out->dlen+sizeof(double)-1)/sizeof(double));
	for (i = 0, t = out; t != NULL; t = t->nexttab, i++)
		(void) fprintf(fp, "static\t%s\t_b%d[%d];\n",
			t->vartype, i, (t->dtype == 'c') ? 2*x : x);
	(void) fprintf(fp, "\nmain()\n{\n\tregister x, y;\n");
	for (i = 0, t = out; t != NULL; t = t->nexttab, i++)
		(void) fprintf(fp, "\tregister %s *_p%d;\n", t->vartype, i);
	(void) fprintf(fp, "\n\tsignal(SIGFPE,SIG_IGN);\n\n");
	(void) fprintf(fp, "\tfor (y = 0; y < %d; y++) {\n", y);
	for (i = 1, t = out; (t = t->nexttab) != NULL; i++) (void) fprintf(fp,
		"\t\tif (fread(_p%d = _b%d,%d,%d,stdin) != %d) exit(%d);\n",
		i, i, t->plen, x, x, i+1);
	(void) fprintf(fp, "\t\tfor (_p0 = _b0, x = 0; x < %d; x++", x);
	for (i = 0, t = out; t != NULL; t = t->nexttab, i++)
		(void) fprintf(fp, ", _p%d%s", i, (t->dtype == 'c') ? " += 2" : "++");
	if (out->dtype == 'c') {
		(void) fprintf(fp, ") {\n#define\tr\t(_p0[0])\n");
		(void) fprintf(fp, "#define\ti\t(_p0[1])\n");
	} else
		(void) fprintf(fp, ") {\n#define\tp\t(*_p0)\n");
	for (i = 1, t = out; (t = t->nexttab) != NULL; i++)
		if (t->dtype == 'c') {
			(void) fprintf(fp, "#define\tr%d\t(_p%d[0])\n", i, i);
			(void) fprintf(fp, "#define\ti%d\t(_p%d[1])\n", i, i);
		} else
			(void) fprintf(fp, "#define\tp%d\t(*_p%d)\n", i, i);
	for (i = 0; s = vars[i]; i++) {
		char	buf[20];
		switch (*s++) {
		case 'i':
			if (ihget(s,(char *)&ivar,0))
				continue;
			(void) fprintf(fp, "#define\t%s\t(%d)\n", s, ivar);
			continue;
		case 'd':
			if (ihget(s,(char *)&dvar,0))
				continue;
			break;
		case 'z':
			dvar = out->dzero;
			break;
		case 'f':
			dvar = out->dfact;
			break;
		default:
			continue;
		}
		(void)sprintf(buf, "%.12g", dvar);
		if (index(buf, '.') == NULL)
			(void)strcat(buf, ".0");
		(void) fprintf(fp, "#define\t%s\t(%s)\n", s, buf);
	}
	(void) fprintf(fp, "\t\t\t%s;\n", afile);
	(void) fprintf(fp, (out->dtype == 'c') ?
		"#undef\tr\n#undef\ti\n" : ";\n#undef\tp\n");
	for (i = 1, t = out; (t = t->nexttab) != NULL; i++)
		(void) fprintf(fp, (t->dtype == 'c') ?
			"#undef\tr%d\n#undef\ti%d\n" : "#undef\tp%d\n", i, i);
	(void) fprintf(fp, "\t\t}\n");
	if (out->flgs & NORM) {
		(void) fprintf(fp, "\t\tfor (_p0 = _b0, x = %d; --x >= 0; _p0++)\n",
			x);
		(void) fprintf(fp, "\t\t\t*_p0 = %.12g * (*_p0 - (%.12g));\n",
			1/out->dfact, out->dzero);
	}
	if (out->flgs & CONV)
		(void) fprintf(fp, "\t\tgconv(_B0,_b0,%d,%d,%d);\n",
			x, ihdcvt(out->dtype), CVT_F);
	(void) fprintf(fp, "\t\tif (fwrite(_%c0,%d,%d,stdout) != %d) exit(1);\n",
		(out->flgs & CONV) ? 'B' : 'b', out->dlen, x, x);
	(void) fprintf(fp, "\t}\n}\n");

	if (pflag)
		exit(0);

	(void)(fclose(fp));
	(void)(chmod(tempsrc,0600));
}
