#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihbox.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1998/05/13 01:01:42 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihbox.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.7 $		Update:		$Date: 1998/05/13 01:01:42 $
**	Function:	Apply boxcar filter to a General Image
**			coordinates.
**	Components:	main()		GIPS/UNIX command
*/

#include	<stdio.h>
#include	<signal.h>
#include	<ctype.h>
#include	"gips.h"
	/*
	 *	flag fields filled by gipsflags()
	 */
static	char	*ifile	= "-";		/* input file */
static	char	*ofile	= "-";		/* output file */
static	char	*afile	= NULL;		/* array file */
static	char	*mfile	= "-";		/* secondary input header */
static	flag	Nflag	;		/* handle dnull "properly" */
static	flag	nflag	;		/* only process null pixels */
static	flag	uflag	;		/* unweight pixels */
static	flag	xflag	;		/* don't process null pixels */
static	flag	Sflag	;		/* generate memory statistics */
static	double	dfacti	= GIPSHIGH;	/* output scale for dscale=n */
static	double	dzeroi	= GIPSHIGH;	/* output base for dscale=n */
	/*
	 *	table used by gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"N",	"1",		&Nflag,
	"n",	"1",		&nflag,
	"u",	"1",		&uflag,
	"x",	"1",		&xflag,
	"S",	"1",		&Sflag,
	"a",	"-> file",	(flag *) &afile,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"0",	"%lf origin",	(flag *) &dzeroi,
	"1",	"%lf factor",	(flag *) &dfacti,
	NULL,	"array [file]",	NULL
};
	/*
	 *	variables initialized from GIPS header
	 */
static	int	xnum	;	/* input x-direction pixel length */
static	int	lpix	;	/* length of GIPS pixel in bytes */
static	char	dtype	;	/* pixel type code */
static	char	dscale = 's';	/* pixel scale code */
static	double	dnull	;	/* null pixel value */
static	double	dfact	;	/* dscale=n scale factor */
static	double	dzero	;	/* dscale=n scale constant */
static	long	inull	;	/* null integer pixel value */
static	int	dflag	;	/* set if dnull not defined */
static	int	inhom	;	/* set if array elements not identical */
static	double	*darr	;	/* boxcar array (float version) */
static	long	*iarr	;	/* boxcar array */
static	int	xarr	;	/* boxcar array x-dimension */
static	int	xarr2	;	/* xarr/2 */
static	int	yarr	;	/* boxcar array y-dimension */
static	int	yarr2	;	/* yarr/2 */

static		copy(), iproc(), dproc();

extern	char	*index();	/* C library */
extern	char	*strcat();	/* C library */
extern	char	*strcpy();	/* C library */
extern	double	fabs();		/* Math library */

main(argc,argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fp;		/* input GIPS file */
	FILE	*fa;		/* file containing boxcar array */
	FILE	*fo;		/* output GIPS file */
	int	ynum;		/* number of rasters in input image */
	int	narr;		/* number of elements in boxcar array */
	int	lrast;		/* length of raster element in byte */
	int	cvtinout;	/* input/output conversion code */
	int	cvtswap;	/* swapbuf element conversion code */
	char	*inout;		/* file input buffer */
	char	*swapbuf;	/* filter output buffer */
	char	**buf;		/* raster ring buffer pointers */
	char	*nullbuf;	/* buffer of null values */
	char	xscale;		/* type of x-axis coordinate */
	char	yscale;		/* type of y-axis coordinate */
	long	raster;		/* byte length of raster */
	long	border;		/* byte length of border */
	char	abuf[FONTLEN];	/* input file name */
	double	xwrap;		/* xscale=w wrap factor */
	double	xmin, xmax;	/* header variables */

	register int i;		/* scratch */
	/*
	 *	save and parse callers argument list
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (afile == NULL)
		if ((afile = *++argv) == NULL)
			error("missing array file","");
		else
			--argc;
	if (argc > 2)
		error("multiple input files: %s",argv[2]);
	if (argc == 2)
		ifile = *++argv;
	if (xflag && nflag)
		error("conflicting flags: -n and -x", "");
	if (Sflag)
		telltale(0);
	/*
	 *	read the input GIPS header, if necessary merging "mfile"
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	if (*mfile != '-')
		if ((fa = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fa));
	/*
	 *	get all required header fields
	 */
	if ((dflag = ihget("dnull",(char *) &dnull,0)) && (nflag || xflag))
		gipserr1(GIPSNOVAR);
	(void)(ihget("dscale",&dscale,1));
	if (dscale == 'f')
		gipserr1(GIPSBADVAL);
	else if (dscale == 'n') {
		if (ihget("dzero",(char *)&dzero,0) ||
		    ihget("dfact",(char *)&dfact,0))
			gipserr();
		if (dfacti == GIPSHIGH)
			dfacti = dfact;
		if (dzeroi == GIPSHIGH)
			dzeroi = dzero;
		(void)(ihput("dzero",(char *)&dzeroi));
		(void)(ihput("dfact",(char *)&dfacti));
		if (dfact == dfacti && dzero == dzeroi) {
			dscale='s';
			if (Nflag)
				dnull = (dnull - dzero) / dfact;
		}
		if (dfacti == 0)
			gipserr1(GIPSBADVAL);
		else
			dfacti = 1/dfacti;
	}
	if (ihget("yscale",&yscale,1))
		yscale = 's';
	else if (yscale == 'w')
		gipserr1(GIPSUNSCALE);
	if (ihget("xnum",(char *) &xnum,0))
		gipserr();
	if (xnum < 1 )
		gipserr1(GIPSBADVAL);
	if (ihget("xscale",&xscale,1))
		xscale = 's';
	/*
	**	Determine whether the margins are really wrapped
	*/
	if (xscale == 'w' && ihget("xwrap", (char *)&xwrap, 0) == 0) {
		if (ihget("xmin", (char *)&xmin, 0))
			xmin = 0;
		if (ihget("xmax", (char *)&xmax, 0))
			xmax = xnum;
		if (fabs(xwrap) != fabs(xmax-xmin))
			xscale = 's';
	}
	if (ihget("ynum",(char *) &ynum,0))
		gipserr();
	if (ynum < 1 )
		gipserr1(GIPSBADVAL);
	if (ihget("dtype",&dtype,1))
		gipserr();
	if ((lpix = ihdtype(dtype)) == 0)
		gipserr1(GIPSBADVAL);
	/*
	 *	write the GIPS header
	 */
	putargs();
	if ((fo = ihwrit(ofile,0)) == NULL)
		gipserr();
	if (fo == stdout) {
		(void)(signal(SIGPIPE,exit));
		ofile = "'stdout'";
	}
	/*
	 *	open the boxcar array definition file or string
	 */
	if (index(afile,' ') || index(afile,'\n') || index(afile,'\t')) {
		fa = fopens(afile);
	} else if ((fa = fopen(afile,"r")) == NULL) {
		(void) strcat(strcpy(abuf, GIPSBOXDIR), afile);
		if ((fa = fopen(abuf,"r")) == NULL)
			gipserr2(GIPSBADOPEN,afile);
		else
			afile = abuf;
	}
	if (fscanf(fa,"%d%d",&xarr,&yarr) != 2 || xarr < 1 || yarr < 1 ||
		(xarr & 1) == 0 || (yarr & 1) == 0)
		error("bad array limits: %s",afile);
	/*
	 *	allocate the float or integer boxcar array
	 */
	narr = xarr*yarr;
	xarr2 = xarr/2;
	yarr2 = yarr/2;
	cvtinout = ihdcvt(dtype);
	/*
	 *	allocate the various buffers...
	 *	inout:	input and output buffer
	 *	buf:	ring of pointers to individual raster buffers
	 *	swapbuf: buffer to hold filtered raster result
	 *	darr:	double-precision boxcar array
	 *	iarr:	integer version of darr
	 *	nullbuf: buffer of dnull values
	 */
	inout = alloc(xnum, lpix);
	buf = (char **) alloc(yarr, sizeof(char *));

	if (dtype == 'd' || dtype == 'f' || dscale == 'n') {
		cvtswap = CVT_D;
		darr = (double *) alloc(narr, lrast = sizeof(double));
	} else	{
		cvtswap = CVT_I;
		iarr = (long *) alloc(narr, lrast = sizeof(long));
		inull = (long) dnull;
	}
	swapbuf = (cvtswap == cvtinout) ? inout : alloc(xnum,lrast);
	if (gipserrno == GIPSNOCORE)
		gipserr();
	/*
	 *	read the boxcar array
	 */
	for (i = 0; i < narr;) {
		double d;
		int	n, ch;

		if (fscanf(fa,"%lf",&d) != 1)
			gipserr2(GIPSBADDATA,afile);
		while ((ch = getc(fa)) != EOF && isspace(ch))
			;
		switch (ch) {
		case '*':
			n = (int)d;
			if (n <= 0 || (double)n != d)
				error("bad repeat count: %d", (char *)n);
			if (fscanf(fa,"%lf",&d) != 1)
				gipserr2(GIPSBADDATA,afile);
			break;
		default:
			ungetc(ch, fa);
		case EOF:
			n = 1;
			break;
		}
		while (--n >= 0 && i < narr)
			if (darr != NULL)
				inhom |= ((darr[i++] = d) != darr[0]);
			else
				inhom |= ((iarr[i++] = (long)d) != iarr[0]);
	}
	if (!isastring(fa))
		(void)(fclose(fa));
	raster	= lrast*xnum;
	border	= lrast*xarr2;
	/*
	 *	fill the dnull buffer
	 */
	nullbuf = alloc(raster+border+border, 1);
	if (darr == NULL) {
		*(long *)nullbuf = inull;
		i = sizeof(char);
	} else {
		*(double *)nullbuf = dnull;
		i = sizeof(double);
	}
	copy(nullbuf, nullbuf+i, raster+border+border-i);
	/*
	 *	loop over input rasters
	 */
	for (i = 0;; i++) {
		register char *save;	/* pointer to most recent raster */
		/*
		 *	begin by allocating raster buffers
		 */
		if (i < yarr) {
			if ((save = buf[i] = alloc(xnum+xarr-1, lrast)) == NULL)
				gipserr();
		} else	{
			register int j;
			/*
			 *	if all raster buffers are allocated
			 *	(and hence filled) they can be convolved
			 *	to "out" and written out
			 */
			if (darr == NULL)
				iproc((long **)(buf+yarr2),(long *)swapbuf);
			else
				dproc((double **)(buf+yarr2),(double *)swapbuf);
			if (inout != swapbuf)
				gconv(inout,swapbuf,xnum,cvtinout,cvtswap);
			if (fwrite(inout,lpix,xnum,fo) != xnum)
				gipserr2(GIPSBADWRITE,ofile);
			/*
			 *	on all but last raster, rotate the ring buffers
			 */
			if (i == yarr+ynum-1)
				break;
			save = buf[0];
			for (j = 1; j < yarr; buf[j-1] = buf[j], j++)
				;
			buf[yarr-1] = save;
		}
		/*
		 *	fill the input buffer
		 */
		if (i < yarr2 || i >= ynum+yarr2) {
			copy(nullbuf, save, raster+border+border);
			continue;
		}
		if (fread(inout,lpix,xnum,fp) != xnum)
			gipserr2(GIPSBADREAD,ifile);
		/*
		 *	handle dscale=n conversion
		 */
		if (dscale == 'n') {
			register double *din = ((double *)save)+xarr2;
			register double *dswap = (double *)swapbuf;
			register j = xnum;

			if (inout != swapbuf)
				gconv(swapbuf,inout,xnum,cvtswap,cvtinout);
			while (--j >= 0)
				*din++ = dfact * *dswap++ + dzero;
		} else
			gconv(save+border,inout,xnum,cvtswap,cvtinout);
		/*
		 *	handle xscale=w wrap-around
		 */
		if (xarr2 == 0)
			;
		else if (xscale == 'w') {
			copy(save+raster,save,border);
			copy(save+border,save+raster+border,border);
		} else if (yarr2 != 0) {
			copy(nullbuf, save, border);
			copy(nullbuf, save+border+raster, border);
		}
	}
	(void)(ihbcl(fp));
	(void)(ihbcl(fo));
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}
/*
 *	copy "a" to "b" byte-by-byte (i.e. extension if a and b overlap)
 */
static
copy(a,b,n)

	register char *a;		/* 'from' field */
	register char *b;		/* 'to' field */
	register long n;		/* number of bytes to copy */
{
	while (--n >= 0)
		*b++ = *a++;
}
/*
 *	convolve integer rasters
 */
static
iproc(ibuf,out)

	long	**ibuf;		/* raster buffer ring pointer */
	long	*out;		/* output buffer */
{
	int	 ix;		/* current pixel x-coordinate */
	int	 ay;		/* index over boxcar */
	register long x;	/* pixel value accumulator */
	register long w;	/* weight accumulator */
	register long *a;	/* pointer to boxcar array */
	register long *b;	/* pointer to input pixel */
	register int ax;	/* index over boxcar */

	for (ix = 0; ix < xnum; ix++, *out++ = x) {
		x = ibuf[0][ix+xarr2];
		if ((xflag && x == inull) || (nflag && x != inull))
			continue;
		for (a = iarr, ay = -yarr2, x=w=0; ay <= yarr2; ay++)
			for (ax=xarr, b=ibuf[ay]+ix; --ax >= 0; a++,b++)
				if (dflag || *b != inull)
					if (inhom) {
					 x += *b * *a;
					 w += (*a > 0) ? *a : -*a;
					} else {
					 x += *b;
					 w++;
					}
		if (w == 0) x = inull;
		else if (! uflag) x /= w;
	}
}
/*
 *	convolve floating point pixels
 */
static
dproc(dbuf,out)

	register double	**dbuf;	/* raster buffer ring pointer */
	register double	*out;	/* output buffer */
{
	double	w;		/* weight accumulator */
	register double *a;	/* pointer to boxcar array */
	register double *b;	/* pointer to input pixel */
	register ix;		/* current pixel x-coordinate */
	register int ax;	/* index over boxcar */
	register int ay;	/* index over boxcar */

	for (ix = 0; ix < xnum; ix++, out++) {
		*out = dbuf[0][ix+xarr2];
		if ((xflag && *out == dnull) || (nflag && *out != dnull)) {
			if (dscale == 'n')
				*out = (*out - dzeroi) * dfacti;
			continue;
		}
		for (a = darr, ay = -yarr2, *out = w = 0; ay <= yarr2; ay++)
			for (ax = xarr, b = dbuf[ay]+ix; --ax >= 0; a++,b++)
				if (dflag || *b != dnull)
					if (inhom) {
						*out += *b * *a;
						w += (*a > 0) ? *a : -*a;
					} else {
						*out += *b;
						w++;
					} 
		if (w == 0) {
			*out = dnull;
			continue;
		} else if (! uflag) {
			*out /= w;
		}
		if (dscale == 'n')
			*out = (*out - dzeroi) * dfacti;
	}
}
