#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihc.c,v $	GIPS $Revision: 2.8 $ (MIT/CSR) $Date: 2004/11/29 18:52:23 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihc.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.8 $		Update:		$Date: 2004/11/29 18:52:23 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Combine GIPS header and binary array to make
**			a General Image output stream
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"

#define		MAXFIL		16

	/*
	 *	variables filled in by gipsflags()
	 */
static	char	*ddir	= "/tmp";	/* default scratch directory */
static	char	*ifile	= "-";		/* input image array */
static	char	*ofile	= "-";		/* output general image */
static	char	*hfile	= "-";		/* input header */
static	char	*mfile	= "-";		/* secondary input header */
static	int	block	= BUFSIZ;	/* input blocksize (dft from stdio) */
static	int	skip;			/* skip these ifile bytes */
static	flag	bflag;			/* byte-swap the pixels */
static	flag	fflag;			/* read ASCII random-access pixels */
static	flag	iflag;			/* invert the image */
static	flag	rflag;			/* reverse raster pixels */
static	flag	wflag;			/* read binary random-access pixels */
static	flag	Sflag;			/* generate memory statistics */
	/*
	 *	argument definition block for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"b",	"1",		(flag *) &bflag,
	"f",	"1",		(flag *) &fflag,
	"i",	"1",		(flag *) &iflag,
	"r",	"1",		(flag *) &rflag,
	"w",	"1",		(flag *) &wflag,
	"S",	"1",		(flag *) &Sflag,
	"",	"%d block",	(flag *) &block,
	"d",	"-> dir",	(flag *) &ddir,
	"h",	"-> file",	(flag *) &hfile,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"s",	"%d bytes",	(flag *) &skip,
	NULL,	"[file]",	NULL
};
	/*
	**	Format of binary input record
	*/
struct	inbuf	{
	unsigned long	x,y;	/* input x and y coordinates */
	float		f;	/* pixel value */
};
	/*
	 *	vital header data values
	 */
static	char	dtype;		/* input pixel type code */
static	int	xnum;		/* input x-direction pixel length */
static	int	ynum;		/* input y-direction pixel length */
static	double	dnull;		/* null pixel data value */
	/*
	 *	temporary file for inversion of tape or pipe arrays
	 */
static	char	*tmpl = "%s/ihc%dXXXXXX";
static	char	file[100];	/* buffer for temporary file name */
static	FILE	*fp;		/* output stream pointer */
static	int	fi;		/* input file descriptor */
static	int	lpix;		/* input pixel length */
static	int	raster;		/* input raster length in bytes */
static	int	nfile;		/* count of temporary files */
static	char	*buf;		/* input buffer */
static	int	*sumbuf;	/* pixel count buffer */
static	float	*addbuf;	/* pixel accumulator */

static		copy(), spool(), sort(), getpix();
extern	char	*mktemp();
extern	off_t	lseek();
	/*
	 *	interrupt handler - erase any temporary files
	 */
static
onint()
{
	while (nfile > 0) {
		(void)sprintf(file,tmpl,ddir,--nfile);
		(void)(unlink(file));
	}
	exit(1);
	return(0);
}

main(argc,argv)

	int	argc;			/* argument count */
	char	**argv;			/* array of argument pointers */
{
	int	i;			/* scratch */
	int	xflag;			/* indicates reblocking */
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	if ((wflag |= fflag) && (bflag || iflag || rflag))
		error("flag conflict: -f","");
	if (argc > 1)
		ifile = *++argv;
	if (Sflag)
		telltale(0);
	/*
	 *	read GIPS header and (optional) merge file
	 */
	if ((fp = ihread(hfile,GIPSVERSIONS)) == NULL)
		gipserr();
	if (fp != stdin)
		(void)(ihbcl(fp));
	if (*mfile != '-')
		if ((fp = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fp));
	/*
	 *	ensure xorg,yorg initialized and check dimensional variables
	 */
	if (ihget("xorg",(char *)&i,i=0)) (void)(ihput("xorg",(char *)&i));
	if (ihget("yorg",(char *)&i,i=0)) (void)(ihput("yorg",(char *)&i));

	if (ihget("xnum",(char *) &xnum,0)) gipserr();
	if (xnum < 1 ) gipserr1(GIPSBADVAL);
	if (ihget("ynum",(char *) &ynum,0)) gipserr();
	if (ynum < 1 ) gipserr1(GIPSBADVAL);

	if (ihget("dtype",&dtype,1)) gipserr();
	if ((lpix = ihdtype(dtype)) == 0) gipserr1(GIPSBADTYPE);
	/*
	 *	add header history entry and write it out
	 */
	if (*ofile != '-')
		(void)(ihput("origin",ofile));
	putargs();
	if ((fp = ihwrit(ofile,0)) == NULL)
		gipserr();
	else if (fp == stdout)
		(void)(signal(SIGPIPE,exit)), ofile = "'stdout'";
	(void)(mktemp(tmpl));
	/*
	 *	get raster length and xflag - whether array must be inverted
	 *	or reblocked
	 */
	if (wflag) {
		addbuf = (float *)alloc(xnum,sizeof(float));
		sumbuf = (int *)alloc(xnum,MAX(lpix,sizeof(int)));
		(void)(ihget("dnull",(char *)&dnull,0));
		gipsint(onint);
		i = sort(0,ynum,0);
	} else {
		raster = xnum*lpix;
		if (*ifile == '-')
			ifile = "'stdin'";
		else if ((fi = open(ifile,0)) == -1)
			gipserr2(GIPSBADOPEN,ifile);
		xflag = iflag || (skip % block) ||
			(rflag && (block % raster) != 0);
		buf = alloc(block,1);
		if (xflag && (fi == 0 || strncmp(ifile,"/dev/",5) == 0))
			spool();
		if (xflag)
			block = raster;
		i = copy();
	}
	if (i)
		gipserr2(GIPSBADWRITE,ofile);
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}
/*
 *	copy file "ifile" to stream *fp, with possible reversal or inversion
 */
static
copy()
{
	register off_t	npix;	/* number of pixels to copy */
	register i;		/* scratch */
	register char *d, *e;	/* scratch */

 	if (bflag && (raster & 1))
		error("-b illegal for odd raster length: %d",(char *)raster);
	/*
	 *	seek past header, if requested
	 */
	if (! iflag)
		if (skip % block)
			(void)(lseek(fi,(off_t) skip,0));
		else
			for (i = 0; i < skip; i += block)
				if (read(fi, buf, block) != block)
					gipserr2(GIPSBADREAD,ifile);
	/*
	 *	loop over output rasters
	 */
	for (npix = raster * ynum; npix > 0; npix -= i) {
		if (iflag)
			(void)(lseek(fi,(off_t) (npix-raster+skip),0));
		if ((i = read(fi,buf,block)) <= 0)
			break;
		if (bflag)
			swab(buf,buf,i & ~1);
		if (rflag) {
			for (d = buf; d < buf+i; d += raster)
				for (e = d + raster; e > d; )
					if (fwrite(e -= lpix,lpix,1,fp)
							!= 1)
						return(1);
		} else if ( fwrite(buf,i,1,fp) != 1)
			return(1);
	}
	if (npix > 0)
		error("unexpected EOF: %s",ifile);
	/*
	 *	if input from stdin or tape, read to the end
	 */
	if (fi == 0 || strncmp(ifile,"/dev/",5) == 0)
		while (read(fi,buf,block) > 0)
			;
	(void)(close(fi));
	(void)(ihbcl(fp));
	return(0);
}
/*
 *	copy input file to temporary
 */
static
spool()
{
	register	fo;			/* output file descriptor */
	register long	l = raster*ynum;/* number of bytes to copy */
	register	i;			/* scratch */

	(void)sprintf(file,tmpl,ddir,nfile++);
	if ((fo = creat(file,0600)) == -1)
		gipserr2(GIPSBADOPEN,file);
	gipsint(onint);
	for (l = raster*ynum+skip; (i = read(fi,buf,block)) > 0; l -= i)
		if (skip >= i)
			skip -= i;
		else if (l >= 0 && write(fo,buf,i) != i)
			gipserr2(GIPSBADWRITE,file);
	if (i == -1)
		gipserr2(GIPSBADREAD,ifile);
	(void)(close(fi));
	(void)(close(fo));
	if ((fi = open(ifile = file,0)) == -1)
		gipserr2(GIPSBADOPEN,ifile);
	(void)(unlink(ifile));
}
/*
 *	recursively spool to temporary files and sift each raster
 */
static
sort(ymin,ymax,ntemp)

	int	ymin;		/* lower input y-coordinate */
	int	ymax;		/* one more than upper input y-coordinate */
	int	ntemp;		/* next temporary file number */
{
	register	i;	/* scratch */
	register	imax;	/* number of output files */
	register	fact;	/* number of rasters per output file */
	register float	*a;	/* accumulator pointer */
	register int	*s;	/* count buffer pointer */
	FILE	*ff[MAXFIL];	/* output stream pointers */
	char	infile[100];	/* name of temporary input file */
	struct	inbuf	inrec;	/* binary input record */

	if (*ifile == '-')
		ifile = "'stdin'";
	else if (freopen(ifile,"r",stdin) == NULL)
		gipserr2(GIPSBADOPEN,ifile);
	while (--skip >= 0)
		(void)(getchar());

	if ((imax = ymax-ymin) > 1) {
		if (imax > MAXFIL)
			imax = MAXFIL;
		fact = (ymax-ymin+imax-1)/imax;
		imax = (ymax-ymin+fact-1)/fact;
		for (i = 0; i < imax; i++) {
			(void)(sprintf(file,tmpl,ddir,nfile = ntemp+i));
			if ((ff[i] = fopen(file,"w")) == NULL)
				gipserr2(GIPSBADOPEN,file);
		}
		while (getpix(&inrec)) {
			if ((i = (inrec.y - ymin)/fact) < imax &&
			    fwrite((char *)&inrec,sizeof(inrec),1,ff[i]) != 1) {
				(void)sprintf(file,tmpl,ddir,ntemp+i);
				gipserr2(GIPSBADWRITE, file);
			}
		}
		fflag = 0;
		for (i = 0; i < imax; (void)(fclose(ff[i++])));
		for (i = 0; i < imax; i++, ymin += fact) {
			(void)(sprintf(ifile = infile,tmpl,ddir,ntemp+i));
			if (sort(ymin,MIN(ymin+fact,ymax),ntemp+imax))
				return(1);
			(void)(unlink(infile));
		}
		return(0);
	}
	clear((char *)(s = sumbuf), xnum*sizeof(int));
	clear((char *)(a = addbuf), xnum*sizeof(float));
	while (getpix(&inrec))
		if ((i = inrec.x) < xnum)
			a[i] += inrec.f, s[i]++;
	(void)(fclose(stdin));
	clearerr(stdin);
	for (i = xnum; --i >= 0; a++, s++)
		*a = *s ? *a / *s : dnull;
	gconv((char *)sumbuf, (char *)addbuf, xnum, ihdcvt(dtype), CVT_F);
	return (fwrite((char *)sumbuf, lpix, xnum, fp) != xnum);
}

static
getpix(in)
	register struct	inbuf	*in;	/* input buffer */
{
	if (fflag) {
		if (scanf(" %ld %ld %f", &in->x, &in->y, &in->f) == 3)
			return 1;
	} else {
		if (fread((char *)in, sizeof(struct inbuf), 1, stdin) == 1)
			return 1;
	}
	if (feof(stdin))
		return 0;
	if (ferror(stdin))
		gipserr2(GIPSBADREAD, ifile);
}
