#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihfft.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 1998/05/13 01:01:45 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihfft.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 1998/05/13 01:01:45 $
**	Components:	main()		UNIX command
**
**	Function:	Perform complex Fourier transform on each raster
**			of a General Image file
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"
	/*
	 *	variables set by gipsflags()
	 */
static	char	*ifile	= "-";		/* input file */
static	char	*ofile	= "-";		/* output file */
static	char	*mfile	= "-";		/* secondary input header */
static	flag	iflag	;		/* sign flag (inverse transform) */
static	flag	Nflag	;		/* generate null pixel warnings */
static	flag	Sflag	;		/* generate memory statistics */
	/*
	 *	argument table for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"i",	"1",		&iflag,
	"N",	"1",		&Nflag,
	"S",	"1",		&Sflag,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	NULL,	"[name]",	NULL
};
	/*
	 *	header variables required by this command
	 */
static	char	dtype	;		/* input pixel type code */
static	char	dscale	= 's';		/* input pixel mapping code */
static	double	*wst	;		/* tables of cos,sin (pi/n) */
static	float	dfact	;		/* dscale=n scaling factor */
static	float	dzero	;		/* dscale=n pixel origin */
static	float	dnull	;		/* null pixel value */
static	int	nulls	;		/* count of input null pixels */
static	int	inull	= 1;		/* flag if dnull present in header */
static	int	lpix	;		/* input pixel length */
static	int	lout	;		/* output pixel length */
static	int	xnum	;		/* input x-direction pixel length */
static	int	ynum	;		/* input y-direction pixel length */
static	int	xlfft	;		/* 2**n >= xnum for fft */
static	int	xlout	;		/* number of output elements */

static		fft();
extern	double	sin(), cos(), sqrt();	/* math library */
extern	double	fabs();			/* math library */

main(argc,argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fp;		/* input GIPS stream pointer */
	FILE	*fo;		/* output GIPS stream pointer */
	char	*in;		/* input buffer pointer */
	float	*out;		/* output buffer pointer */
	char	xscale;		/* input scale type */
	int	incvt;		/* input element conversion code */
	int	outcvt;		/* output element conversion code */
	int	nw;		/* dimension of wst */
	double	d;		/* fraction of pi for wst setup */
	double	xwrap;		/* xscale=w wrap factor */
	double	xmin, xmax;	/* header variables */

	register	int i;		/* scratch */
	register	double	*w;	/* pointer to wst */

	double	pi = 3.1415926535897932;
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("bad arg: %s",argv[2]);
	if (Sflag)
		telltale(0);
	if (argc == 2)
		ifile = *++argv;
	/*
	 *	read input GIPS header, and possible merge header
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	if (*mfile != '-')
		if ((fo = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fo));
	/*
	 *	get vital header variables and test them out
	 */
	if (ihget("dscale",&dscale,1) == 0 &&
		   (dscale == 'f' || dscale == 'l'))
		gipserr1(GIPSUNSCALE);
	if (dscale == 'n') {
		if (ihget("dfact",(char *)&d,0))
			gipserr();
		else
			dfact = (float) d;
		if (ihget("dzero",(char *)&d,0))
			gipserr();
		else
			dzero = (float) d;
		(void)(ihput("dscale","s"));
	}
	/*
	 *	"-i" inverse transform only works for complex-type data
	 */
	if (ihget("dtype",&dtype,1))
		gipserr();
	lpix = ihdtype(iflag ? 'c' : dtype);
	lout = ihdtype(iflag ? dtype : 'c');
	if (lpix == 0 || lout == 0)
		gipserr1(GIPSBADTYPE);
	incvt = iflag ? CVT_C : ihdcvt(dtype);
	outcvt = iflag ? ihdcvt(dtype) : CVT_C;

	if (ihget("xscale",&xscale,1))
		xscale = 's';
	if (ihget("xnum",(char *) &xnum,0))
		gipserr();
	else if (xnum < 1 )
		gipserr1(GIPSBADVAL);
	/*
	**	Determine whether the margins are really wrapped
	*/
	if (xscale == 'w' && ihget("xwrap", (char *)&xwrap, 0) == 0) {
		if (ihget("xmin", (char *)&xmin, 0))
			xmin = 0;
		if (ihget("xmax", (char *)&xmax, 0))
			xmax = xnum;
		if (fabs(xwrap) != fabs(xmax-xmin))
			xscale = 's';
	}

	if (ihget("ynum",(char *) &ynum,0))
		gipserr();
	else if (ynum < 1 )
		gipserr1(GIPSBADVAL);
	/*
	 *	allocate input and output buffers and cos,sin table array
	 */
	if (! iflag) {
		for (xlfft = 1; xlfft < xnum; xlfft <<= 1)
			;
		if (ihput("xlfft",(char *)&xnum))
			gipserr();
		xlout = xlfft;
		if (Nflag && dtype != 'c')
			if ((inull = ihget("dnull",(char *)&d,0)) == 0)
				dnull = (float) d;
	} else if (ihget("xlfft",(char *)&xlout,0) || ihdel("xlfft"))
		gipserr();
	else
		xlfft = xnum;
	if (xnum != xlout && ihput("xnum",(char *)&xlout))
		gipserr();
	out = (float *) alloc(2*xlfft,sizeof(float));
	if (! iflag)
		(void)(ihput("dtype","c"));
	/*
	**	Allocate cos,sin array
	*/
	for (i = nw = 1; i < xlfft; i <<= 1, nw++)
		;
	wst = (double *) alloc(2*nw,sizeof(double));
	in = alloc(xnum,lpix);
	if (gipserrno == GIPSNOCORE)
		gipserr();
	/*
	 *	write history entry and output GIPS header
	 */ 
	putargs();
	if ((fo = ihwrit(ofile,0)) == NULL)
		gipserr();
	if (fo == stdout) {
		(void)(signal(SIGPIPE,exit));
		ofile = "'stdout'";
	}
	/*
	 *	initialize the sin,cos table
	 *	special test for cos,sin(0) when using -lnm [pgf]
	 */
	d = iflag ? pi : -pi;
	for (w = wst, i = 0; i < nw; i++, d *= 0.5) {
		*w++ = cos(d);
		*w++ = sin(d);
	}
	/*
	 *	read and transform the rasters
	 */
	while (--ynum >= 0) {
		register float	*fout;	/* output buffer pointer */
		float		f;	/* scratch */

		if (fread(in,lpix,xnum,fp) != xnum)
			gipserr2(GIPSBADREAD,ifile);

		gconv((char *)out,in,xlfft,CVT_C,incvt);

		if (dscale == 'n' || inull == 0)
			for (i = xlfft, fout = out; --i >= 0; fout++, fout++) {
				if (inull == 0 && *fout == dnull)
					nulls++;
				if (dscale == 'n')
					*fout = dfact * *fout + dzero;
			}
		if (! iflag && (i = xlfft-xnum) > 0) {
			fout = out + 2*xnum;
			if (xscale == 'w')
				move((char*)out,(char*)fout,2*sizeof(float)*i);
			else
				clear((char *)fout, 2*sizeof(float)*i);
		}
		fft(fout = out);
		f = 1/sqrt((double) xlfft);
		for (i = 2*xlout; --i >= 0; *fout++ *= f)
			;
		if (iflag && outcvt != CVT_C) {
			gconv(in,(char *)out,xlout,outcvt,CVT_C);
			fout = (float *)in;
		} else
			fout = out;
		if (fwrite((char *)fout,lout,xlout,fo) != xlout)
			gipserr2(GIPSBADWRITE,ofile);
	}
	(void)(ihbcl(fp));
	(void)(ihbcl(fo));
	if (Sflag)
		telltale(1);
	if (nulls != 0)
		error("warning - null pixels read: %d",(char *)nulls);
	(void)(exit(0));
}
/*
 *	Returns FFT of Fortran complex array FARR(NARR,2) in place.
 *	Author Norman Brenner, MIT; C-conversion Peter Ford MIT DEAPS.
 */
static
fft(farr)
	
	float	*farr;		/* array of pairs of float variables */
{
	register	float *f, *fr, *flim;
	register int	i;
	register double	*w;
	double	wr, wi, tempd;
	float	*f1, tempr, tempi;

	for (f = fr = farr, flim = farr+2*xlfft; f < flim; ) {
		if (f<fr) {
			tempr = f[0]; f[0] = fr[0]; fr[0] = tempr;
			tempi = f[1]; f[1] = fr[1]; fr[1] = tempi;
		}
		for (i = xlfft; i < (fr-farr+1) && i >= 2; i /= 2)
			 fr -= i;
		f += 2;
		fr += i;
	}

	for (w = wst, i = 2; i < 2*xlfft; i <<= 1, w += 2)
		for (wr = 1.0, wi = 0.0, f1 = farr; f1 < (farr+i); ) {
			for (f = f1, fr = f+i; f < flim; fr = i+f) {
				tempr = wr*fr[0] - wi*fr[1];
				tempi = wr*fr[1] + wi*fr[0];
				fr[0] = f[0] - tempr;
				fr[1] = f[1] - tempi;
				f[0] += tempr;
				f[1] += tempi;
				f += i+i;
			}
			tempd = wr,
			wr = wr*w[0] - wi*w[1];
			wi = wi*w[0] + tempd*w[1];
			f1 += 2;
		}
}
