#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihfromfits.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 2004/11/29 18:52:44 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		fits.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 2004/11/29 18:52:44 $
**	Function:	Translate FITS-format image file to GIPS
**	Components:	main()		GIPS/UNIX command
*/

#include        <string.h>
#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"
	
static	flag	bflag;			/* byte-swap binary input */
static	flag	hflag;			/* copy all HISTORY entries */
static	flag	Bflag;			/* skip binary length fields */
static	flag	Sflag;			/* print statistics */
static	char	*mfile	= "-";		/* merged header data */
	/*
	 *	argument table used by gipsflags()
	 */
static	flagentry	flagtable[] = {
	"b",	"1",		&bflag,
	"h",	"1",		&hflag,
	"B",	"1",		&Bflag,
	"S",	"1",		&Sflag,
	"m",	"-> file",	(flag *)&mfile,
	NULL,	"[file]",	NULL
};
	/*
	 *	maximum permissable input coefficient dimensions,
	 *	input array dimensions, size of FITS tape block,
	 *	and the number of input pixels per block
	 */
#define		INDIMS		4
#define		OUTDIMS		2
#define		FITSBLK		2880
#define		FITSSIZE	(FITSBLK / sizeof(short))
	/*
	 *	flag values used in 
	 */
#define		IS_ARRAY	0010
#define		IS_REQD		0020
#define		IS_HISTORY	0040
#define		IS_FOUND	0100
#define		IS_END		0200
#define		IS_UNITS	0400
#define		IS_OPTHIST	01000
	/*
	 *	variable type codes
	 */
#define		LOGICAL		00
#define		DOUBLE		01
#define		FIXED		02
#define		CHAR8		03
	/*
	 *	conversion between n=0,1 x,y indices and float realworld values
	 */
#define		FLOAT(i,n)	(crval[n]+((i)-crpix[n])*cdelt[n])
#define		FIX(f,n)	(crpix[n]-1.0+((f)-crval[n])/cdelt[n])
	/*
	 *	static variables and arrays to which the FITS header is decoded
	 */
static char	simple;		/* =T if file is in standard format */
static int	bitpix;		/* number of bits per image pixel */
static int	naxes;		/* number of axes (or 0 if no image data) */
static int	naxis[INDIMS];	/* number of pixel columns/rows */
static double	bscale = GIPSLOW; /* float = bscale*integer + bzero_ */
static double	bzero_  = GIPSLOW; /* .. mapping of pixels to data */  /* was bzero */
static char	bunit[9];	/* description of float data units */
static double	dnull;		/* pixel value denoting bad data ? */
static char	object[9];	/* description of image */
static char	write_date[9];	/* date file written dd/mm/yy */
static char	date_obs[9];	/* date of data acquisition dd/mm/yy */
static char	origin[9];	/* tape-writing institution */
static char	instrument[9];	/* data acquisition instrument */
static char	observer[9];	/* observer name or id */
static char	history[80];	/* history entry */
static double	crval[INDIMS];	/* r.a./dec reference values */
static double	crpix[INDIMS]; 	/* corresponding pixel values */
static double	cdelt[INDIMS];	/* r.a./dec increment per pixel */
static double	crota[INDIMS];	/* rotation of axis from stated type */
static char	ctype[INDIMS][9];/* type of coordinates */
static double	dmax;		/* maximum true float pixel value */
static double	dmin;		/* minimum true float pixel value */
	/*
	 *	input format definition table
	 */
static struct format	{
	char	*str;		/* scanf() format string */
	int	len;		/* maximum output field length */
} form[] = {
	"%1s",		sizeof(char),
	"%lf",		sizeof(double),
	"%d",		sizeof(int),
	"%*1s%8s",	sizeof(char)*9,
};
	/*
	 *	Miscellaneous strings
	 */
static	char	*initfile = "FITSinit"; /* skeleton GIPS header */
static	char	*morehist;			/* history entry */
	/*
	 *	table to define mapping between FITS and GIPS header variables
	 */
static struct data {
	char *	name;		/* FITS variable name */
	char *	addr;		/* pointer to where to save the value */
	char *	oname;		/* equivalent GIPS name, if any */
	int 	flg;		/* flag fields */
} fitsparm[] = {

"END     ",		NULL,		"",		IS_END,

"SIMPLE  ",		&simple,	"",		IS_REQD | LOGICAL,
"BITPIX  ",	(char*) &bitpix,	"", 		IS_REQD | FIXED,
"NAXIS   ",	(char*) &naxes,		"",		IS_REQD | FIXED,
"NAXIS",	(char*) naxis,		"xnum",		IS_ARRAY | FIXED,

"CRVAL",	(char*) crval,		"",		IS_ARRAY | DOUBLE,
"CRPIX",	(char*) crpix,		"",		IS_ARRAY | DOUBLE,
"CDELT",	(char*) cdelt,		"",		IS_ARRAY | DOUBLE,
"CROTA",	(char*) crota,		"",		IS_ARRAY | DOUBLE,
"CTYPE",		&ctype[0][0],	"xtitle", IS_ARRAY|CHAR8|IS_UNITS,

"BSCALE  ",	(char*) &bscale,	"dfact",	DOUBLE,
"BZERO   ",	(char*) &bzero_,		"dzero",	DOUBLE,
"BUNIT   ",		bunit,		"dtitle",	CHAR8 | IS_UNITS,
"BLANK   ",	(char*) &dnull,		"dnull",	DOUBLE,
"OBJECT  ",		object,		"title",	IS_HISTORY | CHAR8,
"DATE    ",		write_date,	"",		IS_HISTORY | CHAR8,
"DATE-OBS",		date_obs,	"date",		CHAR8,
"ORIGIN  ",		origin,		"",		IS_HISTORY | CHAR8,
"INSTRUME",		instrument,	"",		IS_HISTORY | CHAR8,
"OBSERVER",		observer,	"",		IS_HISTORY | CHAR8,
"HISTORY ",		history,	"",		IS_OPTHIST,

"DATAMAX ",	(char*) &dmax,		"dmax",		DOUBLE,
"DATAMIN ",	(char*) &dmin,		"dmin",		DOUBLE,

NULL
};
	/*
	 *	table to determine translation of data and axis titles
	 */
static struct units {
	char	* oldname;		/* FITS name */
	char	* newname;		/* GIPS name */
} unittab [] = {
	"K",		"Kelvins",
	"JY/BEAM",	"Jy per beam area",
	"JY/PIX",	"Jy per pixel",
	"MAG/PIX",	"Magnutides per pixel",
	"M/SEC",	"Meters per second",
	"DEGREES",	"Angle in degrees",
	"RA",		"Right Ascension (deg)",
	"RA---SIN",	"Right Ascension (deg)",
	"DEC",		"Declination (deg)",
	"DEC--SIN",	"Declination (deg)",
	"LL",		"Tangent plane E-W (deg)",
	"MM",		"Tangent plane N-S (deg)",
	"GLON",		"Galactic longitude (deg)",
	"GLAT",		"Galactic latitude (deg)",
	"ELON",		"Ecliptic longitude (deg)",
	"ELAT",		"Ecliptic latitude (deg)",
	"TIME",		"Time (sec)",
	"FREQ",		"Frequency (Hz)",
	"LAMBDA",	"Wavelength (meters)",
	"VELO",		"Velocity (m/sec)",
	"VELO-LSR",	"Velocity wrt local rest",
	"VELO-HEL",	"Velocity wrt Sun",
	"VELO-OBS",	"Velocity wrt observer",
	"PIXEL",	"Pixel",
	"STOKES",	"Stokes polarization",
	"COMPLEX",	"Complex",
	NULL
};

static		readblock(), reverse(), putedit(), cindex();

main(argc,argv)

	int	argc;			/* number of arguments */
	char	**argv;			/* array of argument pointers */
{
	char	c;
	char	buf[FITSBLK];		/* FITS input block */
	register char *b;		/* pointer to input buffer */
	int	i;
	int	n;
	double	d;
	FILE	*fi = stdin;		/* GIPS input file number */
	FILE	*fp;			/* FITS output file */
	register struct	data *pd;	/* pointer to header name table */
	/*
	 *	save and parse arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input files: %s",argv[2]);
	if (Sflag)
		telltale(0);
	/*
	 *	open the GIPS file
	 */
	if (argc > 1 && (fi = fopen(*++argv,"r")) == NULL)
		gipserr2(GIPSBADOPEN,*argv);
	/*
	 *	read the skeleton GIPS profile into core
	 */
	if ((fp = ihread(initfile,GIPSPROFILE)) == NULL)
		gipserr();
	else
		(void)ihbcl(fp);
	morehist = strcpy(alloc(100,1),"fits: ");
	if (fi != stdin && *argv != NULL) {
		(void)ihput("origin",*argv);
		(void)ihput("title",*argv);
	}
	/*
	 *	read the FITS header block(s), identify the header variables,
	 *	and copy them to static storage.
	 */
	for (i = 1; i && (readblock(fi,buf,FITSBLK) == FITSBLK); )
		for (b = buf; (b < buf+FITSBLK) && i; b += 80) {
			for (pd = fitsparm; cindex(b,pd->name); pd++)
				;
			if (pd->name == NULL)
				continue;
			if (pd->flg & IS_ARRAY)
				if (sscanf(b+5,"%d=",&n) != 1 ||
					--n < 0 || n > INDIMS)
					continue;
				else
					(void)sscanf(b+10,form[pd->flg&07].str,
					    pd->addr+n*(form[pd->flg&07].len));
			else if (pd->flg & IS_END)
				i--;
			else if ((pd->flg & IS_OPTHIST) == 0)
				(void)sscanf(b+10,form[pd->flg&07].str,pd->addr);
			else if (hflag) {
				char *ch = b+79;
				while (*ch == ' ') *ch-- = 0;
				(void)ihput("history",b+8);
			}
			pd->flg |= IS_FOUND;
		}
	/*
	 *	an endfile condition is bad news indeed
	 */
	if (i)
		error("EOF while reading FITS header: %s",*argv);
	/*
	 *	check that vital (IS_REQD) variables were located
	 */
	for (pd = fitsparm; pd->name != NULL; pd++)
		if ((pd->flg & IS_REQD) && (! pd->flg & IS_FOUND))
			(void) fprintf(stderr,"%s\n",(i++,pd->name));
	if (i)
		error("variables missing from FITS header: %s",*argv);
	/*
	 *	construct a history entry from IS_HISTORY variables
	 */
	for (n = 0, pd = fitsparm; pd->name != NULL; pd++)
		if (pd->flg & IS_HISTORY) {
			morehist = realloc(morehist,
				(unsigned)(strlen(morehist)+10));
			if (morehist != 0)
				(void)strncat(strcat(morehist," "),pd->addr,8);
		} else if (*pd->oname && (pd->flg & IS_FOUND))
			n = putedit(pd,0);
	if (pd->name != NULL && n > 0)
		error("error while reformatting ",pd->name);
	/*
	 *	fix up GIPS variables not directly translated from FITS header
	 */
	if (simple != 'T')
		error("This is not a SIMPLE image","");
	if (naxes < 2)
		error("bad NAXES: %d",(char *)naxes);
	switch(bitpix) {
		case 8:
			c = 'b';
			break;
		case 16:
			c = 'h';
			break;
		case 32:
			c = 'i';
			break;
		default:
			error("bad BITPIX: %d",(char *)bitpix);
	}
	(void)ihput("dtype", &c);
	(void)ihput("dscale",
		(bscale != GIPSLOW && bzero_ != GIPSLOW) ? "n" : "s");
	d = (cdelt[0] != 0) ? FLOAT(1, 0) : 0;
	(void)ihput("xmin", (char *) &d);
	d = (cdelt[1] != 0) ? FLOAT(1, 1) : 0;
	(void)ihput("ymin", (char *) &d);
	d = (cdelt[0] != 0) ? FLOAT(naxis[0],0) : naxis[0];
	(void)ihput("xmax", (char *) &d);
	d = (cdelt[1] != 0) ? FLOAT(naxis[1],1) : naxis[1];
	(void)ihput("ymax", (char *) &d);
	(void)ihput("xscale", "s");
	(void)ihput("yscale", "s");
	if (morehist != NULL)
		(void)ihput("history",morehist);
	/*
	**	If specified, merge additional header info
	*/
	if (strcmp(mfile, "-") && (fp = ihmerg(mfile)))
		(void)ihbcl(fp);
	/*
	 *	write the GIPS header
	 */
	if ((fp = ihwrit("-",0)) == NULL)
		gipserr();
	(void)signal(SIGPIPE,exit);
	/*
	 *	read the input file, byte-swap if requested, and write to stdout
	 */
	for (i = (naxis[0]*naxis[1]*bitpix)/8; i > 0 ; i -= n) {
		n = (i > FITSBLK) ? FITSBLK : i;
		if (readblock(fi,buf,n) != n)
			error("EOF before end of image: %s",*argv);
		if (bflag)
			swab(buf,buf,n);
		if (fwrite(buf,n,1,fp) != 1)
			gipserr2(GIPSBADWRITE,"'stdout'");
	}
	/*
	 *	advance to end of input file (if tape, this may be vital)
	 */
	(void)fseek(fi,0L,2);
	(void)fclose(fi);
	(void)ihbcl(fp);
	if (Sflag)
		telltale(1);
	(void)exit(0);
}
	/*
	**	Read next block
	*/
static
readblock(fi, buf, n)
FILE	*fi;
char	*buf;
{
	static	int	blen;
	static	char	*bsave, *bend;
	register l, len;

	if (! Bflag)
		return fread(buf, 1, n, fi);

	for (buf += n, l = n; l > 0; l -= len) {
		if (blen == 0) {
			if (fread((char *)&blen, sizeof(blen), 1, fi) != 1)
				return n-l;
			if (bflag)
				blen = reverse((char *)&blen);
			len = blen + sizeof(blen);
			bsave = bsave ? realloc(bsave, len) : alloc(blen, 1);
			if (! bsave || fread(bsave, len, 1, fi) != 1)
				return n-l;
			bend = bsave + blen;
		}
		len = (l > blen) ? blen : l;
		move (bend-blen, buf-l, len);
		blen -= len;
	}
	return n;
}
	/*
	**	Convert byte-swapped (VAX) integer
	*/
static
reverse(c)
register char	*c;
{
	register n;
	register char *d;

	for (n = 0, d = c+3; d >= c; )
		n = (n << 8) | (*d-- & 0377);
}
/*
 *	change a FITS unit descriptor string to GIPS format
 *	and insert a leading character to describe GIPS array variables
 */
static
putedit(pd,i)

	register struct data *pd;	/* pointer to FITS header table */
	int	i;			/* array index */
{
	register struct units *pu;	/* pointer to unit table */
	register char	*c = pd->addr;	/* pointer to FITS value */
	/*
	 *	if variable describes units, attempt to translate
	 */
	if (pd->flg & IS_UNITS)
		for (pu = unittab; pu->oldname != NULL && c == pd->addr; pu++)
			if (strcmp(pu->oldname,c) == 0)
				c = pu->newname;
	/*
	 *	if variable is array element, set first character in GIPS name
	 */
	if (pd->flg & IS_ARRAY) {
		if (i < OUTDIMS-1) {
			pd->addr += form[pd->flg & 7].len;
			(void)putedit(pd,i+1);
		}
		*pd->oname = "xyzw"[i];
	}
	/*
	 *	update GIPS header
	 */
	return(ihput(pd->oname,c));
}
/*
 *	return 0 if "c" and "d" are identical, non-zero otherwise
 */
static
cindex(c,d)

	register char *c;		/* string pointer */
	register char *d;		/* string pointer */
{
	if (d == NULL)
		return(0);
	while (*d && *c++ == *d++)
		;
	return(*d);
}
