#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihfromgif.c,v $	GIPS $Revision: 1.7 $ (MIT/CSR) $Date: 1998/05/13 01:03:18 $";
#endif

/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1992 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihfromgif.c,v $		Creation Date:	01/01/92
**	Version:	$Revision: 1.7 $		Update:		$Date: 1998/05/13 01:03:18 $
**	Function:	Convert GIF file to GIPS format
**	Components:	main()		GIPS/UNIX main program
*/

#include <stdio.h>
#include <signal.h>
#include <strings.h>
#include "gips.h"

	/*
	**	Input variables
	*/

static	char	*ifile	= NULL;		/* input raster file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= "-";		/* output GIPS file */
static	char	*afile	= "temp";	/* output colormap file */
static	flag	Sflag;			/* statistics flag */
static	flag	vFlag;			/* verbose flag */

	/*
	**	Variable definitions
	*/

static	flagentry	flagtable[] = {
	"S",	"1",			(flag *) &Sflag,
	"v",	"1",			(flag *) &vFlag,
	"a",	"-> map",		(flag *) &afile,
	"m",	"-> file",		(flag *) &mfile,
	"o",	"-> file",		(flag *) &ofile,
	NULL,	"[gif-file]",	NULL
};

	/*
	**	GIF flags
	*/

#define	GIF_ID_87	"GIF87a"
#define	GIF_ID_89	"GIF89a"
#define	GIF_EXTENSION	0x21
#define	GIF_COMMENT	0xfe
#define	IMAGESEP	0x2c
#define	INTERLACEMASK	0x40
#define	COLORMAPMASK	0x80

typedef	unsigned char	uchar;

	/*
	**	GIF file header
	*/

static	struct	{
	char	id[6];		/* GIF unique ID */
	uchar	dwidth[2];	/* display width */
	uchar	dheight[2];	/* display height */
	uchar	flg;		/* image flags */
	uchar	bkgr;		/* background color */
	uchar	aspect;		/* should be zero for GIF87 */
} GifHdr;

	/*
	**	GIF image header
	*/

static	struct	{
	uchar	sep;		/* image separator */
	uchar	left[2];	/* left margin */
	uchar	top[2];		/* top margin */
	uchar	width[2];	/* image width */
	uchar	height[2];	/* image height */
	uchar	interlace;	/* interlace flag */
	uchar	codesize;	/* initial code size */
} GifImg;

	/*
	**	Miscellaneous
	*/

static	FILE	*fp;		/* output image stream */
static	int	xnum, ynum;	/* image dimensions */
static	int	iFlag;		/* =1 if interlaced */
static	int	is89;		/* =1 for GIF89 format */
static	char	*outbuf;	/* interlace buffer */
static	char	*outline;	/* interlace output line */
static	int	OutShft;	/* output pixel bit shift */

static	copy_colormap(), copy_gif_image(), ReadCode(), switch_interlace();

main(argc, argv)
	int	argc;
	char	**argv;
{
	char	buf[256];	/* scratch */
	int	ch, ext;	/* scratch */

	/*
	**	Parse command string
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc < 2) {
		(void)signal(SIGPIPE, exit);
		ifile = "'stdout'";
	} else if (freopen(ifile = argv[1], "r", stdin) == NULL)
		perror(ifile), exit(1);
	if (Sflag)
		telltale(0);

	/*
	**	Read the GIF file header
	*/

	if (fread(&GifHdr, sizeof(GifHdr), 1, stdin) != 1)
		if (ferror(stdin))
			perror(ifile), exit(1);
		else
			error("%s: unexpected EOF in file header", ifile);
	else if (!strncmp(GifHdr.id, GIF_ID_89, 6))
		is89 = 1;
	else if (strncmp(GifHdr.id, GIF_ID_87, 6))
		error("%s: bad GIF header", ifile);
	if (vFlag)
		fprintf(stderr, "screen dimension: %dx%d\n",
			GifHdr.dwidth[0] + (GifHdr.dwidth[1] << 8),
			GifHdr.dheight[0] + (GifHdr.dheight[1] << 8));
	if (!is89 && GifHdr.aspect)
		error("%s: corrupt GIF file - bad end-of-header", ifile);

	/*
	**	If present, write the colormap
	*/

	if (GifHdr.flg & COLORMAPMASK)
		copy_colormap();
 
        printf( "ihfromgif: after copy_colormap() \n" );

	/*
	**	Check for extensions
	*/

	if ((ch = getchar()) == GIF_EXTENSION)
		for (ext = getchar(); (ch = getchar()) & 255; ) {
			(void)fread(buf, ch, 1, stdin);
			if (ext == GIF_COMMENT && vFlag)
				(void)fwrite(buf, ch, 1, stderr);
		}
	else
		ungetc(ch, stdin);

	/*
	**	Read the GIF image header
	*/

	if (fread(&GifImg, sizeof(GifImg), 1, stdin) != 1)
		if (ferror(stdin))
			perror(ifile), exit(1);
		else
			error("%s: unexpected EOF in image header", ifile);
	else if (GifImg.sep != IMAGESEP)
		error("%s: missing GIF image separator", ifile);
	xnum = GifImg.width[0] + (GifImg.width[1] << 8);
	ynum = GifImg.height[0] + (GifImg.height[1] << 8);
	iFlag = GifImg.interlace & INTERLACEMASK;

	/*
	**	Construct in-core GIPS header,
	**	updating from any -m option
	*/

	(void)strcpy(buf, GIPSPROFDIR);
	(void)strcat(buf, GINPROFILE);
	fp = ihread(buf, GIPSPROFILE);
	(void)ihbcl(fp);
	if (mfile != NULL && *mfile != '-')
		if (fp = ihmerg(mfile))
			(void)ihbcl(fp);
		else
			gipserr();
	ihput("xnum", &xnum);
	ihput("ynum", &ynum);
	putargs();

	/*
	**	Copy the image
	*/

	if (iFlag && (outbuf = outline = alloc(xnum, ynum)) == NULL)
		error("%s: insufficient memory for output buffer", ifile);
	if ((fp = ihwrit(ofile, 1)) == NULL)
		perror(ofile), exit(1);
	else if (fp == stdout)
		ofile = "'stdout'";
	copy_gif_image();
	if (! iFlag)
		;
	else if (fwrite(outbuf, xnum, ynum, fp) != ynum)
		gipserr2(GIPSBADWRITE, ofile);
	else
		free(outbuf);

	/*
	**	Test for trailing extensions
	*/

	while ((ch = getchar()) == GIF_EXTENSION)
		for (ext = getchar(); (ch = getchar()) & 255; ) {
			(void)fread(buf, ch, 1, stdin);
			if (ext == GIF_COMMENT && vFlag)
				(void)fwrite(buf, ch, 1, stderr);
		}

	/*
	**	Cleanup
	*/

	(void)ihbcl(fp);
	if (Sflag)
		telltale(1);
	exit(0);
}

	/*
	**	Copy colormap from GIF file
	*/

static
copy_colormap()
{
	FILE	*fa;			/* colormap file pointer */
	char	buf[256];		/* buffer for file name */
	char	*c;			/* scratch */
	int	size;			/* GIF colormap size */
	int	r, g, b;		/* red, green, blue intensities */
	int	i, j;			/* scratch */

	size = 2 << (GifHdr.flg & 7);
        OutShft = 7 - (GifHdr.flg & 7);
	if (! (c = rindex(afile, '.')) || strcmp(c, ".cmap"))
		afile = strcat(strcpy(buf, afile), ".cmap");
	if ((fa = fopen(afile, "w")) == NULL)
		perror(afile), exit(1);
	for (i = 0; i < size; i++) {
		r = getchar() & 255;
		g = getchar() & 255;
		b = getchar() & 255;
		for (j = 0; j < 256; j += size)
			(void)fprintf(fa, "%d\t%d\t%d\n", r, g, b);
	}
	(void)fclose(fa);
	if (vFlag)
		fprintf(stderr, "%s: %d colors copied to %s\n",
			ifile, size, afile);
}

	/*
	**	Decompress the GIF image
	*/

static
copy_gif_image()
{
static	int	Prefix[4096];	/* compression prefix table */
static	int	Suffix[4096];	/* compression suffix table */
static	int	OutCode[2049];	/* decompressor stack */
	int	*OutPtr;	/* Decompressor output stack pointer */
	int	CodeSize;	/* Code size, read from GIF header */
	int	InitCodeSize;	/* Starting code size, used during Clear */
	int	Code;		/* Value returned by ReadCode */
	int	MaxCode;	/* limiting value for current code size */
	int	ClearCode;	/* GIF clear code */
	int	EOFCode;	/* GIF end-of-information code */
	int	CurCode;	/* Decompressor variable */
	int	OldCode;	/* Decompressor variable */
	int	InCode;		/* Decompressor variable */
	int	FirstFree;	/* First free code; generated per GIF spec */
	int	FreeCode;	/* Decompressor, next free slot in hash table */
	int	FinChar;	/* Decompressor variable */
	int	ReadMask;	/* Code AND mask for current code size */
	int	BitMask;	/* AND mask for data size */
	int	XC;		/* interlace sample index */

	if (vFlag)
		fprintf(stderr, "%s: copying %dx%d %sinterlaced image\n",
			ifile, xnum, ynum, iFlag ? "" : "non-");
	XC = 0;
	CodeSize = GifImg.codesize;
	BitMask = (2 << (GifHdr.flg & 7)) - 1;
	OutPtr = OutCode;
	ClearCode = 1 << CodeSize;
	EOFCode = ClearCode + 1;
	FreeCode = FirstFree = ClearCode + 2;
	InitCodeSize = ++CodeSize;
	MaxCode = 1 << CodeSize;
	ReadMask = MaxCode - 1;

	while ((Code = ReadCode(CodeSize) & ReadMask) != EOFCode) {
		if (Code == ClearCode) {
			CodeSize = InitCodeSize;
			MaxCode = 1 << CodeSize;
			ReadMask = MaxCode - 1;
			FreeCode = FirstFree;
			Code = ReadCode(CodeSize) & ReadMask;
			CurCode = OldCode = Code;
			FinChar = CurCode & BitMask;
			if (! iFlag)
				putc(FinChar << OutShft, fp);
			else {
				*outline++ = FinChar;
				if ((++XC % xnum) == 0)
					switch_interlace();
			}
			continue;
		}
	    	CurCode = InCode = Code;
	    	if (CurCode >= FreeCode) {
			CurCode = OldCode;
			*OutPtr++ = FinChar;
	    	}
	    	while (CurCode > BitMask) {
			if (OutPtr > &OutCode[2048])
		    		error("%s: GIF buffer overflow", ifile);
			*OutPtr++ = Suffix[CurCode];
			CurCode = Prefix[CurCode];
	    	}
	    	FinChar = CurCode & BitMask;
	    	*OutPtr++ = FinChar;
		while (--OutPtr >= OutCode)
			if (! iFlag)
				putc(*OutPtr << OutShft, fp);
			else {
				*outline++ = *OutPtr;
				if ((++XC % xnum) == 0)
					switch_interlace();
			}
	    	OutPtr = OutCode;
	    	Prefix[FreeCode] = OldCode;
	    	Suffix[FreeCode] = FinChar;
	    	OldCode = InCode;
	    	if (++FreeCode >= MaxCode && CodeSize < 12) {
		    	CodeSize++;
		    	MaxCode *= 2;
		    	ReadMask = (1 << CodeSize) - 1;
	    	}
	}
}

	/*
	**	Unpack the next code
	*/

static
ReadCode(CodeSize)
	register int	CodeSize;	/* number of bits in code */
{
static	int		BitOffset;	/* bit offset of next code */
static	uchar		InBuf[4];	/* input buffer */
static	int		InOff = -1;	/* byte offset of last byte read */
static	int		BufLen = 1;	/* input buffer length remaining */
	register int	Code;		/* current code */
	register int	Shft;		/* current bit shift */
	register int	CurOff;		/* current byte offset */

	CurOff = BitOffset / 8;

	/*
	**	Get 2 or 3 input bytes, taking them from InBuf if
	**	possible, otherwise read them from stdin.
	*/

	for (Code = Shft = 0; Shft <= CodeSize+8; CurOff++, Shft += 8) {
		if (CurOff <= InOff) {
			Code |= InBuf[CurOff & 3] << Shft;
			continue;
		}

		/*
		**	Strip off the input buffer length bytes
		*/

		if (--BufLen <= 0)
			if (BufLen < 0)
				error("%s: unexpected end of image", ifile);
			else if ((BufLen = getchar()) == EOF)
				error("%s: unexpected EOF", ifile);
			else if (BufLen == 0)
				continue;

		InOff = CurOff;
		Code |= (InBuf[CurOff & 3] = getchar()) << Shft;
	}

	/*
	**	Adjust the input bits and advance the bit offset
	*/

	Code >>= BitOffset % 8;
	BitOffset += CodeSize;
	return Code;
}

	/*
	**	Reset outline pointer
	*/

static
switch_interlace()
{
static	int	Pass;		/* interlace pass */
static	int	YC;		/* output line */

	switch (Pass) {
	case 0:
		YC += 8;
		if (YC >= ynum) {
			Pass++;
			YC = 4;
		}
		break;
	case 1:
		YC += 8;
		if (YC >= ynum) {
			Pass++;
			YC = 2;
		}
		break;
	case 2:
		YC += 4;
		if (YC >= ynum) {
			Pass++;
			YC = 1;
		}
		break;
	case 3:
		YC += 2;
		break;
	default:
		break;
	}
	outline = outbuf + YC * xnum;
}
