#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihfrompix.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/05/13 01:01:48 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihfrompix.c,v $		Creation Date:	05/10/88
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/05/13 01:01:48 $
**	Function:	Convert Sun raster file to GIPS format
**	Components:	main()		GIPS/UNIX main program
*/
#include <stdio.h>
#include <signal.h>
#include <pixrect/pixrect_hs.h>
#include "gips.h"
	/*
	**	Input variables
	*/
static	char	*ifile	= NULL;		/* input raster file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= "-";		/* output GIPS file */
static	char	*afile	= "temp";	/* output colormap file */
static	flag	Sflag;			/* statistics flag */
static	flag	rflag;			/* select red plane of 24 bits */
static	flag	gflag;			/* select green plane of 24 bits */
static	flag	bflag;			/* select blue plane of 24 bits */
	/*
	**	Variable definitions
	*/
static	flagentry	flagtable[] = {
	"a",	"-> map",		(flag *) &afile,
	"S",	"1",			(flag *) &Sflag,
	"m",	"-> file",		(flag *) &mfile,
	"o",	"-> file",		(flag *) &ofile,
	"r",	"1",			(flag *) &rflag,
	"g",	"1",			(flag *) &gflag,
	"b",	"1",			(flag *) &bflag,
	NULL,	"[raster file]",	NULL
};

extern	char	*rindex(), *strcpy(), *strcat();
static		copy_colormap();

main(argc, argv)
	int	argc;
	char	**argv;
{
	struct	rasterfile	hdr;		/* Sun raster header */
	colormap_t		cmap;		/* Sun colormap */
	struct	pixrect		*pr;		/* Sun pixrect */
	short			*bin;		/* input buffer */
	char			buf[256];	/* scratch */
	register	char	*c;		/* scratch */
	register	int	i;		/* scratch */
	register	FILE	*fp;		/* output image stream */
	/*
	**	Parse command string
	*/
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc < 2) {
		(void)signal(SIGPIPE, exit);
		ifile = "'stdout'";
	} else if (freopen(ifile = argv[1], "r", stdin) == NULL)
		perror(ifile), exit(1);
	if (! (rflag | gflag | bflag))
		rflag = gflag = bflag = 1;
	if (Sflag)
		telltale(0);
	/*
	**	Read the input header
	*/
	if (pr_load_header(stdin, &hdr) == PIX_ERR)
		error("%s: cannot load rasterfile header", ifile);
	/*
	**	Copy the colormap
	*/
	if (hdr.ras_maplength)
		copy_colormap(&hdr, &cmap);
	/*
	**	Load the image into a pixrect
	*/
	if ((pr = pr_load_image(stdin, &hdr, &cmap)) == NULL)
		error("%s: cannot load raster image", ifile);
	(void)fclose(stdin);
	bin = mpr_d(pr)->md_image;
	/*
	**	Construct in-core GIPS header,
	**	updating from any -m option
	*/
	(void)strcpy(buf, GIPSPROFDIR);
	(void)strcat(buf, GINPROFILE);
	fp = ihread(buf, GIPSPROFILE);
	(void)ihbcl(fp);
	ihput("xnum", &hdr.ras_width);
	ihput("ynum", &hdr.ras_height);
	if (mfile != NULL && *mfile != '-')
		if (fp = ihmerg(mfile))
			(void)ihbcl(fp);
		else
			gipserr();
	ihget("xnum", &hdr.ras_width, 0);
	ihget("ynum", &hdr.ras_height, 0);
	putargs();
	/*
	**	Write the output image header
	*/
	if ((fp = ihwrit(ofile, 1)) == NULL)
		perror(ofile), exit(1);
	else if (fp == stdout)
		ofile = "'stdout'";
	/*
	**	Copy the image
	*/
	while (--hdr.ras_height >= 0) {
		register short *ic;
		register short	na, nc;

		i = hdr.ras_width;
		switch (hdr.ras_depth) {
		case 1:
			ic = bin;
			for (na = nc = 0; --i >= 0; na <<= 1) {
				if (nc-- == 0)
					na = *ic++, nc = 16;
				putc((na & 0x8000) ? 255 : 0, fp);
			}
			break;
		case 8:
			fwrite((char *)bin, i, 1, fp);
			break;
		case 24:
			for (c = (char *)bin; --i >= 0; c += 3)
				putc((rflag ? c[0] : 0) | 
					(gflag ? c[1] : 0) |
					(bflag ? c[2] : 0), fp);
			break;
		default:
			fprintf(stderr, "%s: depth %d not supported\n",
				ifile, hdr.ras_depth);
			exit(1);
		}
		if (ferror(fp))
			perror(ofile), exit(1);
		bin += (mpr_d(pr)->md_linebytes)/2;
	}
	(void)ihbcl(fp);
	if (Sflag)
		telltale(1);
	exit(0);
}
	/*
	**	Read and possible write the colormap
	*/
static
copy_colormap(hp, cp)
	struct  rasterfile *hp;
	colormap_t	*cp;
{
	char		*c;
	FILE		*fp;
	char		buf[1024];
	int		i;

	cp->type = hp->ras_maptype;
	cp->length = hp->ras_maplength / 3.0;
	cp->map[0] = (unsigned char *)alloc(cp->length, 1);
	cp->map[1] = (unsigned char *)alloc(cp->length, 1);
	cp->map[2] = (unsigned char *)alloc(cp->length, 1);
	if (pr_load_colormap(stdin, hp, cp) == PIX_ERR)
		error("%s: cannot load colormap", ifile);
	if (hp->ras_maptype != RMT_EQUAL_RGB)
		return;
	if (! (c = rindex(afile, ".")) || strcmp(c, ".cmap"))
		afile = strcat(strcpy(buf, afile), ".cmap");
	if ((fp = fopen(afile, "w")) == NULL)
		perror(afile), exit(1);
	for (i = 0; i < 256; i++)
		fprintf(fp, "%d\t%d\t%d\n", cp->map[0][i],
			cp->map[1][i], cp->map[2][i]);
	fclose(fp);
}
