#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihgeom.c,v $	GIPS $Revision: 2.13 $ (MIT/CSR) $Date: 1998/05/13 01:01:49 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihgeom.c	Creation Date:	04/11/84
**	Version:	$Revision: 2.13 $		Update:		$Date: 1998/05/13 01:01:49 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Transform a General Image via a Geometry file
*/

#define		GRID
#define		NUMRING		8

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"
	/*
	 *	variables set by gipsflags()
	 */
static	char	*afile	= NULL;		/* reformatted input image [obs] */
static	char	*ifile	= "'stdin'";	/* input geometry file */
static	char	*mfile	= "-";		/* merged-header file */
static	char	*ofile	= "-";		/* output general image file */
static	int	nrings	= NUMRING;	/* number of ring buffers */
static	flag	Dflag;			/* debug flag */
static	flag	Sflag;			/* generate memory statistics */
static	flag	fflag;			/* input is formatted */
static	flag	nflag;			/* suppress image output */
static	flag	xflag;			/* interpolate to get pixel value */
	/*
	 *	argument table for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"D",	"1",		&Dflag,
	"S",	"1",		&Sflag,
	"f",	"1",		&fflag,
	"n",	"1",		&nflag,
	"x",	"1",		&xflag,
	"",	"%d buffers",	(flag *) &nrings,
	"a",	"-> file",	(flag *) &afile,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	NULL,	"name [name]",	NULL
};
	/*
	 *	control grid file header
	 */
static	int	gxpix;		/* number of x-pixels (columns) */
static	int	gypix;		/* number of y-pixels (rows) */
static	int	nrast;		/* number of y-pixels per output block */
static	char	mapbuf[BUFLEN];	/* mapping parameters */

static	gimage	*gi;		/* pointer to image control block */

static		putblock();
extern	char	*index();	/* C library */
	/*
	 *	open contol grid file and read header
	 */
static	FILE *
getifile(s)
	register char	*s;
{
	char	buf[BUFLEN];
	register FILE	*fp;

	if (s == NULL)
		fp = stdin;
	else if (index(s,' ') || index(s,'\t') || index(s,'\n')) {
		fp = fopens(ifile = s);
		fflag++;
	} else if ((fp = fopen(ifile = s,"r")) == NULL)
		gipserr2(GIPSBADOPEN,ifile);
	/*
	 *	Read the control grid file header
	 */
	if (fgets(buf, sizeof(buf)-1, fp) == NULL)
		gipserr2(GIPSBADREAD, ifile);
	if (sscanf(buf, GIPS_GRID, &gxpix, &gypix, &nrast) != 3)
		gipserr2(GIPSBADDATA, ifile);
	if (Dflag)
		(void) fprintf(stderr,"ihgeom: %s", buf);
	if (fgets(mapbuf, sizeof(mapbuf)-1, fp) == NULL)
		gipserr2(GIPSBADREAD, ifile);
	if (Dflag)
		(void) fprintf(stderr,"ihgeom: %s", mapbuf);
	return(fp);
}

main(argc,argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	char	dtype;		/* output pixel type code */
	register dlen;		/* output pixel length in bytes */
	register ybase;		/* y-index of first buffer raster */
	register i;		/* scratch */
	FILE	*fp;		/* input cell-format image stream pointer */
	FILE	*fm;		/* pointer to -m aux input stream */
	FILE	*fi;		/* input geometry file stream pointer */
	FILE	*fo;		/* output GIPS stream pointer */
	char	*outbuf;	/* output buffer */
	int	buflen;		/* length of outbuf in bytes */
	int	dx, dy;		/* interpolator box dimensions */
	float	ddx, ddy;	/* interpolation weights */
	double	lonlat[4];	/* longitude/latitude limits */
	int	nbuf;		/* count of output buffers */
	int	ncell;		/* count of interpolation cells */
	char	gbuf[BUFLEN];	/* buffer for geometry */

	struct	{		/* control grid cell */
		int	xy[4];	/* box output pixel values */
		float	v[8];	/* box input pixel values */
	}	cell;
	/*
	**	save and parse callers arguments
	*/
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (afile == NULL)
		afile = *++argv, --argc;
	if (argc > 2)
		error("bad arg: %s",argv[2]);
	if (afile == NULL)
		if (nflag)
			--argv;
		else
			error("no input image file specified","");
	if (Sflag)
		telltale(0);
	if (Dflag)
		setbuf(stderr,alloc(BUFSIZ,1));
	/*
	 *	Open input geometry gfile
	 */
	fi = getifile(*++argv);
	/*
	 *	Locate the reformatted input image and
	 *	fix-up the output general image header
	 */
	if (! nflag) {
		if ((fp = ihread(afile,GIPSPROFILE)) == NULL)
			gipserr();
		(void)ihdel("xmin");
		(void)ihdel("xmax");
		(void)ihdel("ymin");
		(void)ihdel("ymax");
		(void)ihdel("xscale");
		if (*mfile != '-')
			if ((fm = ihmerg(mfile)) == NULL)
				gipserr();
			else
				(void)(ihbcl(fm));
		if ((gi = gopen(fp, nrings)) == NULL)
			gipserr();
		if (ihget("geometry", gbuf, sizeof(gbuf)) ||
			sscanf(gbuf, "type=%1s", &dtype) != 1)
			gipserr();
		if (dtype == 'M') {
			dtype = 'f';
			ihput("dscale", "s");
			xflag = 0;
		}
		if ((dlen = ihdtype(dtype)) == 0)
			gipserr();
		(void)ihput("xnum",(char *)&gxpix);
		(void)ihput("ynum",(char *)&gypix);
		(void)ihput("dtype",&dtype);
		(void)ihdel("geometry");
		if (*mapbuf && strncmp(mapbuf, "name=Tiepoint type=0 ", 21)) {
			if (ihput("mapping", mapbuf))
				gipserr();
			if (g_setup())
				gipserr();
			if (g_unmap(0, gypix/2, &lonlat[0]))
				;
			else if (g_unmap(gxpix, gypix/2, &lonlat[2]))
				;
			else if (lonlat[0] != lonlat[2]) {
				ihput("xmin", &lonlat[0]);
				ihput("xmax", &lonlat[2]);
			}
			if (g_unmap(gxpix/2, 0, &lonlat[0]))
				;
			else if (g_unmap(gxpix/2, gypix, &lonlat[2]))
				;
			else if (lonlat[1] != lonlat[3]) {
				ihput("ymin", &lonlat[1]);
				ihput("ymax", &lonlat[3]);
			}
		} else
			(void)ihdel("mapping");
		if (*mapbuf && !strncmp(mapbuf, "name=Mercator type=8 ", 21))
			(void)ihput("yscale","m");
		else
			(void)ihput("yscale","s");
		putargs();
		if ((fo = ihwrit(ofile,0)) == NULL)
			gipserr();
		if (fo == stdout) {
			(void)(signal(SIGPIPE,exit));
			ofile = "'stdout'";
		}
		/*
		 *	allocate the output buffer
		 */
		outbuf = alloc(nrast*gxpix,dlen);
		buflen = nrast*gxpix*dlen;
	}
	/*
	 *	loop over control grid cells
	 */
	for (nbuf = ncell = ybase = 0; ; ncell++) {
		if (fflag) {
			for (i = 0; i < 4; i++)
				if (fscanf(fi," %d",&cell.xy[i]) != 1)
					break;
			if (feof(fi) || ferror(fi))
				break;
			for (i = 0; i < 8; i++)
				if (fscanf(fi," %f",&cell.v[i]) != 1)
					break;
			if (feof(fi) || ferror(fi))
				break;
		} else if (fread((char *)&cell,sizeof(cell),1,fi) != 1)
				break;

		dx = cell.xy[2] - cell.xy[0];
		if (cell.xy[0] < 0 || cell.xy[2] > gxpix || dx <= 0)
			error ("bad x-coordinate in cell: %d",(char *)ncell);

		dy = cell.xy[3] - cell.xy[1];
		if (cell.xy[1] < 0 || cell.xy[3] > gypix || dy <= 0)
			error ("bad y-coordinate in cell: %d",(char *)ncell);

		while (cell.xy[1] >= ybase+nrast) {
			if (! nflag) {
				if (fwrite(outbuf,buflen,1,fo) != 1)
					gipserr2(GIPSBADWRITE,ofile);
				clear(outbuf,buflen);
			}
			ybase += nrast;
			nbuf++;
		}

		if (cell.xy[3] > ybase+nrast)
			error ("cell spans output buffers: %d",(char *)ncell);

		if (Dflag) (void) fprintf(stderr,
		"%d,%d:\t(%g,%g) (%g,%g) (%g,%g) (%g,%g) -> (%d,%d),(%d,%d)\n",
			nbuf,ncell,cell.v[0],cell.v[1],cell.v[2],cell.v[3],
			cell.v[4],cell.v[5],cell.v[6],cell.v[7],
			cell.xy[0],cell.xy[1],cell.xy[2],cell.xy[3]);

		ddx = 1/(float)dx;
		ddy = 1/(float)dy;

		cell.v[4] = (cell.v[4] - cell.v[0])*ddy;
		cell.v[5] = (cell.v[5] - cell.v[1])*ddy;

		if (! nflag) putblock(dx, dy,
			outbuf+(cell.xy[0]+(cell.xy[1]-ybase)*gxpix)*dlen,
			(gxpix-dx)*dlen, dlen,
			xflag ? cell.v[0] : cell.v[0]+0.5,
			xflag ? cell.v[1] : cell.v[1]+0.5,
			(cell.v[2]-cell.v[0])*ddx, (cell.v[3]-cell.v[1])*ddx,
			cell.v[4], cell.v[5],
			((cell.v[6]-cell.v[2])*ddy-cell.v[4])*ddx,
			((cell.v[7]-cell.v[3])*ddy-cell.v[5])*ddx);
	}
	if (ferror(fi))
		gipserr2(GIPSBADREAD,ifile);
	/*
	 *	if necessary, flush the output buffer(s)
	 */
	if (! nflag) {
		while (ybase < gypix) {
			i = gypix - ybase;
			if (i > nrast)
				i = nrast;
			if (fwrite(outbuf,gxpix*i,dlen,fo) != dlen)
				gipserr2(GIPSBADWRITE,ofile);
			clear(outbuf,buflen);
			ybase += nrast;
		}
		(void)(ihbcl(fo));
	}
	if (Sflag)
		telltale(1);
	exit(0);
}

static
putblock(x,y,out,dout,dl,v0x,v0y,v1x,v1y,v2x,v2y,v3x,v3y)

	int		x, y;
	register char	*out;
	register	dout, dl;
	float		v0x,v0y,v1x,v1y,v2x,v2y,v3x,v3y;
{
	register int	xx;
	register char	*c;
	float		vx, vy;

	for (vx = v0x, vy = v0y ; --y >= 0; out += dout) {
		for (xx = x; --xx >= 0; vx += v1x, vy += v1y, out += dl)
			if (c = xflag ? ginterp(vx, vy, gi) :
					gcell((int)vx, (int)vy, gi))
				switch (dl) {
				case 1:
					*out = *c;
					break;
				case 2:
					*((short *)out) = *((short *)c);
					break;
				case 4:
					*((long *)out) = *((long *)c);
					break;
				case 8:
					*((double *)out) = *((double *)c);
					break;
				}
		vx = v0x += v2x; vy = v0y += v2y;
		v1x += v3x; v1y += v3y;
	}
}
