#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihgfmt.c,v $	GIPS $Revision: 2.6 $ (MIT/CSR) $Date: 1998/11/13 23:53:37 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihgfmt.c	Creation Date:	06/13/84
**	Version:	$Revision: 2.6 $		Update:		$Date: 1998/11/13 23:53:37 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Re-structure a General Image to random-access format
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"

#define		OUTBLOCK	1024
	/*
	 *	variables set by gipsflags()
	 */
static	char	*ddir	= "/tmp";	/* scratch directory */
static	char	*ifile	= "-";		/* input file */
static	char	*ofile	= "-";		/* output file */
static	char	*mfile	= "-";		/* merged-header file */
static	int	block	= OUTBLOCK;	/* cell block length in bytes */
static	flag	Nflag;			/* no zero pad after header */
static	flag	Sflag;			/* generate memory statistics */
	/*
	 *	argument table for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"N",	"1",		&Nflag,
	"S",	"1",		&Sflag,
	"d",	"-> dir",	(flag *) &ddir,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"",	"%d block",	(flag *) &block,
	NULL,	"[name]",	NULL
};

static	char	tempfile[100];			/* scratch header file */
static	char	*tmpl = "%s/ihgXXXXXX";		/* scratch header file */
extern	char	*mktemp();			/* C library */

main(argc,argv)

	int		argc;		/* number of arguments */
	char		**argv;		/* array of argument pointers */
{
	FILE		*fi;		/* input stream pointer */
	FILE		*fo;		/* output stream pointer */
	int		xnum;		/* number of x pixels in image */
	int		ydim;		/* number of y pixels in cell */
	char		*buf;		/* ring-buffer pointer */
	char		*inx;		/* x-raster pointer */
	char		dtype;		/* pixel type */
	char		gbuf[100];	/* geometry value */

	register char	*iny;		/* cell raster origin */
	register int	i;		/* scratch */
	register	xdim;		/* number of x pixels in cell */
	register	xlen;		/* number of pixels across image */
	register	dlen;		/* pixel length in bytes */
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("bad arg: %s",argv[2]);
	if (Sflag)
		telltale(0);
	if (argc == 2)
		ifile = *++argv;
	/*
	 *	save input file and pixel type code
	 *	then, if requested, update header and test
	 */
	if ((fi = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fi == stdin)
		ifile = "'stdin'";
	if (*mfile != '-')
		if ((fo = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fo));
	if (ihget("xnum",(char *)&xnum,1) ||
	    ihget("dtype",&dtype,1))
		gipserr();
	/*
	 *	determine the cell size
	 */
	if ((dlen = ihdtype(dtype)) == 0)
		gipserr1(GIPSBADTYPE);
	for (xdim = 1, i = (block/dlen) << 1; i >>= 2; xdim <<= 1)
		;
	ydim = block / (xdim*dlen);
	if (xdim * ydim * dlen != block)
		error("bad -%d specification",(char *)block);
	buf = alloc(xdim+xnum*ydim,dlen);

	sprintf(gbuf, "type=%c block=%d xdim=%d ydim=%d",
		dtype,block,xdim,ydim);
	(void)(ihput("geometry",gbuf));
	(void)(ihput("dtype","G"));
	putargs();

	if (ofile[0] == '-' && ofile[1] == '\0') {
		sprintf(tempfile,tmpl,ddir);
		ofile = mktemp(tempfile);
	}

	if ((fo = ihwrit(ofile,0)) == NULL)
		gipserr();
	i = ftell(fo);
	if (ofile == tempfile) {
		(void)(ihbcl(fo));
		(void)(unlink(ofile));
		if ((fo = ihwrit("-",0)) == NULL)
			gipserr();
		(void)(signal(SIGPIPE,exit));
		ofile = "'stdout'";
	}
	if (! Nflag)
		while (i++ % block)
			(void)(putc('\0',fo));

	xlen = xnum*dlen;
	while (fread(buf,xlen,ydim,fi) > 0)
		for (inx = buf; inx < buf+xlen; inx += xdim*dlen)
			for (iny = inx, i = ydim; --i >= 0; iny += xlen)
				if (fwrite(iny,dlen,xdim,fo) != xdim)
					gipserr2(GIPSBADWRITE,ofile);
	(void) ihbcl(fi);
	(void) ihbcl(fo);
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}
