#ifndef lint
static char *sccsid = "%Z%$RCSfile: box.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1998/05/13 01:21:39 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		box.c		Creation Date:	09/05/84
**	Version:	$Revision: 2.7 $		Update:		$Date: 1998/05/13 01:21:39 $
**	Components:	box()		ihgrid() subroutine
**
**	Function:	Define recursive control grid partition
*/

#include	"ihgrid.h"

static	box();

firstbox(xpix, ypix, nrast)
	register xpix, ypix, nrast;
{
	register x, y;
	struct	cell	cell;

	pix_unmap(cell.xy[0] = 0, 0, &cell, 0);
	pix_unmap(cell.xy[2] = xpix,0, &cell, 1);
	/*
	**	March down the output image, ncell rasters at a time
	*/
	for (y = 0; y < ypix; ) {
		pix_unmap(cell.xy[0],cell.xy[1] = y, &cell, 2);
		y += nrast;
		if (y > ypix) y = ypix;
		pix_unmap(cell.xy[2],cell.xy[3] = y, &cell, 3);
		box(cell,cell);
		cell.v[0] = cell.v[2]; cell.flg[0] = cell.flg[2];
		cell.v[1] = cell.v[3]; cell.flg[1] = cell.flg[3];
	}
}
	/*
	**	Examine an interpolation cell
	*/
static
box(ocell,icell)
	struct	cell	ocell;	/* cell to be interpolated */
	struct	cell	icell;	/* copy of ocell, later a sub-cell */
{
	register	i;	/* scratch */
	register int	n;	/* number of invalid cell vertices */
	register int	m;	/* number of plottable cell vertices */

	register	dx = ocell.xy[2] - ocell.xy[0];
	register	dy = ocell.xy[3] - ocell.xy[1];

	if (dx <= 0 || dy <= 0)
		return;
	/*
	**	if cell not square enough, divide it in 2
	*/
	if (dx >= 2*dy) {
		dx >>= 1;
		ocell.xy[0] += dx;
		icell.xy[2] = ocell.xy[0],
		pix_unmap(icell.xy[2], icell.xy[1], &icell, 1);
		pix_unmap(icell.xy[2], icell.xy[3], &icell, 3);
		box(icell,icell);
		ocell.v[0] = icell.v[1]; ocell.flg[0] = icell.flg[1];
		ocell.v[2] = icell.v[3]; ocell.flg[2] = icell.flg[3];
		box(ocell,ocell);
		return;
	}
	/*
	**	remember validity of corners
	*/
	for (i = n = m = 0; i < 4; i++)
		if (ocell.flg[i] == VERTEX_NONPHYS)
			n++;
		else if (ocell.flg[i] != VERTEX_OFFMAP)
			m++;
	/*
	**	if a single pixel is valid, we gotta write it out
	*/
	if (dx == 1 || (m == 0 && (n == 0 || n == 4))) {
		if (ocell.flg[0] == 0)
			putcell(&ocell);
		return;
	}
	/*
	**	compute input coordinates of central pixel of cell
	*/
	icell.xy[2] = ocell.xy[0] + (dx>>1);
	icell.xy[3] = ocell.xy[1] + (dy>>1);
	pix_unmap(icell.xy[2], icell.xy[3], &icell, 3);
	/*
	**	determine whether cell is "linear", i.e. whether the
	**	central cell coordinates can be linearly interpolated 
	**	from the corner values in both x and y.
	*/
	if (m == 4) {
		float	fx, fy, ddx, ddy;

		ddx = (float)(dx>>1) / (float)dx;
		ddy = (float)(dy>>1) / (float)dy;

		fx = (1-ddx)*(1-ddy)*ocell.v[0].f[0]
			+ ddx*(1-ddy)*ocell.v[1].f[0]
			+ (1-ddx)*ddy*ocell.v[2].f[0]
			+ ddx*ddy*ocell.v[3].f[0];

		if (fabs(fx-icell.v[3].f[0]) < 0.5) {
			fy = (1-ddx)*(1-ddy)*ocell.v[0].f[1]
				+ ddx*(1-ddy)*ocell.v[1].f[1]
				+ (1-ddx)*ddy*ocell.v[2].f[1]
				+ ddx*ddy*ocell.v[3].f[1];
				
			if (fabs(fy-icell.v[3].f[1]) < 0.5) {
				putcell(&ocell);
				return;
			}
		}
	}
	/*
	**	call box recursively for each quadrant of current cell
	*/
	pix_unmap(icell.xy[2], icell.xy[1], &icell, 1);
	pix_unmap(icell.xy[0], icell.xy[3], &icell, 2);
	box(icell,icell);
	ocell.v[0] = icell.v[2]; ocell.flg[0] = icell.flg[2];
	icell.xy[0] = icell.xy[2]; icell.xy[2] = ocell.xy[2];
	icell.v[0] = icell.v[1]; icell.flg[0] = icell.flg[1];
	icell.v[1] = ocell.v[1]; icell.flg[1] = ocell.flg[1];
	icell.v[2] = icell.v[3]; icell.flg[2] = icell.flg[3];
	pix_unmap(icell.xy[2], icell.xy[3], &icell, 3);
	box(icell,icell);
	icell.xy[2] = icell.xy[0]; icell.xy[0] = ocell.xy[0];
	icell.xy[1] = icell.xy[3]; icell.xy[3] = ocell.xy[3];
	icell.v[0] = ocell.v[0]; icell.flg[0] = ocell.flg[0];
	ocell.v[0] = icell.v[3]; ocell.flg[0] = icell.flg[3];
	icell.v[1] = icell.v[2]; icell.flg[1] = icell.flg[2];
	icell.v[2] = ocell.v[2]; icell.flg[2] = ocell.flg[2];
	pix_unmap(icell.xy[2], icell.xy[3], &icell, 3);
	box(icell,icell);
	icell.xy[0] = icell.xy[2]; icell.xy[2] = ocell.xy[2];
	icell.v[0] = icell.v[1]; icell.flg[0] = icell.flg[1];
	icell.v[1] = ocell.v[0]; icell.flg[1] = ocell.flg[0];
	icell.v[2] = icell.v[3]; icell.flg[2] = icell.flg[3];
	icell.v[3] = ocell.v[3]; icell.flg[3] = ocell.flg[3];
	box(icell,icell);
}
