#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihgrid.c,v $	GIPS $Revision: 2.21 $ (MIT/CSR) $Date: 2004/11/02 20:20:36 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1986,1987 Massachusetts Institute of Technology
**
**	Module:		ihgrid.c	Creation Date:	07/27/86
**	Version:	$Revision: 2.21 $		Update:		$Date: 2004/11/02 20:20:36 $
**	Components:	main()		UNIX command
**
**	Function:	Generate an Image Geometry File
*/

#include	"ihgrid.h"
#include	<signal.h>

	/*
	 *	variables set by gipsflags()
	 */
	flag	Dflag;		/* debug flag */
	int	order;		/* tie-point polynomial order */
static	char	*ifile	= "-";	/* input tie-point and/or header file */
static	char	*mfile	= "-";	/* second input image header file */
static	char	*ofile	= "-";	/* output geometry file */
static	flag	fflag;		/* formatted (ASCII) output */
static	flag	Sflag;		/* generate memory statistics */
static	flag	xyrot;		/* x and y axis rotation */
static	char	*centers;	/* Center longitude and latitude */
static	char	*erot;		/* Euler rotation angles */
static	int	nproj;		/* map projection - index to table[] */
static	double	X0 = 1.e20;	/* output x-origin */
static	double	Y0 = 1.e20;	/* output y-origin */
static	int	gxpix;		/* number of output x-pixels */
static	int	gypix;		/* number of output y-pixels */
static	int	nrast;		/* number of rows per output block */
static	double	phi0;		/* First standard parallel */
static	double	phi1;		/* Second standard parallel */
static	double	radius;		/* map dimension */
static	double	radfact	= 1;	/* map dimension scale factor */
	/*
	 *	argument table for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"D",	"1",		&Dflag,
	"R",	"%lf fact",	(flag *) &radfact,
	"S",	"1",		&Sflag,
	"X",	"%lf org",	(flag *) &X0,
	"Y",	"%lf org",	(flag *) &Y0,
	"",	"%d nproj",	(flag *) &nproj,
	"b",	"%d size",	(flag *) &nrast,
	"c",	"-> lon,lat",	(flag *) &centers,
	"f",	"1",		&fflag,
	"n",	"%d order",	(flag *) &order,
	"p",	"%lf lat",	(flag *) &phi0,
	"P",	"%lf lat2",	(flag *) &phi1,
	"l",	"-> l,t,p",	(flag *) &erot,
	"r",	"%lf rad",	(flag *) &radius,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"s",	"1",		(flag *) &xyrot,
	"x",	"%d num",	(flag *) &gxpix,
	"y",	"%d num",	(flag *) &gypix,
	NULL,	"[name]",	NULL
};

extern	int	tiepoint();	/* Tiepoint transformation */
extern	char	*g_getname();	/* return mapping title */
	/*
	 *	local stuff
	 */
static	int	xnum, ynum;	/* number of input pixels */
static	float	fxnum, fynum;	/* size of input image */
static	int	ncell;		/* number of output cells */
static	int	xcell, ycell;	/* cumulative cell range */
static	double	eangle[3];	/* Euler rotation angles */
static	char	mapbuf[BUFLEN];	/* mapping definition */
static	G_PARM	*proj0;		/* pointer to input projection table */
static	char	*mapname = "";	/* mapping title */
static	FILE	*fp;		/* output GIPS stream pointer */

static		get_radius(), setmap();

static
setup()
{
	int	rc;

	fxnum = (float) xnum;
	fynum = (float) ynum;
	/*
	**	If tiepointing, use special setup routine
	*/
	if (nproj == 0) {
		X0 = Y0 = 0;
		(void)g_setproc(0, tiepoint);
		return setmap();
	}
	/*
	**	If unspecified, take coordinate origin as center of map
	*/
	if (centers)
		if (X0 < 1.e20 || Y0 < 1.e20)
			error("-c, and -X/-Y are mutually exclusive", "");
		else
			X0 = Y0 = 0;
	if (X0 >= 1.e20)
		X0 = gxpix/2;
	if (Y0 >= 1.e20)
		Y0 = gypix/2;
	/*
	**	If Euler angles defined, setup rotation matrix
	*/
	if (erot)
		(void)sscanf(erot, "%lg,%lg,%lg",
			&eangle[0], &eangle[1], &eangle[2]);
	/*
	**	If -r argument not specified, try to estimate it
	*/
	if (radius == 0 && (rc = get_radius()))
		return rc;
	radius *= radfact;
	/*
	**	If specified, adjust the center location
	*/
	if (centers) {
		double	clon, clat, cxy[2], dd[2];

		if (sscanf(centers, "%lf,%lf", &clon, &clat) != 2)
			error("bad center longitude or latitude: %s", centers);
		if (setmap())
			gipserr();
		if (g_map(clon, clat, cxy))
			error("illegal center lon,lat (%s)", centers);
		if (g_unmap((int)(cxy[0]), (int)(cxy[1]), dd))
			error("center lon,lat (%s) outside image", centers);
		X0 = gxpix/2 - cxy[0];
		Y0 = gypix/2 - cxy[1];
	}
	/*
	**	Debug printout
	*/
	if (Dflag) {
		(void) fprintf(stderr,
			"ihgrid: %s setup -x%d -y%d -X%.12g -Y%.12g -r %.12g",
			mapname, gxpix, gypix, X0, Y0, radius);
		if (radfact != 1)
			(void) fprintf(stderr, " -R %.12g", radfact);
		if (erot)
			(void) fprintf(stderr, " -l %.12g,%.12g,%.12g",
				eangle[0], eangle[1], eangle[2]);
		if (phi0 != 0)
			(void) fprintf(stderr, " -p %.12g", phi0);
		if (phi1 != 0)
			(void) fprintf(stderr, " -P %.12g", phi1);
		(void) fprintf(stderr, "\n");
		(void) fflush(stderr);
	}
	return setmap();
}
	/*
	**	Estimate radius parameter
	*/
static
get_radius()
{
	double	d, r, x0, y0, pix[2];
	double	xmin, xmax, ymin, ymax;
	register dx, dy, rc;

	if (proj0 && proj0->radius != 0) {
		radius = proj0->radius;
		return 0;
	}
	radius = hypot((double)gxpix, (double)gypix);
	if (rc = setmap())
		return rc;
	for (r = 0, dx = 0; dx != -3; dx = -dx-dx-1)
		for (dy = 0; dy != -3; dy = -dy-dy-1)
			if (g_map(eangle[0]+dx, eangle[1]+dy, pix) == 0)
				if (rc++ == 0) {
					x0 = pix[0];
					y0 = pix[1];
				} else {
					d = hypot(pix[0]-x0, pix[1]-y0);
					if (d > r) r = d;
				}
	if (ihget("xmin", (char *)&xmin, 1)) xmin = 0;
	if (ihget("xmax", (char *)&xmax, 1)) xmax = fxnum;
	if (ihget("ymin", (char *)&ymin, 1)) ymin = 0;
	if (ihget("ymax", (char *)&ymax, 1)) ymax = fynum;
	if (r == 0)
		error("unable to estimate -r value for %s mapping", mapname);
	radius *= radius/(r*hypot(fabs(xmin-xmax), fabs(ymin-ymax)));
	return 0;
}
	/*
	**	Add "mapping" variable to incore gips header
	*/
static
setmap()
{
	(void)sprintf(mapbuf, GIPS_MAPPING_P,
		mapname = g_getname(nproj), nproj, X0, Y0, radius,
		eangle[0], eangle[1], eangle[2], phi0, phi1, xyrot);
	if (ihput("mapping", mapbuf))
		gipserr();
	return g_setup();
}

main(argc, argv)
	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	char	dtype;		/* input data type */
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("bad arg: %s", argv[2]);
	/*
	 *	user wants a list of permitted projections
	 */
	if (argc == 2 && strcmp(ifile = *++argv, "?") == 0) {
		g_listproj(stdout);
		exit(0);
	}
	if (Sflag)
		telltale(0);
	if (Dflag)
		setbuf(stderr, alloc(BUFSIZ, 1));
	/*
	 *	read input header
	 */
	if ((fp = ihread(ifile, GIPSPROFILE)) == NULL)
		gipserr();
	/*
	 *	if the image is cellular, load necessary tables
	 */
	if (!ihget("dtype", &dtype, 1) && dtype == 'G')
		(void)gclose(gopen(fp, 1));
	if (fp == stdin)
		ifile = "'stdin'";
	else
		(void)(ihbcl(fp));
	if (*mfile != '-')
		if ((fp = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fp));
	/*
	 *	Verify output file
	 */
	if (*ofile == '-' && ofile[1] == '\0') {
		(void)(signal(SIGPIPE, exit));
		fp = stdout;
		ofile = "'stdout'";
	} else if ((fp = fopen(ofile, "w")) == NULL)
		gipserr2(GIPSBADOPEN, ofile);
	/*
	 *	get necessary header variable values and check them out
	 */
	if (ihget("xnum", (char *)&xnum, 0))
		gipserr();
	if (gxpix <= 0)
		gxpix = xnum;
	if (ihget("ynum", (char *)&ynum, 0))
		gipserr();
	if (gypix <= 0)
		gypix = ynum;
	/*
	**	check whether the input file is a non-local projection,
	**	if so, save a pointer to the projection parameter table.
	*/
	if (! g_setup()) {
		proj0 = g_proj;
		g_proj = NULL;
	}
	/*
	 *	Determine output buffer depth for ihgeom()
	 */
	if (nrast <= 0)
		nrast = (gypix > 32) ? gypix >> 5 : 32;
	/*
	 *	initialize parameters for this projection
	 */
	switch (setup()) {
	case 0:
		break;
	case 1:
		error ("%s mapping is singular for origin choice", mapname);
	default:
		error ("setup for %s mapping not implemented", mapname);
	}
	/*
	 *	Write output data header
	 */
	if (fprintf(fp, GIPS_GRID, gxpix, gypix, nrast) <= 0 ||
			fprintf(fp, "\n%s\n", mapbuf) <= 0)
		gipserr2(GIPSBADWRITE, ofile);
	/*
	 *	First call to interpolator
	 */
	firstbox(gxpix, gypix, nrast);
	/*
	 *	Tha-a-a-ts all folks!
	 */
	(void)(fclose(fp));
	if (Dflag && ncell)
		(void) fprintf(stderr, "ihgrid: %s cells=%d avg=(%g,%g)\n",
			mapname, ncell, (float)xcell/ncell, (float)ycell/ncell);
	if (Sflag)
		telltale(1);
}
	/*
	**	transform from output pixel value to input address
	*/
pix_unmap(x, y, c, n)
	register struct	cell	*c;
	register	n;
{
	double		lonlat[2], xy[2];
	register G_PARM	*proj1;
	register	ret;
	register float	*p = c->v[n].f;

	switch (g_unmap(x, y, lonlat)) {
	case 0:
		if (! nproj) {
			p[0] = lonlat[0];
			p[1] = lonlat[1];
		} else if (proj0) {
			proj1 = g_proj;
			g_proj = proj0;
			ret = g_map(lonlat[0], lonlat[1], xy);
			g_proj = proj1;
			p[0] = xy[0];
			p[1] = xy[1];
			if (ret)
				return c->flg[n] = VERTEX_OFFMAP;
		} else if (gips(lonlat, (int *)p, 1, GIPS_XY_FVAL))
			gipserr();
		if (p[0] < 0 || p[0] >= fxnum || p[1] < 0 || p[1] >= fynum)
			ret = VERTEX_OFFMAP;
		else
			ret = 0;
		break;
	case 1:
		ret = VERTEX_NONPHYS;
		break;
	case -1:
		error("inverse %s mapping not implemented", mapname);
	default:
		gipserr();
	}
	c->flg[n] = ret;
}
	/*
	 *	write a control grid cell
	 */
putcell(c)
	register struct cell *c;
{
	if ((fflag && fprintf(fp, "%d %d %d %d %g %g %g %g %g %g %g %g\n",
		c->xy[0], c->xy[1], c->xy[2], c->xy[3],
		c->v[0].f[0], c->v[0].f[1], c->v[1].f[0], c->v[1].f[1],
		c->v[2].f[0], c->v[2].f[1], c->v[3].f[0], c->v[3].f[1]) <= 0) ||
	   (!fflag &&
		fwrite((char *)c, 4*sizeof(int)+8*sizeof(float), 1, fp) != 1))
		gipserr2(GIPSBADWRITE, ofile);
	if (Dflag) {
		ncell++;
		xcell += (c->xy[2] - c->xy[0]);
		ycell += (c->xy[3] - c->xy[1]);
	}
}
