#ifndef lint
static char *sccsid = "%Z%$RCSfile: tiepoint.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/05/13 01:21:42 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		tiepoint.c	Creation Date:	09/05/84
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/05/13 01:21:42 $
**	Components:	tiepoint()	ihgrid() function
**
**	Function:	Transform an image via tiepoint mapping
*/

#include	"ihgrid.h"

extern	flag	Dflag;		/* debug flag */
extern	int	order;		/* polynomial order */
static	double	*rhsx, *rhsy;	/* tie point coefficients */
static		settie(), solve();

tiepoint(sw,x,y,xy)
	double		x, y, xy[2];
{
	double		gx, gy;
	register double	*fx = rhsx, *fy = rhsy;
	register int	ny, nx;

	switch (sw) {
	case 0:
		return settie();
	case 2:
		xy[0] = xy[1] = 0;
		for (ny = order, gy = 1; --ny >= 0; gy *= y)
			for (nx = order, gx = gy; --nx >= 0; gx *= x)
				xy[0] += gx * *fx++, xy[1] += gx * *fy++;
		return 0;
	default:
		break;
	}
	return -1;
}

static
settie()
{
	int	ntie = 0, m;
	double	x, y, xp, yp, *lhs, gx, gy;
	register i, nx, ny;
	register double *fx, *fy;

	if (order <= 0)
		error("bad -n value: %d",(char *)order);

	m = ++order;
	m *= m;
	lhs = (double *) alloc(m*m,sizeof(double));
	rhsx = (double *) alloc(2*m,sizeof(double));
	rhsy = rhsx + m;

	while ((i = scanf(" %lf %lf %lf %lf",&xp,&yp,&x,&y)) == 4) {
		i = ntie++ % m;
		if (i == 0) fx = lhs;
		rhsx[i] += xp;
		rhsy[i] += yp;
		for (ny = order, gy = 1; --ny >= 0; gy *= y)
			for (nx = order, gx = gy; --nx >= 0; gx *= x)
				*fx++ += gx;
	}
	if (i != EOF)
		error("illegal numeric in input stream: 0%o",
			(char *)(getchar()));
	if (Dflag) {
		(void) fprintf(stderr,
			"ihgrid: setup %d-order bilinear from %d tie points\n",
			order,ntie);
		for (i = 0, fx = lhs; i < m*m; ) {
			(void) fprintf(stderr,"\tlhs(%d)=%.12g",i,*fx++);
			if ((++i % 4) == 0 || i == m*m)
				(void) fprintf(stderr,"\n");
		}
		for (i = 0, fx = rhsx, fy = rhsy; i < m; ) {
			(void) fprintf(stderr,"\trhsx(%d)=%.12g\trhsy(%d)=%.12g",
				i,*fx++,i,*fy++);
			if ((++i % 2) == 0 || i == m)
				(void) fprintf(stderr,"\n");
		}
	}
	if (solve(lhs,m,rhsx,rhsy))
		error("singular tie-point transformation","");
		
	free((char *)lhs);
	if (Dflag) {
		(void) fprintf(stderr,"ihgrid: interpolation coefficients:\n");
		for (i = 0, fx = rhsx, fy = rhsy; i < m; i++)
			(void) fprintf(stderr,"\tx(%d)=%.12g\ty(%d)=%.12g\n",
				i,*fx++,i,*fy++);
	}
	return 0;
}
	/*
	 *	Solution of a.x = b and a.y = c for square non-symmetric
	 *	matrix 'a' of dimension 'n', leaving 'x' in 'b', 'y' in 'c'
	 */
static
solve(a,n,b,c)
	register double	*a, *b, *c;
	register int	n;
{
	register	i, j, k;
	int		l;
	double		t, u, v;

	for (k = 0; k < n-1; k++) {
		for (t = 0, i = k+1; i < n; i++)
			if (fabs(a[n*i+k]) > t)
				t = fabs(a[n*i+k]), l = i;
		if (t == 0)
			return 1;
		u = b[l];
		v = c[l];
		if (l != k) {
			t = a[l*n+k];
			a[l*n+k] = a[(n+1)*k];
			a[(n+1)*k] = t;
			b[l] = b[k];
			b[k] = u;
			c[l] = c[k];
			c[k] = v;
		}
		for (t = -1/a[(n+1)*k], i = k+1; i < n; i++)
			a[i*n+k] *= t;
		for (j = k+1; j < n; j++) {
			t = a[l*n+j];
			if (l != k) {
				a[l*n+j] = a[k*n+j];
				a[k*n+j] = t;
			}
			for (i = k+1; i < n; i++)
				a[i*n+j] += t*a[i*n+k];
			b[j] += u*a[j*n+k];
			c[j] += v*a[j*n+k];
		}
	}
	for (k = n; --k >= 0; ) {
		if ((t = a[k*(n+1)]) == 0)
			return(1);
		u = -(b[k] /= t);
		v = -(c[k] /= t);
		for (i = 0; i < k; i++) {
			b[i] += u*a[i*n+k];
			c[i] += v*a[i*n+k];
		}
	}
	return 0;
}
