#ifndef lint
static char *sccsid = "%Z%$RCSfile: get_rast.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/05/13 01:22:43 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		get_rast.c	Creation Date:	11/06/84
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/05/13 01:22:43 $
**	Components:	get_rast()	ihm subroutine
**			get_comp()	compare 'map' elements
**			get_rsort()	sort by ascending realworld values
**			get_debug()	output of map and raster structures
**
**	Function:	Compare the input and output maps
**			and create the raster structure
*/

#include	"ihm.h"

#define		max(x,y)	((x)>(y)?(x):(y))
#define		min(x,y)	((x)<(y)?(x):(y))

static		get_rsort();

get_rast(t,xy)

	register TABLE	* t;		/* input file table */
	int		  xy;		/* =0,1 for x,y raster */
{
	register MAP 	* tm = t->map[xy];	/* input map */
	register	  tn = t->num[xy];	/* input pixels */
	register MAP 	* om = out->map[xy];	/* output map */
	register	  on = out->num[xy];	/* output pixels */

	RASTER		* tr0;		/* base of new raster table */
	RASTER		* tr;		/* pointer to new rast table */
	int		  tns = tn;	/* saved tn value */
	static	int	  init[2];	/* flag for first call */
	register	sw;		/* scratch */
	/*
	 *	on first call, sort output raster pointers
	 */
	if (! init[xy]++) {
		get_debug(out,xy,out->num[xy],"unsorted output MAP");
		out->spl[xy] = get_rsort(out,xy);
		get_debug(out,xy,out->num[xy],"sorted output MAP");
	}
	get_debug(t,xy,t->num[xy],"unsorted input MAP");
	/*
	 *	now sort the current raster structure
	 */
	t->spl[xy] = get_rsort(t,xy);
	get_debug(t,xy,t->num[xy],"sorted input MAP");
	/*
	 *	allocate a raster table
	 */
	t->rast[xy] = tr0 = (RASTER *) alloc(on+1,sizeof(struct raster));
	if (t->spl[xy]== 0 || out->spl[xy] == 0 || t->spl[xy] != out->spl[xy]) {
		tr0--;
		t->spl[xy] = 0;
	}
	/*
	 *	use sw as a flag for when to update t->inmap
	 */
	sw = (! qflag && xy == 0);
	/*
	 *	fill raster table elements
	 */
	for ( ; --on >= 0; om++) {
		tr = tr0 + om->mptr;
		if ((om+1)->mwt <= tm->mwt)
			continue;
		for ( ; (tm+1)->mwt <= om->mwt && --tn >= 0; tm++)
			tm->mwt = (tm+1)->mwt - tm->mwt;
		tr->lowlimit = tm;
		tr->lowweight = (tm+1)->mwt - om->mwt;
		if (sw) t->inmap[tm->mptr] = 1;
		for ( ; (om+1)->mwt > (tm+1)->mwt && --tn >= 0; tm++) {
			tm->mwt = (tm+1)->mwt - tm->mwt;
			if (sw) t->inmap[tm->mptr] = 1;
			else if (qflag && tm->mwt > tr->lowweight) {
				tm++;
				tr->lowlimit++;
				break;
			}
		}
		if (tr->lowlimit->mptr >= tns)
			tr->lowlimit = NULL;
		else if (! qflag && tm != tr->lowlimit) {
			tr->highlimit = tm;
			tr->highweight = (om+1)->mwt - tm->mwt;
			if (sw) t->inmap[tm->mptr] = 1;
		}
	}
	for ( ; --tn >= 0; tm++)
		tm->mwt = (tm+1)->mwt - tm->mwt;

	get_debug(t,xy,out->num[xy],"completed RASTER");
}

static
get_comp(s,t)

	register MAP	*s;		/* first map element index */
	register MAP	*t;		/* second map element index */
{
	return((s->mwt > t->mwt || (s->mwt == t->mwt && s > t)) ? 1 : -1);
}

static
get_rsort(t,xy)

	register TABLE	* t;		/* input file table */
	register int	xy;		/* =0,1 for x,y raster */
{
	register	sw;		/* sense of axis */
	register MAP	*m;		/* raster map pointer */
	register	i = t->num[xy]+1;	/* scratch */

	qsort((char *)(m = t->map[xy]),i,sizeof(struct map),get_comp);
	sw = ((m+1)->mptr > m->mptr) ? 1 : -1;
	for ( ; --i > 0; m++)
		if ((sw == 1 && (m+1)->mptr < m->mptr) ||
			      (sw == -1 && (m+1)->mptr > m->mptr))
			return(0);
	return(sw);
}

get_debug(t,xy,n,str)

	TABLE		*t;		/* input or output table */
	int		xy;		/* =0,1 for x or y raster */
	register	n;		/* number of points */
	char		*str;		/* output comment */
{
	register MAP	*tm = t->map[xy];	/* map array */
	register MAP	*m;			/* map pointer */
	register MAP	*s;			/* map pointer */
	register RASTER	*r = t->rast[xy];	/* raster pointer */

	if (! Dflag) return;

	(void) fprintf(stderr,
		"dump %c raster of \"%s\" (%s):\n", "xy"[xy],
			t->fname ? t->fname : "unnamed", str);
	(void) fprintf(stderr,
		"nmap\tptr\tweight\tnlow\tptr\tweight\tnhigh\tptr\tweight\n");

	for (m = tm; n-- >= 0; m++, fflush(stderr)) {
		(void) fprintf(stderr,"%d\t%d\t%.12g",m-tm,m->mptr,m->mwt);
		if (r == NULL) {
			(void) fprintf(stderr,"\tno raster\n");
			continue;
		}
		if (s = r->lowlimit)
			(void) fprintf(stderr,"\t%d\t%d\t%.12g",
				s-tm,s->mptr,r->lowweight);
		else
			(void) fprintf(stderr,"\tno low");
		if (s = r->highlimit)
			(void) fprintf(stderr,"\t%d\t%d\t%.12g\n",
				s-tm,s->mptr,r->highweight);
		else
			(void) fprintf(stderr,"\tno high\n");
		r++;
	}
}
