#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihmed.c,v $	GIPS $Revision: 3.5 $ (MIT/CSR) $Date: 1998/05/13 01:01:52 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990 Massachusetts Institute of Technology
**
**	Module:		ihmed.c		Creation Date:	06/04/83
**	Version:	$Revision: 3.5 $		Update:		$Date: 1998/05/13 01:01:52 $
**	Function:	Apply median filter to a General Image.
**	Components:	main()		GIPS/UNIX command
*/

#include	<stdio.h>
#include	<signal.h>
#include	<strings.h>
#include	<math.h>
#include	"gips.h"
	/*
	 *	flag fields filled by gipsflags()
	 */
static	char	*ifile	= "-";		/* input file */
static	char	*ofile	= "-";		/* output file */
static	char	*mfile	= "-";		/* secondary input header */
static	int	xarr	= 3;		/* x-size of average */
static	int	yarr	= 3;		/* y-size of average */
static	flag	Sflag	;		/* generate memory statistics */
static	flag	Nflag	;		/* only replace null pixels */
static	flag	Gflag	;		/* use for clearing gores */
static	flag	Xflag	;		/* don't replace null pixels */
	/*
	 *	table used by gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"G",	"1",		&Gflag,
	"S",	"1",		&Sflag,
	"X",	"1",		&Xflag,
	"n",	"1",		&Nflag,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"x",	"%d xdim",	(flag *) &xarr,
	"y",	"%d ydim",	(flag *) &yarr,
	NULL,	"[file]",	NULL
};
	/*
	 *	variables initialized from GIPS header
	 */
static	int	xnum	;	/* input x-direction pixel length */
static	int	ynum	;	/* number of input rasters */
static	int	lpix	;	/* length of GIPS pixel in bytes */
static	char	dtype	;	/* pixel type code */
static	double	dnull	;	/* null pixel value */
static	long	inull	;	/* null integer pixel value */
static	int	dflag	;	/* set if dnull not defined */
static	int	xarr2;		/* xarr/2 */
static	int	yarr2;		/* yarr/2 */

static		copy(), iproc(), dproc();

main(argc,argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fp;		/* input GIPS file */
	FILE	*fo;		/* output GIPS file */
	int	lrast;		/* length of raster element in byte */
	int	cvtinout;	/* input/output conversion code */
	int	cvtswap;	/* swapbuf element conversion code */
	char	*inout;		/* file input buffer */
	char	*swapbuf;	/* filter output buffer */
	char	**buf;		/* raster ring buffer pointers */
	char	*nullbuf;	/* buffer of null values */
	char	*medbuf;	/* median sort buffer */
	char	xscale;		/* type of x-axis coordinate */
	char	yscale;		/* type of y-axis coordinate */
	char	dscale;		/* type of dn coordinate */
	long	raster;		/* byte length of raster */
	long	border;		/* byte length of border */
	double	xwrap;		/* xscale=w wrap factor */
	double	xmin, xmax;	/* header variables */

	register int i;		/* scratch */
	/*
	 *	save and parse callers argument list
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input files: %s",argv[2]);
	if (argc == 2)
		ifile = *++argv;
	if (xarr <= 0)
		error("non-positive -x value: %d", (char *)xarr);
	xarr2 = xarr/2;
	if (yarr <= 0)
		error("non-positive -y value: %d", (char *)yarr);
	yarr2 = yarr/2;
	if (Sflag)
		telltale(0);
	/*
	 *	read the input GIPS header, if necessary merging "mfile"
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	if (*mfile != '-')
		if ((fo = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fo));
	/*
	 *	get all required header fields
	 */
	if ((dflag = ihget("dnull",(char *) &dnull,0)) && Gflag)
		error("%s: no dnull value for -G flag", ifile);
	(void)(ihget("dscale",&dscale,1));
	if (dscale == 'f')
		gipserr1(GIPSBADVAL);
	if (ihget("yscale",&yscale,1))
		yscale = 's';
	else if (yscale == 'w')
		gipserr1(GIPSUNSCALE);
	if (ihget("xnum",(char *) &xnum,0))
		gipserr();
	if (xnum < 1 )
		gipserr1(GIPSBADVAL);
	if (ihget("xscale",&xscale,1))
		xscale = 's';
	/*
	**	Determine whether the margins are really wrapped
	*/
	if (xscale == 'w' && ihget("xwrap", (char *)&xwrap, 0) == 0) {
		if (ihget("xmin", (char *)&xmin, 0))
			xmin = 0;
		if (ihget("xmax", (char *)&xmax, 0))
			xmax = xnum;
		if (fabs(xwrap) != fabs(xmax-xmin))
			xscale = 's';
	}
	if (ihget("ynum",(char *) &ynum,0))
		gipserr();
	if (ynum < 1 )
		gipserr1(GIPSBADVAL);
	if (ihget("dtype",&dtype,1))
		gipserr();
	if ((lpix = ihdtype(dtype)) == 0)
		gipserr1(GIPSBADVAL);
	/*
	 *	write the GIPS header
	 */
	putargs();
	if ((fo = ihwrit(ofile,0)) == NULL)
		gipserr();
	if (fo == stdout) {
		(void)(signal(SIGPIPE,exit));
		ofile = "'stdout'";
	}
	cvtinout = ihdcvt(dtype);
	/*
	 *	allocate the various buffers...
	 *	inout:	input and output buffer
	 *	buf:	ring of pointers to individual raster buffers
	 *	swapbuf: buffer to hold filtered raster result
	 *	nullbuf: buffer of dnull values
	 */
	inout = alloc(xnum, lpix);
	buf = (char **) alloc(yarr, sizeof(char *));
	if (dtype == 'd' || dtype == 'f') {
		cvtswap = CVT_D;
		lrast = sizeof(double);
		medbuf = alloc(xarr*yarr,sizeof(double *));
	} else {
		cvtswap = CVT_I;
		inull = (long) dnull;
		lrast = sizeof(long);
		medbuf = alloc(xarr*yarr,sizeof(long));
	}
	swapbuf = (cvtswap == cvtinout) ? inout : alloc(xnum,lrast);
	if (gipserrno == GIPSNOCORE)
		gipserr();
	raster	= lrast*xnum;
	border	= lrast*xarr2;
	/*
	 *	fill the dnull buffer
	 */
	nullbuf = alloc(raster+border+border, 1);
	if (cvtswap == CVT_I) {
		*(long *)nullbuf = inull;
		i = sizeof(char);
	} else {
		*(double *)nullbuf = dnull;
		i = sizeof(double);
	}
	copy(nullbuf, nullbuf+i, raster+border+border-i);
	/*
	 *	loop over input rasters
	 */
	for (i = 0;; i++) {
		register char *save;	/* pointer to most recent raster */
		/*
		 *	begin by allocating raster buffers
		 */
		if (i < yarr) {
			if ((save = buf[i] = alloc(xnum+xarr-1, lrast)) == NULL)
				gipserr();
		} else	{
			register int j;
			/*
			 *	if all raster buffers are allocated
			 *	(and hence filled) they can be convolved
			 *	to "out" and written out
			 */
			if (dtype == 'd' || dtype == 'f')
				dproc(i-yarr, (double **)(buf+yarr2),
					(double *)swapbuf,(double **)medbuf);
			else
				iproc(i-yarr, (long **)(buf+yarr2),
					(long *)swapbuf,(long *)medbuf);
			if (inout != swapbuf)
				gconv(inout,swapbuf,xnum,cvtinout,cvtswap);
			if (fwrite(inout,lpix,xnum,fo) != xnum)
				gipserr2(GIPSBADWRITE,ofile);
			/*
			 *	on all but last raster, rotate the ring buffers
			 */
			if (i == yarr+ynum-1)
				break;
			save = buf[0];
			for (j = 1; j < yarr; buf[j-1] = buf[j], j++)
				;
			buf[yarr-1] = save;
		}
		/*
		 *	fill the input buffer
		 */
		if (i < yarr2 || i >= ynum+yarr2) {
			copy(nullbuf, save, raster+border+border);
			continue;
		}
		if (fread(inout,lpix,xnum,fp) != xnum)
			gipserr2(GIPSBADREAD,ifile);
		gconv(save+border,inout,xnum,cvtswap,cvtinout);
		/*
		 *	handle xscale=w wrap-around
		 */
		if (xarr2 == 0)
			;
		else if (xscale == 'w') {
			copy(save+raster,save,border);
			copy(save+border,save+raster+border,border);
		} else if (yarr2 != 0) {
			copy(nullbuf,save,border);
			copy(nullbuf,save+border+raster,border);
		}
	}
	(void)(ihbcl(fp));
	(void)(ihbcl(fo));
	if (Sflag)
		telltale(1);
	exit(0);
}
/*
 *	copy "a" to "b" byte-by-byte (i.e. extension if a and b overlap)
 */
static
copy(a,b,n)

	register char *a;		/* 'from' field */
	register char *b;		/* 'to' field */
	register long n;		/* number of bytes to copy */
{
	while (--n >= 0)
		*b++ = *a++;
}
/*
 *	median-filter integer rasters
 */
static
iproc(y,ibuf,out,med)

	register int	y;		/* image y-coordinate */
	register long	**ibuf;		/* raster buffer ring pointer */
	register long	*out;		/* output buffer */
	register long	*med;		/* sort buffer */
{
	register long	*am;		/* pointer to median table */
	register long	*b;		/* pointer to input pixel */
	register long	*c;		/* scratch */
	register int	ix;		/* current pixel x-coordinate */
	register int	ax;		/* index over boxcar */
	register int	ay;		/* index over boxcar */
	register long	d1, d2;		/* scratch */

	for (ix = 0; ix < xnum; ix++, out++) {
		*out = d1 = ibuf[0][ix+xarr2];
		if ((Gflag || Xflag) && d1 == inull)
			continue;
		if (Nflag && d1 != inull)
			continue;
		for (am = med, ay = -yarr2; ay <= yarr2; ay++)
			for (ax = 0, b = ibuf[ay]+ix; ax < xarr; ax++, b++)
				if (Gflag) {
					if (*b == inull
						&& (ix+ax-xarr2) >= 0
						&& (ix+ax-xarr2) < xnum
						&& (y+ay) >= 0
						&& (y+ay) < ynum) {
						*out = inull;
						ay = yarr;
						break;
					}
				} else if (dflag || *b != inull)
					*am++ = *b;
		if (Gflag)
			continue;
		for (ax = am-med, ay = ax/2; ay > 0; ay /= 2)
			for (b = med+ay; (c = b) < am; b++)
				while ((c -= ay) >= med &&
						(d1 = *c) > (d2 = *(c+ay)))
					*c = d2, *(c+ay) = d1;
		*out = ax ? med[ax/2] : inull;
	}
}
/*
 *	median-filter floating point rasters
 */
static
dproc(y,dbuf,out,med)

	register int	y;		/* image y-coordinate */
	register double	**dbuf;		/* raster buffer ring pointer */
	register double	*out;		/* output buffer */
	register double	**med;		/* median sort buffer */
{
	register double	**am;		/* pointer to median table */
	register double *b;		/* pointer to input pixel */
	register double **db;		/* scratch */
	register double **c;		/* scratch */
	register double	*d;		/* scratch */
	register int	ix;		/* current pixel x-coordinate */
	register int	ax;		/* index over boxcar */
	register int	ay;		/* index over boxcar */

	for (ix = 0; ix < xnum; ix++, out++) {
		*out = dbuf[0][ix+xarr2];
		if ((Gflag || Xflag) && *out == dnull)
			continue;
		if (Nflag && *out != dnull)
			continue;
		for (am = med, ay = -yarr2; ay <= yarr2; ay++)
			for (ax = 0, b = dbuf[ay]+ix; ax < xarr; ax++, b++)
				if (Gflag) {
					if (*b == dnull
						&& (ix+ax-xarr2) >= 0
						&& (ix+ax-xarr2) < xnum
						&& (y+ay) >= 0
						&& (y+ay) < ynum) {
						*out = dnull;
						ay = yarr;
						break;
					}
				} else if (dflag || *b != dnull)
					*am++ = b;
		if (Gflag)
			continue;
		for (ax = am-med, ay = ax/2; ay > 0; ay /= 2)
			for (db = med+ay; (c = db) < am; db++)
				while ((c -= ay) >= med &&
						*(b = *c) > *(d = *(c+ay)))
					*c = d, *(c+ay) = b;
		*out = ax ? *med[ax/2] : dnull;
	}
}
