#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihnull.c,v $	GIPS $Revision: 2.9 $ (MIT/CSR) $Date: 1998/11/13 23:53:34 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihnull.c	Creation Date:	12/29/83
**	Version:	$Revision: 2.9 $		Update:		$Date: 1998/11/13 23:53:34 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Update a GIPS header and copy the
**			binary image without change
*/

#include	<stdio.h>
#include	<signal.h>
#define	GRID
#include	"gips.h"

	/*
	 *	variables filled in by gipsflags()
	 */
static	char	*ifile	= "-";		/* input image array */
static	char	*ofile	= "-";		/* output general image */
static	char	*afile	= "-";		/* second output general image */
static	char	*mfile	= "-";		/* secondary input header */
static	int	nbuffs	= 64;		/* number of gcell() buffers */
static	int	factor	= 1;		/* shrink factor */
static	flag	Sflag;			/* generate memory statistics */
	/*
	 *	argument definition block for gipsflags()
	 */
static	flagentry	flagtable[]	= {
	"S",	"1",		(flag *) &Sflag,
	"",	"%d buffers",	(flag *) &nbuffs,
	"a",	"-> file",	(flag *) &afile,
	"m",	"-> file",	(flag *) &mfile,
	"n",	"%d factor",	(flag *) &factor,
	"o",	"-> file",	(flag *) &ofile,
	NULL,	"[file]",	NULL
};
static		convert_mapping();
extern	int	_gcount;		/* gcell() extract count */

main(argc,argv)

	int	argc;			/* argument count */
	char	**argv;			/* array of argument pointers */
{
	int	xnum, ynum;		/* input image size */
	char	dtype;			/* pixel type */
	register gimage	*gi	= NULL;	/* cellular image handle */
	register FILE	*fp;		/* input image stream pointer */
	register FILE	*fq;		/* output image stream pointer */
	register FILE	*fa	= NULL;	/* 2nd output image stream pointer */
	register char	*inbuf, *in;	/* input buffer */
	register char	*outbuf, *out;	/* output buffer */
	register int	xlim, ylim;	/* output image size */
	register int	fact;		/* reduction factor */
	register int	dlen;		/* data length */
	register int	i, l, x, y;	/* scratch */
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	if (argc > 1)
		ifile = *++argv;
	if (Sflag)
		telltale(0);
	/*
	 *	read GIPS header
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	putargs();
	/*
	 *	get image dimensions and pixel size
	 */
	if (ihget("xnum", &xnum) || ihget("ynum", &ynum))
		gipserr();
	if (ihget("dtype", &dtype))
		gipserr();
	/*
	 *	if cellular image, fix header
	 */
	if (dtype == 'G') {
		char	geom[256];

		if (ihget("geometry", geom, sizeof(geom)-1))
			gipserr();
		if (sscanf(geom, "type=%c", &dtype) != 1)
			gipserr();
		if (dtype == 'M') {
			dtype = 'f';
			ihput("dscale", "s");
		}
		if (nbuffs <= 0)
			nbuffs = 64;
		if ((gi = gopen(fp, nbuffs)) == NULL)
			gipserr();
		ihput("dtype", &dtype);
		ihdel("geometry");
		xlim = (xnum + factor - 1) / factor;
		ylim = (ynum + factor - 1) / factor;
	} else {
		xlim = xnum;
		ylim = ynum;
		inbuf = alloc(ihdtype(dtype), xlim);
		if (inbuf == NULL)
			gipserr();
	}
	if ((dlen = ihdtype(dtype)) == 0)
		gipserr();
	/*
	 *	handle shrinkage factor
	 */
	if ((fact = factor) <= 0)
		error("bad -n value: %d", (char *)factor);
	else if (fact != 1) {
		xnum = (xnum + fact - 1) / fact;
		ihput("xnum", &xnum);
		ynum = (ynum + fact - 1) / fact;
		ihput("ynum", &ynum);
		convert_mapping();
	}
	/*
	 *	Merge auxiliary header
	 */
	if (*mfile != '-')
		if ((fq = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fq));
	/*
	 *	write output header
	 */
	out = outbuf = alloc(dlen, xnum);
	if (outbuf == NULL || (fq = ihwrit(ofile,1)) == NULL)
		gipserr();
	else if (fq == stdout)
		(void)(signal(SIGPIPE,exit)), ofile = "'stdout'";
	if (*afile != '-' && (fa = ihwrit(afile,1)) == NULL)
		gipserr();
	/*
	 *	copy one raster at a time
	 */
	for (y = 0; y < ylim; y++) {
		if (gi) {
			/*
			**	cellular image
			*/
			for (x = 0; x < xlim; ) {
				if (! (in = gcell(x * fact, y * fact, gi)))
					error("%s: bad cell format", ifile);
				for ( ; _gcount > 0 && x < xlim; x++) {
					for (i = 0; i < dlen; i++)
						*out++ = *in++;
					in += (fact-1) * dlen;
					_gcount -= fact;
				}
			}
		} else {
			/*
			**	image array
			*/
			if (fread(in = inbuf, dlen, xlim, fp) != xlim)
				gipserr2(GIPSBADREAD, ifile);
			if (y % fact)
				continue;
			for (x = 0; x < xlim; x++)
				for (l = dlen; --l >= 0; in++)
					if ((x % fact) == 0)
						*out++ = *in;
		}
		/*
		**	write output raster
		*/
		if (fwrite(out = outbuf, dlen, xnum, fq) != xnum)
			gipserr2(GIPSBADWRITE, ofile);
		if (fa && fwrite(outbuf, dlen, xnum, fa) != xnum)
			gipserr2(GIPSBADWRITE, afile);
	}
	/*
	 *	cleanup
	 */
	if (gi)
		(void) gclose(gi);
	else
		(void) ihbcl(fp);
	(void) ihbcl(fq);
	if (fa)
		(void) ihbcl(fa);
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}

	/*
	**	Convert the mapping parameter
	*/

static
convert_mapping()
{
	char	map[1024], name[64];
	int	type, xyrot;
	double	xorg, yorg, radius, e1, e2, e3, lat1, lat2;

	if (ihget("mapping", map, sizeof(map)-1))
		return;
	xyrot = 0;
	if (sscanf(map, GIPS_MAPPING, name, &type, &xorg, &yorg, &radius,
		&e1, &e2, &e3, &lat1, &lat2, &xyrot) < 10)
		return;
	radius /= factor;
	xorg /= factor;
	yorg /= factor;
	(void)sprintf(map, GIPS_MAPPING_P, name, type, xorg, yorg,
		radius, e1, e2, e3, lat1, lat2, xyrot);
	ihput("mapping", map);
}
