#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihpds.c,v $	GIPS $Revision: 3.30 $ (MIT/CSR) $Date: 1999/06/11 20:58:46 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990 Massachusetts Institute of Technology
**
**	Module:		ihpds.c		Creation Date:	06/04/90
**	Version:	$Revision: 3.30 $		Update:		$Date: 1999/06/11 20:58:46 $
**	Function:	Convert PDS label + image to GIPS image
**	Components:	main()		GIPS/UNIX command
*/

#include <stdlib.h>  /* for malloc */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <strings.h>
#define	GRID
#include "gips.h"

	/*
	**	Conversion of VAX halfword RDWs
	*/

#ifdef vax
#define	getshort(s) *(short *)(s)
#define	ORDER	1
#else
#define getshort(s) \
	(((((unsigned char *)(s))[1])<<8)|(((unsigned char *)(s))[0]))
#define	ORDER	0
#endif vax

	/*
	**	Command flags and options
	*/

static	char	Dflag;			/* debug flag */
static	char	Lflag;			/* ihstokes output in dB */
static	char	Sflag;			/* print execution statistics */
static	char	Vflag;			/* use VICAR header info */
static	char	Wflag;			/* force west longitudes */
static	char	bflag;			/* byte-swap regardless! */
static	char	*conv;			/* ihstokes parameters */
static	char	hflag;			/* print PDS header only */
static	char	lflag;			/* print image statistics only */
static	char	*plane;			/* ihvicar plane */
static	char	vflag;			/* print image name & size */
static	char	*mfile;			/* header merge file */
static	char	*ofile	= "-";		/* output file */

static	flagentry	flagtable[] = {
	"D",	"1",		(char *)&Dflag,
	"L",	"1",		(char *)&Lflag,
	"S",	"1",		(char *)&Sflag,
	"V",	"1",		(char *)&Vflag,
	"W",	"1",		(char *)&Wflag,
	"b",	"1",		(char *)&bflag,
	"c",	"-> conv",	(char *)&conv,
	"h",	"1",		(char *)&hflag,
	"l",	"1",		(char *)&lflag,
	"m",	"-> file",	(char *)&mfile,
	"o",	"-> file",	(char *)&ofile,
	"p",	"-> plane",	(char *)&plane,
	"v",	"1",		(char *)&vflag,
	NULL,	"[file]",	NULL
};

	/*
	**	Items parsed from PDS label
	*/

static	int	iloc		= 1;	/* image record number */
static	int	nline		= 0;	/* number of image lines */
static	int	nsamp		= 0;	/* number of image samples */
static	int	nhist		= 0;	/* number of histogram values */
static	int	bitsperpix	= 8;	/* number of bits per image pixel */
static	int	reclen		= 0;	/* number of bytes per record */
static	int	lineheadbytes	= 0;	/* bytes before each image line */
static	int	linetailbytes	= 0;	/* bytes after each image line */
static	int	hdrlen		= 0;	/* length of detached header */
static	int	lblrec		= 0;	/* number of label records */
static	int	hdrrec		= 0;	/* number of header records */
static	int	linelen		= 0;	/* image line length */
static	int	lsb		= 0;	/* LSB/HSB flag */
static	int	pdsv		= 0;	/* PDS level */
static	int	encoderec;		/* encoding table pointer */

static	char	object[256];		/* current label object */
static	char	imagetype[256];		/* image format */
static	char	imageid[256];		/* image identifier */
static	char	recfm[256];		/* image record format */
static	char	bytetype[256];		/* image pixel type */
static	char	encoding[256];		/* byte encoding type */
static	char	istruct[256];		/* pointer to image name or record */
static	char	old_istruct[256];	/* pointer to image name or record */
static	char	hstruct[256];		/* pointer to header name or record */
static	char	projtype[256];		/* image projection type */
static	char	dsid[256];		/* data set id */

static	double	xmin	= GIPSHIGH;	/* minimum longitude */
static	double	xmax	= GIPSHIGH;	/* maximum longitude */
static	double	ymin	= GIPSHIGH;	/* minimum latitude */
static	double	ymax	= GIPSHIGH;	/* maximum latitude */
static	double	dzero	= GIPSHIGH;	/* pixel mapping origin */
static	double	dfact	= GIPSHIGH;	/* pixel mapping increment */
static	double	dnull	= GIPSHIGH;	/* null pixel value */

	/*
	**	Information from IMAGE_MAP_PROJECTION_CATALOG
	*/

static	char	map_projtype[256];	/* map catalog projection type */
static	char	map_scale[256];		/* map catalog scale + unit */
static	char	map_resolution[256];	/* map catalog resoln + unit */
static	char	map_radius[256];	/* planetary radius + unit */
static	char	map_londir[256];	/* positive longitude direction */
static	int	map_xframe	= 1;	/* X-framelet offset */
static	int	map_yframe	= 1;	/* Y-framelet offset */
static	double	map_xoff	= 0;	/* X-axis projection offset */
static	double	map_yoff	= 0;	/* Y-axis projection offset */
static	double	map_phi0;		/* first standard parallel */
static	double	map_phi1;		/* second standard parallel */
static	double	map_clat;		/* center latitude */
static	double	map_clon;		/* center longitude */
static	double	map_azi;		/* projection azimuth */
static	double	map_rot;		/* Cartesian azimuth */

	/*
	**	Definition of useful PDS label keywords
	*/

static	struct	{
	char	*name;			/* keyword name */
	char	*fmt;			/* sprintf() format */
	char	*value;			/* pointer to value */
} *k, keys[] = {
	".PDS_VERSION_ID",		"%s",	(char *) NULL,
	".RECORD_TYPE",			"%s",	(char *) recfm,
	".RECORD_BYTES",		"%d",	(char *) &reclen,
	".DATA_SET_ID",			"%s",	(char *) dsid,
	".MAP_PROJECTION_TYPE",		"%s",	(char *) projtype,
	".MAXIMUM_LATITUDE",		"%lf",	(char *) &ymin,
	".MINIMUM_LATITUDE",		"%lf",	(char *) &ymax,
	".MAXIMUM_LONGITUDE",		"%lf",	(char *) &xmax,
	".MINIMUM_LONGITUDE",		"%lf",	(char *) &xmin,
	".^IMAGE_HEADER",		"%s",	(char *) hstruct,
	".^IMAGE",			"%s",	(char *) istruct,
	".IMAGE_POINTER",		"%s",	(char *) istruct,
	".IMAGE_LINES",			"%d",	(char *) &nline,
	".IMAGE_RECORD_BYTES",		"%d",	(char *) &linelen,
	".IMAGE_ID",			"%s", 	(char *) imageid,
	".LINE_SAMPLES",		"%d",	(char *) &nsamp,
	".HEADER_RECORD_BYTES",		"%d",	(char *) &hdrlen,
	".HEADER_RECORDS",		"%d",	(char *) &hdrrec,
	".LABEL_RECORDS",		"%d",	(char *) &lblrec,

	".IMAGE.^STRUCTURE",		"%s",	(char *) old_istruct,
	".HEADER.TYPE",			"%s",	(char *) imagetype,
	".IMAGE_HEADER.TYPE",		"%s",	(char *) imagetype,
	".FOREIGN_LABEL.TYPE",		"%s",	(char *) imagetype,
	".IMAGE.LINES",			"%d",	(char *) &nline,
	".IMAGE.LINE_SAMPLES",		"%d",	(char *) &nsamp,
	".IMAGE.SAMPLE_BITS",		"%d",	(char *) &bitsperpix,
	".IMAGE.SAMPLE_TYPE",		"%s",	(char *) bytetype,
	".IMAGE.LINE_PREFIX_BYTES",	"%d",	(char *) &lineheadbytes,
	".IMAGE.LINE_SUFFIX_BYTES",	"%d",	(char *) &linetailbytes,
	".IMAGE.ENCODING",		"%s",	(char *) encoding,
	".IMAGE.ENCODING_TYPE",		"%s",	(char *) encoding,
	".^ENCODING_HISTOGRAM",		"%d",	(char *) &encoderec,
	".ENCODING_HISTOGRAM.ITEMS",	"%d",	(char *) &nhist,
	".IMAGE.SCALING_FACTOR",	"%lf",	(char *) &dfact,
	".IMAGE.OFFSET",		"%lf",	(char *) &dzero,
	".IMAGE.MISSING",		"%lf",	(char *) &dnull,
	".IMAGE_MAP_PROJECTION.MAP_PROJECTION_TYPE",
			       	"%[^\n]",	(char *) map_projtype,
	".IMAGE_MAP_PROJECTION.MAP_SCALE",
				"%[^\n]",	(char *) map_scale,
	".IMAGE_MAP_PROJECTION.MAP_RESOLUTION",
				"%[^\n]",	(char *) map_resolution,
	".IMAGE_MAP_PROJECTION.LINE_PROJECTION_OFFSET",
					"%lf",	(char *) &map_xoff,
	".IMAGE_MAP_PROJECTION.X_AXIS_PROJECTION_OFFSET",
					"%lf",	(char *) &map_xoff,
	".IMAGE_MAP_PROJECTION.SAMPLE_PROJECTION_OFFSET",
					"%lf",	(char *) &map_yoff,
	".IMAGE_MAP_PROJECTION.Y_AXIS_PROJECTION_OFFSET",
					"%lf",	(char *) &map_yoff,
	".IMAGE_MAP_PROJECTION.A_AXIS_RADIUS",
				"%[^\n]",	(char *) map_radius,
	".IMAGE_MAP_PROJECTION.FIRST_STANDARD_PARALLEL",
					"%lf",	(char *) &map_phi0,
	".IMAGE_MAP_PROJECTION.SECOND_STANDARD_PARALLEL",
					"%lf",	(char *) &map_phi1,
	".IMAGE_MAP_PROJECTION.CENTER_LATITUDE",
					"%lf",	(char *) &map_clat,
	".IMAGE_MAP_PROJECTION.CENTER_LONGITUDE",
					"%lf",	(char *) &map_clon,
	".IMAGE_MAP_PROJECTION.LINE_FRAMELET_OFFSET",
					"%d",	(char *) &map_xframe,
	".IMAGE_MAP_PROJECTION.X_AXIS_FRAMELET_OFFSET",
					"%d",	(char *) &map_xframe,
	".IMAGE_MAP_PROJECTION.SAMPLE_FRAMELET_OFFSET",
					"%d",	(char *) &map_yframe,
	".IMAGE_MAP_PROJECTION.Y_AXIS_FRAMELET_OFFSET",
					"%d",	(char *) &map_yframe,
	".IMAGE_MAP_PROJECTION.SPHERICAL_AZIMUTH",
					"%lf",	(char *) &map_azi,
	".IMAGE_MAP_PROJECTION.CARTESIAN_AZIMUTH",
					"%lf",	(char *) &map_rot,
	".IMAGE_MAP_PROJECTION.MAP_PROJECTION_ROTATION",
					"%lf",	(char *) &map_rot,
	".IMAGE_MAP_PROJECTION.POSITIVE_LONGITUDE_DIRECTION",
					"%s",	(char *) map_londir,
	".IMAGE_MAP_PROJECTION_CATALOG.MAP_PROJECTION_TYPE",
					"%s",	(char *) map_projtype,
	".IMAGE_MAP_PROJECTION_CATALOG.MAP_SCALE",
				"%[^\n]",	(char *) map_scale,
	".IMAGE_MAP_PROJECTION_CATALOG.LINE_PROJECTION_OFFSET",
					"%lf",	(char *) &map_xoff,
	".IMAGE_MAP_PROJECTION_CATALOG.X_AXIS_PROJECTION_OFFSET",
					"%lf",	(char *) &map_xoff,
	".IMAGE_MAP_PROJECTION_CATALOG.SAMPLE_PROJECTION_OFFSET",
					"%lf",	(char *) &map_yoff,
	".IMAGE_MAP_PROJECTION_CATALOG.Y_AXIS_PROJECTION_OFFSET",
					"%lf",	(char *) &map_yoff,
	".IMAGE_MAP_PROJECTION_CATALOG.A_AXIS_RADIUS",
				"%[^\n]",	(char *) map_radius,
	".IMAGE_MAP_PROJECTION_CATALOG.FIRST_STANDARD_PARALLEL",
					"%lf",	(char *) &map_phi0,
	".IMAGE_MAP_PROJECTION_CATALOG.SECOND_STANDARD_PARALLEL",
					"%lf",	(char *) &map_phi1,
	".IMAGE_MAP_PROJECTION_CATALOG.CENTER_LATITUDE",
					"%lf",	(char *) &map_clat,
	".IMAGE_MAP_PROJECTION_CATALOG.CENTER_LONGITUDE",
					"%lf",	(char *) &map_clon,
	".IMAGE_MAP_PROJECTION_CATALOG.LINE_FRAMELET_OFFSET",
					"%d",	(char *) &map_xframe,
	".IMAGE_MAP_PROJECTION_CATALOG.X_AXIS_FRAMELET_OFFSET",
					"%d",	(char *) &map_xframe,
	".IMAGE_MAP_PROJECTION_CATALOG.SAMPLE_FRAMELET_OFFSET",
					"%d",	(char *) &map_yframe,
	".IMAGE_MAP_PROJECTION_CATALOG.Y_AXIS_FRAMELET_OFFSET",
					"%d",	(char *) &map_yframe,
	".IMAGE_MAP_PROJECTION_CATALOG.SPHERICAL_AZIMUTH",
					"%lf",	(char *) &map_azi,
	".IMAGE_MAP_PROJECTION_CATALOG.CARTESIAN_AZIMUTH",
					"%lf",	(char *) &map_rot,
	".IMAGE_MAP_PROJECTION_CATALOG.MAP_PROJECTION_ROTATION",
					"%lf",	(char *) &map_rot,
	".IMAGE_MAP_PROJECTION_CATALOG.POSITIVE_LONGITUDE_DIRECTION",
					"%s",	(char *) map_londir,
	NULL
};

	/*
	**	Format of image
	*/

#define		FMT_FIXED	0	/* fixed-length records */
#define		FMT_VARYING	1	/* varying-length records */
#define		FMT_STREAM	2	/* stream-delimited records */
static	char	infm;			/* label file format */
static	char	dtype;			/* GIPS image pixel type */
static	int	nrec;			/* record number */
static	char	iname[256];		/* image file name */

	/*
	**	Miscellaneous
	*/

static	char	line[256];		/* input label line */
static	int	lablrecl;		/* fixed record length */
static	int	byteoff;		/* =1 if image offset is bytes */
/*static	FILE	*fp	= "stdin";*/	/* input label/image stream */
static	FILE	*fp;	/* input label/image stream */
static	FILE	*fq;			/* output image stream */
static	char	*ifile	= "'stdin'";	/* input file name */
static		aux();
/*  extern	char	*malloc();/*		/* C library */

main(argc, argv)
	int	argc;
	char	*argv[];
{
	double	d;
        fp = stdout;
	/*
	**	Parse command arguments
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s...", argv[2]);
	if (argc < 2)
		ifile = "'stdin'";
	else if (freopen(ifile = argv[1], "r", stdin) == NULL)
		gipserr2(GIPSBADOPEN, ifile);
	if (Sflag)
		telltale(0);

	/*
	**	Determine label format and read first "record"
	*/

	if (get_label_format())
		error("%s: unknown label format", ifile);

	/*
	**	Read file label and convert keywords
	*/

	read_label();
	if (hflag)
		exit(0);

	/*
	**	Darling Clementine patch
	*/

	if (strcmp(encoding, "CLEM-JPEG-0 DECOMPRESSED") == 0)
		encoding[0] = 0;

	/*
	**	Decode image location and format
	*/

	get_image_name();
	get_image_format();

	/*
	**	If -l specified, print info and exit
	*/

	if (lflag)
		list_label();

	/*
	**	Print image statistics
	*/
	if (vflag)
		fprintf(stderr, "%s: %d lines x %d samples x %d bits\n",
			iname, nline, nsamp, bitsperpix);

	/*
	**	Special processing for VICAR2 and STOKES images
	*/

	if (Vflag && strcmp(imagetype, "VICAR2") == 0)
		aux("ihvicar");

	if (strcmp(encoding, "COMPRESSED_STOKES_MATRIX") == 0)
		aux("ihstokes");

	if (dtype == '?')
		error("bad bits/pix: %d\n", (char *)bitsperpix);

	/*
	**	Construct output GIPS header
	*/

	(void) strcat(strcpy(line, GIPSPROFDIR), GINPROFILE);
	ihbcl(ihread(line, GIPSPROFILE));
	ihput("xnum", &nsamp);
	ihput("ynum", &nline);
	ihput("title", iname);
	ihput("dtype", &dtype);
	if (dnull != GIPSHIGH)
		ihput("dnull", &dnull);
	if (dzero != GIPSHIGH || dfact != GIPSHIGH) {
		double d0 = (dzero == GIPSHIGH) ? 0 : dzero;
		double d1 = (dfact == GIPSHIGH) ? 1 : dfact;
		ihput("dzero", &d0);
		ihput("dfact", &d1);
		ihput("dscale", "n");
		if (dtype == 'b') {
			ihput("dmin", &d0);
			ihput("dmax", (d = d0+255*d1, &d));
		} else if (dtype == 's') {
			ihput("dmin", (d = d0-128*d1, &d));
			ihput("dmax", (d = d0+127*d1, &d));
		} else if (dtype == 'h') {
			ihput("dmin", &d0);
			ihput("dmax", (d = d0+65535*d1, &d));
		} else if (dtype == 'u') {
			ihput("dmin", (d = d0-32768*d1, &d));
			ihput("dmax", (d = d0+32767*d1, &d));
		}
	}

	ihput("xscale", "w");
	d = 360;
	ihput("xwrap", &d);
	if (xmin != GIPSHIGH) {
		if (xmin > 180)
			xmin -= 360;
		ihput("xmin", (char *)&xmin);
	}
	if (xmax != GIPSHIGH) {
		if (xmax > 180)
			xmax -= 360;
		if (xmin != GIPSHIGH && xmin > xmax) {
			ihput("xmin", (char *)&xmax);
			xmax = xmin;
		}
		ihput("xmax", (char *)&xmax);
	}
	if (ymin != GIPSHIGH)
		ihput("ymin", (char *)&ymin);
	if (ymax != GIPSHIGH)
		ihput("ymax", (char *)&ymax);
	if (strcmp(projtype, "MERCATOR") == 0)
		ihput("yscale", "m");

	/*
	**	Check for IMAGE_MAP_PROJECTION_CATALOG object
	*/

	if (*map_projtype)
		get_mapping();

	/*
	**	If requested, merge header variables
	*/

	if (mfile)
		if (fq = ihmerg(mfile))
			ihbcl(fq);
		else gipserr();

	/*
	**	Load any encoding histogram
	*/

	if (encoding[0])
		find_hist();

	/*
	**	Locate start of input image
	*/

	find_image();

	/*
	**	Open the output GIPS file
	*/

	putargs();
	if ((fq = ihwrit(ofile, 1)) == NULL)
		gipserr();

	/*
	**	Copy the image
	*/

	if (copy_file())
		gipserr2(GIPSBADREAD, iname);

	/*
	**	Clean up
	*/

	fclose(fp);
	ihbcl(fq);
	if (Sflag)
		telltale(1);
	exit(0);
}


	/*
	**	Determine label format, read first record
	*/

get_label_format()
{
	int	i, n;
	char	*c;

	/*
	**	Read the first few bytes and check if VARYING_LENGTH
	*/

	for (c = line; ; ) {
		if ((i = getchar()) == EOF)
			gipserr2(GIPSBADREAD, ifile);
		if (i == EOF || i == '\n' || c >= line+sizeof(line)-1)
			return 1;
		switch (*c++ = i) {
		case 'L':
			if (c <= line+10)
				continue;
			if (strncmp(c-10, "SFDU_LABEL", 10))
				continue;
			break;
		case '1':
			if (c <= line+12)
				continue;
			if (strncmp(c-12, "PDSX00000001", 12))
				continue;
			break;
		case '3':
			if (strncmp(c-4, "PDS3", 4) == 0)
				break;
			continue;
		case '"':
			if (strncmp(c-6, "\"PDS3\"", 4) == 0)
				break;
		default:
			continue;
		}
		break;
	}

	if (!strncmp(line+2, "CCSD", 4) || !strncmp(line+2, "NJPL", 4)
			|| !strncmp(line+2, "PDS_VERSION_ID", 14)) {
		infm = FMT_VARYING;
		for (n = getshort(line); c < &line[n+2]; *c++ = i)
			if ((i = getchar()) == EOF)
				return 1;
		if (n & 1)
			(void)getchar();
	} else if (!strncmp(line, "CCSD", 4) || !strncmp(line, "NJPL", 4)
			|| !strncmp(line, "PDS_VERSION_ID", 14)) {
		while ((i = getchar()) == ' ' || i == '\t' || i == '\r')
			*c++ = i;
		if (i == EOF)
			return 1;
		if (i == '\n')
			infm = FMT_STREAM;
		else {
			ungetc(i, stdin);
			infm = FMT_FIXED;
			lablrecl = c-line;
		}
	} else
		return 1;

	/*
	**	Leave "line" NUL-terminated
	*/

	*c = 0;
	return 0;
}

	/*
	**	Remove whitespace and CR from end of line
	*/

char	*
strip(c)
	register char	*c;
{
	register char	*s = c;

	c += strlen(c);
	while (--c >= s && (isspace(*c) || *c == '\r'))
		;
	*++c = 0;
	return s;
}

	/*
	**	Read and parse a PDS label
	*/

read_label()
{
	register unsigned short	len;
	register char		*c;
	register int		sw	= 0;

	c = &line[2*(infm == FMT_VARYING)];

	if (hflag)
		printf("%s\n", strip(c));

	for (nrec = 1; strcmp(c, "END"); nrec++) {
		switch (infm) {
		case FMT_STREAM:
			if (fgets(line, sizeof(line)-1, stdin) == NULL)
				if (pdsv == 3 && feof(stdin))
					return;
				else
					gipserr2(GIPSBADREAD, ifile);
			break;
		case FMT_VARYING:
			fread(line, 2, 1, stdin);
			len = getshort(line);
			if (len > 0 && fread(line, len, 1, stdin) != 1)
				if (pdsv == 3 && feof(stdin))
					return;
				else
					gipserr2(GIPSBADREAD, ifile);
			line[len] = 0;
			if (len & 1)
				(void)getchar();
			break;
		case FMT_FIXED:
			if (fread(line, lablrecl, 1, stdin) != 1)
				if (pdsv == 3 && feof(stdin))
					return;
				else
					gipserr2(GIPSBADREAD, ifile);
			line[lablrecl] = 0;
			break;
		}
		for (c = line; isspace(*c); c++)
			;
		if (*c)
			(void)strip(c);
		else
			line[0] = 0;
		if (hflag)
			printf("%s\n", line);
		if (! sw)
			sw = get_keyword(c);
		else if (index(c, '"'))
			sw = 0;
	}
}

	/*
	**	Parse keyword=value in line
	*/

get_keyword(str)
	register char	*str;
{
	register char	*c;
	char		name[256];

	/*
	**	Locate the keyword
	*/

	for (c = str; *c != '='; c++)
		if (isspace(*c))
			*c = 0;
		else if (*c)
			;
		else if (strcmp(str, "END_OBJECT"))
			return 0;
		else {
			c[1] = 0;
			break;
		}
	*c = 0;

	/*
	**	Ignore comments
	*/

	if (str[0] == '/' && str[1] == '*')
		return 0;
	while (isspace(*++c))
		;

	/*
	**	Strip single and double quotation marks
	*/

	if (*c == '"')
		if (index(++c, '"'))
			*rindex(c, '"') = 0;
		else
			return 1;

	if (*c == '\'')
		if (index(++c, '\''))
			*rindex(c, '\'') = 0;
		else
			return 0;

	/*
	**	Go parse the keyword
	*/

	if (strcmp(str, "OBJECT") == 0)
		(void)sprintf(object+strlen(object), ".%s", c);
	else if (strcmp(str, "END_OBJECT") == 0) {
		if (! (c = rindex(object, '.')))
			c = object;
		*c = 0;
	} else {
		(void)sprintf(name, "%s.%s", object, str);
		for (k = keys; k->name && strcmp(name, k->name); k++)
			;
		if (! k->name)
			return 0;
		if (k == keys && strncmp(c, "PDS", 3) == 0) {
			sscanf(c+3, "%d", &pdsv);
			return 0;
		} else if (! k->name && !strcmp(k->fmt, "%s")) {
			strcpy(k->value, c);
			return 0;
		} else if (! k->name || !strcmp(c, "N/A") ||
				sscanf(c, k->fmt, k->value) == 1)
			return 0;
		(void)sprintf(name, "%s.%s: bad value: '%s'",
			object+1, str, c);
		error(name, "");
	}
	return 0;
}

	/*
	**	Determine location of input image
	*/

get_image_name()
{
	register char	*b, *c;
	char		file[256], ch;

	(void)strcpy(iname, ifile);
	byteoff = 0;

	if (pdsv < 3 && old_istruct[0] && ! istruct[0])
		strcpy(istruct, old_istruct);

	switch (istruct[0]) {
	case '\0':
		strcpy(istruct, ifile);
		return;
		/*error("image location not specified in PDS label", "");*/
	case '(':
		if (sscanf(istruct, "(\"%[^\"]\",%d <BYTES%c)",
			file, &iloc, &ch) == 3)
			byteoff = 1;
		else if (sscanf(istruct, "(\"%[^\"]\",%d)", file, &iloc) != 2)
			error("%s: bad image structure", istruct);
		b = file;
		break;
	default:
		if (isdigit(istruct[0])) {
			(void)sscanf(istruct, "%d", &iloc);
			if (sscanf(istruct, "%d <BYTES%c", &iloc, &ch) == 2)
				byteoff++;
			return;
		}
		b = istruct;
		break;
	}
	c = (c = rindex(iname, '/')) ? c+1 : iname;
	while (*c = *b++) {
		if (isupper(*c))
			*c = tolower(*c);
		c++;
	}
}

	/*
	**	Display image information
	*/

list_label()
{
	printf("%s:\n", ifile);
	if (imagetype[0])
		printf("\t%18.18s = %s\n", "header type", imagetype);
	if (recfm[0])
		printf("\t%18.18s = %s\n", "record format", recfm);
	if (encoding[0])
		printf("\t%18.18s = %s\n", "compression", encoding);
	if (byteoff)
		printf("\t%18.18s = %d\n", "starting byte", iloc);
	else
		printf("\t%18.18s = %d\n", "starting record", iloc);
	printf("\t%18.18s = %d\n", "image lines", nline);
	printf("\t%18.18s = %d\n", "image samples", nsamp);
	printf("\t%18.18s = %d\n", "bits/pixel", bitsperpix);
	printf("\t%18.18s = %s\n", "byte format", bytetype);
	printf("\t%18.18s = %d\n", "record length", reclen);
	printf("\t%18.18s = %d\n", "line header bytes", lineheadbytes);
	printf("\t%18.18s = %s\n", "image file name", iname);
	printf("\t%18.18s = %c\n", "GIPS format", dtype);
	exit(0);
}

	/*
	**	Decode pixel format
	*/

get_image_format()
{
	char	*b = bytetype;

	if (!strncmp(b, "VAX_", 4) || !strncmp(b, "LSB_", 4))
		b += 4, lsb = 1;
	else if (!strncmp(b, "MSB_", 4) || !strncmp(b, "HSB_", 4))
		b += 4, lsb = 0;
	else if (!strncmp(b, "IEEE_", 4))
		b += 5, lsb = 0;

	if (bflag) lsb = 1 - lsb;

	switch (bitsperpix) {
	case 1:
	case 4:
	case 8:
		dtype = 'b';
		break;
	case 16:
		dtype = strncmp(b, "UNSIGNED", 8) ? 'h' : 'u';
		break;
	case 32:
		if (!strncmp(b, "REAL", 4))
			dtype = 'f';
		else
			dtype = strncmp(b, "UNSIGNED", 8) ? 'i' : 'l';
		break;
	default:
		dtype = '?';
		break;
	}
}

	/*
	**	Parse IMAGE_MAP_PROJECTION_CATALOG and
	**	construct a "mapping" header variable.
	*/

static	struct	{
	int	nproj;			/* GIPS projection index */
	char	*name;			/* GIPS projection name */
	char	*pdsname;		/* PDS projection name */
} *pt, ptab[] = {
	1,	"Gnomonic",		"GNOMONIC",
	2,	"Stereographic",	"STEREOGRAPHIC",
	3,	"Orthographic",		"ORTHOGRAPHIC",
	4,	"Equidistant",		"EQUIDISTANT",
	5,	"Equivalent",		"EQUIVALENT",
	6,	"Lambert_Conic_1",	"LAMBERT_CONIC_ONE_PARALLEL",
	7,	"Lambert_Conic_2",	"LAMBERT_CONIC_TWO_PARALLEL",
	8,	"Mercator",		"MERCATOR",
	9,	"Stereo_Polar",		"POLAR_STEREOGRAPHIC",
	10,	"Mollweide",		"MOLLWEIDE",
	11,	"Albers_1",		"ALBERS_ONE_PARALLEL",
	12,	"Albers_2",		"ALBERS_TWO_PARALLEL",
	13,	"Lambert_Cylindrical",	"LAMBERT_CYLINDRICAL",
	14,	"Lambert_Polar",	"LAMBERT_POLAR",
	15,	"Bonne",		"BONNE",
	16,	"Sinusoidal",		"SINUSOIDAL",
	17,	"Werner",		"WERNER",
	19,	"Aitoff",		"AITOFF",
	20,	"Hammer",		"HAMMER",
	21,	"Briesemeister",	"BRIESEMEISTER",
	22,	"Simple_Cylindrical",	"CYLINDRICAL",
	22,	"Simple_Cylindrical",	"SIMPLE CYLINDRICAL",
	0
};

get_mapping()
{
	char	buf[256];		/* scratch */
	int	xyrot;			/* Cartesian rotation */
	double	rad;			/* radius */
	double	scale;			/* scale */

	for (pt = ptab; strcmp(map_projtype, pt->pdsname); pt++)
		if (pt[1].nproj == 0) {
			fprintf(stderr, "%s: warning: %s not supported\n",
				ifile, map_projtype);
			return;
		}

	switch (sscanf(map_radius, "%lf %s", &rad, buf)) {
	case 1:
		break;
	case 2:
		if (!strcmp(buf, "<M>"))
			rad *= 0.001;
		else if (strcmp(buf, "<KM>"))
			return;
		break;
	default:
		return;
	}

	if (map_scale[0])
		switch (sscanf(map_scale, "%lf %s", &scale, buf)) {
		case 1:
			break;
		case 2:
			if (!strcmp(buf, "<M/PIXEL>"))
				scale *= 0.001;
			else if (strcmp(buf, "<KM/PIXEL>"))
				return;
			break;
		default:
			return;
		}
	else if (map_resolution[0]) {
		double resolution;
		switch (sscanf(map_resolution, "%lf %s", &resolution, buf)) {
		case 1:
			break;
		case 2:
			if (!strcmp(buf, "<DEGREES>"))
				resolution *= atan(1.0)/45;
			else if (strcmp(buf, "<RADIANS>"))
				return;
			break;
		default:
			return;
		}
		scale = rad * resolution;
	} else
		return;

	if (Wflag || strncmp(map_londir, "WEST", 4) == 0)
		map_clon = -map_clon;

	while (map_rot < -180)
		map_rot += 360;
	while (map_rot > 180)
		map_rot -= 360;
	if (map_rot == 90)
		xyrot = 1;
	else if (map_rot == -90)
		xyrot = -1;
	else if (map_rot == 0)
		xyrot = 0;
	else {
		fprintf(stderr,
			"%s: warning: CARTESIAN_AZIMUTH %g not supported\n",
			ifile, map_rot);
		return;
	}

	/* fix Magellan FMAPs */
	if (strcmp(dsid, "MGN-V-RDRS-5-DIM-V1.0") == 0) {
		map_xoff = -map_xoff;
		map_yoff = -map_yoff;
	}

	sprintf(buf, GIPS_MAPPING_P, pt->name, pt->nproj, map_yoff, map_xoff,
		rad/scale, map_clon, map_clat, map_azi, map_phi0, map_phi1, xyrot);
	ihput("mapping", buf);
}

	/*
	**	Read to start of image
	*/

find_image()
{
	register int	n, len, infm2;
	
	if (strcmp(recfm, "FIXED_LENGTH") == 0)
		infm2 = FMT_FIXED;
	else if (strcmp(recfm, "VARYING_LENGTH") == 0)
		infm2 = FMT_VARYING;
	else if (strcmp(recfm, "VARIABLE_LENGTH") == 0)
		infm2 = FMT_VARYING;
	else if (*recfm)
		error("bad image format: %s", recfm);
	else
		error("unspecified image format", "");
	if (! strcmp(iname, ifile) && infm == infm2)
		return;
	if (freopen(iname, "r", stdin) == NULL) {
		register char *c = rindex(iname, '/');
		for (c = c ? c+1 : iname; *c; c++)
			if (isupper(*c))
				*c = tolower(*c);
			else if (islower(*c))
				*c = toupper(*c);
		if (freopen(iname, "r", stdin) == NULL)
			gipserr2(GIPSBADOPEN, iname);
	}
	infm = infm2;
	if (hdrlen > 0)
		fseek(stdin, (long)hdrlen, 0);
	nrec = 0;
}

	/*
	**	Read the encoding histogram
	*/

find_hist()
{
	register int	len, hlen, rc;
	register char	*c, *h;

	if (strcmp(encoding, "HUFFMAN_FIRST_DIFFERENCE"))
		error("unsupported encoding type: %s", encoding);

	hlen = nhist * sizeof(long);
	c = h = alloc(1, hlen+reclen);

	for ( ; c < h+hlen; nrec++) {
		switch (infm) {
		case FMT_FIXED:
			rc = fread(c, len = reclen, 1, stdin);
			break;
		case FMT_VARYING:
			if (fread(line, 2, 1, fp) != 1)
				return 1;
			len = getshort(line);
			rc = fread(c, len, 1, stdin);
			if (len & 1)
				(void)getchar();
			break;
		case FMT_STREAM:
			error("bad histogram format: STREAM", "");
		}
		if (rc != 1)
			gipserr2(GIPSBADREAD, ifile);
		if (++nrec > encoderec)
			c += len;
	}

	decmpinit((long *)h);
	free(h);
}

	/*
	**	Copy the image
	*/

copy_file()
{
	int	dlen, len;
	register int	i, l;
	register char	*inrec, *cmprec, *s;

	/*
	**	Allocate buffers
	*/

	if (hdrrec+lblrec > iloc)
		iloc = hdrrec+lblrec+1;
	dlen = (bitsperpix*nsamp+7)/8;
	l = (linelen > reclen) ? linelen : reclen;
	inrec = (l < sizeof(line)) ? line : malloc(l);
	cmprec = encoding[0] ? malloc(l) : inrec;

	while (nline > 0) {
		if (++nrec >= iloc && linelen > 0)
			reclen = linelen;
		switch (infm) {
		case FMT_FIXED:
			if (fread(inrec, reclen, 1, fp) != 1)
				return 1;
			break;
		case FMT_VARYING:
			if (fread(cmprec, 2, 1, fp) != 1)
				return 1;
			len = getshort(cmprec);
			if (fread(cmprec, len, 1, fp) != 1)
				return 1;
			if (len & 1)
				(void) getc(fp);
			break;
		case FMT_STREAM:
			error("bad image format: STREAM", "");
		}

		if (nrec < iloc)
			continue;

		if (encoding[0])
			decompress(cmprec, inrec, &len, &reclen);

		s = inrec+lineheadbytes;
		switch (bitsperpix) {
		case 1:
			for (i = 0; i < nsamp; i++, l <<= 1) {
				if ((i & 7) == 0)
					l = *s++;
				putc((l >> 7) & 1, fq);
			}
			break;
		case 4:
			for (i = 0; i < nsamp; i++, l <<= 4) {
				if ((i & 1) == 0)
					l = *s++;
				putc((l >> 4) & 15, fq);
			}
			break;
		case 32:
			if (lsb != ORDER) {
				for (i = 0; i < nsamp; i++, s += 4)
					if (dtype == 'f') {
#ifdef vax
						putc(s[1], fq);
						putc(s[0] ? s[0]+1 : 0, fq);
#else
						putc(s[1] ? s[1]-1 : 0, fq);
						putc(s[0], fq);
#endif vax
						putc(s[3], fq);
						putc(s[2], fq);
					} else {
						putc(s[3], fq);
						putc(s[2], fq);
						putc(s[1], fq);
						putc(s[0], fq);
					}
				break;
			}
		case 16:
			if (lsb != ORDER)
				swab(s, s, dlen);
		case 8:
			fwrite(s, dlen, 1, fq);
			break;
		}
		if (ferror(fq))
			gipserr2(GIPSBADWRITE, ofile);
		nline--;
	}
	return 0;
}

static
aux(cmd)
	char		*cmd;
{
	char		buf[256], *c, *arg[20], **ap = arg;
	register char	*inrec;
	register int	len, l, sw;
	register FILE	*fq;

	/*
	**	Locate start of input image
	*/

	if (hstruct[0]) {
		(void)strcpy(istruct, hstruct);
		iloc = 1;
		get_image_name();
	}

	if (strcmp(iname, ifile)) {
		if (iloc != 1)
			fprintf(stderr,
				"%s: warning: bad offset to image header: %d\n",
				iname, (char *)iloc);
		find_image();
	}

	*ap++ = cmd;
	
	/*
	**	Fix Magellan MIDR CD-ROM goofs
	*/

	if (strcmp(cmd, "ihvicar") == 0) {
		if (bflag)
			*ap++ = "-b";
		if (! strncmp(iname+strlen(iname)-10, "BROWSE.IMG", 10) ||
			! strncmp(iname+strlen(iname)-10, "browse.img", 10))
			if (! strncmp(imageid, "F-MIDR", 6) ||
				! strncmp(imageid, "C1-MIDR", 7) ||
				! strncmp(imageid, "C2-MIDR", 7) ||
				! strncmp(imageid, "C3-MIDR", 7))
					*ap++ = "-M8";
			else if (! strncmp(imageid, "P-MIDR.N", 8) ||
				! strncmp(imageid, "P-MIDR.S", 8))
					*ap++ = "-M10";
		if (plane) {
			*ap++ = "-p";
			*ap++ = plane;
		}
	}

	if (vflag)
		*ap++ = "-D";
	if (Lflag)
		*ap++ = "-L";
	if (conv) {
		*ap++ = "-c";
		*ap++ = conv;
	}
	if (mfile) {
		*ap++ = "-m";
		*ap++ = mfile;
	}
	*ap++ = "-o";
	*ap++ = ofile;
	*ap++ = iname;
	*ap++ = 0;

	switch (infm) {
	case FMT_FIXED:

		/*
		**	Fixed length - let the command do the job
		*/

		(void)fclose(fp);
		(void)close(0);
		execvp(cmd, arg);
		error("cannot display image - can't find %s", cmd);
		break;

	case FMT_VARYING:

		/*
		**	Varying length - reformat and pipe to command
		*/

		inrec = malloc(reclen);
		buf[0] = 0;
		for (ap = arg; *ap; ap++)
			sprintf(buf+strlen(buf), 
				index(*ap, ' ') ? "'%s' " : "%s ", *ap);
		if ((fq = popen(buf, "w")) == NULL)
			gipserr2(GIPSBADOPEN, cmd);
		while (fread(inrec, 2, 1, fp) == 1 && ! ferror(fq)) {
			for (l = len = getshort(inrec); --l >= 0; )
				(void) putc(getc(fp), fq);
			if (len & 1)
				(void) getc(fp);
		}
		(void) pclose(fq);
		if (ferror(fp))
			gipserr2(GIPSBADREAD, iname);
		if (ferror(fq))
			gipserr2(GIPSBADWRITE, cmd);
		break;

	case FMT_STREAM:
		error("bad %s image format: STREAM", cmd);
	}

	(void) fclose(fp);
	if (Sflag)
		telltale(1);
	exit(0);
}
