#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihstokes.c,v $	GIPS $Revision: 1.7 $ (MIT/CSR) $Date: 1993/01/27 22:19:52 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990 Massachusetts Institute of Technology
**
**	Module:		ihstokes.c	Creation Date:	07/02/91
**	Version:	$Revision: 1.7 $		Update:		$Date: 1993/01/27 22:19:52 $
**	Function:	Convert PDS compressed Stokes image
**	Components:	main()		GIPS/UNIX command
*/

#include <string.h>  /* for strcat and strcpy */
#include <stdio.h>
#include <math.h>
#include "gips.h"

#define	DEG2RAD(x) (M_PI*(x)/90)

	/*
	**	Command flags and options
	*/

static	char	Dflag;			/* debug */
static	char	Lflag;			/* output in dB */
static	char	Sflag;			/* print execution statistics */
static	char	bflag;			/* unused */
static	char	*mfile;			/* merged header file */
static	char	*ofile	= "-";		/* output file */
static	char	*tr_rv;			/* conversion */

static	flagentry	flagtable[] = {
	"D",	"1",			(char *)&Dflag,
	"L",	"1",			(char *)&Lflag,
	"S",	"1",			(char *)&Sflag,
	"b",	"1",			(char *)&bflag,
	"c",	"-> te,ta,re,ra",	(char *)&tr_rv,
	"m",	"-> file",		(char *)&mfile,
	"o",	"-> file",		(char *)&ofile,
	NULL,	"[file]",		NULL
};

static	double		rec_e, rec_a;
static	double		tra_e, tra_a;
static	float		norm[256][256];
static	char		line[1024];
static	int		samplen;	/* bytes per sample */
static	int		off[3];		/* offset to old/user header, data */
static	char		type[1024];	/* image type */

main(argc, argv)
	int	argc;
	char	*argv[];
{
	char		*ifile;		/* input file name */
	unsigned char	*inbuf;		/* input buffer */
	FILE		*fq;		/* output stream */
	float		*convbuf;	/* intermediate buffer */
	float		*outbuf;	/* output buffer */
	float		s[4][4];	/* Stokes matrix */
	float		sr[4];		/* receive Stokes vector */
	float		st[4];		/* transmit Stokes vector*/
	float		srt[4][4];	/* sr x st */
	float		d;		/* power */
	float		conv[256];	/* input conversion table */
	float		aconv[256];	/* input conversion table */
	int		ynum;		/* number of lines */
	int		xnum;		/* number of samples */
	int		dlen;		/* output pixel length */
	char		dtype;		/* output pixel type */
	char		dscale;		/* output pixel type */
	double		dzero, dfact;	/* conversion factors */
	double		dnull;		/* null pixel value */
	register int	x, i, j;	/* scratch */
	register unsigned char	*b;	/* scratch */

	/*
	**	Parse command arguments
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s...", argv[2]);
	if (argc < 2)
		ifile = "'stdin'";
	else if (freopen(ifile = argv[1], "r", stdin) == NULL)
		gipserr2(GIPSBADOPEN, ifile);
	if (Sflag)
		telltale(0);

	/*
	**	Construct Stokes vectors
	*/

	if (tr_rv && sscanf(tr_rv, "%lg,%lg,%lg,%lg",
			&tra_e, &tra_a, &rec_e, &rec_a) != 4)
		error("bad -c value: %s", tr_rv);

	st[0] = 1;
	st[1] = cos(DEG2RAD(tra_a)) * cos(DEG2RAD(tra_e));
	st[2] = sin(DEG2RAD(tra_a)) * cos(DEG2RAD(tra_e));
	st[3] = sin(DEG2RAD(tra_e));

	sr[0] = 1;
	sr[1] = cos(DEG2RAD(rec_a)) * cos(DEG2RAD(rec_e));
	sr[2] = sin(DEG2RAD(rec_a)) * cos(DEG2RAD(rec_e));
	sr[3] = sin(DEG2RAD(rec_e));

	if (Dflag) {
		fprintf(stderr, "ihstokes parameters:\n\n");
		fprintf(stderr,
			"\tInput file:                 %s\n", ifile);
		fprintf(stderr,
			"\tTransmitted phase and ecc:  { %g, %g }\n",
			tra_a, tra_e);
		fprintf(stderr,
			"\tTransmitted Stokes Vector:  { %g, %g, %g, %g }\n",
			st[0], st[1], st[2], st[3]);
		fprintf(stderr,
			"\tReceived phase and ecc:     { %g, %g }\n",
			rec_a, rec_e);
		fprintf(stderr,
			"\tReceived Stokes Vector:     { %g, %g, %g, %g }\n",
			sr[0], sr[1], sr[2], sr[3]);
		fprintf(stderr, "\nNew Header:\n\n");
	}

	/*
	**	Read image header
	*/

	for (i = 50; ; i += 50) {
		if (fread(line, 50, 1, stdin) != 1)
			gipserr2(GIPSBADREAD, ifile);
		if (! line[0])
			break;
		if (Dflag)
			fprintf(stderr, "\t%s\n", line);
		sscanf(line, "NUMBER OF SAMPLES PER RECORD = %d", &xnum);
		sscanf(line, "NUMBER OF LINES IN IMAGE = %d", &ynum);
		sscanf(line, "NUMBER OF BYTES PER SAMPLE = %d", &samplen);
		sscanf(line, "DATA TYPE = %s", type);
		sscanf(line, "BYTE OFFSET OF OLD HEADER = %d", &off[0]);
		sscanf(line, "BYTE OFFSET OF USER HEADER = %d", &off[1]);
		sscanf(line, "BYTE OFFSET OF FIRST DATA RECORD = %d", &off[2]);
	}

	if (xnum == 0 || ynum == 0)
		error("%s: image dimensions missing from header", ifile);
	if (samplen != 10 || strcmp(type, "COMPRESSED"))
		error("%s: not a Stokes image", ifile);
	if (off[2] < i)
		error("%s: cannot find start of data", ifile);

	/*
	**	Print old and user headers
	*/

	for (j = 0; j < 2; j++) {
		if (! Dflag || ! off[j] || i > off[j])
			continue;
		while (i < off[j])
			getchar(), i++;
		while ((i += 50) && fread(line, 50, 1, stdin) == 1 && line[0]) {
			if (off[j])
				fprintf(stderr, "\n%s Header:\n\n",
					j ? "User" : "Old", off[j] = 0);
			fprintf(stderr, "\t%s\n", line);
		}
	}

	while (i++ < off[2])
		getchar();
		
	/*
	**	Take outer product of Stokes vectors and
	**	adjust the diagonal components
	*/

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			srt[i][j] = sr[i] * st[j] + 
				(i == j ? 0 : sr[j] * st[i]);

	srt[0][0] += srt[1][1];
	srt[2][2] -= srt[1][1];
	srt[3][3] -= srt[1][1];

	/*
	**	Construct lookup tables
	*/

	for (i = -128; i < 128; i++) {
		conv[i & 255] = i / 127.0;
		aconv[i & 255] = conv[i & 255] * fabs(conv[i & 255]);
	}

	for (i = -128; i <= 128; i++) {
		d = pow(2.0, (float)i);
		for (j = -128; j < 128; j++)
			norm[i & 255][j & 255] = d * (j / 254.0 + 1.5);
	}

	/*
	**	Construct output GIPS header
	*/

	(void) strcat(strcpy(line, GIPSPROFDIR), GINPROFILE);
	ihbcl(ihread(line, GIPSPROFILE));
	ihput("xnum", &xnum);
	ihput("ynum", &ynum);
	ihput("dtype", "f");
	ihput("dscale", "s");
	ihput("dtitle", Lflag ? "Power (dB)" : "Power (lin)");
	inbuf = (unsigned char *)alloc(sizeof(unsigned char), xnum * 10);
	convbuf = outbuf = (float *)alloc(sizeof(float), xnum);

	/*
	**	If requested, merge header variables
	*/

	if (mfile)
		if (fq = ihmerg(mfile))
			ihbcl(fq);
		else gipserr();

	/*
	**	Open the output GIPS file
	*/

	putargs();
	if ((fq = ihwrit(ofile, 1)) == NULL)
		gipserr();

	/*
	**	Now determine the output format
	*/

	if (ihget("dscale", &dscale))
		dscale = 's';
	else if (dscale == 'n')
		if (ihget("dzero", &dzero) || ihget("dfact", &dfact))
			gipserr();
		else if (dfact == 0)
			gipserr();
		else
			dfact = 1/dfact;

	ihget("dtype", &dtype);
	if ((dlen = ihdtype(dtype)) == 0)
		gipserr();
	if (dtype != 'f')
		outbuf = (float *)alloc(dlen, xnum);
	if (ihget("dnull", &dnull))
		dnull = 0;
	
	/*
	**	Copy the image
	*/

	while (--ynum >= 0) {
		if (fread(b = inbuf, xnum, 10, stdin) != 10)
			gipserr2(GIPSBADREAD, ifile);
		for (x = 0; x < xnum; convbuf[x++] = d) {
			i = *b++;
			j = *b++;
			d = srt[0][0];
			d += srt[0][1] * conv[*b++];
			d += srt[0][2] * aconv[*b++];
			d += srt[0][3] * aconv[*b++];
			d += srt[1][2] * aconv[*b++];
			d += srt[1][3] * aconv[*b++];
			d += srt[2][2] * conv[*b++];
			d += srt[2][3] * conv[*b++];
			d += srt[3][3] * conv[*b++];
			d *= norm[i][j];
			if (d <= 0) {
				d = dnull;
				continue;
			}
			d = Lflag ? 5*log10(d) : sqrt(d);
			if (dscale == 'n')
				d = (d - dzero) * dfact;
		}
		if (dtype != 'f')
			gconv(outbuf, convbuf, xnum, ihdcvt(dtype), CVT_F);
		if (fwrite(outbuf, dlen, xnum, fq) != xnum)
			gipserr2(GIPSBADWRITE, ofile);
	}

	/*
	**	Clean up
	*/

	ihbcl(fq);
	if (Sflag)
		telltale(1);
	exit(0);
}
