#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihvicar.c,v $	GIPS $Revision: 2.40 $ (MIT/CSR) $Date: 1998/11/13 23:25:54 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihvicar.c,v $		Creation Date:	05/10/88
**	Version:	$Revision: 2.40 $		Update:		$Date: 1998/11/13 23:25:54 $
**	Function:	Convert VICAR image file to GIPS format
**	Components:	main()		GIPS/UNIX main program
*/

#include <string.h>   /* for strcat and strcpy */
#include <stdio.h>
#include <signal.h>
#include <math.h>
#define	GRID
#include "gips.h"

#define	UNINIT	(1.e30)			/* uninitialized */

#ifdef vax
#define	INT_FMT		"LOW"
#define	REAL_FMT	"VAX"
#else
#define	INT_FMT		"HIGH"
#define	REAL_FMT	"IEEE"
#endif vax

	/*
	**	Input variables
	*/
	
static	char	*ifile	= NULL;		/* input VICAR file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= "-";		/* output GIPS file */
static	int	block;			/* input (tape) block size */
static	int	plane;			/* desired image plane */
static	double	magn	= 1;		/* PIXSIZ magnification */
static	flag	bflag;			/* byte swap flag */
static	flag	vflag;			/* VAX float flag */
static	flag	Dflag;			/* debug flag */
static	flag	Sflag;			/* statistics flag */
static	flag	Xflag;			/* by-pass special MGN midr code */

	/*
	**	Variable definitions
	*/
	
static	flagentry	flagtable[] = {
	"",	"%d block",	(flag *) &block,
	"b",	"1",		(flag *) &bflag,
	"v",	"1",		(flag *) &vflag,
	"M",	"%lf magn",	(flag *) &magn,
	"D",	"1",		(flag *) &Dflag,
	"S",	"1",		(flag *) &Sflag,
	"X",	"1",		(flag *) &Xflag,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"p",	"%d plane",	(flag *) &plane,
	NULL,	"[image]",	NULL
};

static	char	buf[4096];
static	void	swap2(), swapv(), swap4();
static		get_mapping();

static	char	*FORMAT, *TYPE, *ORG, *TASK, *USER, *DAT_TIM;
static	char	*PIXSIZ_U, *RADIUS_U, *PRODUCT;
static	char	*MAP_PROJ, *CIMAGE, *DN_UNITS, *INTFMT, *REALFMT;
static	int	LBLSIZE, BUFSIZE, DIM, EOL, RECSIZE, SUBFRAME, SUBF_TOT;
static	int	NL, NS, NB, N1, N2, N3, NBB, NLB, IFILE;
static	double	SPECLINE = UNINIT, SPECSAMP = UNINIT, PROJSAMP = UNINIT;
static	double	SUBF_COL = UNINIT, SUBF_ROW = UNINIT, RADIUS = UNINIT; 
static	double	LOW_DN = UNINIT, HI_DN = UNINIT;
static	double	LAT_UR = UNINIT, LAT_LR = UNINIT;
static	double	LAT_UL = UNINIT, LAT_LL = UNINIT;
static	double	LON_UR = UNINIT, LON_LR = UNINIT;
static	double	LON_UL = UNINIT, LON_LL = UNINIT;
static	double	LON_CL = UNINIT, LON_CR = UNINIT;
static	double	LAT_UC = UNINIT, LAT_LC = UNINIT;
static	double	LOW_REP = UNINIT, HI_REP = UNINIT;
static	double	PIXSIZ = UNINIT, PROJ_LON = UNINIT, SPDN_1 = UNINIT;

#define	OK(d)	((d) != UNINIT)

#define	REQD	01		/* keyword must be present */
#define	FOUND	02		/* keyword is present */
#define	CHAR	04		/* keyword is character string */
#define	INT	010		/* keyword is integer */
#define	DOUBLE	020		/* keyword is double */

static	struct	keytab	{
	char	*v_name;	/* header keyword */
	char	*v_ptr;		/* where to copy the value */
	int	v_flg;		/* flag fields */
} keys[] = {
	"LBLSIZE",	(char *) &LBLSIZE,	INT | REQD,
	"TASK",		(char *) &TASK,		CHAR,
	"USER",		(char *) &USER,		CHAR,
	"DAT_TIM",	(char *) &DAT_TIM,	CHAR,
	"FORMAT",	(char *) &FORMAT,	CHAR | REQD,
	"TYPE",		(char *) &TYPE,		CHAR,
	"INTFMT",	(char *) &INTFMT,	CHAR,
	"PRODUCT",	(char *) &PRODUCT,	CHAR,
	"REALFMT",	(char *) &REALFMT,	CHAR,
	"BUFSIZ",	(char *) &BUFSIZE,	INT | REQD,
	"DIM",		(char *) &DIM,		INT,
	"EOL",		(char *) &EOL,		INT,
	"RECSIZE",	(char *) &RECSIZE,	INT | REQD,
	"ORG",		(char *) &ORG,		CHAR,
	"NL",		(char *) &NL,		INT | REQD,
	"NS",		(char *) &NS,		INT | REQD,
	"NB",		(char *) &NB,		INT,
	"N1",		(char *) &N1,		INT,
	"N2",		(char *) &N2,		INT,
	"N3",		(char *) &N3,		INT,
	"NBB",		(char *) &NBB,		INT,
	"NLB",		(char *) &NLB,		INT,
	"LOW_DN",	(char *) &LOW_DN,	DOUBLE,
	"HI_DN",	(char *) &HI_DN,	DOUBLE,
	"LOW_REP",	(char *) &LOW_REP,	DOUBLE,
	"HI_REP",	(char *) &HI_REP,	DOUBLE,
	"LAT_LL",	(char *) &LAT_LL,	DOUBLE,
	"LAT_LR",	(char *) &LAT_LR,	DOUBLE,
	"LAT_UL",	(char *) &LAT_UL,	DOUBLE,
	"LAT_UR",	(char *) &LAT_UR,	DOUBLE,
	"LON_LL",	(char *) &LON_LL,	DOUBLE,
	"LON_LR",	(char *) &LON_LR,	DOUBLE,
	"LON_UL",	(char *) &LON_UL,	DOUBLE,
	"LON_UR",	(char *) &LON_UR,	DOUBLE,

	"MAP_PROJ",	(char *) &MAP_PROJ,	CHAR,
	"IMAGE",	(char *) &CIMAGE,	CHAR,
	"DN_UNITS",	(char *) &DN_UNITS,	CHAR,
	"RADIUS_U",	(char *) &RADIUS_U,	CHAR,
	"PIXSIZ_U",	(char *) &PIXSIZ_U,	CHAR,
	"PROJSAMP",	(char *) &PROJSAMP,	DOUBLE,
	"SUBFRAME",	(char *) &SUBFRAME,	INT,
	"FILE",		(char *) &IFILE,	INT,
	"SUBF_TOT",	(char *) &SUBF_TOT,	INT,
	"SPECLINE",	(char *) &SPECLINE,	DOUBLE,
	"SPECSAMP",	(char *) &SPECSAMP,	DOUBLE,
	"SUBF_COL",	(char *) &SUBF_COL,	DOUBLE,
	"SUBF_ROW",	(char *) &SUBF_ROW,	DOUBLE,
	"LAT_UC",	(char *) &LAT_UC,	DOUBLE,
	"LON_CL",	(char *) &LON_CL,	DOUBLE,
	"LAT_LC",	(char *) &LAT_LC,	DOUBLE,
	"LON_CR",	(char *) &LON_CR,	DOUBLE,
	"PIXSIZ",	(char *) &PIXSIZ,	DOUBLE,
	"RADIUS",	(char *) &RADIUS,	DOUBLE,
	"PROJ_LON",	(char *) &PROJ_LON,	DOUBLE,
	"SPDN_1",	(char *) &SPDN_1,	DOUBLE,
	NULL
};

main(argc, argv)
	int	argc;
	char	**argv;
{
	int		xnum;		/* number of samples */
	int		dlen;		/* number of bytes per pixel */
	int		inlen;		/* input raster length */
	int		outlen;		/* output raster length */
	int		ynum;		/* number of rows */
	int		ptype;		/* band interleave type */
	int		noseek;		/* ==0 of seekable */
	char		dtype;		/* pixel type code */
	char		*bin;		/* input buffer */
	char		*bout;		/* output buffer */
	char		*dummy;		/* output buffer */
	void		(*sproc)();	/* field reformatting routine */
	double		xmin, xmax;	/* x-axis limits */
	double		ymin, ymax;	/* y-axis limits */
	register FILE	*fp;		/* output image stream */
	register int	n, sw;		/* scratch */
	double		d;		/* scratch */
	
	/*
	**	Parse command string
	*/
	
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...", argv[2]);
	else if (argc < 2)
		(void)signal(SIGPIPE, exit), ifile = "'stdin'";
	else if (freopen(ifile = argv[1], "r", stdin) == NULL)
		perror(ifile), exit(1);
	if (Sflag)
		telltale(0);

	/*
	**	If requested, change the input block size
	*/

	if (block > 0)
		setbuffer(stdin, alloc(block, 1), block);
		
	/*
	**	Read the VICAR header
	*/
	
	scan_header();

	/*
	**	Determine pixel type
	*/
	
	if (strcmp(FORMAT, "BYTE") == 0) {
		dlen = 1;
		dtype = 'b';
	} else if (strcmp(FORMAT, "HALF") == 0) {
		dlen = 2;
		dtype = 'u';
	} else if (strcmp(FORMAT, "FULL") == 0) {
		dlen = 4;
		dtype = 'l';
	} else if (strcmp(FORMAT, "REAL") == 0) {
		dlen = 4;
		dtype = 'f';
		if (! REALFMT)
			;
		else if (strcmp(REALFMT, REAL_FMT))
			vflag = 1;
		else if (vflag)
			error("-v flag conflicts with REALFMT='%s'", REALFMT);
	} else
		error("illegal FORMAT value: %s", FORMAT);

	/*
	**	Determine pixel formatting routine
	*/

	if (! INTFMT)
		;
	else if (strcmp(INTFMT, INT_FMT))
		bflag = 1;
	else if (bflag)
		error("-b flag conflicts with INTFMT='%s'", INTFMT);

	if (dtype == 'u' && bflag)
		sproc = swap2;
	else if (dtype == 'l' && bflag)
		sproc = swap4;
	else if (dtype == 'f' && bflag && ! vflag)
		sproc = swap4;
	else if (dtype == 'f' && vflag)
		sproc = swapv;
	else
		sproc = NULL;

	/*
	**	Determine band type
	*/

	if (DIM == 0)
		DIM = 2;
	else if (DIM < 2 || DIM > 3)
		error("unsupported DIM value: %d\n", (char *)DIM);
	if (NB == 0)
		NB = 1;
	if (DIM == 2 || NB < 2)
		ptype = 0;
	else if (! ORG || !strcmp(ORG, "BSQ"))
		ptype = 1;
	else if (!strcmp(ORG, "BIL"))
		ptype = 2;
	else if (!strcmp(ORG, "BXX"))
		ptype = 3;
	else
		error("unsupported ORG value: %d\n", ORG);
	if (plane >= NB)
		error("bad -p value: %d\n", (char *)plane);

	/*
	**	Construct in-core GIPS header,
	*/
	
	if ((fp = ihread(strcat(strcpy(buf,
			GIPSPROFDIR), GINPROFILE), GIPSPROFILE)) == NULL)
		gipserr();
	(void)ihbcl(fp);
	ihput("dtype", &dtype);
	ihput("xnum", &NS);
	ihput("ynum", &NL);
	ihput("xscale", "w", 1);
	ihput("yscale", "s", 1);
	inlen = NBB + NS * dlen;
	if (CIMAGE && *CIMAGE)
		ihput("title", CIMAGE);
	if (DN_UNITS && *DN_UNITS)
		ihput("dtitle", DN_UNITS);
	if (PRODUCT && *PRODUCT)
		ihput("origin", PRODUCT);
	else if (TASK && *TASK)
		ihput("origin", TASK);
	if (USER && *USER)
		ihput("owner", USER);
	if (DAT_TIM && *DAT_TIM)
		ihput("date", DAT_TIM);
	if (OK(LON_UL) && OK(LON_UR)) {
		ihput("xmin", &LON_UL, 1);
		ihput("xmax", &LON_UR, 1);
	}
	if (OK(LAT_UL) && OK(LAT_LL)) {
		ihput("ymin", &LAT_UL, 1);
		ihput("ymax", &LAT_LL, 1);
	}
	if (OK(LOW_REP))
		ihput("dmin", &LOW_REP, 0);
	else if (OK(LOW_DN))
		ihput("dmin", &LOW_DN, 0);
	if (OK(HI_REP))
		ihput("dmax", &HI_REP, 0);
	else if (OK(HI_DN))
		ihput("dmax", &HI_DN, 0);
	if (OK(SPDN_1))
		ihput("dnull", &SPDN_1, 0);

	/*
	**	Apply fudge factor to PIXSIZ, SPECLINE, and PROJSAMP
	*/

	PIXSIZ *= magn;
	SPECLINE /= magn;
	PROJSAMP /= magn;

	/*
	**	Check for Magellan MIDR format
	*/

	if (!MAP_PROJ)
		;
	else if (!strcmp(MAP_PROJ, "SINUSOIDAL"))
		get_mapping("sinusoidal", 16);
	else if (!strcmp(MAP_PROJ, "MERCATOR"))
		get_mapping("mercator", 8);
	else if (!strcmp(MAP_PROJ, "STEREOGRAPHIC"))
		get_mapping("stereographic", 2);
	else {
		(void)fprintf(stderr,
			"%s: warning: MAP_PROJ=%s not supported.\n",
			ifile, MAP_PROJ);
		if (OK(LAT_UC))
			ihput("ymin", &LAT_UC, 0);
		if (OK(LAT_LC))
			ihput("ymax", &LAT_LC, 0);
		if (OK(LON_CL))
			ihput("xmin", &LON_CL, 0);
		if (OK(LON_CR))
			ihput("xmax", &LON_CR, 0);
	}

	/*
	**	Process any dscale=n options
	*/

	if (OK(LOW_DN) && OK(HI_DN) && OK(LOW_DN)
			&& OK(LOW_REP) && OK(HI_REP)) {
		ihput("dscale", "n", 1);
		if (DN_UNITS && ! strcmp(DN_UNITS, "DECIBELS"))
			HI_DN += 1;
		(void)sprintf(buf, "%g", (HI_REP-LOW_REP)/(HI_DN-LOW_DN));
		(void)sscanf(buf, "%lf", &d);
		ihput("dfact", &d, 1);
		(void)sprintf(buf, "%g", LOW_REP - d * LOW_DN);
		(void)sscanf(buf, "%lf", &d);
		ihput("dzero", &d, 1);
	}

	/*
	**	update from any -m option
	*/
	
	if (mfile != NULL && *mfile != '-')
		if (fp = ihmerg(mfile))
			(void)ihbcl(fp);
		else
			gipserr();
	ihget("xnum", &xnum, 0);
	if (xnum > NS) {
		xnum = NS;
		ihput("xnum", &NS);
	}
	ihget("ynum", &ynum, 0);
	if (ynum > NL) {
		ynum = NL;
		ihput("ynum", &NL);
	}
	outlen = xnum * dlen;
	bin = alloc(1, inlen);
	dummy = alloc(1, inlen);
	bout = bin + NBB;
	putargs();

	/*
	**	Check axis limits for validity
	*/

	if (ihget("xmin", &xmin))
		xmin = 0;
	if (ihget("xmax", &xmax))
		xmax = xnum;
	if (xmin == xmax) {
		(void)ihdel("xmin");
		(void)ihdel("xmax");
	}
	if (ihget("ymin", &ymin))
		ymin = 0;
	if (ihget("ymax", &ymax))
		ymax = ynum;
	if (ymin == ymax) {
		(void)ihdel("ymin");
		(void)ihdel("ymax");
	}
	
	/*
	**	Write the output image header
	*/
	
	if ((fp = ihwrit(ofile, 1)) == NULL)
		perror(ofile), exit(1);
	else if (fp == stdout)
		ofile = "'stdout'";

	/*
	**	If seekable, set read pointer
	*/

	noseek = fseek(stdin, 0L, 1);

	if (ptype == 3 && (n = plane * (NLB + ynum)) > 0)
		if (! noseek)
			fseek(stdin, n * inlen, 1);
		else
			while (--n >= 0)
				fread(bin, inlen, 1, stdin);

	/*
	**	Copy the image
	*/

	for (sw = 0;;) {
		if (ptype != 2)
			sw = fread(bin, inlen, 1, stdin);
		else if (! noseek) {
			fseek(stdin, inlen * plane, 1);
			sw = fread(bin, inlen, 1, stdin);
			fseek(stdin, inlen * (NB - plane - 1), 1);
		} else for (n = 0; n < NB; n++)
			sw = fread(n == plane ? bin : dummy, inlen, 1, stdin);
		if (sw != 1)
			break;
		if (--NLB >= 0)
			continue;
		if (--ynum < 0)
			break;
		if (sproc)
			sproc(bout, bout, xnum);
		if (fwrite(bout, outlen, 1, fp) != 1)
			perror(ofile), exit(1);
	}

	/*
	**	Cleanup
	*/

	if (noseek)
		while (getchar() != EOF)
			;
	if (ferror(stdin))	
		perror(ifile), exit(1);
	(void)ihbcl(fp);
	if (Sflag)
		telltale(1);
	exit(0);
}

	/*
	**	Construct "mapping" variable from SPECLINE, PROJSAMP
	*/

static
get_mapping(title, projtype)
	char	*title;
	int	projtype;
{
	int	x, y;
	double	xmin, xmax, ymin, ymax;
	double	euler0, euler1, d, dl, lonlat[2];
	char	*msg = "%s: warning: no %s in VICAR2 label.\n";

	/*
	**	Need PIXSIZ and PROJ_LON
	*/

	if (!OK(PIXSIZ)) {
		(void)fprintf(stderr, msg, ifile, "PIXSIZ");
		return;
	} else if (!OK(PROJ_LON)) {
		(void)fprintf(stderr, msg, ifile, "PROJ_LON");
		return;
	}

	/*
	**	Scale PIXSIZ and RADIUS to meters
	*/

	if (! PIXSIZ_U)
		;
	else if (!strcmp(PIXSIZ_U, "KM/PIXEL"))
		PIXSIZ *= 1000;
	else if (strcmp(PIXSIZ_U, "M/PIXEL"))
		(void)fprintf(stderr, msg, ifile, "PIXSIZ_U");
	if (! OK(RADIUS))
		RADIUS = 6051000.0;
	else if (! RADIUS_U)
		;
	else if (! strcmp(RADIUS_U, "KM"))
		RADIUS *= 1000.0;
	else if (strcmp(RADIUS_U, "M"))
		(void)fprintf(stderr, msg, ifile, "RADIUS_U");

	/*
	**	If LON_UL and LAT_UL given for a SINUSOIDAL projection,
	**	use Magellan MIDR mapping definition (since PROJSAMP
	**	and SPECLINE are incorrect in some MIDR framelets).
	*/

	if (projtype == 16 && OK(LON_UL) && OK(LAT_UL) && ! Xflag) {
		d = RADIUS * DEG_RAD / PIXSIZ;
		if (OK(LON_LL) && OK(LAT_LL) && fabs(LAT_LL) < fabs(LAT_UL)) {
			dl = PROJ_LON - LON_LL;
			if (dl < -180)
				dl += 360;
			else if (dl > 180)
				dl -= 360;
			lonlat[0] = dl * d * cos(LAT_LL * DEG_RAD);
		} else {
			dl = PROJ_LON - LON_UL;
			if (dl < -180)
				dl += 360;
			else if (dl > 180)
				dl -= 360;
			lonlat[0] = dl * d * cos(LAT_UL * DEG_RAD);
		}
		lonlat[1] = LAT_UL * d;
		euler0 = PROJ_LON;
		euler1 = 0;
	} else if (!OK(PROJSAMP) || !OK(SPECLINE)) {
		(void)fprintf(stderr, msg, ifile, "PROJSAMP or SPECLINE");
		return;
	} else {
		if (projtype != 2) {
			euler0 = PROJ_LON;
			euler1 = 0;
		} else if (OK(LAT_UC) && OK(LAT_LC) && (LAT_UC + LAT_LC) < 0) {
			euler0 = 180-PROJ_LON;
			euler1 = -90;
		} else {
			euler0 = PROJ_LON;
			euler1 = 90;
		}
		lonlat[0] = PROJSAMP;
		lonlat[1] = SPECLINE;
	}

	while (euler0 > 180)
		euler0 -= 360;
	while (euler0 < -180)
		euler0 += 360;

	(void)sprintf(buf, GIPS_MAPPING_P,
		title, projtype, lonlat[0], lonlat[1], RADIUS/PIXSIZ,
		euler0, euler1, 0.0, 0.0, 0.0, 0);
	(void)ihput("mapping", buf);

	(void)ihput("xtitle", "Longitude");
	(void)ihput("ytitle", "Latitude");

	if (g_setup())
		gipserr();

	for (x = 0; x <= NS; x += NS)
		for (y = 0; y <= NL; y += NL) {
			if (g_unmap(x, y, lonlat))
				continue;
			if (fabs(lonlat[0]) < 1.e-10)
				lonlat[0] = 0;
			if (fabs(lonlat[1]) < 1.e-10)
				lonlat[1] = 0;
			if (lonlat[0] < 1.e20)
				(void)ihput(x ? "xmax" : "xmin", lonlat);
			if (lonlat[1] < 1.e20)
				(void)ihput(y ? "ymax" : "ymin", lonlat+1);
		}
}

	/*
	**	swap byte pairs
	*/

static	void
swap2(in, out, n)
	register unsigned short	*in, *out;
	register int	n;
{
	while (--n >= 0) {
		register unsigned short i = *in++;
		*out++ = (i << 8) | ((i >> 8) & 0xff);
	}
}

	/*
	**	swap 4-byte fields
	*/

static	void
swap4(in, out, n)
	register unsigned long *in, *out;
	register int	n;
{
	while (--n >= 0) {
		register unsigned long i = *in++;
#ifdef vax
		register unsigned long j;
		if (j = i & 0xff000000)
			j += 0x1000000;
		*out++ = (((i & 0xff0000) | j) >> 16) | ((i & 0xffff) << 16);
#else
		*out++ = ((i & 0xff) << 24) | ((i & 0xff00) << 8) |
			((i & 0xff0000) >> 8) | ((i & 0xff000000) >> 24);
#endif vax
	}
}

	/*
	**	convert VAX to IEEE or IEEE to VAX
	*/

static	void
swapv(in, out, n)
	register unsigned long *in, *out;
	register int	n;
{
	while (--n >= 0) {
		register unsigned long j, i = *in++;
#ifdef vax
		if (j = i & 0xff)
			j += 0x1;
		*out++ = (((i & 0xff0000) | j) << 8) | ((i >> 8) & 0xff00ff);
#else
		if (j = i & 0xff0000)
			j -= 0x10000;
		*out++ = (((i & 0xff) | j) << 8) | ((i >> 8) & 0xff00ff);
#endif vax
	}
}

	/*
	**	Header parse table
	*/
	
static	int	state_table[][4] = {
	/*	blank	equals	quote	other		state */
	{	0,	-1,	-1,	1	}, /* 0 blank */
	{	-1,	2,	-1,	1	}, /* 1 keyword */
	{	7,	-1,	3,	6	}, /* 2 assignment */
	{	4,	4,	5,	4	}, /* 3 starting quote */
	{	4,	4,	5,	4	}, /* 4 string value */
	{	0,	-1,	-1,	-1	}, /* 5 ending quote */
	{	7,	-1,	-1,	6	}, /* 6 numeric value */
	{	0,	-1,	-1,	1	}  /* 7 end numeric */
};

	/*
	**	Read the VICAR header
	*/

scan_header()
{
	int	nb;		/* input byte count */
	int	ch;		/* current character */
	int	ct;		/* character type code */
	int	nul;		/* =1 if NUL read */
	int	state;		/* state code */
	char	*p = buf;	/* pointer to buf */
	struct	keytab	*kp;	/* pointer to keys[] */
	
	/*
	**	Read the input header
	*/
	
	for (state = nb = ch = nul = 0; ch != EOF; nb++) {
		if ((keys[0].v_flg & FOUND) && nb >= LBLSIZE) {
			ct = 0;
			ch = EOF;
		} else if (nul) {
			(void)getchar();
			continue;
		} else switch (ch = getchar()) {
		case EOF:
			error("%s: unexpected EOF", ifile);
		case 0:
			nul++;
		case '\r':
		case '\n':
			continue;
		case ' ':
			ct = 0;
			break;
		case '=':
			ct = 1;
			break;
		case '\'':
			ct = 2;
			break;
		default:
			ct = 3;
			break;
		}
		switch (state = state_table[state][ct]) {
		case -1:
			error("VICAR label error in byte %d", (char *)nb); 
		case 0:			/* blank before keyword */
			break;
		case 1:			/* keyword */
			*p++ = ch;
			break;
		case 2:			/* assignment = */
			*p = 0;
			for (kp = keys; kp->v_name &&
					strcmp(buf, kp->v_name); kp++)
				;
			if (! kp->v_name)
				kp = NULL;
			p = buf;
			break;
		case 3:			/* starting quote */
			if (kp && ! (kp->v_flg & CHAR))
				error("non-numeric %s value", kp->v_name);
			break;
		case 4:			/* string value */
			*p++ = ch;
			break;
		case 5:			/* ending quote */
			if ((ch = getchar()) == '\'') {
				*p++ = ch;
				state = 4;
				nb++;
				break;
			}
			(void)ungetc(ch, stdin);
			if (kp) {
				if (Dflag)
					fprintf(stderr, "%-8.8s = %.*s\n",
						kp->v_name, p-buf, buf);
				*(char **)(kp->v_ptr) = strncpy(
					alloc(1, p-buf+1), buf, p-buf);
				kp->v_flg |= FOUND;
			}
			p = buf;
			break;
		case 6:			/* numeric value */
			if (kp && (kp->v_flg & CHAR))
				error("non-string %s value", kp->v_name);
			*p++ = ch;
			break;
		case 7:			/* end numeric */
			if (kp) {
				*p = 0;
				if (Dflag)
					fprintf(stderr, "%-8.8s = %s\n",
						kp->v_name, buf);
				if (sscanf(buf[0] ? buf : "0",
					(kp->v_flg & DOUBLE) ? "%lf" : "%d",
						kp->v_ptr) != 1)
					error("bad %s value", kp->v_name);
				kp->v_flg |= FOUND;
			}
			p = buf;
			break;
		}
	}
	
	/*
	**	Check for required parameters
	*/
	
	for (kp = keys; kp->v_name; kp++)
		if ((kp->v_flg & (REQD | FOUND)) == REQD)
			error("keyword not in header: %s", kp->v_name);
}
