#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihpho.c,v $	GIPS $Revision: 1.11 $ (MIT/CSR) $Date: 2004/11/02 20:02:11 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990,1991 Massachusetts Institute of Technology
**
**	Module:		ihpho.c		Creation Date:	03/27/90
**	Version:	$Revision: 1.11 $		Update:		$Date: 2004/11/02 20:02:11 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Create and edit a cellular GIPS image file
*/

#include	<stdio.h>
#include	<strings.h>
#include	<ctype.h>
#include	<math.h>
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/time.h>
#include	"gips.h"

#ifdef SVR4
#define	SYS_mmap
#include	<unistd.h>
#include	"raster.h"
#else
#include	<sys/syscall.h>
#include	<rasterfile.h>
#endif

#define		VICAR_SIZE	16384	/* VICAR tape block size */

#define	uchar	unsigned char
#define	ushort	unsigned short

static	char	Dflag;			/* =1 for debug printout */
static	char	Hflag;			/* =1 for halfword pixels */
static	char	Sflag;			/* =1 to list statistics */
static	char	Zflag;			/* =1 for truncated input pixels */
static	int	color;			/* current drawing color */
static	char	*mfile	= "-";		/* merged header file */
static	int	xnum;			/* x-dimension of new image */
static	int	ynum;			/* y-dimension of new image */

static	flagentry	flagtable[]	= {
	"D",	"1",		(char *) &Dflag,
	"H",	"1",		(char *) &Hflag,
	"S",	"1",		(char *) &Sflag,
	"Z",	"1",		(char *) &Zflag,
	"c",	"%d color",	(char *) &color,
	"m",	"-> file",	(char *) &mfile,
	"x",	"%d xnum",	(char *) &xnum,
	"y",	"%d ynum",	(char *) &ynum,
	NULL,	"image [script]",	NULL
};

static	char	*dfile;			/* image file */
static	char	*ifile;			/* input command file */
static	int	pixlen;			/* length of pixel 1 or 2 */
static	uchar	rgb[3][65536];		/* RGB values */
static	gimage	*gi;			/* photoproduct image pointer */
extern	int	_gcount;		/* gcell() remainder */
static	char	cmd[1024];		/* subcommand */
static	char	text[4096];		/* subcommand text */
static	int	xorg, yorg;		/* relative coordinate origins */
static	int	x_dimn, y_dimn;		/* size of last inserted image */
static	int	map_sw;			/* ==1 if no mapping */
static	int	origin;			/* text drawing origin */
static	double	angle;			/* drawing angle (degrees) */
static	FILE	*fin = stdin;		/* command stream */
static	FILE	*fdir;			/* directory stream pointer */
static	char	*ifile2;		/* secondary command stream */
static	time_t	tclock;			/* process start time */
#if __SWR
char *cuserid(char *dummy) { return getlogin(); }
#else
extern	char	*cuserid();		/* login user name */
#endif

static	open_image(), create_image(), read_colormap(), read_header();
static	insert_image(), insert_lonlat(), select_font(), select_input();
static	select_reset(), select_parm(), insert_text(), rotate_text();
static	insert_line(), insert_wedge(), insert_vwedge(), insert_circle();
static	insert_arc(), write_pnmfile(), write_vicarfile(), write_rasterfile();
static	copy_image(), get_coord(), scan_err(), write_ansi_label(), line();
static	rect(), wedge(), vwedge(), circle(), arc();

#define	SPIX(c)	if (pixlen==2) *(ushort *)(c) = color; else *(c) = color

main(argc, argv)

	int	argc;			/* number of arguments */
	char	**argv;			/* array of argument pointers */
{
	int	i;			/* scratch */
	char	*c;			/* scratch */

	/*
	**	Parse arguments, open command file
	*/

	saveargs(argv);
	tclock = time((time_t *)NULL);
	argc = gipsflags(argc, argv, flagtable);

	if (argc > 3)
		error("too many args", argv[3]);
	else if (argc < 2 || (dfile = argv[1]) == NULL)
		error("missing image file name", "");
	else if (argc == 2)
		ifile = "'stdin'";
	else if (freopen(ifile = argv[2], "r", stdin) == NULL)
		perror(ifile), exit(-1);
	if (Sflag)
		telltale(0);

	open_image();

	/*
	**	Initialize the colormap
	*/

	for (i = 0; i < (Hflag ? 65536 : 256); i++)
		rgb[0][i] = rgb[1][i] = rgb[2][i] = Hflag ? (i>>8) : i;

	/*
	**	Set default text mode
	*/

	(void)setfont("times.r.12");

	/*
	**	Loop over input command file
	*/

	for (c = text;; ) {
		if (fgets(c, sizeof(text)-(c-text)-1, fin) == NULL)
			if (c != text)
				error("bad contined line", ifile);
			else if (fin == stdin)
				break;
			else {
				(void)fclose(fin);
				fin = stdin;
				free(ifile);
				ifile = ifile2;
				c = text;
				continue;
			}

		if (Dflag && fin == stdin)
			(void)fprintf(stderr, "%s", c);
		c = c + strlen(c)-2;
		if (!strcmp(c, "\\\n"))
			continue;
		if (sscanf(c = text, "%s", cmd) != 1 || cmd[0] == '#')
			continue;

		if (!strcmp(cmd, "color") || !strcmp(cmd, "origin"))
			select_parm();
		else if (!strcmp(cmd, "font") || !strcmp(cmd, "italic"))
			select_font();
		else if (!strcmp(cmd, "bold") || !strcmp(cmd, "symbol"))
			select_font();
		else if (!strcmp(cmd, "angle"))
			select_angle();
		else if (!strcmp(cmd, "include"))
			select_input();
		else if (!strcmp(cmd, "reset"))
			select_reset();
		else if (!strcmp(cmd, "colormap"))
			read_colormap();
		else if (!strcmp(cmd, "header"))
			read_header();
		else if (!strcmp(cmd, "image"))
			insert_image(1);
		else if (!strcmp(cmd, "ximage"))
			insert_image(0);
		else if (!strcmp(cmd, "longitude"))
			insert_lonlat(0);
		else if (!strcmp(cmd, "latitude"))
			insert_lonlat(1);
		else if (!strcmp(cmd, "vicarfile"))
			write_vicarfile();
		else if (!strcmp(cmd, "rasterfile"))
			write_rasterfile();
		else if (!strcmp(cmd, "pnmfile"))
			write_pnmfile();
		else if (!strcmp(cmd, "text") || !strcmp(cmd, "label"))
			insert_text();
		else if (!strcmp(cmd, "line") || !strcmp(cmd, "rect"))
			insert_line();
		else if (!strcmp(cmd, "wedge"))
			insert_wedge();
		else if (!strcmp(cmd, "vwedge"))
			insert_vwedge();
		else if (!strcmp(cmd, "circle"))
			insert_circle();
		else if (!strcmp(cmd, "arc"))
			insert_arc();
		else
			error("bad command: %s", cmd);
	}

	(void)gclose(gi);

	(void)fclose(fin);
	if (fdir) {	
		(void)fprintf(fdir, "\n");
		(void)fclose(fdir);
	}
	if (Sflag)
		telltale(1);
	exit(0);
}

	/*
	**	Open or create photoproduct file
	*/

static
open_image()
{
	FILE	*fp;
	char	geom[1024], dtype;

	/*
	**	If no photoproduct file exists, create one
	*/

	if (access(dfile, W_OK))
		create_image();
	else if (strcmp(mfile, "-"))
		error("%s: cannot use -m option with existing file", dfile);

	/*
	**	Read the photoproduct header into memory
	*/

	if ((fp = ihread(dfile, GIPSPROFILE)) == NULL)
		gipserr();

	if (ihget("xnum", (char *)&xnum, 0) || ihget("ynum", (char *)&ynum, 0))
		gipserr();

	if (ihget("geometry", geom, sizeof(geom)-1))
		gipserr();

	if (sscanf(geom, "type=%c", &dtype) != 1)
		gipserr();

	switch (pixlen = ihdtype(dtype)) {
	case 0:
		error("%s: unrecognized dtype value", dfile);
		break;
	case 1:
		if (Hflag)
			error("%s: pixel size conflicts with -H flag", dfile);
		break;
	case 2:
		Hflag = 1;
		break;
	default:
		error("%s: unsupported dtype value", dfile);
		break;
	}

	/*
	**	Open the photoproduct file for read/write access
	*/

	if ((gi = gopen(fp, 128)) == NULL)
		gipserr();

	if ((fp = freopen(dfile, "r+", fp)) == NULL)
		perror(dfile), exit(-1);

	gwrite(gi, fp);

	x_dimn = xnum;
	y_dimn = ynum;
	map_sw = g_setup();
}

	/*
	**	Create a new cellular image
	*/

static
create_image()
{
	FILE	*fp;
	int	len;
	int	size, nx, ny, block;
	char	geometry[64];

	block = 8192;
#ifdef	SYS_mmap
	if (block % getpagesize())
		block = getpagesize();
#endif	SYS_mmap

	if (xnum <= 0)
		error("%s: -x value required for new file", dfile);
	if (ynum <= 0)
		error("%s: -y value required for new file", dfile);
	if ((fp = ihread(GINPROFILE, GIPSPROFILE)) == NULL)
		gipserr();
	(void)fclose(fp);
	if (strcmp(mfile, "-"))
		(void)ihbcl(ihmerg(mfile));
	pixlen = Hflag ? 2 : 1;
	(void)ihput("xnum", (char *)&xnum);
	(void)ihput("ynum", (char *)&ynum);
	(void)ihput("dtype", "G");
	(void)sprintf(geometry,
		"type=%c block=%d xdim=%d ydim=%d",
		Hflag ? 'u' : 'b', block, 128 / pixlen, block / 128);
	(void)ihput("geometry", geometry);
	putargs();
	if ((fp = ihwrit(dfile, 1)) == NULL)
		gipserr();
	(void)fflush(fp);
	nx = (xnum + (128/pixlen) - 1) & ~((128/pixlen) - 1);
	ny = (ynum + (block/128) - 1) & ~((block/128) - 1);
	size = nx * ny * pixlen;
	len = (ftell(fp) + block - 1) & ~(block - 1);
#ifdef	SYS_mmap
	if (color == 0)
		gclear(fp, len, size);
	else
#endif	SYS_mmap
	{
		len -= ftell(fp);
		while (--len >= 0)
			(void)putc(0, fp);
		len = nx * ny;
		while (--len >= 0) {
			if (Hflag)
				(void)putc(color >> 8, fp);
			(void)putc(color, fp);
		}
	}
	(void)ihbcl(fp);
	if (Dflag)
		(void)fprintf(stderr, "%s: created\n", dfile);
}

	/*
	**	Read an RGB colormap table
	*/

static
read_colormap()
{
	char		name[1024];
	int		r, g, b;
	register FILE	*fp;
	register char	*c;
	register int	i;

	scan_err(sscanf(text, "%*s %s", name), 1);

	if ((fp = fopen(name, "r")) == NULL)
		perror(name), exit(-1);

	for (i = 0; i < (Hflag ? 65536 : 256); ) {
		if (fgets(c = text, sizeof(text)-1, fp) == NULL)
			fprintf(stderr, "%s: unexpected EOF in line %d\n",
				name, (char *)(i+1)), exit(-1);
		while (*c && isspace(*c))
			c++;
		if (*c == '\0' || *c == '#')
			continue;
		if (sscanf(text, "%d %d %d", &r, &g, &b) != 3 ||
				((r|g|b) & 0xffffff00))
			fprintf(stderr, "bad r,g,b value in line %d: %s\n",
				text, (char *)(i+1)), exit(-1);
		rgb[0][i] = r;
		rgb[1][i] = g;
		rgb[2][i] = b;
		i++;
	}

	(void)fclose(fp);
	if (Dflag)
		(void)fprintf(stderr, "%s: colormap file read\n", name);
}

	/*
	**	Read image header
	*/

static
read_header()
{
	char	file[1024];
	FILE	*fp;

	scan_err(sscanf(text, "%*s %[^\n]", file), 1);

	if ((fp = ihread(file, GIPSPROFILE)) == NULL)
		gipserr();

	if (ihget("xnum", (char *)&x_dimn, 0) ||
			ihget("ynum", (char *)&y_dimn, 0))
		gipserr();

	(void)ihbcl(fp);

	map_sw = g_setup();
}

	/*
	**	Insert GIPS image into scratch image
	*/

static
insert_image(sw)
	register int	sw;
{
	char	dtype, file[1024], sx[32], sy[32];
	register FILE	*fp;
	register int	x, y, z, dz, x0, y0, n, pl;
	register char	*c;

	scan_err(sscanf(text, "%*s %s %s %[^\n]", sx, sy, file), 3);
	if (get_coord(sx, sy, &x_dimn, &y_dimn))
		return;

	for (c = file+strlen(file); isspace(*--c); *c = 0)
		;
	x0 = x_dimn + xorg;
	y0 = y_dimn + yorg;
	if ((fp = ihread(file, GIPSPROFILE)) == NULL)
		gipserr();
	if (ihget("xnum", (char *)&x_dimn, 0))
		gipserr();
	if (ihget("ynum", (char *)&y_dimn, 0))
		gipserr();
	if (ihget("dtype", &dtype, 1))
		gipserr();
	dz = ihdtype(dtype);
	pl = pixlen;
	if (!Zflag && dz != pixlen)
		fprintf(stderr, "%s: incompatible dtype: %c\n",
			file, dtype), exit(-1);
	for (y = 0; y < y_dimn; y++)
		for (x = 0; x < x_dimn; )
			if ((c = gcell(x+x0, y+y0, gi)) && _gcount > 0) {
				while (_gcount-- > 0 && x++ < x_dimn) {
					n = 0;
					switch (dz) {
					case 8:
						(void)getc(fp);
						(void)getc(fp);
						(void)getc(fp);
						(void)getc(fp);
					case 4:
						(void)getc(fp);
						(void)getc(fp);
					case 2:
						n = (getc(fp) & 255) << 8;
					case 1:
						n |= getc(fp) & 255;
						break;
					}
					if (pl == 2) {
						if (n || sw)
							*(short *)c = n;
						c += 2;
					} else {
						if (n || sw)
							*c = n;
						c++;
					}
				}
			} else {
				for (z = 0; z < dz; z++)
					(void)getc(fp);
				x++;
			}
	(void)ihbcl(fp);
	map_sw = g_setup();
	if (Dflag) {
		if (*(c = file) == '|') {
			while (isspace(*++c))
				;
			while (*c && !isspace(*c))
				c++;
			(void)strcpy(c, "...");
		}
		(void)fprintf(stderr, "%s: %dx%d inserted\n",
			file, x_dimn, y_dimn);
	}
}

	/*
	**	Insert latitude or longitude scale line
	*/
static
insert_lonlat(sw)
	int	sw;			/* sw=0 lon, =1 lat */
{
	double	lonlat[2];		/* longitude and latitude */
	double	xinc;			/* stepping increment */
	double	xmax;			/* limit of co-axis */
	double	xy[2];			/* pixel address */
	int	w;			/* line width */
	int	pix[2], pix1[2];	/* pixel address */
	int	n;			/* flag */

	scan_err(sscanf(text, "%*s %lf %lf %lf %lf %d",
		&lonlat[sw], &lonlat[1-sw], &xmax, &xinc, &w), 5);

	if (xinc == 0)
		return;

	for (n = 0; ; lonlat[1-sw] += xinc) {
		if (xinc > 0 && lonlat[1-sw] > xmax)
			break;
		if (xinc < 0 && lonlat[1-sw] < xmax)
			break;
		if (! map_sw && ! g_map(lonlat[0], lonlat[1], xy)) {
			pix[0] = xy[0] + 0.5;
			pix[1] = xy[1] + 0.5;
		} else if (! map_sw || gips(lonlat, pix, 1, GIPS_XY_IVAL)) {
			n = 0;
			continue;
		}
		if (pix[0] < 0 || pix[0] >= x_dimn)
			;
		else if (pix[1] < 0 || pix[1] >= y_dimn)
			;
		else if (n++ == 0) {
			pix1[0] = pix[0];
			pix1[1] = pix[1];
			continue;
		} else if (2 * abs(pix[0] - pix1[0]) > x_dimn)
			;
		else if (2 * abs(pix[1] - pix1[1]) > y_dimn)
			;
		else {
			line(gi, pix1[1]+yorg, pix1[0]+xorg,
				pix[1]+yorg, pix[0]+xorg, color, w);
			pix1[0] = pix[0];
			pix1[1] = pix[1];
			continue;
		}
		n = 0;
	}
}

	/*
	**	Select drawing font
	*/

static
select_font()
{
	char	name[1024];

	scan_err(sscanf(text, "%*s %s", name), 1);

	switch (cmd[0]) {
	case 'f':
		(void)setfont(name);
		break;
	case 'i':
		(void)setauxfont(name, 1);
		break;
	case 'b':
		(void)setauxfont(name, 2);
		break;
	case 's':
		(void)setauxfont(name, 3);
		break;
	}
}

	/*
	**	Select secondary input file
	*/

static
select_input()
{
	char	name[1024];

	scan_err(sscanf(text, "%*s %s", name), 1);

	if (fin != stdin)
		error("illegal nesting: %s", text);
	if ((fin = fopen(name, "r")) == NULL)
		perror(name), exit(-1);
	ifile2 = ifile;
	ifile = strcpy(alloc(strlen(name)+1, 1), name);
}

	/*
	**	Select drawing origin
	*/

static
select_reset()
{
	char	sx[32], sy[32];

	scan_err(sscanf(text, "%*s %s %s", sx, sy), 2);
	if (get_coord(sx, sy, &xorg, &yorg))
		return;
	if (yorg < 0 || yorg >= ynum)
		error("bad y origin: %d", (char *)yorg);
	if (xorg < 0 || xorg >= xnum)
		error("bad x origin: %d", (char *)xorg);
}

	/*
	**	Select drawing parameter
	*/

static
select_parm()
{
	int	parm;

	scan_err(sscanf(text, "%*s %d", &parm), 1);
	if (cmd[0] == 'c')
		(void)setink(color = parm);
	else
		(void)setorigin(origin = parm);
}

	/*
	**	Select text angle
	*/

select_angle()
{
	scan_err(sscanf(text, "%*s %lf", &angle), 1);
}

	/*
	**	Write text into scratch image
	*/

static
insert_text()
{
	int	y, x;
	char	str[1024], *c, sx[32], sy[32];

	scan_err(sscanf(text, "%*s %s %s %[^\n]", sx, sy, str), 3);
	if (get_coord(sx, sy, &x, &y))
		return;
	for (c = str+strlen(str); isspace(*--c); *c = 0)
		;
	if (angle != 0)
		rotate_text(y+yorg, x+xorg, str);
	else if (cmd[0] == 'l')
		(void)gilabel(gi, y+yorg, x+xorg, str);
	else
		(void)gimarker(gi, y+yorg, x+xorg, str);
}

static
rotate_text(y1,x1,str)
	int y1, x1;
	char *str;
{
	int dx1, dy1, dx2, dy2, ox1, oy1, ox2, oy2;
	int x, y, xx, yy, xxx, yyy, cos_a, sin_a, ink;
	double d;
	char *b, *c, *cbuf, *malloc();

	dx1 = comwidth(str)+1;
	dy1 = comheight(str)+2;
	c = cbuf = alloc(1, dx1*dy1);
	settype(IMAGE);
	ink = color;
	setink(255);
	setorigin(1);
	(void)memmarker(cbuf, dy1, dx1, 0, 0, str);
	setink(ink);
	setorigin(origin);
	switch (origin) {
	case 0:
		ox1 = -dx1/2;
		oy1 = -dy1/2;
		break;
	case 1:
		ox1 = 0;
		oy1 = 0;
		break;
	case 2:
		ox1 = -dx1/2;
		oy1 = 0;
		break;
	case 3:
		ox1 = -dx1;
		oy1 = 0;
		break;
	case 4:
		ox1 = 0;
		oy1 = -dy1/2;
		break;
	case 5:
		ox1 = -dx1;
		oy1 = -dy1/2;
		break;
	case 6:
		ox1 = 0;
		oy1 = -dy1;
		break;
	case 7:
		ox1 = -dx1/2;
		oy1 = -dy1;
		break;
	case 8:
		ox1 = -dx1;
		oy1 = -dy1;
		break;
	}
#define	x_rot(x,y)	(( cos_a * (x) + sin_a * (y)) >> 16)
#define	y_rot(x,y)	((-sin_a * (x) + cos_a * (y)) >> 16)
	/*
	**	Find size of rotated pixrect
	*/
	d = angle * atan(1.0) / 45.0;
	cos_a = 65536.0 * cos(d);
	sin_a = 65536.0 * sin(d);
	x = x_rot(dx1, 0);
	ox2 = MIN(0, x);
	dx2 = MAX(0, x);
	y = y_rot(dx1, 0);
	oy2 = MIN(0, y);
	dy2 = MAX(0, y);
	x = x_rot(0, dy1);
	ox2 = MIN(ox2, x);
	dx2 = MAX(dx2, x);
	y = y_rot(0, dy1);
	oy2 = MIN(oy2, y);
	dy2 = MAX(dy2, y);
	x = x_rot(dx1, dy1);
	ox2 = MIN(ox2, x);
	dx2 = MAX(dx2, x);
	y = y_rot(dx1, dy1);
	oy2 = MIN(oy2, y);
	dy2 = MAX(dy2, y);
	/*
	**	Get rotated text origin address
	*/
	x1 += x_rot(ox1 + dx1/2, oy1 + dy1/2) - (dx2 - ox2)/2;
	y1 += y_rot(ox1 + dx1/2, oy1 + dy1/2) - (dy2 - oy2)/2;
	/*
	**	Project the text, pixel by pixel
	*/
	xx = (-ox2) << 16;
	yy = (-oy2) << 16;
	for (y = 0; y < dy1; y++, xx += sin_a, yy += cos_a)
		for (x = 0, xxx = xx, yyy = yy; x < dx1; x++) {
			if (*c++ && (b = gcell((xxx>>16)+x1, (yyy>>16)+y1, gi)))
				SPIX(b);
			xxx += cos_a;
			yyy -= sin_a;
		}
	free(cbuf);
}

	/*
	**	Insert line or rectangle into scratch image
	*/

static
insert_line()
{
	int	y1, x1, y2, x2, w;
	char	sx1[32], sx2[32], sy1[32], sy2[32];

	scan_err(sscanf(text, "%*s %s %s %s %s %d", sx1, sy1, sx2, sy2, &w), 5);
	if (get_coord(sx1, sy1, &x1, &y1) || get_coord(sx2, sy2, &x2, &y2))
		return;
	if (cmd[0] == 'l')
		line(gi, y1+yorg, x1+xorg, y2+yorg, x2+xorg, color, w);
	else
		rect(gi, y1+yorg, x1+xorg, y2+yorg, x2+xorg, color, w);
}

	/*
	**	Insert horizontal wedge
	*/

static
insert_wedge()
{
	int	y1, x1, y2, x2, c1, c2, mask = 0, fill = 0;
	char	sx1[32], sx2[32], sy1[32], sy2[32];

	scan_err(sscanf(text, "%*s %s %s %s %s %d %d %d %d",
		sx1, sy1, sx2, sy2, &c1, &c2, &mask, &fill), -6);
	if (get_coord(sx1, sy1, &x1, &y1) || get_coord(sx2, sy2, &x2, &y2))
		return;
	wedge(gi, y1+yorg, x1+xorg, y2+yorg, x2+xorg, c1, c2, mask, fill);
}

	/*
	**	Insert vertical wedge
	*/

static
insert_vwedge()
{
	int	y1, x1, y2, x2, c1, c2, mask = 0, fill = 0;
	char	sx1[32], sx2[32], sy1[32], sy2[32];

	scan_err(sscanf(text, "%*s %s %s %s %s %d %d %d %d",
		sx1, sy1, sx2, sy2, &c1, &c2, &mask, &fill), -6);
	if (get_coord(sx1, sy1, &x1, &y1) || get_coord(sx2, sy2, &x2, &y2))
		return;
	vwedge(gi, y1+yorg, x1+xorg, y2+yorg, x2+xorg, c1, c2, mask, fill);
}

	/*
	**	Write circle into scratch image
	*/

static
insert_circle()
{
	int	y, x, r, w;
	char	sx[32], sy[32];

	scan_err(sscanf(text, "%*s %s %s %d %d", sx, sy, &r, &w), 4);
	if (get_coord(sx, sy, &x, &y))
		return;
	circle(gi, y+yorg, x+xorg, r, color, w);
}

	/*
	**	Write arc into scratch image
	*/

static
insert_arc()
{
	int	y, x, r, w;
	char	sx[32], sy[32];
	double	a1, a2;

	scan_err(sscanf(text, "%*s %s %s %d %lf %lf %d",
			sx, sy, &r, &a1, &a2, &w), 6);
	if (get_coord(sx, sy, &x, &y))
		return;
	arc(gi, y+yorg, x+xorg, r, a1, a2, color, w);
}

	/*
	**	Write SUN rasterfile
	*/

static
write_rasterfile()
{
	char			name[1024];	/* output file name */
	FILE			*fp;		/* output file stream */
	int			fact;		/* reduction factor */
	int			depth;		/* image depth */
	char			*ofile;		/* output file/pipe */
	struct rasterfile	r;		/* header */

	scan_err(sscanf(text, "%*s %d %d %[^\n]", &fact, &depth, name), 3);
	if (fact < 1)
		error("bad reduction factor: %d", (char *)fact);
	if (depth != 1 && depth != 3)
		error("bad depth: %d", (char *)depth);
	for (ofile = name+strlen(name); isspace(*--ofile); *ofile = 0)
		;

	/*
	**	Open output file
	*/

	if (name[0] == '|')
		fp = popen(ofile = &name[1], "w");
	else
		fp = fopen(ofile = name, "w");

	if (fp == NULL)
		perror(ofile), exit(-1);

	/*
	**	Write rasterfile header
	*/

	r.ras_magic = RAS_MAGIC;
	r.ras_width = xnum / fact;
	r.ras_height = ynum / fact;
	r.ras_depth = depth == 3 ? 24 : 8;
	r.ras_length = (r.ras_width + (r.ras_width & 1))
		* r.ras_height * depth;
	r.ras_type = RT_STANDARD;

	/*
	**	Write the image
	*/

	if (depth == 1 && pixlen == 1) {
		char	lin[256];
		int	i;

		r.ras_maptype = RMT_EQUAL_RGB;
		r.ras_maplength = 3 * 256;
		if (fwrite((char *)&r, sizeof(r), 1, fp) != 1)
			perror(ofile), exit(-1);
		for (i = 0; i < 3; i++)
			if (fwrite((char *)(rgb[i]), 256, 1, fp) != 1)
				perror(ofile), exit(-1);
		(void)fflush(fp);
		for (i = 0; i < 256; i++)
			lin[i] = i;
		copy_image(lin, fact, fp);
	} else {
		r.ras_maptype = RMT_NONE;
		r.ras_maplength = 0;
		if (fwrite((char *)&r, sizeof(r), 1, fp) != 1)
			perror(ofile), exit(-1);
		(void)fflush(fp);
		copy_image((char *)(rgb[0]), fact, fp);
		if (depth == 3) {
			copy_image((char *)(rgb[1]), fact, fp);
			copy_image((char *)(rgb[2]), fact, fp);
		}
	}

	/*
	**	Cleanup
	*/

	if (name[0] == '|')
		(void)pclose(fp);
	else
		(void)fclose(fp);
	if (Dflag)
		(void)fprintf(stderr, "%s: written\n", ofile);
}

	/*
	**	Write portable anymap file
	*/

static
write_pnmfile()
{
	char			name[1024];	/* output file name */
	FILE			*fp;		/* output file stream */
	int			fact;		/* reduction factor */
	int			depth;		/* image depth */
	char			*ofile;		/* output file/pipe */

	scan_err(sscanf(text, "%*s %d %d %[^\n]", &fact, &depth, name), 3);
	if (fact < 1)
		error("bad reduction factor: %d", (char *)fact);
	if (depth != 1 && depth != 3)
		error("bad depth: %d", (char *)depth);
	for (ofile = name+strlen(name); isspace(*--ofile); *ofile = 0)
		;

	/*
	**	Open output file
	*/

	if (name[0] == '|')
		fp = popen(ofile = &name[1], "w");
	else
		fp = fopen(ofile = name, "w");

	if (fp == NULL)
		perror(ofile), exit(-1);

	/*
	**	Write the header
	*/

	fprintf(fp, "P%d\n# %s\n%d %d\n255\n", 
		(depth == 3) ? 6 : 5, dfile, xnum/fact, ynum/fact);

	/*
	**	Write the image
	*/

	if (depth == 1 && pixlen == 1) {
		char	lin[256];
		int	i;
		for (i = 0; i < 256; i++)
			lin[i] = i;
		copy_image(lin, fact, fp);
	} else {
		copy_image((char *)(rgb[0]), fact, fp);
		if (depth == 3) {
			copy_image((char *)(rgb[1]), fact, fp);
			copy_image((char *)(rgb[2]), fact, fp);
		}
	}

	/*
	**	Cleanup
	*/

	if (name[0] == '|')
		(void)pclose(fp);
	else
		(void)fclose(fp);
	if (Dflag)
		(void)fprintf(stderr, "%s: written\n", ofile);
}

	/*
	**	Write out VICAR header and image
	*/

static
write_vicarfile()
{
	char		ofile[1024];		/* output file name */
	char		volid[64];		/* ANSI tape volume serial */
	char		name[64];		/* ANSI tape file name */
	char		cdepth[64];		/* image depth/plane */
	int		istape = 0;		/* =1 if tape output */
	int		lrecl;			/* output record length */
	int		depth;			/* output image depth */
	register int	x;			/* scratch */

	scan_err(sscanf(text, "%*s %s %s %s %s",
		cdepth, ofile, volid, name), 4);

	if (strcmp(cdepth, "1") == 0)
		depth = 1;
	else if (strcmp(cdepth, "3") == 0)
		depth = 3;
	else if (strcmp(cdepth, "red") == 0)
		depth = -1;
	else if (strcmp(cdepth, "green") == 0)
		depth = -2;
	else if (strcmp(cdepth, "blue") == 0)
		depth = -3;
	else
		error("bad depth: %c", cdepth);

	/*
	**	Check for tape output
	*/

	if (strncmp(ofile, "/dev/", 5) == 0)
		istape = 1;

	/*
	**	Open output file
	*/

	if (strcmp(ofile, "-") && freopen(ofile, "w", stdout) == NULL)
		perror(ofile), exit(-1);

	/*
	**	Determine if it is a tape
	*/

	if (istape) {
		write_ansi_label(ofile, volid, name, "HDR");
		setbuffer(stdout, alloc(VICAR_SIZE, 1), VICAR_SIZE);
	}

	/*
	**	Write VICAR header
	*/

	lrecl = xnum + (xnum & 1);
	x = printf("LBLSIZE=%d ", lrecl);
	x += printf("FORMAT='BYTE' ");
	x += printf("BUFSIZ=%d ", VICAR_SIZE);
	x += printf("RECSIZE=%d ", lrecl);
	x += printf("NL=%d ", ynum);
	x += printf("NS=%d ", xnum);
	x += printf("NB=%d ", (depth == 3) ? 3 : 1);
	x += printf("ORG='%s' ", (depth == 3) ? "BIL" : "BSQ");
	x += printf("TASK='MGMPHO' ");
	x += printf("TYPE='IMAGE' ");
	x += printf("USER='%s' ", cuserid((char *)NULL));
	x += printf("DAT_TIM='%24.24s' ", ctime(&tclock));

	while (x++ < lrecl)
		(void)putchar(0);

	/*
	**	Write the image
	*/

	copy_image((depth > 0) ? (char *)(rgb[0]) : (char *)(rgb[-1-depth]),
		(depth == 3) ? -1 : 1, stdout);
	
	/*
	**	Cleanup
	*/

	if (istape)
		write_ansi_label(ofile, volid, name, "EOF");
	(void)fclose(stdout);
	if (Dflag)
		(void)fprintf(stderr, "%s: written\n", ofile);
}

	/*
	**	Copy scratch image
	*/

static
copy_image(map, fact, fp)
	register char	*map;		/* colormap array */
	register int	fact;		/* resample factor */
	register FILE	*fp;		/* output stream */
{
	register int	x, y, nx, dlen;	/* scratch */
	register char	*c;		/* scratch */
	register uchar	*outbuf, *o;	/* output buffer */
	register uchar	*outlim;	/* end of output buffer */
	register ushort	*outbuf2, *o2;	/* output buffer */
	register ushort	*outlim2;	/* end of output buffer */

	outbuf = NULL;
	outbuf2 = NULL;

	if (fact != -1) {
		dlen = pixlen;
		nx = (xnum/fact)*fact;
	} else if (Hflag) {
		fact = 1;
		nx = xnum;
		dlen = sizeof(ushort);
		outbuf2 = (ushort *)alloc(xnum, dlen);
		outlim2 = outbuf2 + xnum;
	} else {
		fact = 1;
		nx = xnum;
		dlen = sizeof(uchar);
		outbuf = (uchar *)alloc(xnum, dlen);
		outlim = outbuf + xnum;
	}

	for (y = 0; y < ynum; y += fact) {
		for (o = outbuf, o2 = outbuf2, x = 0; x < nx; )
			if ((c = gcell(x, y, gi)) && _gcount > 0) {
				for ( ;_gcount-- > 0 && x < nx; c += dlen)
					if (outbuf)
						*o++ = *(uchar *)c, x++;
					else if (outbuf2)
						*o2++ = *(ushort *)c, x++;
					else if ((x++ % fact) == 0)
						(void)putc(dlen == 2
							? map[*(ushort *)c]
							: map[*(uchar *)c],
							fp);
			} else if (! outbuf && ! outbuf2 && (x++ % fact) == 0)
				(void)putc(0,fp);
		if (outbuf) {
			for (x = 0; x <= 131072 ; x += 65536)
				for (o = outbuf; o < outlim; o++)
					(void)putc(map[x + *o], fp);
		} else if (outbuf2) {
			for (x = 0; x <= 131072 ; x += 65536)
				for (o2 = outbuf2; o2 < outlim2; o2++)
					(void)putc(map[x + *o2], fp);
		} else if ((xnum/fact) & 1)
			(void)putc(0,fp);
	}
}

	/*
	**	Decode x,y coordinate pair
	*/

static
get_coord(sx, sy, px, py)
	char	*sx, *sy;
	int	*px, *py;
{
	double	dx, dy, xy[2];

	if (sscanf(sx, "%lf", &dx) != 1)
		error("bad x-coordinate value: %s", text);
	else if (sscanf(sy, "%lf", &dy) != 1)
		error("bad y-coordinate value: %s", text);
	else if (map_sw || ! index(sx, '.') || ! index(sy, '.'))
		*px = dx, *py = dy;
	else if (g_map(dx, dy, xy)) {
		fprintf(stderr, "coordinates outside map: %s\n", text);
		return 1;
	} else
		*px = xy[0], *py = xy[1];
	return 0;
}

	/*
	**	Error in scanf()
	*/

static
scan_err(rc, narg)
	int	rc, narg;
{
	if (rc == EOF)
		error("bad args: %s", text);
	else if ((narg > 0 && rc != narg) || (narg < 0 && rc < -narg))
		fprintf(stderr, "bad #%d arg: %s\n", (char *)(rc+1), text),
		exit(-1);
}

	/*
	**	Write ANSI tape labels
	*/

static
write_ansi_label(tape, volid, name, type)
	char		*tape;		/* tape name */
	char		*volid;		/* tape volume */
	char		*name;		/* file name */
	char		*type;		/* HDR or EOF */
{
	int		nblk;		/* block count */
	int		n;		/* scratch */
	struct	tm	*tm;		/* process start time */
static	int		nseq;		/* tape file sequence number */

	tm = localtime(&tclock);

	if (type[0] == 'H') {
		if (nseq++ == 0) {
			n = printf("VOL1%-6.6s %-26.26s%-17.17s%-25.25s%1d",
				volid, "", "", "", 3);
			while (n++ < 80) (void)putchar(' ');
			(void)fflush(stdout);
		}
		nblk = 0;
	} else {
		for (nblk = ftell(stdout); nblk % VICAR_SIZE; nblk++)
			(void)putchar('\0');
		if (freopen(tape, "w", stdout) == NULL)
			perror(tape), exit(-1);
	}

	n = printf("%-3.3s1%-17.17s%-6.6s%04d%04d%04d%02d %05d %05d %06d%",
		type, name, volid, 1, nseq, 1, 1, 
		1000 * tm->tm_year + tm->tm_yday + 1, 99999, nblk/VICAR_SIZE);
	while (n++ < 80) (void)putchar(' ');
	(void)fflush(stdout);

	n = printf("%-3.3s2%1s%05d%05d%-35.35s00",
		type, "F", VICAR_SIZE, VICAR_SIZE, "");
	while (n++ < 80) (void)putchar(' ');

	if (freopen(tape, "w", stdout) == NULL)
		perror(tape), exit(-1);
}

	/*
	**	Draw line
	*/

static
line(gim, y1, x1, y2, x2, color, width)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* starting point */
	register int	y2, x2;		/* ending point */
	register int	color;		/* fill color */
	register int	width;		/* line width */
{
	register int	x, y, w, i;	/* scratch */
	register int	xmin, xmax;	/* actual line limits */
	register int	ymin, ymax;	/* actual line limits */
	register char	*c;		/* pointer to image pixel */

	w = width/2;

	if (x1 <= x2)
		xmin = x1-w, xmax = x2+w;
	else
		xmin = x1+w, xmax = x2-w;
	if (y1 <= y2)
		ymin = y1-w, ymax = y2+w;
	else
		ymin = y1+w, ymax = y2-w;

	if (abs(x2-x1) > abs(y2-y1))
		for (x = xmin; ; x += (x2 > x1) ? 1 : -1 ) {
			y = (y2*(x-x1)+y1*(x2-x))/(x2-x1)-w;
			for (i = 0; i < width; i++)
				if (c = gcell(x, y+i, gim))
					SPIX(c);
			if (x == xmax)
				break;
		}
	else if (y1 == y2) {
		x = x1-w;
		for (i = 0; i < width; i++)
			if (c = gcell(x+i, y1, gim))
				SPIX(c);
	} else
		for (y = ymin; ; y += (y2 > y1) ? 1 : -1) {
			x = (x2*(y-y1)+x1*(y2-y))/(y2-y1)-w;
			for (i = 0; i < width; i++)
				if (c = gcell(x+i, y, gim))
					SPIX(c);
			if (y == ymax)
				break;
		}
}

	/*
	**	Draw rectangle
	*/

static
rect(gim, y1, x1, y2, x2, color, width)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* top left */
	register int	y2, x2;		/* bottom right */
	register int	color;		/* fill color */
	register int	width;		/* =0 to fill, otherwise =width */
{
	register char	*c;		/* pointer to image pixel */
	register int	x, y, w;	/* scratch */

	if (width > 0) {
		w = width/2;
		for (y = y1-w; y <= y2+w; y++)
			for (x = -w; x < width-w; x++) {
				if (c = gcell(x1+x, y, gim))
					SPIX(c);
				if (c = gcell(x2+x, y, gim))
					SPIX(c);
			}
		for (x = x1-w; x <= x2+w; x++)
			for (y = -w; y < width-w; y++) {
				if (c = gcell(x, y1+y, gim))
					SPIX(c);
				if (c = gcell(x, y2+y, gim))
					SPIX(c);
			}
	} else if (width == 0) {
		for (y = y1; y <= y2; y++)
			for (x = x1; x <= x2; )
				if (!(c = gcell(x, y, gim)) || _gcount <= 0)
					x++;
				else for ( ;--_gcount >= 0 && x++ <= x2;
							c += pixlen)
					SPIX(c);
	}
}

	/*
	**	Draw horizontal wedge
	*/

static
wedge(gim, y1, x1, y2, x2, c1, c2, mask, fill)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* top left */
	register int	y2, x2;		/* bottom right */
	register int	c1, c2;		/* color range */
	register int	mask, fill;	/* for masked wedge */
{
	register char	*c;		/* pointer to image pixel */
	register int	x, y;		/* scratch */
	register int	color;		/* color value */

	if (x1 == x2)
		return;

	for (x = x1; x <= x2; x++) {
		color = (c1 * (x2 - x) + c2 * (x - x1)) / (x2 - x1);
		if (mask)
			color = (color & ~mask) | (fill & mask);
		for (y = y1; y <= y2; y++)
			if (c = gcell(x, y, gim))
				SPIX(c);
	}
}

	/*
	**	Draw vertical wedge
	*/

static
vwedge(gim, y1, x1, y2, x2, c1, c2, mask, fill)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* top left */
	register int	y2, x2;		/* bottom right */
	register int	c1, c2;		/* color range */
	register int	mask, fill;	/* for masked wedge */
{
	register char	*c;		/* pointer to image pixel */
	register int	x, y;		/* scratch */
	register int	color;		/* color value */

	if (y1 == y2)
		return;

	for (y = y1; y <= y2; y++) {
		color = (c1 * (y2 - y) + c2 * (y - y1)) / (y2 - y1);
		if (mask)
			color = (color & ~mask) | (fill & mask);
		for (x = x1; x <= x2; x++)
			if (c = gcell(x, y, gim))
				SPIX(c);
	}
}

	/*
	**	Draw circle
	*/

static
circle(gim, y1, x1, r, color, width)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* center point */
	register int	r;		/* radius */
	register int	color;		/* fill color */
	register int	width;		/* =0 to fill, otherwise =width */
{
	register char	*c;		/* pointer to image pixel */
	register int	x, y;		/* scratch */
	register int	xa, xb;		/* arc widths */

	for (y = 0; y <= r; y++) {
		xa = sqrt((double)((r-y)*(r+y)));
		if (width > 0 && y < (r-width))
			xb = sqrt((double)((r-width-y)*(r-width+y)));
		else
			xb = 0;
		for (x = xb; x <= xa; )
			if (!(c = gcell(x1+x, y1-y, gim)) || _gcount <= 0)
				x++;
			else for ( ;--_gcount >= 0 && x++ <= xa; c += pixlen)
				SPIX(c);
		for (x = xb; x <= xa; )
			if (!(c = gcell(x1+x, y1+y, gim)) || _gcount <= 0)
				x++;
			else for ( ;--_gcount >= 0 && x++ <= xa; c += pixlen)
				SPIX(c);
		xb = x1 - xb;
		xa = x1 - xa;
		for (x = xa; x <= xb; )
			if (!(c = gcell(x, y1-y, gim)) || _gcount <= 0)
				x++;
			else for ( ;--_gcount >= 0 && x++ <= xb; c += pixlen)
				SPIX(c);
		for (x = xa; x <= xb; )
			if (!(c = gcell(x, y1+y, gim)) || _gcount <= 0)
				x++;
			else for ( ;--_gcount >= 0 && x++ <= xb; c += pixlen)
				SPIX(c);
	}
}

	/*
	**	Draw finite-width arc
	*/

static
arc(gim, y1, x1, r, ang1, ang2, color, width)
	register gimage	*gim;		/* cell image descriptor */
	register int	y1, x1;		/* center point */
	register int	r;		/* radius */
	double		ang1, ang2;	/* arc measured from center */
	register int	color;		/* fill color */
	register int	width;		/* =0 to fill, otherwise =width */
{
	register char	*c;		/* pointer to image pixel */
	double		d, yy;		/* scratch */
	double		t1, t2;		/* tangents */
	register int	i, sw, x, y;	/* scratch */
	register int	xa, xb;		/* arc widths */
	int		q[4];		/* quadrant flags */

#define	F_TEST_ANG1	0x01		/* test the first angle */
#define	F_BAD_ANG1	0x02		/* not within this quadrant */
#define	F_TEST_ANG2	0x04		/* test the second angle */
#define	F_BAD_ANG2	0x08		/* not within this quadrant */
#define F_NEG_X		0x10		/* x-offset is negative */
#define F_NEG_Y		0x20		/* y-offset is negative */

	if (ang1 < 0) {
		ang1 = 360 + ang1;
		if (ang2 >= 0) {
			arc(gim, y1, x1, r, 0.0, ang2, color, width);
			ang2 = 360;
		}
	}
	if (ang2 < 0)
		ang2 = 360 + ang2;
	if (ang1 > ang2)
		return;

	q[0] = F_NEG_Y;
	q[1] = 0;
	q[2] = F_NEG_X;
	q[3] = F_NEG_X | F_NEG_Y;

	d = atan(1.0)/45;

	for (i = 0; i < 4; i++) {
		if (ang1 <= 90*i)
			;
		else if (ang1 < 90*(i+1)) {
			q[i] |= F_TEST_ANG1;
			if (ang1 == 0 || ang1 == 180)
				t1 = 1.e20;
			else
				t1 = tan(d*fabs(ang1-(i > 1 ? 270 : 90)));
		} else
			q[i] |= F_BAD_ANG1;
		if (ang2 > 90*(i+1))
			;
		else if (ang2 >= 90*i) {
			q[i] |= F_TEST_ANG2;
			if (ang2 == 0 || ang2 == 180)
				t2 = 1.e20;
			else
				t2 = tan(d*fabs(ang2-(i > 1 ? 270 : 90)));
		} else
			q[i] |= F_BAD_ANG2;
	}
 
	for (y = 0; y <= r; y++) {
		xa = sqrt((double)((r-y)*(r+y)));
		if (width > 0 && y < (r-width))
			xb = sqrt((double)((r-width-y)*(r-width+y)));
		else
			xb = 0;
		yy = y;
		for (i = 0; i < 4; i++) {
			if ((sw = q[i]) & (F_BAD_ANG1 | F_BAD_ANG2))
				continue;
			for (x = xb; x <= xa; x++) {
				if (sw & (F_TEST_ANG1 | F_TEST_ANG2))
					d = x ? yy / x : 1.e20;
				if ((sw & F_TEST_ANG1) &&
					((i & 1) ? (d < t1) : (d > t1)))
					continue;
				if ((sw & F_TEST_ANG2) &&
					((i & 1) ? (d > t2) : (d < t2)))
					continue;
				if (c = gcell((sw & F_NEG_X) ? x1-x : x1+x,
						(sw & F_NEG_Y) ? y1-y : y1+y,
						gim))
					SPIX(c);
			}
		}
	}
}

#ifdef SYS_mmap

	/*
	**	If SunOS memory mapping available, replace all gcell()`
	**	calls with virtual memory references
	*/

#include	<sys/mman.h>

#define	GIPS_GEOM	"type=%c block=%d xdim=%d ydim=%d"

typedef	struct	g_mem	{
	char	*gm_ptr;	/* virtual origin of image */
	int	gm_size;	/* length of image in bytes */
	int	gm_off;		/* length of image header */
	int	gm_xmask;	/* number of x pixels in cell - 1 */
	int	gm_ymask;	/* number of y pixels in cell - 1 */
	int	gm_xcell;	/* number of x cells */
	int	gm_shft;	/* block length shift */
	int	gm_xshft;	/* x pixel address shift */
	int	gm_yshft;	/* y pixel address shift */
	int	gm_dshft;	/* pixel length shift */
} GMEM;

int	_gcount;		/* available pixel count */

shift(n)
{
	int	i = 0;

	while (n >>= 1)
		i++;
	return i;
}

/*ARGSUSED*/
gimage	*
gopen(fp, n)
	FILE	*fp;
	int	n;
{
	GMEM	*gm;
	int	nx, ny, block;
	char	geom[128], dtype;

	if (ihget("geometry", geom, sizeof(geom)-1))
		return NULL;
	if (sscanf(geom, GIPS_GEOM, &dtype, &block, &nx, &ny) != 4)
		return NULL;
	if (dtype != "bu"[Hflag])
		fprintf(stderr, "%s: bad dtype value: %c\n",
			dfile, dtype), exit(-1);
	if (block % getpagesize())
		fprintf(stderr, "Memory page size %d != %d\n",
			(char *)getpagesize(), (char *)block), exit(-1);
	gm = (GMEM *)alloc(1, sizeof(GMEM));
	gm->gm_shft = shift(block);
	gm->gm_xmask = nx - 1;
	gm->gm_ymask = ny - 1;
	gm->gm_xshft = shift(nx);
	gm->gm_yshft = shift(ny);
	gm->gm_dshft = shift(pixlen);
	gm->gm_xcell = (xnum + gm->gm_xmask) >> gm->gm_xshft;
	nx = (xnum + gm->gm_xmask) & ~gm->gm_xmask;
	ny = (ynum + gm->gm_ymask) & ~gm->gm_ymask;
	gm->gm_size = nx * ny * ihdtype(dtype);
	gm->gm_off = (ftell(fp) + block - 1) & ~(block - 1);
	return (gimage *)gm;
}

gwrite(gi, fp)
	gimage	*gi;
	FILE	*fp;
{
	GMEM	*gm = (GMEM *)gi;

	gm->gm_ptr = mmap((char *)0, gm->gm_size, PROT_READ|PROT_WRITE,
		MAP_SHARED, fileno(fp), (off_t)(gm->gm_off));
	if ((int)gm->gm_ptr == -1)
		perror(dfile), exit(-1);
	madvise((caddr_t)gm->gm_ptr, gm->gm_size, MADV_RANDOM);
#ifdef DEBUG
	(void)fprintf(stderr, "%s: memory mapped at 0x%08x-0x%08x\n",
		dfile, gm->gm_ptr, gm->gm_ptr + gm->gm_size - 1);
#endif DEBUG
}

gclear(fp, off, len)
	FILE		*fp;
	int		off, len;
{
	register uchar	*c;
	register int	i;

	if (ftruncate(fileno(fp), (off_t)(off + len)) == EOF)
		perror(dfile), exit(-1);
	if (! color)
		return;
	if (freopen(dfile, "r+", fp) == NULL)
		perror(dfile), exit(-1);
	c = (uchar *)mmap((char *)0, len, PROT_WRITE, MAP_SHARED,
		fileno(fp), (off_t)off);
	if ((int)c == -1)
		perror(dfile), exit(-1);
	madvise((caddr_t)c, len, MADV_SEQUENTIAL);
	for (i = 0; i < len; i += pixlen)
		if (Hflag)
			*(ushort *)(c+i) = color;
		else
			c[i] = color;
	(void)munmap((caddr_t)c, len);
}

char	*
gcell(x, y, gi)
	register int	x, y;
	gimage		*gi;
{
	register GMEM	*gm = (GMEM *)gi;
	register char	*c;

	_gcount = 0;
	if (gm == NULL || y < 0 || y >= ynum || x < 0 || x >= xnum)
		return NULL;
	c = gm->gm_ptr + (((x >> gm->gm_xshft)
		+ gm->gm_xcell * (y >> gm->gm_yshft)) << gm->gm_shft);
	x &= gm->gm_xmask;
	_gcount = gm->gm_xmask + 1 - x;
	return c + ((x + ((y & gm->gm_ymask) << gm->gm_xshft)) << gm->gm_dshft);
}

gclose(gi)
	gimage	*gi;
{
	GMEM	*gm = (GMEM *)gi;

	if (gm == NULL)
		return EOF;
	free((char *)gm);
	return munmap((caddr_t)gm->gm_ptr, gm->gm_size);
}
#endif	SYS_mmap
