#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihrot.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 1998/05/13 01:01:56 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihrot.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 1998/05/13 01:01:56 $
**	Components:	main()		UNIX command
**
**	Function:	Rotate, invert, or reverse a General Image
*/

#include	<string.h>    /* for strlen */
#include        <stdlib.h>   /* for malloc */
#include	<stdio.h>
#include	"gips.h"
	/*
	 *	maximum number of concurrently open files
	 */
#define		MAXFIL		16
	/*
	 *	parameters affected by gipsflags()
	 */
static	char	*ddir	= "/tmp";	/* spooling directory */
static	char	*ifile	= "-";		/* input general image */
static	char	*hfile	= "-";		/* replacement general image header */
static	char	*mfile	= "-";		/* merged general image header */
static	char	*ofile	= "-";		/* output general image */
static	flag	rflag	;		/* reverse raster pixels */
static	flag	iflag	;		/* invert the image */
static	flag	nflag	;		/* don't turn the image */
static	flag	Sflag	;		/* generate memory statistics */
static	int	block	= NIMAGE;	/* max in-core buffer space */
	/*
	 *	argument descriptor table for gipsflags()
	 */
static	flagentry	flagtable[] = {
	"r",	"1",		&rflag,
	"i",	"1",		&iflag,
	"n",	"1",		&nflag,
	"S",	"1",		&Sflag,
	"d",	"-> dir",	(flag *) &ddir,
	"h",	"-> file",	(flag *) &hfile,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	"",	"%d bytes",	(flag *) &block,
	NULL,	"[file]",	NULL
};
	/*
	 *	header fields that must be swapped after a rotation
	 */
static	char	type[2] = {'d','d'};	/* variable type (always d) */
static	char	scale[2] = {'s', 's'};	/* Type of axis mapping */
static	char	title[2][BUFLEN];	/* Description of variable */
static	char	font[2][FONTLEN];	/* Font for writing title */
static	int	num[2];			/* Number of axis pixels. */
static	int	org[2];			/* Lower pixel limit */
static	int	lim[2];			/* Upper pixel limit */
static	int	lfft[2];		/* Number of pixels before fft */
static	double	max[2];			/* Maximum float axis value */
static	double	min[2];			/* Minimum float axis value */
static	double	wrap[2];		/* Axis wrap limits */
static	char	file[2][BUFLEN];	/* File for scale=f mapping. */
static	int	npoly[2];		/* Order of polynomial if scale=p */
static	char	parm[2][BUFLEN];	/* Polynomial coefficients if scale=p */
static	char	axis[2][BUFLEN];	/* Axis location */
	/*
	 *	header processing table
	 */
/*
   Moved keytab initialization to inside main to avoid
   compiler error - Jeff Bytof  Nov. 2009
*/


static	char	dtype;		/* input pixel type code */
static	int	lpix;		/* input pixel length */
static	int	tfile;		/* temporary file counter */
static	char	buf[256];	/* name of temporary file */
static	FILE	*fp;		/* input or scratch file pointer */
static	FILE	*fo;		/* output file pointer */

static	char	*tmpl = "%s/ihrXXXXXX";	/* scratch file names */

static 		invert(), reverse(), rotate(), copy(), copy2();
static	FILE	*copen();	/* open file with error checking */
static	char	*fname();	/* make temporary filename */
/*
 *	interrupt handler
 */
static
onint()
{
	while (--tfile >= 0)
		(void)(unlink(fname(tfile)));
	exit(0);
	return(0);
}

main(argc,argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* argument pointer array */
{
	double	d;			/* scratch */
	register i;			/* scratch */
	register struct key *kp;	/* -> table of header variables */

/*
    moved keytab initialization to inside main from above
    to avoid compile error - Jeff Bytof  Nov. 2009
    Also removed static struct declaration
*/

struct	key {
	char	name[12];	/* name of header variable */
	char	*addr[2];	/* pointers to x- and x-values */
	int	len;		/* max length of string values */
	int	flg[2];		/* x- and y-value search flags */
}  keytab[] = {
"xtype",	(char *) &type[0],	(char *) &type[1],	1,	1, 1,
"xscale",	(char *) &scale[0],	(char *) &scale[1],	1,	1, 1,
"xtitle",	(char *) title[0],	(char *) title[1],	BUFLEN,	1, 1,
"xfont",	(char *) font[0],	(char *) font[1],	FONTLEN,1, 1,
"xnum",		(char *) &num[0],	(char *) &num[1],	0,	0, 0,
"xorg",		(char *) &org[0],	(char *) &org[1],	0,	1, 1,
"xlim",		(char *) &lim[0],	(char *) &lim[1],	0,	1, 1,
"xlfft",	(char *) &lfft[0],	(char *) &lfft[1],	0,	1, 1,
"xmax",		(char *) &max[0],	(char *) &max[1],	0,	1, 1,
"xmin",		(char *) &min[0],	(char *) &min[1],	0,	1, 1,
"xwrap",	(char *) &wrap[0],	(char *) &wrap[1],	0,	1, 1,
"xfile",	(char *) file[0],	(char *) file[1],	BUFLEN,	1, 1,
"xpoly",	(char *) &npoly[0],	(char *) &npoly[1],	0,	1, 1,
"xparm",	(char *) parm[0],	(char *) parm[1],	BUFLEN,	1, 1,
"xaxis",	(char *) axis[0],	(char *) axis[1],	BUFLEN,	1, 1,
""
};

	/*
	 *	verify calling arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	if (argc > 1)
		ifile = *++argv;
	if (Sflag)
		telltale(0);
	/*
	 *	read the header of the input file
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	putargs();

	if (*hfile != '-')
		if ((fo = ihread(hfile,GIPSPROFILE)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fo));
	
	if (ihget("dtype", &dtype, 1))
		gipserr();
	/*
	 *	read the header variables into the keytab table
	 */
	for (i = 0; i < 2; i++) {
		for (kp = keytab; kp->name[0] != 0; kp++) {
			*kp->name = "xy"[i];
			if (ihget(kp->name,kp->addr[i],kp->len) == 0)
				kp->flg[i] = 0;
			else
				if (kp->flg[i] == 0) 
					gipserr();
		}
		if (num[i] < 1)
			error("%.1snum < 1","xy"+i);
	}
	/*
	 *	rearrange and update the header variables as required
	 */
	for (i = 0; i < 2; i++) {
		if (rflag && nflag && i == 0 || iflag && i == 1)
			if (scale[i] == 'f')
				error("unsupported %.1sscale value","xy"+i);
			else	{
				d = max[i];
				max[i] = min[i];
				min[i] = d;
			}
		if (! nflag && *font[i])
			if (font[i][strlen(font[i])-1] == 'r')
				font[i][strlen(font[i])-1] = '\0';
			else
				(void)(strcat(font[i],"r"));
		for (kp = keytab; kp->name[0] != 0; kp++) {
			*kp->name = "xy"[nflag ? i : 1-i];
			if (kp->flg[i] != 0)
				(void)(ihdel(kp->name));
			else
				(void)(ihput(kp->name,kp->addr[i]));
		}
	}
	/*
	 *	write the output header, merging with '-m' option if requested
	 */
	if (*mfile != '-')
		if ((fo = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fo));
	if ((fo = ihwrit(ofile,0)) == NULL)
		gipserr();
	if (fo == stdout)
		ofile = "'stdout'";

	if ((lpix = ihdtype(dtype)) == 0)
		error("bad dtype value: %.1s",&dtype);

	if (nflag) 
		if (rflag || iflag)
			invert(num[1]);
		else
			while ((i = getc(fp)) != EOF)
				(void)(putc(i,fo)); 
	else	
		rotate(num[0]);

	(void)(ihbcl(fo));
	if (Sflag)
		telltale(1);
	(void)(onint());
}
/*
 *	invert 'ynum' rasters without rotation
 */
static
invert(ynum)

	int	ynum;		/* number of rasters to invert */
{
	int	size;		/* size of incore buffer */
	int	i;		/* scratch */
	long	loff = 0;	/* current offset in input file */
	register raster;	/* length of raster */
	register char *in;	/* input buffer pointer */

	int	incore = 0;	/* =1 if inversion performed incore */
	FILE	*ff;		/* temporary output file stream for ! incore */

	raster = num[0]*lpix;
	size = raster*ynum;
	/*
	 *	if incore inversion, read the whole image into core
	 */
	if (iflag &&
	    (incore = (size <= block &&
			(in = malloc((unsigned)size)) != NULL))) {
		if (fread(in,size,1,fp) != 1)
			gipserr2(GIPSBADREAD,ifile);
		in += size-raster;
	} else	{
		/*
		 *	if input image is coming from a disk file, seek it
		 */
		in = alloc(raster,sizeof(char));
		if (iflag)
			if (fp != stdin && strncmp(ifile,"/dev/rmt",8) != 0)
				loff = ftell(fp);
			else	{
				/*
				 *	if input image is a pipe or tape file,
				 *	copy it to a disk temporary,
				 *	using the incore buffer
				 */
				ff = copen(fname(tfile++),"w");
				for (i = ynum; --i >= 0; )
					if (fread(in,raster,1,fp) != 1)
						gipserr2(GIPSBADREAD,ifile);
					else if (fwrite(in,raster,1,ff) != 1)
						gipserr2(GIPSBADWRITE,ofile);
				(void)(ihbcl(fp));
				(void)(fclose(ff));
				fp = copen(ifile = buf, "r");
			}
	}
	/*
	 *	write out the inverted image
	 */
	while (--ynum >= 0) {
		if (! incore) {
			if (iflag)
				(void)(fseek(fp,(long)(loff+raster*ynum),0));
			if (fread(in,raster,1,fp) != 1)
				gipserr2(GIPSBADREAD,ifile);
		}
		if (reverse(in,raster,fo))
			gipserr2(GIPSBADWRITE,ofile);
		else if (incore)
			in += iflag ? -raster : raster;
	}
	(void)(ihbcl(fp));
	/*
	 *	if a temporary was created, remove it
	 */
}
/*
 *	write the raster to 'fo' (in reverse order if 'rflag' set)
 */
static
reverse(in,raster,ptr)

	register char	*in;	/* input raster pointer */
	register int	raster;	/* number of raster elements */
	register FILE	*ptr;	/* output stream pointer */
{
	register	l;
	register char	*out;

	if (! rflag)
		return(fwrite(in,raster,1,ptr) != 1);
	while ((raster -= (l = lpix)) >= 0)
		for (out = in+raster; --l >= 0; (void)(putc(*out++,ptr)))
			;
	return(ferror(ptr));
}
/*
 *	rotate 'xnum'-element raster
 */
static
rotate(xnum)

	int	xnum;
{
	int	nfile;		/* number of files at this level */
	int	size;		/* size of incore buffer */
	int	raster;		/* length of raster */
	int	xnum0;		/* number of elements in first split */
	int	xnum1;		/* number of elements in remaining splits */
	int	rast0;		/* length of first split in bytes */
	int	rast1;		/* length of remaining splits in bytes */
	int	fn[MAXFIL];	/* output file numbers */
	FILE	*fid[MAXFIL];	/* output file descriptors */
	char	*in;		/* input raster */
	char	tname[256];	/* temporary file name */
	register i;		/* scratch */
	register j;		/* scratch */
	register char *out;	/* output raster pointer */
	/*
	 *	if rotation can be performed incore,
	 *	do it immediately and return
	 */
	if ((size = (raster = xnum*lpix)*num[1]) <= block &&
	    (in = malloc((unsigned)size)) != NULL) {
		if (fread(in,size,1,fp) != 1)
			gipserr2(GIPSBADREAD,ifile);
		copy(in,xnum,lpix);
		(void)(ihbcl(fp));
		free(in);
		return;
	}
	if (xnum <= 1)
		error("cannot invert","");
	/*
	 *	determine the number of times the raster can be split,
	 *	and the length of each split. The first split may be
	 *	shorter than the others
	 */
	if ((nfile = (size/block)+1) < 2)
		nfile = 2;
	if (nfile > MAXFIL)
		nfile = MAXFIL;
	rast1 = (xnum1 = (xnum+nfile-1)/nfile) *lpix;
	rast0 = (xnum0 = xnum - xnum1*(nfile-1)) *lpix;
	/*
	 *	open the scratch files, one for each split raster
	 */
	for (i = nfile; --i >= 0; )
		fid[i] = copen(ifile = fname(fn[i] = tfile++),"w");
	/*
	 *	allocate the input buffer
	 */
	in = alloc(raster,sizeof(char));
	/*
	 *	read each input raster and write it to the several output files
	 */
	for (i = num[1]; --i >= 0; )
		if (fread(out = in,raster,1,fp) != 1)
			gipserr2(GIPSBADREAD,ifile);
		else for (j = nfile; --j >= 0; out += rast1)
			if (fwrite(out,(j == 0) ? rast0 : rast1,1,fid[j]) != 1)
				gipserr2(GIPSBADWRITE,ofile);
	/*
	 *	close files and free buffers
	 */
	for (i = nfile; --i >= 0; (void)(fclose(fid[i])))
		;
	(void)(ihbcl(fp));
	free(in);
	/*
	 *	open the temporary files one at a time and call rotate()
	 *	again to copy them to the output image
	 */
	for (i = nfile; --i >= 0; ) {
		ifile = fname(fn[j = rflag ? i : nfile-i-1]);
		fp = copen(strcpy(tname, ifile), "r");
		rotate((j == 0) ? xnum0 : xnum1);
		(void)(unlink(tname));
	}
}
/*
 *	open file 'fil' and report any errors
 */
static	FILE *
copen(fil,code)

	register char *fil;
	register char *code;
{
	register FILE *ff;

	if ((ff = fopen(fil,code)) != NULL)
		return(ff);
	gipserr2(GIPSBADOPEN,fil);
	return(NULL);
}
/*
 *	return temporary filename
 */
static	char *
fname(i)

	register i;
{
	static	char name[256];

	if (! *name) {
		sprintf(name,tmpl,ddir);
		(void)(mktemp(name));
		gipsint(onint);
	}
	sprintf(buf,"%s%d",name,i);
	return(buf);
}
/*
 *	split raster into segments 
 */
static
copy(in,xnum,llpix)

	register char	*in;
	register	xnum;
	register	llpix;
{
	register int	raster = xnum*llpix;

	if (! rflag) {
		in += raster-llpix;
		llpix = -llpix;
	}
	while (--xnum >= 0)
		if (copy2(in,raster,num[1],fo))
			gipserr2(GIPSBADWRITE,ofile);
		else
			in += llpix;
}
/*
 *	write raster segment to output file
 */
static
copy2(in,raster,ynum,ptr)

	register char	*in;
	register int	raster;
	register	ynum;
	register FILE	*ptr;
{
	register	l;

	if (iflag) {
		in += raster*(ynum-1);
		raster = -raster;
	}
	for (l = lpix; --ynum >= 0; in += raster)
		if (l == 1)
			(void)(putc(*in,ptr));
		else {
			while (--l >= 0)
				(void)(putc(*in++,ptr));
			in -= (l = lpix);
		}
	return(ferror(ptr));
}
