#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihs.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 2004/11/29 18:52:23 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihs.c		Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 2004/11/29 18:52:23 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Copy GIPS stream to display image plane
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"

#define		NIX	-32768
	/*
	 *	fields updated by gipsflags() from callers arguments
	 */
static	flag	cflag;			/* =1 to close input stream when done */
static	flag	iflag;			/* =1 to insert image into plane */
static	flag	qflag;			/* =1 to suppress warning messages */
static	flag	nflag;			/* =1 to skip reading image */
static	flag	sflag;			/* =1 to copy image without change */
static	flag	Sflag;			/* generate memory statistics */
static	int	block;			/* output blocksize */
static	int	headlen	= 1;		/* Max length of header line */
static	int	xorg	= NIX;		/* x-coord of image origin */
static	int	xnum;			/* image x-dimension */
static	int	xlim	= NIX;		/* width of image */
static	int	yorg	= NIX;		/* y-coord of image origin */
static	int	ynum;			/* image y-dimension */
static	int	ylim	= NIX;		/* height of image */
static	char	*ifile	= "-";		/* input image file name */
static	char	*ofile	= "-";		/* output image file name */
static	char	*mfile	= "-";		/* input header merge file name */
static	char	*ffile;			/* output histogram file name */
static	char	*hfile;			/* output header file name */
	/*
	 *	table used by gipsflags()
	 */
static	flagentry	flagtable[] = {
	"S",	"1",		(flag *) &Sflag,
	"X",	"%d lim",	(flag *) &xlim,
	"Y",	"%d lim",	(flag *) &ylim,
	"",	"%d block",	(flag *) &block,
	"c",	"1",		(flag *) &cflag,
	"f",	"-> file",	(flag *) &ffile,
	"h",	"-> file",	(flag *) &hfile,
	"i",	"1",		(flag *) &iflag,
	"l",	"%d length",	(flag *) &headlen,
	"m",	"-> file",	(flag *) &mfile,
	"n",	"1",		(flag *) &nflag,
	"o",	"-> file/n",	(flag *) &ofile,
	"q",	"1",		(flag *) &qflag,
	"s",	"1",		(flag *) &sflag,
	"x",	"%d org",	(flag *) &xorg,
	"y",	"%d org",	(flag *) &yorg,
	NULL,	"[file]",	NULL
};

static	char	dtype;			/* GIPS image pixel type */
static	int	*ftable	= NULL;		/* output histogram */

static		puthdr(), puthist(), ihsc();
extern	char	*strcpy();
extern	off_t	lseek();

main(argc,argv)

	int		argc;		/* number of arguments */
	char		**argv;		/* array of argument pointers */
{
	FILE		*fm;		/* input GIPS merge-header stream */
	register FILE	*fp;		/* input GIPS file stream */
	register FILE	*fo;		/* output GIPS header stream */
	int		ixy;		/* number of bytes read */
	int		num;		/* scratch */
	int		nimage = 0;	/* > 0 for output to image plane */
	/*
	 *	save and parse callers arguments
	 */
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	if (argc > 1)
		ifile = *++argv;
	if (Sflag)
		telltale(0);
	/*
	 *	check desired header length
	 */
	if (headlen < 0 || headlen > BUFLEN)
		error("bad -l length: %d",(char *)headlen);
	/*
	 *	if output is to display, check that it is available
	 */
	if (*ofile != '-' && sscanf(ofile,"%d",&nimage) == 1 && nimage > 0) {
		if (command_open() == NULL)
			gipserr();
	} else if (iflag)
		error("-i conflicts with -o %s",ofile);
	/*
	 *	read input GIPS header and possible merge file
	 */
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (*mfile != '-')
		if ((fm = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fm));
	/*
	 *	get necessary header variable values and check them out
	 */
	if (ihget("dtype",&dtype,1))
		gipserr();
	if (ihdtype(dtype) == 0 && dtype != 'G')
		gipserr1(GIPSBADVAL);
	if (ihget("xnum",(char *)&xnum,0)|| ihget("ynum",(char *)&ynum,0))
		gipserr();

	if (xorg != NIX)
		(void)(ihput("xorg",(char *)&xorg));
	else
		xorg = (ihget("xorg",(char *)&num,0) == 0) ? num : 0;
	if (xorg >= COLS  || (xorg+xnum) < 0)
		gipserr1(GIPSBADVAL);
	if (xlim != NIX)
		(void)(ihput("xlim",(char *)&xlim));
	else
		xlim = (ihget("xlim",(char *)&num,0) == 0) ? num : COLS;
	if (xlim > COLS)
		gipserr1(GIPSBADVAL);
	if (yorg != NIX)
		(void)(ihput("yorg",(char *)&yorg));
	else
		yorg = (ihget("yorg",(char *)&num,0) == 0) ? num : 0;
	if (yorg >= ROWS || (yorg+ynum) < 0)
		gipserr1(GIPSBADVAL);
	if (ylim != NIX)
		(void)(ihput("ylim",(char *)&ylim));
	else
		ylim = (ihget("ylim",(char *)&num,0) == 0) ? num : ROWS;
	if (ylim > ROWS)
		gipserr1(GIPSBADVAL);
	if (! sflag && ! qflag && ! nflag &&
	    ( xorg < 0 || xorg + xnum > COLS || yorg < 0 || yorg + ynum > ROWS))
		(void) fprintf(stderr,
			"ihs: warning - image truncated on:%s%s%s%s\n",
			(xorg < 0) ? " left" : "",
			(xorg + xnum > COLS) ? " right" : "",
			(yorg < 0) ? " top" : "",
			(yorg + ynum > ROWS) ? " bottom" : "");
	/*
	 *	for -n option, just write the header
	 */
	if (nflag && hfile == NULL)
		hfile = "-";
	if (! sflag && hfile != NULL)
		if (nimage <= 0 || *hfile != '-' || hfile[1]) {
			puthdr();
			if (nflag++) {
				if (Sflag)
					telltale(1);
				(void)(exit(0));
			}
		}
	if (dtype == 'G')
		error("cannot process dtype=G image: ",ifile);
	/*
	 *	open the COLSxROWS image output file
	 */
	if (ofile[0] == '-' && ofile[1] == '\0') {
		if (! sflag &&  hfile != NULL &&
				hfile[0] == '-' && hfile[1] == '\0')
			error("-h and -o cannot both be 'stdout'","");
		(void)(signal(SIGPIPE,exit));
		fo = stdout;
	} else if (nimage > 0) {
		if ((num = image_open(nimage,iflag ? 2 : 1)) < 0 ||
		    (fo = fdopen(num,iflag ? "w+" : "w")) == NULL)
			gipserr();
	} else if ((fo = fopen(ofile,"w")) == NULL)
		gipserr2(GIPSBADOPEN,ofile);
	/*
	 *	-s flag: copy the image without reformatting
	 */
	if (sflag) {
		register i;
		if (iflag || nflag || ffile != NULL)
			error("-s flag conflicts with -f,-i,-n","");
		if (hfile != NULL)
			puthdr();
		else if (block > 0)
			setbuffer(fo, alloc(1, block), block);
		for (num = 0; num < ynum; num++) {
			for (i = ihdtype(dtype)*xnum; --i >= 0; )
				(void) putc(getc(fp),fo);
			if (ferror(fp))
				gipserr2(GIPSBADREAD,ifile);
			if (ferror(fo))
				gipserr2(GIPSBADWRITE,ofile);
		}
	} else {
		if ((ixy = ihsc(ihdtype(dtype),fp,fo)) == 0)
			gipserr2(GIPSBADWRITE,(nimage > 0) ? *gipserrs : ofile);
		(void)(ihbcl(fo));
		if (! cflag && ixy != xnum*ynum) {
			(void) fprintf(stderr,
				"ihs: bad pixel count: %d not %d\n",
					ixy,xnum*ynum);
			(void)(exit(1));
		}
		if (ftable != NULL)
			puthist(ixy);
		if (! nflag && hfile != NULL)
			puthdr();
	}
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}
	/*
	 *	add a history entry and write out the GIPS header
	 */
static
puthdr()
{
	register	FILE	*fh;

	putargs();
	if ((fh = ihwrit(hfile,headlen)) == NULL)
		gipserr();
	else if (fh == stdout)
		(void)(signal(SIGPIPE,exit));
	else
		(void)(ihbcl(fh));
}
	/*
	 *	generate histogram lookup table
	 */
static
puthist(ixy)
{
	FUNCTION shist[256];
	register i, fo;
	int	num;
	double	scale;

	if (sscanf(ffile,"%d",&num) == 1) {
		if (num < 1)
			gipserr2(GIPSBADOPEN,ffile);
		if ((fo = function_open(num,1)) < 0)
			gipserr();
	} else {
		(void)(strcpy(*gipserrs,ffile));
		if ((fo = creat(ffile,0644)) < 0)
			gipserr1(GIPSBADOPEN);
	}
	scale = 255.9999 / ixy;
	for (i = 256; --i >= 0; )
		shist[i] = scale * (ixy -= ftable[i]);
	num = write (fo,(char *)shist,sizeof(FUNCTION)*256);
	(void)(close(fo));
	if (num != sizeof(FUNCTION)*256)
		gipserr1(GIPSBADWRITE);
}

static
ihsc(size,fp,fo)

	int	size;		/* size of input pixel in bytes */
	FILE	*fp;		/* input stream pointer */
	FILE	*fo;		/* output stream pointer */
{
	int	insize, ncopy, count = 0;
	char	*in, out[COLS], *inoff, *outoff = out;
	register int irow, i;

	inoff = in = alloc(insize = xnum,size);
	if (ffile != NULL)
		ftable = (int *) alloc(256,sizeof(int));
	if (gipserrno == GIPSNOCORE)
		gipserr();

	clear(out,sizeof(out));
	irow = ylim;

	for (i = yorg; i; ) 
		if (i > 0)
			if (! iflag &&  fwrite(out,sizeof(out),1,fo) != 1)
				return(0);
			else	{
				irow--;
				i--;
			}
		else if (fread(in,size,insize,fp) == insize) {
			count += insize;
			i++;
		} 
		else
			return(count);

	if (xorg > 0)
		outoff += xorg;
	else
		inoff -= xorg*size;

	ncopy = (xorg+xnum > xlim ? xlim : xorg+xnum) - (outoff-out);

	while	(fread(in,size,insize,fp) == insize)  {
		count += insize;
		if (irow == 0)
			if (cflag)
				break;
			else
				continue;

		(void)(gips((double *)inoff,(int *)outoff,ncopy,GIPS_D_VAL));
		
		if (ftable != NULL) {
			register char *b = outoff;
			for (i = ncopy; --i >= 0; ftable[*b++ & 0377]++)
				;
		}
		if (iflag) {
			(void)(lseek(fileno(fo),
				(off_t)((ylim-irow)*COLS+outoff-out),0));
			if (write(fileno(fo),outoff,ncopy) != ncopy)
				return(0);
		} else if (fwrite(out,sizeof(out),1,fo) != 1)
			return(0);
		irow--;
	};

	if (! iflag && irow > 0) {
		clear(outoff,ncopy);
		while (--irow >= 0)
			if (fwrite(out,sizeof(out),1,fo) != 1)
				return(0);
	}

	if (fp != stdin)
		(void)(ihbcl(fp));
	return(count);
}
