#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihsamp.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 1991/04/24 16:44:59 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1990 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihsamp.c,v $		Creation Date:	05/10/88
**	Version:	$Revision: 2.4 $		Update:		$Date: 1991/04/24 16:44:59 $
**	Function:	Convert cellular image to raster format
**	Components:	main()		GIPS/UNIX main program
*/

#include <stdio.h>
#include <signal.h>
#define	GRID
#include "gips.h"

	/*
	**	Input variables
	*/
	
static	char	*ofile	= "-";		/* output GIPS file */
static	char	*mfile	= "-";		/* output GIPS file */
static	int	factor	= 1;		/* reduction factor */
static	flag	Sflag;			/* statistics flag */

	/*
	**	Variable definitions
	*/
	
static	flagentry	flagtable[] = {
	"S",	"1",		(flag *) &Sflag,
	"m",	"-> file",	(flag *) &mfile,
	"n",	"%d factor",	(flag *) &factor,
	"o",	"-> file",	(flag *) &ofile,
	NULL,	"[file]",	NULL
};

extern	int	exit();
extern	int	_gcount;

main(argc, argv)
	int	argc;
	char	**argv;
{
	register FILE	*fp, *fq;
	register gimage	*gi;
	register int	x, y, fact, dlen, i;
	register char	*c, *d;
	double		dnull;
	int		xnum, ynum;
	char		dtype = 'b', geom[256], *ifile, *nullbuf;

	/*
	**	Parse command string
	*/
	
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...", argv[2]);
	else if (argc < 2)
		(void)signal(SIGPIPE, exit), ifile = "-";
	else
		ifile = argv[1];

	if ((fact = factor) < 1)
		error("bad -n value: %d", (char *)factor);

	if (Sflag)
		telltale(0);
	
	if ((fp = ihread(ifile, GIPSPROFILE)) == NULL)
		gipserr();
	if (fp != stdin)
		fclose(stdin);
	else
		ifile = "'stdin'";

	if (*mfile != '-')
		if (fq = ihmerg(mfile))
			(void)ihbcl(fq);
		else
			gipserr();

	putargs();
	if (ihget("xnum", &xnum) || ihget("ynum", &ynum) ||
		ihget("geometry", geom, sizeof(geom)-1))
			gipserr();
	sscanf(geom, "type=%c", &dtype);
	if (dtype == 'M') {
		dtype = 'f';
		ihput("dscale", "s");
	}
	if ((dlen = ihdtype(dtype)) == 0 || (gi = gopen(fp, 64)) == NULL)
		gipserr();


	ihdel("geometry");
	convert_mapping();
	xnum /= fact;
	ynum /= fact;
	if (ihput("xnum", &xnum) || ihput("ynum", &ynum) ||
		ihput("dtype", &dtype))
			gipserr();

	nullbuf = alloc(dlen, 1);
	if (ihget("dnull", &dnull) == 0)
		gconv(nullbuf, &dnull, 1, ihdcvt(dtype), CVT_D);

	if ((fp = ihwrit(ofile, 1)) == NULL)
		gipserr();
	else if (fp == stdout)
		ofile = "'stdout'";

	for (y = 0; y < ynum; y++)
		for (x = 0; x < xnum; ) {
			if (c = gcell(x * fact, y * fact, gi))
				while (_gcount > 0 && x < xnum) {
					_gcount -= fact;
					for (i = 0; i < dlen; i++)
						(void)putc(c[i], fp);
					c += fact * dlen;
					x++;
				}
			else {
				for (i = 0, d = nullbuf; i < dlen; i++)
					(void)putc(*d++, fp);
				x++;
			}
		}

	/*
	**	Cleanup
	*/

	if (ferror(stdin))	
		perror(ifile), exit(1);
	(void)ihbcl(fp);
	(void)gclose(gi);
	if (Sflag)
		telltale(1);
	exit(0);
}

	/*
	**	Convert the mapping parameter
	*/

static
convert_mapping()
{
	char	map[1024], name[64];
	int	type, xorg, yorg, xyrot;
	double	radius, e1, e2, e3, lat1, lat2;

	if (ihget("mapping", map, sizeof(map)-1))
		return;
	xyrot = 0;
	if (sscanf(map, GIPS_MAPPING, name, &type, &xorg, &yorg, &radius,
		&e1, &e2, &e3, &lat1, &lat2, &xyrot) < 10)
		return;
	radius /= factor;
	xorg /= factor;
	yorg /= factor;
	ihput("mapping", sprintf(map, GIPS_MAPPING_P, name, type, xorg, yorg,
		radius, e1, e2, e3, lat1, lat2, xyrot));
}
