#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihtofits.c,v $	GIPS $Revision: 2.3 $ (MIT/CSR) $Date: 1998/05/13 01:01:59 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		ihtofits.c	Creation Date:	06/25/86
**	Version:	$Revision: 2.3 $		Update:		$Date: 1998/05/13 01:01:59 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Translate GIPS-format image file to FITS
*/

#include	<stdio.h>
#include	<ctype.h>
#include	<signal.h>
#include	"gips.h"

#define		FITSLEN		2880
	
static	flag	bflag;			/* byte-swap binary output */
static	flag	Sflag;			/* print statistics */
static	char	*mfile	= "-";		/* merged header data */
static	char	*ofile	= NULL;		/* output FITS file */
	/*
	 *	argument table used by gipsflags()
	 */
static	flagentry	flagtable[] = {
	"b",	"1",		&bflag,
	"S",	"1",		&Sflag,
	"m",	"-> file",	(flag *)&mfile,
	"o",	"-> file",	(flag *)&ofile,
	NULL,	"[file]",	NULL
};
	/*
	**	Header variables
	*/
static	int	xnum;		/* number of x-pixels */
static	int	ynum;		/* number of y-pixels */
static	char	dtype;		/* pixel type code */
static	double	dfact;		/* slope of pixel normalization */
static	double	dzero;		/* base level of pixel normalization */
static	double	dnull;		/* unassigned value of pixel */
static	double	dmin = 0;	/* minimum displayable pixel value */
static	double	dmax = 256;	/* maximum displayable pixel value */
static	double	xmin = 0;		/* minimum realworld x-value */
static	double	xmax = GIPSHIGH;	/* maximum realworld x-value */
static	double	ymin = 0;		/* minimum realworld y-value */
static	double	ymax = GIPSHIGH;	/* maximum realworld y-value */
static	char	title[9];	/* image title */
static	char	xtitle[9];	/* x-axis title */
static	char	ytitle[9];	/* y-axis title */
static	char	dtitle[9];	/* data title */
	/*
	**	Fields defined only in FITS headers
	*/
static	double	crxdel;		/* width of x-pixel */
static	double	crydel;		/* width of y-pixel */
	/*
	**	field definition flags
	*/
#define		RCHAR		040	/* right-adjusted character string */
#define		INT		020	/* integer field */
#define		FLOAT		010	/* double precision field */
#define		CHAR8		04	/* 8-byte left justified string */
#define		FORMAT		(RCHAR|INT|FLOAT|CHAR8)
#define		GREQD		02	/* required GIPS field */
#define		INIT		01	/* item initialized */
	/*
	**	field definition table
	*/
static	struct	vars	{
	int	var_flag;	/* processing flags */
	char	*gips_name;	/* name for ihget() */
	char	*gips_val;	/* pointer to value */
	char	*fits_name;	/* name in fits header */
} variables[] = {
	GREQD,		"dtype",	(char *)&dtype,		NULL,
	FLOAT,		"xmax",		(char *)&xmax,		NULL,
	FLOAT,		"ymax",		(char *)&ymax,		NULL,

	RCHAR|INIT,	NULL,		"T",			"SIMPLE",
	RCHAR|INIT,	NULL,		"16",			"BITPIX",
	RCHAR|INIT,	NULL,		"2",			"NAXIS",
	INT|GREQD,	"xnum",		(char *)&xnum,		"NAXIS1",
	INT|GREQD,	"ynum",		(char *)&ynum,		"NAXIS2",
	FLOAT,		"dfact",	(char *)&dfact,		"BSCALE",
	FLOAT,		"dzero",	(char *)&dzero,		"BZERO",

	CHAR8,		NULL,		"GIPS V.1",		"ORIGIN",
	CHAR8,		"title",	(char *)title,		"OBJECT",
	FLOAT|INIT,	"dmin",		(char *)&dmin,		"DATAMIN",
	FLOAT|INIT,	"dmax",		(char *)&dmax,		"DATAMAX",
	CHAR8,		"dtitle",	(char *)dtitle,		"BUNIT",
	FLOAT,		"dnull",	(char *)&dnull,		"BLANK",

	CHAR8,		"xtitle",	(char *)xtitle,		"CTYPE1",
	RCHAR|INIT,	NULL,		"1",			"CRPIX1",
	FLOAT|INIT,	"xmin",		(char *)&xmin,		"CRVAL1",
	FLOAT|INIT,	NULL,		(char *)&crxdel,	"CDELT1",

	CHAR8,		"ytitle",	(char *)ytitle,		"CTYPE2",
	RCHAR|INIT,	NULL,		"1",			"CRPIX2",
	FLOAT|INIT,	"ymin",		(char *)&ymin,		"CRVAL2",
	FLOAT|INIT,	NULL,		(char *)&crydel,	"CDELT2",
	0
};

main(argc, argv)
	int	argc;			/* number of arguments */
	char	**argv;			/* array of argument pointers */
{
	int	dlen;			/* number of bytes per pixel */
	int	outline	= 0;		/* number of header lines written */
	char	*inbuf;			/* input buffer */
	char	*convbuf;		/* halfword buffer */
	short	outbuf[FITSLEN/2];	/* output buffer */
	char	*ifile	= "-";		/* input GIPS file */
	FILE	*fp;			/* input file stream */
	FILE	*fm;			/* merge file stream */
	FILE	*fo;			/* output file stream */
	register struct vars *vp;	/* pointer to variables */
	register i;			/* scratch */
	/*
	**	save and parse arguments
	*/
	saveargs(argv);
	argc = gipsflags(argc,argv,flagtable);
	if (argc == 2)
		ifile = *++argv;
	else if (argc > 2)
		error("multiple input files: %s",argv[2]);
	if (Sflag)
		telltale(0);
	/*
	**	open the GIPS file
	*/
	if ((fp = ihread(ifile,GIPSPROFILE)) == NULL)
		gipserr();
	if (mfile[0] != '-')
		if ((fm = ihmerg(mfile)) == NULL)
			gipserr();
		else
			ihbcl(fm);
	/*
	**	Read useful header fields
	*/
	for (vp = variables; vp->var_flag; vp++)
		if (vp->gips_name) {
			if (ihget(vp->gips_name, vp->gips_val, 9) == 0 ||
					gipserrno == GIPSTRUNC) {
				if (vp->var_flag & CHAR8) {
					register char *c;
					for (c = vp->gips_val; *c; c++)
						if (islower(*c))
							*c = toupper(*c);
				}
				vp->var_flag |= INIT;
			} else if (vp->var_flag & GREQD)
				gipserr();
		}
	/*
	**	Allocate buffers
	*/
	if ((dlen = ihdtype(dtype)) == 0)
		gipserr2(GIPSBADVAL, "dtype");
	inbuf = alloc(dlen, xnum);
	convbuf = alloc(2, xnum);
	if (inbuf == NULL || convbuf == NULL)
		gipserr();
	/*
	**	Open output file with FITSLEN buffer
	*/
	if (ofile == NULL) {
		fo = stdout;
		ofile = "'stdout'";
		(void) signal(SIGPIPE, exit);
	} else if ((fo = fopen(ofile, "w")) == NULL)
		gipserr2(GIPSBADOPEN, ofile);
	setbuffer(fo, (char *)outbuf, FITSLEN);
	/*
	**	Compute FITS pixel spacings
	*/
	crxdel = (((xmax >= GIPSHIGH) ? (xnum-1) : xmax) - xmin)/(xnum-1);
	crydel = (((ymax >= GIPSHIGH) ? (ynum-1) : ymax) - ymin)/(ynum-1);
	/*
	**	Write the FITS header
	*/
	for (vp = variables; vp->var_flag; vp++) {
		if (! (vp->var_flag & INIT) || ! vp->fits_name)
			continue;
		fprintf(fo, "%-8.8s= ", vp->fits_name);
		switch (vp->var_flag & FORMAT) {
		case RCHAR:
			fprintf(fo, "%20.20s", vp->gips_val);
			break;
		case INT:
			fprintf(fo, "%20d", *((int *)(vp->gips_val)));
			break;
		case FLOAT:
			fprintf(fo, "%20lg", *((double *)(vp->gips_val)));
			break;
		case CHAR8:
			fprintf(fo, "'%-8.8s'          ", vp->gips_val);
			break;
		}
		for (i = 30; i < 80; i++)
			putc(' ', fo);
		outline++;
	}
	fprintf(fo, "END");
	for (i = (outline*80 + 3) % FITSLEN; i > 0 && i < FITSLEN; i++)
		putc(' ', fo);
		
	while (--ynum >= 0) {
		if (fread(inbuf, dlen, xnum, fp) != xnum)
			gipserr2(GIPSBADREAD, *argv);
		gconv(convbuf, inbuf, xnum, CVT_H, ihdcvt(dtype));
		if (bflag)
			swab(convbuf, convbuf, xnum*2);
		if (fwrite(convbuf, 2, xnum, fo) != xnum)
			gipserr2(GIPSBADWRITE, fo);
	}

	ihbcl(fp);
	if (Sflag)
		telltale(1);
	exit(0);
}
	
