#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihtogif.c,v $	GIPS $Revision: 1.4 $ (MIT/CSR) $Date: 1998/05/13 01:02:00 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1991 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihtogif.c,v $		Creation Date:	05/18/91
**	Version:	$Revision: 1.4 $		Update:		$Date: 1998/05/13 01:02:00 $
**	Function:	Convert general image to GIF format
**	Components:	main()		GIPS/UNIX main program
*/

#include <stdio.h>
#include <signal.h>
#include <strings.h>
#include <memory.h>
#include <ctype.h>
#include "gips.h"

#define	getpix() ((--pix_cnt>=0) ? *pix_ptr++ : readblock())

	/*
	**	Input variables
	*/
static	char	*cfile	= NULL;		/* comment file */
static	char	*ifile	= "-";		/* input GIPS file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= NULL;		/* output GIF file */
static	char	*afile	= NULL;		/* input colormap file */
static	flag	Mflag;			/* mono colormap */
static	flag	Sflag;			/* statistics flag */
	/*
	**	Variable definitions
	*/
static	flagentry	flagtable[] = {
	"M",	"1",			(flag *) &Mflag,
	"S",	"1",			(flag *) &Sflag,
	"a",	"-> map",		(flag *) &afile,
	"c",	"-> textfile",		(flag *) &cfile,
	"m",	"-> file",		(flag *) &mfile,
	"o",	"-> file",		(flag *) &ofile,
	NULL,	"[file]",		NULL
};

static	FILE	*fi;			/* input file stream */
static	int	xnum, ynum;		/* image size */
static	int	dlen;			/* length of input pixel */
static	char	*inbuf;			/* input buffer */
static	unsigned char	*outbuf;		/* converted buffer */
static	unsigned char	*pix_ptr;		/* pointer to outbuf */
static	int	pix_cnt;		/* size of outbuf */

static	putword(), compress(), output(), cl_block(), char_out(), flush_char();

main(argc, argv)
	int	argc;
	char	**argv;
{
	int	i, j;		/* scratch */
	char	dtype;		/* pixel type */
	char	buf[BUFSIZ];	/* scratch */
	FILE	*fp;		/* cmap input */

	/*
	**	Parse command string
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc == 2)
		ifile = argv[1];
	if (Sflag)
		telltale(0);

	/*
	**	Read the input header
	*/

	if ((fi = ihread(ifile, GIPSPROFILE)) == NULL)
		gipserr();
	else if (fi == stdin)
		ifile = "'stdin'";
	if (mfile != NULL && *mfile != '-')
		if (fp = ihmerg(mfile))
			(void)ihbcl(fp);
		else
			gipserr();
	if (ihget("xnum", &xnum) || ihget("ynum", &ynum))
		gipserr();
	if (ihget("dtype", &dtype, 1))
		dtype = 'b';
	if ((dlen = ihdtype(dtype)) == 0)
		gipserr();
	inbuf = alloc(dlen, xnum);
	outbuf = (unsigned char *)alloc(sizeof(unsigned char), xnum);

	/*
	**	Open the output file
	*/

	if (ofile == NULL) {
		(void)signal(SIGPIPE, exit);
		ofile = "'stdout'";
	} else if (freopen(ofile, "w", stdout) == NULL)
		perror(ofile), exit(1);

	/*
	**	Write GIF header
	*/

	fwrite(cfile ? "GIF89a" : "GIF87a", 6, 1, stdout);
	putword(xnum);
	putword(ynum);
	putchar(0x80 | ((8-1)<<4) | (8-1));
	putchar(0);
	putchar(0);

	/*
	**	If specified, copy the colormap
	*/

	if (afile != NULL) {
		fp = fopen(strcpy(buf, afile), "r");
		if (fp == NULL)
			fp = fopen(strcat(buf,".cmap"), "r");
		if (fp == NULL) {
			sprintf(buf,"%s%s",GIPSCMAPDIR,afile);
			fp = fopen(buf, "r");
		}
		if (fp == NULL)
			fp = fopen(strcat(buf,".cmap"), "r");
		if (fp == NULL)	
			perror(afile), exit(1);
		for (i = 0; i < 256;) {
			float	r, g, b;
			int	ir, ig, ib;
			if (fgets(buf, sizeof(buf)-1, fp) == NULL)
				perror(afile), exit(1);
			for (j = 0; isspace(buf[j]); j++)
				;
			if (buf[j] == '\0' || buf[j] == '#')
				continue;
			if (sscanf(&buf[j], "%f %f %f", &r, &g, &b) != 3)
				fprintf(stderr, "%s: bad record: %s\n",
					afile, buf), exit(1);
			ir = (int)r & 0xff;
			ig = (int)g & 0xff;
			ib = (int)b & 0xff;
			if (Mflag) {
				j = (ir*11 + ig*16 + ib*5) >> 5;
				putchar(j);
				putchar(j);
				putchar(j);
			} else {
				putchar(ir);
				putchar(ig);
				putchar(ib);
			}
			i++;
		}
		fclose(fp);
	} else
		for (i = 0; i < 256; i++) {
			putchar(i);
			putchar(i);
			putchar(i);
		}

	/*
	**	Insert comments
	*/

	if (cfile) {
		FILE	*fc = fopen(cfile, "r");

		if (fc == NULL)
			perror(cfile), exit(-1);
		putchar(0x21);
		putchar(0xfe);
		while (fgets(buf, sizeof(buf)-1, fc)) {
			i = strlen(buf);
			while (--i >= 0 && (isspace(buf[i]) || buf[i] == '\r'))
				;
			if (i > 252)
				i = 252;
			buf[++i] = '\r';
			buf[++i] = '\n';
			putchar(++i);
			fwrite(buf, i, 1, stdout);
		}
		(void)fclose(fc);
		putchar(0);
	}

	/*
	**	Write the Image
	*/

	putchar(',');		/* image separator */
	putword(0);
	putword(0);
	putword(xnum);		/* x dimension */
	putword(ynum);		/* y dimension */
	putchar(0);		/* no interlace */
	putchar(8);		/* 8 bit */
	compress();
	putchar(0);		/* Write Zero-length packet (EOF) */
	putchar(';');		/* Write GIF file terminator */

	/*
	**	Clean up
	*/

	(void)ihbcl(fi);
	if (Sflag)
		telltale(1);
	exit(0);
}

static	unsigned char
readblock()
{
	if (fread(inbuf, dlen, xnum, fi) != xnum)
		perror(ifile), exit(1);
	if (gips((double *)inbuf, (int *)outbuf, xnum, GIPS_D_VAL))
		gipserr();
	pix_cnt = xnum-1;
	pix_ptr = outbuf;
	return *pix_ptr++;
}

static
putword(w)
	int	w;
{
	putchar(w & 0xff);
	putchar((w>>8)&0xff);
}

#define BITS		12
#define	INIT_BITS	9
#define HSIZE		5003			/* 80% occupancy */
#define	MAXCODE(n)	((1<<(n))-1)
#define	MAXMAXCODE	(MAXCODE(BITS)+1)
#define	CLEARCODE	(MAXCODE(INIT_BITS-1)+1)
#define	EOFCODE		(MAXCODE(INIT_BITS-1)+2)

static	unsigned long masks[] = { 
	0x0000, 0x0001, 0x0003, 0x0007, 0x000F, 0x001F, 0x003F, 0x007F, 0x00FF,
	0x01FF, 0x03FF, 0x07FF, 0x0FFF, 0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF
};
static	unsigned long	cur_accum;
static	int	cur_bits;
static	int	a_count;
static	char	accum[256];
static	int	n_bits = INIT_BITS;		/* number of bits/code */
static	long	htab [HSIZE];
static	unsigned short	codetab [HSIZE];
static	int	hsize = HSIZE;			/* for dynamic table sizing */
static	int	clear_flg;
static	int	maxcode = MAXCODE(INIT_BITS);
static	int	free_ent = MAXCODE(INIT_BITS-1)+3;	/* first unused entry */

static
compress()
{
	register long	fcode;
	register long	len;
	register int	i;
	register int	c;
	register int	ent;
	register int	disp;
	register int	hshift;

	ent = getpix();	
	hshift = 0;
	for (fcode = hsize; fcode < 65536; fcode *= 2)
		hshift++;
	hshift = 8 - hshift;
	memset(htab, 0xff, sizeof(htab));
	output(CLEARCODE);

	for (len = xnum * ynum; --len > 0;) {
		c = getpix();  
		fcode = ((c << BITS) + ent);
		i = ((c << hshift) ^ ent);
		if (htab[i] == fcode) {
			ent = codetab[i];
			continue;
		}
		if (htab[i] >= 0) {
			disp = hsize - i;
			if (i == 0)
				disp = 1;
probe:
			if ((i -= disp) < 0)
				i += hsize;
			if (htab[i] == fcode) {
				ent = codetab[i];
				continue;
			}
			if (htab[i] > 0)
				goto probe;
		}
		output(ent);
		ent = c;
		if (free_ent < MAXMAXCODE) {
			codetab[i] = free_ent++;
			htab[i] = fcode;
		} else
			cl_block();
	}
	output(ent);
	output(EOFCODE);
}

static
output(code)
	int code;
{
	cur_accum &= masks[cur_bits];
	if (cur_bits > 0)
		cur_accum |= (code << cur_bits);
	else
		cur_accum = code;
	cur_bits += n_bits;
	while (cur_bits >= 8) {
		char_out((unsigned int) (cur_accum & 0xff));
		cur_accum >>= 8;
		cur_bits -= 8;
	}
	if (free_ent <= maxcode && ! clear_flg)
		;
	else if (clear_flg) {
		maxcode = MAXCODE(n_bits = INIT_BITS);
		clear_flg = 0;
	} else {
		n_bits++;
		maxcode = n_bits == BITS ? MAXMAXCODE : MAXCODE(n_bits);
	}
	if (code != EOFCODE)
		return;
	while (cur_bits > 0) {
		char_out((unsigned int)(cur_accum & 0xff));
		cur_accum >>= 8;
		cur_bits -= 8;
	}
	if (a_count > 0)
		flush_char();
	fflush(stdout);
	if (ferror(stdout))
		error("%s: unable to write GIF file", ofile);
}

static
cl_block()
{
	memset(htab, 0xff, sizeof(htab));
	free_ent = MAXCODE(INIT_BITS-1)+3;
	clear_flg = 1;
	output(CLEARCODE);
}

static
char_out(c)
	int c;
{
	accum[a_count] = c;
	if (++a_count >= 254)
		flush_char();
}

static
flush_char()
{
	putchar(a_count);
	fwrite(accum, a_count, 1, stdout);
	a_count = 0;
}
