#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihtopix.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 1998/11/13 23:53:33 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihtopix.c,v $		Creation Date:	05/10/88
**	Version:	$Revision: 2.4 $		Update:		$Date: 1998/11/13 23:53:33 $
**	Function:	Convert general image to Sun raster file format
**	Components:	main()		GIPS/UNIX main program
*/
#include <stdio.h>
#include <signal.h>
#include <strings.h>
#include <ctype.h>
#ifdef SVR4
#include "raster.h"
#else
#include <rasterfile.h>
#endif
#include "gips.h"
	/*
	**	Input variables
	*/
static	char	*ifile	= "-";		/* input GIPS file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= NULL;		/* output raster file */
static	char	*afile	= NULL;		/* input colormap file */
static	flag	Sflag;			/* statistics flag */
	/*
	**	Variable definitions
	*/
static	flagentry	flagtable[] = {
	"a",	"-> map",		(flag *) &afile,
	"S",	"1",			(flag *) &Sflag,
	"m",	"-> file",		(flag *) &mfile,
	"o",	"-> file",		(flag *) &ofile,
	NULL,	"[file]",		NULL
};

main(argc, argv)
	int	argc;
	char	**argv;
{
	struct	rasterfile	hdr;		/* Sun raster header */
	unsigned	char	rgb[3][256];	/* colormap */
	register	int	i, j;		/* scratch */
	register	int	dlen;		/* length of input pixel */
	register	int	nout;		/* length of output raster */
	char			buf[BUFSIZ];	/* scratch */
	char			dtype;		/* pixel type */
	char			*bin;		/* input buffer */
	char			*bout;		/* output buffer */
	register	FILE	*fi;		/* input image */
	register	FILE	*fp;		/* cmap input */
	/*
	**	Parse command string
	*/
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc == 2)
		ifile = argv[1];
	if (Sflag)
		telltale(0);
	/*
	**	Read the input header
	*/
	if ((fi = ihread(ifile, GIPSPROFILE)) == NULL)
		gipserr();
	else if (fi == stdin)
		ifile = "'stdin'";
	if (mfile != NULL && *mfile != '-')
		if (fp = ihmerg(mfile))
			(void)ihbcl(fp);
		else
			gipserr();
	/*
	**	Construct the raster header
	*/
	hdr.ras_magic = RAS_MAGIC;
	if (ihget("xnum", &hdr.ras_width) ||
			ihget("ynum", &hdr.ras_height) ||
			ihget("dtype", &dtype) ||
			(dlen = ihdtype(dtype)) == 0)
		gipserr();
	hdr.ras_depth = 8;
	bin = alloc(dlen, hdr.ras_width);
	nout = (1 + hdr.ras_width) & ~1;
	bout = alloc(1, nout);
	hdr.ras_length = nout * hdr.ras_height;
	hdr.ras_type = RT_STANDARD;
	if (afile == NULL) {
		hdr.ras_maptype = RMT_NONE;
		hdr.ras_maplength = 0;
	} else {
		hdr.ras_maptype = RMT_EQUAL_RGB;
		hdr.ras_maplength = sizeof(rgb);
	}
	/*
	**	Write the output header
	*/
	if (ofile == NULL) {
		(void)signal(SIGPIPE, exit);
		ofile = "'stdout'";
	} else if (freopen(ofile, "w", stdout) == NULL)
		perror(ofile), exit(1);
	if (fwrite(&hdr, sizeof(hdr), 1, stdout) != 1)
		perror(ofile), exit(1);
	/*
	**	If specified, write the colormap
	*/
	if (afile != NULL) {
		fp = fopen(strcpy(buf, afile), "r");
		if (fp == NULL)
			fp = fopen(strcat(buf,".cmap"), "r");
		if (fp == NULL) {
			sprintf(buf,"%s%s",GIPSCMAPDIR,afile);
			fp = fopen(buf, "r");
		}
		if (fp == NULL)
			fp = fopen(strcat(buf,".cmap"), "r");
		if (fp == NULL)	
			perror(afile), exit(1);
		for (i = 0; i < 256; ) {
			float	r, g, b;
			if (fgets(buf, sizeof(buf)-1, fp) == NULL)
				perror(afile), exit(1);
			for (j = 0; isspace(buf[j]); j++)
				;
			if (buf[j] == '\0' || buf[j] == '#')
				continue;
			if (sscanf(&buf[j], "%f %f %f", &r, &g, &b) != 3)
				fprintf(stderr, "%s: bad record: %s\n",
					afile, buf), exit(1);
			rgb[0][i] = (int)r;
			rgb[1][i] = (int)g;
			rgb[2][i] = (int)b;
			i++;
		}
		fclose(fp);
		hdr.ras_type = RT_STANDARD;
		if (fwrite(rgb, sizeof(rgb), 1, stdout) != 1)
			perror(afile), exit(1);
	}
	/*
	**	Translate the GIPS image to 8-bit pixels,
	**	and write out to the raster file
	*/
	for (i = hdr.ras_width, j = hdr.ras_height; --j >= 0; ) {
		if (fread(bin, dlen, i, fi) != i)
			perror(ifile), exit(1);
		if (gips((double *)bin, (int *)bout, i, GIPS_D_VAL))
			gipserr();
		if (fwrite(bout, 1, nout, stdout) != nout)
			perror(ofile), exit(1);
	}
	(void)ihbcl(fi);
	if (Sflag)
		telltale(1);
	exit(0);
}
