#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihtrans.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 1998/05/13 01:02:02 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihtrans.c	Creation Date:	12/29/83
**	Version:	$Revision: 2.4 $		Update:		$Date: 1998/05/13 01:02:02 $
**	Components:	main()		GIPS/UNIX command
**
**	Function:	Translate a GIPS image to/from an external format
*/

#include	<stdio.h>
#include	<signal.h>
#include	"gips.h"
	/*
	**	variables filled in by gipsflags()
	*/
static	char	*ifile	= "-";		/* input image array */
static	char	*ofile	= "-";		/* output general image */
static	char	*mfile	= "-";		/* update file */
static	flag	Sflag;			/* generate memory statistics */
static		swap2();
	/*
	**	argument definition block for gipsflags()
	*/
static	flagentry	flagtable[]	= {
	"S",	"1",		(flag *) &Sflag,
	"m",	"-> file",	(flag *) &mfile,
	"o",	"-> file",	(flag *) &ofile,
	NULL,	"[file]",	NULL
};
	/*
	**	Conversion from ieee to vax floating point
	*/
static	tovaxf(in, n, out)
	register float *in;
	register unsigned char *out;
	register n;
{
	(void) swap2((unsigned short *)in, 2*n, (unsigned short *)out);
	while (--n >= 0) {
		if (out[1])
			out[1]++;
		out += 4;
	}
	return 0;
}
	/*
	**	Conversion from ieee to vax double precision
	*/
static	tovaxd(in, n, out)
	register unsigned long *in, *out;
	register n;
{
	unsigned long *out2 = out;
	int	m = n;
#ifdef vax
	(void) swap4(in, n, out);
	in = out;
#endif
	while (--n >= 0) {
		register unsigned long i = *in++;
		register unsigned long j;

		if (j = i & 0x7ff00000) {
			register unsigned long k = *in++;
			*out++ = (i & 0x80000000) | ((j - 0x37e00000) << 3) |
				((i & 0xfffff) << 3) | (k >> 29);
			*out++ = k << 3;
		} else {
			in++;
			*out++ = 0;
			*out++ = 0;
		}
	}
#ifdef vax
	return swap1(out2, m, out2);
#else
	return swap2((unsigned short *)out2, 2*m, (unsigned short *)out2);
#endif
}
	/*
	**	Conversion from vax to ieee floating point
	*/
static	fromvaxf(in, n, out)
	register float *in;
	register unsigned char *out;
	register n;
{
	(void) swap2((unsigned short *)in, 2*n, (unsigned short *)out);
	while (--n >= 0) {
		if (out[0])
			out[0]--;
		out += 4;
	}
	return 0;
}
	/*
	**	Conversion from vax to ieee double precision
	*/
static	fromvaxd(in, n, out)
	register double *in;
	register unsigned long *out;
	register n;
{
#ifdef vax
	unsigned long *out2 = out;
	int	m = n;
	(void) swap1((unsigned long *)in, 2*n, out2);
#else
	(void) swap2((unsigned short *)in, 4*n, (unsigned short *)out);
#endif
	while (--n >= 0) {
		register unsigned long i = *out;
		register unsigned long j;

		if (j = i & 0x7f800000) {
			*out++ = (i & 0x80000000) | ((i & 0x7fffff) >> 3) |
				((j >> 3) + 0x37e00000);
			j = *out;
			*out++ = (j >> 3) | (i << 29);
		} else {
			*out++ = 0;
			*out++ = 0;
		}
	}
#ifdef vax
	(void) swap4(out2, 2*m, out2);
#endif
}
	/*
	**	Swap halfword pairs
	*/
static	swap1(in, n, out)
	register unsigned long *in, *out;
	register n;
{
	while (--n >= 0) {
		register unsigned long i = *in++;
		*out++ = (i << 16) | ( i >> 16);
	}
	return 0;
}
	/*
	**	Swap byte pairs
	*/
static	swap2(in, n, out)
	register unsigned short *in, *out;
	register n;
{
	while (--n >= 0) {
		register unsigned short i = *in++;
		*out++ = (i << 8) | ((i >> 8) & 0xff);
	}
	return 0;
}
	/*
	**	Reverse order of 4 bytes
	*/
static	swap4(in, n, out)
	register unsigned long *in, *out;
	register n;
{
	while (--n >= 0) {
		register unsigned long i = *in++;
		*out++ = ((i & 0xff) << 24) | ((i & 0xff00) << 8) |
			((i & 0xff0000) >> 8) | ((i & 0xff000000) >> 24);
	}
	return 0;
}
	/*
	**	Other conversions
	*/
static	tovaxc(in, n, out) float *in; unsigned char *out; {
	return tovaxf(in, 2*n, out);
}
static	fromvaxc(in, n, out) float *in; unsigned char *out; {
	return fromvaxf(in, 2*n, out);
}
static	icvtc(in, n, out) float *in; unsigned char *out; {
	return icvtf(in, 2*n, out);
}
static	toibmf(in, n, out) float *in, *out; {
	return 1;
}
static	toibmd(in, n, out) double *in, *out; {
	return 1;
}
static	toibmc(in, n, out) float *in, *out; {
	return 1;
}
extern	int	icvtf(), icvtd();
	/*
	**	structure defining an external format
	*/
typedef	struct	format	{
	char	*desc;		/* description */
	char	xtype;		/* external dtype value */
	char	itype;		/* corresponding internal dtype value */
	int	length;		/* length of external binary field */
	int	(*iproc)();	/* input->internal translation routine */
	int	(*oproc)();	/* internal->output translation routine */
} FORM;
	/*
	**	external format for ieee machines
	*/
static	FORM	ieee[]	= {
	"unsigned byte",	'b',	'b',	1,	0,	0,
	"signed byte",		's',	's',	1,	0,	0,
#ifdef sun
	"signed halfword",	'h',	'h',	2,	0,	0,
	"unsigned halfword",	'u',	'u',	2,	0,	0,
	"signed fullword",	'i',	'i',	4,	0,	0,
	"unsigned fullword",	'l',	'l',	4,	0,	0,
	"fullword float",	'f',	'f',	4,	0,	0,
	"doubleword float",	'd',	'd',	8,	0,	0,
	"complex float",	'c',	'c',	8,	0,	0,
#endif
#ifdef vax
	"signed halfword",	'h',	'h',	2,	swap2,	swap2,
	"unsigned halfword",	'u',	'u',	2,	swap2,	swap2,
	"signed fullword",	'i',	'i',	4,	swap4,	swap4,
	"unsigned fullword",	'l',	'l',	4,	swap4,	swap4,
	"fullword float",	'f',	'f',	4,	tovaxf,	fromvaxf,
	"doubleword float",	'd',	'd',	8,	tovaxd,	fromvaxd,
	"complex float",	'c',	'c',	8,	tovaxc,	fromvaxc,
#endif
	0
};
	/*
	**	external format for DEC 11/VAX
	*/
static	FORM	decvax[]	= {
	"unsigned byte",	'b',	'b',	1,	0,	0,
	"signed byte",		's',	's',	1,	0,	0,
#ifdef vax
	"signed halfword",	'h',	'h',	2,	0,	0,
	"unsigned halfword",	'u',	'u',	2,	0,	0,
	"signed fullword",	'i',	'i',	4,	0,	0,
	"unsigned fullword",	'l',	'l',	4,	0,	0,
	"fullword float",	'f',	'f',	4,	0,	0,
	"doubleword float",	'd',	'd',	8,	0,	0,
	"complex float",	'c',	'c',	8,	0,	0,
#endif
#ifdef sun
	"signed halfword",	'h',	'h',	2,	swap2,	swap2,
	"unsigned halfword",	'u',	'u',	2,	swap2,	swap2,
	"signed fullword",	'i',	'i',	4,	swap4,	swap4,
	"unsigned fullword",	'l',	'l',	4,	swap4,	swap4,
	"fullword float",	'f',	'f',	4,	fromvaxf,	tovaxf,
	"doubleword float",	'd',	'd',	8,	fromvaxd,	tovaxd,
	"complex float",	'c',	'c',	8,	fromvaxc,	tovaxc,
#endif
	0
};
	/*
	**	external format for IBM 360
	*/
static	FORM	ibm360[]	= {
	"unsigned byte",	'b',	'b',	1,	0,	0,
	"signed byte",		's',	's',	1,	0,	0,
#ifdef vax
	"signed halfword",	'h',	'h',	2,	swap2,	swap2,
	"unsigned halfword",	'u',	'u',	2,	swap2,	swap2,
	"signed fullword",	'i',	'i',	4,	swap4,	swap4,
	"unsigned fullword",	'l',	'l',	4,	swap4,	swap4,
	"fullword float",	'f',	'f',	4,	icvtf,	toibmf,
	"doubleword float",	'd',	'd',	8,	icvtd,	toibmd,
	"complex float",	'c',	'c',	8,	icvtc,	toibmc,
#endif
#ifdef sun
	"signed halfword",	'h',	'h',	2,	0,	0,
	"unsigned halfword",	'u',	'u',	2,	0,	0,
	"signed fullword",	'i',	'i',	4,	0,	0,
	"unsigned fullword",	'l',	'l',	4,	0,	0,
	"fullword float",	'f',	'f',	4,	icvtf,	toibmf,
	"doubleword float",	'd',	'd',	8,	icvtd,	toibmd,
	"complex float",	'c',	'c',	8,	icvtc,	toibmc,
#endif
	0
};
	/*
	**	table of recognized external formats
	**	the local machine format must appear first
	*/
static	struct	tab	{
	char	*desc;		/* CPU description */
	char	*name;		/* version name */
	FORM	*form;		/* allowed dtype formats */
} vertab[] = {
#ifdef vax
	"DEC 11/VAX",		"vax",		decvax,
	"IEEE",			"ieee",		ieee,
#endif
#ifdef sun
	"IEEE",			"ieee",		ieee,
	"DEC 11/VAX",		"vax",		decvax,
#endif
	"DEC 11/VAX",		"1.0",		decvax,
	"IEEE",			"2.0",		ieee,
	"IEEE",			"ieee",		ieee,
	"SUN 2/M86010",		"sun2",		ieee,
	"SUN 3/M68020",		"sun",		ieee,
	"IBM/360",		"ibm",		ibm360,
	"IBM/360",		"ibm360",	ibm360,
	0
};
	/*
	**	tell about the available transformations
	*/
static
inform()
{
	register struct tab *pt;
	register struct format *f;

	for (pt = vertab; pt->desc; pt++) {
		(void) printf("\n%s: version = %s\n\n", pt->desc, pt->name);
		(void) printf("\textern\tintern\textern\tinput\toutput\tfield\n");
		(void) printf("\tdtype\tdtype\tlength\tproc?\tproc?\tdescription\n");
		for (f = pt->form; f->desc; f++)
			(void) printf("\t%c\t%c\t%d\t%s\t%s\t%s\n",
				f->xtype, f->itype, f->length,
				f->iproc ? "yes" : "no",
				f->oproc ? "yes" : "no",
				f->desc);
	}
	(void) printf("\n");
	exit(0);
}
	/*
	**	return pointer to FORM entry
	*/
static	FORM *
gettab(io)
register io;
{	
	register struct tab *pt;
	register struct format *f;
	char	version[100];
	char	dtype;
	/*
	**	point pt at version table
	*/
	if (ihget("version", version, sizeof(version)) ||
	    ihget("dtype", &dtype, 1))
		gipserr();
	for (pt = vertab; pt->desc && strcmp(pt->name, version); pt++)
		;
	if (! pt->desc)
		error("unsupported version: %s", version);
	/*
	**	point f at supported dtype entry
	*/
	for (f = pt->form; f->desc; f++)
		if (dtype == (io ? f->itype : f->xtype))
			return(f);
	error("unsupported dtype: %.1s", &dtype);
	return(NULL);
}

main(argc, argv)

	int	argc;			/* argument count */
	char	**argv;			/* array of argument pointers */
{
	register FORM	*in;		/* input format pointer */
	register FORM	*out;		/* output format pointer */
	register FILE	*fp;		/* input image stream pointer */
	register FILE	*fq;		/* output image stream pointer */
	char		*inbuf;		/* input buffer */
	char		*outbuf;	/* output buffer */
	char		*inconv;	/* converted input buffer */
	char		*outconv;	/* pre-converted output buffer */
	int		xnum;		/* x-dimension */
	int		ynum;		/* y-dimension */
	/*
	**	save and parse callers arguments
	*/
	if (argc == 2 && strcmp(argv[1], "?") == 0)
		inform();
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...", argv[2]);
	if (argc > 1)
		ifile = *++argv;
	if (Sflag)
		telltale(0);
	/*
	**	read GIPS header and (optional) merge file
	*/
	if ((fp = ihread(ifile, GIPSVERSIONS)) == NULL)
		gipserr();
	if (fp == stdin)
		ifile = "'stdin'";
	/*
	**	lookup input version numbers
	*/
	in = gettab(0);
	(void)(ihput("version", vertab[0].name));
	/*
	**	possibly augment header
	*/
	if (*mfile != '-')
		if ((fq = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void)(ihbcl(fq));
	/*
	**	lookup output version number
	**	and fix up the output header
	*/
	out = gettab(1);
	if (ihput("dtype", &out->xtype) ||
	    ihget("xnum", (char *)&xnum, 1) ||
	    ihget("ynum", (char *)&ynum, 1))
		gipserr();
	if (in->iproc == out->oproc)
		in->iproc = out->oproc = 0;
	/*
	**	allocate buffers
	*/
	inbuf = alloc(in->length, xnum);
	inconv = outconv = in->iproc ? alloc(ihdtype(in->itype), xnum) : inbuf;
	if (in->itype != out->itype)
		outconv = alloc(ihdtype(out->itype), xnum);
	outbuf = out->oproc ? alloc(out->length, xnum) : outconv;
	/*
	**	write the output header
	*/
	putargs();
	if ((fq = ihwrit(ofile, 1)) == NULL)
		gipserr();
	else if (fq == stdout)
		(void)(signal(SIGPIPE, exit)), ofile = "'stdout'";
	/*
	**	copy and convert the rasters
	*/
	while (--ynum >= 0) {
		if (fread(inbuf, in->length, xnum, fp) != xnum)
			gipserr2(GIPSBADREAD, ifile);
		if (inbuf != inconv && (*in->iproc)(inbuf, xnum, inconv))
			error("input conversion error: %s", ifile);
		if (inconv != outconv)
			gconv(outconv, inconv, xnum,
				ihdcvt(out->itype), ihdcvt(in->itype));
		if (outbuf != outconv && (*out->oproc)(outconv, xnum, outbuf))
			error("output conversion error: %s", ofile);
		if (fwrite(outbuf, out->length, xnum, fq) != xnum)
			gipserr2(GIPSBADWRITE, ofile);
	}
	/*
	**	cleanup
	*/
	(void) ihbcl(fp);
	(void) ihbcl(fq);
	if (Sflag)
		telltale(1);
	(void)(exit(0));
}
