#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihfromwave.c,v $	GIPS $Revision: 1.3 $ (MIT/CSR) $Date: 2004/11/29 18:56:25 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1993 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihfromwave.c,v $		Creation Date:	04/08/93
**	Version:	$Revision: 1.3 $		Update:		$Date: 2004/11/29 18:56:25 $
**	Function:	Undo general image wavelet compression
**	Components:	main()		GIPS/UNIX main program
*/

#include <stdio.h>
#include <sys/syscall.h>
#include <stdlib.h>
#include "gips.h"

#define WAVE_GEOM "Wavelet dtype=%c xnum=%d ynum=%d nlvl=%d fact=%lf"

	/*
	**	Input variables
	*/
static	char	*ifile	= "-";		/* input GIPS file */
static	char	*mfile	= "-";		/* input GIPS header */
static	char	*ofile	= "-";		/* output WAVE file */
	flag	Dflag;			/* debug flag */
static	flag	Sflag;			/* statistics flag */

	/*
	**	Variable definitions
	*/

static	flagentry	flagtable[] = {
	"D",	"1",			(flag *) &Dflag,
	"S",	"1",			(flag *) &Sflag,
	"m",	"-> file",		(flag *) &mfile,
	"o",	"-> file",		(flag *) &ofile,
	NULL,	"[file]",		NULL
};

static	int	xNum, yNum;	/* input image size */
static	int	dLen;		/* length of input pixel */
static	char	dType;		/* pixel type */
static	FILE	*fIn;		/* input file stream */
static	long	lOff;		/* input read offset */

extern	char	*check_malloc();

parse_unepic_args(argc, argv, pFp, pnLev, pxDim, pyDim, pcFact, pFq)
	int	argc;		/* number of input parameters */
	char	**argv;		/* input arguments */
	FILE	**pFp;		/* output stream pointer */
	int	*pnLev;		/* number of compression levels */
	int	*pxDim;		/* input image x-dimension */
	int	*pyDim;		/* input image y-dimension */
	double	*pcFact;	/* compression factor */
	FILE	**pFq;		/* output stream pointer */
{
	char	aGeom[256];	/* geometry value */
	int	i;		/* scratch */

	/*
	**	Parse command string
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc == 2)
		ifile = argv[1];
	if (Sflag)
		telltale(0);

	/*
	**	Read the input header
	*/

	if ((*pFp = fIn = ihread(ifile, GIPSPROFILE)) == NULL)
		gipserr();
	else if (fIn == stdin)
		ifile = "'stdin'";
	lOff = -ftell(fIn);
	if (mfile != NULL && *mfile != '-')
		if (*pFq = ihmerg(mfile))
			(void)ihbcl(*pFq);
		else
			gipserr();
	if (ihget("xnum", pxDim) || ihget("ynum", pyDim))
		gipserr();
	if (ihget("dtype", &dType, 1) || dType != 'W')
		gipserr();
	if (ihget("geometry", aGeom, sizeof(aGeom)-1))
		gipserr();
	if (sscanf(aGeom, WAVE_GEOM,
		&dType, &xNum, &yNum, pnLev, pcFact) != 5)
		gipserr();
	if ((dLen = ihdtype(dType)) == 0)
		gipserr();

	/*
	**	Fix the output header
	*/

	ihput("xnum", &xNum);
	ihput("ynum", &yNum);
	ihput("dtype", &dType);
	ihdel("geometry");

	/*
	**	Open the output file
	*/

	if ((*pFq = ihwrit(ofile, 1)) == NULL)
		gipserr2(GIPSBADOPEN, ofile);

	/*
	**	Read compressed header
	*/

	if ((i = (getc(fIn) & 255)) != 0377)
		error("%s: not a wavelet-compressed image\n", ifile);
	if ((i = (getc(fIn) & 255)) != *pnLev)
		error("%s: inconsistent nlev in geometry\n", ifile);
	i = getc(fIn) & 255;
	if (((i << 8) | (getc(fIn) & 255)) != *pxDim)
		error("%s: inconsistent xnum in header\n", ifile);
	i = getc(fIn) & 255;
	if (((i << 8) | (getc(fIn) & 255)) != *pyDim)
		error("%s: inconsistent ynum in header\n", ifile);
	i = getc(fIn) & 255;
	*pcFact = (i << 8) | (getc(fIn) & 255);
}

write_byte_image(fq, pRes, xDim, yDim)
	FILE		*fq;
	int		*pRes;
	int		xDim;
	int		yDim;
{
	char		*pOut;
	register int	y;

	pOut = check_malloc(dLen * xNum);

	for (y = 0; y < yNum; y++) {
		gconv(pOut, &pRes[y*xDim], xNum, ihdcvt(dType), CVT_I);
		fwrite(pOut, xNum, dLen, fq);
	}

	ihbcl(fq);
	check_free((char *)pOut);

	if (Sflag)
		telltale(1);
}

close_ihwave(fp)
	FILE		*fp;
{
	return ihbcl(fp);
}

/* Temporary variables for use in writing/reading files. See epic.h. */
	char    temp_byte;
	short   temp_short;
	int     temp_int;

char
*check_malloc(size)
	int	size;
{
	char	*s;

	if ((s = malloc(size)) == NULL)
		error("insufficient memory", "");
	return s;
}

check_free(p)
	char	*p;
{
	if (p)
		free (p);
}

read(fid, buf, len)
	int	fid, len;
	char	*buf;
{
	int	rc;

	rc = syscall(SYS_read, fid, buf, len);
	if (rc > 0 && fIn && fid == fileno(fIn))
		lOff += rc;
	return rc;
}

long
ftell(fp)
	FILE	*fp;
{
	if (fp == fIn)
		return lOff - fp->_cnt;
	else
		return (long)lseek(fileno(fp), (off_t)0, 1) - fp->_cnt;
}
