#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihtowave.c,v $	GIPS $Revision: 1.3 $ (MIT/CSR) $Date: 2004/11/29 18:56:25 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1993 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: ihtowave.c,v $		Creation Date:	04/08/93
**	Version:	$Revision: 1.3 $		Update:		$Date: 2004/11/29 18:56:25 $
**	Function:	Convert general image to wavelet-compressed
**	Components:	parse_epic_args()	Input filter
*/

#include <stdio.h>
#include <sys/syscall.h>
#include <stdlib.h>
#include "gips.h"

#define WAVE_GEOM "Wavelet dtype=%c xnum=%d ynum=%d nlvl=%d fact=%g"

	/*
	**	Input variables
	*/
static	char	*iFile	= "-";		/* input GIPS file */
static	char	*mFile	= "-";		/* input GIPS header */
static	char	*oFile	= "-";		/* output WAVE file */
static	int	nLev	= 4;		/* number of levels */
static	int	cFact = 12;		/* bin size */
	flag	Dflag;			/* debug flag */
static	flag	Sflag;			/* statistics flag */

	/*
	**	Variable definitions
	*/

static	flagentry	flagtable[] = {
	"D",	"1",			(flag *) &Dflag,
	"S",	"1",			(flag *) &Sflag,
	"b",	"%d binsize",		(flag *) &cFact,
	"l",	"%d levels",		(flag *) &nLev,
	"m",	"-> file",		(flag *) &mFile,
	"o",	"-> file",		(flag *) &oFile,
	NULL,	"[file]",		NULL
};

static	FILE	*fOut;		/* output file stream */
static	long	lOff;		/* output read offset */

extern	char	*check_malloc();

parse_epic_args(argc, argv, ppImg, pxDim, pyDim, pnLev, pcFact, pFq)
	int	argc;		/* number of input parameters */
	int	*pxDim;		/* output image x-dimension */
	int	*pyDim;		/* output image y-dimension */
	int	*pnLev;		/* number of compression levels */
	char	**argv;		/* input arguments */
	float	**ppImg;	/* input image array */
	double	*pcFact;	/* compression factor */
	FILE	**pFq;		/* output stream pointer */
{

	FILE	*fp;		/* input file stream */
	int	xNum, yNum;	/* input image size */
	int	xyDim;		/* image size */
	int	dLen;		/* length of input pixel */
	char	dType;		/* pixel type */
	char	*pIn;		/* input buffer */
	char	aGeom[256];	/* scratch */
	int	i;		/* scratch */

	/*
	**	Parse command string
	*/

	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		error("multiple input file(s): %s ...",argv[2]);
	else if (argc == 2)
		iFile = argv[1];
	if (Sflag)
		telltale(0);

	if ((*pnLev = nLev) > 10)
		error("max -l value is %d", (char *)10);
	*pcFact = cFact;

	/*
	**	Read the input header
	*/

	if ((fp = ihread(iFile, GIPSPROFILE)) == NULL)
		gipserr();
	else if (fp == stdin)
		iFile = "'stdin'";
	if (mFile != NULL && *mFile != '-')
		if (fOut = ihmerg(mFile))
			(void)ihbcl(fOut);
		else
			gipserr();
	if (ihget("xnum", &xNum) || ihget("ynum", &yNum))
		gipserr();
	if (ihget("dtype", &dType, 1))
		dType = 'b';
	if ((dLen = ihdtype(dType)) == 0)
		gipserr();

	/*
	**	Check dimensions
	*/

	for (*pxDim = 1; *pxDim < xNum; *pxDim <<= 1)
		;
	for (*pyDim = 1; *pyDim < yNum; *pyDim <<= 1)
		;
	if (*pxDim < *pyDim)
		*pxDim = *pyDim;
	else if (*pyDim < *pxDim)
		*pyDim = *pxDim;
	if (*pxDim > (1<<15-1))
		error("max x-dimension is %d", (char *)(1<<15-1));
	if (*pyDim > (1<<15-1))
		error("max y-dimension is %d", (char *)(1<<15-1));
	if ((*pxDim % (1<<*pnLev)) || (*pyDim % (1<<*pnLev)) ||
		(*pxDim / (1<<*pnLev)) < 15 ||
		(*pyDim / (1<<*pnLev)) < 15)
		error("illegal image dimensions", "");

	/*
	**	Fix the output header
	*/

	ihput("xnum", pxDim);
	ihput("ynum", pyDim);
	ihput("dtype", "W");
	sprintf(aGeom, WAVE_GEOM, dType, xNum, yNum, *pnLev, *pcFact);
	ihput("geometry", aGeom);

	/*
	**	Read the input file, convert to float
	*/

	pIn = check_malloc(dLen*xNum);
	*ppImg = (float *)check_malloc(sizeof(float) * *pxDim * *pyDim);
	for (i = 0; i < yNum; i++) {
		fread(pIn, xNum, dLen, fp);
		gconv(*ppImg + *pxDim*i, pIn, xNum, CVT_F, ihdcvt(dType));
	}
	check_free(pIn);

	/*
	**	Open the output file
	*/

	putargs();
	if ((*pFq = fOut = ihwrit(oFile, 1)) == NULL)
		gipserr2(GIPSBADOPEN, oFile);
	lOff = -ftell(fOut);
}

	/*
	**	Called from main() to clean up
	*/

close_ihwave(fp)
	FILE	*fp;
{
	if (Sflag)
		telltale(1);
	return ihbcl(fp);
}

/* Temporary variables for use in writing/reading files. See epic.h. */
	char	temp_byte;   
	short	temp_short;
	int	temp_int;

char
*check_malloc(size)
	int	size;
{
	char	*s;

	if ((s = malloc(size)) == NULL)
		error("insufficient memory", "");
	return s;
}

check_free(p)
	char	*p;
{
	if (p)
		free(p);
}

write(fid, buf, len)
	int	fid, len;
	char	*buf;
{
	int	rc;

	rc = syscall(SYS_write, fid, buf, len);
	if (rc > 0 && fOut && fid == fileno(fOut))
		lOff += rc;
	return rc;
}

long
ftell(fp)
	FILE	*fp;
{
	if (fp == fOut)
		return lOff + (fp->_ptr - fp->_base);
	else
		return (long)lseek(fileno(fp), (off_t)0, 1) + (fp->_ptr - fp->_base);
}
