#ifndef lint
static char *sccsid = "%Z%$RCSfile: info.c,v $	GIPS $Revision: 2.6 $ (MIT/CSR) $Date: 1998/11/13 23:35:11 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1989 Massachusetts Institute of Technology
**
**	Module:		info.c			Creation Date:	07/19/89
**	Version:	$Revision: 2.6 $		Update:		$Date: 1998/11/13 23:35:11 $
**	Components:	info_events		menu handler
**			set_info_panel		create line of display
**			display_info		create info window
**
**	Function:	Display information window
*/

#include	"gipstool.h"
#include	"selections.h"

#define		FONT_MASK	0x8000	/* flag to specify font type */
	/*
	**	Sunview info popup structures
	*/
static	Frame		info_frame;	/* describes the info window */
static	Panel		info_panel;	/* describes the info panel */
static	Panel_item	info_item[64];	/* describes the info items */
static	int		info_line;	/* line counter */
static	void		(*info_proc)();	/* panel event handler */
	/*
	**	Event handler to pass menu button
	*/
static
handle_info_event(canvas, event, arg)
	Canvas	canvas;
	Event	*event;
	caddr_t	arg;
{
	switch (event_id(event)) {
	case MS_LEFT:
		info_proc(canvas, event, arg);
		break;
	case MS_MIDDLE:
		break;
	case MS_RIGHT:
		menu_show(menu, canvas, event, 0);
		break;
	default:
		if (event_is_ascii(event) && event_is_down(event))
			image_selected(image_sw, event, 0);
	}
}
	/*
	**	Panel button handler
	*/
static
handle_info_item(item, event)
	Panel_item	item;
	Event		*event;
{
	int	id;
	Event	ie;
	Rect	r;

	if (event_id(event) != MS_LEFT)
		return;
	id = (int)panel_get(item, PANEL_CLIENT_DATA);
	ie = *event;
	event_set_down(&ie);
	if (id & FONT_MASK) {
		event_set_id(&ie, 'F');
		image_selected(image_sw, &ie, 0);
		event_set_id(&ie, id & ~FONT_MASK);
		image_selected(image_sw, &ie, 0);
		id = ' ';
	}
	xv_set(image_sw, WIN_MOUSE_XY,
		(int)xv_get(image_sw, WIN_WIDTH)/2,
		(int)xv_get(image_sw, WIN_HEIGHT)/2,
		0);
	event_set_id(&ie, id);
	image_selected(image_sw, &ie, 0);
}
	/*
	**	Create a pair of panel messages on one line
	*/
static
set_info_panel(str1, value, str2)
	int	value;
	char	*str1, *str2;
{
	static	int	width;

	if (! info_frame) {
		info_frame = xv_create(base_frame, FRAME_CMD,
			WIN_SHOW,		FALSE,
			FRAME_SHOW_LABEL,	TRUE,
			WIN_RIGHT_OF,		base_frame,
			0);
		info_panel = (Panel)xv_get(info_frame, FRAME_CMD_PANEL);
		info_proc = (void (*)())xv_get(info_panel, WIN_EVENT_PROC);
		xv_set(info_panel,
			PANEL_ACCEPT_KEYSTROKE,	TRUE,
			WIN_EVENT_PROC,		handle_info_event,
			0);
		width = xv_col(info_panel, 4);
	}
	if (! info_item[info_line]) {
		int	y = 2*xv_row(info_panel,info_line)/3;
		int	dy = xv_row(info_panel,1)/8;
		if (value) {
			char *b, buf[3];
			if (value & FONT_MASK)
				sprintf(b = buf, "F%c", value & ~FONT_MASK);
			else
				b = form_char(value);
			(void)panel_create_item(info_panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	handle_info_item,
			PANEL_ITEM_X,		10,
			PANEL_ITEM_Y,		y,
			PANEL_LABEL_STRING,	b, 
			PANEL_LABEL_WIDTH,	15,
			PANEL_CLIENT_DATA,	(caddr_t)value,
			0);
		}
		(void)panel_create_item(info_panel, PANEL_MESSAGE,
			PANEL_ITEM_X,		10 + width + 10,
			PANEL_ITEM_Y,		y+dy,
			PANEL_LABEL_STRING,	str1,
			0);
		info_item[info_line] = panel_create_item(info_panel,
			PANEL_MESSAGE,
			PANEL_ITEM_X,		20+width+xv_col(info_panel,20),
			PANEL_ITEM_Y,		y+dy,
			0);
	}
	xv_set(info_item[info_line++],
		PANEL_LABEL_STRING,	str2 ? str2 : "none",
		0);
}
	/*
	**	Return file basename
	*/
static	char	*
basename(name)
	char	*name;
{
	char	*s, *rindex();

	if (name && (s = rindex(name, '/')))
		name = s+1;
	return name;
}
	/*
	**	Display image information
	*/
display_info()
{
	register i, n;
	int	create;
	static	char *title[] = { "Roman", "Italic", "Bold", "Symbol" };
	char	buf[256], *s, *label;

	if (no_display)
		return;
	create = info_frame == GIPSWIN(NULL);
	info_line = 0;
	set_info_panel("Frame title:", SET_FRAME_LABEL,
		label = (char *)xv_get(base_frame, FRAME_LABEL));
	set_info_panel("Program version:", 0,
		index(TOOL_NAME(3.2), '%') ? "test" : TOOL_NAME(3.2));
	set_info_panel("Original image:", 0,
		basename(imfile));
	(void)sprintf(buf, "%dx%d pixels", xnum, ynum);
	set_info_panel("Original size:", 0, buf);
	(void)sprintf(buf, "%dx%d pixels", xdim, ydim);
	set_info_panel("In-core size:", 0, buf);
	(void)sprintf(buf, "%d,%d", xorg, yorg);
	set_info_panel("In-core origin:", 0, buf);
	(void)sprintf(buf, "%dx%d pixels",
			image_rect.r_width, image_rect.r_height);
	set_info_panel("Display size:", SET_BOTTOM_RIGHT, buf);
	(void)sprintf(buf, "%d,%d", xoff, yoff);
	set_info_panel("Display origin:", 0, buf);
	(void)sprintf(buf, "%d,%d", image_rect.r_left, image_rect.r_top);
	set_info_panel("Display margins:", SET_TOP_LEFT, buf);
	(void)sprintf(buf, "%d,%d", image_mark[0], image_mark[1]);
	set_info_panel("Image mark:", SET_LINE_MARK, buf);
	set_info_panel("Colormap name:", 0, depth > 1 ? colormap : NULL);
	set_info_panel("Colormap file:", SET_COLORMAP,
		depth > 1 ? basename(afile) : NULL);
	(void)sprintf(buf, "%d",stretch_percent);
	set_info_panel("Color stretch:", SET_COLORMAP_STRETCH, buf);
	(void)sprintf(buf, "%d", graphic_index);
	set_info_panel("Graphic index:", ADD_GRAPHIC, buf);
	(void)sprintf(buf, "%d", brush_size);
	set_info_panel("Brush size:", SET_PAINT_SIZE, buf);
	(void)sprintf(buf, "%d", graphic_color);
	set_info_panel("Graphic color:", SET_TEXT_COLOR, buf);
	(void)sprintf(buf, "%d degrees", graphic_angle);
	set_info_panel("Graphic angle", SET_TEXT_ANGLE, buf);
	for (n = 0; n < 2; n++) {
		buf[0] = 0;
		for (i = 0; i < NGRAPHICS; i++)
			if ((n == 0 && graphic_sw[i] == 1) ||
					(n == 1 && graphic_sw[i] > 1))
				(void)sprintf(buf+strlen(buf), "%d ", i);
		set_info_panel(n ? "Indices `off':" : "Indices `on': ",
			ADD_GRAPHIC, buf[0] ? buf : (char *)NULL);
	}
	for (i = 0; i < 4; i++) {
		(void)sprintf(buf, "%s font:", title[i]);
		set_info_panel(buf,
			FONT_MASK | title[i][0], basename(setauxfont(0, i)));
	}
	sprintf(buf, "%d%s", graphic_origin, rotated_font ? " (rotated)" : "");
	set_info_panel("Text orientation:", SET_TEXT_ORIGIN, buf);
	set_info_panel("Command log file:", LOG_INPUT_COMMANDS,
		ofile && *ofile ? basename(ofile) : NULL);
	set_info_panel("Pixel log file:", BEGIN_PIXEL_LOGGING,
		pixel_stream ? basename(pixel_log_file) : NULL);
	set_info_panel("Tag file:", READ_TAG_FILE,
		basename(tag_file));
	if (i = command_index[left_button_value[0] & 0377]) {
		sprintf(buf, "[%s] %s", command_text[i-1], left_button_value+1);
		set_info_panel("Left button:", SET_LEFT_BUTTON, buf);
	} else
		set_info_panel("Left button:", SET_LEFT_BUTTON, "unassigned");
	set_info_panel("Middle button:", 0,
		"display zoomed insert");
	set_info_panel("Right button:", 0,
		"display pop-up menus");
	window_fit(info_panel);
	window_fit(info_frame);

	if (create) {
		Rect	r, v;
		frame_get_rect(base_frame, &r);
		frame_get_rect(info_frame, &v);
		xv_set(info_frame,
			XV_X, MAX(MIN(display_rect.r_width - v.r_width,
				r.r_left + r.r_width), 0),
			XV_Y, MAX(r.r_top, 0),
			0);
	}

	xv_set(info_frame,
		FRAME_LABEL,	label,
		WIN_SHOW,	TRUE,
		0);
}

update_info()
{
	if (info_frame && (int)xv_get(info_frame, WIN_SHOW))
		display_info();
}

set_info_frame(show)
	int	show;
{
	if (info_frame)
		if ((int)xv_get(info_frame, FRAME_CMD_PUSHPIN_IN) == show)
			xv_set(info_frame, XV_SHOW, show, 0);
}
