#ifndef lint
static char *sccsid = "%Z%$RCSfile: main.c,v $	GIPS $Revision: 2.6 $ (MIT/CSR) $Date: 1998/05/13 01:34:00 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		main.c		Creation Date:	03/19/86
**	Version:	$Revision: 2.6 $		Update:		$Date: 1998/05/13 01:34:00 $
**	Function:	Display GIPS image within customized Sun window
**	Components:	main()		GIPS/UNIX SunTool main program
*/


#include        <stdlib.h>    /* for getenv */
#define		MAIN
#include	"gipstool.h"
#include	<signal.h>
	/*
	**	Pointers to compound input variables
	*/
static	char	*xbuf;		/* input pixel width, origin, offset */
static	char	*ybuf;		/* input pixel width, origin, offset */
static	char	*xview;		/* output window size, left-right border */
static	char	*yview;		/* output window size, left-right border */
static	int	nring;		/* number of ring buffers for dtype=G */
static	flag	Kflag;		/* run as zombie */
	/*
	**	table used by gipsflags()
	*/
static	flagentry	flagtable[] = {
	"",	"%d depth",		(flag *) &depth,
	"C",	"-> map",		(flag *) &colormap,
	"D",	"1",			(flag *) &Dflag,
	"I",	"-> file",		(flag *) &gipsrc,
	"K",	"1",			(flag *) &Kflag,
	"N",	"1",			(flag *) &Nflag,
	"S",	"1",			(flag *) &Sflag,
	"X",	"-> size,left,right",	(flag *) &xview,
	"Y",	"-> size,left,right",	(flag *) &yview,
	"a",	"-> file",		(flag *) &afile,
	"c",	"1",			(flag *) &cflag,
	"e",	"-> coeffs",		(flag *) &ecoeffs,
	"i",	"-> file",		(flag *) &ifile,
	"m",	"-> file",		(flag *) &mfile,
	"n",	"%d colors",		(flag *) &colors,
	"o",	"-> file",		(flag *) &ofile,
	"r",	"%d ringbufs",		(flag *) &nring,
	"x",	"-> size,org,off",	(flag *) &xbuf,
	"y",	"-> size,org,off",	(flag *) &ybuf,
	"z",	"%d xzoom",		(flag *) &xzoom,
	"Z",	"%d yzoom",		(flag *) &yzoom,
	NULL,	"[file]",		NULL
};
extern	Pixfont	*xv_pf_default();
	/*
	**	The tool icon
	*/
static	short	image_icon[] = {
#include "gipstool.icon"
};
mpr_static(icon, ICON_DEFAULT_WIDTH, ICON_DEFAULT_HEIGHT, 1, image_icon);

static	Rect	icon_rect = { 0, 48, 64, 16 };
	/*
	**	The usual image cursor
	*/
static	short	image_cursor_array[16] = {
	0x8000,0xC000,0xE000,0xF000,0xF800,0xFC00,0xFE00,0xF000,
	0xD800,0x9800,0x0C00,0x0C00,0x0600,0x0600,0x0300,0x0300
};
mpr_static(image_cursor_pr, 16, 16, 1, image_cursor_array);
	/*
	**	Cursor displayed while dragging
	*/
static	short	drag_cursor_array[16] = {
	0x0FE0,0x07C0,0x07C0,0x0380,0x8102,0xC106,0xE00E,0xFC7E,
	0xE00E,0xC106,0x8102,0x0100,0x0380,0x07C0,0x0FE0,0x0000
};
mpr_static(drag_cursor_pr, 16, 16, 1, drag_cursor_array);
	/*
	**	The hourglass cursor
	*/
static	short	hourglass_cursor_array[16] = {
	0x7FFE,0x4002,0x200C,0x1A38,0x0FF0,0x07E0,0x03C0,0x0180,
	0x0180,0x0240,0x0520,0x0810,0x1108,0x23C4,0x47E2,0x7FFE
};
mpr_static(hourglass_cursor_pr, 16, 16, 1, hourglass_cursor_array);

static	Notify_value	base_frame_notifier();
static	char		*geom[] = { "", "-geometry", "-0-0" };
	/*
	**	Start here
	*/
main(argc, argv)

	int	argc;		/* number of arguments */
	char	**argv;		/* array of argument pointers */
{
	FILE	*fm;		/* input GIPS merge-header file stream */
	char	buf[BUFLEN];	/* scratch */
	char	*icon_name;	/* title for icon */
	int	width;		/* initial width of image sub-window */
	int	height;		/* initial height of image sub-window */
	int	nolocate = 0;	/* =1 if user specifies window location */
	short	rborder	= 0;	/* right viewport border */
	short	bborder	= 0;	/* bottom viewport border */
	char	*c;		/* scratch */
	int	i;		/* scratch */
	/*
	**	Check for window placement override
	*/
	for (i = 1; i < argc && argv[i]; i++)
		if (!strcmp(argv[i], "-Wp") ||
			!strcmp(argv[i], "-position") ||
				!strcmp(argv[i], "-geometry"))
			nolocate++;
	/*
	**	parse tool-specific arguments
	*/
	geom[0] = argv[0];

        printf( "xgips: before xv_init \n" );


	xv_init(XV_INIT_ARGS, nolocate ? 1 : 3, geom,
		XV_INIT_ARGC_PTR_ARGV, &argc, argv,
		0);

        printf( "xgips: after xv_init \n" );

	/*
	**	save and parse callers arguments
	*/
	saveargs(argv);
	argc = gipsflags(argc, argv, flagtable);
	if (argc > 2)
		Error("multiple input file(s): %s ...", argv[2]);
	if (argc == 2)
		imfile = argv[1];
	if (depth != 0 && depth != 1 && depth != 8)
		Error("bad pixel depth: %d", (char *)depth);
	if (Sflag)
		telltale(0);
	/*
	**	do we run as a zombie?
	*/
	if (Kflag && fork()) {
		i = getdtablesize();
		while (i >= 0)
			(void)close(i--);
		exit(0);
	}
	/*
	**	read input GIPS header and possible merge file
	*/

        printf( "xgips: before read input GIPS header \n" );

	if ((fp = ihread(imfile, GIPSPROFILE)) == NULL)
		gipserr();
	if (*mfile != '-')
		if ((fm = ihmerg(mfile)) == NULL)
			gipserr();
		else
			(void) ihbcl(fm);
	/*
	**	If we are reading the standard input,
	**	get the title from the GIPS header.
	**	Then construct a short name for the icon
	*/
	if (fp == stdin)
		imfile = save_string(ihget("title", buf, BUFLEN-1) ?
			"'stdin'" : buf, (char *)NULL);
	if (imfile[0] == '\'')
		icon_name = "GipsTool";
	else if (icon_name = rindex(imfile, '/'))
		icon_name++;
	else
		icon_name = imfile;

        printf( "xgips: before xv_create  base_frame \n" );

	base_frame = xv_create(GIPSWIN(NULL), FRAME,
		WIN_ERROR_MSG,		"Unable to create window.",
		FRAME_ICON,	xv_create(GIPSWIN(NULL), ICON,
					ICON_IMAGE,	&icon,
					ICON_LABEL,	icon_name,
					ICON_LABEL_RECT,&icon_rect,
					0),
		0);

        printf( "xgips: after xv_create  base_frame \n" );

	/*
	**	If input image is seek-able, record its location
	*/
	if (cflag && (strncmp(imfile, "/dev/", 5) == 0 ||
		(image_seek = ftell(fp)) <= 0 || ferror(fp)))
		cflag = 0;
	/*
	**	get necessary header variable values and check them out
	*/
	if (ihget("dtype", &dtype, 1))
		gipserr();
	/*
	**	see whether this is a cellular image
	*/
	if (dtype == 'G') {
		(void)ihget("xnum", (char *)&xdim, 0);
		(void)ihget("ynum", (char *)&ydim, 0);
		if ((input_gimage = gopen(fp, nring)) == NULL)
			gipserr2(GIPSBADOPEN, imfile);
		if (ihget("geometry", buf, sizeof(buf)))
			gipserr();
		dtype = (buf[5] == 'M') ? 'c' : buf[5];
		(void)ihput("dtype", &dtype);
		cflag++;
	}
	if (ihdtype(dtype) == 0)
		gipserr1(GIPSBADVAL);
	if (ihget("xnum", (char *)&xnum, 0) ||
			ihget("ynum", (char *)&ynum, 0))
		gipserr();
	map_sw = (g_setup() == 0);
	/*
	**	get input windowing parameters
	*/

        printf( "xgips: before get input windowing parameters \n" );

	if (xbuf && sscanf(xbuf, "%d,%d,%d", &xdim, &xorg, &xoff) < 1)
		Error("bad x-pixel origin, size, or offset: %s", xbuf);

	if (ybuf && sscanf(ybuf, "%d,%d,%d", &ydim, &yorg, &yoff) < 1)
		Error("bad y-pixel origin, size, or offset: %s", ybuf);

	if (xorg != NIX) (void) ihput("xorg", (char *)&xorg);
	else if (ihget("xorg", (char *)&xorg, 0)) xorg = 0;

	if (xdim != NIX) (void) ihput("xlim", (char *)&xdim);
	else if (ihget("xlim", (char *)&xdim, 0)) xdim = xnum;

	if (yorg != NIX) (void) ihput("yorg", (char *)&yorg);
	else if (ihget("yorg", (char *)&yorg, 0)) yorg = 0;

	if (ydim != NIX) (void) ihput("ylim", (char *)&ydim);
	else if (ihget("ylim", (char *)&ydim, 0)) ydim = ynum;
	/*
	**	inspect window limits for consistency
	*/
	if (xorg < 0 || xorg >= xnum)
		Error("x-limits outside image", "");
	if (yorg < 0 || yorg >= ynum)
		Error("y-limits outside image", "");

        printf( "xgips: xdim = %d, ydim = %d \n", xdim, ydim );

	/*
	**	Determine size of saved pixrect
	*/
	if (yzoom == 0 && xzoom == 1)
		Zflag++, yzoom = 1;
	else if (yzoom == NIX)
		yzoom = xzoom;
	if (xzoom < 1)
		Error("bad -z value: %d", (char *)xzoom);
	if (yzoom < 1)
		Error("bad -Z value: %d", (char *)yzoom);
	image_rect.r_width = saved_rect.r_width = xzoom*MIN(xnum-xorg, xdim);
	image_rect.r_height = saved_rect.r_height = yzoom*MIN(ynum-yorg, ydim);
	/*
	**	Read user-supplied viewport dimensions
	*/
	if (xview && sscanf(xview, "%hd,%hd,%hd",
			&image_rect.r_width, &image_rect.r_left, &rborder) < 1)
		Error("bad x-view size, or left/right border: %s", xview);
	if (yview && sscanf(yview, "%hd,%hd,%hd",
			&image_rect.r_height, &image_rect.r_top, &bborder) < 1)
		Error("bad y-view size, or left/right border: %s", yview);
	/*
	**	create the base_frame, using the command name if
	**	TOOL_NAME is an SCCS temporary.
	*/

        printf( "xgips: before xv_get \n" );

	display_rect = *((Rect *)xv_get(base_frame, WIN_SCREEN_RECT));

	width = MIN(image_rect.r_width + image_rect.r_left + rborder,
			display_rect.r_width
				- (int)xv_get(base_frame, WIN_LEFT_MARGIN)
				- (int)xv_get(base_frame, WIN_RIGHT_MARGIN));
	height = MIN(image_rect.r_height + image_rect.r_top + bborder,
		display_rect.r_height
			- (int)xv_get(base_frame, WIN_TOP_MARGIN)
			- (int)xv_get(base_frame, XV_BOTTOM_MARGIN));

        printf( "xgips: width = %d, height = %d \n", width, height );

	if (depth == 0)
		depth = xv_get(base_frame, WIN_DEPTH);

	if (depth == 1) {
		if (xzoom > 4)
			Error("-z value too large: %d", (char *)xzoom);
		if (xzoom != yzoom)
			Error("xzoom != yzoom not implemented for depth=1", "");
	}
	image_cursor = cursor_create(CURSOR_IMAGE, &image_cursor_pr, 0);
	drag_cursor = cursor_create(CURSOR_IMAGE, &drag_cursor_pr,
		CURSOR_XHOT, 7, CURSOR_YHOT, 7, 0);
	hourglass_cursor = cursor_create(CURSOR_IMAGE, &hourglass_cursor_pr, 0);
	/*
	**	create a canvas sub-window and associated pixwin
	*/

        printf( "xgips: before image_sw xv_create \n");

	image_sw = xv_create(base_frame, CANVAS,
		WIN_WIDTH,		width,
		WIN_HEIGHT,		height,
		CANVAS_AUTO_SHRINK,	TRUE,
		CANVAS_AUTO_EXPAND,	TRUE,
		CANVAS_AUTO_CLEAR,	FALSE,
		WIN_DYNAMIC_VISUAL,	Nflag ? TRUE : FALSE,
		WIN_CURSOR,		image_cursor,
		WIN_ERROR_MSG,		"Can't create image sub-window.",
		0);
	c = rindex(imfile, '/');
	xv_set(base_frame,
		FRAME_LABEL,	c ? c+1 : imfile,
		0);

        printf( "xgips: before canvas_paint_window \n" );

	image_pw = (Pixwin *)canvas_paint_window(image_sw);
	xv_set((int)image_pw,
		WIN_WIDTH,		width,
		WIN_HEIGHT,		height,
		CANVAS_CMS_REPAINT,	TRUE,
		CANVAS_REPAINT_PROC,	image_repaint,
		WIN_EVENT_PROC,		image_selected, 
		WIN_CONSUME_EVENTS,
			WIN_NO_EVENTS,
			WIN_ASCII_EVENTS,
			WIN_MOUSE_BUTTONS,
			WIN_UP_EVENTS,
			LOC_WINENTER,
			LOC_WINEXIT,
			LOC_DRAG,
			KEY_UP_HALF,
			KEY_DOWN_HALF,
			KEY_LEFT_HALF,
			KEY_RIGHT_HALF,
			KEY_TOP_EDGE,
			KEY_BOTTOM_EDGE,
			KEY_LEFT_EDGE,
			KEY_RIGHT_EDGE,
			KEY_REDRAW,
			0,
		0);

        
        printf( "xgips: after canvas_paint_window .....123..... \n" );

        printf( "xgips: before xv_get \n" );

	image_rect.r_width = (int)xv_get(image_sw, WIN_WIDTH)
		- image_rect.r_left - rborder;

        printf( "xgips: after xv_get width \n" );

	image_rect.r_height = (int)xv_get(image_sw, WIN_HEIGHT)
		- image_rect.r_top - bborder;

        printf( "xgips: after xv_get height \n" );

        printf( "xgips:  image_rect.r_width = %d, image_rect.r_height = %d \n", 
                image_rect.r_width, image_rect.r_width );

        printf( "xgips: before window_fit \n" );


	window_fit(base_frame);

        printf( "xgips: nolocate = %d \n", nolocate );

	if (! nolocate) {
		Rect	r;
		frame_get_rect(base_frame, &r);
		xv_set(base_frame,
			XV_X,	MAX(0, (display_rect.r_width - r.r_width)/2),
			XV_Y,	MAX(0, (display_rect.r_height - r.r_height)/2),
			0);
	}

        printf( "xgips: after nolcate section \n" );

	/*
	**	verify colormap depth
	*/

        printf( "xgips: before verify colormap depth \n" );

	if (Nflag || Zflag)
		colors = 256;
	for (cshift = 0, i = 256; i != colors; i >>= 1, cshift++)
		if (i == 2)
			Error("illegal number of colors: %d", (char *)colors);


        printf( "xgips: after verify colormap depth \n" );

	/*
	**	load the colormap and install the menu tree and event mask
	*/
	graphic_color = 255;
	graphic_sw[0] = 1;

        printf( "xgips: color depth = %d \n", depth );

        printf( "xgips: afile = %s, init_color() = %d \n", 
                 afile, init_color() );


	if (depth > 1 && init_color())
		perror(afile), exit(1);

        printf( "xgips: gipsdir() = %s, gipsprofdir() = %s, GIPSPROFDIR = %s, getenv(\"GIPSDIR\") = %s \n", 
                 gipsdir(), gipsprofdir(), GIPSPROFDIR, getenv("GIPSDIR") );

        printf( "xgips: before setup_command_menu \n" );

	setup_command_menu();

        printf( "xgips: after setup_command_menu \n" );

	default_font = xv_pf_default();

        printf( "xgips: after xv_pf_default \n" );

	(void)setfont(strcpy(font, GIPSTOOLFONT));
	(void)setauxfont("S.6", 3);

        printf( "xgips: after setauxfont \n" );


	/*
	**	If requested, open the logging file
	*/

        printf( "xgips: before open logging file \n" );

	if (ofile != NULL) {
		ofile = save_string(ofile, (char *)NULL);
		if ((output_stream = fopen(ofile, "w")) == NULL)
			gipserr2(GIPSBADOPEN, ofile);
	}
	/*
	**	Interpose a general trap
	*/
	(void)notify_interpose_event_func(base_frame,
		base_frame_notifier, NOTIFY_SAFE);
	/*
	**	Finally, enable interrupts and start up the tool
	**	Let the window manager read the image
	*/

        printf( "xgips: before window_main_loop near exit(0) \n" );

        printf( "xgips: sizeof(  base_frame ) = %d \n", sizeof( base_frame ) );

	window_main_loop(base_frame);


	exit(0);
}

	/*
	**	Frame notifier
	*/

static	Notify_value
base_frame_notifier(frame, event, arg, type)
	Frame			frame;
	Event			*event;
	Notify_arg		arg;
	Notify_event_type	type;
{
	int		old, new;
	Notify_value	ret;

	old = (int)xv_get(frame, FRAME_CLOSED);
	ret = notify_next_event_func(frame, (int)event, arg, type);
	new = (int)xv_get(frame, FRAME_CLOSED);
	if (old != new) {
		set_palette_frame(old);
		set_info_frame(old);
		set_tag_frame(old);
		set_zoom_value_frames(old);
	}
	return ret;
}
