#ifndef lint
static char *sccsid = "%Z%$RCSfile: mempix.c,v $   GIPS $Revision: 1.1 $ (MIT/CSR) $Date: 1998/05/13 01:34:49 $";
#endif
/*                 General Image Processing System
**                         Peter G. Ford
**                    Center for Space Research
**      Copyright 1998 Massachusetts Institute of Technology
**
**	Module:		$RCSfile: mempix.c,v $		Creation Date:	04/29/98
**	Version:	$Revision: 1.1 $		Update:		$Date: 1998/05/13 01:34:49 $
**	Components:	mem_rop		copy a memory pixrect
**			mem_get		return pixel value
**			mem_put		store pixel value
**			mem_vec		write vector into pixrect
**			mem_create	create memory pixrect
**			mem_point	create secondary memory pixrect
**			mem_destroy	remove memory pixrect
**
**	Function:	Provide some memory pixrect methods for XView
*/

#include <string.h>     /* for memcpy */
#include <pixrect/pixrect.h>
#include <pixrect/memvar.h>

Pixrect *xv_mem_create();
Pixrect *xv_mem_point();
int xv_mem_destroy();

int
mem_rop( dst, x, y, w, h, op, src, x0, y0 )
  Pixrect *dst, *src;
  int x, y, w, h, x0, y0, op;
{
  if (src && dst) {
    register struct mpr_data *mdsrc = mpr_d(src);
    register struct mpr_data *mddst = mpr_d(dst);
    if (mdsrc && mddst) {
      if (src->pr_depth == 8 && dst->pr_depth == 8 && op == PIX_SRC) {
	register unsigned char *dstdata = (unsigned char *)(mddst->md_image);
	register unsigned char *srcdata = (unsigned char *)(mdsrc->md_image);
	x0 += mdsrc->md_offset.x;
	y0 += mdsrc->md_offset.y;
	x += mddst->md_offset.x;
	y += mddst->md_offset.y;
	dstdata += y*mddst->md_linebytes+x;
	srcdata += y0*mdsrc->md_linebytes+x0;
	if (h > (dst->pr_size.y-y))
	  h = dst->pr_size.y-y;
	if (h > (src->pr_size.y-y0))
	  h = src->pr_size.y-y0;
	if (w > (dst->pr_size.x-x))
	  w = dst->pr_size.x-x;
	if (w > (src->pr_size.x-x0))
	  w = src->pr_size.x-x0;
	while (--h >= 0) {
	  memcpy(dstdata, srcdata, w);
	  dstdata += mddst->md_linebytes;
	  srcdata += mdsrc->md_linebytes;
	}
      } else if (src->pr_depth == 1 && dst->pr_depth == 1 && op == PIX_SRC) {
	for (; --h >= 0; y++, y0++) {
	  if (y >= dst->pr_size.y || y0 >= src->pr_size.y)
	    break;
	  if (y >= 0 && y0 >= 0) {
	    register int xx = x, xx0 = x0, ww = w;
	    for (; --w >= 0; xx++, xx0++) {
	      if (xx >= dst->pr_size.x || xx0 >= src->pr_size.x)
		break;
	      if (xx >= 0 && xx0 >= 0)
		mem_put(dst, xx, y, mem_get(src, xx0, y0));
	    }
	  }
	}
      } else {
	(*src->pr_ops->pro_stencil)();
      }
    }
  }
  return 0;
}

int
mem_get( pr, x, y )
  Pixrect *pr;
  int x, y;
{
  register int val = 0;

  if (pr) {
    register struct mpr_data *md = mpr_d(pr);
    if (md) {
      x += md->md_offset.x;
      y += md->md_offset.y;
      if (x >= 0 && x < pr->pr_size.x && y >= 0 && y < pr->pr_size.y) {
	register unsigned char *data =
	  (unsigned char *)(md->md_image)+y*md->md_linebytes;
	switch (pr->pr_depth) {
	case 1:
	  val = (data[x >> 3] >> (7 - (x & 7))) & 1;
	  break;
	case 8:
	  val = data[x];
	  break;
	default:
	  (*pr->pr_ops->pro_stencil)();
	  break;
	}
      }
    }
  }
  return val;
}

int
mem_put( pr, x, y, val )
  Pixrect *pr;
  int x, y, val;
{
  register int rc = 0;

  if (pr) {
    register struct mpr_data *md = mpr_d(pr);
    if (md) {
      x += md->md_offset.x;
      y += md->md_offset.y;
      if (x >= 0 && x < pr->pr_size.x && y >= 0 && y < pr->pr_size.y) {
	register unsigned char *data =
	  (unsigned char *)(md->md_image)+y*md->md_linebytes;
	switch (pr->pr_depth) {
	case 1:
	  data[x >> 3] |= (val & 1) << (7 - (x & 7));
	  break;
	case 8:
	  data[x] = val;
	  break;
	default:
	  (*pr->pr_ops->pro_stencil)();
	  break;
	}
      }
    }
  }
  return 0;
}

int
mem_vec( pr, x1, y1, x2, y2, op, color )
  Pixrect *pr;
  int x1, y1, x2, y2, op, color;
{
  if (pr) {
    register struct mpr_data *md = mpr_d(pr);
    if (md && (pr->pr_depth == 1 || pr->pr_depth == 8) && op == PIX_SRC) {
      register unsigned char *data = (unsigned char *)(md->md_image);
      register int x, y, dx = x2 - x1, dy = y2 - y1;
      x1 += md->md_offset.x;
      y1 += md->md_offset.y;
      x2 += md->md_offset.x;
      y2 += md->md_offset.y;
      if (abs(dy) > abs(dx)) {
        data += y1*md->md_linebytes;
        for (x = x1, y = y1; ; x = (x1*(y2-y)+x2*(y-y1)+dy/2)/dy) {
	  if (x >= 0 && x < pr->pr_size.x && y >= 0 && y < pr->pr_size.y)
	    if (pr->pr_depth == 8)
	      data[x] = color;
	    else
	      data[x>>3] |= (color & 1) << (7 - (x & 7));
	  if (y == y2)
	    break;
	  else if (dy > 0) {
	    y++;
	    data += md->md_linebytes;
	  } else {
	    y--;
	    data -= md->md_linebytes;
	  }
        }
      } else {
	for (x = x1, y = y1; ; y = (y1*(x2-x)+y2*(x-x1)+dx/2)/dx) {
	  if (x >= 0 && x < pr->pr_size.x && y >= 0 && y < pr->pr_size.y)
	    if (pr->pr_depth == 8)
	      data[y*md->md_linebytes+x] = color;
	    else
	      data[y*md->md_linebytes+(x>>3)] |= (color & 1) << (7 - (x & 7));
	  if (x == x2)
	    break;
	  else if (dx > 0)
	    x++;
	  else
	    x--;
	}
      }
    } else {
      (*pr->pr_ops->pro_stencil)();
    }
  }
  return 0;
}

Pixrect *
mem_create(w, h, depth)
    int w, h, depth;
{
  Pixrect *pr = xv_mem_create(w, h, depth);

  if (pr && pr->pr_ops) {
    pr->pr_ops->pro_rop = mem_rop;
    pr->pr_ops->pro_get = mem_get;
    pr->pr_ops->pro_put = mem_put;
    pr->pr_ops->pro_vector = mem_vec;
  }

  return(pr);
}

Pixrect *
mem_point(w, h, depth, image)
  int w, h, depth;
  short	*image;
{
  Pixrect *pr = xv_mem_point(w, h, depth, image);

  if (pr && pr->pr_ops) {
    pr->pr_ops->pro_rop = mem_rop;
    pr->pr_ops->pro_get = mem_get;
    pr->pr_ops->pro_put = mem_put;
    pr->pr_ops->pro_vector = mem_vec;
  }

  return(pr);
}

int
mem_destroy(pr)
  Pixrect *pr;
{
  return xv_mem_destroy(pr);
}
