#ifndef lint
static char *sccsid = "%Z%$RCSfile: menu.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 2004/11/29 20:08:51 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		menu.c		Creation Date:	03/19/86
**	Version:	$Revision: 2.5 $		Update:		$Date: 2004/11/29 20:08:51 $
**	Components:	gips_win_error()	error handler
**			set_gips_error()	initialize error handler
**			display_help_window()	add help window to display
**			setup_command_menu()	read command file
**
**	Function:	Handle menu prompts and help messages
*/

#include	"gipstool.h"
#include	<xview/panel.h>
#include	<xview/notice.h>
	/*
	**	Static
	*/
static	Event	menu_event;		/* event that displayed the menu */

	/*
	**	Menu done
	*/

/* ARGSUSED */
void
canvas_menu_done(done_menu, result)
	Menu	done_menu;
	int	result;
{
	if (done_menu != menu)
		menu_destroy(done_menu);
	else {
		event_set_id(&menu_event, result);
		image_selected(image_sw, &menu_event, NULL);
	}
}

	/*
	**	read the menu structures
	*/
setup_command_menu()
{
	off_t	l, lseek();
	int	menu_count;
	char	*buf, file[256], **sub_menu_name;
	Menu	*sub_menu, parent_menu;
	register i, n;
	register char *c;
	/*
	**	Firstly, read the menu file into *buf
	*/
	(void) sprintf(file, GIPSTOOLMENU);
	if ((i = open(file, 0)) < 0 || (l = lseek(i, (off_t)0, 2)) < 0)
		perror(file), exit(1);
	if ((buf = alloc((int)l+2, 1)) == NULL)
		exit(2);
	if (lseek(i, (off_t)0, 0) != 0 || read(i, buf, (int)l) != l)
		perror(file), exit(1);
	(void)close(i);
	/*
	**	Determine number of root menus
	*/
	for (c = buf, menu_count = 0; *c != '\n'; *c++ = 0, menu_count++)
		c = index(c, '\n');
	/*
	**	Allocate menu structures
	*/
	menu = menu_create(MENU_TITLE_ITEM, "Gipstool",
#ifdef	MENU_LEFT_MARGIN
		MENU_LEFT_MARGIN,	4,
#endif	MENU_LEFT_MARGIN
		0);
	sub_menu = (Menu *)alloc(sizeof(Menu), menu_count);
	sub_menu_name = (char **)alloc(sizeof(char *), menu_count);
	/*
	**	Fill the menu structure fields
	*/
	for (i = 0, c = buf; *c != '\n'; c++, i++) {
		while (isspace(*c)) c++;
		sub_menu_name[i] = c;
		while (!isspace(*c)) c++;
		*c++ = 0;
		while (isspace(*c)) c++;
		sub_menu[i] = menu_create(MENU_TITLE_ITEM, c,
#ifdef	MENU_LEFT_MARGIN
			MENU_LEFT_MARGIN,       4,
#endif	MENU_LEFT_MARGIN
			0);
		sprintf(file, "   %s", c);
		menu_set(menu, MENU_PULLRIGHT_ITEM,
			save_string(file, (char *)NULL),
			sub_menu[i],
			0);
		c += strlen(c);
	}
	/*
	**	Count the number of sub-menu items
	*/
	for (buf = c, i = 0; *++c; i++)
		do c = index(c, '\n');
			while (*++c && *c != '\n');
	/*
	**	Allocate pointer tables to the sub-menu items
	*/
	command_prompt = (char **)alloc(sizeof(char *), i);
	command_text = (char **)alloc(sizeof(char *), i);
	command_help = (char **)alloc(sizeof(char *), i);
	/*
	**	Scan the items again and fill structures
	*/
	for (c = buf, i = 0; *++c; i++) {
		/*
		**	First determine which menu it belongs in
		*/
		parent_menu = GIPSWIN(NULL);
		for (n = 0; ! parent_menu && n < menu_count; n++) {
			char *d = sub_menu_name[n];
			if (strncmp(d, c, strlen(d)) == 0)
				parent_menu = sub_menu[n];
		}
		/*
		**	Now isolate any command prompt
		*/
		while (!isspace(*c)) c++;
		while (*c == ' ' || *c == '\t') c++;
		if (*c != '\n') command_prompt[i] = c;
		c = index(c, '\n'); *c++ = 0;
		/*
		**	Save the menu item string and decode the key value
		*/
		command_text[i] = c;
		if (c[0] == 'M' && c[1] == '-')
			n = 0200, c += 2;
		else
			n = 0;
		if (c[0] == 'D' && c[1] == 'E' && c[2] == 'L')
			n += 0177;
		else 
			n += (*c == '^' && ! isspace(c[1])) ?
				*++c & 037 : *c & 0177;
		command_index[n] = i+1;
		c = index(c, '\n'); *c++ = 0;
		menu_set(parent_menu ? parent_menu : menu,
			MENU_STRING_ITEM, command_text[i], (caddr_t)n,
			0);
		/*
		**	Finally, save the help lines,
		**	replacing single newlines with blanks
		*/
		command_help[i] = c;
		do	{
			c = index(c, '\n') + 1;
		} while (*c && *c != '\n');
		c[-1] = 0;
	}
	free ((char *)sub_menu_name);
	/*
	**	Add the frame menu
	*/
	menu_set(menu, 
		MENU_NOTIFY_PROC,	menu_return_value,
		MENU_DONE_PROC,		canvas_menu_done,
#ifdef OLD
		MENU_PULLRIGHT_ITEM, "   Frame",
		(Menu)xv_get(base_frame, WIN_MENU),
#endif OLD
		0);
}

	/*
	**	Display a menu
	*/
display_canvas_menu(buf, ie)
	char	buf[NUMBUF][TEXTLEN];
	Event	*ie;
{
	Menu	temp_menu;

	if (! event_shift_is_down(ie))
		temp_menu = menu;
	else temp_menu = menu_create(MENU_STRINGS,
		get_coordinates(buf[0], "x: %x, %tx (%rx,%rX)", ie),
		get_coordinates(buf[1], "y: %y, %ty (%ry,%rY)", ie),
		get_coordinates(buf[2], "d: %d, %td (%rd,%rD)", ie),
		0,
		0);
	menu_event = *ie;
	menu_show(temp_menu, base_frame, ie, 0);
	return 0;
}

	/*
	**	Display the help message for a particular command key
	*/
display_help_window(code, str1, str2)
	int		code;
	char		*str1, *str2;
{
	int		narg, i, rc;
	char		ch, *b, *c, buf[50], *arg[20];

	if (i = command_index[code])
		b = c = command_help[i-1];
	else
		sprintf(b = c = buf, "Key %s not assigned", form_char(code));
	for (narg = 0; narg < 18; c = b) {
		while ((ch = *b) && ch != '\n') b++;
		*b = 0;
		arg[narg++] = save_string(c, NULL);
		if (! (*b++ = ch))
			break;
	}
	arg[narg] = arg[narg+1] = NULL;
	sprintf(buf, "The '%s' key:", form_char(code)),
	rc = notice_prompt(base_frame, (Event *)NULL,
		NOTICE_BUTTON_YES,	str1,
		(str2 && *str2) ? NOTICE_BUTTON_NO : NOTICE_NO_BEEPING,
		(str2 && *str2) ? (caddr_t)str2 : (caddr_t)TRUE,
		NOTICE_FOCUS_XY,		0, 0,
		NOTICE_MESSAGE_STRINGS, buf,
			arg[0], arg[1], arg[2], arg[3], arg[4],
			arg[5], arg[6], arg[7], arg[8], arg[9],
			arg[10], arg[11], arg[12], arg[13], arg[14],
			arg[15], arg[16], arg[17], arg[18], arg[19],
			0,
		0);
	for (i = 0; i < narg; i++)
		free(arg[narg]);
	switch (rc) {
	case NOTICE_YES:
		return 1;
	case NOTICE_NO:
		return 2;
	default:
		return 0;
	}
}
