#ifndef lint
static char *sccsid = "%Z%$RCSfile: setup.c,v $	GIPS $Revision: 2.3 $ (MIT/CSR) $Date: 2004/11/29 20:08:51 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		setup.c		Creation Date:	03/19/86
**	Version:	$Revision: 2.3 $		Update:		$Date: 2004/11/29 20:08:51 $
**	Components:	get_hist		save pixel histogram
**			pixrep			replicate zoomed pixels
**			pointer_xor		insert xor-ed line
**			read_image		read image, write to screen
**
**	Function:	Read and convert the input image
*/
#include	"gipstool.h"

static	get_hist(), pointer_xor(), pixrep();
	/*
	**	allocate input buffer and pixrect
	*/
read_image()
{
	struct	pixrect	*input_pr;	/* one-raster input pixrect */
	struct	pixrect	*byte_pr;	/* one-byte-per-pixel input */
	char		*inbuf;		/* raw input buffer */
	char		*inoff;		/* start of data in input buffer */
	short		*outbuf;	/* pixrect data area */
	char		*convbuf;	/* converted input buffer */
	short		*zsave;		/* error diffusion buffer */
	short		*zsave1;	/* error diffusion pointer */
	short		*zsave2;	/* error diffusion pointer */
	int		size;		/* pixel size in bytes */
	int		nconv;		/* number of conversions desired */
	int		width;		/* width of saved_pr */
	int		flen;		/* length of useful input raster */
	int		height;		/* height of saved_pr */
	register	y;		/* input y-raster count */
	register	yy;		/* output y-raster address */
	register	yp;		/* y-pixel count */
	extern	int	_gcount;	/* number of pixels -> gcell() */

	/*
	**	Determine size of image to save, allocate a pixrect
	*/
	xv_set(image_sw, WIN_CURSOR, hourglass_cursor, 0);
	sw_rect = *((Rect *)xv_get(image_sw, WIN_RECT));
	width = saved_rect.r_width;
	height = saved_rect.r_height;
	nconv = width / xzoom;
	if (saved_pr == NULL) {
		saved_pr = mem_create(width, height, depth);
		if (depth == 1)
			data_pr = mem_create(nconv, height/yzoom, 8);
	}
	input_pr = mem_create(width, yzoom, depth);
	if (saved_pr == NULL || input_pr == NULL)
		Error("cannot create saved pixrect","");
	/*
	**	Allocate buffers for each raster
	**	create an input pixrect to define a single raster
	*/
	size = ihdtype(dtype);
	outbuf = mpr_d(input_pr)->md_image;
	inbuf = alloc(xnum,size);
	convbuf = (depth != 1 && xzoom == 1) ?
		(char *)outbuf : alloc(nconv,sizeof(char));
	if (depth == 1)
		byte_pr = mem_point(nconv, 1, 8, (short *)convbuf);
	if (Zflag) {
		zsave = (short *)alloc(2*(nconv+2), sizeof(short));
		zsave1 = zsave+1;
		zsave2 = zsave1+nconv+2;
		setup_diffuse_error();
	}
	if (inbuf == NULL || convbuf == NULL)
		gipserr();
	inoff = &inbuf[size * xorg];
	/*
	**	clear the display background and initialize image window
	*/
	pw_writebackground((int)image_pw, 0, 0,
		sw_rect.r_width, sw_rect.r_height, PIX_CLR);
	if (cflag) {
		xorg += xoff;
		yorg += yoff;
		xoff = yoff = 0;
		nconv = MIN(nconv, xnum - xorg);
		flen = nconv*size;
	}
	saved_rect.r_left = xoff * xzoom;
	saved_rect.r_top = yoff * yzoom;
	bzero((char *)image_hist, sizeof(image_hist));
	/*
	**	write the pixels into saved_pr, then rop them to screen
	*/
	for (y = 0; y < ynum; y++) {
		if (! cflag && fread(inbuf,xnum,size,fp) != size)
			gipserr2(GIPSBADREAD, imfile);
		if ((yp = (y - yorg) * yzoom) < 0)
			continue;
		if (y >= yorg + ydim)
			if (cflag)
				break;
			else
				continue;
		if (cflag) 
			if (input_gimage) {
				register int x;
				register char *c;
				for (x = 0; x < nconv; x += _gcount) {
					c = gcell(xorg+x, y, input_gimage);
					if (_gcount < -1) {
						bcopy(c, inoff+size*x, size);
						c = inoff+size*x;
						x++;
						_gcount = -_gcount - 1;
					}
					if (_gcount > (nconv - x))
						_gcount = nconv - x;
					else if (_gcount <= 0)
						_gcount = 1;
					bcopy(c, inoff+x*size, _gcount*size);
				}
			} else {
				off_t loff, lseek();
				loff = image_seek + size*(y*xnum+xorg);
				if (lseek(fileno(fp), loff, 0) < image_seek ||
					read(fileno(fp), inoff, flen) != flen)
					gipserr2(GIPSBADREAD, imfile);
			}
		if (gips((double *)inoff,(int *)convbuf,nconv,GIPS_D_VAL))
			gipserr();
		if (depth == 1) {
			(void)pr_rop(data_pr, 0, y-yorg, nconv, 1,
					PIX_SRC, byte_pr, 0, 0);
			if (Zflag) {
				short *zs = zsave2;
				zsave2 = zsave1;
				diffuse_error(convbuf, outbuf, nconv,
					zsave1 = zs, zsave2);
			} else
				shift_bw(convbuf, outbuf, nconv);
		} else if (cshift || xzoom > 1 || yzoom > 1)
			pixrep(convbuf, (char *)outbuf, xzoom, nconv,
				mpr_d(input_pr)->md_linebytes);
		else
			get_hist(convbuf, nconv);
		(void)pr_rop(saved_pr, 0, yp, width, yzoom,
				PIX_SRC, input_pr, 0, 0);
		if (image_rect.r_height < saved_rect.r_height)
			pointer_xor(y);
		if (yp < saved_rect.r_top ||
			yp >= saved_rect.r_top + image_rect.r_height)
			continue;
		yy = image_rect.r_top + yp - saved_rect.r_top;
		pw_write((int)image_pw,
			image_rect.r_left,
			yy,
			MIN(image_rect.r_width, width - saved_rect.r_left),
			yzoom,
			PIX_SRC,
			input_pr,
			saved_rect.r_left,
			0);
	}
	if (! cflag)
		(void) ihbcl(fp);
	if (Zflag)
		free((char *)zsave);
	if (convbuf != (char *)outbuf)
		free(convbuf);
	if (depth == 1)
		pr_close(byte_pr);
	pr_close(input_pr);
	free(inbuf);
	xv_set(image_sw, WIN_CURSOR, image_cursor, 0);
	if (cflag && Nflag && stretch_percent)
		stretch_colormap(stretch_percent);
}
	/*
	**	Accumulate histogram
	*/
static
get_hist(c, n)
register char *c;
register n;
{
	register int	*h = image_hist;
	register int	i;

	while (--n >= 0)
		if (i = *c++ & 0377)
			h[i]++;
}
	/*
	**	Display a line to mark fraction of image filled
	*/
static
pointer_xor(y)
	register y;
{
	y = (image_rect.r_height * (y - yorg))/ydim;
	if (y > 0 && y < image_rect.r_height) {
		y += image_rect.r_top;
		pw_vector((int)image_pw, image_rect.r_left, y,
			image_rect.r_left + image_rect.r_width, y,
			PIX_SRC ^ PIX_DST, 255);
		pw_vector((int)image_pw, image_rect.r_left, y,
			image_rect.r_left + image_rect.r_width, y,
			PIX_SRC ^ PIX_DST, 255);
	}
}

static
pixrep(in, out, rep, count, len)
	register char	*in, *out;
	register	rep, count;
	int		len;
{
	register 	i, ch, m = cshift;
	char		*out0	= out;

	while (--count >= 0) {
		ch = (*in++ & 0377) >> m;
		image_hist[ch]++;
		for (i = rep; --i >= 0; *out++ = ch)
			;
	}
	for (in = out0, i = 0; ++i < yzoom; in = out)
		move(in, out = in + len, len);
}
