#ifndef lint
static char *sccsid = "%Z%$RCSfile: tags.c,v $	GIPS $Revision: 2.5 $ (MIT/CSR) $Date: 1998/11/14 06:37:13 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		tags.c		Creation Date:	09/23/86
**	Version:	$Revision: 2.5 $		Update:		$Date: 1998/11/14 06:37:13 $
**	Components:	alloc_tag_item()	create tag entry
**			get_tag_item()		locate tag entry
**			get_tag_name()		return last tag name
**			recall_named_tag()	format tag entries
**			scroll_tag_list()	list tag entries
**			set_named_tag()		add tag entry
**			write_tag_list()	format one tag entry
**
**	Function:	Save, list and restore cursor location via tags
*/

#include	"gipstool.h"
#undef		flag
#include	"selections.h"
#include	<xview/textsw.h>
#include	<xview/seln.h>
#include	<ctype.h>
	/*
	**	Saved image tag structure
	*/
static	struct	tg_item	{
	struct	tg_item	*tg_low;	/* 'left' fork pointer */
	struct	tg_item	*tg_high;	/* 'right' fork pointer */
	double	tg_dval[2];		/* realworld coordinates */
	int	tg_loc[2];		/* image location */
	char	*tg_comment;		/* comment text */
	char	tg_name[1];		/* name array */
} *anchor_tag;				/* head of binary tree */

static	int	tg_loaded;		/* =1 if text loaded */
static		write_tag_list(), free_tag_list();

	/*
	**	Return last accessed tag name
	*/
char	*
get_tag_name()
{
	Seln_holder	holder;		/* selection service handle */
	Seln_request	*buffer;	/* pointer to select buffer */

	holder = seln_inquire(SELN_PRIMARY);
	buffer = seln_ask(&holder, SELN_REQ_CONTENTS_ASCII, 0, 0);
	return buffer->data + sizeof(Seln_attribute);
}
	/*
	**	Convert realworld value to character
	*/
static	char	*
conv_real(d, s)
	double	d;			/* value to convert */
	char	*s;			/* output buffer */
{
	if (d <= GIPSLOW)
		return strcpy(s, "MIN");
	else if (d >= GIPSHIGH)
		return strcpy(s, "MAX");
	else
		return gcvt(d, precision, s);
}

static	Frame	tag_frame;		/* Frame for tag text */
static	Panel	tag_panel;		/* Panel for tag buttons */
static	Textsw	tag_window;		/* Scrolling window for tag text */
static	char	*tag_temp;		/* Scratch tag file name */

static
tag_point_button(item, event)
	Panel_item	item;
	Event		*event;
{
	recall_named_tag("", event);
}

static
tag_set_button(item, event)
	Panel_item	item;
	Event		*event;
{
	event_set_id(event, SET_NAMED_TAG);
	image_selected(image_sw, event, 0);
}

static
tag_load_button(item, event)
	Panel_item	item;
	Event		*event;
{
	read_tag_file(get_tag_name());
}

static
tag_save_button(item, event)
	Panel_item	item;
	Event		*event;
{
	write_tag_file(get_tag_name());
}

static
tag_clear_button(item, event)
	Panel_item	item;
	Event		*event;
{
	xv_set(tag_window, TEXTSW_READ_ONLY, FALSE, 0);
	textsw_erase(tag_window, 0, TEXTSW_INFINITY);
	xv_set(tag_window, TEXTSW_INSERTION_POINT, 0, 0);
	xv_set(tag_window, TEXTSW_READ_ONLY, TRUE, 0);
	free_tag_list(anchor_tag);
	anchor_tag = NULL;
	tg_loaded = 1;
}

static
tag_done_button(item, event)
	Panel_item	item;
	Event		*event;
{
	xv_set(tag_frame, WIN_SHOW, FALSE, 0);
}

display_tag_list()
{
	Rect	r, v;
	int	line, i, width;
	char	*mktemp();

	if (! tag_frame) {
		defaults_set_string("openWindows.scrollbarPlacement", "Left");
		tag_frame = xv_create(base_frame, FRAME_CMD,
			WIN_SHOW,		FALSE,
			/*WIN_WIDTH,		512,
			WIN_HEIGHT,		128,*/
			FRAME_NO_CONFIRM,	TRUE,
			FRAME_SHOW_LABEL,	TRUE,
			FRAME_SHOW_RESIZE_CORNER,	TRUE,
			WIN_RIGHT_OF,		base_frame,
			FRAME_LABEL,	xv_get(base_frame, FRAME_LABEL),
			0);
		tag_panel = (Panel)xv_get(tag_frame, FRAME_CMD_PANEL);
		width = xv_col(tag_panel, 7);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"FIND",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_point_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,line = 0)/3,
			0);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"SET",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_set_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,++line)/3,
			0);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"LOAD",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_load_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,++line)/3,
			0);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"SAVE",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_save_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,++line)/3,
			0);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"CLEAR",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_clear_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,++line)/3,
			0);
		panel_create_item(tag_panel, PANEL_BUTTON,
			PANEL_LABEL_STRING,	"DONE",
			PANEL_LABEL_WIDTH,	width,
			PANEL_NOTIFY_PROC,	tag_done_button,
			PANEL_ITEM_X,		5,
			PANEL_ITEM_Y,		2*xv_row(tag_panel,++line)/3,
			0);
		window_fit(tag_panel);
		tag_window = xv_create(tag_frame, TEXTSW,
			WIN_RIGHT_OF,		tag_panel,
			WIN_X,			xv_col(tag_panel, 8),
			WIN_Y,			0,
			WIN_WIDTH,		650,
			WIN_HEIGHT,		(int)xv_get(
						tag_panel, WIN_HEIGHT),
			TEXTSW_DISABLE_CD,	TRUE,
			TEXTSW_DISABLE_LOAD,	TRUE,
			0);
		close(creat(tag_temp = mktemp("/tmp/gtXXXXXX"), 0600));
		xv_set(tag_window,
			TEXTSW_FILE,	 	tag_temp,
			0);
		window_fit(tag_frame);
		frame_get_rect(base_frame, &r);
		frame_get_rect(tag_frame, &v);
		xv_set(tag_frame,
			XV_X, MIN(display_rect.r_width - v.r_width,
				r.r_left + r.r_width),
			XV_Y, MAX(r.r_top, 0),
			0);
	}
	if (!tg_loaded) {
		xv_set(tag_window, TEXTSW_READ_ONLY, FALSE, 0);
		textsw_erase(tag_window, 0, TEXTSW_INFINITY);
		xv_set(tag_window, TEXTSW_INSERTION_POINT, 0, 0);
		write_tag_list(anchor_tag, NULL);
		xv_set(tag_window, TEXTSW_READ_ONLY, TRUE, TEXTSW_FIRST, 0, 0);
		tg_loaded = 1;
	}
	xv_set(tag_frame, WIN_SHOW, TRUE, 0);
}

destroy_tag_frame()
{
	if (tag_window)
		textsw_reset(tag_window, 0, 0);
	if (tag_temp)
		unlink(tag_temp);
	tg_loaded = 0;
}

read_tag_file(name)
	char		*name;		/* input tag file name */
{
	char		*b, buf[1024];	/* input line */
	char		*c;		/* scratch */
	FILE		*fp;		/* input stream */
	int		i[2];		/* fixed image coordinates */
	double		f[2];		/* float image coordinates */
	double		d[2];		/* realworld coordinates */
	int		status;		/* saved value of display_status */

	if (! name) {
		display_error_message("no tag file", "");
		return;
	}
	if (! (fp = fopen(name, "r"))) {
		(void)sprintf(buf, "%s/%s", GIPSPROFDIR, name);
		if (! (fp = fopen(buf, "r"))) {
			display_system_message(name);
			return;
		}
	}
	tag_file = save_string(name, tag_file);
	if (! no_display)
		xv_set(image_sw, WIN_CURSOR, hourglass_cursor, 0);
	while (fgets(c = buf, sizeof(buf)-1, fp) != NULL) {
		while (*c && isspace(*c)) c++;
		if (! *c || *c == '#')
			continue;
		while (*c && ! isspace(*c)) c++;
		if (! *c || sscanf(c, " %lf %lf", &d[0], &d[1]) != 2) {
			if (! no_display)
				xv_set(image_sw, WIN_CURSOR, image_cursor, 0);
			display_error_message("bad tag line", c);
			return;
		}
		status = no_display;
		no_display = 1;
		b = c;
		while (isspace(*c)) c++;
		while (*c && ! isspace(*c)) c++;
		while (isspace(*c)) c++;
		while (*c && ! isspace(*c)) c++;
		(void)strcpy(b, c);
		while (*b)
			if (*b++ == '\n')
				*--b = 0;
		if (map_sw) {
			if (g_map(d[0], d[1], f))
				i[0] = -1, i[1] = -1;
			else
				i[0] = f[0] + 0.5, i[1] = f[1] + 0.5;
		} else
			(void)gips(d, i, 1, GIPS_XY_IVAL);
		allocate_tag(buf, i, d);
		no_display = status;
	}
	(void)fclose(fp);
	if (! no_display) {
		display_tag_list();
		xv_set(tag_frame, FRAME_LABEL, name, 0);
		xv_set(image_sw, WIN_CURSOR, image_cursor, 0);
	}
	update_info();
}

write_tag_file(name)
	char		*name;		/* output tag file name */
{
	FILE		*fp;		/* output stream */

	if (! name) {
		display_error_message("no tag file", "");
		return;
	}
	if (! (fp = fopen(name, "w"))) {
		display_system_message(name);
		return;
	}
	tag_file = save_string(name, tag_file);
	write_tag_list(anchor_tag, fp);
	(void)fclose(fp);
	update_info();
}

static
write_tag_list(tg, fp)
	struct tg_item	*tg;		/* current chain entry */
	FILE		*fp;		/* output stream */
{
	char		str[256], scr1[32], scr2[32], scr3[32];

	if (! tg)
		return;
	write_tag_list(tg->tg_low, fp);
	if (fp) {
		(void) fprintf(fp,
			"%-16.16s %8.8s %8.8s %s\n",
			tg->tg_name,
			conv_real(tg->tg_dval[0], scr1),
			conv_real(tg->tg_dval[1], scr2),
			tg->tg_comment);
	} else {
		(void) sprintf(scr3, "(%d,%d)",
			tg->tg_loc[0], tg->tg_loc[1]);
		(void) sprintf(str,
			"%-16.16s %8.8s %8.8s %-11.11s %s\n",
			tg->tg_name,
			conv_real(tg->tg_dval[0], scr1),
			conv_real(tg->tg_dval[1], scr2),
			scr3,
			tg->tg_comment);
		(void)textsw_insert(tag_window, str, strlen(str));
	}
	write_tag_list(tg->tg_high, fp);
}

static
free_tag_list(tg)
	struct tg_item	*tg;		/* current chain entry */
{
	if (! tg)
		return;
	free_tag_list(tg->tg_low);
	free_tag_list(tg->tg_high);
	free((char *)tg);
}

	/*
	**	Insert new tag item in chain
	*/
static	struct	tg_item	*
alloc_tag_item(tg, tnew)
	register struct tg_item	*tg;	/* current chain entry */
	register struct tg_item	*tnew;	/* new chain entry */
{
	register		 i;	/* scratch */
	/*
	**	End of chain: return new pointer
	*/
	if (tg == NULL)
		return tnew;
	/*
	**	Within chain: search deeper
	*/
	if ((i = strcmp(tg->tg_name, tnew->tg_name)) > 0) {
		tg->tg_low = alloc_tag_item(tg->tg_low, tnew);
		return tg;
	}
	if (i < 0) {
		tg->tg_high = alloc_tag_item(tg->tg_high, tnew);
		return tg;
	}
	/*
	**	Duplicate entry: free the old one
	*/
	tnew->tg_high = tg->tg_high;
	tnew->tg_low = tg->tg_low;
	free((char *)tg);
	tg_loaded = 0;
	return tnew;
}
	/*
	**	Convert tag coordinates to realworld, allocate tag
	*/
set_named_tag(s, xloc, yloc)
	register char	*s;		/* tag name */
	register int	xloc, yloc;	/* pixel address */
{
	int		i[2];		/* image addresses */
	double		d[2];		/* realworld values */

	i[0] = X_RECT_TO_IMAGE(xloc);
	i[1] = Y_RECT_TO_IMAGE(yloc);

	if (map_sw) {
		if (g_unmap(i[0], i[1], d))
			d[0] = d[1] = GIPSHIGH;
	} else
		(void)egips(i, d, 1, EGIPS_XY_VAL);

	allocate_tag(s, i, d);
}

allocate_tag(s, i, d)
	register char	*s;		/* tag name */
	int		i[2];		/* image addresses */
	double		d[2];		/* realworld address */
{
	register struct tg_item	*tg;	/* tag list pointer */

	if (! *s)
		s = get_tag_name();

	while (isspace(*s)) s++;

	if (! *s) {
		display_error_message("no tag name selected", "");
		return;
	}

	tg = (struct tg_item *)alloc(sizeof(struct tg_item) + strlen(s), 1);
	if (tg == NULL) {
		no_more_memory();
		return;
	}
	for (s = strcpy(tg->tg_name, s); *s && ! isspace(*s); s++)
		if (isupper(*s))
			*s = tolower(*s);
	if (*s)
		for (*s++ = 0; isspace(*s); s++)
			;
	tg->tg_comment = s;
	tg->tg_loc[0] = i[0];
	tg->tg_loc[1] = i[1];
	tg->tg_dval[0] = d[0];
	tg->tg_dval[1] = d[1];
	anchor_tag = alloc_tag_item(anchor_tag, tg);
	if (! no_display)
		display_tag_list();
}
	/*
	**	Find item in tag chain by name
	*/
static	struct	tg_item	*
get_tag_item(tg, s)
	register struct tg_item	*tg;	/* tag list pointer */
	register char		*s;	/* tag name */
{
	register 		i;	/* scratch */

	if (tg == NULL)
		return NULL;
	i = strcmp(tg->tg_name, s);
	if (i > 0)
		return get_tag_item(tg->tg_low, s);
	if (i < 0)
		return get_tag_item(tg->tg_high, s);
	return tg;
}
	/*
	**	Recall a tag list item
	*/
recall_named_tag(s, ie)
	register char	*s;		/* tag name */
	register Event	*ie;		/* Event pointer */
{
	register struct tg_item	*tg;	/* tag list pointer */
	register char	*c;		/* scratch */

	if (! *s) {
		s = get_tag_name();
		if (! *s)
			display_error_message("no tag name selected", "");
	}
	for (c = s; *c; c++)
		if (isupper(*c))
			*c = tolower(*c);
	if (! (tg = get_tag_item(anchor_tag, s)))
		display_error_message("unknown tag", s);
	else if (tg->tg_loc[0] < 0 || tg->tg_loc[0] >= xnum ||
		tg->tg_loc[1] < 0 || tg->tg_loc[1] >= ynum)
		display_error_message("tag not within image boundary", 1);
	else {
		set_image_cursor(tg->tg_loc[0], tg->tg_loc[1], ie);
		set_image_mark(X_IMAGE_TO_RECT(tg->tg_loc[0]),
			X_IMAGE_TO_RECT(tg->tg_loc[1]));
		clear_cursor_window();
		if (! no_display)	
			xv_set(tag_frame, WIN_SHOW, TRUE, 0);
	}
}
	/*
	**	reset image mark to current cursor location
	*/
set_image_mark(locx, locy)
	register locx, locy;		/* cursor location */
{
	image_mark[0] = X_RECT_TO_IMAGE(locx);
	image_mark[1] = Y_RECT_TO_IMAGE(locy);
	update_info();
}
	/*
	**	Test whether image address is within the image
	*/
test_image_mark(x, y)
	register int	x, y;
{
        if (x < 0 || x >= xnum)
                display_error_message("x-address '%d' out of bounds",
			(char *)x);
        else if (! cflag && (x < xorg || x >= xorg+xdim))
                display_error_message("x-address '%d' not stored",
			(char *)x);
        else if (y < 0 || y >= ynum)
                display_error_message("y-address '%d' out of bounds",
			(char *)y);
        else if (! cflag && (y < yorg || y >= yorg+ydim))
                display_error_message("y-address '%d' not stored",
			(char *)y);
	else
		return 0;
	return 1;
}
	/*
	**	Position cursor at given image pixel.
	**	If necessary to shift the displayed image,
	**	call image_repaint() before setting the
	**	cursor in case the pixel cannot be centered
	*/
set_image_cursor(xpix, ypix, ie)
	register int	xpix, ypix;
	register Event	*ie;
{
	register int	locx, locy;

	locx = X_IMAGE_TO_RECT(xpix);
	locy = Y_IMAGE_TO_RECT(ypix);
	if (locx < image_rect.r_left ||
			locx >= image_rect.r_left + image_rect.r_width ||
			locy < image_rect.r_top ||
			locy >= image_rect.r_top + image_rect.r_height) {
		xoff = xpix - xorg - image_rect.r_width/(2*xzoom);
		yoff = ypix - yorg - image_rect.r_height/(2*yzoom);
		image_repaint(image_sw, image_pw, (Rectlist *)NULL);
		locx = X_IMAGE_TO_RECT(xpix);
		locy = Y_IMAGE_TO_RECT(ypix);
		xv_set(base_frame, WIN_SHOW, TRUE, 0);
	}
	/*
	**	It is necessary to remember the mouse position in ie
	**	when commands are being read from an external file.
	*/
	event_set_x(ie, locx);
	event_set_y(ie, locy);
	xv_set(image_sw, WIN_MOUSE_XY, locx, locy, 0);
}

set_tag_frame(show)
	int	show;
{
	if (tag_frame)
		if ((int)xv_get(tag_frame, FRAME_CMD_PUSHPIN_IN) == show)
			xv_set(tag_frame, XV_SHOW, show, 0);
}
