#ifndef lint
static char *sccsid = "%Z%$RCSfile: aedsubs.c,v $	GIPS $Revision: 2.1 $ (MIT/CSR) $Date: 1986/07/27 03:30:44 $";
#endif
/*		   Comtal Image Processor Utility
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1985 Massachusetts Institute of Technology
**
**	Module:		aedsubs.c	Creation Date:	11/21/85
**	Version:	$Revision: 2.1 $		Update:		$Date: 1986/07/27 03:30:44 $
**	Function:	Device independent GIPS display routines
**	Components:
**		image_open()	return fildes to opened image plane
**		graphic_open()	return fildes to opened graphic plane
**		pseudo_open()	return fildes to opened ps table
**		function_open()	return fildes to opened stretch table
**		command_open()	return stream ptr to command input
**		display_put()	set display origin
**		target_get()	return target coordinates
**		target_put()	reposition target
**		target_roam()	set target for roaming
**		terget_dump()	set target for dumping
**		color_get()	return graphic color
**		color_put()	set graphic color
**		print_subtab()	write command table to stream
**		comtal_error()	error message string generator
*/

#include	<stdio.h>
#include	<ctype.h>
#include	"gips.h"

#define		WHITE	15

static	char genclrc[]		= "%s/.clrc";
static	char genimage[]		= "/dev/aed";
static	char gengraphic[]	= "";
static	char gentarget[]	= "";
static	char gencommand[]	= "/dev/aedint";
static	char genpseudocolor[]	= "";
static	char genfunction[]	= "";

static	struct	ctab {
	char	*cname;
	char	*insert[2];
	char	*ccmd;
} subtab[]	= {
	"image_open",	NULL,	NULL,	"",
	"graphic_open",	NULL,	NULL,	"",
	"pseudo_open",	NULL,	NULL,	"",
	"function_open",NULL,	NULL,	"",
	"display_put",	NULL,	NULL,	"",
	"target_get",	NULL,	NULL,	"",
	"target_put",	NULL,	NULL,	"",
	"target_roam",	NULL,	NULL,	"",
	"target_dump",	NULL,	NULL,	"",
	NULL
};

extern	char	*sprintf(), *strcpy(), *getenv(), *index();
static	char	buf[256];

static
comtal_error()
{
	register	char *s, *t;

	for (s = buf+5, t = *gipserrs; *t = *s; s++, t++)
		if (*t == '.')
			*t = ' ';
	gipserrno = GIPSBADOPEN;
	return(-1);
}

print_subtab(s,fp)
register char *s;
register FILE *fp;
{
	register struct ctab *t;

	if (s == NULL) s = "";
	if (fp == NULL) fp = stdout;
	for (t = subtab; t->cname; t++)
		fprintf(fp,"%s%s\t: \"%s\"\n",s,t->cname,t->ccmd);
}

FILE *
command_open()
{
	register	i, l;
	register struct ctab *t;
	register char	*s;
	static	 FILE	*fcom;
	static	 char	*combuf;

	if (combuf == NULL && (combuf = getenv("HOME")) != NULL &&
			(i = open(sprintf(buf,genclrc,combuf),0)) >= 0) {
		(void) lseek(i,0L,2);
		combuf = alloc(l = lseek(i,0L,1)+1,1);
		(void) lseek(i,0L,0);
		(void) read(i,s = combuf,l);
		(void) close(i);
		for (s = combuf; *s; *s++ = 0) {
			for (t = subtab; t->cname != NULL &&
				strncmp(s,t->cname,i = strlen(t->cname)); t++)
				;
			if (t->cname != NULL)
				t->ccmd = s+i+1;
			if ((s = index(s,'\n')) == NULL)
				break;
		}
	}
	(void) strcpy(*gipserrs,"aed");
	if (fcom == NULL || (fcom->_flag & (_IOWRT | _IORW)) == 0)
		fcom = fopen(gencommand,"w");
	if (fcom != NULL)
		return(fcom);
	gipserrno = GIPSBADOPEN;
	return(NULL);
}

static
comout(n,i1,i2)
{
	register FILE *fp;
	register char *cp, *s;
	register	i;
	char *fmt = "%d"; 

	cp = subtab[n].ccmd;
	if (! *cp)
		return(0);
	if ((fp = command_open()) == NULL)
		return(-1);
	for (; *cp; cp++)
		if (*cp != '%')
			putc(islower(*cp) ? toupper(*cp) : *cp,fp);
		else
			for (i = 0; i < 2 && (s = subtab[n].insert[i]); i++)
				if (strncmp(s+1,cp+2,strlen(s+1)) == 0) {
					cp += strlen(s+1)+1;
					fmt[1] = *s;
					fprintf(fp,fmt,i ? i2 : i1);
					break;
				}
	fputs(" \n",fp);
	(void) fflush(fp);
	return(ferror(fp) ? -1 : 0);
}

image_open(ni,mode)
{
	register	fildes;

	if ((fildes = open(sprintf(buf,genimage,ni),mode)) < 0)
		return(comtal_error());
	(void) comout(0,ni,0);
	return(fildes);
}

graphic_open(ng,mode)
{
	register	fildes, nc;

	if ((fildes = open(sprintf(buf,gengraphic,ng),mode)) < 0)
		return(comtal_error());
	(void) color_put(ng,WHITE);
	return(fildes);
}

pseudo_open(np,mode)
{
	register	fildes;
	static	char	*colors[] = { "red", "green", "blue" };

	if (np < 0 || np > 2)
		return(-1);
	if ((fildes = open(sprintf(buf,genpseudocolor,"rgb"[np]),mode)) < 0) {
		(void) strcpy(buf+17,colors[np]);
		return(comtal_error());
	}
	(void) comout(2,"RGB"[np],0);
	return(fildes);
}

function_open(nf,mode)
{
	register	fildes;

	if ((fildes = open(sprintf(buf,genfunction,nf),mode)) < 0)
		return(comtal_error());
	(void) comout(3,nf,0);
	return(fildes);
}

display_put(nt,xy)

	register	int	*xy;
{
	return(comout(4,xy[0],xy[1]));
}

target_get(nt,xy)

	register	int	*xy;
{
	register	nbytes;
	register	fildes;
	short	spixel[2];

	if (comout(5,xy[0],xy[1]))
		return(-1);
	(void) strcpy(*gipserrs,"target");
	if ((fildes = open(gentarget,0)) == -1) {
		gipserrno = GIPSBADOPEN;
		return(-1);
	}
	nbytes = read(fildes,(char *)spixel,4);
	(void) close(fildes);
	if (nbytes != 4) {
		gipserrno = GIPSBADREAD;
		return(-1);
	}
	xy[0] = spixel[0];
	xy[1] = spixel[1];
	return(gipserrno = 0);
}

target_put(nt,xy)

	register	int	*xy;
{
	return(comout(6,xy[0],xy[1]));
}

target_roam(nt)
{
	return(comout(7,nt,0));
}

target_dump(nt)
{
	return(comout(8,nt,0));
}

color_get(ng)
{
	return(WHITE);
}

color_put(ng,nc)
{
	static char *colors[] = {
		"BLA", "RED", "DGR", "LOR", "FGR", "ORN", "GRN", "YEL",
		"BLU", "MAG", "DTU", "PNK", "TUR", "LPI", "CYN", "WHT"
	};
	if (nc < 0 || nc >= 16)
		return(-1);
	return(comout(1,ng,colors[nc]));
}
