#include	"tailor.h"
#ifndef lint
statop char *comid initially "@(#)cin.c	4.2 (MIT/CSR) 01/21/84";
#endif
/*		   Comtal Image Processor Utility
 *			   Peter G. Ford
 *		      Center for Space Research
 *	Copyright 1984 Massachusetts Institute of Technology
 *
 *	Module:		cin.c		Creation Date:	12/29/83
 *	Version:	1.0		Update:		01/21/84
 *	Function:	Transfer data stream from UNIX to Comtal
 *	Components:	cin()		UNIX command
 */
#include	<stdio.h>
#include	"gips.h"

statop	int	number();	/* verify that 'ext' is numeric */
statop	int	input();	/* verify that this is 'cin' call */
statop	int	rgb();		/* verify that 'ext' is rgb */
statop	int	macro();	/* verify that 'ext' is macro */
statop	int	null();		/* verify that there is no extension */

statop	struct	tab	{
	char	*dev;		/* name of Comtal feature accepted by comio */
	char	*abbrev;	/* abbreviation of feature */
	int	bs;		/* length of data element */
	int	bn;		/* number of elements per block */
	int	(* proc)();	/* routine to check extension */
} funtab[] initially {
	"target",	"t",	2,	2,	null,
	"image",	"i",	2,	8192,	number,
	"graphic",	"g",	2,	8192, 	number,
	"command",	"c",	1,	1,	input,
	"function",	"f",	2,	256,	number,
	"table",	"tb",	2,	256,	number,
	"small",	"s",	2,	256,	number,
	"pseudocolor",	"p",	2,	256,	rgb,
	"targetshape",	"ts",	2,	256,	input,
	"macro",	"m",	1,	128,	macro,
	"code",		"code",	1,	4096,	null,
	NULL
};

statop	io	initially 0;			/* input/output switch */

extern	char	*sprintf(), *strcpy(), *strcat(), *alloc(), *index(), *rindex();

main(argc,argv)

	int	argc;			/* number of arguments */
	char	**argv;			/* argument pointer array */
{
	register struct tab *t;		/* pointer to pseudo-device table */
	register char *s;		/* scratch */
	register int i;			/* scratch */
	char	*buf;			/* i/o buffer */

	char	*ifile initially "'stdin'";	/* input file name */
	char	*ofile initially "'stdout'";	/* output file name */
	char	*ext initially "";		/* extension to pseudo-device */
	statop	char file[50] initially "/dev/";/* Comtal file name */

	if (argc == 2 && argv[1][0] == '=' && argv[1][1] == '\0') {
		fprintf(stderr,"Usage: %s function [file]\n",*argv);
		exit(TAMEEXITVAL);
	}
	if (argc < 2)			/* check number of args */
		error("missing args","");
	if (argc > 3)			/* check number of args */
		error("too many arguments: %s",argv[3]);
	if ((s = rindex(*argv,'c')) != NULL)
		use(strcpy(*argv,s));
	saveargs(argv);
	/*
	 *	set i/o switch sw=0,1 for cin,cout
	 */
	io = ((*argv)[1] == 'o');
	/*
	 *	locate pseudo-device extension (period following name
	 */
	if ((s = index(*++argv,'.')) != NULL)
		ext = s;	
	else
		s = *argv + strlen(*argv);
	/*
	 *	locate pseudo-device name or abbreviation in table
	 */
	for (t = funtab; *t->dev; t++)
		if (strncmp(*argv,t->dev,s-*argv) == 0 ||
		    strncmp(*argv,t->abbrev,s-*argv) == 0)
			break;
	if (! *t->dev)
		error("unknown comtal function: %s",*argv);
	/*
	 *	check the legality of the extension for this pseudo-device
	 */
	if ((*t->proc)(ext) == 0) 
		error("illegal comtal extension: %s",ext);
	else
		use(strcat(strcat(file,t->dev),ext));
	/*
	 *	locate the i/o buffer
	 */
	buf = alloc(t->bn,t->bs);
	/*
	 *	check that the Comtal is accessible and attempt to clear its
	 *	keyboard input buffer
	 */
	if (command_open() == NULL)
		gipserr();
	/*
	 *	unless they are pipes, open the input and output streams
	 */
	if (io || argc == 3)
		if (freopen(ifile = io ? file : argv[1],"r",stdin) == NULL)
			gipserr2(GIPSBADOPEN,ifile);
	if (! io || argc == 3)
		if (freopen(ofile = io ? argv[1] : file,"w",stdout) == NULL)
			gipserr2(GIPSBADOPEN,ofile);
	/*
	 *	copy the stream
	 */
	while ((i = Fread(buf,t->bs,t->bn,stdin)) == t->bn) {
		if (*t->dev == 'g')
			swab(buf,buf,t->bs * i);
		if (Fwrite(buf,t->bs,i,stdout) != i)
			gipserr2(GIPSBADWRITE,ofile);
	}
	if (i < 0)
		gipserr2(GIPSBADREAD,ifile);
	exit(TAMEEXITVAL);
}
/*
 *	verify numeric extension 1-128
 */
statop
number(ext)

	register char *ext;
{
	int	i;

	return(sscanf(ext,".%d",&i) == 1 && i > 0 && i <= 128 &&
		*++ext > '0' && *ext <= '9');
}
/*
 *	verify single character extension "rgb"
 */
statop
rgb(ext)

	register char *ext;
{
	return(*ext++ == '.' && index("rgb",*ext) != NULL && *++ext == '\0');
}
/*
 *	verify single character macro name extension [a-o]
 */
statop
macro(ext)

	register char *ext;
{
	return(*ext++ == '.' && *ext >= 'a' && *ext <= 'o' && *++ext == '\0');
}
/*
 *	verify that this is the 'cin' command - not 'cout'
 */
statop
input(ext)

	register char *ext;
{
	if (io)
		error("illegal output function: %s",ext);
	return(null(ext));
}
/*
 *	verify that there is no extension
 */
statop
null(ext)

	char *ext;
{
	return(! *ext);
}
