/* @(#)comtal.h		4.2 (MIT/CSR) 07/03/84
 *
 *		   Comtal Image Processor Utility
 *			   Peter G. Ford
 *		      Center for Space Research
 *	Copyright 1984 Massachusetts Institute of Technology
 *
 *	Module:		comtal.c	Creation Date:	07/03/84
 *	Version:	1.0		Update:		07/03/84
 *	Function:	Define Comtal hardware command codes
 */

/* Names of COMTAL transfer bits */

#define D_LR		0100000		/* l-r screen split */
#define D_FUNC		040000		/* function memory enable */
#define D_PSEUD		020000		/* pseudo color enable */
#define D_TEST		0200		/* test bit on */
#define D_TR5		040		/* test pattern selector */
#define D_TR6		0100		/* test pattern selector */
#define D_COLOR		01		/* true color */
#define D_G0		0400		/* graphics 0 switch */
#define D_G1		01000		/* graphics 1 switch, old diallib only */
#define D_GCOLTA	010000		/* graphics color & target */
#define D_GG		01		/* green on graphics */
#define D_GR		02		/* red on graphics */
#define D_GB		04		/* blue on graphics */
#define D_TAR		020000		/* target enable */
#define D_CLAS		040000		/* class enable */

#define I_MASK		0177		/* image mask */
#define G0_CMASK	07		/* graphics 0 color mask */
#define G1_CMASK	070		/* graphics 1 color mask, old diadlib only */

#define C_READ		0100000		/* read image, graphics, target, memory */
#define C_ROLL		02000		/* roll enable */
#define C_DOWN		01		/* roll down, not up */
#define C_GRAPH		01000		/* graphics not image */
#define C_FUNCT		040000		/* function not pseudo-color */
#define C_BLK		020000		/* write or read in block mode */
#define C_CLAS		02000		/* class not function */

#define TAR_Y		040000		/* target y */
#define CLAS_NO		01		/* classification number */
#define CLAS_CT		010000		/* classified pixel counter read */

#define KEY		'c'		/* key for comtal commands */

#define DO_GS		020000		/* gray scale */
#define DO_DOT		040000		/* dot pattern */
#define DO_BAR		060000		/* bar pattern */
#define DO_COLOR	010000		/* color */
#define DO_BW		002000		/* black-white */
#define DO_GRPH		001000		/* overlay */

/*
 * definition of GO2 bit.
 * Since all trcodes are as described in the manual (0-3),
 * and the Go2 bit is needed for some commands, we define it here
 * so it can be used (ala c.c_trcode = 1 | GO2;), and is shifted
 * to its correct position by the driver.
 */

#define	GO2	04


/* ioctl commands */
#define	CGETSTATE	(('c'<<8)|0)
#define	CSETSTATE	(('c'<<8)|1)
#define	CDEBUG		(('c'<<8)|2)

/* Debug state bits */
#define	CD_OPEN		0x1
#define	CD_CLOSE	0x2
#define	CD_STRATEGY	0x4
#define	CD_INTR		0x8
#define	CD_WAIT		0x10
#define	CD_START	0x20

/* structure of all COMTAL commands */

struct combuf {
	char	c_key;
	char	c_trcode;
	short	c_trdata;
	char	c_buf[16384];
};
