#ifndef lint
static char *sccsid = "%Z%$RCSfile: comio.c,v $	GIPS $Revision: 2.1 $ (MIT/CSR) $Date: 1986/07/27 04:56:47 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1983 Massachusetts Institute of Technology
**
**	Module:		gipstool.c	Creation Date:	11/12/83
**	Version:	$Revision: 2.1 $		Update:		$Date: 1986/07/27 04:56:47 $
**
**	Function:	Mediate between UNIX I/O calls & Comtal Image Unit
**	Acknowledgment:	This code adapted from A. Olson, NCI
*/

#include	"stdio.h"
#include	"errno.h"
#include	"sgtty.h"
#include	"comtal.h"
#include	"comsubs.h"
#include	"signal.h"
#include	"sys/types.h"
#include	"sys/stat.h"
#include	"syscall.h"
#include	"ctype.h"
	/*
	** For use with Comtal Vision One/20 V23B54 operating system.
	*/
#define		NFIDS		28
#define		TAME		0
#define		WILD		(-1)
#define		FREAD		1
#define		FWRITE		2
#define		MAXBYTES	(sizeof comraw.c_buf)
#define		HEADSIZE	(sizeof comraw - MAXBYTES)
#define		IMAGE		0
#define		GRAPHIC		1
#define		PSEUDOCOLOR	0
#define		FUNCTION	1
#define		errnoset(x)	((errno = x), WILD)

static	struct	device {
	char *	d_name;
	short	d_station;
	short	d_fid;
	short	d_count;
} comdevs[] = {
	"/dev/comtala",	0,	0,	0,
	"/dev/comtalb",	1,	0,	0
};

static	struct	feature {
	struct	device	*f_devptr;	/* points into above table */
	unsigned short	f_number;	/* image number/graphic number */
	long		f_filesize;
	long		f_offset;
	short (*	f_iop)();
	char		f_trcode;
	short		f_trdata;
} comfeats[NFIDS];

static	short		comtargetio(), comcodeio(), command();
static	comdefaultstation;	/* default station to use */
static	struct	combuf	comraw;
static	comcurfid;
static	fidbase;
extern	char		*sprintf();
extern	errno;

static	struct	nameinfo {
	char		*n_name;
	short		n_filesize;
	short		(*n_iop)();
	char		n_trcode;
	short		n_trdata[2];
} names[] = {
	"command*",	0,	command,	0,	0,	0,
	"image*.#",	0,	0,		1,	0,	004000,
	"graphic*.#",	0,	0,		1,	001000,	005000,
	"function*.#",	0,	0,		2,	060000,	060400,
	"small*.#",	0,	0,		2,	062000,	062400,
	"pseudocolor*.@", 256*2, 0,		2,	020000,	020000,
	"targetshape*",	256*2,	0,		2,	021400,	021400,
	"target*",	2*2,	comtargetio,	0,	0,	0,
	"code*",	0,	comcodeio,	0,	0,	0,
	"macro*.!",	128,	0,		3,	020000,	020000,
	"",		0,	0,		0,	0,	0
};

comio_() { return 0; }

open(name, mode)
	char	*name;
{
	return comopen(SYS_open, name, mode);
}

creat(name, mode)
	char	*name;
{
	return comopen(SYS_creat, name, mode);
}

access(name, mode)
	char	*name;
{
	return comopen(SYS_access, name, mode);
}

stat(name, ptr)
	char		*name;
	struct	stat	*ptr;
{
	return comopen(SYS_stat, name, (int)ptr);
}

fstat(fildes, ptr)
	struct	stat	*ptr;
{
	struct	feature	*comgetf();

	if (fildes < getdtablesize() && fildes >= fidbase &&
			comgetf(fildes) != 0)
		fildes = comcurfid;
	return syscall(SYS_stat, fildes, (int)ptr);
}

read(fildes, buffer, nbytes)
	register	fildes;
	char		*buffer;
{
	if (fildes > getdtablesize() || fildes < fidbase)
		return syscall(SYS_read, fildes, buffer, nbytes);
	else	return comio(fildes, buffer, nbytes, FREAD);
}

write(fildes, buffer, nbytes)
	register	fildes;
	char		*buffer;
{
	if (fildes > getdtablesize() || fildes < fidbase)
		return syscall(SYS_write, fildes, buffer, nbytes);
	else	return comio(fildes, buffer, nbytes, FWRITE);
}

long
lseek(fildes, offset, whence)
	register	fildes;
	long		offset;
{
	long	comlseek();

	if (fildes > getdtablesize() || fildes < fidbase)
		return syscall(SYS_lseek, fildes, offset, whence);
	else	return comlseek(fildes, offset, whence);
}

close(fildes)
	register fildes;
{
	if (fildes > getdtablesize() || fildes < fidbase)
		return (syscall(SYS_close, fildes) == WILD) ? WILD : TAME;
	else	return comclose(fildes);
}

getdtablesize()
{
	if (fidbase == 0)
		fidbase = syscall(SYS_getdtablesize);
	return fidbase + NFIDS;
}
	
static	commute(from, to, count)
	char	*from;
	char	*to;
{
	if (from != to && count != 0)
		asm("	movc3	12(ap),*4(ap),*8(ap)");
}

static	combufio(trcode, trdata, nbytes, mode)
	char	trcode;
	short	trdata;
	short	nbytes;
	char	mode;
{
	comraw.c_key = KEY;
	comraw.c_trcode = trcode;
	comraw.c_trdata = trdata;
	if (nbytes != 0)
		comraw.c_trcode |= GO2;
	if (mode == FREAD)
		comraw.c_trdata |= C_READ;
	else if (nbytes != 0)
		comraw.c_trdata &= ~C_READ;
	return (write(comcurfid, (char *) &comraw, HEADSIZE + nbytes) ==
		HEADSIZE + nbytes) ? nbytes : WILD;
}

static
comnbpl(fp)
	register struct	feature *fp;
{
	if (fp->f_iop != 0 || fp->f_trcode != 1)
		return errnoset(EBADF);
	switch (fp->f_trdata) {
		case 000000:
		case 004000:
		case 010000:
		case 014000:
			return 512;
		case 001000:
		case 005000:
		case 011000:
		case 015000:
			return 512/8;
		default:
			return errnoset(EBADF);
	}
}

static
comopen(syssw, name, mode)
	char	*name;
{
	register struct	nameinfo *np;
	register struct	feature *fp;
	int			number;
	int			trmask;
	int			station;
	register char *		s;
	register char *		t;
	char			*getenv();

	if (strncmp(name, "/dev/", 5) != 0)
		return syscall(syssw, name, mode);
	for (np = names; ; ) {
		for (t = name+5, s = np->n_name; *s; s++, t++) {
			trmask = number = station = 0;
			switch (*s) {
			case '*':
				if (*t == '.' || *t == '\0')
					--t;
				else
					station = *t - 'a' + 1;
				continue;
			case '!':
				if (*t < 'a' || *t > 'o')
					break;
				trmask = *t - 'a' + 1;
				continue;
			case '@':
				if (*t == 'r')
					trmask = 0400;
				else if (*t == 'b')
					trmask = 01000;
				else if (*t != 'g')
					break;
				continue;
			case '#':
				while (isdigit(*t))
					number = 10 * number + *t++ - '0';
				if (*t || number <= 0 || number > 128)
					break;
				--t;
				continue;
			default:
				if (*s == *t)
					continue;
				break;
			}
			break;
		}
		if (! *s && ! *t)
			break;
		if (! (++np)->n_name[0])
			return syscall(syssw, name, mode);
	}
	if (syssw == SYS_open || syssw == SYS_creat)
		++mode;
	/*
	** "command" is write only!
	*/
	if (np->n_iop == command)
		if ((syssw == SYS_open && (mode & FREAD) != 0) ||
		    (syssw == SYS_access && (mode & 03) != 0))
			return errnoset(EACCES);
	/*
	** Find an open slot.
	*/
	fp = comfeats;
	while (fp->f_devptr != 0)
		if (++fp >= &comfeats[NFIDS])
			return errnoset(EMFILE);
	/*
	** See if we get to figure out which station to use;
	** if we get to do so, take it from the COMTAL environment entry - pgf
	*/
	if (comdefaultstation == 0)
		if ((s = getenv("COMTAL")) == NULL || strlen(s) != 1)
			comdefaultstation = 1;
	else
			comdefaultstation = *s + 1 - (isupper(*s) ? 'A' : 'a');
	if (station == 0)
		station = comdefaultstation;
	if (station < 1 || station > 2)
		return errnoset(EACCES);
	if (syssw == SYS_stat || syssw == SYS_access)
		return syscall(syssw, comdevs[station-1].d_name, mode);
	if (comdevs[--station].d_count == 0) {
		register int	fid;

		if ((fid = comdevs[station].d_fid =
			syscall(SYS_open, comdevs[station].d_name, 2)) == WILD)
			return fid;
		else {
			register olderrno;

			olderrno = errno;
			ioctl(fid, FIOCLEX, (struct sgttyb *) NULL);
			errno = olderrno;
		}
	}
	fp->f_devptr = &comdevs[station];
	++(fp->f_devptr->d_count);
	fp->f_number = number;
	fp->f_offset = 0;
	fp->f_filesize = np->n_filesize;
	fp->f_iop = np->n_iop;
	fp->f_trcode = np->n_trcode;
	fp->f_trdata = np->n_trdata[station] | trmask;
	return fp - comfeats + getdtablesize() - NFIDS;
}

static	struct	feature *
comgetf(featdes)
	register featdes;
{
	register struct	feature *fp;

	if (featdes < getdtablesize() && featdes >= fidbase &&
		(fp = &comfeats[featdes - fidbase])->f_devptr != 0) {
			comcurfid = fp->f_devptr->d_fid;
			return fp;
	}
	errno = EBADF;
	return 0;
}

static
comclose(featdes)
{
	register struct	feature *fp;
	register int	olderrno;

	if ((fp = comgetf(featdes)) == 0)
		return WILD;
	if (--(fp->f_devptr->d_count) == 0) {
		olderrno = errno;
		(void)close(fp->f_devptr->d_fid);
		errno = olderrno;
	}
	fp->f_devptr = 0;
	return TAME;
}

static
comio(featdes, buffer, nbytes, mode)
	register char *	buffer;
	register	nbytes;
{
	register struct	feature *fp;
	register		filesize;
	register		blocksize;
	register		ndone;
	register		ntodo;
	register		donethistime;

	if (nbytes <= 0)
		return errnoset(EINVAL);
	if ((fp = comgetf(featdes)) == 0)
		return errnoset(EBADF);
	if (fp->f_iop == command)
		fp->f_offset = 0;
	if (fp->f_offset < 0)
		return errnoset(EINVAL);
	if ((filesize = fp->f_filesize) != 0)
		blocksize = filesize;
	else {
		if ((blocksize = comprep(fp, nbytes, mode)) <= 0)
			return WILD;
		filesize = fp->f_filesize;
		fp->f_filesize = 0;
	}
	if (fp->f_offset >= filesize)
		return 0;
	if (fp->f_offset + nbytes > filesize)
		nbytes = filesize - fp->f_offset;
	for (ndone = 0; ndone < nbytes; ndone += ntodo) {
		ntodo = nbytes - ndone;
		if (ntodo > blocksize - fp->f_offset % blocksize)
			ntodo = blocksize - fp->f_offset % blocksize;
		if (mode == FREAD || ntodo != blocksize) {
			/*
			** We get to do some reading.
			*/
			if (fp->f_iop == 0)
				donethistime = combufio(fp->f_trcode,
					fp->f_trdata|(fp->f_offset/blocksize),
					blocksize, FREAD);
			else	donethistime = (*fp->f_iop)(FREAD,
					fp->f_offset-fp->f_offset%blocksize,
					blocksize);
			if (donethistime != blocksize)
				return WILD;
			if (mode == FREAD)
				commute(&comraw.c_buf[fp->f_offset%blocksize],
					buffer, ntodo);
		}
		if (mode == FWRITE) {
			/*
			** We get to do some writing.
			*/
			commute(buffer, &comraw.c_buf[fp->f_offset%blocksize],
				ntodo);
			if (fp->f_iop == 0)
				donethistime = combufio(fp->f_trcode,
					fp->f_trdata|(fp->f_offset/blocksize),
					blocksize, FWRITE);
			else	donethistime = (*fp->f_iop)(FWRITE,
					fp->f_offset-fp->f_offset%blocksize,
					blocksize);
			if (donethistime != blocksize)
				return WILD;
		}
		buffer += ntodo;
		fp->f_offset += ntodo;
	}
	return ndone;
}

static long
comlseek(featdes, offset, whence)
	register long	offset;
{
	register struct	feature	*fp;

	if ((fp = comgetf(featdes)) == 0)
		return WILD;
	if (whence == 1)
		offset += fp->f_offset;
	else if (whence == 2)
		if (fp->f_filesize != 0)
			offset += fp->f_filesize;
		else {
			if (comprep(fp, 0, 0) <= 0)
				return WILD;
			offset += fp->f_filesize;
			fp->f_filesize = 0;
		}
	return fp->f_offset = offset;
}

static
comassoc(station, iorg, number, linesize, blocksize)
{
	register struct {
		unsigned hostnum	:2;
		unsigned		:5;
		unsigned iorg		:1;
		unsigned		:4;
		unsigned mode		:3;
	} i;
	struct {
		short	igp_hostnum;
		short	igp_visionnum;
		short	igp_nwpb;	/* Number of Words per Block */
		short	igp_nwpl;	/* Number of Words per Line */
		short	igp_nlpb;	/* Number of Lines per Block */
		short	igp_window;
	} igp;

	i.hostnum = station;
	igp.igp_hostnum = station;
	i.iorg = iorg;
	i.mode = 3;
	igp.igp_visionnum = number - 1;
	igp.igp_nwpb = blocksize / 2;
	igp.igp_nwpl = linesize / 2;
	igp.igp_nlpb = blocksize / linesize;
	igp.igp_window = 0;
	commute((char *)&igp, comraw.c_buf, sizeof igp);
	return (combufio(3, i, sizeof igp, FWRITE)==sizeof igp) ?
		TAME : WILD;
}

static
comprep(fp, nbytes, mode)
	register struct	feature *fp;
{
	register	blocksize;
	register	linesize;
	register	iorg;

	if (fp->f_iop == comcodeio) {
		if (mode == FWRITE)
			/* Just get as much code space as possible. */
			return fp->f_filesize = 4096;
		blocksize = fp->f_offset + nbytes;
		if ((blocksize % 256) != 0)
			blocksize += 256 - (blocksize % 256);
		if (blocksize > 4096)
			blocksize = 4096;
		return fp->f_filesize = blocksize;
	}
	if (fp->f_iop == command)
		return fp->f_filesize = nbytes;
	if (fp->f_iop != 0)
		abort();
	switch (fp->f_trdata) {
		case 060000:		/* function memory */
		case 060400:		/* station b function memory */
		case 062000:		/* small area memory */
		case 062400:		/* station b small area mem. */
			fp->f_filesize = 256*2;
			if (nbytes != 0)
				if (comassoc(fp->f_devptr->d_station,
					IMAGE, (int) fp->f_number,
					(int) fp->f_filesize,
					(int) fp->f_filesize) != TAME)
						return 0;
			return fp->f_filesize;
		case 000000:		/* image */
		case 004000:
		case 010000:
		case 014000:
			linesize = comnbpl(fp);
			fp->f_filesize = 512 * linesize;
			iorg = IMAGE;
			break;
		case 001000:		/* graphic */
		case 005000:
		case 011000:
		case 015000:
			linesize = comnbpl(fp);
			fp->f_filesize = 512 * linesize;
			iorg = GRAPHIC;
			break;
		default:		/* Internal glitch */
			abort();
	}
	if (nbytes == 0)
		return 0;
	if (fp->f_offset >= fp->f_filesize)
		return 0;
	if (fp->f_offset + nbytes > fp->f_filesize)
		nbytes = fp->f_filesize - fp->f_offset;
	blocksize = compute(fp->f_offset, (long)nbytes, linesize, fp->f_filesize, mode);
	if (comassoc(fp->f_devptr->d_station, iorg,
		(int) fp->f_number, linesize, blocksize) != TAME)
			return 0;
	return blocksize;
}

static
compute(firstbyte, nbytes, linesize, filesize, mode)
	long	firstbyte, nbytes, filesize;
{
	register	firstline, nlines, maxlines, linesper, firstio;

	firstline = firstbyte / linesize;
	nlines = (firstbyte + nbytes)/linesize - firstline;
	if (((firstbyte + nbytes) % linesize) != 0)
		++nlines;
	maxlines = MAXBYTES / linesize;
	if ((firstline % nlines) == 0)
		if (nlines <= maxlines)
			return nlines * linesize;
		else if ((nlines % maxlines) == 0)
			return maxlines * linesize;
	/*
	** So much for the clear-cut cases.
	*/
	filesize /= linesize;
	/*
	** If maxlines is small do a direct calculation.
	*/
	if (maxlines <= 32) {	/* 512 x 512 image, for example */
		register	bestlinesper, bestcalls, calls, extras;

		bestcalls = nlines + 6;	/* soon to be bested! */
		bestlinesper = 1;	/* just in case, though. . .*/
		linesper = nlines/((nlines - 1)/maxlines + 2);
		for (;linesper <= maxlines;++linesper) {
			firstio = firstline - firstline % linesper;
			calls = (firstline + nlines - firstio) / linesper + 1;
			if (firstio + calls * linesper > filesize)
				continue;
			if ((mode & FWRITE) != 0) {
				extras = ((firstbyte%(linesper*linesize))!=0)+
				(((firstbyte+nbytes)%(linesper*linesize))!=0);
				if (extras > 1 && calls == 1)
					extras = 1;
				calls += extras;
			}
			if (calls < bestcalls) {
				bestcalls = calls;
				bestlinesper = linesper;
				if (calls <= (nlines - 1) / maxlines + 1)
					break;
			}
		}
		return bestlinesper * linesize;
	}
	/*
	** Try to find one block that covers it all.
	*/
	for (linesper = nlines + 1;linesper <= maxlines;) {
		register	shy;

		firstio = firstline - firstline % linesper;
		if ((shy = firstline + nlines - (firstio + linesper)) <= 0)
			if (firstio + linesper <= filesize)
				return linesper * linesize;
			else	shy += linesper;
		linesper += (shy - 1) / (firstio /linesper + 1) + 1;
	}
	/*
	** We've messed with it long enough.
	*/
	return maxlines * linesper;
}

static short
comtargetio(mode)
{
	register struct {
		unsigned location	:9;
		unsigned		:3;
		unsigned mode		:3;
	} i;
	register short *	p;

	i.mode = 0;
	if (mode == FREAD)
		return combufio(3, i, 2*2, FREAD);
	p = (short *) comraw.c_buf;
#ifdef MIT-ERL
	{
		register n, x, y;
		x = *p++;
		y = *p;
		(void)sprintf(comraw.c_buf, "S TAR %d %d\n", x, y);
		n = command(0, 0, strlen(comraw.c_buf));
		*p = y;
		*--p = x;
		return n;
	}
#else
	i.location = *p++;
	if (combufio(3, i, 0, FWRITE) != 0)
		return WILD;
	i.location = *p++;
	i.mode = 4;
	return (combufio(3, i, 0, FWRITE) == 0) ? 2*2 : WILD;
#endif
}

/*ARGSUSED*/
static short
comcodeio(mode, offset, nbytes)
	long	offset;
{
	register struct {
		unsigned length	:7;
		unsigned zeroes	:4;
		unsigned r11	:1;
		unsigned mode	:3;
	} i;
	register		olderrno;
	register char *		cp;
	register		n;

	i.length = nbytes / 256;
	i.zeroes = 0;
	i.r11 = 0;
	i.mode = 6;
	olderrno = errno;
	if (combufio(3, i, nbytes, mode) == nbytes)
		return nbytes;
	i.length = 1;	/* Set up for upcoming tests */
	if (mode == FWRITE) {
		char	savearea[20];

		commute(comraw.c_buf, savearea, sizeof savearea);
		if (nbytes != 256 && combufio(3, i, 256, FWRITE) == 256) {
			/* There's code to release */
			commute("rel cod $\0", comraw.c_buf, 10);
			(void)combufio(3, 024001, 128*2, FWRITE);
		}
		(void)sprintf(comraw.c_buf, "s cod %d $", nbytes / 256);
		(void)combufio(3, 024001, 128*2, FWRITE);
		commute(savearea, comraw.c_buf, sizeof savearea);
		errno = olderrno;
		i.length = nbytes / 256;
		return combufio(3, i, nbytes, mode);
	}
	cp = comraw.c_buf;
	n = nbytes;
	do *cp++ = 0;
		while (--n != 0);
	if (combufio(3, i, 256, FREAD) != 256) {
		/* nothing to read */
		errno = olderrno;
		return nbytes;
	}
	for (n = nbytes - 256; n > 256; n -= 256) {
		i.length = n / 256;
		if (combufio(3, i, n, FREAD) == n) {
			errno = olderrno;
			return nbytes;
		}
	}
	return WILD;
}

/*ARGSUSED*/
static short
command(mode, offset, nbytes)
{
	register struct {
		unsigned length	:6;
		unsigned zeroes	:5;
		unsigned r11	:1;
		unsigned mode	:3;
	} i;
	static char	saved[128*2];
	static char	*to = saved;
	register char	*from;
	register	nslots;
	register	ntodo;
	char		temp[128*2];

	from = comraw.c_buf;
	do {
		ntodo = nbytes - (from - comraw.c_buf);
		nslots = (sizeof saved - 2) - (saved - to);
		if (ntodo > nslots)
			ntodo = nslots;
		while ((*to++ = *from++) != '\n' && --ntodo != 0);
		if (ntodo == 0) { /* we've yet to find a '\n' */
			if (from < &comraw.c_buf[nbytes]) {
				to = saved;
				return errnoset(EINVAL);
			}
			return nbytes;
		}
		/* '\n' found!! */
		--to;	/* point to '\n' */
		if (to == saved)
			return errnoset(EINVAL);
		--to;	/* point to last char */
		if (*to != '$') {
			if (*to != ' ')
				*++to = ' ';
			*++to = '$';
		}
		*++to = '\0';
		++to;
		ntodo = to - saved;
		to = saved;
		/*
		** First, get the mule's attention.
		*/
		i.length = 0;
		i.r11 = 1;
		i.mode = 2;
		i.zeroes = 0;
		if (combufio(3, i, 0, FWRITE) != 0)
			return COMWILD;
		i.length = 1;
		/*
		** "Efficiency" bufs, prepare to cringe.
		*/
		commute(comraw.c_buf, temp, ntodo);
		commute(saved, comraw.c_buf, ntodo);
		if (combufio(3, i, 128*2, FWRITE) != 128*2)
			return WILD;
		commute(temp, comraw.c_buf, ntodo);
	} while (from < &comraw.c_buf[nbytes]);
	return nbytes;
}

comfstat(fid, statp)
	istat	*statp;
{
	static struct {
		unsigned i_hostnum	:2;
		unsigned 		:5;
		unsigned i_iorg		:1;
		unsigned 		:4;
		unsigned i_mode		:3;
		unsigned 		:1;
	} i;
	static struct {
		unsigned j_visionnum	:8;
		unsigned j_usernum	:2;
		unsigned j_one		:1;
		unsigned j_hostnum	:2;
		unsigned j_anotherone	:2;
		unsigned j_iorg		:1;
	} j;
	register	iorg;
	register struct	feature *fp;

	if ((fp = comgetf(fid)) == 0)
		return WILD;
	switch (fp->f_trdata) {
		case 000000:
		case 004000:
		case 010000:
		case 014000:
			iorg = 0;
			break;
		case 001000:
		case 005000:
		case 011000:
		case 015000:
			iorg = 1;
			break;
		default:
			return errnoset(EBADF);
	}
	j.j_visionnum = fp->f_number - 1;
	j.j_usernum = j.j_hostnum = fp->f_devptr->d_station;
	j.j_one = 1;
	j.j_anotherone = 1;
	j.j_iorg = iorg;
	if (combufio(1, j, 0, FWRITE) != 0)
		return errnoset(EIO);
	i.i_hostnum = fp->f_devptr->d_station;
	i.i_mode = 7;
	i.i_iorg = iorg;
	if (combufio(3, i, sizeof *statp, FREAD) != sizeof *statp)
		return errnoset(EIO);
	commute(comraw.c_buf, (char *)statp, sizeof *statp);
	return TAME;
}

comstat(name, statp)
	char	*name;
	istat	*statp;
{
	register fid;
	register retval;

	fid = comopen(SYS_creat, name, 3);
	if (fid >= getdtablesize() || fid < fidbase)
		return WILD;
	retval = comfstat(fid, statp);
	(void)close(fid);
	return retval;
}
