#ifndef lint
static char *sccsid = "%Z%$RCSfile: comsubs.c,v $	GIPS $Revision: 2.1 $ (MIT/CSR) $Date: 1986/07/27 04:56:50 $";
#endif
/*		   Comtal Image Processor Utility
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		comsubs.c	Creation Date:	04/28/84
**	Version:	$Revision: 2.1 $		Update:		$Date: 1986/07/27 04:56:50 $
**	Function:	Device independent GIPS display routines
**	Components:
**		image_open()	return fildes to opened image plane
**		graphic_open()	return fildes to opened graphic plane
**		pseudo_open()	return fildes to opened ps table
**		function_open()	return fildes to opened stretch table
**		command_open()	return stream ptr to command input
**		display_put()	set display origin
**		target_get()	return target coordinates
**		target_put()	reposition target
**		target_roam()	set target for roaming
**		terget_dump()	set target for dumping
**		color_get()	return graphic color
**		color_put()	set graphic color
**		print_subtab()	write command table to stream
**		comtal_error()	error message string generator
*/

#include	<stdio.h>
#include	<ctype.h>
#include	"comsubs.h"
#include	"gips.h"

static	char	genclrc[]		= "%s/.clrc";
static	char	genimage[]		= "/dev/image.%d";
static	char	gengraphic[]		= "/dev/graphic.%d";
static	char	gentarget[]		= "/dev/target";
static	char	gencommand[]		= "/dev/command";
static	char	genpseudocolor[]	= "/dev/pseudocolor.%c";
static	char	genfunction[]		= "/dev/function.%d";

static	struct	ctab {
	char	*cname;
	char	*insert[2];
	char	*ccmd;
} comsubtab[]	= {
	"image_open",	"dni)", NULL,	"SU A SU B D I %(ni) A PS ",
	"graphic_open",	"dng)", "snc)",	"CO G %(ng) %(nc) A G %(ng)",
	"pseudo_open",	"cnp)", NULL,	"A PS %(np)",
	"function_open","dnf)", NULL,	"A F %(nf)",
	"display_put",	"dx)",  "dy)",	"S D %(x) %(y)",
	"target_get",	"dnt)", NULL,	"",
	"target_put",	"dx)",  "dy)",	"A T S TAR %(x) %(y)",
	"target_roam",	"dnt)", NULL,	"SU A A B A T RO I",
	"target_dump",	"dnt)", NULL,	"SU A A B A T DU I",
	NULL
};

static	int	zooms[] = { 0, 0, 2, 1 };
static	char	buf[128];
extern	char	*sprintf(), *strcpy(), *getenv(), *index();

static
comtal_error()
{
	register char 	s, *t;

	for (s = buf+5, t = *gipserrs; *t = *s; s++, t++)
		if (*t == '.')
			*t = ' ';
	gipserrno = GIPSBADOPEN;
	return -1;
}

print_subtab(s, fp)
	register char	*s;
	register FILE	*fp;
{
	register struct	ctab	*t;

	if (s == NULL) s = "";
	if (fp == NULL) fp = stdout;
	for (t = comsubtab; t->cname; t++)
		fprintf(fp, "%s%s\t: \"%s\"\n", s, t->cname, t->ccmd);
}

FILE *
command_open()
{
	register	i, l;
	register struct ctab *t;
	register char	*s;
	static	 FILE	*fcom;
	static	 char	*combuf;

	if (combuf == NULL && (combuf = getenv("HOME")) != NULL &&
			(i = open(sprintf(buf, genclrc, combuf), 0)) >= 0) {
		(void)lseek(i, 0L, 2);
		combuf = alloc(l = lseek(i, 0L, 1)+1, 1);
		(void)lseek(i, 0L, 0);
		(void)read(i, s = combuf, l);
		(void)close(i);
		for (s = combuf; *s; *s++ = 0) {
			for (t = comsubtab; t->cname != NULL &&
				strncmp(s, t->cname, i = strlen(t->cname)); t++)
				;
			if (t->cname != NULL)
				t->ccmd = s+i+1;
			if ((s = index(s, '\n')) == NULL)
				break;
		}
	}
	(void)strcpy(*gipserrs, "Comtal");
	if (fcom == NULL || (fcom->_flag & (_IOWRT | _IORW)) == 0)
		fcom = fopen(gencommand, "w");
	if (fcom == NULL) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	fprintf(fcom, "$\n");
	(void)fflush(fcom);
	if (! ferror(fcom))
		return fcom;
	(void)fclose(fcom);
	gipserrno = GIPSBADWRITE;
	return fcom = NULL;
}

static
comout(n, i1, i2)
{
	register FILE	*fp;
	register char	*cp, *s;
	register	i;
	char		*fmt = "%d"; 

	if ((fp = command_open()) == NULL)
		return -1;
	cp = comsubtab[n].ccmd;
	if (! *cp)
		return 0;
	for (; *cp; cp++)
		if (*cp != '%')
			putc(islower(*cp) ? toupper(*cp) : *cp, fp);
		else
			for (i = 0; i < 2 && (s = comsubtab[n].insert[i]); i++)
				if (strncmp(s+1, cp+2, strlen(s+1)) == 0) {
					cp += strlen(s+1)+1;
					fmt[1] = *s;
					fprintf(fp, fmt, i ? i2 : i1);
					break;
				}
	fputs(" \n", fp);
	(void)fflush(fp);
	return ferror(fp) ? -1 : 0;
}

image_open(ni, mode)
{
	register fildes;

	if ((fildes = open(sprintf(buf, genimage, ni), mode)) < 0)
		return comtal_error();
	(void)comout(0, ni, 0);
	return fildes;
}

graphic_open(ng, mode)
{
	register fildes, nc;
	gstat	gs;

	if ((fildes = open(sprintf(buf, gengraphic, ng), mode)) < 0)
		return comtal_error();
	if (comfstat(fildes, (istat *)&gs) != COMTAME)
		return fildes;
	if ((nc = (gs.gs_color & 0170) >> 3) == 0)
		nc = 15;
	(void)color_put(ng, nc);
	return fildes;
}

pseudo_open(np, mode)
{
	register	fildes;
	static	char	*colors[] = { "red", "green", "blue" };

	if (np < 0 || np > 2)
		return -1;
	if ((fildes = open(sprintf(buf, genpseudocolor, "rgb"[np]), mode)) < 0) {
		(void)strcpy(buf+17, colors[np]);
		return comtal_error();
	}
	(void)comout(2, "RGB"[np], 0);
	return fildes;
}

function_open(nf, mode)
{
	register	fildes;

	if ((fildes = open(sprintf(buf, genfunction, nf), mode)) < 0)
		return comtal_error();
	(void)comout(3, nf, 0);
	return fildes;
}

display_put(nt, xy)

	register	int	*xy;
{
	return comout(4, xy[0], xy[1]);
}

target_get(nt, xy)

	register int	*xy;
{
	register nbytes;
	register fildes;
	short	spixel[2];
	istat	is;

	if (comout(5, xy[0], xy[1]))
		return -1;
	(void)strcpy(*gipserrs, "target");
	if ((fildes = open(gentarget, 0)) == NULL) {
		gipserrno = GIPSBADOPEN;
		return -1;
	}
	nbytes = read(fildes, (char *)spixel, 4);
	(void)close(fildes);
	if (nbytes != 4) {
		gipserrno = GIPSBADREAD;
		return -1;
	}
	if (comstat(sprintf(buf, genimage, nt), &is) != COMTAME)
		return comtal_error();
	xy[0] = is.is_shownx + (spixel[0] >> zooms[is.is_zoom]);
	xy[1] = is.is_showny + (spixel[1] >> zooms[is.is_zoom]);
	return gipserrno = 0;
}

target_put(nt, xy0)

	register int	*xy0;
{
	int	xy[2];
	istat	is;
	static	count;

	if (comstat(sprintf(buf, genimage, nt), &is) != COMTAME) {
		(void)strcpy(buf, gentarget);
		return comtal_error();
	}
	xy[0] = (xy0[0] - is.is_shownx) << zooms[is.is_zoom];
	xy[1] = (xy0[1] - is.is_showny) << zooms[is.is_zoom];
	if ((xy[0] | xy[1]) & ~0777) {
		if (count) {
			count = 0;
			return 1;
		}
		if ((xy[0] = xy0[0] - (256>>zooms[is.is_zoom])) < 0) xy[0] = 0;
		if ((xy[1] = xy0[1] - (256>>zooms[is.is_zoom])) < 0) xy[1] = 0;
		if (display_put(nt, xy) != 0)
			return -1;
		count++;
		return target_put(nt, xy);
	}
	count = 0;
	return comout(6, xy[0], xy[1]);
}

target_roam(nt)
{
	return comout(7, nt, 0);
}

target_dump(nt)
{
	return comout(8, nt, 0);
}

color_get(ng)
{
	gstat	gs;

	if (comstat(sprintf(buf, gengraphic, ng), (istat *)&gs) != COMTAME)
		return comtal_error();
	else
		return (gs.gs_color >> 3) & 017;
}

color_put(ng, nc)
{
	static char *colors[] = {
		"BLA", "RED", "DGR", "LOR", "FGR", "ORN", "GRN", "YEL",
		"BLU", "MAG", "DTU", "PNK", "TUR", "LPI", "CYN", "WHT"
	};
	if (nc < 0 || nc >= 16)
		return -1;
	return comout(1, ng, colors[nc]);
}
