#ifndef lint
static char *sccsid = "%Z%$RCSfile: gipscolor.c,v $	GIPS $Revision: 2.3 $ (MIT/CSR) $Date: 1987/06/22 11:01:51 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1986 Massachusetts Institute of Technology
**
**	Module:		gipscolor.c	Creation Date:	03/19/86
**	Version:	$Revision: 2.3 $		Update:		$Date: 1987/06/22 11:01:51 $
**	Function:	Display GIPS colormap wedge in Sun window
**	Components:
**		main()			GIPS/UNIX SunTool colormap display
**		sigwinched		main window-change entry
**		image_sigwinched()	Handle our window changes
**		image_selected()	Handle keyboard interrupts
**		set_input_mask		set SunTool keyboard interrupts
**		read_colormap()		set colormap from input stream
*/
#include	<stdio.h>
#include	<suntool/tool_hs.h>
#include	<sunwindow/cms.h>
#include	<suntool/menu.h>
#include	<signal.h>
#include	<ctype.h>
#include	"gips.h"

#define		CTRL(c)		('c' & 037)
#define		TOOL_NAME	"GipsColor $Revision: 2.3 $"
#define		RGBSIZE		(Nflag ? 256 : 128)
	/*
	**	Input flags
	*/
static	flag	Nflag;		/* use 8-bit colormap */
static	flag	Sflag;		/* print statistics */
static	flag	rflag;		/* set for reversed color wedge */
static	flag	vflag;		/* set for vertical color wedge */
static	char	*Cname = "bw";	/* colormap name */
	/*
	**	table used by gipsflags()
	*/
static	flagentry	flagtable[] = {
	"N",	"1",			(flag *) &Nflag,
	"S",	"1",			(flag *) &Sflag,
	"r",	"1",			(flag *) &rflag,
	"v",	"1",			(flag *) &vflag,
	"C",	"-> mapname",		(flag *) &Cname,
	NULL,	"[colormap file]]"
};
	/*
	**	The tool icon
	*/
static	short	image_icon[] = {
#include "gipscolor.icon"
};
 
static	mpr_static(icon_pr, 64, 64, 1, image_icon);
 
static	struct	icon	icon = {
	TOOL_ICONWIDTH, TOOL_ICONHEIGHT, NULL,
	{ 0, 0, TOOL_ICONWIDTH, TOOL_ICONHEIGHT },
	&icon_pr, { 0, 0, 0, 0 }, NULL, NULL, 0
};
	/*
	**	Tool structures
	*/
static	struct	tool		*tool;		/* describes the tool */
static	struct	toolsw		*image_sw;	/* image subwindow */
static	int			image_fd;	/* image file descriptor */
static	struct	pixwin		*image_pw;	/* current image sub-area */
static	struct	pixrect		*saved_pr;	/* saved color wedge */
static	int			must_update;	/* window repaint */
static	char	ifile[BUFSIZ];			/* input colormap file name */
static	char	colormap[CMS_NAMESIZE];		/* shared colormap name */
static	char	rgb[3][256];;			/* colormap values */
	/*
	**	The cursor menu
	*/
static	struct	menuitem	menu_item[] = {
{ MENU_IMAGESTRING,	"^I  Read new Colormap File",	(caddr_t) CTRL(I) },
{ MENU_IMAGESTRING,	"^R  Reverse Color Wedge",	(caddr_t) CTRL(R) },
{ MENU_IMAGESTRING,	"^V  Rotate Color Wedge",	(caddr_t) CTRL(V) },
{ MENU_IMAGESTRING,	"^C  Quit GipsColor",		(caddr_t) CTRL(C) }
};
static	struct	menu	menu = {
	MENU_IMAGESTRING,	NULL,	4,	menu_item
};

main(argc,argv)

	int	argc;			/* number of arguments */
	char	**argv;			/* array of argument pointers */
{
	int	planes = -1;		/* image plane mask */
	char	*strncpy();		/* C library */
	char	**attrs	= NULL;		/* tool attribute list */
	/*
	**	parse tool-specific arguments
	*/
	argc--;
	if (tool_parse_all(&argc, argv+1, &attrs, *argv) == -1) {
		tool_usage(*argv);
		exit(1);
	}
	/*
	**	save and parse callers arguments
	*/
	saveargs(argv);
	argc = gipsflags(argc+1, argv, flagtable);
	if (--argc > 1)
		error("bad arg: %s", argv[1]);
	else if (argc == 1 &&
			strcmp(*++argv, "-") &&
			! exist(".", *argv, "") &&
			! exist(".", *argv, ".cmap") &&
			! exist(GIPSCMAPDIR, *argv, "") &&
			! exist(GIPSCMAPDIR, *argv, ".cmap"))
			gipserr2(GIPSBADOPEN, *argv);
	if (Sflag)
		telltale(0);
	/*
	**	create the tool
	*/
	if ((tool = tool_make(
		WIN_NAME_STRIPE, 0,
		WIN_ICON, &icon,
		WIN_ICON_LABEL, "GipsColor",
		WIN_ATTR_LIST, attrs, 0)) == NULL)
			error("can't create tool","");
	tool_free_attribute_list(attrs);
	/*
	**	create a generic sub-window and associated pixwin
	*/
	if ((image_sw = tool_createsubwindow(tool, "", 
		TOOL_SWEXTENDTOEDGE, TOOL_SWEXTENDTOEDGE,
			NULL)) == NULL)
		error("can't create image sub-window","");
	image_fd = image_sw->ts_windowfd;
	if ((image_pw = pw_open(image_fd)) == NULL)
		error("can't open image sub-window","");
	if (image_pw->pw_pixrect->pr_depth != 8)
		error("depth not 8", "");
	pw_putattributes(image_pw, &planes);
	pw_setcmsname(image_pw, strncpy(colormap, Cname, CMS_NAMESIZE-1));
	pw_putcolormap(image_pw, 0, RGBSIZE, rgb[0], rgb[1], rgb[2]);
	/*
	**	Finally, enable interrupts and start up the tool
	**	Let the window manager read the image
	*/
	set_inputmask();
	tool_install(tool);
	tool_select(tool, 0);
	tool_destroy(tool);
}
	/*
	**	The tool manager handles all tool damage.
	**	set_inputmask() is in select.c
	*/
static
sigwinched()
{
	tool_sigwinch(tool);
}
	/*
	**	Repair damage
	*/
static
image_sigwinched(sw)
	caddr_t	sw;
{
	struct	rect	r;
	static	int	notinit;
	/*
	**	Get a damage rectlist
	*/
	if (! notinit++ && ifile[0])
		read_colormap();
	pw_damaged(image_pw);
	win_getsize(image_fd, &r);
	/*
	**	Determine whether window has stretched
	*/
	if (saved_pr == NULL || must_update ||
			r.r_width != saved_pr->pr_size.x ||
			r.r_height != saved_pr->pr_size.y) {

		register x, y, i, l;
		register char *p;

		pw_donedamaged(image_pw);
		if (saved_pr != NULL) pr_close(saved_pr);
		saved_pr = mem_create(r.r_width, r.r_height, 8);
		if (saved_pr == NULL) error("no more memory", "");

		p = (char *)mpr_d(saved_pr)->md_image;
		l = mpr_d(saved_pr)->md_linebytes;
		if (vflag) {
			for (y = 0; y < r.r_height; y++, p += l) {
				i = ((rflag ? (r.r_height - y) : y)
					<< (Nflag ? 8 : 7)) / r.r_height;
				for (x = 0; x < r.r_width; x++)
					p[x] = i;
			}
		} else {
			for (x = 0; x < r.r_width; x++)
				p[x] = ((rflag ? (r.r_width - x) : x)
					<< (Nflag ? 8 : 7)) / r.r_width;
			for (y = 1; y < r.r_height; y++, p += l)
				bcopy(p, p+l, r.r_width);
		}
		must_update++;
	}
	/*
	**	Paint the image
	*/
	pw_write(image_pw, 0, 0, r.r_width, r.r_height,
		PIX_SRC, saved_pr, 0, 0);

	if (must_update)
		must_update = 0;
	else
		pw_donedamaged(image_pw);
}
	/*
	**	Keyboard/mouse handler
	*/
/*ARGSUSED*/
static
image_selected(sw, ibits, obits, ebits, timer)
	caddr_t	sw;				/* image sub-window */
	int	*ibits,	*obits,	*ebits;		/* interrupt masks */
	struct	timeval	**timer;		/* interval timer */
{
	struct	inputevent ie;

	if (*ibits & (1 << image_fd)) {
		if (input_readevent(image_fd, &ie))
			error("select dies", "");
		if (ie.ie_code == MS_RIGHT) {
			struct menu *menu_ptr = &menu;
			struct	menuitem *mi;
			char	buf[256];
			menu.m_imagedata = (caddr_t)sprintf(buf,
				"%s: Colormap \"%s\"", TOOL_NAME, colormap);
			if (mi = menu_display(&menu_ptr, &ie, image_fd))
				ie.ie_code = (short) mi->mi_data;
		}
		switch (ie.ie_code) {
		case CTRL(I):
			read_colormap();
			break;
		case CTRL(R):
			rflag ^= 1;
			must_update++;
			image_sigwinched();
			break;
		case CTRL(V):
			vflag ^= 1;
			must_update++;
			image_sigwinched();
			break;
		case CTRL(C):
			tool_destroy(tool);
			if (Sflag)
				telltale(1);
			exit(0);
		}
	}
	*ibits = *obits = *ebits = 0;
}
	/*
	**	Set the keyboard and mouse interupt mask
	*/
static
set_inputmask()
{
	struct inputmask im;

	image_sw->ts_io.tio_selected = image_selected;
	input_imnull(&im);
	im.im_flags |= IM_ASCII;
	win_setinputcodebit(&im, MS_LEFT);
	win_setinputcodebit(&im, MS_MIDDLE);
	win_setinputcodebit(&im, MS_RIGHT);
	win_setinputmask(image_fd, &im, NULL, WIN_NULLLINK);
	image_sw->ts_io.tio_handlesigwinch = image_sigwinched;
	signal(SIGWINCH, sigwinched);
}
	/*
	**	Read the colormap
	*/
read_colormap()
{
	char	buf[BUFSIZ];		/* input buffer */
	double	d[3];			/* input {rgb} color triplet */
	register char *c;		/* buffer pointer */
	register i, n;			/* scratch */
	/*
	**	Disregard comment lines beginning with "#"
	**	If -N flag set, accept all RGBSIZE entries.
	**	Otherwise, skip every other entry.
	*/
	for (i = n = 0; i < 256; i++) {
		if (fgets(buf, sizeof(buf)-1, stdin) == NULL)
			break;
		for (c = buf; isspace(*c); c++)
			;
		if (*c == '#')
			continue;
		if ((c[0] == 'E' || c[0] == 'e') &&
		    (c[1] == 'N' || c[1] == 'n') &&
		    (c[2] == 'D' || c[2] == 'd') &&
		    (c[3] == '\0' || isspace(c[3])))
			if (i == 0)
				continue;
			else
				break;
		if (sscanf(c, " %lf %lf %lf ", &d[0], &d[1], &d[2]) != 3) {
			fprintf("%s: bad line: %s\n", ifile, buf);
			break;
		}
		if (! Nflag && (i & 01))
			continue;
		gconv(buf, (char *)d, 3, CVT_B, CVT_D);
		rgb[0][n] = buf[0];
		rgb[1][n] = buf[1];
		rgb[2][n] = buf[2];
		n++;
	}
	pw_putcolormap(image_pw, 0, RGBSIZE, rgb[0], rgb[1], rgb[2]);
}
	/*
	**	Try to open the colormap file
	*/
static
exist(dir, name, suf)
     char	*dir, *name, *suf;
{
	return freopen(sprintf(ifile, "%s/%s%s", dir, name, suf), "r", stdin)
		!= NULL;
}
