#ifndef lint
static char *sccsid = "%Z%$RCSfile: azimuthal.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1990/11/25 20:22:34 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		azimuthal.c	Creation Date:	06/29/87
**	Version:	$Revision: 2.7 $		Update:		$Date: 1990/11/25 20:22:34 $
**	Components:	equidistant()	ihgrid() function
**			equivalent()	""
**			gnomonic()	""
**			lambert4()	""
**			orthographic()	""
**			polstereo()	""
**			stereographic()	""
**
**	Function:	Transform an image via azimuthal mappings
*/

#define		GRID
#include	<math.h>
#include	"gips.h"

equidistant(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = M_PI_2 * t->radius;
		return 0;
	case 1:
		v = sin(y);
		vv = sin(x) * cos(y);
		if (v == 0 && vv == 0) return 1;
		vv = atan2(v, vv);
		v = t->radius * acos(cos(x) * cos(y));
		xy[0] = v * cos(vv);
		xy[1] = v * sin(vv);
		return 0;
	case 2:
		v = hypot(x, y);
		if (v > t->rlim) return 1;
		v = (v == 0) ? y / t->radius : y * sin(v / t->radius) / v;
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		if (y == 0) { xy[0] = x / t->radius; return 0; }
		v = x * tan(xy[1]) / y;
		if (v > 1 || v < -1) return 1;
		xy[0] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

equivalent(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = 2 * t->radius * t->radius;
		return 0;
	case 1:
		v = 0.5 * (1 + cos(x) * cos(y));
		if (v <= 0) return 1;
		v = t->radius / sqrt(v);
		xy[0] = v * cos(y) * sin(x);
		xy[1] = v * sin(y);
		return 0;
	case 2:
		v = t->rlim + t->rlim - x * x - y * y;
		if (v < t->rlim) return 1;
		vv = sqrt(v) / t->rlim;
		v = y * vv;
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		v = (y == 0) ? x * vv : x * tan(xy[1]) / y;
		if (v > 1 || v < -1) return 1;
		xy[0] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

gnomonic(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;
	
	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		v = cos(x);
		if (v == 0) return 1;
		xy[0] = t->radius * tan(x);
		xy[1] = t->radius * tan(y) / v;
		return 0;
	case 2:
		if (t->radius == 0 && x == 0) return 1;
		xy[0] = atan2(x, t->radius);
		xy[1] = atan2(y, hypot(x, t->radius));
		return 0;
	default:
		break;
	}
	return -1;
}

lambert4(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = 2 * t->radius;
		return 0;
	case 1:
		v = t->rlim * sin(M_PI_4 - 0.5 * y);
		xy[0] = v * sin(x);
		xy[1] = -v * cos(x);
		return 0;
	case 2:
		if (x == 0 && y == 0) return 1;
		xy[0] = atan2(x, -y);
		v = hypot(x, y) / t->rlim;
		if (v > 1 || v < -1) return 1;
		xy[1] = M_PI_2 - 2 * asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

orthographic(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		xy[0] = t->radius * cos(y) * sin(x);
		xy[1] = t->radius * sin(y);
		return 0;
	case 2:
		v = y / t->radius;
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		v = (y == 0) ? x / t->radius : x * tan(xy[1]) / y;
		if (v > 1 || v < -1) return 1;
		xy[0] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

polstereo(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = 2 * t->radius;
		return 0;
	case 1:
		v = t->rlim * tan(M_PI_4 - 0.5 * y);
		if (errno) return 1;
		xy[0] = sin(x) * v;
		xy[1] = -cos(x) * v;
		return 0;
	case 2:
		if (x == 0 && y == 0) return 1;
		xy[0] = atan2(x, -y);
		xy[1] = M_PI_2 - 2 * atan2(hypot(x, y), t->rlim);
		if (errno) return 1;
		return 0;
	default:
		break;
	}
	return -1;
}

stereographic(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = 4 * t->radius;
		return 0;
	case 1:
		v = 1 + cos(x) * cos(y);
		if (v == 0) return 1;
		xy[0] = 2 * t->radius * cos(y) * sin(x) / v;
		xy[1] = 2 * t->radius * sin(y) / v;
		return 0;
	case 2:
		v = (x * x + y * y) / t->rlim;
		if (v > t->radius) return 1;
		v = y / (v + t->radius);
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		v = (y == 0) ? x / (t->radius + x * x / t->rlim)
			: x * tan(xy[1]) / y;
		if (v > 1 || v < -1) return 1;
		xy[0] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}
