#ifndef lint
static char *sccsid = "%Z%$RCSfile: comlabel.c,v $	GIPS $Revision: 2.8 $ (MIT/CSR) $Date: 2007/11/19 18:38:27 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		comlabel.c		Creation Date:	12/30/83
**	Version:	$Revision: 2.8 $			Update:		$Date: 2007/11/19 18:38:27 $
**	Components:	setfont()		Change or report font name
**			setauxfont()		Change auxilliary font
**			settype()		Change or report image/graphic
**			setink()		Change or report pixel value
**			setorigin()		Change or report text origin
**			comlabel()		Update image/graphic file
**			memlabel()		Update image/graphic array
**			commarker()		Update file with offset
**			memmarker()		Update array with offset
**			gilabel()		Update cell-format image
**			gimarker()		Update cell-format with offset
**			comheight()		Return max height in pixels
**			comwidth()		Return max width in pixels
**			memory()		Internal - write into memory
**			label()			Internal - buffer image file
**			getfont()		Internal - read a font
**			getdisp()		Internal - get dispatch pointer
**			marker()		Internal - get pixel offset
**
**	Function:	Write Berkeley-Font character graphics to image
**			(byte pixel) or graphic (bit pixel) array or file.
*/

#include        <stdlib.h>   /* for malloc */
#include        <string.h>   /* for strlen */
#include	<stdio.h>
#include	<gips.h>
#include	<vfont.h>
#include	<ctype.h>
	/*
	**	mapping of font definition file
	*/
static struct packed {
	char	fontname[FONTLEN];	/* name of font */
	char	fontskel[FONTLEN];	/* skeleton font name */
	char	fontrot;		/* 'r' if rotated, '\0' if not */
	char	fonttype;		/* type "RIBS" */
	int	fontsize;		/* point size */
	struct header	h;		/* font file header */
	struct dispatch ds[256];	/* character sizes */
	char		maps[1];	/* character bit maps */
} *psave[4], *p;
	/*
	**	table of \(xx codes
	*/
static	struct	spectab	{
	char	key[3];		/* two-character mnemonic */
	char	val;		/* CAT value */
	char	flg;		/* =1 if on symbol font */
} spectab[] = {
	'f', 'i', 0, 0001, 0, 'f', 'l', 0, 0002, 0, 'f', 'f', 0, 0003, 0,
	'h', 'y', 0, 0004, 0, 'r', 'u', 0, 0005, 0, 'e', 'm', 0, 0006, 0,
	'b', 'u', 0, 0007, 0, 's', 'q', 0, 0010, 0, 'F', 'i', 0, 0011, 0,
	'F', 'l', 0, 0012, 0, 'f', 'I', 0, 0011, 0, 'f', 'L', 0, 0012, 0,
	'd', 'e', 0, 0013, 0, 'd', 'g', 0, 0014, 0, 'f', 'm', 0, 0015, 0,
	'c', 'o', 0, 0016, 0, 'r', 'g', 0, 0017, 0, 'c', 't', 0, 0020, 0,
	'1', '4', 0, 0021, 0, '1', '2', 0, 0022, 0, '3', '4', 0, 0023, 0,
	'i', 'f', 0, 0001, 1, 'i', 'p', 0, 0002, 1, 'p', 't', 0, 0003, 1,
	'r', 'h', 0, 0004, 1, 'c', 'u', 0, 0005, 1, 'r', 'n', 0, 0006, 1,
	'b', 's', 0, 0007, 1, '+', '-', 0, 0010, 1, '<', '=', 0, 0011, 1,
	'>', '=', 0, 0012, 1, 's', 'r', 0, 0013, 1, 't', 's', 0, 0014, 1,
	'i', 's', 0, 0015, 1, 's', 'l', 0, 0016, 1, 'b', 'v', 0, 0017, 1,
	'l', 'f', 0, 0020, 1, 'r', 'f', 0, 0021, 1, 'l', 'c', 0, 0022, 1,
	'r', 'c', 0, 0023, 1, 'l', 't', 0, 0024, 1, 'l', 'b', 0, 0025, 1,
	'r', 't', 0, 0026, 1, 'r', 'b', 0, 0027, 1, 'l', 'k', 0, 0030, 1,
	'r', 'k', 0, 0031, 1, 's', 'b', 0, 0032, 1, 's', 'p', 0, 0033, 1,
	'c', 'a', 0, 0034, 1, 'n', 'o', 0, 0035, 1, 'l', 'h', 0, 0036, 1,
	'm', 'o', 0, 0037, 1, 'a', 'a', 0, 0047, 1, 'g', 'r', 0, 0050, 1,
	'm', 'u', 0, 0052, 1, 'p', 'l', 0, 0053, 1, 'm', 'i', 0, 0055, 1,
	'd', 'i', 0, 0057, 1, '=', '=', 0, 0060, 1, '~', '=', 0, 0061, 1,
	'a', 'p', 0, 0062, 1, '!', '=', 0, 0063, 1, '<', '-', 0, 0064, 1,
	'-', '>', 0, 0065, 1, 'u', 'a', 0, 0066, 1, 'd', 'a', 0, 0067, 1,
	's', 'c', 0, 0070, 1, '*', '*', 0, 0071, 1, 'e', 'q', 0, 0075, 1,
	'*', 'A', 0, 0101, 1, '*', 'B', 0, 0102, 1, '*', 'G', 0, 0103, 1,
	'*', 'D', 0, 0104, 1, '*', 'E', 0, 0105, 1, '*', 'Z', 0, 0106, 1,
	'*', 'Y', 0, 0107, 1, '*', 'H', 0, 0110, 1, '*', 'I', 0, 0111, 1,
	'*', 'K', 0, 0112, 1, '*', 'L', 0, 0113, 1, '*', 'M', 0, 0114, 1,
	'*', 'N', 0, 0115, 1, '*', 'C', 0, 0116, 1, '*', 'O', 0, 0117, 1,
	'*', 'P', 0, 0120, 1, '*', 'R', 0, 0121, 1, '*', 'S', 0, 0122, 1,
	'*', 'T', 0, 0123, 1, '*', 'U', 0, 0124, 1, '*', 'F', 0, 0125, 1,
	'*', 'X', 0, 0126, 1, '*', 'Q', 0, 0127, 1, '*', 'W', 0, 0130, 1,
	'd', 'd', 0, 0131, 1, 'b', 'r', 0, 0132, 1, 'i', 'b', 0, 0133, 1,
	'c', 'i', 0, 0135, 1, 'g', 'a', 0, 0140, 1, '*', 'a', 0, 0141, 1,
	'*', 'b', 0, 0142, 1, '*', 'g', 0, 0143, 1, '*', 'd', 0, 0144, 1,
	'*', 'e', 0, 0145, 1, '*', 'z', 0, 0146, 1, '*', 'y', 0, 0147, 1,
	'*', 'h', 0, 0150, 1, '*', 'i', 0, 0151, 1, '*', 'k', 0, 0152, 1,
	'*', 'l', 0, 0153, 1, '*', 'm', 0, 0154, 1, '*', 'n', 0, 0155, 1,
	'*', 'c', 0, 0156, 1, '*', 'o', 0, 0157, 1, '*', 'p', 0, 0160, 1,
	'*', 'r', 0, 0161, 1, '*', 's', 0, 0162, 1, '*', 't', 0, 0163, 1,
	'*', 'u', 0, 0164, 1, '*', 'f', 0, 0165, 1, '*', 'x', 0, 0166, 1,
	'*', 'q', 0, 0167, 1, '*', 'w', 0, 0170, 1, 'p', 'd', 0, 0171, 1,
	'e', 's', 0, 0172, 1, 'o', 'r', 0, 0174, 1,   0,   0, 0,    0, 0
};

static	int	maptype;	/* IMAGE or GRAPHIC */
static	int	gicell;		/* =1 for cellular general image */
static	int	maporg;		/* origin code 0 - 8 from setorigin() */
static	int	ink;		/* ink density 0 - 255 from setink() */
static	int	majsize;	/* default point size of major font */
static	int	rotated;	/* =1 if major font is rotated */
static	int	psize;		/* current point size */
static	int	ptype;		/* current font type */
static	int	lastptype;	/* previous font type */
static	int	sheight;	/* height of string */
static	int	swidth;		/* width of string */
static	int	top2baseline;	/* height of string above baseline */
static	int	vertshift;	/* additional vertical shifts */
static	int	horizshift;	/* additional horizontal shifts */
static	int	halfsize;	/* value of halfline shifts in pixels */
static	char	oldname[FONTLEN];	/* previous font name */

static		label(), setmap(), marker();

extern	off_t	lseek();	/* C library */
/*extern	char	*malloc();*/ 	/* C library */
/*extern	char	*realloc();*/	/* C library */
extern	char	*index();	/* C library */
extern	char	*rindex();	/* C library */
/*extern	char	*strcpy();*/	/* C library */
/*extern	char	*strncpy();*/	/* C library */
/*extern	char	*strcat();*/	/* C library */
/*extern	char	*strncat();*/	/* C library */

	/*
	**	Read and format a fresh font
	*/
static	struct packed *
getfont(name, r)

	register char		*name;	/* name of font definition file */
	register struct packed	*r;	/* pointer to font storage area */
{
	register struct dispatch *dp;	/* font header pointer */
	unsigned	psize;		/* size of font table */
	int	 	size;		/* size of font data */
	register int	fid;		/* font file descriptor */
	register int	i;		/* scratch */
	register char 	*s;		/* scratch */
	char	buffer[FONTLEN];	/* scratch */
	/*
	**	First find the file, either in the current directory
	**	or in GIPSFONTDIR or (sun's only) in GIPSAUXFONTDIR
	*/
	if ((fid = open(name, 0)) == EOF && *name != '/') {
		s = getenv("GIPSFONTDIR");
		(void)sprintf(buffer, "%s/%s", s ? s : GIPSFONTDIR, name);
		if ((fid = open(buffer, 0)) != EOF)
			name = buffer;
	}
#ifdef sun
	if (fid == EOF && *name != '/') {
		s = getenv("GIPSAUXFONTDIR");
		sprintf(buffer, "%s/%s", s ? s : GIPSAUXFONTDIR, name);
		fid = open(name = buffer, 0);
	}
#endif
	if (fid == EOF)
		return NULL;
	else if (r != NULL && strcmp(r->fontname, name) == 0) {
		(void)close(fid);
		return r;
	}
	/*
	**	find font length in bytes, allocate a "packed" structure
	*/
	(void)lseek(fid, (off_t)0, 2);
	size = lseek(fid, (off_t)0, 1);
	psize = size + sizeof(struct packed) - sizeof(struct header)
		- 256*sizeof(struct dispatch);
	r = (struct packed *) (r ? realloc((char *)r, psize) : malloc(psize));
	if (r == NULL) {
		(void)close(fid);
		return NULL;
	}
	/*
	**	rewind and read the definition file
	*/
	(void)lseek(fid, (off_t)0, 0);
	if (read(fid, (char *)&r->h, size) != size ||
			r->h.magic != VFONT_MAGIC) {
		(void)close(fid);
		free((char *) r);
		return NULL;
	}
	(void)close(fid);
	/*
	**	if the font is rotated, switch up-down and
	**	left-right dimensions in header
	*/
	r->fontsize = r->fonttype = 0;
	if ((r->fontrot = name[strlen(name)-1]) != 'r')
		r->fontrot = 0;
	else for (dp = &r->ds[0]; dp < &r->ds[256]; dp++) {
		i = dp->left;
		dp->left = dp->down;
		dp->down = dp->right;
		dp->right = dp->up;
		dp->up = i;
	}
	/*
	**	Try hard to give ' ' a non-zero width
	*/
	for (s = " n0N"; *s; s++)
		if (r->ds[' '].width = r->ds[*s].width)
			break;
	/*
	**	Create font skeleton name, for size & style changes
	*/
	if (s = rindex(strcpy(r->fontname, name), '/')) 
		name = ++s;
	if ((s = name = rindex(strcpy(r->fontskel, name), '.')) == NULL)
		return r;
	if (index("rRiIbBS", *--s) != NULL)
		r->fonttype = *s;
	while (isdigit(*++name))
		r->fontsize = 10 * r->fontsize + *name - '0';
	if (r->fonttype != 0 && r->fontsize != 0)
		(void)sprintf(s, "%%c.%%d%c", r->fontrot);
	return r;
}
	/*
	**	return name of current font (if name=NULL) or change fonts
	*/
char *
setfont(name)

	register char *name;		/* name of font definition file */
{
	register i;			/* scratch */
	char	buffer[FONTLEN];	/* scratch */
	/*
	**	just return the pointer to the previous font name
	*/
	p = psave[0];
	if (name == NULL)
		return (p == NULL) ? NULL : p->fontname;
	else
		(void)strcpy(oldname, (p == NULL) ? "" : p->fontname);
	/*
	**	locate the desired font definition file
	*/
	if ((p = psave[ptype = 0] = getfont(name, p)) == NULL)
		return NULL;
	rotated = p->fontrot == 'r';
	halfsize = (p->ds['I'].up + p->ds['I'].down)/2;
	/*
	**	attempt to load italic, boldface, and special
	*/
	if ((majsize = p->fontsize) != 0) {
		char *ch = isupper(p->fonttype) ? "RIBS" : "ribS";
		for (i = 1; i < 4; i++) {
			sprintf(buffer, (i == 3) ? "%c.%d%c" : p->fontskel,
				ch[i], p->fontsize, p->fontrot);
			psave[i] = getfont(buffer, psave[i]);
		}
	}
	/*
	**	return the previous name
	*/
	return oldname;
}
	/*
	**	define special fonts
	*/
char *
setauxfont(name, i)

	register char *name;	/* name of font definition file */
	register i;		/* font number */
{
	register struct	packed *r;

	if (i < 0 || i > 3)
		return NULL;
	if (name != NULL) {
		r = psave[i] = getfont(name, psave[i]);
		if (i == 0) {
			rotated = r->fontrot == 'r';
			halfsize = (r->ds['I'].up + r->ds['I'].down)/2;
		}
	}
	return psave[i] ? psave[i]->fontname : NULL;
}
	/*
	**	determine the pixel value with which to fill
	**	subsequent images, return the previous density value
	*/
char
setink(c)

	register char c;	/* density at which to write into image plane */
{
	register char retval;

	retval = ink;
	ink = c;
	return retval;
}
	/*
	**	determine the location relative to the specified coordinates
	**	at which strings are to be inserted into images and graphics
	*/
setorigin(i)

	register i;		/* origin code 0 - 8 */
{
	register retval;

	retval = maporg;
	if (i >= 0 && i <= 8) maporg = i;
	return retval;
}
	/*
	**	set data type of subsequent structures - IMAGE or GRAPHIC
	**	i.e. 8-bit or 1-bit pixels
	*/
settype(i)

	register int i;
{
	register retval;

	retval = maptype;
	maptype = (i == IMAGE) ? IMAGE : GRAPHIC;
	return retval;
}
	/*
	**	Parse numeric shift fields
	*/
static char *
getscale(s, ip, in)

	register char *s;		/* scan pointer */
	register int *ip;		/* returned value */
	register int in;		/* default value */
{
	register char *t;		/* numeric scan */
	register char *u;		/* possible sign field */
	register n;			/* scratch */

	u = (s[1] == '\'') ? s+2 : s+1;
	t = (*u == '-' || *u == '+') ? u+1 : u;
	if (! isdigit(*t))
		return s;
	for (n = 0; isdigit(*t); n = 10*n + *t++ - '0')
		;
	if (s[1] == '\'')
		if (*t == '\'')
			t++;
		else
			return s;
	switch (*u) {
	case '+':
		*ip += n;
		break;
	case '-':
		*ip -= n;
		break;
	default:
		*ip = n ? n : in;
		break;
	}
	return t;
}
	/*
	**	decode the next text character and rearrange font pointer
	*/
static	struct dispatch *
getdisp(cp)

	register char **cp;		/* pointer to string pointer */
{
	register char	*c = *cp;	/* pointer to string */
	char	buffer[FONTLEN];	/* scratch */
	register ttype	= ptype;	/* temporary font type */
	register sw	= 0;		/* end escape search switch */
	register i, n;
	/*
	**	Handle in-line escape codes
	*/
	while (*c == '\\' && !sw)
		switch(*++c) {
		case 'f':
			switch(*++c) {
			case 'P':
				ptype = lastptype;
				break;
			case '0':
			case 'R':
				lastptype = ptype;
				ptype = 0;
				break;
			case '1':
			case 'I':
				lastptype = ptype;
				ptype = 1;
				break;
			case '2':
			case 'B':
				lastptype = ptype;
				ptype = 2;
				break;
			case '3':
			case 'S':
				lastptype = ptype;
				ptype = 3;
				break;
			default:
				c -= 2;
				break;
			}
			ttype = ptype;
			c++;
			break;
		case 's':
			c = getscale(c, &psize, majsize);
			break;
		case 'h':
			c = getscale(c, &horizshift, 0);
			break;
		case 'v':
			c = getscale(c, &vertshift, 0);
			break;
		case '(': {
			register struct spectab *s;
			for (s = spectab; s->key[0]; s++)
				if (c[1] == s->key[0] && c[2] == s->key[1]) {
					c += 2;
					i = s->val & 0377;
					ttype = s->flg ? 3 : ptype;
					sw++;
					break;
				}
			}
			break;
		case 'd':
			vertshift -= halfsize;
			c++;
			break;
		case 'u':
			vertshift += halfsize;
			c++;
			break;
		case '\'':
			i = 0140;
			ttype = 3;
			sw++;
			break;
		case 'e':
			i = 0134;
			ttype = 3;
			sw++;
			break;
		default:
			i = n = 0;
			while (n++ < 3 && *c >= '0' && *c <= '7')
				i = (i << 3) + *c++ - '0';
			if (i)
				--c, sw++;
			break;
		}
	/*
	**	If c -> output character, see if symbol font must be used
	*/
	if (! sw)
		if (! (i = *c & 0377))
			return NULL;
		else if (index("\"#<=>@^{}~", i))
			ttype = 3;
	/*
	**	If the desired font isn't loaded, go get it
	*/
	p = psave[ttype] ? psave[ttype] : psave[0];
	if (psize != p->fontsize) {
		(void)sprintf(buffer, p->fontskel,
			p->fonttype, psize, p->fontrot);
		if (p = getfont(buffer, p))
			psave[ttype] = p;
		else
			p = psave[ttype] ? psave[ttype] : psave[0];
	}
	/*
	**	Update the input string pointer and return the dispatch
	*/
	*cp = c+1;
	return p ? &p->ds[i] : NULL;
}
	/*
	**	write string 'cp' into graphic or image file 'fp'
	**	(nlines x ncols) baseline, basecol denote
	**	coordinates of top left of string
	*/
comlabel(fildes, nlines, ncols, baseline, basecol, cp)

	int	fildes;		/* image or graphic file number */
	int	nlines;		/* y-dimension of array */
	int	ncols;		/* x-dimension of array */
	int	baseline;	/* y-coordinate of top of string */
	int	basecol;	/* x-coordinate of left of string */
	char	*cp;		/* string to put into file */
{
	return marker(rotated ? &baseline : (int *)NULL,
			rotated ? (int *)NULL : &basecol, cp) ||
		label(fildes, nlines, ncols, baseline, basecol, cp);
}
	/*
	**	write string "cp" into image or graphic file "fildes".
	**	centerline/centercol determine the string location,
	**	according to last value passed to setorigin()
	*/
commarker(fildes, nlines, ncols, centerline, centercol, cp)

	int	fildes;		/* image or graphic file number */
	int	nlines;		/* y-dimension of array */
	int	ncols;		/* x-dimension of array */
	int	centerline;	/* y-coordinate of origin of string */
	int	centercol;	/* x-coordinate of origin of string */
	char	*cp;		/* string to put into file */
{
	return marker(&centerline, &centercol, cp) ||
		label(fildes, nlines, ncols, centerline, centercol, cp);
}
static
label(fildes, nlines, ncols, baseline, basecol, cp)

	int	fildes;		/* image or graphic file number */
	int	nlines;		/* y-dimension of array */
	int	ncols;		/* x-dimension of array */
	int	baseline;	/* y-coordinate of top of string */
	int	basecol;	/* x-coordinate of left of string */
	char	*cp;		/* string to put into file */
{
	register char *mp;	/* image or graphic buffer */
	register nsl;		/* height of character string in pixels */
	unsigned size;		/* size of buffer required */
	int	first2read;	/* first y-coordinate to update */
	int	last2read;	/* last y-coordinate to update */
	int	offset;		/* offset in file to first2read */
	int	ret;		/* return code */
	int	nread;		/* number of y-rasters to read */
	/*
	**	verify callers arguments and get height of font in y-pixels
	*/
	if (nlines <= 0 || ncols <= 0)
		return 0;
	else if ((nsl = rotated ? swidth : sheight) == 0)
		return 1;
	/*
	**	determine the size of incore buffer required to hold the string
	*/
	first2read = rotated ? baseline : baseline-top2baseline;
	if ((last2read = first2read+nsl-1) >= nlines)
		last2read = nlines-1;
	if (first2read < 0)
		first2read = 0;
	nread = (last2read-first2read)+2;
	if (nread <= 0 || first2read >= nlines || last2read < 0)
		return 1;
	size = nread*ncols;
	offset = first2read*ncols;
	/*
	**	adjust coordinates for graphic files which are 8 times smaller
	*/
	if (maptype == GRAPHIC) {
		offset = ncols*((first2read &= ~7) >> 3);
		if ((size = (size+7) >> 3)+offset > (nlines*ncols+7) >> 3)
			size--;
	}
	/*
	**	allocate the buffer, read the file into it, call memlabel() to
	**	fill it with the unpacked characters, and write the buffer again
	*/
	if ((mp = malloc(size)) == NULL)
		return 0;
	ret = (lseek(fildes, (off_t)offset, 0) != EOF &&
		read(fildes, mp, (int) size) == size &&
		memory(mp, nread, ncols, baseline-first2read, basecol, cp) &&
		lseek(fildes, (off_t)offset, 0) != EOF &&
		write(fildes, mp, (int)size) == size);
	free(mp);
	return ret;
}
	/*
	**	write string "cp" into incore image or graphic array "addr".
	**	centerline/centercol determine the string location,
	**	according to last value passed to setorigin()
	*/
memmarker(addr, nlines, ncols, centerline, centercol, cp)

	char	*addr;		/* pointer to incore image or graphic */
	int	nlines;		/* y-dimension of array */
	int	ncols;		/* x-dimension of array */
	int	centerline;	/* y-coordinate of origin of string */
	int	centercol;	/* x-coordinate of origin of string */
	char	*cp;		/* string to put into array */
{
	return marker(&centerline, &centercol, cp) ||
		memory(addr, nlines, ncols, centerline, centercol, cp);
}
	/*
	**	Write string into array - top-left origin coordinates
	*/
memlabel(array, nlines, ncols, baseline, basecol, cp)

	char	*array;		/* pointer to incore image or graphic */
	int	nlines;		/* y-dimension of array */
	int	ncols;		/* x-dimension of array */
	int	baseline;	/* y-coordinate of top of string */
	int	basecol;	/* x-coordinate of left of string */
	char	*cp;		/* string to put into file */
{
	return marker(rotated ? &baseline : (int *)NULL,
			rotated ? (int *)NULL : &basecol, cp) ||
		memory(array, nlines, ncols, baseline, basecol, cp);
}
	/*
	**	write string "cp" into cell-format general image "gim".
	**	centerline/centercol determine the string location,
	**	according to last value passed to setorigin()
	*/
gimarker(gim, centerline, centercol, cp)

	gimage	*gim;		/* pointer to cell image structure */
	int	centerline;	/* y-coordinate of origin of string */
	int	centercol;	/* x-coordinate of origin of string */
	char	*cp;		/* string to put into array */
{
	int	rc;

	gicell = 1;
	rc = marker(&centerline, &centercol, cp) ||
		memory((char *)gim, 1<<24, 65536, centerline, centercol, cp);
	gicell = 0;
	return rc;
}
	/*
	**	Write string into image - top-left origin coordinates
	*/
gilabel(gim, baseline, basecol, cp)

	gimage	*gim;		/* pointer to cell image structure */
	int	baseline;	/* y-coordinate of top of string */
	int	basecol;	/* x-coordinate of left of string */
	char	*cp;		/* string to put into file */
{
	int	rc;

	gicell = 1;
	rc = marker(rotated ? &baseline : (int *)NULL,
			rotated ? (int *)NULL : &basecol, cp) ||
		memory((char *)gim, 1<<24, 65536, baseline, basecol, cp);
	gicell = 0;
	return rc;
}
	/*
	**	Write string into array
	*/
memory(array, nl, nc, line, col, cp)

	char	*array;		/* pointer to incore image or graphic */
	int	nl;		/* y-dimension of incore array */
	int	nc;		/* x-dimension of incore array */
	int	line;		/* y-coordinate of top of string */
	int	col;		/* x-coordinate of left of string */
	char	*cp;		/* string to put into array */
{
	register struct dispatch *dp;	/* pointer to font entry */
	register l;		/* scratch */
	register char * byte;	/* pointer to font definition entry */
	int	line2;		/* upper y-limit of byte */
	int	col1;		/* lower x-limit of byte */
	int	col2;		/* upper x-limit of byte */
	int	offset;		/* offset in array to start of byte */
	int	nb;		/* number of font definition bytes */
	/*
	**	verify callers arguments
	*/
	if (nl < 0 || nc < 0 || array == NULL)
		return 0;
	/*
	**	handle the input string byte-by-byte
	*/
	while ((dp = getdisp(&cp)) != NULL) {
		if (dp->nbytes != 0) {
			/*
			**	dp -> entry in font definition file
			*/
			l = col + (rotated ? vertshift : horizshift);
			col1 = l - dp->left;
			col2 = l + dp->right;
			l = line + (rotated ? horizshift : -vertshift);
			line2 = l + dp->down;
			if (line2 >= nl) line2 = nl-1;
			if (col2 >= nc) col2 = nc-1;
			nb = ((dp->left + dp->right) + 7)/8;
			byte = p->maps + dp->addr;
			l -= dp->up;
			if (l < 0) {
				offset = col1;
				byte -= l*nb;
				l = 0;
			} else
				offset = col1 + l*nc;
			/*
			**	loop over y-rasters 
			*/
			while (l++ < line2) {
				setmap(array, byte, col1, col2, offset, nc);
				byte += nb;
				offset += nc;
			}
		}
		/*
		**	advance to next byte in "cp"
		*/
		if (rotated)
			line += dp->width;
		else
			col += dp->width;
	}
	return 1;
}
	/*
	**	fill array (image or graphic) starting at array+offset
	**	from col to col2 according to bit map
	*/
static
setmap(array, map, col, col2, offset, nc)

	register char *array;	/* image or graphic array */
	register char	*map;	/* font bit map */
	register col;		/* starting column */
	int	col2;		/* ending column */
	register offset;	/* starting offset in array */
	int	nc;		/* number of columns */
{
	register i;		/* column loop index */
	register char *c;	/* scratch */

	for (i = 0; col < col2; offset++, i++)
		if (col++ >= 0 && (map[i >> 3] & 0200 >> (i & 7)))
			if (gicell) {
				if (c = gcell(offset & 0xffff,
						offset >> 16, array))
					*c = ink;
			} else if (maptype == IMAGE) {
				array[offset] = ink;
			} else if (ink) {
#ifdef IGED
				extern int debug;
				extern FILE *fio;
				if (debug & 2 && fio != NULL) {
					extern FILE *fio;
					register x = offset % nc;
					register y = offset / nc;
					(void)fprintf(fio,
						"%d %d %d %d ", x, y, x, y);
				}
#endif
				array[GRAPHADDR(offset>>3)] |= 0200>>(offset&7);
			} else
				array[GRAPHADDR(offset>>3)] &= ~(0200>>(offset&7));
}
	/*
	**	get maximum height of string (in pixels) from font file header
	*/
comheight(cp)

	char	*cp;			/* string to encode */
{
	(void)marker(NULL, NULL, cp);
	return sheight;
}
	/*
	**	get maximum length of string (in pixels) from font file header
	*/
comwidth(cp)

	char	*cp;		/* string to encode */
{
	(void)marker(NULL, NULL, cp);
	return swidth;
}
	/*
	**	tables of shift coefficients
	*/
#define		YT	0
#define		YC	010
#define		YB	04
#define		XL	0
#define		XC	02
#define		XR	01

static	int	orgtab[9] = {
 /*	0,	1	2	3	4	5	6	7	8    */
	XC|YC,	XL|YT,	XC|YT,	XR|YT,	XL|YC,	XR|YC,	XL|YB,	XC|YB,	XR|YB
};
	/*
	**	adjust center line and column according to setorigin() code
	*/
static
marker(lptr, cptr, cp)

	int	*lptr;			/* pointer to line origin */
	int	*cptr;			/* pointer to column origin */
	char	*cp;			/* string to be coded */
{
	register ht;			/* height of text */
	register wd;			/* width of text */
	register flg;			/* origin type flag */
	register struct dispatch *dp;	/* pointer to font buffer */

	ptype = lastptype = vertshift = horizshift = 0;
	sheight = swidth = top2baseline = 0;
	psize = majsize;
	if (psave[0] == NULL || cp == NULL)
		return 1;
	while ((dp = getdisp(&cp)) != NULL) {
		swidth += dp->width;
		if (dp->nbytes) {
			if (dp->down-vertshift > sheight)
				sheight = dp->down-vertshift;
			if (dp->up+vertshift > top2baseline)
				top2baseline = dp->up+vertshift;
		}
	}
	swidth += horizshift;
	ptype = lastptype = vertshift = horizshift = 0;
	psize = majsize;
	sheight += top2baseline;
	if (rotated)
		sheight += sheight/2;
	if ((flg = orgtab[maporg]) & YC)
		ht = rotated ? sheight/2 : top2baseline/2;
	else if (flg & YB)
		ht = rotated ? 0 : top2baseline-sheight;
	else
		ht = rotated ? sheight : top2baseline;

	if (flg & XC)
		wd = swidth/2;
	else if (flg & XR)
		wd = swidth;
	else
		wd = 0;

	if (lptr != NULL)
		*lptr += rotated ? -wd : ht;
	if (cptr != NULL)
		*cptr -= rotated ? ht : wd;
	return 0;
}
