#ifndef lint
static char *sccsid = "%Z%$RCSfile: conformal.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1993/05/26 15:58:16 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		conformal.c	Creation Date:	06/29/87
**	Version:	$Revision: 2.7 $		Update:		$Date: 1993/05/26 15:58:16 $
**	Components:	lambert1()	ihgrid() function
**			lambert2()	""
**			mercator()	""
**
**	Function:	Transform an image via conformal mappings
*/

#define		GRID
#include	<math.h>
#include	"gips.h"

lambert1(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		if (x == GIPSLOW)
			return GIPSBADVAL;
		v = fabs(DEG_RAD * x);
		t->tphi2 = tan(M_PI_4 - 0.5 * v);
		t->sphi = sin(v);
		if (errno || t->tphi2 == 0 || t->sphi == 0) return 1;
		t->radius = xy[0];
		t->csphi = 1 / t->sphi;
		t->rcphi = t->radius * cos(v) * t->csphi;
		if (t->rcphi == 0) return 1;
		if (x < 0) t->rcphi = -t->rcphi;
		return 0;
	case 1:
		vv = tan(M_PI_4 - 0.5 * fabs(y));
		if (errno) return 1;
		v = vv / t->tphi2;
		if (v < 0) return 1;
		vv = fabs(t->rcphi) * pow(v, t->sphi);
		if (errno) return 1;
		v = x * t->sphi;
		xy[0] = vv * sin(v);
		if (t->rcphi < 0) vv = -vv;
		xy[1] = t->rcphi - vv * cos(v);
		return 0;
	case 2:
		v = t->rcphi - y;
		if (t->rcphi < 0) v = -v;
		if (x == 0 && v == 0) return 1;
		xy[0] = t->csphi * atan2(x, v);
		v = hypot(v, x) / fabs(t->rcphi);
		if (v < 0) return 1;
		xy[1] = M_PI_2 - 2 * atan(t->tphi2 * pow(v, t->csphi));
		if (errno) return 1;
		if (t->rcphi < 0) xy[1] = -xy[1];
		return 0;
	default:
		break;
	}
	return -1;
}

lambert2(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		if (x == GIPSLOW || y == GIPSLOW || x == y || (x > 0 && y < 0))
			return GIPSBADVAL;
		v = fabs(DEG_RAD * y);
		vv = fabs(DEG_RAD * x);
		t->radius = xy[0];
		t->fact0 = log(t->radius * cos(v));
		t->fact1 = log(t->radius * cos(vv));
		v = tan(M_PI_4 - 0.5 * v);
		vv = tan(M_PI_4 - 0.5 * vv);
		if (errno || v <= 0 || vv <= 0) return 1;
		vv = log(v) - log(vv);
		if (vv == 0) return 1;
		t->sphi0 = (t->fact0 - t->fact1) / vv;
		if (t->sphi0 == 0) return 1;
		vv = 1 - t->sphi0 * t->sphi0;
		if (vv < 0) return 1;
		t->csphi0 = 1 / t->sphi0;
		t->rcphi0 = t->radius * sqrt(vv) * t->csphi0;
		if (x < 0 || y < 0) t->rcphi0 = -t->rcphi0;
		v = pow(v, t->sphi0);
		if (errno) return 1;
		v *= t->sphi0;
		t->fact0 = t->radius * cos(DEG_RAD * y);
		if (v == 0 || t->fact0 == 0) return 1;
		t->fact0 = v / t->fact0;
		return 0;
	case 1:
		vv = tan(M_PI_4 - 0.5 * fabs(y));
		if (errno || vv < 0) return 1;
		vv = pow(vv, t->sphi0) / t->fact0;
		if (errno) return 1;
		v = x * t->sphi0;
		xy[0] = vv * sin(v);
		if (t->rcphi0 < 0) vv = -vv;
		xy[1] = t->rcphi0 - vv * cos(v);
		return 0;
	case 2:
		v = t->rcphi0 - y;
		if (t->rcphi0 < 0) v = -v;
		if (x == 0 && v == 0) return 1;
		xy[0] = t->csphi0 * atan2(x, v);
		v = t->fact0 * hypot(v, x);
		if (v < 0) return 1;
		xy[1] = M_PI_2 - 2 * atan(pow(v, t->csphi0));
		if (errno) return 1;
		if (t->rcphi0 < 0) xy[1] = -xy[1];
		return 0;
	default:
		break;
	}
	return -1;
}

mercator(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = log(GIPSHIGH);
		return 0;
	case 1:
		v = tan(M_PI_4 + 0.5 * y);
		if (errno || v <= 0) return 1;
		xy[0] = t->radius * x;
		xy[1] = t->radius * log(v);
		return 0;
	case 2:
		xy[0] = x / t->radius;
		v = y / t->radius;
		if (v < -t->rlim || v > t->rlim) return 1;
		xy[1] = 2 * atan(exp(v)) - M_PI_2;
		return 0;
	default:
		break;
	}
	return -1;
}
