#ifndef lint
static char *sccsid = "%Z%$RCSfile: equivalent.c,v $	GIPS $Revision: 2.8 $ (MIT/CSR) $Date: 1990/11/25 20:22:38 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		equivalent.c	Creation Date:	06/29/87
**	Version:	$Revision: 2.8 $		Update:		$Date: 1990/11/25 20:22:38 $
**	Components:	albers1()	ihgrid() function
**			albers2()		""
**			bonne()		""
**			sinusoidal()	""
**			werner()	""
**
**	Function:	Transform an image via equivalent mappings
*/

#define		GRID
#include	<math.h>
#include	"gips.h"

albers1(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		if (x == GIPSLOW)
			return GIPSBADVAL;
		t->cphi = cos(DEG_RAD * x);
		t->sphi = sin(DEG_RAD * x);
		if (t->sphi == 0) return 1;
		t->rctphi = xy[0] / t->sphi;
		return 0;
	case 1:
		v = x * t->sphi;
		vv = 1 + t->sphi * (t->sphi - 2 * sin(y));
		if (vv < 0) return 1;
		vv = sqrt(vv);
		xy[0] = sin(v) * vv * t->rctphi;
		xy[1] = cos(v) * (t->cphi - vv) * t->rctphi;
		return 0;
	case 2:
		break;
	default:
		break;
	}
	return -1;
}

albers2(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		if (x == GIPSLOW || y == GIPSLOW || x == y)
			return GIPSBADVAL;
		t->sphi0 = sin(DEG_RAD * x);
		t->fact0 = 0.5 * (t->sphi0 + sin(y * DEG_RAD));
		if (t->fact0 == 0) return 1;
		t->fact1 = 2 * xy[0] * xy[0] / t->fact0;
		if (t->fact1 == 0) return 1;
		t->fact2 = xy[0] * cos(DEG_RAD * x) / t->fact0;
		return 0;
	case 1:
		v = t->fact2 * t->fact2 + (t->sphi0 - sin(y)) * t->fact1;
		if (v < 0) return 1;
		v = sqrt(v);
		x *= t->fact0;
		xy[0] = v * sin(x);
		xy[1] = -v * cos(x);
		return 0;
	case 2:
		if (x == 0 && y == 0) return 1;
		xy[0] = atan2(x, -y) / t->fact0;
		v = t->sphi0 + (t->fact2 * t->fact2 - x * x - y * y) / t->fact1;
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

bonne(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv;

	switch (sw) {
	case 0:
		if (x == GIPSLOW)
			return GIPSBADVAL;
		v = DEG_RAD * x;
		t->radius = xy[0];
		vv = tan(v);
		if (errno || vv == 0) return 1;
		t->rctphi = t->radius / vv;
		t->fact0 = t->rctphi + t->radius * v;
		return 0;
	case 1:
		v = t->fact0 - t->radius * y;
		if (v == 0) return 1;
		vv = t->radius * cos(y) * x / v;
		xy[0] = v * sin(vv);
		xy[1] = t->rctphi - v * cos(vv);
		return 0;
	case 2:
		v = hypot(t->rctphi - y, x);
		xy[1] = (t->fact0 - v) / t->radius;
		vv = cos(xy[1]) * t->radius;
		if (vv == 0 || (x == 0 && t->rctphi == y)) return 1;
		xy[0] = v * atan2(x, t->rctphi - y) / vv;
		return 0;
	default:
		break;
	}
	return -1;
}

sinusoidal(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		xy[0] = t->radius * x * cos(y);
		xy[1] = t->radius * y;
		return 0;
	case 2:
		xy[1] = y / t->radius;
		v = t->radius * cos(xy[1]);
		if (v == 0) return 1;
		xy[0] = x / v;
		return 0;
	default:
		break;
	}
	return -1;
}

werner(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v, vv, vvv;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		v = M_PI_2 - y;
		if (v == 0) return 1;
		vv = cos(y) * x / v;
		xy[0] = t->radius * v * sin(vv);
		xy[1] = -t->radius * v * cos(vv);
		return 0;
	case 2:
		if (y > 0) return 1;
		xy[1] = M_PI_2 - hypot(x, y) / t->radius;
		vv = cos(xy[1]);
		v = M_PI_2 - xy[1];
		if (vv == 0 || v == 0) return 1;
		vvv = x / (t->radius * v);
		if (vvv > 1 || vvv < -1) return 1;
		xy[0] = v * asin(vvv) / vv;
		return 0;
	default:
		break;
	}
	return -1;
}
