#ifndef lint
static char *sccsid = "%Z%$RCSfile: gcell.c,v $	GIPS $Revision: 2.15 $ (MIT/CSR) $Date: 2004/11/29 18:51:20 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984, 1990 Massachusetts Institute of Technology
**
**	Module:		gcell.c		Creation Date:	04/22/84
**	Version:	$Revision: 2.15 $		Update:		$Date: 2004/11/29 18:51:20 $
**
**	Components:	gopen()		Access general image in cell format
**			gcell()		Return pointer to pixel
**			ginterp()	Return pointer to interpolation
**			gclose()	Close and release image
**			gwrite()	Mark image for update
**
**	Function:	Open and access general image in cell order
*/

#include	"gips.h"
#include	<strings.h>
#define		NUMRING		8
#define		GI_MAGIC	04321
#define		GI_MERGE	0x80
#define		GI_WRITE	0x40

#ifdef BIDR
#define		BIDR_MAGIC	0x7007
extern	gimage	*bidr_open();	/* Open a Magellan BIDR file */
extern	char	*bidr_cell();	/* Point to a BIDR raster */
extern	char	*bidr_interp(); /* Interpolate BIDR pixel value */
#endif BIDR

	/*
	**	Control table for cell images
	*/
struct	gi	{
	short	gi_magic;	/* the magic word */
	short	gi_nrings;	/* number of ring buffers */
	short	gi_shft;	/* shift power of block */
	short	gi_xshft;	/* log2 number of x pixels in cell */
	short	gi_yshft;	/* log2 number of y pixels in cell */
	short	gi_dshft;	/* log2 pixel length in bytes */
	short	gi_cvt;		/* input pixel type */
	char	*gi_buf;	/* ring-buffer pointer */
	char	*gi_ptr;	/* previous buffer pointer value */
	char	gi_flag;	/* flags */
	off_t	gi_off;		/* offset of first data block */
	double	gi_dnull;	/* null variable value */
	float	gi_fnull;	/* null variable value */
	int	gi_fid;		/* input image file descriptor */
	int	gi_xnum;	/* input x-direction pixel length */
	int	gi_ynum;	/* input y-direction pixel length */
	int	gi_xmask;	/* mask for x addressing within cell */
	int	gi_ymask;	/* mask for y addressing within cell */
	int	gi_block;	/* ring buffer record size */
	int	gi_xcell;	/* number of x-cells */
	int	gi_xsize;	/* number of bytes in cell x-raster */
	int	gi_last;	/* last record read into buf */
	int	gi_next;	/* next free portion of buf */
	int	gi_nrec[1];	/* ring-buffer record numbers */
};
	int	_gcount;	/* number of pixels pointed to by gcell() */
extern	off_t	lseek();	/* C library */
static		shift();
	/*
	**	Create control table from open file descriptor
	*/
gimage *
gopen(fp, nring)

	FILE	*fp;		/* pointer to General Image */
	int	nring;		/* number of ring buffers to allocate */
{
	register struct gi	*gi;		/* pointer to file table */
	register		i;		/* scratch */
	double			d;		/* scratch */
	long			off;		/* header offset */
	char			dtype;		/* "dtype" value */
	char			gbuf[BUFLEN];	/* "geometry" value */
	/*
	**	Read and test input file
	*/
	if (fp == NULL) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	/*
	**	Check for bidr header
	*/
	if (ihget("dtype", &dtype, 1) || dtype != 'G' ||
	    ihget("geometry", gbuf, sizeof(gbuf)))
		return NULL;
#ifdef BIDR
	if (strncmp(gbuf, "type=b bidr=", 12) == 0) {
		(void)fclose(fp);
		return bidr_open(gbuf);
	}
#endif BIDR
	if ((off = ftell(fp)) == -1) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	/*
	**	Allocate gimage control block
	*/
	if (nring < 1) nring = NUMRING;
	gi = (struct gi *) alloc(sizeof(struct gi) + nring*sizeof(int), 1);
	if (gi == NULL)
		return NULL;
	gi->gi_fid = fileno(fp);
	gi->gi_nrings = nring;
	/*
	**	Check other header fields
	*/
	if (ihget("xnum", (char *)&gi->gi_xnum, 1) ||
	    ihget("ynum", (char *)&gi->gi_ynum, 1) ||
	    ihget("geometry", gbuf, sizeof(gbuf)))
		return NULL;

	if (sscanf(gbuf, "type=%1s block=%d xdim=%d ydim=%d",
	    &dtype, &gi->gi_block, &gi->gi_xmask, &gi->gi_ymask) != 4)
		return NULL;
	if (dtype == 'M') {
		gi->gi_flag |= GI_MERGE;
		dtype = 'c';
	}
	/*
	**	convert dnull to appropriate pixel type
	*/
	gi->gi_xsize = ihdtype(dtype) * gi->gi_xmask;
	gi->gi_cvt = ihdcvt(dtype);
	if (ihget("dnull", (char *)&d, 1) == 0) {
		gconv((char *)&gi->gi_dnull, (char *)&d, 1, gi->gi_cvt, CVT_D);
		gi->gi_fnull = d;
	}
	gi->gi_off = gi->gi_block*((off+gi->gi_block-1)/gi->gi_block);
	/*
	**	determine the cell size
	*/
	gi->gi_shft = shift(gi->gi_block);
	gi->gi_xshft = shift(gi->gi_xmask--);
	gi->gi_yshft = shift(gi->gi_ymask--);
	gi->gi_dshft = shift(ihdtype(dtype));
	if (gi->gi_shft - gi->gi_xshft - gi->gi_yshft - gi->gi_dshft)
		return NULL;
	/*
	**	allocate the ring buffer
	*/
	if ((gi->gi_buf = alloc(gi->gi_nrings, gi->gi_block)) == NULL)
		return NULL;
	for (i = 0; i < gi->gi_nrings; i++)
		gi->gi_nrec[i] = -1;
	gi->gi_last = -1;
	gi->gi_magic = GI_MAGIC;
	gi->gi_xcell = (gi->gi_xnum + gi->gi_xmask) >> gi->gi_xshft;
	return (gimage *)gi;
}
	/*
	**	Return pointer to a pixel
	*/
char	*
gcell(x, y, gim)

	register int	x;		/* x-pixel address */
	register int	y;		/* y-pixel address */
	gimage		*gim;		/* pointer to gimage table */
{
	register struct gi *gi = (struct gi *) gim;
	register int	i;		/* ring buffer number */
	register float	*f;		/* pointer to cell */
	register long	off;		/* offset to input block */
	/*
	**	Test valid gimage
	*/
	if (gi == NULL)
		return NULL;
#ifdef BIDR
	if (gi->gi_magic == BIDR_MAGIC)
		return bidr_cell(x, y, gim);
#endif BIDR
	if (gi->gi_magic != GI_MAGIC)
		return NULL;
	/*
	**	Test if (x,y) within image
	*/
	if (y < 0 || y >= gi->gi_ynum || x >= gi->gi_xnum) {
		_gcount = -gi->gi_xnum;
		return (char *)&gi->gi_dnull;
	}
	if (x < 0) {
		_gcount = x;
		return (char *)&gi->gi_dnull;
	}
	/*
	**	Get buffer number and offset
	**	Store in _gcount the number of pixels remaining
	*/
	i = (x >> gi->gi_xshft) + gi->gi_xcell * (y >> gi->gi_yshft);
	y = ((x &= gi->gi_xmask) +
		((y & gi->gi_ymask) << gi->gi_xshft)) << gi->gi_dshft;
	_gcount = (gi->gi_flag & GI_MERGE) ? 1 : gi->gi_xmask + 1 - x;
	/*
	**	If same buffer as last time, we've got it
	*/
	if (i == gi->gi_last) {
		if ((gi->gi_flag & (GI_MERGE|GI_WRITE)) == GI_MERGE) {
			f = (float *)(gi->gi_ptr + y);
			if (f[1])
				f[0] /= f[1], f[1] = 0;
			if (f[0] == 0)
				return (char *)&gi->gi_dnull;
		}
		return gi->gi_ptr + y;
	}
	/*
	**	Go see whether buffer is in core.
	*/
	gi->gi_last = i;
	for (x = 0; x < gi->gi_nrings; x++)
		if (i == gi->gi_nrec[x]) {
			gi->gi_ptr = gi->gi_buf + (x << gi->gi_shft);
			if ((gi->gi_flag & (GI_MERGE|GI_WRITE)) == GI_MERGE) {
				f = (float *)(gi->gi_ptr + y);
				if (f[1])
					f[0] /= f[1], f[1] = 0;
				if (f[0] == 0)
					return (char *)&gi->gi_dnull;
			}
			return gi->gi_ptr + y;
		}
	/*
	**	If not there, locate next ring buffer number
	*/
	x = (gi->gi_next++) % gi->gi_nrings;
	gi->gi_ptr = gi->gi_buf + (x << gi->gi_shft);
	/*
	**	If block is to be updated, write out the old one
	*/
	if ((gi->gi_flag & GI_WRITE) && gi->gi_nrec[x] >= 0)
		if (lseek(gi->gi_fid,
		(off_t)(gi->gi_off+(gi->gi_nrec[x] << gi->gi_shft)), 0) == -1
	    || write(gi->gi_fid, gi->gi_ptr, gi->gi_block) != gi->gi_block)
		gipserrno = GIPSBADWRITE;
	/*
	**	Read in the new block
	*/
	gi->gi_nrec[x] = i;
	if (lseek(gi->gi_fid, (off_t)(gi->gi_off+(i << gi->gi_shft)), 0) == -1) {
		gipserrno = GIPSBADREAD;
		gi->gi_magic = 0;
		return NULL;
	}
	if ((i = read(gi->gi_fid, gi->gi_ptr, gi->gi_block)) < 0) {
		gipserrno = GIPSBADREAD;
		gi->gi_magic = 0;
		return NULL;
	}
	if (i != gi->gi_block)
		clear(gi->gi_ptr + i, gi->gi_block - i);
	if ((gi->gi_flag & (GI_MERGE|GI_WRITE)) == GI_MERGE) {
		f = (float *)(gi->gi_ptr + y);
		if (f[1])
			f[0] /= f[1], f[1] = 0;
		if (f[0] == 0)
			return (char *)&gi->gi_dnull;
	}
	return gi->gi_ptr + y;
}
	/*
	**	Return pointer to cell and mark it writeable
	*/
gwrite(gim, fp)
	gimage	*gim;
	FILE	*fp;
{
	if (((struct gi *)gim)->gi_magic == GI_MAGIC) {
		((struct gi *)gim)->gi_flag |= GI_WRITE;
		((struct gi *)gim)->gi_fid = fileno(fp);
	}
}
	/*
	**	Interpolate cell value
	*/
#ifdef	TEST
char *
ginterp(vx, vy, gim)
	float	vx, vy;
	gimage	*gim;
{
	register struct gi *gi = (struct gi *)gim;
	register	x = vx, y = vy;
	register	i, nil = *((int *)&gi->gi_fnull);

	float	vxvy;
	static	union	{
		double	d[1];
		float	f[2];
		long	l[2];
		char	s[8];
	} u[4];

#ifdef BIDR
	if (gi->gi_magic == BIDR_MAGIC)
		return bidr_interp(vx, vy, gim);
#endif BIDR
	if (gi->gi_magic != GI_MAGIC)
		return NULL;

#define COPY(to,from,n) \
	if (gi->gi_cvt==CVT_C || gi->gi_cvt==CVT_F) \
		bcopy((from),(to),(n)<<(gi->gi_dshft)); \
	else \
		gconv((to),(from),(n),CVT_F,gi->gi_cvt)

	if (gi->gi_flag & GI_MERGE) {
		register float *pf;
#define	FCOPY(x,y) (pf = (float *)gcell((x),(y),gi)) ? *pf : gi->gi_fnull
		u[0].f[0] = FCOPY(x,y);
		u[1].f[0] = FCOPY(x+1,y);
		u[2].f[0] = FCOPY(x,y+1);
		u[3].f[0] = FCOPY(x+1,y+1);
	} else
	if ((x & gi->gi_xmask) != gi->gi_xmask) {
		register char *c;
		c = gcell(x,y,gi);
		COPY(u[0].s, c, 2);
		c = (u[0].l[0] != nil && (y & gi->gi_ymask) != gi->gi_ymask) ?
			c + gi->gi_xsize : gcell(x,y+1,gi);
		COPY(u[2].s, c, 2);
		if (gi->gi_cvt != CVT_C) {
			u[1].f[0] = u[0].f[1];
			u[3].f[0] = u[2].f[1];
		}
	} else {
		COPY(u[0].s, gcell(x,y,gi), 1);
		COPY(u[1].s, gcell(x+1,y,gi), 1);
		COPY(u[2].s, gcell(x,y+1,gi), 1);
		COPY(u[3].s, gcell(x+1,y+1,gi), 1);
	}
	vx -= x;
	vy -= y;
	vxvy = vx*vy;
	for (i = 0; u[i].l[0] == nil; i++)
		if (i == 3)
			return (char *)&gi->gi_dnull;
	u[0].f[0] =
		 (1-vx-vy+vxvy) * (u[0].l[0] == nil ? u[i].f[0] : u[0].f[0])
		+ (vx-vxvy)	* (u[1].l[0] == nil ? u[i].f[0] : u[1].f[0])
		+ (vy-vxvy)	* (u[2].l[0] == nil ? u[i].f[0] : u[2].f[0])
		+ vxvy		* (u[3].l[0] == nil ? u[i].f[0] : u[3].f[0]);
	if (gi->gi_cvt != CVT_C) {
		gconv(u[1].s, u[0].s, 1, gi->gi_cvt, CVT_F);
		return u[1].s;
	}
	u[0].f[1] =
		 (1-vx-vy+vxvy) * (u[0].l[1] == nil ? u[i].f[1] : u[0].f[1])
		+ (vx-vxvy)	* (u[1].l[1] == nil ? u[i].f[1] : u[1].f[1])
		+ (vy-vxvy)	* (u[2].l[1] == nil ? u[i].f[1] : u[2].f[1])
		+ vxvy		* (u[3].l[1] == nil ? u[i].f[1] : u[3].f[1]);
	return u[0].s;
}
#else
char *
ginterp(vx, vy, gim)
	float	vx, vy;
	gimage	*gim;
{
	register struct gi *gi = (struct gi *)gim;
	register	x = vx, y = vy;
	register	i, nil = *((int *)&gi->gi_fnull);
	float	vxvy;
	static	double	d;
	union	{
		float	f;
		long	l;
		char	s[4];
	} u[4];

#ifdef BIDR
	if (gi->gi_magic == BIDR_MAGIC)
		return bidr_interp(vx, vy, gim);
#endif BIDR
	if (gi->gi_magic != GI_MAGIC)
		return NULL;

	if ((x & gi->gi_xmask) != gi->gi_xmask) {
		register char *c;
		c = gcell(x,y,gi);
		gconv(u[0].s, c, 2, CVT_F, gi->gi_cvt);
		c = (u[0].l != nil && (y & gi->gi_ymask) != gi->gi_ymask) ?
			c + gi->gi_xsize : gcell(x,y+1,gi);
		gconv(u[2].s, c, 2, CVT_F, gi->gi_cvt);
	} else {
		gconv(u[0].s, gcell(x,y,gi), 1, CVT_F, gi->gi_cvt);
		gconv(u[1].s, gcell(x+1,y,gi), 1, CVT_F, gi->gi_cvt);
		gconv(u[2].s, gcell(x,y+1,gi), 1, CVT_F, gi->gi_cvt);
		gconv(u[3].s, gcell(x+1,y+1,gi), 1, CVT_F, gi->gi_cvt);
	}
	vx -= x;
	vy -= y;
	vxvy = vx*vy;
	for (i = 0; u[i].l == nil; i++)
		if (i == 3)
			return (char *)&gi->gi_dnull;
	u[0].f = (1-vx-vy+vxvy) * (u[0].l == nil ? u[i].f : u[0].f)
		+ (vx-vxvy)	* (u[1].l == nil ? u[i].f : u[1].f)
		+ (vy-vxvy)	* (u[2].l == nil ? u[i].f : u[2].f)
		+ vxvy		* (u[3].l == nil ? u[i].f : u[3].f);
	gconv((char *)&d, u[0].s, 1, gi->gi_cvt, CVT_F);
	return (char *)&d;
}
#endif	TEST
	/*
	**	Close image and release table
	*/
gclose(gim)
	gimage	*gim;
{
	register struct gi *gi = (struct gi *)gim;
	register int	i;

	gipserrno = 0;
	if (gi == NULL)
		return EOF;
#ifdef BIDR
	if (gi->gi_magic == BIDR_MAGIC)
		return bidr_close(gim);
#endif BIDR
	gi->gi_magic = 0;
	if (gi->gi_flag & GI_WRITE)
		for (i = 0; i < gi->gi_nrings; i++) 
			if (gi->gi_nrec[i] < 0)
				;
			else if (lseek(gi->gi_fid, (off_t)(gi->gi_off
					+ (gi->gi_nrec[i] << gi->gi_shft)),
					0) == -1)
				gipserrno = GIPSBADWRITE;
			else if (write(gi->gi_fid,
					gi->gi_buf + (i << gi->gi_shft),
					gi->gi_block) != gi->gi_block)
				gipserrno = GIPSBADWRITE;
	free(gi->gi_buf);
	free((char *)gi);
	i = close(gi->gi_fid);
	return gipserrno ? gipserrno : i;
}
	/*
	**	Return number of significant binary digits
	*/
static
shift(m)
	register	m;
{
	register	i = 0;
	while (m >>= 1) i++;
	return(i);
}
