#ifndef lint
static char *sccsid = "%Z%$RCSfile: gips.c,v $	GIPS $Revision: 2.8 $ (MIT/CSR) $Date: 1998/05/13 00:51:45 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		gips.c		Creation Date:	12/30/83
**	Version:	$Revision: 2.8 $		Update:		$Date: 1998/05/13 00:51:45 $
**	Components:	gips()		Realworld -> Image (C)
**			egips()		Image ->Realworld (C)
**			gips_()		Realworld -> Image (Fortran)
**			egips_()	Image ->Realworld (Fortran)
**			pgips()		x/y Realworld -> Image
**			dgips()		Data pixel Realworld -> Image
**			makegips()	Setup Coefficients
**			makefile()	Initialize scale=f Coefficients
**			makepoly()	Initialize Polynomial Coefficients
**			maketree()	Comparison routine for qsort()
**			galloc()	'realloc' with error handling
**			get()		Lookup GIPS header variable
**
**	Function:	Translate image arrays between image and realworld
**			coordinates.
*/

#include        <stdlib.h>   /* for calloc, etc. */
#include	"gips.h"
#include	<math.h>

#define		R2D	114.591559026164642

static	struct	tab {
	double	max, min;	/* Float value limits */
	double	wrap;		/* scale=w limit */
	double	wnum;		/* double of num */
	double	linfac;		/* Linear scaling factor */
	double	*conv;		/* Pointers to conversion buffers */
	int	*iconv;		/* Pointers to inverse index buffers */
	char	*cconv;		/* Byte-data lookup array */
	int	num;		/* Number of pixels. */
	int	npoly;		/* Number of polynomial terms. */
	int	dlen;		/* Byte length of data type. */
	char	type;		/* Variable length code */
	char	scale;		/* Type of mapping */
	char	ch;		/* Type character x, y, or d */
} table[3];

static	double	dmin, dfact;	/* Data pixel scaling factors */

extern	int	arginit;	/* =0 if header empty or altered */
static	char	*galloc();
extern	char	*strcpy();
static	char	*galloc();
static		pgips(), dgips(), xgips(), makegips(), makefile();
static		maketree(), makepoly(), get();

	/*
	**	realworld value -> pixel address or value
	*/
gips(real, pix, nout, mode)

	register double	*real;		/* Input Array of Realworld values */
	register int	*pix;		/* Output Array of Image values */
	register int	nout;		/* Number of items to transform */
	register int	mode;		/* mode=0,1,2 for x/y/d axes */
{
	register char	typsav;
	/*
	**	bomb if bad count, mode. arginit==0 from libIH.a routines if
	**	the header has been updated - so re-make the coefficients
	*/
	if (nout < 1)
		return(gipserrno = GIPSBADARG);
	if (arginit == 0 && makegips())
		return(gipserrno);
	switch (mode) {
	case GIPS_X_VAL:
		pgips(real, pix, nout, table, mode);
		break;
	case GIPS_Y_VAL:
		pgips(real, pix, nout, table+1, mode);
		break;
	case GIPS_D_VAL:
		dgips((char *) real, (char *) pix, nout, table+2);
		break;
	case GIPS_XY_IVAL:
	case GIPS_XY_FVAL:
		while (--nout >= 0) {
			pgips(real++, pix++, 1, table, mode);
			pgips(real++, pix++, 1, table+1, mode);
		}
		break;
	case GIPS_D_DVAL:
		typsav = table[2].type;
		table[2].type = 'd';
		dgips((char *) real, (char *) pix, nout, table+2);
		table[2].type = typsav;
		break;
	case GIPS_R_VAL:
		xgips((char *) real, (double *) pix, nout, table+2);
		break;
	default:
		return(gipserrno = GIPSBADARG);
	}
	return(gipserrno = 0);
}
	/*
	**	pixel address or value -> realworld coordinate or value
	*/
egips(pix, real, nout, mode)

	register int	*pix;		/* Input Array of Int Image values */
	register double	*real;		/* Output Array of Realworld values */
	register int	nout;		/* Number of items to transform */
	register int	mode;		/* mode=0,1,2 for x/y/d axes */
{
	register struct tab *t;		/* table pointer */
	register i	= -1;		/* pix value */
	double	d;			/* scratch for scale=w */
	/*
	**	bomb if bad count, mode. arginit==0 from libIH.a routines if
	**	the header has been updated - so re-make the coefficients
	*/
	if (nout < 1)
		return(gipserrno = GIPSBADARG);
	switch (mode) {
	case EGIPS_X_VAL:
	case EGIPS_XY_VAL:
		t = table;
		break;
	case EGIPS_Y_VAL:
		t = table+1;
		break;
	case EGIPS_D_VAL:
		t = table+2;
		break;
	default:
		return(gipserrno = GIPSBADARG);
	}
	if (arginit == 0 && makegips())
		return(gipserrno);
	/*
	**	loop over input integers
	*/
	while (--nout >= 0) {
		i = pix ? *pix++ : i+1;		/* get pixel value */
		if (i < 0)			/* GIPSLOW if below limit */
			*real++ = GIPSLOW;
		else if (i > t->num)		/* HIGH if above limit */
			*real++ = GIPSHIGH;
		else if (mode == EGIPS_D_VAL &&
				t->scale != 'l' && t->scale != 'f')
			*real++ = dmin + i * dfact;
		else
			switch(t->scale) {
			case 's':
				*real++ = i;
				break;
			case 'p':		/* from polynomial */
				if (mode == EGIPS_X_VAL || mode == EGIPS_Y_VAL)
					goto index;
			default:		/* linear mappings */
			case 'S':
			case 'n':
			case 'w':
			case 'N':
				*real++ = t->min + i /t->linfac;
				break;
			case 'm':		/* mercator */
				*real++ =
					R2D * atan(exp(t->min+i/t->linfac))-90;
				break;
			case 'l':		/* logarithmic */
				*real++ = pow((double)10, t->min+i/t->linfac);
				break;
			case 'f':		/* from lookup tree */
				if (t->type == 's') {
					*real++ = t->conv[i+128];
					break;
				}
			index:
				*real++ = t->conv[i];
				break;
			}
		if (mode == EGIPS_XY_VAL)
			if (t++ == table)
				nout++;
			else
				t = table;
	}
	return(gipserrno = 0);
}
	/*
	**	realworld x,y coordinate to pixel address
	*/
static
pgips(in, out, nout, t, mode)

	register double	*in;			/* Input Realworld x,y array */
	register int	*out;			/* Output pixel x,y addresses */
	register	nout;			/* Number of array elements */
	register struct tab *t;			/* table[0,1] for x,y axes */
	int		mode;			/* if =5, use fout */
{
	float		*fout	= (float *)out;	/* float output pointer */
	double		d;			/* temporary */

	while (--nout >= 0) {
		switch (t->scale) {
		case 's':
			d = *in++;	
			break;
		case 'S':
			d = t->linfac * (*in++ - t->min);
			break;
		case 'w':
			if ((d = t->linfac * (*in++ - t->min)) < 0)
				d += t->wrap * (1 + (int)(d/t->wrap));
			else if (d > t->wnum)
				d -= t->wrap * (1 + (int)((d-t->wnum)/t->wrap));
			break;
		case 'f':
		case 'p': {
				register double *dp	= t->conv;
				register int	*ip	= t->iconv;
				register	i;

				i = t->num/2;
				d = *in++;
				while (i > 0 && *in != dp[*(ip += i)])
					i = (d > dp[*ip]) ? i/2 : -i/2;
				d = i;
			}
			break;
		case 'm':
			d = *in++ + 90;
			d = (d <= 0 || d >= 180) ? t->wnum :
				t->linfac*(log(tan(d/R2D))-t->min);
			break;
		case 'l':
			if (*in <= 0)
				d = *in++;
			else
				d = t->linfac*(log10(*in++)-t->min);
			break;
		default:
			break;
		}
		if (mode == GIPS_XY_FVAL)
			*fout++ = (float) d;
		else
			*out++ = (int) d;
	}
}
	/*
	**	array value -> pixel value
	*/
static
dgips(in, out, nout, t)

	register char	*in;		/* Input Realworld Array */
	register char	*out;		/* Output Image Pixel Array */
	register int	nout;		/* Array dimension */
	register struct	tab *t;		/* pointer to table */
{
	/*
	**	'Simple' Realworld -> Pixel mapping, i.e. neither dmin nor dmax
	**	defined in header, so just copy the fields, with possible change
	**	of data type to unsigned integer
	*/
	if (t->scale == 's' || t->scale == 'N') {
		gconv(out, in, nout, CVT_B, ihdcvt(t->type));
		return;
	}
	/*
	**	lookup of byte realworld data - 'iconv' can be used as a
	**	lookup table because the number of possible realworld values is
	**	limited to 0 - 255.
	*/
	if (t->cconv != NULL) {
		register char *c =  t->cconv;

		if (t->type == 'b')
			while (--nout >= 0)
				*out++ = c[*in++ & 255];
		else if (t->type == 's')
			while (--nout >= 0)
				*out++ = c[128 + *in++];
		return;
	}
	/*
	**	remaining types of operations are inherently less efficient,
	**	so put the type-dependent logic inside the array loop.
	*/
	for ( ; --nout >= 0; in += t->dlen) {
		double	fo;			/* float realworld value */
		double	fi;			/* conversions */
		register i;
		/*
		**	convert input realworld element to float
		*/
		switch (t->type) {
		case 'b':	fo = *in & 0377;		break;
		case 's':	fo = *in;			break;
		case 'h':	fo = *((short *)in);		break;
		case 'u':	fo = *((unsigned short *)in);	break;
		case 'i':	fo = *((long *)in);		break;
		case 'l':	fo = *((unsigned long *)in);	break;
		case 'f':	fo = *((float *)in);		break;
		case 'd':	fo = *((double *)in);		break;
		case 'c':	fo = *((float *)in);
				fi = ((float *)in)[1];
				fo = sqrt(fo*fo + fi*fi);	break;
		}
		/*
		**	file lookup employs binary lookup of 't->conv'.
		*/
		if (t->scale == 'f') {
			register double *dp	= t->conv;
			register int	*ip	= t->iconv;

			i = 128;
			while (i > 0 && fo != dp[*(ip += i)])
				i = (fo > dp[*ip]) ? (i >> 1) : -(i >> 1);
			*out++ = i;
			continue;
		}
		/*
		**	normalized linear mapping...
		*/
		if (t->scale == 'n')
			fo = t->conv[0] + fo * t->conv[1];
		/*
		**	polynomial and logarithmic mappings...
		*/
		else {
			if (t->scale == 'p') {
				register double *d = t->conv;

				for (i=t->npoly, fi = fo, fo = *d++; --i >= 0; )
					fo = *d++ + fo*fi;
			} else if (t->scale == 'l')
				fo = (fo > 0) ? log10(fo) : GIPSLOW;
			fo = t->linfac * (fo - t->min);
		}
		if (fo < 0)
			*out++ = 0;
		else if (fo >= 254.5)
			*out++ = 255;
		else
			*out++ = (int) (fo + 0.5);
	}
}
	/*
	**	array value -> realworld value
	*/
static
xgips(in, out, nout, t)

	register char	*in;		/* Input Image Array */
	register double	*out;		/* Output Realworld Value */
	register int	nout;		/* Array dimension */
	register struct	tab *t;		/* pointer to table */
{
	double	f1, f2;

	gconv((char *)out, in, nout, CVT_D, ihdcvt(t->type));

	switch (t->scale) {
	case 'n':
	case 'N':
		if (t->linfac == 0)
			f1 = GIPSLOW;
		else {
			f1 = t->min + t->conv[0] / t->linfac;
			f2 = t->conv[1] / t->linfac;
		}
	case 'p':
	case 'l':
		for (; --nout >= 0; out++) {
			if (t->scale == 'p') {
				register double *d	= t->conv;
				register	i	= t->npoly;
		
				for (f1 = *out, *out = *d++; --i >= 0; )
					*out = *d++ + *out*f1;
			} else if (t->scale == 'l')
				*out = (*out > 0) ? log10(*out) : GIPSLOW;
			else
				*out = (f1 == GIPSLOW) ? f1 : f1 + f2 * *out;
		}
	default:
		break;
	}
}
	/*
	**	reconstruct the conversion buffers used by above routines
	*/
static
makegips()
{
	register i;			/* axis index i=0,1,2 for x,y,d */
	register struct tab *t;		/* table pointer */
	register char *chtype;		/* data type */

	for (chtype = "xyd", t = table; *chtype; t++, chtype++) {
		/*
		**	get correct type and scale from header.
		**	x- and y-axes must always have x/ytype = d 
		*/
		t->ch = *chtype;
		if (t->ch != 'd')
			t->type = 'd';
		else if (get(t, "xtype", &t->type, 1))
			t->type = 'b';
		/*
		**	fix up cellular geometry
		*/
		if (t->ch == 'd' && t->type == 'G') {
			char	buf[7];
			ihget("geometry", buf, sizeof(buf));
			sscanf(buf, "type=%c", &t->type);
		}
		t->dlen = ihdtype(t->type);
		if (get(t, "xscale", &t->scale, 1) != 0)
			t->scale = 's';
		/*
		**	get dimension of the image coordinate -
		**	always 256 for pixels
		*/
		if (t->ch == 'd') t->num = 256;
		else if (get(t, "xnum", (char *)&t->num, 0))
			return(gipserrno);
		t->wnum = (double)t->num;
		/*
		**	get minimum and maximum realworld limits of coordinates.
		**	these will be disregarded for scale=f specifications
		*/
		if (get(t, "xmin", (char *)&t->min, 0))
			if (t->scale == 'm')
				return(gipserrno);
			else
				t->min = 0;
		if (get(t, "xmax", (char *)&t->max, 0))
			if (t->scale == 'm')
				return(gipserrno);
			else
				t->max = t->wnum;
		if (t->scale != 'f')
			if (t->max == t->min)
				return(gipserrno = GIPSBADLIM);
			else
				t->linfac = t->wnum / (t->max-t->min);
		if (t->ch == 'd') {
			dmin = t->min;
			dfact = 1 / t->linfac;
		}
		/*
		**	each scale type requires very different initialization
		**	note that scale=s,w,n with min and/or max defined is
		**	renamed to scale=S,W,N for efficiency of conversion
		*/
		switch(t->scale) {
		case 's':
			if (t->min != 0 || t->max != t->wnum) {
				t->scale = 'S';
				break;
			}
			continue;
		case 'l':
			if (t->min <= 0) {
				(void)(strcpy(*gipserrs+1, "min"));
				return(gipserrno = GIPSBADLIM);
			} else
				t->min = log10(t->min);
			if (t->max <= 0)
				return(gipserrno = GIPSBADLIM);
			else
				t->max = log10(t->max);
			t->linfac = t->wnum/(t->max-t->min);
			break;
		case 'p':
			if (makepoly(t))
				return(gipserrno);
			break;
		case 'f':
			if (makefile(t))
				return(gipserrno);
			continue;
		case 'n':
			if (t->ch == 'd') {
				t->conv = (double *)galloc((char *)t->conv,
						2, sizeof(double));
				if (t->conv == NULL ||
				    ihget("dzero", (char *)(&t->conv[0]), 0) ||
				    ihget("dfact", (char *)(&t->conv[1]), 0))
					return(gipserrno);
				t->conv[0] = (t->conv[0] - t->min) * t->linfac;
				t->conv[1] *= t->linfac;
				if (t->conv[0] == 0 && t->conv[1] == 1.0) {
					t->scale = 'N';
					continue;
				}
				break;
			}
		case 'm':
			if (t->ch != 'd') {
				if (t->min <= -90 || t->min >= 90) {
					(void)(strcpy(*gipserrs+1, "max"));
					return(gipserrno = GIPSBADLIM);
				}
				if (t->max <= -90 || t->max >= 90)
					return(gipserrno = GIPSBADLIM);
				t->max = log(tan((t->max+90)/R2D));
				t->min = log(tan((t->min+90)/R2D));
				t->linfac = t->wnum/(t->max-t->min);
				break;
			}
		case 'w':
			if (t->ch != 'd') {
				if (get(t, "xwrap", &t->wrap, 1))
					t->wrap = t->max - t->min;
				if (t->wrap == 0)
					return gipserrno = GIPSBADLIM;
				t->wrap = fabs(t->wrap * t->linfac);
				break;
			}
		default:
			(void)(strcpy(*gipserrs+1, "scale"));
			return(gipserrno = GIPSUNSCALE);
		}
		/*
		**	for byte -> byte pixel conversions,
		**	construct a lookup table
		**	i=2,dscale=s,N,f will not get this far
		*/
		if (t->ch == 'd' && t->dlen == 1 &&
				t->scale != 's' && t->scale != 'N') {
			char inbuf[256], *tmp;
			register i;

			tmp = galloc(t->cconv, 256, sizeof(char));
			t->cconv = NULL;
			if (tmp == NULL)
				return(gipserrno);
			for (i = 0; i < 256; inbuf[i] = i, i++)
				;
			dgips(inbuf, tmp, 256, t);
			t->cconv = tmp;
		} else if (t->cconv) {
			free(t->cconv);
			t->cconv = NULL;
		}
	}
	/*
	**	arginit is reset to zero in libIH.a whenever a header variable
	**	is updated, so it is an indication to gips() and egips() that 
	**	makegips() must be repeated.
	*/
	arginit = 1;
	return(gipserrno = 0);
}
	/*
	**	galloc() is a combination of calloc() and realloc()
	**	with error message generation
	*/
static	char *
galloc(s, n, m)

	register char	*s;		/* area to be reallocated */
	register int	n;		/* element size */
	register int	m;		/* number of elements */
{
/*	extern	char	*calloc(), *realloc(); */

	if (s == NULL)
		s = calloc((unsigned) n, (unsigned) m);
	else
		s = realloc(s, (unsigned) (n * m));
	if (! s)
		gipserrno = GIPSNOCORE;
	return(s);
}
	/*
	**	scale=f implementation
	*/
static
makefile(t)

	register struct tab *t;		/* pointer to table */
{
	char	fname[BUFLEN];		/* filename buffer */
	FILE	*ff;			/* file stream pointer */
	register double	*d;		/* realworld table */
	register int	n;		/* number of input values */
	register int	m;		/* number of input values */

	if (get(t, "xfile", fname, sizeof(fname)-1))
		return(gipserrno = GIPSNOVAR);

	if ((ff = fopen(fname, "r")) == NULL)
		return(gipserrno = GIPSBADOPEN);

	m = n = (t->ch == 'd' && ihdtype(t->type) == 1) ? 256 : t->num+1;

	d = t->conv = (double *) galloc((char *)t->conv, n, sizeof(double));
	if (d == NULL)
		return(gipserrno);
	/*
	**	input file can contain the values in any format
	**	compatible with the following fscanf() call
	*/
	while (--n >= 0 && fscanf(ff, "%lf", d++) == 1)
		;
	(void)(fclose(ff));
	if (n >= 0)
		return(gipserrno = GIPSBADDATA);
	/*
	**	if byte data, construct translate table
	*/
	if (t->ch == 'd' && (t->type == 'b' || t->type == 's')) {
		register char *c;

		t->iconv = (int *)
			(c = galloc((char *)t->iconv, m, sizeof(int)));
		if (c == NULL)
			return(gipserrno);
		t->min = (t->type == 'b') ? 0 : -128;
		t->max = (t->type == 'b') ? 255 : 127;
		for (n = 0; n < m; )
			if (*d >= t->min && *d <= t->max) {
				register j = (int) *d++;
				if (t->type == 'b')
					j += 128;
				c[j] = n++;
			}
		(void)(ihput("dmin", (char *)&t->min));
		(void)(ihput("dmax", (char *)&t->max));
		return(gipserrno);
	}
	return(gipserrno = maketree(t, m));
}
	/*
	**	generate the iconv inverted list 
	*/
static
maketree(t, number)

	struct	 tab *t;		/* table pointer */
	int	 number;		/* number of t->conv elements */
{
	register	n = number;	/* number of conv[] elements */
	register int	*ip;		/* pointer to iconv[] */
	register int	*p;		/* pointer to iconv[] */
	register double	*dp = t->conv;	/* pointer to conv[] */
	register	i;		/* scratch */
	register 	m = 0;		/* scratch */
	register 	g;		/* scratch */

	/*
	**	allocate t->iconv
	*/
	ip = t->iconv = (int *) galloc((char *)t->iconv, n, sizeof(int));
	if (ip == NULL)
		return(gipserrno);
	/*
	**	for byte data, create binary lookup table
	*/
	while (m < n)
		*ip++ = m++;
	/*
	**	shell sort of ip[] such that dp[ip[n]] <= dp[ip[n+1]]
	*/
	for (ip = t->iconv, g = n/2; g > 0; g /= 2)
		for (i = g; i < n; i++)
			for (p=ip+i-g; p >= ip && dp[*p] > dp[m=p[g]]; ) {
				p[g] = *p;
				*p = m;
				p -= g;
			}
	ip = t->iconv;
	t->min = dp[ip[0]];
	t->max = dp[ip[number-1]];
	(void)(ihput((t == table) ? "xmin" : "ymin", (char *)&t->min));
	(void)(ihput((t == table) ? "xmax" : "ymax", (char *)&t->max));
	return(gipserrno);
}
	/*
	**	handle scale=p polynomial mapping specifications. the
	**	polynomial index is in x/y/dpoly and the coefficients in
	**	x/y/dparm (which is a string variable). for x and y axis
	**	polynomials, first allocate some t->conv to hold the float
	**	x/yparm elements, then make a lookup table to convert from
	**	image x,y (range 0->num) to their realworld values. this
	**	table then replaces t->conv. for dscale=p, conv always
	**	contains the polynomial coefficients because it is never
	**	necessary to invert the polynomial mapping.
	*/ 
static
makepoly(t)

	register struct tab *t;		/* table pointer */
{
	FILE	*ff;			/* string 'stream' pointer */
	char	parm[BUFLEN];		/* buffer for ASCII polynomial terms */
	double	d;			/* realworld value */
	register double	*pol;		/* polynomial coefficients */
	register double	*dd;		/* poly index pointer */
	register double	*dp;		/* poly value pointer */
	register int	n;		/* polynomial index */
	register int	j;		/* polynomial index */

	if (get(t, "xpoly", (char *)&t->npoly, 0))
		return(gipserrno);
	if (t->npoly < 0)
		return(gipserrno = GIPSBADVAL);
	if (get(t, "xparm", parm, sizeof(parm)-1))
		return(gipserrno);
	/*
	**	(re)allocate the polynomial array
	*/
	t->conv = (double *) galloc((char *)t->conv,
			j = t->npoly+1, sizeof(double));
	if (t->conv == NULL)
		return(gipserrno);
	/*
	**	read npoly+1 elements from the 'parm' string
	*/
	ff = fopens(parm);
	dp = t->conv + j;
	while (--j >= 0)
		if (fscanf(ff, "%lf", --dp) != 1)
			return(gipserrno = GIPSBADVAL);
	if (t == table+2)
		return(gipserrno = 0);
	/*
	**	reallocate the conversion array - the polynomial 
	**	coefficient array pointer is saved in 'pol'
	*/
	pol = t->conv;
	dd = t->conv = (double *)
		galloc((char *)NULL, j = t->num+1, sizeof(double));
	if (dd == NULL)
		return(gipserrno);
	/*
	**	fill 'conv' with realworld values for coordinates
	**	spanning the image
	*/
	for (d = 0; --j >= 0; dd++, d++)
		for (n = t->npoly, dp = pol, *dd = *dp++; --n >= 0; )
			*dd = *dp++ + *dd * d;
	free((char *) pol);
	/*
	**	construct iconv inverted list array
	*/
	return(maketree(t, t->num+1));
}
	/*
	**	lookup value of header item, returning code from ihget()
	**	(non-zero if item not found for any reason)
	*/
static
get(t, s, v, m)

	register struct tab *t;		/* table pointer */
	register char *s;		/* pointer to header item name */
	char	 *v;			/* pointer to header item value */
	int	 m;			/* length of header item value */
{
	char	 buf[10];		/* item name buffer */
	register char *c;		/* scratch */

	for (c = buf, *c++ = t->ch; *c++ = *++s ; )
		;
	return(ihget(buf, v, m));	/* return value */
}
	/*
	**	Call gips() from fortran
	*/
gips_(real, pix, nout, mode)

	double	*real;		/* Input Array of Realworld values */
	int	*pix;		/* Output Array of Image values */
	int	*nout;		/* Number of items to transform */
	int	*mode;		/* mode=0,1,2 for x/y/d axes */
{
	return gips(real, pix, *nout, *mode);
}
	/*
	**	Call gips() from fortran
	*/
egips_(pix, real, nout, mode)

	int	*pix;		/* Input Array of Image values */
	double	*real;		/* Output Array of Realworld values */
	int	*nout;		/* Number of items to transform */
	int	*mode;		/* mode=0,1,2 for x/y/d axes */
{
	return egips(pix, real, *nout, *mode);
}
