#ifndef lint
static char *sccsid = "%Z%$RCSfile: gtrans.c,v $	GIPS $Revision: 2.22 $ (MIT/CSR) $Date: 1998/05/13 00:51:49 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		gtrans.c	Creation Date:	06/29/87
**	Version:	$Revision: 2.22 $		Update:		$Date: 1998/05/13 00:51:49 $
**	Components:	g_listproj()	Write projection info
**			g_setup()	Initialize a projection
**			g_map()		Map from input to output
**			g_unmap()	Map from output to input
**
**	Function:	Transform an image via non-local mappings
*/

#define		GRID
#include	<math.h>
#include	"gips.h"

#define	SGN(x)	(((x)>0)?1:-1)
#define	MAXPROJ	(sizeof(table)/sizeof(G_TRANS)-1)

G_PARM	*g_proj;			/* master projection structure */

	/*
	 *	projection routines
	 */
extern	gnomonic(), stereographic(), orthographic(), cylindrical(); 
extern	equidistant(), equivalent(), lambert1(), lambert2(); 
extern	mercator(), polstereo(), mollweide(), albers1(), albers2(); 
extern	lambert3(), lambert4(), bonne(), sinusoidal(), werner(), user(); 
extern	aitoff(), hammer(), briesemeister();
static	euler(), seteuler();
	/*
	 *	master projection table
	 */
static	G_TRANS	 table[] = {
	NULL,		"Tiepoint",	"Polynomial Bi-variant",	"-n",
	gnomonic,	"Gnomonic",	"Azimuthal r=tan(d)",		"",
	stereographic,	"Stereographic","Azimuthal r=2tan(d/2)",	"",
	orthographic,	"Orthographic",	"Azimuthal r=sin(d)",		"",
	equidistant,	"Equidistant",	"Azimuthal r=d\t",		"",
	equivalent,	"Equivalent",	"Azimuthal r=2sin(d/2)",	"",
	lambert1,	"Lambert_Conic_1",	"Conformal One parallel",	"-p",
	lambert2,	"Lambert_Conic_2",	"Conformal Two parallels",	"-pP",
	mercator,	"Mercator",	"Conformal Equatorial",		"",
	polstereo,	"Stereo_Polar",	"Conformal Stereographic",	"",
	mollweide,	"Mollweide",	"Equivalent Non-conical",	"",
	albers1,	"Albers_1",	"Equivalent One parallel",	"-p",
	albers2,	"Albers_2",	"Equivalent Two parallels",	"-pP",
	lambert3,	"Lambert_Cylindrical",	"Equivalent Cylindrical",	"",
	lambert4,	"Lambert_Polar","Equivalent Polar azimuthal",	"",
	bonne,		"Bonne",	"Equivalent Pseudo-conical",	"",
	sinusoidal,	"Sinusoidal",	"Equivalent",			"",
	werner,		"Werner",	"Equivalent",			"",
	user,		"User_Defined",	"via subroutine \"user()\"",	"",
	aitoff,		"Aitoff",	"Equivalent Non-conical",	"",
	hammer,		"Hammer",	"Equivalent Non-conical",	"",
	briesemeister,	"Briesemeister","Equivalent Non-conical",	"",
	cylindrical,	"Simple_Cylindrical",	"Cylindrical",		"",
	NULL
};
	/*
	**	Write projection types to stream
	*/
void
g_listproj(fp)
	FILE	*fp;
{
	register G_TRANS	*tp;

	fprintf(fp, "-n\tNAME\t\t\tDESCRIPTION\t\t\tOPTIONS\n");
	for (tp = table; tp->name; tp++)
		fprintf(fp, "%2d\t%-19.19s\t%-26.26s\t%s\n", tp-table,
			tp->name, tp->desc, tp->init);
}
	/*
	**	Initialize mapping parameters
	*/
g_setup()
{
	char	buf[BUFSIZ], pname[64];
	double	ea[3], phi0, phi1, radius;
	int	nproj;
	register G_PARM	*t;

	if (ihget("mapping", buf, sizeof(buf)-1))
		return gipserrno;
	if (g_proj)
		free((char *)g_proj);
	t = g_proj = (G_PARM *)alloc(sizeof(G_PARM), 1);
	if (! t)
		return gipserrno = GIPSNOCORE;
	t->xorg = t->yorg = t->xyrot = 0;
	radius = ea[0] = ea[1] = ea[2] = phi0 = phi1 = 0;
	if (sscanf(buf, GIPS_MAPPING, pname, &nproj, &t->xorg, &t->yorg,
		&radius, &ea[0], &ea[1], &ea[2], &phi0, &phi1, &t->xyrot) < 1)
		return gipserrno = GIPSBADMAP;
	if (nproj < 0 || nproj > MAXPROJ || table[nproj].proc == NULL ||
			t->xyrot > 1 || t->xyrot < -1)
		return gipserrno = GIPSBADMAP;
	else	
		t->tproj = &table[nproj];
	if (ea[0] != 0 || ea[1] != 0 || ea[2] != 0) {
		if (seteuler(ea))
			return gipserrno = GIPSBADMAP;
		t->euler_sw = 1;
	} else
		t->euler_sw = 0;
	t->deg_sw = t->euler_sw || (t->tproj != table && *t->tproj->proc != user);
	errno = 0;
	if ((*t->tproj->proc)(0, phi0, phi1, &radius) != 0)
		return gipserrno = GIPSBADMAP;
	return gipserrno = 0;
}
	/*
	**	Return projection name
	*/
char *
g_getname(n)
{
	return (n < 0 || n > MAXPROJ) ? "" : table[n].name;
}
	/*
	**	Supply projection procedure
	*/
g_setproc(n, p)
	int	n, (*p)();
{
	if (n < 0 || n > MAXPROJ)
		return 1;
	table[n].proc = p;
	return 0;
}
	/*
	**	Map from input realworld value to output pixel address
	*/
g_map(lon, lat, xy)
	double lon, lat, xy[2];
{
	register ret;
	register G_PARM	*t = g_proj;

	if (t->deg_sw) {
		if (lon >= 360 || lon <= -360)
			lon -= 360*(int)(lon/360);
		if (lon > 180)
			lon -= 360;
		else if (lon <= -180)
			lon += 360;
		lon *= DEG_RAD;
		if (lat > 90 || lat < -90)
			return 1;
		lat *= DEG_RAD;
		if (t->euler_sw && euler(&lon, &lat, 0))
			return 1;
	}
	errno = 0;
	ret = (*t->tproj->proc)(1, lon, lat, xy);
	if (t->tproj == table)
		return ret;
	switch (t->xyrot) {
	case 1:
		lon = -xy[1], xy[1] = xy[0], xy[0] = lon;
		break;
	case -1:
		lon = xy[1], xy[1] = -xy[0], xy[0] = lon;
		break;
	}
	if (ret == 0) {
		xy[0] = xy[0] + t->xorg;
		xy[1] = t->yorg - xy[1];
	}
	return ret;
}
	/*
	**	Map from output pixel address to input realworld address
	*/
g_unmap(x, y, lonlat)
	double	lonlat[2];
{
	register ret;
	register G_PARM	*t = g_proj;

	errno = 0;
	if (t->tproj == table)
		return (*t->tproj->proc)(2, (double)x, (double)y, lonlat);
	switch (t->xyrot) {
	case 1:
		ret = (*t->tproj->proc)(2, t->yorg - y, t->xorg - x, lonlat);
		break;
	case 0:
		ret = (*t->tproj->proc)(2, x - t->xorg, t->yorg - y, lonlat);
		break;
	case -1:
		ret = (*t->tproj->proc)(2, y - t->yorg, x - t->xorg, lonlat);
		break;
	}
	if (ret != 0)
		return ret;
	if (t->deg_sw) {
		if (lonlat[0] < -M_PI || lonlat[0] > M_PI)
			return 1;
		else if (lonlat[1] < -M_PI_2 || lonlat[1] > M_PI_2)
			return 1;
		if (t->euler_sw && euler(&lonlat[0], &lonlat[1], 1))
			return 1;
		lonlat[0] *= RAD_DEG, lonlat[1] *= RAD_DEG;
	}
	return 0;
}
	/*
	**	Rotate by Euler transformation
	*/
static
euler(lon, lat, sw)
	register double	*lon, *lat;
	register sw;
{
	double	p[3], q[3], r;
	register i, j;
	register G_PARM	*t = g_proj;

	r = cos(*lat);
	p[1] = sin(*lon) * r;
	p[0] = cos(*lon) * r;
	p[2] = sin(*lat);
	for (i = 0; i < 3; q[i++] = r)
		for (r = 0, j = 0; j < 3; j++)
			r += p[j] * (sw ? t->e[j][i] : t->e[i][j]);
	if ((q[0] == 0 && q[1] == 0) || q[2] > 1 || q[2] < -1) return 1;
	*lon = atan2(q[1], q[0]);
	*lat = asin(q[2]);
	return 0;
}
	/*
	 *	Construct transformation matrix from -tpr options
	 */
static
seteuler(a)
	double	a[3];		/* Euler angles */
{
	double	lon, lat, rho;
	double	costh, sinth, tanth, costhc, sinthc, costhp, sinthp;
	double	thetap, thetac, lamc, lamp;
	register G_PARM	*t = g_proj;

	lon = a[0] * DEG_RAD;
	lat = a[1] * DEG_RAD;
	rho = a[2] * DEG_RAD;

	if (fabs(lat) < M_PI_2) {
		costh = cos(lat);
		sinth = sin(lat);
		tanth = sinth/costh;
		sinthp = costh * cos(rho);
		thetap = asin(sinthp);
		costhp = cos(thetap);
		sinthc = costh * sin(rho);
		thetac = asin(sinthc);
		costhc = cos(thetac);
		if (fabs(thetap) == M_PI_2)
			lamp = M_PI * SGN(thetap);
		else {
			lamp = tan(thetap)*tanth;
			if (fabs(lamp) > 1 && fabs(lamp) < 1.0000001)
				lamp = SGN(lamp);
			lamp = lon - SGN(rho)*acos(-lamp);
		}
		if (fabs(thetac) == M_PI_2)
			lamc = lon;
		else {
			lamc = tan(thetac)*tanth;
			if (fabs(lamc) > 1 && fabs(lamc) < 1.0000001)
				lamc = SGN(lamc);
			lamc = lon + acos(-lamc);
		}
	} else if (fabs(lat) == M_PI_2) {
		sinth = (lat > 0) ? 1 : -1;
		costhp = costhc = 1;
		costh = sinthp = sinthc = 0;
		lamp = (lat > 0) ? lon + M_PI : lon;
		lamc = lon + M_PI_2; 
	} else
		return 1;

	t->e[0][0] = costh * cos(lon);
	t->e[0][1] = costh * sin(lon);
	t->e[0][2] = sinth;
	t->e[1][0] = costhc * cos(lamc);
	t->e[1][1] = costhc * sin(lamc);
	t->e[1][2] = sinthc;
	t->e[2][0] = costhp * cos(lamp);
	t->e[2][1] = costhp * sin(lamp);
	t->e[2][2] = sinthp;

	return 0;
}
