#ifndef lint
static char *sccsid = "%Z%$RCSfile: icvtf.c,v $	GIPS $Revision: 2.4 $ (MIT/CSR) $Date: 2004/11/02 20:44:49 $";
#endif

/*
 *		   General Image Processing System
 *			   Peter G. Ford
 *		      Center for Space Research
 *	Copyright 1984 Massachusetts Institute of Technology
 *
 *	Module:		icvtf.c		Creation Date:	04/28/84
 *	Version:	$Revision: 2.4 $		Update:		$Date: 2004/11/02 20:44:49 $
 *	Function:	Convert an IBM binary array to a VAX binary array.
 *			Called from Fortran or C routines.
 *	Syntax:		ierr = icvti(iibm, nitem, iarr);
 *			ierr = icvtf(fibm, nitem, farr);
 *			ierr = icvtd(dibm, nitem, darr);
 *			ierr:	number of conversion errors (see note 3)
 *			iibm:	input array of 4-byte IBM integers
 *			iarr:	output array of 4-byte VAX integers
 *			fibm:	input array of 4-byte IBM floats
 *			farr:	output array of 4-byte VAX F-format floats
 *			dibm:	input array of 8-byte IBM doubles (REAL*8)
 *			darr:	output array of 8-byte VAX doubles (D-format)
 *			nitem:	number of items to convert
 *
 *	Notes:	[1]	It is assumed that the IBM input fields have
 *			NOT already been byte-swapped.
 *		[2]	Unnormalized IBM floats are interpreted correctly.
 *		[3]	While there are no specific bit patterns that represent
 *			illegal IBM floating point fields, the IBM exponent
 *			range is larger than the VAX. Fields that would
 *			cause overflow or underflow when converting are
 *			set to the hexadecimal values described below. The
 *			value returned by icvtf or icvtd tells the calling
 *			program the number of such fix-ups that were made.
 *			icvti always returns a zero value.
 *			Routine		Condition	Output Value
 *			icvtf		underflow 	0x00000100
 *			icvtf		overflow	0xffffffff
 *			icvtd		underflow	0x0000010000000000
 *			icvtd		overflow	0xffffffffffffffff
 */

static	int	errors		= -1;		/* init flag & error count */
static	double	table[256]	= { 0.0 };	/* powers of 2 */

static
init()
{
	register i = 127;
	register double *d1 = &table[128];
	register double *d2;

	for (*(d2 = d1) = 1.0; --i > 0; d1++)
		*--d2 = 1.0/(d1[1] = 2 * *d1);
	d1[1] = *d1;
	d2[-1] = *d2;
}

static double
dconv(buf, sw)

	register unsigned char	*buf;
	register		sw;
{
	register unsigned char	*b	= buf;
	register unsigned long	fract	= 0;
	register		exp	= 4;
	register		shft;
	double			f;
	static	short		shift[]	= {-1,3,2,2,1,1,1,1,0,0,0,0,0,0,0,0};

	while (--exp >= 0)
		fract = (fract << 8) | *b++;
	shft = shift[(fract >> 20) & 0xf];
	if (shft < 0) {
		if (fract) errors++;
		return 0;
	}
	exp = ((int)((fract >> 24) & 0177)-64)*4+128-shft;
	fract = (fract << shft) & 0xffffff;
	if (exp <= 1)
		if (exp == 1)
			fract >>= 1;
		else {
			exp = 1;
			fract = 0x800000;
			errors++;
		}
	else if (exp >= 255)
		if (exp == 255)
			fract <<= 1;
		else {
			exp = 255;
			fract = 0x7fffff;
			errors++;
		}
	f = fract * table[104];
	if (sw) {
		for (fract = 0, sw = 4; --sw >= 0; )
			fract = (fract << 8) | *b++ ;
		if (fract & 0x80000000) {
			fract &= ~0x80000000;
			f += table[104-1+shft];
		}
		f += fract * table[104-32+shft];
	}
	return f * ((*buf & 0x80) ? -table[exp] : table[exp]);
}

icvtf(fibm, nitem, farr)

	register char	*fibm;
	register int	nitem;
	register float	*farr;
{
	if (errors < 0) init();
	for (errors = 0; --nitem >= 0; fibm += 4)
		*farr++ = (float) dconv((unsigned char *)fibm, 0);
	return errors;
}

icvtd(dibm, nitem, darr)

	register char	*dibm;
	register int	nitem;
	register double	*darr;
{
	if (errors < 0) init();
	for (errors = 0; --nitem >= 0; dibm += 8)
		*darr++ = dconv((unsigned char *)dibm, 1);
	return errors;
}

icvti(iibm, nitem, iarr)

	register char	*iibm;
	register int	nitem;
	register int	*iarr;
{
	register i;
	register n;

	for ( ; --nitem >= 0; *iarr++ = n)
		for (n = 0, i = 4; --i >= 0; )
			n = (n << 8) | (*iibm++ & 0377);
	return 0;
}

icvtf_(fibm, pts, farr)

	char	*fibm;
	int	*pts;
	float  	*farr;
{
	return icvtf(fibm, *pts, farr);
}

icvtd_(dibm, pts, darr)

	char	*dibm;
	int	*pts;
	double	*darr;
{
	return icvtd(dibm, *pts, darr);
}

icvti_(iibm, pts, iarr)

	char	*iibm;
	int	*pts;
	int	*iarr;
{
	return icvti(iibm, *pts, iarr);
}

#include <stdio.h>
#if __SWR
char *gcvt(double d, int p, char *s)
{
	(void)sprintf(s, "%.*g", p, d);
	return s;
}
#endif
