#ifndef lint
static char *sccsid = "%Z%$RCSfile: ihread.c,v $	GIPS $Revision: 2.7 $ (MIT/CSR) $Date: 1998/11/13 23:08:03 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		ihread.c	Creation Date:	12/30/83
**	Version:	$Revision: 2.7 $		Update:		$Date: 1998/11/13 23:08:03 $
**	Components:	ihread()	construct incore header from named file
**			ihmerg()	append data to incore header
**			ihwrit()	write incore header to named file
**			ihget()		return value of named item
**			ihput()		set value of named item
**			ihdel()		remove value of named item
**			ihtype()	return type and description of item
**			ihbcl()		close file or pipe
**			ihseek()	locate item in incore table
**			ihfree()	deallocate incore table
**			ihclose()	close input file on errors
**			hashit()	update incore hash tables
**			hash()		hash an item name
**			iherrm_()	write error message to stderr
**
**	Function:	Initialize and fill the incore GIPS header table
*/

#include        <string.h>   /* for strlen */
#include	<stdlib.h>
#include	"gips.h"
#include	<ctype.h>
#include	<sys/types.h>
	/*
	**	header switch, =0 if incore header updated
	*/
int	arginit;
	/*
	**	Allowed variable type codes
	*/
#define	HDR_CHAR	'c'
#define	HDR_DOUBLE	'd'
#define	HDR_END		'E'
#define	HDR_ENUM	'V'
#define	HDR_HISTORY	'h'
#define	HDR_INTEGER	'i'
#define	HDR_REQUIRED	'R'
#define	HDR_STRING	's'

static	char	allowed_types[]	= {
		HDR_CHAR, HDR_DOUBLE, HDR_END, HDR_ENUM,
		HDR_HISTORY, HDR_INTEGER, HDR_REQUIRED, HDR_STRING, '\0'
};

static	char	string_types[]	= {
		HDR_ENUM, HDR_HISTORY, HDR_REQUIRED, HDR_STRING, '\0'
};

#define	OUTLEN		72	/* max length of header line */
#define	HASHNUM		100	/* dimension of variable name hash table */
#define	VAR_INPROFILE	01	/* variable was defined in profile */
#define	VAR_INHEADER	02	/* variable was found in header */
#define	VAR_ADDED	04	/* variable was added after profile */
#define	VAR_UPDATED	010	/* variable has been altered */
#define	VAR_AVAILABLE	(VAR_INHEADER|VAR_ADDED|VAR_UPDATED)
	/*
	**	structure defining a single header variable
	*/
typedef	struct	key {
	struct	key *next;		/* chain pointer */
	struct	key *hashptr;		/* pointer to hash-lookup aliases */
	char	type;			/* variable type */
	char	flgs;			/* VAR_ flags */
	union	{
		double	dval;		/* double precision value */
		int	ival;		/* integer value */
		char	cval;		/* character value */
		char	*sval;		/* string value */
	}	uval;
	char	name[1];		/* variable name or history item */

} KEY;

static	KEY	keyanch = { NULL, &keyanch };	/* storage anchor */
static	KEY	*hashtab[HASHNUM];		/* hash chain anchors */
static	char	isapipe[256];			/* pipe stream flags */

static	check_required(), hashit(), hash(), ihfree(), bufcor();

extern	char	*argsav[];		/* routine name and argv list */
extern	char	*ctime(), *index();
extern	time_t	time();			/* C library - returns date & time */
	/*
	**	Check for valid item request
	*/
static
ihcheck(s, t, u)

	register char *s;		/* item name */	
	register char *t;		/* item value */
	register char *u;		/* default error title */
{
	if (! argsav[0])
		argsav[0] = u;
	if (! s || ! t || ! *s)
		return gipserrno = GIPSBADARG;
	else if (! keyanch.next)
		return gipserrno = GIPSNOHEAD;
	else
		return gipserrno = 0;
}
	/*
	**	return pointer to incore table entry for item name "s"
	*/
static	KEY *
ihseek(s, u, na)

	register	char *s;	/* item name */
	register	char *u;	/* variable pointer */
	register	char *na;	/* routine name */
{
	register	KEY *t;		/* pointer to incore element */

	if (ihcheck(s, u, na))
		return NULL;
	/*
	**	lookup through the hash table,
	**	not through the linked incore table
	*/
	for (t = hashtab[hash(s)]; t; t = t->hashptr)
		if (! strcmp(s, t->name))
			return t;
	gipserrno = GIPSBADVAR;
	return NULL;
}
	/*
	**	close input file, write message to stderr,
	**	and return to caller
	*/
static	FILE *
ihclose(f)

	register FILE	*f;
{
	if (f && ! isastring(f) && fileno(f) > 2)
		(void)(ihbcl(f));
	ihfree();
	return NULL;
}
	/*
	**	Read header profile table and header
	*/
FILE *
ihread(fname, prof)

	char	*fname;			/* name of input header file */
	char	*prof;			/* name of header definition file */
{
	register FILE	*fp;		/* input file pointer */
	char		buf[BUFLEN];	/* buffer to hold description field */
	register	efind = 0;	/* =1 if type E variable defined */
	register	retc = 0;	/* error code */
	register KEY	*t;		/* pointer to incore table */
	/*
 	**	free any previous incore table
 	*/
	ihfree();
	/*
 	**	verify callers arguments
 	*/
	if (ihcheck(fname, prof, "ihread"))
		if (gipserrno != GIPSNOHEAD)
			return NULL;
		else
			gipserrno = 0;
	/*
 	**	open the profile file. if pathname relative,
	**	try working directory then 'GIPSPROFDIR'
 	*/
	if (! (fp = fopen(strcpy(*gipserrs, prof), "r")) && *prof != '/')
		fp = fopen(strcat(strcpy(buf, GIPSPROFDIR), prof), "r");
	if (! fp) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	/*
 	**	profile read loop
 	*/
	for ( t = &keyanch; ;) {
		char	c;		/* variable type code */
		char	d;		/* first byte of val */
		register char	*f;	/* input format */
		register char	*s;	/* pointer to buf[] */
		register n;		/* scratch */
		/*
 		**	read variable name, type, and first byte of description
		**	verify allowable type code
 		*/
		if ((retc = fscanf(fp, "%s %1c %1c", buf, &c, &d)) != 3 ||
				! index(allowed_types, c))
			break;
		/*
		**	read variable description into remainder of buf
		*/
		n = strlen(buf);
		s = buf+n;
		if (d == '"')
			f = "%[^\"]\"";
		else
			*++s = d, f = "%[^\n]";
		if ((retc = fscanf(fp, f, s+1)) != 1)
			break;
		buf[n] = ' ';
		/*
		**	locate and chain new table element and item name
		*/
		if (! (t->next = (KEY *) alloc(1, sizeof(KEY)+strlen(buf))))
			return ihclose(fp);
		(keyanch.hashptr = t = t->next)->next = NULL;
		if ((t->type = c) == HDR_END)
			efind = 1;
		(strcpy(t->name, buf))[n] = 0;
		hashit(t);
	}
	/*
	**	if errors occurred, write message and return NULL
	*/ 
	if (gipserrno == 0)
		if (! efind)
			gipserrno = GIPSNOEOF;
		else if (retc != EOF)
			gipserrno = GIPSBADDATA;
	if (fileno(fp) > 2)
		(void)(fclose(fp));
	return gipserrno ? NULL : ihmerg(fname);
}
	/*
	**	merge a GIPS header into the incore table
	*/
FILE *
ihmerg(fname)

	char	*fname;		/* input GIPS file name ("-" = stdin) */
{
	register FILE	*ff;	/* input file pointer */
	char	buf[BUFLEN];	/* input buffer */
	register char	*s;	/* pointer to buf[] */
	register int	i;	/* scratch */
	register KEY	*u;	/* current incore item pointer */
	register KEY	*t;	/* pointer to history list */
	/*
	**	check callers arguments and open the GIPS file
	*/
	if (ihcheck(fname, fname, "ihmerg"))
		return NULL;
	/*
	**	If 'fname' contains '=', assume that it is string input
	**	If it begins with "|", assume it is a pipe
	*/
	if (fname[0] == '|' && fname[1] != '\0') {
		if (ff = popen(++fname, "r"))
			isapipe[fileno(ff)]++;
	} else if (fname[0] == '-' && fname[1] == '\0') {
		ff = stdin;
		fname = "'stdin'";
	} else if (index(fname, '=')) {
		ff = fopens(fname);
	} else if (! (ff = fopen(fname, "r")) && *fname != '/') {
		(void)(strcat(strcpy(buf, GIPSPROFDIR), fname));
		ff = fopen(buf, "r");
	}
	(void)(strcpy(*gipserrs, fname));
	if (! ff) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	/*
	**	read the GIPS header
	*/
	for (t = keyanch.hashptr; ; ) {
		for (s = buf; (i = getc(ff)) != EOF && i != '='; )
			if (! isspace(i))
				*s++ = i;
		*s = 0;
		if (s == buf || (i == EOF && ! isastring(ff)))
			break;
		/*
		**	buf contains null-terminated variable name
		*/	
		if (! (u = ihseek(buf, buf, "ihmerg")))
			return ihclose(ff);

		switch (u->type) {
		case HDR_END:
			u->flgs |= VAR_INHEADER;
			(void)(fgetc(ff));
			return check_required() ? NULL : ff;
		case HDR_INTEGER:
			i = fscanf(ff, "%d", &u->uval.ival);
			break;
		case HDR_DOUBLE:
			i = fscanf(ff, "%lf", &u->uval.dval);
			break;
			/*
			**	link new history element onto history chain
			**	then process it as any other string item
			*/
		case HDR_HISTORY:
			t->next= (KEY *)alloc(1, sizeof(KEY)+strlen(u->name)+1);
			if (! t->next)
				return ihclose(ff);
			(void)(strcpy(t->next->name, u->name));
			t->next->type = u->type;
			(keyanch.hashptr = t = u = t->next)->next = NULL;
			hashit(t);
			/*
			**	copy the string into fname
			*/
		case HDR_CHAR:
		case HDR_STRING:
		case HDR_REQUIRED:
		case HDR_ENUM:
			i = fscanf(ff, "%1s", &u->uval.cval);
			if (i != 1 || u->type == HDR_CHAR)
				break;
			s = buf;
			if ((i = u->uval.cval) == '"')
				while ((i = getc(ff)) != EOF && i != '"') {
					if (i == '\\' && (i = getc(ff)) != '"')
						*s++ = '\\';
					*s++ = i;
				}
			else
				do *s++ = i;
				while ((i = getc(ff)) != EOF && ! isspace(i));
			/*
			**	copy fname into the table entry
			*/
			*s++ = 0;
			u->uval.sval = alloc(s-buf, sizeof(char));
			if (! u->uval.sval)
				return ihclose(ff);
			(void)(strcpy(u->uval.sval, buf));
			if (i == EOF)
				break;
			s = u->name + strlen(u->name)+1;
			switch (u->type) {
			case HDR_REQUIRED:
				i = strcmp(s, u->uval.sval) ? 0 : 1;
				break;
			case HDR_ENUM:
				for (i = strlen(u->uval.sval); ; ) {
					if (! strncmp(s, u->uval.sval, i) &&
						(s[i] == ' ' || s[i] == 0)) {
						i = 1;
						break;
					}
					s = index(s, ' ');
					if (! s || ! *++s) {
						i = 0;
						break;
					}
				}
				break;
			default:
				i = 1;
				break;
			}
			break;
		default:
			i = 0;
			break;
		}
		u->flgs |= VAR_INHEADER;
		arginit = 0;		/* indicate header altered */
		if (i != 1)
			break;
	}
	if (isastring(ff))
		return check_required() ? NULL : ff;
	gipserrno = (i == EOF) ? GIPSNOEOF : GIPSBADVAL;
	return ihclose(ff);
}
	/*
	**	Determine whether required variables have been found
	*/
static
check_required()
{
	register KEY	*t;	/* pointer to list */

	for (t = &keyanch; t = t->next; )
		if (t->type == HDR_REQUIRED || t->type == HDR_ENUM)
			if ((t->flgs & VAR_AVAILABLE) == 0) {
				(void)strcpy(gipserrs[0], t->name);
				return gipserrno = GIPSNOVAL;
			}
	return gipserrno = 0;
}
	/*
	**	hash t->name and link t to hash table
	*/
static
hashit(t)

	register	KEY	*t;	/* pointer to incore element */
{
	register	i;		/* hashed t->name */

	i = hash(t->name);
	t->hashptr = hashtab[i];
	hashtab[i] = t;
}
	/*
	**	hashing function - returns integer from 0 to HASHNUM-1
	*/
static
hash(c)

	register	char	*c;	/* pointer to item name */
{
	register	i;		/* accumulator */
	register	char	*s;	/* pointer to error message */

	for (i = 0, s = *gipserrs; *s++ = *c; i += *c++ & 0377)
		;
	return i % HASHNUM;
}
	/*
	**	free an incore table and hash array
	*/
static
ihfree()
{
	register KEY	*t;	/* incore element to free */
	register i;		/* hash table element to clear */

	for (i = HASHNUM; --i >= 0; hashtab[i] = NULL)
		;
	for (t=keyanch.hashptr= &keyanch; t = t->next; free((char *) t))
		if (index(string_types, t->type) && t->uval.sval)
			free(t->uval.sval);
	keyanch.next = NULL;
	arginit = 0;
}

ihdel(item)

	register char *item;		/* name of header item */
{
	register KEY *t;		/* incore header item pointer */

	if (! (t = ihseek(item, item, "ihdel")))
		return gipserrno;
	t->flgs &= ~(VAR_INHEADER|VAR_ADDED|VAR_UPDATED);
	return arginit = 0;
}

ihget(item, value, lvalue)

	register char *item;	/* name of header item */
	register char *value;	/* pointer to returned value */
	register lvalue;		/* length of values in bytes */
{
	register KEY	*t;	/* pointer to incore header item */
	register char	*s;	/* pointer to string value */

	if (! (t = ihseek(item, value, "ihget")))
		return gipserrno;

	if ((t->flgs & (VAR_INHEADER|VAR_ADDED|VAR_UPDATED)) == 0)
		return gipserrno = GIPSNOVAL;

	switch (t->type) {
		case HDR_CHAR:
			*value = t->uval.cval;
			break;
		case HDR_HISTORY:
		case HDR_STRING:
		case HDR_REQUIRED:
		case HDR_ENUM:
			if (! (s = t->uval.sval))
				return gipserrno = GIPSNOSTR;
			while (--lvalue > 0 && (*value++ = *s++))
				;
			if (lvalue <= 0) {
				*value = '\0';
				return gipserrno = GIPSTRUNC;
			}
			break;
		case HDR_INTEGER:
			*(int *) value = t->uval.ival;
			break;
		case HDR_DOUBLE:
			*(double *) value = t->uval.dval;
			break;
		default:
			return gipserrno = GIPSBADTYPE;
	}
	return gipserrno = 0;
}

ihput(item, value)

	char		*item;		/* name of incore header item */
	register char	*value;		/* pointer to new item value */
{
	register KEY	*t;		/* pointer to header item */
	register KEY	*u;		/* pointer to previous history entry */
	time_t	clock;			/* current date/time integer */

	if (! (t = ihseek(item, value, "ihput")))
		return gipserrno;

	switch (t->type) {
		case HDR_CHAR:
			t->uval.cval = *value;
			break;
		case HDR_INTEGER:
			t->uval.ival = *(int *) value;
			break;
		case HDR_DOUBLE:
			t->uval.dval = *(double *) value;
			break;
			/*
			**	special processing to add a history entry
			**	to the current end of the incore table
			*/
		case HDR_HISTORY:
			u = t;
			t = keyanch.hashptr;
			t->next= (KEY *)alloc(1, sizeof(KEY)+strlen(t->name)+1);
			if (! t->next)
				return gipserrno;
			(keyanch.hashptr = t = t->next)->next = NULL;
			(void)(strcpy(t->name, u->name));
			t->type = u->type;
			hashit(t);
			t->uval.sval = alloc(strlen(value)+26, sizeof(char));
			if (! t->uval.sval)
				return gipserrno;
			clock = time((time_t *) NULL);
			(void)(strcat(strcpy(t->uval.sval, ctime(&clock)), value));
			(t->uval.sval)[24] = '\t';
			break;
		case HDR_STRING:
		case HDR_ENUM:
			if (t->uval.sval)
				free(t->uval.sval);
			t->uval.sval=alloc(strlen(value)+1, sizeof(char));
			if (! t->uval.sval)
				return gipserrno;
			(void)(strcpy(t->uval.sval, value));
			break;
		default:
			return gipserrno = GIPSBADTYPE;
	}
	/*
	**	indicate that this item has been altered, and that gips() needs
	**	to recompute its translation tables
	*/
	t->flgs |= VAR_UPDATED;
	arginit = 0;
	return gipserrno = 0;
}

ihtype(item, dsc, ldsc)

	register char	*item;	/* pointer to item name */
	register char	*dsc;	/* pointer to output field */
	register	ldsc;	/* maximum length of output field (bytes) */
{
	register KEY	*t;	/* pointer to incore item */
	register char	*s;	/* pointer to string value */

	if (! (t = ihseek(item, dsc, "ihtype")))
		return gipserrno;

	*dsc++ = t->type;

	if (--ldsc < 1) 
		return gipserrno = 0;
	s = t->name + strlen(t->name);
	while (--ldsc > 0 && (*dsc++ = *++s))
		;
	if (ldsc == 0) {
		*dsc = '\0';
		return gipserrno = GIPSTRUNC;
	}
	return gipserrno = 0;
}

FILE *
ihwrit(fname, llength)

	char	*fname;			/* name of output file ("-" = stdout0 */
	int	llength;		/* output line length */
{
	register char	*buf;		/* output buffer */
	register char	*c;		/* output pointer */
	register KEY	*t;		/* pointer to incore element */
	register FILE	*fp;		/* output file stream pointer */
	char		*e	= NULL;		/* pointer to type-E item */
	int		buflen	= -1;		/* length of output buffer */
	int		bufmax;			/* maximum output line length */
	/*
	**	insert routine name for error()
	*/
	if (ihcheck(fname, fname, "ihwrit"))
		return NULL;
	buf = *gipserrs;
	if (llength <= 0 || llength >= BUFLEN)
		bufmax = OUTLEN;
	else
		bufmax = llength;
		
	if (fname[0] == '-' && fname[1] == '\0') {
		fp = stdout;
		(void)(strcpy(buf, "'stdout'"));
	} else if (fname[0] == '|' && fname[1] != '\0') {
		if (fp = popen(strcpy(buf, fname+1), "w"))
			isapipe[fileno(fp)]++;
	} else
		fp = fopen(strcpy(buf, fname), "w");
	if (! fp) {
		gipserrno = GIPSBADOPEN;
		return NULL;
	}
	buf += strlen(strcat(buf, ": "));
	/*
	**	loop over incore item list
	*/
	for (t = &keyanch; t = t->next; ) {
			/*
			**	uppercase type code implies item must be present
			*/
		if ((t->flgs & VAR_AVAILABLE) == 0)
			if (isupper(t->type)) {
				(void)(strcpy(buf, t->name));
				gipserrno = GIPSNOVAL;
				return ihclose(fp);
			} else
				continue;
			/*
			**	copy "item = value" to buf[]
			*/
		if (t->type != HDR_END)
			c = buf + strlen(strcat(strcpy(buf, t->name), " = "));
		switch (t->type) {
		case HDR_CHAR: 
			*c++ = t->uval.cval;
			*c = 0;
			break;
		case HDR_INTEGER: 
			(void)(sprintf(c, "%d", t->uval.ival));
			break;
		case HDR_DOUBLE: 
			(void)(sprintf(c, "%.15g", t->uval.dval));
			break;
		case HDR_STRING: 
		case HDR_REQUIRED: 
		case HDR_ENUM: 
		case HDR_HISTORY:  {
			/*
			**	special care with string values in case they
			**	contain blanks and/or double quotes
			*/
			register char	*d;

			for (d = t->uval.sval; *d && ! isspace(*d); d++)
				;
			if (! *d)
				(void)(strcpy(c, t->uval.sval));
			else if (d == t->uval.sval)
				*c++ = '"', *c = '"';
			else for (*c = '"', d = t->uval.sval; *c; *++c = *d++)
				if (*d == '"')
					*++c = '\\';
				else if (! *d)
					*++c = '"';
			break;
		}
		case HDR_END:
			e = t->name;	/* remember type=E, but don't copy it */
			continue;
		default:
			gipserrno = GIPSBADVAR;
			return ihclose(fp);
		}
		/*
		**	if the buffer has grown past bufmax, add a newline
		*/
		buflen = bufcor(buf, fp, buflen, bufmax);
	}
	/*
	**	write the end-of-header string, and flush the output
	*/
	if (! e) {
		gipserrno = GIPSNOEOF;
		return ihclose(fp);
	}
	(void)sprintf(buf, "%s=\n", e);
	(void)(bufcor(buf, fp, buflen, bufmax));
	(void)(fflush(fp));
	gipserrno = 0;
	return fp;
}

static
bufcor(buf, fp, len, max)

	register char	*buf;
	register FILE	*fp;
	register int	len;
	register int	max;
{
	register int	i;

	i = strlen(buf);

	if (len < 0)
		len = i;
	else if (len + i > max) {
		(void)(putc('\n', fp));
		len = i;
	} else if (len > 0) {
		fputs("   ", fp);
		len += i + 3;
	}
	fputs(buf, fp);
	return len;
}

ihbcl(fp)

	register FILE *fp;	/* GIPS stream pointer */
{
	if (isapipe[fileno(fp)]) {
		isapipe[fileno(fp)] = 0;
		return pclose(fp);
	} else
		return fclose(fp);
}

FILE *
ihread_(fname, prof)

	char	*fname;			/* name of input header file */
	char	*prof;			/* name of header definition file */
{
	return ihread(fname, prof);
}

FILE *
ihmerg_(fname)

	char	*fname;		/* input GIPS file name ("-" = stdin) */
{
	return ihmerg(fname);
}

ihdel_(item)

	char	*item;		/* name of header item */
{
	return ihdel(item);
}

/*ARGSUSED*/
ihget_(item, value, litem, lvalue)

	char	*item;		/* name of header item */
	char	*value;		/* pointer to returned value */
	int	litem;		/* length of name in bytes */
	int	lvalue;		/* max length of value in bytes */
{
	return ihget(item, value, lvalue);
}

ihput_(item, value)

	char	*item;		/* name of incore header item */
	char	*value;		/* pointer to new item value */
{
	return ihput(item, value);
}

/*ARGSUSED*/
ihtype_(item, dsc, litm, ldsc)

	char	*item;		/* pointer to item name */
	char	*dsc;		/* pointer to output field */
	int	litm;		/* maximum length of item name */
	int	ldsc;		/* maximum length of output field (bytes) */
{
	return ihtype(item, dsc, ldsc);
}

FILE *
ihwrit_(fname, lptr)

	char	*fname;			/* name of output file ("-" = stdout0 */
	int	*lptr;			/* pointer to output line length */
{
	return ihwrit(fname, lptr ? *lptr : 0);
}

ihbcl_(fp)
	FILE	**fp;		/* file pointer */
{
	return ihbcl(*fp);
}

iherrm_()
{
	if (gipserrno)
		(void) fprintf(stderr, gipserrs[gipserrno], gipserrs[0]);
	return gipserrno;
}
