#ifndef lint
static char *sccsid = "%Z%$RCSfile: nonconical.c,v $	GIPS $Revision: 2.10 $ (MIT/CSR) $Date: 1998/05/13 00:51:52 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984,1987 Massachusetts Institute of Technology
**
**	Module:		nonconical.c	Creation Date:	06/29/87
**	Version:	$Revision: 2.10 $		Update:		$Date: 1998/05/13 00:51:52 $
**	Components:	aitoff()	ihgrid() function
**			briesemeister()	""
**			hammer()	""
**			lambert3()	""
**			mollweide()	""
**			case1()		forward mapping
**			case2()		inverse mapping
**
**	Function:	Transform an image via non-conical mappings
*/

#define		GRID
#include	<math.h>
#include	"gips.h"

static	case1(), case2();

aitoff(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		v = t->radius*sqrt(2/(1+cos(y)*cos(x/2)));
		xy[0] = 2*v*cos(y)*sin(x/2);
		xy[1] = v*sin(y);
		return 0;
	case 2:
		x /= t->radius;
		y /= t->radius;
		v = 1 - x*x/16 - y*y/4;
		if (v < 0) return 1;
		y *= sqrt(v);
		if (y < -1 || y > 1) return 1;
		xy[1] = asin(y);
		xy[0] = 2*atan2(x*sqrt(v),4*v-2);
		return 0;
	default:
		break;
	}
	return -1;
}

briesemeister(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = sqrt((double)2) * t->radius;
		return 0;
	case 1:
		return case1(x, y, xy,
			2 * sin(acos(cos(y) * cos(x / 2)) / 2), 1.75);
	case 2:
		v = hypot(4 * x / 7, y);
		if (v == 0 || v > t->rlim || v < -t->rlim) return 1;
		return case2(x, y, xy,
			y / v, cos(2 * asin(v / (2 * t->radius))));
	default:
		break;
	}
	return -1;
}

hammer(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = sqrt((double)2) * t->radius;
		return 0;
	case 1:
		return case1(x, y, xy,
			2 * sin(acos(cos(y) * cos(x / 2)) / 2), 2.0);
	case 2:
		v = hypot(x / 2, y);
		if (v == 0 || v > t->rlim || v < -t->rlim) return 1;
		return case2(x, y, xy,
			y / v, cos(2 * asin(v / (2 * t->radius))));
	default:
		break;
	}
	return -1;
}

lambert3(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	v;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		xy[0] = t->radius * x;
		xy[1] = t->radius * sin(y);
		return 0;
	case 2:
		xy[0] = x / t->radius;
		v = y / t->radius;
		if (v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

#define		PREC		(1.e-12)

mollweide(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;
	double	c, ct, v, vv, dv;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		t->rlim = sqrt((double)2) * t->radius;
		return 0;
	case 1:
		c = 0;
		dv = ct = M_PI * sin(y);
		while (fabs(dv) > PREC && fabs(vv = 1 + cos(c)) > PREC)
			c += (dv = ct - c - sin(c)) / vv;
		xy[0] = 2 * t->rlim * x * cos(c / 2) / M_PI;
		xy[1] = t->rlim * sin(c / 2);
		return 0;
	case 2:
		v = y / t->rlim;
		if ( v > 1 || v < -1) return 1;
		v = asin(v);
		xy[0] = M_PI_2 * x / (t->rlim * cos(v));
		v += v;
		v = (v + sin(v)) / M_PI;
		if ( v > 1 || v < -1) return 1;
		xy[1] = asin(v);
		return 0;
	default:
		break;
	}
	return -1;
}

cylindrical(sw, x, y, xy)
	double	x, y, xy[2];
{
	register G_PARM	*t = g_proj;

	switch (sw) {
	case 0:
		t->radius = xy[0];
		return 0;
	case 1:
		xy[0] = t->radius * x;
		xy[1] = t->radius * y;
		return 0;
	case 2:
		xy[0] = x / t->radius;
		xy[1] = y / t->radius;
		return 0;
	default:
		break;
	}
	return -1;
}

static
case1(x, y, xy, vv, vf)
	double	x, y, xy[2], vv, vf;
{
	register G_PARM	*t = g_proj;
	double	v;

	v = sin(acos(cos(x) * cos(y)));
	if (v == 0) { xy[0] = xy[1] = 0; return 0; }
	v = sin(y) / v;
	if (v > 1 || v < -1) return 1;
	vv *= t->radius;
	xy[0] = vf * vv * sin((x > 0) ? acos(v) : -acos(v));
	xy[1] = v * vv;
	return 0;
}

static
case2(x, y, xy, s, u)
	double	x, y, xy[2], s, u;
{
	double	v, a, b, c;

	a = 4 * s * s * u * u;
	b = 1 - s * s;
	c = b - 2 * u * u;
	if (a == 0) {
		if (b == 0) return 1;
		v = -c / b;
	} else {
		v = b * b - 2 * a * c;
		if (v < 0) return 1;
		v = (sqrt(v) - b) / a;
	}
	if (v > 1 || v <= -1) return 1;
	xy[0] = (x > 0) ? acos(v) : -acos(v);
	v = cos(xy[0] / 2);
	if (v == 0) return 1;
	v = u / v;
	if (v > 1 || v < -1) return 1;
	xy[1] = (y > 0) ? acos(v) : -acos(v);
	return 0;
}
