#ifndef lint
static char *sccsid = "%Z%$RCSfile: rdwr.c,v $	GIPS $Revision: 2.2 $ (MIT/CSR) $Date: 1988/12/13 15:51:17 $";
#endif
/*		   General Image Processing System
**			   Peter G. Ford
**		      Center for Space Research
**	Copyright 1984 Massachusetts Institute of Technology
**
**	Module:		rdwr.c		Creation Date:	12/30/83
**	Version:	$Revision: 2.2 $		Update:		$Date: 1988/12/13 15:51:17 $
**	Components:	fread()		Replacement to C library function
**			fwrite()	Replacement to C library function
**
**	Function:	Speedup of C library routines for handling large
**			data arrays
*/

#include	<stdio.h>

#define		min(x,y)	((x)<(y)?(x):(y))
#define		_IOFLGS		(_IOREAD|_IOWRT|_IORW|_IOSTRG|_IONBF)

fread(ptr, size, count, iop)

	register char	*ptr;		/* pointer to input buffer */
	int	 	size;		/* input element size */
	register int	count;		/* number of elements to be read */
	register FILE	*iop;		/* pointer to buffered i/o table */
{
	register char	*base	= ptr;	/* saved input buffer pointer */
	register int	len;		/* length moved per operation */
	/*
	**	replace 'count' with the total number of bytes to read
	*/
	for (count *= size; count > 0; *ptr++ = len, count--) {
		/*
		**	if there is anything special about input file,
		**	don't to anything fancy - pass it to getc()
		*/
		if ((iop->_flag & _IOFLGS) == _IOREAD) {
			/*
			**	if the stdio input buffer is non-empty, move
			**	out the data via move() - a speedup routine
			*/
			if (iop->_cnt > 0) {
				len = min(count, iop->_cnt);
				move(iop->_ptr, ptr, len);
				iop->_ptr += len;
				ptr += len;
				iop->_cnt -= len;
				if ((count -= len) <= 0)
					break;
			}
			/*
			**	if what remains to be read is larger than stdio
			**	block, read it directly to the users buffer
			*/
			if (count >= iop->_bufsiz) {
				iop->_cnt = 0;
				if (iop == stdin && (stdout->_flag & _IOLBF))
					(void)(fflush(stdout));
				while (count > 0 && (len =
				    read(fileno(iop), ptr, count)) > 0) {
					ptr += len;
					count -= len;
				}
				if (count > 0)
					iop->_flag |= (len==0) ? _IOEOF:_IOERR;
				break;
			}
			/*
			**	come here to bypass speedup reading
			*/
		}
		if ((len = getc(iop)) == EOF)
			break;
	}
	/*
	**	use 'ptr-base' to determine how many elements have been read
	*/
	return((size > 0) ? (ptr-base)/size : 0);
}

fwrite(ptr, size, count, iop)

	register char	*ptr;		/* pointer to output buffer */
	int	 	size;		/* output element size */
	register int	count;		/* number of elements to be written */
	register FILE	*iop;		/* pointer to buffered i/o table */
{
	register char	*base	= ptr;	/* saved output pointer */
	register int	len;		/* length of last item moved */
	/*
	**	reset count to be the number of bytes to move.
	*/
	count *= size;
	for ( ; count > 0 && ! ferror(iop); (void)(putc(*ptr++, iop)), count--) {
		/*
		**	bypass the speedup code if there is
		**	anything special about iop
		*/
		if ((iop->_flag & _IOFLGS) != _IOWRT)
			continue;
		/*
		**	if there is any data already in the stdio output buffer
		**	append enough to fill it (if poss) using speedup move().
		*/
		if (iop->_cnt > 0) {
			len = min(count, iop->_cnt);
			move(ptr, iop->_ptr, len);
			iop->_ptr += len;
			ptr += len;
			iop->_cnt -= len;
			if ((count -= len) <= 0)
				break;
		}
		/*
		**	if the stdio buffer is full, write it out
		*/
		if (iop->_cnt <= 0 && iop->_base != NULL &&
		    iop->_ptr > iop->_base)
			(void)(fflush(iop));
		/*
		**	if what remains to be written is larger than a stdio
		**	buffer, write it in a single unbuffered operation 
		*/
		if (count >= iop->_bufsiz) {
			iop->_cnt = 0;
			while (count > 0 && (len =
			    write(fileno(iop), ptr, count)) > 0) {
				ptr += len;
				count -= len;
			}
			if (count > 0)
				iop->_flag |= (len == 0) ? _IOEOF : _IOERR;
			break;
		}
	}
	/*
	**	use 'ptr-base' to determine number of elements written
	*/
	return((size > 0) ? (ptr-base)/size : 0);
}
